////////////////////////////////////////////////////////////////////////////////
/*! @file RigakuScan.cpp   XRD Rigaku Scanner.
- Part of Rigaku_APRI - Research Assistant Plugin package.
- Uses  RAGUI - Research Assistant Graphical User Interface.
- Uses  QtGui v.4.6  -  http://qt.nokia.com/
- Uses  RANet - Research Assistant Net Library (based on ANSI C++).
- Copyright(C) 2010, Viktor E. Bursian, St.Petersburg, Russia.
                     Viktor.Bursian@mail.ioffe.ru
- Copyright(C) 2010, Dmitry K. Nelson, St.Petersburg, Russia.
                     D.Nelson@mail.ioffe.ru
*///////////////////////////////////////////////////////////////////////////////
#include "RigakuScan.h"
#include "attr.h"
#include "Symbols.h"
#include <QMessageBox>
#include <QDebug>
namespace RA {
//------------------------------------------------------------------------------

csString                      sRigakuScan::SpeedUnits(sString(Symbol::Degree)
                                                     +"/min");
DEFINE_CLASS_TAG(sRigakuScan)

//-------------------------------------------------------------- sRigakuScan ---

void  sRigakuScan::Preconfigure (sNodePtr  DeviceCfg)
{
  sPtr<sFolderNode>              DeviceSettings;
  sPtr<sFolderNode>              DeviceLinks;

  DeviceCfg |= "Links";
  DeviceLinks = DeviceCfg >> "Links";
  if( DeviceLinks.IsCorrect() ){
    DeviceLinks |= "Gage";
    DeviceLinks >> "Gage" |= attr("Class","sAPRI_ML327_Counter");
  };
  DeviceCfg |= "Settings";
  DeviceSettings = DeviceCfg >> "Settings";
  if( DeviceSettings.IsCorrect() ){
    DeviceSettings |= attr("StartQuantum",sPhysValue(1.0,Symbol::Degree));

    DeviceSettings |= attr("NumOFSpeeds",3);

    sPtr<sFolderNode>              SpeedsConf;
    DeviceSettings |= "Speeds";
    SpeedsConf=DeviceSettings >> "Speeds");
    if( SpeedsConf.IsCorrect() ){
      for( int i = 0 ; i < 12 ; ++i ){
        SpeedsConf |= attr(i+1,0.5*(i+1));
      };
    };

    //изменение умолчательных шкал для sProtoScanner
    sChartScales      ChartScales(sChartScale("Angle"
                                             ,sPhysRange(0,180,Symbol::Degree))
                                 ,sChartScale(""
                                             ,sPhysRange(0,900,"kphoton/s"))
                                 );
    ChartScales.StoreSettings(DeviceSettings);

    //изменение умолчательных значений атрибутов для sPeriodicGageRecorder
    //! @todo{Rigaku} review!!!
    DeviceSettings |= attr("ValidScanRange",sPhysRange(0,360,Symbol::Degree));
    DeviceSettings |= attr("ScanRange",sPhysRange(0,360,Symbol::Degree,0.5));
    DeviceSettings |= attr("ValidScanSteps",sPhysRange(0.01,10.0,Symbol::Degree));
    DeviceSettings |= attr("ScanStep",sPhysValue(0.25,0.005,Symbol::Degree));

//    DeviceSettings |= attr("ValidSpeeds",sPhysRange(0.5,2.0,SpeedUnits));
//    DeviceSettings |= attr("Speed",sPhysValue(0.3,SpeedUnits));
  };
  sPeriodicGaugeRecorder::Preconfigure(DeviceCfg);
};


sRigakuScan::~sRigakuScan()
{
  if( DeviceSettings.IsCorrect() ){
    DeviceSettings << attr("SpeedIndex",FldSpeed->currentIndex());
  };
  PhotonCounter.SetToNULL();
};


sRigakuScan::sRigakuScan (sString   DevicetId
                         ,sNodePtr  DeviceCfg
                         ,sNodePtr  WidgetCfg)
    :sPeriodicGaugeRecorder(DevicetId,DeviceCfg,WidgetCfg)
    ,PhotonCounter(this)
    ,StartQuantum(1.0,Symbol::Degree)
    ,NumOfSpeeds(3)
{
  Multiplier.X = sPhysValue(1,Symbol::Degree);
  Multiplier.Y = sPhysValue(1,"photon/s");

  //! @todo{Rigaku} review!!!

  LblSpeed = new QLabel(tr("Speed"),ParamArea);
  LblSpeed->setObjectName("LblSpeed");
  ParamLayout->addWidget(LblSpeed,Row,0,1,1);
  FldSpeed = new QComboBox(ParamArea);
  FldSpeed->setObjectName("FldSpeed");
  ParamLayout->addWidget(FldSpeed,Row,1,1,1);
  LblSpeedUnit = new QLabel(ToQString(Symbol::Degree)+"/min",ParamArea);
  LblSpeedUnit->setObjectName("LblSpeedUnit");
  ParamLayout->addWidget(LblSpeedUnit,Row++,2,1,1);
  for( int i = 0 ; i < MaxNumOfSpeeds ; ++i )
    Speeds[i]=0.5*(i+1);

  if( DeviceSettings.IsCorrect() ){
    sPtr<sIntegerNode>          I;
    sPtr<sPhysValueNode>        PV;

    PV=DeviceSettings >> "StartQuantum";
    if( PV.IsCorrect() )  StartQuantum = (sPhysValue)(*PV);

    I=DeviceSettings >> "NumOFSpeeds";
    if( I.IsCorrect() )  NumOfSpeeds = (int)(*I);

    sPtr<sFolderNode>              SpeedsConf;
    SpeedsConf=DeviceSettings >> "Speeds";
    if( SpeedsConf.IsCorrect() ){
      for( int i = 0 ; i < NumOfSpeeds ; ++i ){
        PV=SpeedsConf >> i+1;
        if( PV.IsCorrect() )
          Speeds[i] = ((sPhysValue)(*PV)).ArbValue();
      };
    };

    for (int Index = 0; Index < NumOfSpeeds; Index++ ) {
        FldSpeed->insertItem(Index,"ASpeed");
        FldSpeed->setItemText(Index
                             ,QString().setNum((double)Speeds[Index],'f',2)
                             );
    }
  };
  PhotonCounter.SetTo(Gage->DeviceIdentifier());
  if( ! PhotonCounter.IsCorrect() ){
    //! @todo{Rigaku} можно сказать "Нефиг конфиг ломать, если ты чайник!"
    SetDriverState(Invalid);
  };
};


void  sRigakuScan::RestoreState ()
{
  sPeriodicGaugeRecorder::RestoreState();
  if( ! WidgetConfig.IsNULL() ){
    sPtr<sIntegerNode>          I;

    WidgetConfig |= attr("SpeedIndex",FldSpeed->currentIndex());
    I=WidgetConfig >> "SpeedIndex";
    if( I.IsCorrect() )
      FldSpeed->setCurrentIndex((int)(*I));
  };
};


void  sRigakuScan::StoreState ()
{
  sPeriodicGaugeRecorder::StoreState();
  if( ! WidgetConfig.IsNULL() ){
    WidgetConfig << attr("SpeedIndex",FldSpeed->currentIndex());
  };
};


/*! Просит включить мотор и инициирует ожидание сигнала
    sAPRI_ML327_Counter::GotMarkSignal() о прохождении целого значения угла.
*/
void  sRigakuScan::StartScan ()
{
  //! @todo{Rigaku} review!!!
  if (PhotonCounter->CountIsStarted())    PhotonCounter->Stop();

  Speed = sPhysValue(Speeds[FldSpeed->currentIndex()],0.001,SpeedUnits);

  //правим поле ScanRangeField (его часть From):
  try{
    sPhysValue                  ArbFrom(_Unitsless_); //гарантирует,
                                //что From безразмерная (если в конфиге лажа, то
                                //на следующем присваивании случится xRAlgebra)
    ArbFrom = ScanRangeField->Value().From() / StartQuantum;
    ScanRangeField->SetValue(sPhysRange( Round(ArbFrom.ArbValue()) * StartQuantum
                                       , ScanRangeField->Value().To() ));
  }catch(rxRAlgebra){
    //! @todo{Rigaku} можно сказать "Нефиг конфиг ломать!"
    SetDriverState(Invalid);
    return;
  };

  QMessageBox::information(this, tr("Start scanning")
      ,tr("Set the starting angle %1 and switch the motor on.")
      .arg(ToQString( (ScanRangeField->Value().From() - StartQuantum/2)
                                                     .Text(RA::Plain,Laconic) ))
      ,QMessageBox::Ok);

  disconnect(PhotonCounter.operator ->() ,SIGNAL(GotMarkSignal()) ,0 ,0);
                    //! @todo{Rigaku} Это перебдёж (подумать, нужно ли и от чего именно отсоединять)
  connect(PhotonCounter.operator ->() ,SIGNAL(GotMarkSignal())
         ,this                        ,SLOT(StartRecording())  );
  PhotonCounter->StartWaitingForMarkSignal();
  qDebug() << "Scan started";
};


/*! Запускает запись угловой зависимости.

Вызывается по сигналу sAPRI_ML327_Counter::GotMarkSignal() о прохождении
целого значения угла.

Собственно вся дальнейшая работа возложена на sPeriodicGageRecorder.
*/
void  sRigakuScan::StartRecording ()
{
  disconnect(PhotonCounter.operator ->() ,SIGNAL(GotMarkSignal()) ,0 ,0);
                    //! @todo{Rigaku} подумать, от чего именно нужно отсоединять
  sPeriodicGaugeRecorder::StartScan();
  Dependence << attr("speed",Speed);
};


/*! Останавливает запись угловой зависимости.

Может вызываться как по достижении конца диапазона сканирования, так и по
инициативе юзера или ядра.
*/
void  sRigakuScan::FinishScan ()
{
  sPeriodicGaugeRecorder::FinishScan();
  QMessageBox::information(this, tr("Scanning is finished")
                                ,tr("Switch off the motor.")
                                ,QMessageBox::Ok);
};

//------------------------------------------------------------------------------
}; //namespace RA
