////////////////////////////////////////////////////////////////////////////////
/*! @file ColorsUI.cpp Dialogs and menu generators for colors and color sets.
- Part of RAGUI - Research Assistant Graphical User Interface.
- Uses  RANet - Research Assistant Net Library.
- Copyright(C) 2015, Viktor E. Bursian, St.Petersburg, Russia.
                     Viktor_dot_Bursian_at_mail_dot_ioffe_dot_ru
*///////////////////////////////////////////////////////////////////////////////
#include "ColorsUI.h"
#include "Navigator.h"
#include "ColorNodes.h"
#include "StringsAndQStrings.h"
#include "Environment.h"
#include "Globals.h"
#include "AttrDlg.h"
#include "ColorDialog.h"
#include "NetExceptions.h"
#include <QMessageBox>
#include <QToolButton>
#include <QDialogButtonBox>
namespace RA {
//------------------------------------------------------------------------------
//---------------------------------------------------------------- sColorsUI ---

sColorsUI::sColorsUI ()
{
  sNavigatorViewport::MenuGenerators.push_front(&MenuGenerator);
}


void  sColorsUI::MenuGenerator (psNetViewMenu                context_menu
                               ,sNet2TreeEvolution::pcsItem  tree_item)
{
  if( sColorNode::Type()->CanCast(tree_item->AttrValue().operator ->()) ){
    context_menu->Insert( new sNetViewAction
        (context_menu->parentWidget()
        ,&EditColor  ,/*checkable*/false ,/*checked*/false
        ,sNetViewMenu::tr("Edit color")
        ,""
        ));
  }
  if( sColorSetNode::Type()->CanCast(tree_item->AttrValue().operator ->()) ){
    context_menu->Insert( new sNetViewAction
        (context_menu->parentWidget()
        ,&EditColorSet  ,/*checkable*/false ,/*checked*/false
        ,sNetViewMenu::tr("Edit the color set")
        ,""
        ));
  }
  if( sEnvironment::User.AccessLevel >= sUser::cAdmin ){
    if( sFolderNode::Type()->CanCast(tree_item->AttrValue().operator ->()) ){
      context_menu->AddNew->addAction( new sNetViewAction
          (context_menu->parentWidget()
          ,&AddNewColor  ,/*checkable*/false ,/*checked*/false
          ,sNetViewMenu::tr("color")
          ,sNetViewMenu::tr("Add new attribute with a color value")
          ));
      context_menu->AddNew->addAction( new sNetViewAction
          (context_menu->parentWidget()
          ,&AddNewColorSet  ,/*checkable*/false ,/*checked*/false
          ,sNetViewMenu::tr("color set")
          ,sNetViewMenu::tr("Add new attribute with a color set value")
          ));
    }
  }
}


void  sColorsUI::EditColor (sNet2TreeEvolution::pcsItem  tree_item
                           ,sPoint                       /*pos*/)
{
  sPtr<sColorNode>            CN(tree_item->AttrValue());
  ASSERT(CN.IsCorrect());
  sColor                      C(CN->Value());
  sColorDialog                D(C);
  if( D.exec() ){
    CN->Assign(D.SelectedColor());
  }
}


void  sColorsUI::EditColorSet (sNet2TreeEvolution::pcsItem  tree_item
                              ,sPoint                       /*pos*/)
{
  sPtr<sColorSetNode>         CSN(tree_item->AttrValue());
  ASSERT(CSN.IsCorrect());
  sColorSetEdit               D(CSN->Value()
                               ,( RAGUI::NightRegime ? sColor::cBlack
                                                     : sColor::cWhite )
                               );
  if( D.exec() ){
    CSN->Assign( D.Colors() );
  }
}


void  sColorsUI::AddNewColor (sNet2TreeEvolution::pcsItem  tree_item
                             ,sPoint                       /*pos*/)
{
  sAttrDlg                    D;
  D.setWindowTitle( tr("Add color attribute") );
  sPtr<sColorNode>            NewNode;
  while( D.exec() ){
    NewNode = sNodePtr(new sColorNode);
    try{
      tree_item->AttrValue() << sAttribute(D.AttrName(),NewNode);
      break;
    }catch( rxNetInvalidOp ){
      QMessageBox::critical(NULL ,D.windowTitle()
          ,tr("Uups!  Iz-z-zya-a-a!","Add attribute") );
    }
  }
}


void  sColorsUI::AddNewColorSet (sNet2TreeEvolution::pcsItem  tree_item
                                ,sPoint                       /*pos*/)
{
  sAttrDlg                    D;
  D.setWindowTitle( tr("Add color set attribute") );
  sPtr<sColorSetNode>         NewNode;
  while( D.exec() ){
    NewNode = sNodePtr(new sColorSetNode);
    try{
      tree_item->AttrValue() << sAttribute(D.AttrName(),NewNode);
      break;
    }catch( rxNetInvalidOp ){
      QMessageBox::critical(NULL ,D.windowTitle()
          ,tr("Uups!  Iz-z-zya-a-a!","Add attribute") );
    }
  }
}

//------------------------------------------------------------ sColorSetView ---

sColorSetView::sColorSetView (sColorSet  color_set
                             ,sColor     background
                             ,bool       editable
                             ,QWidget *  parent)
    :QFrame(parent)
    ,ColorSet(color_set)
    ,Background(background)
    ,Editable(editable)
    ,HasSelection( ! ColorSet.Empty() )
{
  ColorSelectMapper = new QSignalMapper(this);
  NameEditedMapper = new QSignalMapper(this);
  QGridLayout *               Layout(new QGridLayout(this));
  setLayout(Layout);
  layout()->setSpacing(0);
  layout()->setMargin(0);
  for (int i = 0; (i < ColorSet.Size()) && (i < cMaxSize) ; ++i) {
    Button[i] = new QToolButton(this);
    Button[i]->setText(QString("██"));
    connect(Button[i], SIGNAL(clicked()), ColorSelectMapper, SLOT(map()));
    ColorSelectMapper->setMapping(Button[i],i);
    Layout->addWidget(Button[i],i,0);
    Name[i] = new QLineEdit(this);
    Name[i]->setText(ToQString(ColorSet[i].Name));
    connect(Name[i], SIGNAL(editingFinished()) , NameEditedMapper, SLOT(map()));
    NameEditedMapper->setMapping(Name[i],i);
    Layout->addWidget(Name[i],i,1);
  }
  connect(ColorSelectMapper ,SIGNAL(mapped(int)) ,this ,SIGNAL(Choosed(int)));
  connect(NameEditedMapper ,SIGNAL(mapped(int)) ,this ,SLOT(ChangeName(int)));
  connect(this ,SIGNAL(Choosed(int)) ,this ,SLOT(Select(int)));
  Restyle();
}


int  sColorSetView::SelectedIndex () const
{
  return ( HasSelection ? ColorSet.CurrentIndex() : -1 );
}


void  sColorSetView::Select (sColor  color)
{
  int                         ColorIndex = -1;
  for( int i = 0 ; i < ColorSet.Size() ; i++ ){
    if( color == ColorSet[i].Color ){
      ColorIndex = i;
      break;
    }
  }
  Select(ColorIndex);
}


void  sColorSetView::Select (int  index)
{
  HasSelection = (0 <= index) && (index < ColorSet.Size());
  if( HasSelection ){
    ColorSet.SetCurrentIndex(index);
  }
  Restyle();
}


void  sColorSetView::Restyle ()
{
  sColor                      Alt( Background==sColor::cWhite ? sColor::cBlack
                                                              : sColor::cWhite);
  QString                     CSS("border-style:solid");
  CSS += "; background-color:";
  CSS += Background.HexARGB();
  setStyleSheet(CSS+"; border-width:0px");
  for( int i = 0 ; (i < ColorSet.Size()) && (i < cMaxSize) ; i++ ){
    Button[i]->setStyleSheet(CSS
                            +"; color:"
                            +ToQString(ColorSet[i].Color.HexARGB())
                                  +"; border-width:1px; border-color:"
                            +( i==SelectedIndex()
                                   ? Alt.HexARGB()
                                   : Background.HexARGB()
                             )
                            );
    Name[i]->setReadOnly( ! ( (i==SelectedIndex()) && Editable ) );
    Name[i]->  setStyleSheet(CSS
                            +"; color:"
                            +ToQString(ColorSet[i].Color.HexARGB())
                            );
    Name[i]->hide();
    Name[i]->show();
  }
}


void  sColorSetView::ChangeName (int index)
{
  ColorSet.Change(index
                 ,ColorSet[index].Color
                 ,FromQString(Name[index]->text()));
}

//---------------------------------------------------------- sColorSetSelect ---

sColorSetSelect::sColorSetSelect (sColorSet  color_set
                                 ,sColor     background
                                 ,QWidget *  parent )
    :QFrame(parent)
    ,Background(background)
{
  StandardColors = new sColorSetView(color_set,Background,false,this);

  CustomColorButton = new QToolButton(this);
  CustomColorButton->setText("██ custom");

  setLayout(new QVBoxLayout(this));
  layout()->setSpacing(2);
  layout()->setMargin(0);
  layout()->setAlignment(Qt::AlignTop);
  layout()->addWidget(StandardColors);
  layout()->addWidget(CustomColorButton);

  connect(StandardColors ,SIGNAL(Choosed(int))
         ,this           ,SLOT(OnStandardColorChoosed(int)));
  connect(CustomColorButton ,SIGNAL(clicked())
         ,this              ,SLOT(EditAndSetCustomColor()));
}


void  sColorSetSelect::SetColor (rcsColor color)
{
  TheColor = color;
  StandardColors->Select(TheColor);
  RestyleCustomColorButton();
  emit ColorChanged();
}


void  sColorSetSelect::OnStandardColorChoosed (int color_index)
{
  TheColor = StandardColors->ColorSet[color_index].Color;
  RestyleCustomColorButton();
  emit ColorChanged();
}


void  sColorSetSelect::EditAndSetCustomColor ()
{
  sColorDialog                D(TheColor);
  if( D.exec() ){
    SetColor( D.SelectedColor() );
  }
}


void  sColorSetSelect::RestyleCustomColorButton ()
{
  sColor                      Alt( Background==sColor::cWhite ? sColor::cBlack
                                                              : sColor::cWhite);
  QString                     CSS("border-style:solid");
  CSS += "; background-color:";
  CSS += Background.HexARGB();
  setStyleSheet(CSS+"; border-width:0px");
  setStyleSheet(CSS);
  CustomColorButton->setStyleSheet(CSS
                                  +"; color:"
                                  +ToQString(TheColor.HexARGB())
                                  +"; border-width:1px; border-color:"
                                  +( ! StandardColors->Selected()
                                   ? Alt.HexARGB()
                                   : Background.HexARGB()
                                   )
                                  );
}

//------------------------------------------------------------ sColorSetEdit ---

sColorSetEdit::sColorSetEdit (sColorSet  color_set
                             ,sColor     background
                             ,QWidget *  parent )
    :QDialog(parent)
    ,ColorSetView(NULL)
    ,Background(background)
{
  MoveUpButton = new QPushButton(tr("Move Up"),this);
  EditButton = new QPushButton(tr("Edit"),this);
  RemoveButton = new QPushButton(tr("Remove"),this);
  MoveDownButton = new QPushButton(tr("Move Down"),this);
  AddButton = new QPushButton(tr("Add"),this);
  AddBasicColorsButton = new QPushButton(tr("Add basic colors"),this);

  QDialogButtonBox *  ButtonBox = new QDialogButtonBox(this);
  ButtonBox->setOrientation(Qt::Horizontal);
  ButtonBox->setStandardButtons(QDialogButtonBox::Cancel|QDialogButtonBox::Ok);

  Layout = new QGridLayout;
  setLayout(Layout);
//  layout()->setSpacing(2);
//  layout()->setMargin(0);
  Layout->setAlignment(Qt::AlignTop);

  Layout->addWidget(MoveUpButton         ,0,1);
  Layout->addWidget(EditButton           ,1,1);
  Layout->addWidget(RemoveButton         ,2,1);
  Layout->addWidget(MoveDownButton       ,3,1);
  Layout->addWidget(AddButton            ,4,1);
  Layout->addWidget(AddBasicColorsButton ,5,1);
  Layout->addWidget(new QLabel(" ",this),10,0);
  Layout->addWidget(ButtonBox           ,11,0,1,2);

  connect(MoveUpButton ,SIGNAL(clicked()) ,this ,SLOT(MoveUp()));
  connect(EditButton ,SIGNAL(clicked()) ,this ,SLOT(Edit()));
  connect(RemoveButton ,SIGNAL(clicked()) ,this ,SLOT(Remove()));
  connect(MoveDownButton ,SIGNAL(clicked()) ,this ,SLOT(MoveDown()));
  connect(AddButton ,SIGNAL(clicked()) ,this ,SLOT(Add()));
  connect(AddBasicColorsButton ,SIGNAL(clicked())
         ,this ,SLOT(AddBasicColors()));
  connect(ButtonBox, SIGNAL(accepted()), this, SLOT(accept()));
  connect(ButtonBox, SIGNAL(rejected()), this, SLOT(reject()));

  SetColors(color_set);
}


void  sColorSetEdit::SetColors (rcsColorSet  colors)
{
  psColorSetView              Old(ColorSetView);
  ColorSetView = new sColorSetView(colors,Background,true,this);
  if( Old ){
    Layout->replaceWidget(Old,ColorSetView);
    Old->deleteLater();
  }else{
    Layout->addWidget(ColorSetView,0,0,9,1);
  }
  connect(ColorSetView ,SIGNAL(Choosed(int)) ,this ,SLOT(Refresh()));
  Refresh();
}


void  sColorSetEdit::Refresh ()
{
  MoveUpButton->setEnabled(ColorSetView->SelectedIndex() > 0);
  MoveDownButton->setEnabled(ColorSetView->SelectedIndex() < Colors().Size()-1);
  EditButton->setEnabled(ColorSetView->Selected());
  RemoveButton->setEnabled(ColorSetView->Selected());
  AddButton->setEnabled(Colors().Size() < ColorSetView->cMaxSize);
  AddBasicColorsButton->setEnabled(Colors().Size() + sColorSet::Basic().Size()
                                   <= ColorSetView->cMaxSize);
}


void  sColorSetEdit::MoveUp ()
{
  int                         Moving = ColorSetView->SelectedIndex();
  sColorSet                   New;
  for( int i = 0 ; i < ColorSetView->ColorSet.Size() ; i++ ){
    if( i == Moving-1 ){
      New.Add( ColorSetView->ColorSet[Moving].Color
             , ColorSetView->ColorSet[Moving].Name
             );
    }else if( i == Moving ){
      New.Add( ColorSetView->ColorSet[i-1].Color
             , ColorSetView->ColorSet[i-1].Name
             );
    }else{
      New.Add( ColorSetView->ColorSet[i].Color
             , ColorSetView->ColorSet[i].Name
             );
    }
  }
  SetColors(New);
  ColorSetView->Select(Moving-1);
  Refresh();
}


void  sColorSetEdit::MoveDown ()
{
  int                         Moving = ColorSetView->SelectedIndex();
  sColorSet                   New;
  for( int i = 0 ; i < ColorSetView->ColorSet.Size() ; i++ ){
    if( i == Moving ){
      New.Add( ColorSetView->ColorSet[i+1].Color
             , ColorSetView->ColorSet[i+1].Name
             );
    }else if( i == Moving+1 ){
      New.Add( ColorSetView->ColorSet[Moving].Color
             , ColorSetView->ColorSet[Moving].Name
             );
    }else{
      New.Add( ColorSetView->ColorSet[i].Color
             , ColorSetView->ColorSet[i].Name
             );
    }
  }
  SetColors(New);
  ColorSetView->Select(Moving+1);
  Refresh();
}


void  sColorSetEdit::Remove ()
{
  int                         Removing = ColorSetView->SelectedIndex();
  sColorSet                   New;
  for( int i = 0 ; i < ColorSetView->ColorSet.Size() ; i++ ){
    if( i != Removing ){
      New.Add( ColorSetView->ColorSet[i].Color
             , ColorSetView->ColorSet[i].Name
             );
    }
  }
  SetColors(New);
  ColorSetView->Select( ( Removing < Colors().Size() ? Removing
                                                     : Colors().Size()-1 ) );
  Refresh();
}


void  sColorSetEdit::Edit ()
{
  int                         Editing = ColorSetView->SelectedIndex();
  sColorSet::sItem            CSI(ColorSetView->ColorSet[Editing]);
  sColorDialog                D(CSI.Color);
  if( D.exec() ){
    ColorSetView->ColorSet.Change(Editing
                                 ,D.SelectedColor()
                                 ,CSI.Name);
    ColorSetView->Select(Editing);
  }
}


void  sColorSetEdit::Add ()
{
  sColorSet                   New(ColorSetView->ColorSet);
  New.Add( ( Background==sColor::cBlack ? sColor::cWhite : sColor::cBlack )
         , "custom" );
  SetColors(New);
  ColorSetView->Select(ColorSetView->ColorSet.Size()-1);
  Refresh();
  Edit();
}


void  sColorSetEdit::AddBasicColors ()
{
  sColorSet                   B(sColorSet::Basic());
  sColorSet                   New(ColorSetView->ColorSet);
  for( int i = 0 ; i < B.Size() ; i++ ){
    New.Add(B[i].Color,B[i].Name);
  }
  SetColors(New);
  ColorSetView->Select(ColorSetView->ColorSet.Size()-1);
  Refresh();
}


//------------------------------------------------------------------------------
} //namespace RA
