////////////////////////////////////////////////////////////////////////////////
/*! @file ConfigureDeviceDialog.cpp   Class sConfigureDeviceDialog.
- Part of RAGUI - Research Assistant Graphical User Interface Library.
- Uses  Qt v.5.0.2  - http://qt-project.org/
- Uses  RANet - Research Assistant Net Library.
- Copyright(C) 2010-2014, Viktor E. Bursian, St.Petersburg, Russia.
                          Viktor.Bursian@mail.ioffe.ru
*///////////////////////////////////////////////////////////////////////////////
#include "ConfigureDeviceDialog.h"
#include "Device.h"
#include "BasicNodes.h"
#include "attr.h"
#include "StringsAndQStrings.h"
#include <QFormLayout>
#include <QLabel>
#include <QMessageBox>
namespace RA {
//------------------------------------------------------------------------------
//--------------------------------------------------- sConfigureDeviceDialog ---

sConfigureDeviceDialog::~sConfigureDeviceDialog ()
{
  DeviceCfg.SetToNULL();
  DeviceLinks.SetToNULL();
}


sConfigureDeviceDialog::sConfigureDeviceDialog (sString    device_id
                                               ,QWidget *  parent)
  :QDialog(parent)
  ,DeviceId(device_id)
{
  ASSERT( sDevice::DeviceList.IsCorrect() );
  DeviceCfg = sDevice::DeviceList >> DeviceId;
  ASSERT( DeviceCfg.IsCorrect() );

  setWindowTitle(tr("Device confuguration"));
  Layout = new QVBoxLayout(this);

  BtnUninstall = new QPushButton(tr("Uninstall"),this);
  Layout->addWidget(BtnUninstall,0,Qt::AlignRight);
  connect(BtnUninstall,SIGNAL(clicked()),this,SLOT(Uninstall()));

  sPtr<sBooleanNode>          DisabledNP ( DeviceCfg >> "Disabled" );
  if( ! DisabledNP.IsCorrect() ){
//    DeviceCfg << bool_attr("Disabled",false);
    DeviceCfg << attr("Disabled",false);
    Disabled = false;
  }else{
    Disabled = (bool)(*DisabledNP);
  }

  BoxDisabled = new QCheckBox(tr("Disabled"),this);
  Layout->addWidget(BoxDisabled,0,Qt::AlignRight);
  BoxDisabled->setChecked(Disabled);

  QFormLayout *               FormLayout = new QFormLayout;
  Layout->addLayout(FormLayout);

  FormLayout->setWidget(0,QFormLayout::LabelRole
                       ,new QLabel(tr("Device identifier:"),this));
  FldDeviceId = new QLineEdit(ToQString(DeviceId),this);
  connect(FldDeviceId,SIGNAL(textEdited(QString))
         ,this,SLOT(CheckNewDeviceId(QString)));
//  FldDeviceId->setEnabled(false);

  FormLayout->setWidget(0,QFormLayout::FieldRole,FldDeviceId);
  FormLayout->setWidget(1,QFormLayout::LabelRole
                       ,new QLabel(tr("Driver class name:"),this));
  sPtr<sTextNode>             ClassNameNP;
  ClassNameNP = DeviceCfg >> "Class";
  if( ! ClassNameNP.IsCorrect() ){
    QLabel *                E = new QLabel(tr("attribute ClassName is absent "
                                              "or has a wrong type")
                                          ,this);
    E->setStyleSheet("color: red");
    FormLayout->setWidget(1,QFormLayout::FieldRole,E);
  }else{
    sString                     ClassName = (sString)(*ClassNameNP);
    FormLayout->setWidget(1,QFormLayout::FieldRole
                         ,new QLabel(ToQString(ClassName),this));
    if( ! sInstrument::Exists(ClassName) ){
      QLabel *                    E = new QLabel(tr("wrong class name"),this);
      E->setStyleSheet("color: red");
      FormLayout->setWidget(2,QFormLayout::FieldRole,E);
    }else{
      DeviceLinks = DeviceCfg >> "Links";
      if( DeviceLinks.IsCorrect() ){
        FormLinks();
      }
//      DeviceCfg |= "Settings";
      sPtr<sFolderNode>              DeviceSettings;
      DeviceSettings = DeviceCfg >> "Settings";
      if( DeviceSettings.IsCorrect() ){
        Layout->addWidget(new QLabel(tr("Settings:"),this));
        Evolution = new sNet2TreeEvolution(sAttribute(sAttrName()
                                                     ,DeviceSettings   ));
        Navigator = new sNavigator(Evolution ,this);
        Navigator->setFrameShape(QFrame::StyledPanel);
        Navigator->setFrameShadow(QFrame::Raised);

        Layout->addWidget(Navigator);

        connect(Evolution,SIGNAL(Changed()),Navigator,SLOT(OnNetChange()));
        Evolution->Initiate();

        UpdatingTimer = new QTimer(this);
        connect(UpdatingTimer, SIGNAL(timeout()), Evolution, SLOT(Update()));
        UpdatingTimer->start(sNavigator::UpdatingPeriod);
      }
    }
    BtnDone = new QPushButton(tr("Done"),this);
    BtnDone->setDefault(true);
    Layout->addWidget(BtnDone,0,Qt::AlignRight);
    connect(BtnDone,SIGNAL(clicked()),this,SLOT(Done()));
  }
}


sConfigureDeviceDialog::sLinkTableRow::sLinkTableRow (sString        link_name
                                                     ,sString        class_name
                                                     ,sString        dev_id
                                                     ,QGridLayout *  layout
                                                     ,int            row)
    :Link(link_name)
    ,Class(class_name)
    ,DevId(dev_id)
{
  ComboBox = new QComboBox(layout->parentWidget());
  layout->addWidget(new QLabel(ToQString(Link),layout->parentWidget())
                   ,row,0,1,1);
  layout->addWidget(new QLabel(ToQString(Class),layout->parentWidget())
                   ,row,1,1,1);
  layout->addWidget(ComboBox,row,2,1,1);
  FillComboBox(DevId);
}


void  sConfigureDeviceDialog::sLinkTableRow::FillComboBox
                                                      (sString  current_dev_id)
{
  ComboBox->clear();
  ComboBox->addItem(ToQString(current_dev_id));
  for_each_attribute( A , sDevice::DeviceList ){
    if( A->Name().Text() != current_dev_id ){
      sPtr<sTextNode>           CandidateClassNP = A->Value() >> "Class";
      sString                   CandidateClass;
      if( CandidateClassNP.IsCorrect() ){
        CandidateClass = (sString)(*CandidateClassNP);
        if( sDevice::Matches(CandidateClass,Class) ){
          ComboBox->addItem(ToQString(A->Name().Text()));
        }
      }
    }
  }end_attr_loop()
}


void  sConfigureDeviceDialog::FormLinks ()
{
  int                         Row = -1;
  for_each_attribute( AF , DeviceLinks ){
    if( Row < 0 ){
      Layout->addWidget(new QLabel(tr("Links:"),this));
      LinkFrame = new QFrame(this);
      LinkLayout = new QGridLayout;
      LinkLayout->setHorizontalSpacing(30);
      LinkFrame->setLayout(LinkLayout);
      LinkFrame->setFrameShape(QFrame::Box);
      Layout->addWidget(LinkFrame);
      QLabel *                  LN = new QLabel(QString("<i><u>")
                                               +tr("Link"),this);
      QLabel *                  CN = new QLabel(QString("<i><u>")
                                               +tr("Class"),this);
      QLabel *                  DI = new QLabel(QString("<i><u>")
                                               +tr("Device Id (choose one)")
                                               ,this);
      LinkLayout->addWidget(LN,++Row,0,1,1);
      LinkLayout->addWidget(CN,Row,1,1,1);
      LinkLayout->addWidget(DI,Row,2,1,1);
    }
    sString                   LinkName = AF->Name().Text();
    sPtr<sTextNode>           ClassNP = AF->Value() >> "Class";
    sString                   Class;
    sPtr<sTextNode>           DevIdNP = AF->Value() >> "DeviceId";
    sString                   DevId;
    if( ClassNP.IsCorrect() ){
      Class = (sString)(*ClassNP);
      if( ! DevIdNP.IsCorrect() ){
        DevId = ""; //"please, select...";
      }else{
        DevId = (sString)(*DevIdNP);
      }
      LinkTable.push_back( new sLinkTableRow(LinkName,Class,DevId
                                            ,LinkLayout,++Row) );
    }
  }end_attr_loop()
}


void  sConfigureDeviceDialog::Uninstall ()
{
  if( QMessageBox::question(this,tr("Uninstall the device")
                           ,tr("Are you sure?"))
      == QMessageBox::Yes ){
    sDevice::DeviceList.DelAttr(DeviceId);
    close();
  }
}


void  sConfigureDeviceDialog::CheckNewDeviceId (QString  NewId)
{
  if( NewId.isEmpty() ){
    BtnDone->setEnabled(false);
  }else{
    sNodePtr      P = sDevice::DeviceList >> FromQString(NewId);
    if( P.IsNULL() || (FromQString(NewId) == DeviceId) ){
      FldDeviceId->setStyleSheet("");
      BtnDone->setEnabled(true);
    }else{
      FldDeviceId->setStyleSheet("color: red");
      BtnDone->setEnabled(false);
    }
  }
}


void  sConfigureDeviceDialog::Done ()
{
  while( ! LinkTable.empty() ){
    if( FromQString( LinkTable.front()->ComboBox->currentText() )
          != LinkTable.front()->DevId ){
      DeviceLinks >> LinkTable.front()->Link
          << attr("DeviceId"
                 ,FromQString( LinkTable.front()->ComboBox->currentText() )
                 );
    }
    LinkTable.pop_front();
  }
  if( BoxDisabled->isChecked() != Disabled ){
//    DeviceCfg << bool_attr("Disabled",BoxDisabled->isChecked());
    DeviceCfg << attr("Disabled",BoxDisabled->isChecked());
  }
  if( FromQString(FldDeviceId->text()) != DeviceId ){
    try{
      sDevice::DeviceList
          << sAttribute(sAttrName(FromQString(FldDeviceId->text()))
                       ,DeviceCfg   );
      sDevice::DeviceList.DelAttr(DeviceId);
    }catch(...){
    }
  }
  close();
}

//------------------------------------------------------------------------------
} //namespace RA




