////////////////////////////////////////////////////////////////////////////////
/*! @file Console.cpp   Консоль прибора (окно управления и индикации).
- Part of RAGUI - Research Assistant Graphical User Interface.
- Uses  QtGui v.4.6  -  http://qt.nokia.com/
- Uses  RANet - Research Assistant Net Library (based on ANSI C++).
- Copyright(C) 2010-2013, Viktor E. Bursian, St.Petersburg, Russia.
                          Viktor.Bursian@mail.ioffe.ru
*///////////////////////////////////////////////////////////////////////////////
#include "Console.h"
#include "FolderNode.h"
#include "BasicNodes.h"
#include "StringsAndQStrings.h"
#include "Log.h"
#include <QSignalTransition>
#include <QMessageBox>
namespace RA {
//------------------------------------------------------------------------------

//-------------------------------------------------------------- sConsoleTag ---

sConsoleTag::sConsoleTag (literal         classname
                         ,pfPreconfigure  preconf
                         ,pfNewConsole    createfunc)
    :sDeviceTag(classname,cConsole,preconf)
    ,TheNewConsoleFunctionPtr(createfunc)
{
}

//----------------------------------------------------------------- sConsole ---
/*! @class sConsole

@par The state machine
@dotfile "Console.state_graph"

Legend: 
- thick solid black arrow - unconditional transition (when the state finishes);
- thin solid black arrow - transition on signal or command or directive;
- dotted black circular arrow - reaction on signal or command or directive
  (place the mouse over the line to bring up a hint);
- dashed black arrow - an alternative transition
  (place the mouse over the line to bring up a hint);
- dashed violet arrow - an additional transition, which appears later, when
  the state machine is working
  (place the mouse over the line to bring up a hint).
  
*/


void  sConsole::Preconfigure (sNodePtr  DeviceCfg)
{
  sDevice::Preconfigure(DeviceCfg);
}


                                  
/*! @details
 
Creates an object of a class derived from sConsole; its unique identifier
DeviceId and the particular class must be registered in the configurational
part of the @ref datanet.
*/
psConsole  sConsole::CreateConsole (sString    device_id
                                   ,sNodePtr   widget_cfg
                                   ,QWidget *  parent    )
{
  psConsole                   Console = NULL;
  sPtr<sFolderNode>           DeviceCfg;
  bool                        Disabled = false;
  sString                     ClassName;

  if( ! IsRunning(device_id) ){
    DeviceCfg = DeviceList >> device_id;
    if( ! DeviceCfg.IsCorrect() )
      throw xProgramError("sConsole::CreateConsole(,,): #1.",__FILE__,__LINE__);
    if( ! get_from<sTextNode>(DeviceCfg>>"Class",ClassName) )
      throw xDeviceLaunch(
          sString("'Class' attribute is not found for the device '")
          +device_id+"'.  Reinstall the device.");
    get_from<sBooleanNode>(DeviceCfg>>"Disabled",Disabled);
    if( Disabled )
      throw xDeviceLaunch(sString("The device '")+device_id+"' is disabled");
    
    pcsTag                      Tag = sTag::ByName(ClassName);
    if( ! Tag )
      throw xDeviceLaunch(sString("Unknown class: ")+ClassName);
          
    pcsConsoleTag               CTag = dynamic_cast<pcsConsoleTag>(Tag);
    if( ! CTag )
      throw xProgramError("sConsole::CreateConsole(,,): #2",__FILE__,__LINE__);
  
    (*(CTag->ThePreconfigureFunctionPtr))(DeviceCfg);
    Console = (*(CTag->TheNewConsoleFunctionPtr))
                                        (device_id,DeviceCfg,widget_cfg,parent);
    if( ! Console )
      throw xProgramError("sConsole::CreateConsole(,,): #3",__FILE__,__LINE__);
    
    if( ! Console->IsInvalid() )
      Console->CallAppearanceMakeUp();
    if( Console->IsInvalid() ){
      QString                   Reason(Console->InvalidityReason());
      Console->deleteLater();  Console=NULL;
      throw xDeviceLaunch(sString("The device '")+device_id
                                +"' is invalid:\n"+FromQString(Reason));
    }
    Console->States->start();
  }
  return Console;
}


sConsole::~sConsole()
{
}


sConsole::sConsole (sString    device_id
                   ,sNodePtr   device_cfg
                   ,sNodePtr   widget_cfg
                   ,QWidget *  parent    )
    :sInstrumentWindow(widget_cfg,parent)
    ,sDevice(device_id,device_cfg)
    ,States(new sDeviceStateMachine(this))
{
  connect(this   ,SIGNAL(ErrorMessage(QString))
         ,this   ,SLOT(ReportError(QString)));
  connect(States ,SIGNAL(NewUpperLevelState(sDeviceStateMachine::eState))
         ,this   ,SIGNAL(NewUpperLevelState(sDeviceStateMachine::eState)));
  connect(States ,SIGNAL(NewUpperLevelState(QString))
         ,this   ,SIGNAL(NewUpperLevelState(QString)));

  connect(this ,SIGNAL(WarmUp())   ,States  ,SIGNAL(WarmUp()));
//  connect(this ,SIGNAL(CoolDown()) ,States  ,SIGNAL(CoolDown()));

  connect(States ,SIGNAL(started())
         ,this   ,SLOT(AfterStartInitiation()));
  State_Warming()
      ->OnEnterCall(SLOT(OnEnterToWarmingState()))
      ->AddTransition( SIGNAL(CoolDown()) ,State_Cooling() )
      ->OnExitCall(SLOT(OnExitFromWarmingState()));
  State_Operating()
      ->OnEnterCall(SLOT(OnEnterToOperatingState()))
      ->AddTransition( SIGNAL(CoolDown()) ,State_Cooling() )
      ->AddTransition<QString>(SIGNAL(Failure(QString)) 
                              ,SLOT(ReportFailure(QString))
                              ,State_Cooling() )
      ->OnExitCall(SLOT(OnExitFromOperatingState()));
  State_Cooling()
      ->OnEnterCall(SLOT(OnEnterToCoolingState()))
      ->OnExitCall(SLOT(OnExitFromCoolingState()));
  State_Frozen()
        ->AddTransition(this ,SIGNAL(CoolDown()) ,State_Frozen() );

  bool                        AutoWarming = true;
  if( ! DeviceConfig.IsNULL() ){
    sPtr<sBooleanNode>          B (DeviceConfig>>"AutoWarming");
    if( B.IsCorrect() )  AutoWarming = (bool)(*B);
  }
  if( AutoWarming ){
    States->setInitialState(State_Warming());
  }else{
    States->setInitialState(State_Frozen());
  }
  Translate();
}


void  sConsole::Translate ()
{
  TextWarming = QString(tr("warming"));
  TextOperating = QString(tr("operating"));
  TextCooling = QString(tr("cooling"));
  TextFrozen = QString(tr("frozen"));
}


void  sConsole::DoClose ()
{
  RANet::Log.Put(sLog::Debug,"Quitting"
                ,sString("Closing: '") + DeviceIdentifier() + "'");
  State_Frozen()
      ->AddTransition(State_Final());
  connect(States,SIGNAL(finished()),this,SIGNAL(QuittingIsAllowed()));
  emit CoolDown();
}


/*! Connects a signal from a driver with  a slot (or signal) of this console.

@param sender          Pointer to the driver.  Should be set to a certain
                       device before the call of this function, otherwise the
                       function does nothing and returns @a false.
@param signal          Signal of that driver.  Use SIGNAL macro in this place.
@param slot_or_signal  Slot or signal of @b this console.  Use SLOT (or SIGNAL)
                       macro here.

The connection will be @a Qt::QueuedConnection.

@sa Disconnect(rcsDriverLink,const char *,const char *).
*/
bool  sConsole::Connect (rcsDriverLink  sender
                        ,const char *  signal
                        ,const char *  slot_or_signal)
{
  sString                     Signal (signal);
  sString                     SlotOrSignal (slot_or_signal);
  if( Signal[0] == '1' ){
    throw xProgramError(sString("sConsole::Connect: not a signal: ")+signal
                       ,__FILE__,__LINE__);
  }else{
    Signal = sString("SIGNAL(")+Signal.Tail(1)+")";
  }
  if( SlotOrSignal[0] == '1' ){
    SlotOrSignal = sString("SLOT(")+SlotOrSignal.Tail(1)+")";
  }else{
    SlotOrSignal = sString("SIGNAL(")+SlotOrSignal.Tail(1)+")";
  }
  if( ! sender.IsCorrect() )
    return false;
  if( TraceStates() || sender->TraceStates() )
    RANet::Log.Put(sLog::Debug,"States"
                  ,sString("Connecting: '") + sender->DeviceIdentifier()
                  +sString("'.'") + Signal
                  +sString("' --> '") + DeviceIdentifier()
                  +sString("'.'") + SlotOrSignal + "'");
  if( ! QObject::connect(sender.operator ->() ,signal
                        ,this                 ,slot_or_signal
                        ,Qt::QueuedConnection) ){
    RANet::Log.Put(sLog::ProgramError,"States"
                  ,sString("Connect failed: '") + sender->DeviceIdentifier()
                  +sString("'.'") + Signal
                  +sString("' --> '") + DeviceIdentifier()
                  +sString("'.'") + SlotOrSignal + "'");
    return false;
  }   
  return true;
}



/*! Disconnects a signal from a driver from a slot (or signal) of this console.

@param sender          Pointer to the driver.  If it is set to NULL the
                       function does nothing and returns @a false.
@param signal          Signal of that driver.  Use SIGNAL macro in this place.
@param slot_or_signal  Slot or signal of @b this console.  Use SLOT (or SIGNAL)
                       macro here.

The 2nd and/or 3rd argument can be 0 (or NULL) - meaning "all".

@sa Connect(rcsDriverLink,const char *,const char *).
*/
bool  sConsole::Disconnect (rcsDriverLink  sender
                           ,const char *  signal
                           ,const char *  slot_or_signal)
{
  if( ! sender.IsCorrect() )
    return false;
  return QObject::disconnect(sender.operator ->() ,signal
                            ,this                 ,slot_or_signal);
}



/*! Sends a command to a submissive driver.

@param receiver        Link to the submissive driver.  Should be set to a 
                       certain device before the call of this function, 
                       otherwise the function does nothing and returns @a false.
@param command         Use reference to a static csCommand object, declared in
                       the submissive device's class (or in its ancestor).
@param control_change  Optional: take or release an exclusive control over the
                       submissive device.
@return @a true, if succeed.

This function posts the command event into the driver's state machine queue, if
the device is not under exclusive control by another device.  Otherwise, the
function returns @a false and the command is ignored (it won't wait until the
other device releases control).
Simultaneously, if @a control_change==cTakeControl it takes exclusive control 
over the @a receiver, or releases it, if @a control_change==cReleaseControl.
  
Usage:
@code
    if( ! SendCommand(Gauge ,sGauge::cStart ,cTakeControl) ){
      //command had not been sent
    };
@endcode

@sa SendDirective<...>(...).

*/
bool  sConsole::SendCommand (rcsDriverLink   receiver
                            ,rcsCommand      command
                            ,eControlChange  control_change)
{
  if( receiver.IsCorrect() ){
    return receiver->GetCommand(new sState::sCommandEvent(this,command
                                                          ,control_change));
  }
  return false;
}


void  sConsole::ReportError (QString  error_text)
{
  QMessageBox::critical(this ,tr("Instrument error")
                       ,tr("Error in '%1' console:\n\n%2")
                         .arg(ToQString(DeviceIdentifier()))
                         .arg(error_text)     
                       );
}


void  sConsole::ReportFailure (QString  error_text)
{
  QMessageBox::critical(this ,tr("Instrument failure")
                       ,tr("Failure in '%1' console:\n\n%2")
                         .arg(ToQString(DeviceIdentifier()))
                         .arg(error_text)     
                       );
}



/*! @fn  template<class data_type> void sConsole::SendDirective(rcsDriverLink,const sDirective<data_type> &,data_type,eControlChange)

@brief Sends a directive to a submissive driver 
       (same as SendCommand(...), but a directive carries data with it).

@param receiver        Link to the submissive driver.  Should be set to a 
                       certain device before the call of this function, 
                       otherwise the function does nothing and returns @a false.
@param directive       Use reference to a static const sDirective object, 
                       declared in the submissive device's class (or in its 
                       ancestor).
@param data            An expression of the @a data_type type.  It will be sent
                       with the directive as the attached data.
@param control_change  Optional: take or release an exclusive control over the
                       submissive device.
@return @a true, if succeed.

This function posts the directive event into the driver's state machine queue, 
if the device is not under exclusive control by another device.  Otherwise, the
function returns @a false and the command is ignored (it won't wait until the
other device releases control).
Simultaneously, if @a control_change==cTakeControl it takes exclusive control 
over the @a receiver, or releases it, if @a control_change==cReleaseControl.
  
Usage:
@code
    if( ! SendDirective(Gauge ,sGauge::cSetPeriod, sPhysValue(0.6,"s")
                       ,cTakeControl) ){
      //directive had not been sent
    };
@endcode

@sa SendCommand(...).
*/



/*! @fn void sConsole::AppearanceMakeUp()
 
@brief Should create visual representation of the console - all layouts,
       subwidgets, etc.

This method is called by the core right after constructor (before
RestoreState() and before the start of the state machine).
In sConsole class this function does nothing, but you can reimplement it in a
descendant. 

The purpose of this method is to separate the appearance of the console from
it's functionality (state machine structure and behaviour). This cannot be
done, if the appearance is created in the constructor (virtual methods do not
work within constructors).

After creation and placing into layouts, the subwidgets may be connected to the
signals and slots of @b this instanse of the console, which, in turn,
interact with the console's state machine.

You may call an ancestor's AppearanceMakeUp() at the beginning of this
function, then extend or modify the inherited appearance.

@note
An ancestor does not necessary mean parent; it can be grand-parent or
great-grand-parent.

Alternatively, you may re-implement this method without calling an ancestor's
implementation of it, which gives you freedom to completely change the
appearance.  In other words, for the consoles, the manners should be inherited
(and, perhaps, extended), while the appearance could have nothing in common.
*/



/*! @fn void sConsole::AfterStartInitiation()
 
@brief Called when the state machine starts
       (before entering its initial state).

In sConsole class this function does nothing, but you can reimplement it in a
descendant. 
*/



/*! @fn void sConsole::OnEnterToWarmingState()
 
@brief Called by the core on enter into the State_Warming().

Added mainly for the completeness. In sConsole class this function does nothing,
but you can reimplement it in a descendant. 
*/



/*! @fn void sConsole::OnExitFromWarmingState()
 
@brief Called by the core on exit from the State_Warming(). 

Added mainly for the completeness. In sConsole class this function does nothing,
but you can reimplement it in a descendant. 
*/



/*! @fn void sConsole::OnEnterToOperatingState()

@brief Called by the core on enter into the State_Operating().  

In sConsole class this function does nothing, but you can reimplement it in a
descendant. Usually, you should connect the signals of the submissive drivers 
with the signals (or slots) of this device:
@code
  Connect(NewDevice1 ,SIGNAL(NewData(sPhysValue)),SIGNAL(ShowData(sPhysValue)));
@endcode
Note, this could not be done, say, in constructor, before the link to
the submissive driver got its value.

Note, there is another way to do something on enter into State_Operating(),
namelly, you can add 
@code
    State_Operating()->OnEnterCall(...);
@endcode
into the constructor.  This will call your slot after all actions,
made by the ancestor's class.  Here, however, you can do something BEFORE
the ancestor's actions.

@sa OnExitFromOperatingState().
*/



/*! @fn void sConsole::OnExitFromOperatingState()

@brief Called by the core on exit from the State_Operating().  

In sConsole class this function does nothing, but you can reimplement it in a
descendant. Usually, you should disconnect the signals of the submissive 
drivers, which were connected on enter into this state:
@code
  Disconnect(NewDevice1 ,SIGNAL(NewData(sPhysValue)) 
            ,SIGNAL(ShowData(sPhysValue)));
@endcode

Note, there is another way to do something on exit from State_Operating(),
namelly, you can add 
@code
    State_Operating()->OnExitCall(...);
@endcode
into the constructor.  This will call your slot after all actions,
made by the ancestor's class.  Here, however, you can do something BEFORE
the ancestor's actions.

@sa OnEnterToCoolingState().
*/



/*! @fn void sConsole::OnEnterToCoolingState()

@brief Called by the core on enter into the State_Cooling().

In sConsole class this function does nothing, but you can reimplement it in a
descendant. Usually, you should disconnect all submissive drivers, like this:
@code
  Disconnect(NewDevice1 ,0 ,0);
  NewDevice1.ReleaseExclusiveControl();
  NewDevice1.SetToNULL();
@endcode

Note, there is another way to do something on enter into State_Cooling(),
namelly, you can add 
@code
    State_Cooling()->OnEnterCall(...);
@endcode
into the constructor.  This will call your slot after all actions,
made by the ancestor's class.  Here, however, you can do something BEFORE
the ancestor's actions.

@sa OnExitFromOperatingState().
*/



/*! @fn void sConsole::OnExitFromCoolingState()

@brief Called by the core on exit from the State_Cooling().  

Added mainly for the completeness. In sConsole class this function does nothing,
but you can reimplement it in a descendant. 
*/

//------------------------------------------------------------------------------
} //namespace RA


