////////////////////////////////////////////////////////////////////////////////
/*! @file Console.h   Консоль прибора (окно управления и индикации).
- Part of RAGUI - Research Assistant Graphical User Interface.
- Uses  QtGui v.4.6  -  http://qt.nokia.com/
- Uses  RANet - Research Assistant Net Library (based on ANSI C++).
- Copyright(C) 2010-2013, Viktor E. Bursian, St.Petersburg, Russia.
                          Viktor.Bursian@mail.ioffe.ru
*///////////////////////////////////////////////////////////////////////////////
#ifndef Console_H
#define Console_H
#include "Device.h"
#include "DeviceStates.h"
#include "InstrumentWindow.h"
#include "Driver.h"
namespace RA {
//------------------------------------------------------------------------------

ANNOUNCE_CLASS(sConsoleTag)
ANNOUNCE_ABSTRACT_CLASS(sConsole)

//-------------------------------------------------------------- sConsoleTag ---

class RAGUI_EXPORT  sConsoleTag : public sDeviceTag
{
  private: // types
    typedef psConsole         fNewConsole (sString,sNodePtr,sNodePtr,QWidget*);
    typedef fNewConsole *     pfNewConsole;

  protected:
                              sConsoleTag (literal         classname
                                          ,pfPreconfigure  preconf
                                          ,pfNewConsole    createfunc);

  private: // fields
    pfNewConsole              TheNewConsoleFunctionPtr;

  friend class sConsole;
};


//------------------------------------------------------- CONSOLE(CLASSNAME) ---
/*! Этот макрос должен стоять в начале декларации класса, публично
    порождённого от RA::sConsole, однако из этого правила есть исключение

Исключение:
- Абстрактные классы, публично порождённые от RA::sConsole, вместо макроса
  CONSOLE используют макрос ABSTRACT_CONSOLE().

Декларирует protected конструктор с параметрами
(RA::sString DeviceId ,RA::sNodePtr DeviceCfg ,RA::sNodePtr WidgetCfg).
Он @b должен @b быть @b определён в соответствующем cpp-файле.
Смысл параметров:
@arg @c DeviceId задаваемый в конфигурации уникальный идентификатор
                 реально существующего "в железе" экземпляра устройства
@arg @c DeviceCfg специального вида указатель на узел сети данных,
                  содержащий параметры устройства и, возможно, состояние
                  в котором оно было оставлено после последнего использования
                  (например, текущая длина волны для спектрометра с шаговым
                  управлением развёрткой, но без датчика угол-код)
@arg @c WidgetCfg специального вида указатель на узел сети данных,
                  содержащий состояние виджета (т.е. пользовательского
                  интерфейса) прибора; сюда заносятся данные, сохраняемые при
                  закрытии программы без закрытия окна прибора, для
                  восстановления при следующем запуске программы

Также декларирует некий статический член - тэг класса, обеспечивающий
динамическое создание объектов класса (т.е. экземпляров устройств), их поиск
и т.п.  В соответствующем cpp-файле необходимо определить этот тэг при помощи
макро DEFINE_CLASS_TAG() с именем класса в качестве аргумента.

Поскольку класс sConsole является виджетом и, следовательно, наследником
класса QObject, в декларации порождаемых от sConsole классов следует ставить
ещё и макро Q_OBJECT (причём, в @b самом начале).

Использование:
@code
class sMyConsole : public sConsole
{
  Q_OBJECT
  CONSOLE(sMyConsole)
  //...в классе не должно быть чистых (pure) виртуальных методов
};
@endcode

@macro
*/
#define  CONSOLE(CLASSNAME)                                                    \
  public:                                                                      \
    class sIntrinsicTag : public sConsoleTag                                   \
    {                                                                          \
      public:                                                                  \
                                  sIntrinsicTag ()                             \
                                      :sConsoleTag(#CLASSNAME                  \
                                                  ,Preconfigure,NewConsole)    \
                                    {};                                        \
        virtual bool              CanCast (psAdam P) const                     \
                                    { return bool(                             \
                                          dynamic_cast<p##CLASSNAME>(P)        \
                                                 );                            \
                                    };                                         \
    };                                                                         \
    static sIntrinsicTag      AdamInheritorTag;                                \
  public:                                                                      \
    static void               Preconfigure (sNodePtr  DevCfg);                 \
  protected:                                                                   \
                              CLASSNAME (sString,sNodePtr,sNodePtr,QWidget*);  \
  private:                                                                     \
    static psConsole          NewConsole (sString   Id                         \
                                         ,sNodePtr  DevCfg                     \
                                         ,sNodePtr  WdgtCfg                    \
                                         ,QWidget*  parent)                    \
                                { return new CLASSNAME(Id,DevCfg               \
                                                      ,WdgtCfg,parent); };     \
  private:




//---------------------------------------------- ABSTRACT_CONSOLE(CLASSNAME) ---
/*! Этот макрос должен стоять в начале декларации @b абстрактного класса,
    публично порождённого от RA::sConsole.

Декларирует protected конструктор с параметрами
(RA::sString DeviceId ,RA::sNodePtr DeviceCfg ,RA::sNodePtr WidgetCfg).
Он @b должен @b быть @b определён в соответствующем cpp-файле.
Смысл параметров:
@arg @c DeviceId задаваемый в конфигурации уникальный идентификатор
                 реально существующего "в железе" экземпляра устройства
@arg @c DeviceCfg специального вида указатель на узел сети данных,
                  содержащий параметры устройства и, возможно, состояние
                  в котором оно было оставлено после последнего использования
                  (например, текущая длина волны для спектрометра с шаговым
                  управлением развёрткой, но без датчика угол-код)
@arg @c WidgetCfg специального вида указатель на узел сети данных,
                  содержащий состояние виджета (т.е. пользовательского
                  интерфейса) прибора; сюда заносятся данные, сохраняемые при
                  закрытии программы без закрытия окна прибора, для
                  восстановления при следующем запуске программы

Также декларирует некий статический член - тэг класса, обеспечивающий
динамическое создание объектов класса (т.е. экземпляров устройств), их поиск
и т.п.  В соответствующем cpp-файле необходимо определить этот тэг при помощи
макро DEFINE_CLASS_TAG() с именем класса в качестве аргумента.

Поскольку класс sConsole является виджетом и, следовательно, наследником
класса QObject, в декларации порождаемых от sConsole классов следует ставить
ещё и макро Q_OBJECT (причём, в @b самом начале).

Использование:
@code
class sMyConsole : public sConsole
{
  Q_OBJECT
  ABSTRACT_CONSOLE(sMyConsole)
  //... в классе могут быть чистые (pure) виртуальные методы
  //... а могут и не быть, но класс по замыслу абстрактен,
  //... т.е. объекты этого класса никогда не создаются
};
@endcode

@macro
*/
#define  ABSTRACT_CONSOLE(CLASSNAME)                                           \
  ADAM(CLASSNAME)                                                              \
  public:                                                                      \
    static void               Preconfigure (sNodePtr  DevCfg);                 \
  protected:                                                                   \
                              CLASSNAME (sString,sNodePtr,sNodePtr,QWidget*);  \
  private:



//----------------------------------------------------------------- sConsole ---
/*! Абстрактный базовый класс для консоли прибора
         (окна управления и индикации).

*/
class RAGUI_EXPORT  sConsole : public sInstrumentWindow
                             , public sDevice
{
  Q_OBJECT
  ABSTRACT_CONSOLE(sConsole)
  
  public:
    virtual                   ~sConsole ();

    psState                   State_Warming () const
                                { return States->Warming; }
    psState                   State_Operating () const
                                { return States->Operating; }
    psState                   State_Cooling () const
                                { return States->Cooling; }
    psState                   State_Frozen () const
                                { return States->Frozen; }
    QFinalState *             State_Final () const
                                { return States->FinalState; }
  
  public: //reimplementing inherited
    virtual void              DoClose ();

  signals: //outgoing
    void                      NewUpperLevelState (sDeviceStateMachine::eState);
    void                      NewUpperLevelState (QString  new_state_name);

  public:
    sDeviceStateMachine *     States;

  signals: //for internal needs
    void                      WarmUp ();
    void                      CoolDown ();
    void                      ErrorMessage (QString = QString());
    void                      Failure (QString = QString());

  protected: //inherited
    virtual void              Retranslate ()
                                {
                                  sInstrumentWindow::Retranslate();
                                  Translate();
                                }

  protected:
    bool                      Connect (rcsDriverLink  sender
                                      ,const char *  signal
                                      ,const char *  slot_or_signal);
    bool                      Disconnect (rcsDriverLink  sender
                                         ,const char *  signal
                                         ,const char *  slot_or_signal);
    bool                      SendCommand (rcsDriverLink   receiver
                                          ,rcsCommand      command
                                          ,eControlChange  control_change
                                                                  =cLeaveControl
                                          );
    template<class data_type>
    bool                      SendDirective (rcsDriverLink     receiver
                                            ,const sDirective<data_type> &
                                                               directive
                                            ,data_type         data
                                            ,eControlChange    control_change
                                                                  =cLeaveControl
                                   )
                                {
                                  if( receiver.IsCorrect() ){
                                    return receiver->GetCommand
                                        (new sState::sDirectiveEvent<data_type>
                                          (this,directive,data,control_change));
                                  };
                                  return false;
                                }

    virtual void              AppearanceMakeUp ()
                                {}

  protected slots:
    virtual void              ReportError (QString);
    virtual void              ReportFailure (QString);
    virtual void              AfterStartInitiation ()
                                {}
    virtual void              OnEnterToWarmingState ()
                                {}
    virtual void              OnExitFromWarmingState ()
                                {}
    virtual void              OnEnterToOperatingState ()
                                {}
    virtual void              OnExitFromOperatingState ()
                                {}
    virtual void              OnEnterToCoolingState ()
                                {}
    virtual void              OnExitFromCoolingState ()
                                {}

  protected: //fields
    QString                   TextWarming;
    QString                   TextOperating;
    QString                   TextCooling;
    QString                   TextFrozen;

  private: // static
    static psConsole          CreateConsole (sString    device_id
                                            ,sNodePtr   widget_cfg
                                            ,QWidget *  parent    );

  private:
    void                      CallAppearanceMakeUp ()
                                { AppearanceMakeUp(); }
    void                      Translate ();
  friend class sRAMainWindow;
};

//------------------------------------------------------------------------------
} //namespace RA
#endif
