////////////////////////////////////////////////////////////////////////////////
/*! @file DBLinkingDialogs.cpp Classes sEditDBInfoDialog and sEditDBLinkDialog.
- Part of RAGUI - Research Assistant Graphical User Interface.
- Uses  RANet - Research Assistant Net Library (based on ANSI C++).
- Uses  Qt v.5  - http://qt.io/
- Copyright(C) 2013-2017, Viktor E. Bursian, St.Petersburg, Russia.
                          VBursian@gmail.com
*///////////////////////////////////////////////////////////////////////////////
#include "DBLinkingDialogs.h"
#include "StringsAndQStrings.h"
#include "Helper.h"
#include <QMessageBox>
#include <QDialogButtonBox>
#include <QVBoxLayout>
#include <QHBoxLayout>
#include <QDir>
#include <QFileDialog>
#include <QIcon>
#include <QDebug>
namespace RA {
//------------------------------------------------------------------------------

const QRegExp      sEditDBInfoDialog::PrefixRegExpr("[a-z][_a-z0-9]+");
const QRegExp      sEditDBInfoDialog::SchemaRegExpr("[a-z][_a-z0-9]+");
const QRegExp      sEditDBInfoDialog::DBNameRegExpr("[a-z][_a-z0-9]+");

const sEditDBInfoDialog::sLinker      sEditDBInfoDialog::Linker;

//----------------------------------------------- sEditDBInfoDialog::sLinker ---

sEditDBInfoDialog::sLinker::sLinker ()
{
  sDB_Info::ModifyInfo = & sEditDBInfoDialog::Modify;
}

//-------------------------------------------------------- sEditDBInfoDialog ---

bool  sEditDBInfoDialog::Modify (psDB_Info  info)
{
  sEditDBInfoDialog *             D = new sEditDBInfoDialog(info
                                            ,tr("Select database"));
  D->setAttribute(Qt::WA_DeleteOnClose);
  return  D->exec() == QDialog::Accepted;
}


sEditDBInfoDialog::sEditDBInfoDialog (psDB_Info  info
                                     ,QString    title
                                     ,QWidget *  parent)
    :QDialog(parent)
    ,Info(info)
{
  setWindowTitle(title);
  resize(600, 100);
  setLayout(new QVBoxLayout());

  URLLabel = new QLabel(tr("URL:   "),this);
  URLField = new QLabel(this);

  TypeField = new QComboBox(this);

  HostLabel = new QLabel(tr("Host"),this);
  HostField = new QLineEdit(this);

  PortLabel = new QLabel(tr("Port"),this);
  PortField = new QSpinBox(this);
  PortField->setRange(-1,0xffff);
  PortField->setSpecialValueText(tr("auto"));

  LoginLabel = new QLabel(tr("Login"),this);
  LoginField = new QLineEdit(this);

  PasswordLabel = new QLabel(tr("Password"),this);
  PasswordField = new QLineEdit(this);
  PasswordField->setEchoMode(QLineEdit::Password);

  ShowPasswordButton = new QToolButton(this);
  ShowPasswordButton->setIcon(QIcon(":/RAGUI/icons/eye.png"));

  DBNameLabel = new QLabel("",this);
  DBNameField = new QLineEdit(this);
  BrowseButton = new QPushButton(tr("Browse"),this);
  RelativeCheckBox = new QCheckBox(tr("relative to the current folder:"),this);
  CurrentDirField = new QLabel(QDir::currentPath(),this);
  DBNameValidator = new QRegExpValidator(DBNameRegExpr,this);

  SchemaLabel = new QLabel(tr("Schema name"),this);
  SchemaField = new QLineEdit(this);

  PrefixLabel = new QLabel(tr("Table naming prefix"),this);
  PrefixField = new QLineEdit(this);

  OptionsLabel = new QLabel(tr("Connection options"),this);
  OptionsField = new QLineEdit(this);

  QDialogButtonBox *  ButtonBox = new QDialogButtonBox(this);
  ButtonBox->setOrientation(Qt::Horizontal);
  ButtonBox->setStandardButtons( QDialogButtonBox::Help
                               | QDialogButtonBox::Cancel
                               | QDialogButtonBox::Ok
//                               | QDialogButtonBox::Open
                               );
  //layout
  {
    QHBoxLayout *             L;

    (L = new QHBoxLayout())->setAlignment(Qt::AlignLeft);
    ((QVBoxLayout*)layout())->addItem(L);
    L->addWidget(URLLabel);
    L->addWidget(URLField);

    (L = new QHBoxLayout())->setAlignment(Qt::AlignLeft);
    ((QVBoxLayout*)layout())->addItem(L);
    L->addWidget(new QLabel(tr("Type"),this));
    L->addWidget(TypeField);

    (L = new QHBoxLayout())->setAlignment(Qt::AlignLeft);
    ((QVBoxLayout*)layout())->addItem(L);
    L->addWidget(HostLabel);
    L->addWidget(HostField);
    L->addWidget(PortLabel);
    L->addWidget(PortField);

    (L = new QHBoxLayout())->setAlignment(Qt::AlignLeft);
    ((QVBoxLayout*)layout())->addItem(L);
    L->addWidget(LoginLabel);
    L->addWidget(LoginField);
    L->addWidget(PasswordLabel);
    L->addWidget(PasswordField);
    L->addWidget(ShowPasswordButton);

    (L = new QHBoxLayout())->setAlignment(Qt::AlignLeft);
    ((QVBoxLayout*)layout())->addItem(L);
    L->addWidget(DBNameLabel);
    L->addWidget(DBNameField);
    L->addWidget(BrowseButton);

    (L = new QHBoxLayout())->setAlignment(Qt::AlignLeft);
    ((QVBoxLayout*)layout())->addItem(L);
    L->addWidget(RelativeCheckBox);

    (L = new QHBoxLayout())->setAlignment(Qt::AlignLeft);
    ((QVBoxLayout*)layout())->addItem(L);
    L->addWidget(CurrentDirField);

    (L = new QHBoxLayout())->setAlignment(Qt::AlignLeft);
    ((QVBoxLayout*)layout())->addItem(L);
    L->addWidget(SchemaLabel);
    L->addWidget(SchemaField);
    L->addWidget(PrefixLabel);
    L->addWidget(PrefixField);

    layout()->addWidget(OptionsLabel);
    layout()->addWidget(OptionsField);

    ((QVBoxLayout*)layout())->addStretch(1);
    layout()->addWidget(ButtonBox);
  }

  PrefixField->setValidator(new QRegExpValidator(PrefixRegExpr));
  SchemaField->setValidator(new QRegExpValidator(SchemaRegExpr));

  TypeField->addItem(ToQString( sDB_Info::TypeName[sDB_Info::cSQLite] ));
  TypeField->addItem(ToQString( sDB_Info::TypeName[sDB_Info::cMySQL ] ));
  TypeField->addItem(ToQString( sDB_Info::TypeName[sDB_Info::cPostgreSQL ] ));
  TypeField->setCurrentIndex( Info->Type );
  RelativeCheckBox->setChecked(true);

  OnTypeChanged(TypeField->currentIndex());

  connect(TypeField ,SIGNAL(currentIndexChanged(int))
         ,this      ,SLOT(OnTypeChanged(int)));
  connect(RelativeCheckBox ,SIGNAL(clicked())
         ,this      ,SLOT(OnRelativeCheckBoxChange()));
  connect(BrowseButton ,SIGNAL(clicked())
         ,this         ,SLOT(Browse()));
  connect(HostField ,SIGNAL(textChanged(QString))
         ,this      ,SLOT(ShowURL()));
  connect(PortField ,SIGNAL(valueChanged(int))
         ,this      ,SLOT(ShowURL()));
  connect(LoginField ,SIGNAL(textChanged(QString))
         ,this      ,SLOT(ShowURL()));
  connect(PasswordField ,SIGNAL(textChanged(QString))
         ,this      ,SLOT(ShowURL()));
  connect(ShowPasswordButton ,SIGNAL(pressed())
         ,this      ,SLOT(ShowPassword()) );
  connect(ShowPasswordButton ,SIGNAL(released())
         ,this      ,SLOT(MaskPassword()) );
  connect(DBNameField ,SIGNAL(textChanged(QString))
         ,this      ,SLOT(ShowURL()));
  connect(SchemaField ,SIGNAL(textChanged(QString))
         ,this      ,SLOT(ShowURL()));
  connect(PrefixField ,SIGNAL(textChanged(QString))
         ,this      ,SLOT(ShowURL()));
  connect(OptionsField ,SIGNAL(textChanged(QString))
         ,this         ,SLOT(ShowURL()));
  connect(ButtonBox ,SIGNAL(accepted()) ,this ,SLOT(OnOK()));
  connect(ButtonBox ,SIGNAL(rejected()) ,this ,SLOT(reject()));
  connect(ButtonBox,SIGNAL(helpRequested()),this,SLOT(ShowHelp()));
}


void  sEditDBInfoDialog::ShowHelp ()
{
  sHelper::ShowKeyId("sEditDBInfoDialog");
};


void  sEditDBInfoDialog::OnTypeChanged (int  type)
{

  switch( sDB_Info::eType(type) ){
    case sDB_Info::cSQLite:
          HostLabel->setVisible( false );
          HostField->setVisible( false );
          PortLabel->setVisible( false );
          PortField->setVisible( false );
          LoginLabel->setVisible( false );
          LoginField->setVisible( false );
          PasswordLabel->setVisible( false );
          PasswordField->setVisible( false );
          ShowPasswordButton->setVisible( false );
          RelativeCheckBox->setVisible( true );
          CurrentDirField->setVisible( true );
          DBNameLabel->setText(tr("SQLite file"));
          DBNameField->setValidator(NULL);
          BrowseButton->setVisible( true );
          SchemaLabel->setVisible( false );
          SchemaField->setVisible( false );
          break;
    default:
          HostLabel->setVisible( true );
          HostField->setVisible( true );
          PortLabel->setVisible( true );
          PortField->setVisible( true );
          LoginLabel->setVisible( true );
          LoginField->setVisible( true );
          PasswordLabel->setVisible( true );
          PasswordField->setVisible( true );
          ShowPasswordButton->setVisible( true );
          RelativeCheckBox->setVisible( false );
          CurrentDirField->setVisible( false );
          DBNameLabel->setText(tr("Database name"));
          DBNameField->setValidator(DBNameValidator);
          BrowseButton->setVisible( false );
          SchemaLabel->setVisible( true );
          SchemaField->setVisible( true );
          break;
  }
  if( sDB_Info::eType(type) == Info->Type ){
    HostField->setText(ToQString( Info->Host ));
    PortField->setValue(Info->Port);
    LoginField->setText(ToQString( Info->Login ));
    PasswordField->setText(ToQString( Info->Password ));
    DBNameField->setText(ToQString( Info->DBName ));
    SchemaField->setText(ToQString( Info->SchemaName ));
    PrefixField->setText(ToQString( Info->TableNamePrefix ));
    OptionsField->setText(ToQString( Info->ConnectOptions ));
  }else{
    sDB_Info          I( (sDB_Info::eType) type );
    HostField->setText(ToQString( I.Host ));
    PortField->setValue(I.Port);
    LoginField->setText(ToQString( I.Login ));
    PasswordField->setText(ToQString( I.Password ));
    DBNameField->setText(ToQString( I.DBName ));
    SchemaField->setText(ToQString( I.SchemaName ));
    PrefixField->setText(ToQString( I.TableNamePrefix ));
    OptionsField->setText(ToQString( I.ConnectOptions ));
  }
  ShowURL();
}


void  sEditDBInfoDialog::OnRelativeCheckBoxChange ()
{
  if( RelativeCheckBox->isChecked() ){
    DBNameField->setText( QDir::current().relativeFilePath(DBNameField->text()) );
  }else{
    DBNameField->setText( QFileInfo(DBNameField->text()).absoluteFilePath() );
  }
}


void  sEditDBInfoDialog::ShowURL ()
{
  TempInfo.Type = sDB_Info::eType( TypeField->currentIndex() );
  TempInfo.DBName = FromQString( DBNameField->text() );
  TempInfo.SchemaName = FromQString( SchemaField->text() );
  TempInfo.TableNamePrefix = FromQString( PrefixField->text() );
  TempInfo.Host = FromQString( HostField->text() );
  TempInfo.Port = PortField->value();
  TempInfo.Login = FromQString( LoginField->text() );
  TempInfo.Password = FromQString( PasswordField->text() );
  TempInfo.ConnectOptions = FromQString( OptionsField->text() );
  URLField->setText(ToQString(sString("<b>")+TempInfo.URL())+"</b>");

  //back translation
  {
//    TempInfo.SetByURL(TempInfo.URL());
//    URLField->setText(URLField->text()
//        +"<br/><small>"
//        +"<br/>&nbsp;&nbsp;type="+ToQString(sDB_Info::TypeName[TempInfo.Type])
//        +"<br/>&nbsp;&nbsp;login="+ToQString(TempInfo.Login)
//        +"<br/>&nbsp;&nbsp;password="+ToQString(TempInfo.Password)
//        +"<br/>&nbsp;&nbsp;host="+ToQString(TempInfo.Host)
//        +"<br/>&nbsp;&nbsp;port="+QString().setNum(TempInfo.Port)
//        +"<br/>&nbsp;&nbsp;dbname="+ToQString(TempInfo.DBName)
//        +"<br/>&nbsp;&nbsp;schema="+ToQString(TempInfo.SchemaName)
//        +"<br/>&nbsp;&nbsp;prefix="+ToQString(TempInfo.TableNamePrefix)
//        +"<br/>&nbsp;&nbsp;options="+ToQString(TempInfo.ConnectOptions)
//        +"</small>"
//        );
  }
}


void  sEditDBInfoDialog::Browse ()
{
  QString                     Name = DBNameField->text();

  Name = QFileDialog::getOpenFileName(DBNameField
      ,""
      ,QFileInfo(Name).path()
      ,"RANet sqlite files (*.RANet.sqlite);;All files (*)"
      ,0
      ,QFileDialog::DontUseCustomDirectoryIcons
      );

  //!@todo{UI} There should be different dialog styles:
  //!  - when setting a link to related part
  //!    - as is, but the file must exist;
  //!  - when selecting a root part to open
  //!    - RelativeCheckBox is senseless,
  //!    - file must exist;
  //!  - when creating new root part
  //!    - RelativeCheckBox is senseless,
  //!    - file should not exist,
  //!    - and it should be possible to create a folder.

//  Dir = QFileDialog::getExistingDirectory(NameField
//      ,""
//      ,QFileInfo(Name).path()
//      );

//  Name = QFileDialog::getSaveFileName(NameField
//      ,""
//      ,Name
//      ,"RANet sqlite files (*.RANet.sqlite);;All files (*)"
//      ,0
//      ,QFileDialog::DontUseCustomDirectoryIcons
//      );

  if( ! Name.isEmpty() ){
    if( RelativeCheckBox->isChecked() ){
      Name = QDir::current().relativeFilePath(Name);
    }
    DBNameField->setText(Name);
  }
}


void  sEditDBInfoDialog::ShowPassword ()
{
  PasswordField->setEchoMode(QLineEdit::Normal);
}


void  sEditDBInfoDialog::MaskPassword ()
{
  PasswordField->setEchoMode(QLineEdit::Password);
}


void  sEditDBInfoDialog::OnOK ()
{
  if( Validate() ){
    AcceptChanges();
    accept();
  }
}


bool  sEditDBInfoDialog::Validate ()
{
  //... analyse and transform DBName, etc.
  /*! @todo{SQL} temporary allow all values for DBName, etc.
  QString                     DBName = DBNameField->text();
  if( DBName.isEmpty() ){
    QMessageBox::information(this, tr("No database name given")
                            ,tr("Please fill up the name field"));
    DBNameField->setFocus();
    return false;
  }
  */
  return true;
}


void  sEditDBInfoDialog::AcceptChanges ()
{
  Info->Type = sDB_Info::eType( TypeField->currentIndex() );
  Info->DBName = FromQString( DBNameField->text() );
  Info->SchemaName = FromQString( SchemaField->text() );
  Info->TableNamePrefix = FromQString( PrefixField->text() );
  Info->Host = FromQString( HostField->text() );
  Info->Port = PortField->value();
  Info->Login = FromQString( LoginField->text() );
  Info->Password = FromQString( PasswordField->text() );
  Info->ConnectOptions = FromQString( OptionsField->text() );
}

//-------------------------------------------------------- sEditDBLinkDialog ---

sEditDBLinkDialog::sEditDBLinkDialog (psDB_Link  link
                                     ,QString    title
                                     ,QWidget *  parent)
    :sEditDBInfoDialog(link,title,parent)
    ,Link(link)
    ,IdDisplay(NULL)
{
  if( ! Link->Id.Undefined() ){
    IdDisplay = new QLabel(ToQString(Link->Id.ToString()
                                    +"    v."+Link->Version.Text()
                                    )
                          ,this);
  }

  //layout
  {
    int                       I = 0;
    QHBoxLayout *             L;

    if( IdDisplay ){
      (L = new QHBoxLayout())->setAlignment(Qt::AlignLeft);
      ((QVBoxLayout*)layout())->insertItem(I++,L);
      L->addWidget(new QLabel(tr("DB Id   "),this));
      L->addWidget(IdDisplay);
    }
  }
}

//------------------------------------------------------------------------------
} //namespace RA
