////////////////////////////////////////////////////////////////////////////////
/*! @file DeviceStates.cpp Basic classes for device state machine definition.
- Part of RAGUI - Research Assistant Graphical User Interface Library.
- Uses  Qt v.5.0.2  - http://qt-project.org/
- Uses  RANet - Research Assistant Net Library.
- Copyright(C) 2010-2014, Viktor E. Bursian, St.Petersburg, Russia.
                          Viktor.Bursian@mail.ioffe.ru
*///////////////////////////////////////////////////////////////////////////////
#include "DeviceStates.h"
#include "Driver.h"
#include "StringsAndQStrings.h"
#include "Log.h"
namespace RA {
//------------------------------------------------------------------------------

const QEvent::Type            sCommand::cDummyCommandEventType
                                  (QEvent::Type(QEvent::registerEventType()));

//csCommand                     sDeviceStateMachine::cResetCommand;

//------------------------------------------------------------------- sState ---
/*! @class sState

It is recommended instead of QState as an element of device's state machine.
It is a composite state with @b exclusive substates.  By default, it has one
child, being simultaneously the initial and the final state (see Fin()).

Further evolution of the state machine may (or may not) add other child states
to the sState instance (also changing the initial child state).  Otherwise (if
left without any additional substates) its behaviour is the same as that of the
simple (non-composite) QState object with two extra features:
-# it always emits the @a finished() signal (right after entering);
-# if a timeout value is set (either in constructor, or with
   @ref SetTimeout(...)), then the state will emit the Timeout() signal.

@note While a state of a device is often defined in the device's abstract base
class, its internal structure may appear later in a derivative.  The presense of
the final state garantees that one can use the @a finished() signal in the
implementation of the base class.

Also, there are some covenience functions, like @ref OnEnterCall(...),
which make the description of the device's state machine shorter and easier
for understanding.

@sa sParallelStateGroup.

*/


void  sState::Initiate (int      timeout_in_ms
                       ,QString  object_name  )
{
  setObjectName(object_name);
  if( TheFinalState )
    setInitialState(TheFinalState);
  connect(this,SIGNAL(finished()),this,SLOT(OnFinish()));
  TheTimeoutValue = timeout_in_ms;

  //!@bug Sometimes, after RA finishes there comes the following error
  //!     QCoreApplication::postEvent: Unexpected null receiver
  //!     Случается, когда на экране консоль sStepperFantastic317.
  //!     Или, когда она была открыта, но закрыта до выхода из RA
  //!     (причём, тоже выскакивает по завершении RA)
  //!     Is it because of this timer?
  //!     Why sState has no destructor?

  TimeoutClock = new QTimer(this);
  connect(TimeoutClock ,SIGNAL(timeout()) ,this , SIGNAL(Timeout())
          ,Qt::QueuedConnection );
}


void  sState::SetTimeout (int  timeout_in_ms)
{
  TheTimeoutValue = timeout_in_ms;
}


void  sState::SetTimeout (sPhysValue  timeout)
{
  TheTimeoutValue = Round( (timeout/sPhysValue(1,"ms")).ToReal() );
}


psState  sState::SetAsInitial ()
{
  if( ! parentState() ){
    DeviceSetInvalid(tr("Error while forming the state machine:\n"
                        "State '%1'->SetAsInitial()\n"
                        "State has no parent.")
                    .arg(objectName())  );
  }
  parentState()->setInitialState(this);
  return this;
}


psState  sState::OnEnterCall (QObject * object, literal slot)
{
  bool                  OK = false;
  psDeviceStateMachine  Machine (dynamic_cast<psDeviceStateMachine>(machine()));
  if( Machine ){
    OK = connect(this,SIGNAL(entered()),object,slot);
  }
  if( ! OK ){
    DeviceSetInvalid(tr("Error while forming the state machine:\n"
                        "%1->OnEnterCall(...,%2).")
                    .arg(objectName())
                    .arg(SoS_Text(slot))   );
  }
  return this;
}


psState  sState::OnFinishCall (QObject * object, literal slot)
{
  bool                  OK = false;
  psDeviceStateMachine  Machine (dynamic_cast<psDeviceStateMachine>(machine()));
  if( Machine ){
    OK = connect(this,SIGNAL(finished()),object,slot);
  }
  if( ! OK ){
    DeviceSetInvalid(tr("Error while forming the state machine:\n"
                        "%1->OnFinishCall(...,%2).")
                    .arg(objectName())
                    .arg(SoS_Text(slot))   );
  }
  return this;
}


psState  sState::OnExitCall (QObject * object, literal slot)
{
  bool                  OK = false;
  psDeviceStateMachine  Machine (dynamic_cast<psDeviceStateMachine>(machine()));
  if( Machine ){
    OK = connect(this,SIGNAL(exited()),object,slot);
  }
  if( ! OK ){
    DeviceSetInvalid(tr("Error while forming the state machine:\n"
                        "%1->OnExitCall(...,%2).")
                    .arg(objectName())
                    .arg(SoS_Text(slot))   );
  }
  return this;
}



/*! Adding a transition, triggered by the given signal.

@param sender    A signal emitting object, normally, a part of the device
                 (like a button, internal timer, etc.).
@param signal    Use SIGNAL macro in this place - the transition will
                 happen when @a sender emits the corresponding signal.
@param to_state  The target state.
@return          Returns @a this, which allows to add transitions and reactions
                 in a chain.

While @a signal may have parameters, they will be ignored.

*/
psState  sState::AddTransition (const QObject *   sender
                               ,const char *      signal
                               ,QAbstractState *  to_state)
{
  if( ! QState::addTransition(sender,signal,to_state) ){
    DeviceSetInvalid(tr("Error while forming the state machine:\n"
                        "%1->AddTransition(sender,%2,to_state).")
                    .arg(objectName())
                    .arg(SoS_Text(signal))  );
  }
  return this;
}



/*! @fn sState::AddTransition(const char *,QAbstractState *)

@brief Adding a transition, triggered by the given signal, emited by the device,
       to which this state belongs ("the hosting device").

@param signal    Use SIGNAL macro in this place - the transition will
                 happen when the hosting device emits the corresponding signal.
@param to_state  The target state.
@return          Returns @a this, which allows to add transitions and reactions
                 in a chain.

While @a signal may have parameters, they will be ignored.

*/



/*! Adding a transition, which happens, when this state is finished.

@param to_state  The target state.
@return          Returns @a this, which allows to add transitions and reactions
                 in a chain.

@note  Unlike QState::addTransition(QAbstractState*) the added transition
is NOT a conditionless transition, allthough it works in the same way for
the primitive sState instance (without any additional substates).

*/
psState  sState::AddTransition (QAbstractState *  to_state)
{
  if( TheOnFinishTransition ){
    removeTransition(TheOnFinishTransition);
    TheOnFinishTransition->deleteLater();
    TheOnFinishTransition = NULL;
    RANet::Log.Put(sLog::Debug,"States"
                ,sString("Substituting OnFinishTransition for state '")
                +FromQString(objectName())
                +sString("' in device '")+DeviceIdentifier()+"'.");
  }
  TheOnFinishTransition =
      QState::addTransition(this,SIGNAL(finished()),to_state);
  if( ! TheOnFinishTransition ){
    DeviceSetInvalid(tr("Error while forming the state machine:\n"
                        "%1->AddTransition(to_state).")
                    .arg(objectName())   );
  }
  return this;
}



/*! Removes the transition, which happens, when this state is finished.

@return          Returns @a this, which allows to add transitions and reactions
                 in a chain.

*/
psState  sState::RemoveOnFinishTransition ()
{
  if( TheOnFinishTransition ){
    removeTransition(TheOnFinishTransition);
    TheOnFinishTransition->deleteLater();
    TheOnFinishTransition = NULL;
  }
  return this;
}



/*! @fn  template<class data_type> psState sState::AddTransition(literal,QAbstractState *,data_type)

@brief Adding a transition, triggered by the given signal, but only if the
    argument of the signal is equal to the password.

@param signal_with_data  Use SIGNAL macro in this place - the transition will
                         happen when @b this device emits the corresponding
                         signal.
@param to_state        The target state.
@param password        An expression convertable to data_type type, to which
                       the signal's data should match in order the transition to
                       happen.
@return          Returns @a this, which allows to add transitions and reactions
                 in a chain.

The @a signal_with_data must have one parameter of the same
type, as the type of the parameter of the template.

Usage:
@code
    State_Operating()
        ->AddTransition(SIGNAL(Got(sTelex))
                       ,State_O_BeingOffended
                       ,sTelex("dura")  );
                   //the template parameter data_type will be taken from the
                   //type of the password
    //or
    State_Operating()
        ->AddTransition<sTelex>(SIGNAL(Got(sTelex))
                               ,State_O_BeingOffended
                               ,"dura"  );
                   //the template parameter data_type is given explicitly,
                   //the expression "dura" will be converted to sTelex type.
@endcode
*/



/*! Adding a transition, triggered by the given signal, with simultaneous call
    of the the given slot (or emission of another signal).

@param signal          Use SIGNAL macro in this place - the transition will
                       happen when @b this device emits the corresponding
                       signal.
@param slot_or_signal  Use SLOT (or SIGNAL) macro in this place - the
                       corresponding slot of @b this device will be called
                       (or the signal emited).
@param to_state        The target state.
@return          Returns @a this, which allows to add transitions and reactions
                 in a chain.

While @a signal may have parameters, they will be ignored.
The @a slot_or_signal should have no parameters. If it has one parameter, use
template<class data_type>AddTransition(literal,literal,QAbstractState*).

*/
psState  sState::AddTransition (const char *      signal
                               ,const char *      slot_or_signal
                               ,QAbstractState *  to_state)
{
  QSignalTransition *     T (new QSignalTransition(DeviceObject(),signal));
  T->setTargetState(to_state);
  QString                 SlotDataType( SoS_Type(slot_or_signal) );
  if( SlotDataType.isEmpty() ){
    connect(T ,SIGNAL(triggered()) ,DeviceObject() ,slot_or_signal);
  }else{
    DeviceSetInvalid(tr("AddTransition on %1 for the device %2 "
                        "won't invoke %3. "
                        "Use AddTransition<%4>(...) instead.")
                    .arg(SoS_Text(signal))
                    .arg(ToQString(DeviceIdentifier()))
                    .arg(SoS_Text(slot_or_signal))
                    .arg(SlotDataType)
                    );
  }
  addTransition(T);
  return this;
}



/*! @fn  template<class data_type> psState sState::AddTransition(literal,literal,QAbstractState *)

@brief Adding a transition, triggered by the given signal, with simultaneous
       call of the the given slot (or emission of another signal).

@param signal_with_data  Use SIGNAL macro in this place - the transition will
                         happen when @b this device emits the corresponding
                         signal.
@param slot_or_signal_with_data  Use SLOT (or SIGNAL) macro in this place - the
                       corresponding slot of @b this device will be called
                       (or the signal emited).
@param to_state        The target state.
@return          Returns @a this, which allows to add transitions and reactions
                 in a chain.

The @a signal_with_data and the @a slot_or_signal_with_data must have one
parameter of the same type, with this type being also the parameter of the
template.

Usage:
@code
  //evolution of the inherited state machine (behaviour)
  {
    //evolution of the State_Operating
    {
      State_O_A
          ->AddTransition<sTelex>(SIGNAL(Got(sTelex))
                                 ,SLOT(Process(sTelex))
                                 ,State_O_B);
    };
  };
@endcode

@note The compiler cannot determine the type of the template from the arguments
of SIGNAL(...) or SLOT(...) macros.  Therefore, it has to be given directly.

@sa  AddTransition(const char *,const char *,QAbstractState*)
*/



/*! Adding a transition, triggered by the given command.

@param command         Use reference to a static csCommand object, declared in
                       the declaration of the hosting device's class (or in its
                       ancestor).
@param to_state        The target state.

Usage:
@code
    State_O_Standby->AddTransition(cMove ,State_O_Moving);
@endcode

*/
psState  sState::AddTransition (rcsCommand        command
                               ,QAbstractState *  to_state)
{
  psCommandTransition     T (new sCommandTransition(HostDriver(),command));
  T->setTargetState(to_state);
  addTransition(T);
  return this;
}



/*! Adding a transition, triggered by the given command, with simultaneous call
    of the the given slot (or emission of the signal).

@param command         Use reference to a static csCommand object, declared in
                       the declaration @b this device's class (or in its
                       ancestor).
@param slot_or_signal  Use SLOT (or SIGNAL) macro in this place - the
                       corresponding slot of @b this device will be called
                       (or the signal emited).
@param to_state        The target state.

The @a slot_or_signal should have no parameters.

*/
psState  sState::AddTransition (rcsCommand        command
                               ,literal           slot_or_signal
                               ,QAbstractState *  to_state)
{
  psCommandTransition   T (new sCommandTransition(HostDriver(),command
                                                 ,slot_or_signal));
  T->setTargetState(to_state);
  addTransition(T);
  return this;
}


/*! @fn template<class data_type> psState sState::AddTransition(const sDirective<data_type> &,QAbstractState *)

@brief Adding a transition, triggered by the given directive.

@param directive       Use reference to a static const sDirective object,
                       declared in the submissive device's class (or in its
                       ancestor).
@param to_state        The target state.

@note The directive usually carries data (given as the 3d argument when calling
SendDirective<...>(...) method.  In this case, the transition will happen, but
the attached data will be lost.  In order to process the attached data in this
driver, one should use another template function:
AddTransition<...>(from_state,directive,slot_or_signal_with_data,to_state)

Usage:
@code
    State_Operating()->AddTransition(cCheat,State_O_Cheating);
@endcode

@sa template<class data_type> void sDriver::AddTransition(psState,const sDirective<data_type> &,literal,QAbstractState *)

*/



/*! @fn  template<class data_type> psState sState::AddTransition(const sDirective<data_type> &,literal,QAbstractState *)

@brief Adding a transition, triggered by the given directive, with simultaneous
       call of the the given slot (or emission of another signal).

@param directive       Use reference to a static const sDirective object,
                       declared in the submissive device's class (or in its
                       ancestor).
@param slot_or_signal_with_data    Use SLOT (or SIGNAL) macro in this place - the
                       corresponding slot of @b this device will be called
                       (or the signal emited).  The data carried by the
                       directive will be passed as its argument.
@param to_state        The target state.


The @a slot_or_signal_with_data must have one parameter of the same type, as
the type of the parameter of the template.

Usage:
@code
    State_O_Measuring
        ->AddTransition(cCheat
                       ,SIGNAL(NewData(sPhysValue))
                       ,State_O_ConfessingInCheating);
@endcode

*/



/*! Adding state-depended reaction to a device state machine: whenever the
    device will emit the given signal, a slot will be called (or another signal
    emited), but only if the device is in a certain state.

@param signal          Use SIGNAL macro in this place - the reaction will happen
                       when @b this device emits the corresponding signal.
@param slot_or_signal  Use SLOT (or SIGNAL) macro in this place - the
                       corresponding slot of @b this device will be called
                       (or the signal emited).

While @a signal may have parameters, they will be ignored.
The @a slot_or_signal should have no parameters. If it has one parameter, use
@a AddReaction<...>(...).

@sa template<class data_type>AddReaction(literal,literal).

*/
psState  sState::AddReaction (const char *  signal
                             ,const char *  slot_or_signal)
{
  if( ! SoS_Type(slot_or_signal).isEmpty() ){
    DeviceSetInvalid(tr("AddReaction on %1 for the device %2 "
                        "won't invoke %3. "
                        "Use AddReaction<%4>(...) instead.")
                    .arg(SoS_Text(signal))
                    .arg(ToQString(DeviceIdentifier()))
                    .arg(SoS_Text(slot_or_signal))
                    .arg(SoS_Type(slot_or_signal))
                    );
  }else{
    QSignalTransition *     T (new QSignalTransition(DeviceObject(),signal));
    connect(T ,SIGNAL(triggered()) ,DeviceObject() ,slot_or_signal);
    addTransition(T);
  }
  return this;
}


/*! @fn  template<class data_type> psState sState::AddReaction(literal,literal)

@brief Adding state-depended reaction to a device state machine: whenever the
    device will emit the given signal, a slot will be called (or another signal
    emited), but only if the device is in a certain state.

@param signal_with_data  Use SIGNAL macro in this place - the reaction will
                         happen when @b this device emits the corresponding
                         signal.
@param slot_or_signal_with_data  Use SLOT (or SIGNAL) macro in this place - the
                       corresponding slot of @b this device will be called
                       (or the signal emited).

The @a signal_with_data and the @a slot_or_signal_with_data must have one
parameter of the same type, with this type being also the parameter of the
template.

@sa AddReaction(const char *,const char *).

*/



/*! @fn template<class data_type> psState sState::AddReaction(const char *,data_type,const char *)

@brief Adding state-depended reaction to a device state machine: whenever the
    device will emit the given signal, a slot will be called (or another signal
    emited), but only if the device is in a certain state and the argument of
    the signal is equal to the password.

@param signal_with_data  Use SIGNAL macro in this place - the reaction will
                         happen when @b this device emits the corresponding
                         signal.
@param password        An expression convertable to data_type type, to which
                       the signal's data should match in order the reaction to
                       happen.
@param slot_or_signal  Use SLOT (or SIGNAL) macro in this place - the
                       corresponding slot of @b this device will be called
                       (or the signal emited).

Usage:
@code
    State_Operating()
        ->AddReaction(SIGNAL(Got(sTelex)) ,sTelex("dura")
                     ,SLOT(SaySamDurak()) );
                              //suitable template instantiation will be done,
                              //in according with the type of the 3d argument
      //or
    State_Operating()
        ->AddReaction<sTelex>(SIGNAL(Got(sTelex)) ,"dura"
                             ,SLOT(SaySamDurak()) );
                                   //the template instantiation is defined by
                                   //the argument of the template, therefore
                                   //the 3d argument of the function will be
                                   //converted to sTelex type.
@endcode
*/



/*! Adding state-depended reaction to a device state machine: whenever the
    device will get the given command, a slot will be called (or another signal
    emited), but only if the device is in a certain state.

@param command         Use a static const sCommand member of @b this device's
                       class (or its ancestor).
@param slot_or_signal  Use SLOT (or SIGNAL) macro in this place - the
                       corresponding slot of @b this device will be called
                       (or the signal emited).

The @a slot_or_signal should have no parameters.

*/
psState  sState::AddReaction (rcsCommand  command
                             ,literal     slot_or_signal)
{
  psCommandTransition   T (new sCommandTransition(HostDriver(),command
                                                 ,slot_or_signal));
  addTransition(T);
  return this;
}



/*! @fn template<class data_type> psState sState::AddReaction(const sDirective<data_type> &,literal)

@brief  Adding state-depended reaction to a device state machine: whenever the
    device will get the given directive, a slot will be called (or another
    signal emited), but only if the device is in a certain state.

@param directive       Use a static const sDirective<data_type> member of
                       the hosting device's class (or its ancestor).
@param slot_or_signal_with_data  Use SLOT (or SIGNAL) macro in this place -
                       the corresponding slot of @b this device will be called
                       (or the signal emited).

The @a slot_or_signal should have one parameter of the @a data_type type.
The data attached to the directive will be passed to it.

*/



void  sState::onEntry (QEvent *  event)
{
  if( ! objectName().isEmpty() ){
    psDeviceStateMachine          Machine (dynamic_cast<psDeviceStateMachine>
                                                                   (machine()));
    if( Machine && Machine->Device()->TraceStates() ){
      RANet::Log.Put(sLog::Debug,"States"
                    ,sString("Device '")+Machine->Device()->DeviceIdentifier()
                    +"' is entering into "+FromQString(objectName()));
    }
  }
  QState::onEntry(event);
  if( TheTimeoutValue > 0 ){
    TimeoutClock->setSingleShot(true);
    TimeoutClock->start(TheTimeoutValue);
  }
}


void  sState::onExit (QEvent *  event)
{
  TimeoutClock->stop();
  QState::onExit(event);
  if( ! objectName().isEmpty() ){
    psDeviceStateMachine          Machine (dynamic_cast<psDeviceStateMachine>
                                                                   (machine()));
    if( Machine && Machine->Device()->TraceStates() ){
      RANet::Log.Put(sLog::Debug,"States"
                    ,sString("Device '")+Machine->Device()->DeviceIdentifier()
                    +"' is exiting from "+FromQString(objectName()));
    }
  }
}


void sState::OnFinish()
{
  if( ! objectName().isEmpty() && (initialState() != Fin()) ){
    psDeviceStateMachine          Machine (dynamic_cast<psDeviceStateMachine>
                                                                   (machine()));
    if( Machine && Machine->Device()->TraceStates() ){
      RANet::Log.Put(sLog::Debug,"States"
                    ,sString("Device '")+Machine->Device()->DeviceIdentifier()
                    +"' is finishing "+FromQString(objectName()));
    }
  }
}


QString  sState::SoS_Name (const char * signal_or_slot)
{
  QString   N ( QString::fromLatin1(signal_or_slot) );
  int       P ( N.indexOf("(") );
  if( (P > 0) && (N.startsWith("1") || N.startsWith("2")) ){
    N.truncate(P);
    N.remove(0,1);
    return N;
  };
  qDebug()<<"Parsing name error: signal_or_slot='"
          <<QString::fromLatin1(signal_or_slot)<<"'.";
  return QString();
}


QString  sState::SoS_Type (const char * signal_or_slot)
{
  QString   T ( QString::fromLatin1(signal_or_slot) );
  int       P ( T.indexOf("(") );
  if( (P > 0) && (T.endsWith(")")) ){
    T = T.mid(P+1,T.length()-P-2);
    return T;
  }
  qDebug()<<"Parsing data typename error: signal_or_slot='"
          <<QString::fromLatin1(signal_or_slot)<<"'.";
  return QString();
}


QString  sState::SoS_Text (const char * signal_or_slot)
{
  QString   N ( QString::fromLatin1(signal_or_slot) );
  if( N.startsWith("1") ){
    return QString("SLOT(")+SoS_Name(signal_or_slot)
                       +"("+SoS_Type(signal_or_slot)+"))";
  }else if( N.startsWith("2") ){
    return QString("SIGNAL(")+SoS_Name(signal_or_slot)
                       +"("+SoS_Type(signal_or_slot)+"))";
  };
  qDebug()<<"Parsing text error: signal_or_slot='"
          <<QString::fromLatin1(signal_or_slot)<<"'.";
  return QString();
}


psDriver  sState::HostDriver ()
{
  psDeviceStateMachine  Machine (dynamic_cast<psDeviceStateMachine>(machine()));
  if( Machine ){
    psDriver   D (dynamic_cast<psDriver>(Machine->Device()));
    if( D )
      return D;
  }
  DeviceSetInvalid(tr("Error while forming the state machine:\n"
                      "State '%1'doesn't belong to a sDeviceStateMachine.")
                  .arg(objectName())  );
  return NULL;
}


QObject *  sState::DeviceObject ()
{
  psDeviceStateMachine  Machine (dynamic_cast<psDeviceStateMachine>(machine()));
  if( Machine ){
    QObject *   D (dynamic_cast<QObject *>(Machine->Device()));
    if( D )
      return D;
  }
  DeviceSetInvalid(tr("Error while forming the state machine:\n"
                      "State '%1'doesn't belong to a sDeviceStateMachine.")
                  .arg(objectName())  );
  return NULL;
}


sString  sState::DeviceIdentifier ()
{
  psDeviceStateMachine  Machine (dynamic_cast<psDeviceStateMachine>(machine()));
  if( Machine ){
    return Machine->Device()->DeviceIdentifier();
  }
  DeviceSetInvalid(tr("Error while forming the state machine:\n"
                      "State '%1'doesn't belong to a sDeviceStateMachine.")
                  .arg(objectName())  );
  return sString();
}


void  sState::DeviceSetInvalid (QString  reason)
{
  psDeviceStateMachine  Machine (dynamic_cast<psDeviceStateMachine>(machine()));
  if( Machine ){
    Machine->Device()->SetInvalid(reason);
  }
}


//----------------------------------------------- sState::sCommandTransition ---
/*! @class sState::sCommandTransition

@brief Transition on command.

Не срабатывает, если драйвер взят под исключительное управление другим
устройством (то есть не тем, чья команда поступила).

@note
Часть функций этого класса ему не нужны, но нужны производным от него
sDirectiveTransition<class data_type>.  Эта кривость сделана потому, что я не
научился делать шаблоны-члены не полностью inline.

*/


bool  sState::sCommandTransition::eventTest (QEvent * event)
{
  if( event->type() == Identifier ){
    sState::psCommandEvent      Command = static_cast<sState::psCommandEvent>
                                                       (event);
    if(   (Command->ControlChange == sDevice::cReleaseControl)
        &&(HostDriver->TheExclusiveMaster == Command->Sender) ){
      HostDriver->TheExclusiveMaster = NULL;
    }
    return true;
  }
  return false;
}


void  sState::sCommandTransition::onTransition (QEvent *)
{
  if( IvokesSlotOrSignal ){
    QMetaObject::invokeMethod((QObject*)HostDriver
                             ,SlotOrSignal.toLatin1().constData()
                             ,Qt::DirectConnection              );
  };
}


//------------------------------------------------------ sDeviceStateMachine ---

sDeviceStateMachine::sDeviceStateMachine (psDevice  host_device)
    :QStateMachine()
    ,Warming(new sState(this,"State_Warming"))
    ,Operating(new sState(this,"State_Operating"))
    ,Cooling(new sState(this,"State_Cooling"))
    ,Frozen(new sState(this,"State_Frozen"))
    ,FinalState(new QFinalState(this))
    ,TheHostDevice(host_device)
{
  //----- State_Warming -----
  Warming
      ->AddTransition(Operating);

  //----- State_Operating -----
  connect(Operating ,SIGNAL(entered())
         ,this      ,SIGNAL(Warmed())
         ,Qt::QueuedConnection);//QueuedConnection is necessary here.
  //reaction on WarmUp signal in State_Operating
  {
    EmitWarmedAndReturn = new QSignalTransition(this,SIGNAL(WarmUp()));
    connect(EmitWarmedAndReturn,SIGNAL(triggered())
           ,this               ,SIGNAL(Warmed())
           ,Qt::QueuedConnection);//QueuedConnection is necessary here.
    Operating->addTransition(EmitWarmedAndReturn);
  }

  //----- State_Cooling -----
  Cooling->AddTransition(Frozen);

  //----- State_Frozen -----
  connect(Frozen ,SIGNAL(entered())
         ,this   ,SIGNAL(Cooled())
         ,Qt::QueuedConnection);//QueuedConnection is necessary here.
  Frozen
        ->AddTransition(this ,SIGNAL(WarmUp()) ,Warming );

  //generating NewUpperLevelState signals (kind of signal mapping)
  {
    Warming->OnEnterCall(this ,SLOT(Into_Warming()));
    Operating->OnEnterCall(this ,SLOT(Into_Operating()));
    Cooling->OnEnterCall(this ,SLOT(Into_Cooling()));
    Frozen->OnEnterCall(this ,SLOT(Into_Frozen()));
  }
}


void sDeviceStateMachine::Into_Warming()
{
  emit NewUpperLevelState(cWarming);
  emit NewUpperLevelState("Warming");
}

void sDeviceStateMachine::Into_Operating()
{
  emit NewUpperLevelState(cOperating);
  emit NewUpperLevelState("Operating");
}

void sDeviceStateMachine::Into_Cooling()
{
  emit NewUpperLevelState(cCooling);
  emit NewUpperLevelState("Cooling");
}

void sDeviceStateMachine::Into_Frozen()
{
  emit NewUpperLevelState(cFrozen);
  emit NewUpperLevelState("Frozen");
}

//------------------------------------------------------------------------------
} //namespace RA



