////////////////////////////////////////////////////////////////////////////////
/*! @file Driver.cpp   Драйвер устройства (части экспериментальной установки).
- Part of RAGUI - Research Assistant Graphical User Interface Library.
- Uses  Qt v.5.0.2  - http://qt-project.org/
- Uses  RANet - Research Assistant Net Library.
- Copyright(C) 2010-2014, Viktor E. Bursian, St.Petersburg, Russia.
                          Viktor.Bursian@mail.ioffe.ru
*///////////////////////////////////////////////////////////////////////////////
#include "Driver.h"
#include "BasicNodes.h"
#include "StringsAndQStrings.h"
#include "Log.h"
namespace RA {
//------------------------------------------------------------------------------
//--------------------------------------------------------------- sDriverTag ---

sDriverTag::sDriverTag (literal         classname
                       ,pfPreconfigure  preconf
                       ,pfNewDriver     createfunc)
    :sDeviceTag(classname,cDriver,preconf)
    ,TheNewDriverFunctionPtr(createfunc)
{
}


//------------------------------------------------------------------ sDriver ---
/*! @class sDriver

@par The state machine
@dotfile "Driver.state_graph"

Legend: 
- thick solid black arrow - unconditional transition (when the state finishes);
- thin solid black arrow - transition on signal or command or directive;
- dotted black circular arrow - reaction on signal or command or directive
  (place the mouse over the line to bring up a hint);
- dashed black arrow - an alternative transition
  (place the mouse over the line to bring up a hint);
- dashed violet arrow - an additional transition, which appears later, when
  the state machine is working
  (place the mouse over the line to bring up a hint).
  
*/


void  sDriver::Preconfigure (sNodePtr  DeviceCfg)
{
  sDevice::Preconfigure(DeviceCfg);
}


sDriver::~sDriver ()
{
  if( TraceStates() )
    RANet::Log.Put(sLog::Debug,"States"
                  ,DeviceIdentifier()+" destructor");
}



/*! Constructor.
*/
sDriver::sDriver (sString   DeviceId
                 ,sNodePtr  DeviceCfg)
    :sDeviceStateMachine(this)
    ,sDevice(DeviceId,DeviceCfg)
    ,TheExclusiveMaster(NULL)
    ,SuicidingFlag(false)
{
}


/*! Called by the core right after constructor 
    (before the start of the state machine).

The purpose of this method is to use virtual methods in the initiation,
which does not work within constructors.

@note
Always call the ancestor's PostConstructor() at the beginning of this function,
if you reimplement it.

*/
void  sDriver::PostConstructor ()
{
  connect(this ,SIGNAL(started())
         ,this ,SLOT(AfterStartInitiation()));
  State_Warming()
      ->SetAsInitial()
      ->OnEnterCall(SLOT(OnEnterToWarmingState()))
      ->AddTransition( SIGNAL(CoolDown()) ,State_Cooling() )
      ->OnExitCall(SLOT(OnExitFromWarmingState()));
  State_Operating()
      ->OnEnterCall(SLOT(OnEnterToOperatingState()))
      ->AddTransition( SIGNAL(CoolDown()) ,State_Cooling() )
      ->OnExitCall(SLOT(OnExitFromOperatingState()));
  State_Cooling()
      ->OnEnterCall(SLOT(OnEnterToCoolingState()))
      ->OnExitCall(SLOT(OnExitFromCoolingState()));
  State_Frozen()
        ->AddTransition(this ,SIGNAL(CoolDown()) ,State_Frozen() );

  connect(this,SIGNAL(finished()),this,SLOT(Harakiri()));
}



/*! Connects a signal from a submissive driver with a slot (or signal)
    of this driver.

@param sender          Link to the submissive driver.  Should be set to a 
                       certain device before the call of this function, 
                       otherwise the function does nothing and returns @a false.
@param signal          Signal of that driver.  Use SIGNAL macro in this place.
@param slot_or_signal  Slot or signal of @b this driver.  Use SLOT (or SIGNAL)
                       macro here.

The @a signal and the @a slot_or_signal parameters may have data 
(should match each other).

The connection will be @a Qt::QueuedConnection.

@sa Disconnect(rcsDriverLink,literal,literal).
*/
bool  sDriver::Connect (rcsDriverLink  sender
                       ,literal        signal
                       ,literal        slot_or_signal)
{
  sString                     Signal (signal);
  sString                     SlotOrSignal (slot_or_signal);
  if( Signal[0] == '1' ){
    throw xProgramError(sString("sDriver::Connect: not a signal: ")+signal
                       ,__FILE__,__LINE__);
  }else{
    Signal = sString("SIGNAL(")+Signal.Tail(1)+")";
  };
  if( SlotOrSignal[0] == '1' ){
    SlotOrSignal = sString("SLOT(")+SlotOrSignal.Tail(1)+")";
  }else{
    SlotOrSignal = sString("SIGNAL(")+SlotOrSignal.Tail(1)+")";
  };
  if( ! sender.IsCorrect() )
    return false;
  if( TraceStates() || sender->TraceStates() )
    RANet::Log.Put(sLog::Debug,"States"
                  ,sString("Connecting: '") + sender->DeviceIdentifier()
                  +sString("'.'") + Signal
                  +sString("' --> '") + DeviceIdentifier()
                  +sString("'.'") + SlotOrSignal + "'");
  if( ! QObject::connect(sender.operator ->() ,signal
                        ,this                 ,slot_or_signal
                        ,Qt::QueuedConnection) ){
    RANet::Log.Put(sLog::ProgramError,"States"
                  ,sString("Connect failed: '") + sender->DeviceIdentifier()
                  +sString("'.'") + Signal
                  +sString("' --> '") + DeviceIdentifier()
                  +sString("'.'") + SlotOrSignal + "'");
    return false;
  };    
  return true;
}



/*! Disconnects a signal from a submissive driver from a slot (or signal)
    of this driver.

@param sender          Link to the submissive driver.  Should be set to a 
                       certain device before the call of this function, 
                       otherwise the function does nothing and returns @a false.
@param signal          Signal of that driver.  Use SIGNAL macro in this place.
@param slot_or_signal  Slot or signal of @b this driver.  Use SLOT (or SIGNAL)
                       macro here.

The 2nd and/or 3rd argument can be 0 (or NULL) - meaning "all".

@sa Connect(rcsDriverLink,literal,literal).
*/
bool  sDriver::Disconnect (rcsDriverLink  sender
                          ,literal        signal
                          ,literal        slot_or_signal)
{
  if( ! sender.IsCorrect() )
    return false;
  return QObject::disconnect(sender.operator ->() ,signal
                            ,this                 ,slot_or_signal);
}



bool  sDriver::GetCommand (sState::psCommandEvent  command_event)
{
  if( TraceStates() && ! command_event->NameForDebugging.Empty() )
    RANet::Log.Put(sLog::Debug,"States"
                ,sString("Device '")+DeviceIdentifier()
                +"' gets "+command_event->NameForDebugging
                +" from '"+command_event->Sender->DeviceIdentifier()+"'");
                
  if(    (TheExclusiveMaster == NULL)
      || (TheExclusiveMaster == command_event->Sender) ){
    if( command_event->ControlChange == cTakeControl ){
      TheExclusiveMaster = command_event->Sender;
    };
    postEvent(command_event);
    return true;
  };
  return false;
}



/*! Sends a command to a submissive driver.

@param receiver        Link to the submissive driver.  Should be set to a 
                       certain device before the call of this function, 
                       otherwise the function does nothing and returns @a false.
@param command         Use reference to a static csCommand object, declared in
                       the submissive device's class (or in its ancestor).
@param control_change  Optional: take or release an exclusive control over the
                       submissive device.
@return @a true, if succeed.

This function posts the command event into the driver's state machine queue, if
the device is not under exclusive control by another device.  Otherwise, the
function returns @a false and the command is ignored (it won't wait until the
other device releases control).
Simultaneously, if @a control_change==cTakeControl it takes exclusive control 
over the @a receiver, or releases it, if @a control_change==cReleaseControl.
  
Usage:
@code
    if( ! SendCommand(Gauge ,sGauge::cStart ,cTakeControl) ){
      //command had not been sent
    };
@endcode

@sa SendDirective<...>(...).

*/
bool  sDriver::SendCommand (rcsDriverLink   receiver
                           ,rcsCommand      command
                           ,eControlChange  control_change)
{
  if( receiver.IsCorrect() ){
    return receiver->GetCommand(new sState::sCommandEvent(this,command
                                                         ,control_change));
  };
  return false;
}



/*! @fn template<class data_type> void sDriver::SendDirective(rcsDriverLink,const sDirective<data_type> &,data_type,eControlChange)

@brief Sends a directive to a submissive driver 
       (same as SendCommand(...), but a directive carries data with it).

@param receiver        Link to the submissive driver.  Should be set to a 
                       certain device before the call of this function, 
                       otherwise the function does nothing and returns @a false.
@param directive       Use reference to a static const sDirective object, 
                       declared in the submissive device's class (or in its 
                       ancestor).
@param data            An expression of the @a data_type type.  It will be sent
                       with the directive as the attached data.
@param control_change  Optional: take or release an exclusive control over the
                       submissive device.
@return @a true, if succeed.

This function posts the directive event into the driver's state machine queue, 
if the device is not under exclusive control by another device.  Otherwise, the
function returns @a false and the command is ignored (it won't wait until the
other device releases control).
Simultaneously, if @a control_change==cTakeControl it takes exclusive control 
over the @a receiver, or releases it, if @a control_change==cReleaseControl.
  
Usage:
@code
    if( ! SendDirective(Gauge ,sGauge::cSetPeriod, sPhysValue(0.6,"s")
                       ,cTakeControl) ){
      //directive had not been sent
    };
@endcode

@sa SendCommand(...).

*/


void  sDriver::Suicide ()
{
  if( TraceStates() )
    RANet::Log.Put(sLog::Debug,"States"
                  ,sString("Suicide: '") + DeviceIdentifier() + "'");
  SuicidingFlag = true;
  State_Frozen()
      ->AddTransition(State_Final());
  emit CoolDown();
}


void sDriver::Harakiri()
{
  if( TraceStates() )
    RANet::Log.Put(sLog::Debug,"States"
                  ,sString("Harakiri: '") + DeviceIdentifier() + "'");
  deleteLater();
}



/*! @fn void sDriver::AfterStartInitiation()
 
@brief Called when the state machine starts
       (before entering its initial state).

In sDriver class this function does nothing, but you can reimplement it in a
descendant. 
*/



/*! @fn void sDriver::OnEnterToWarmingState()
 
@brief Called by the core on enter into the State_Warming().

Added mainly for the completeness. In sDriver class this function does nothing,
but you can reimplement it in a descendant. 
*/



/*! @fn void sDriver::OnExitFromWarmingState()
 
@brief Called by the core on exit from the State_Warming(). 

Added mainly for the completeness. In sDriver class this function does nothing,
but you can reimplement it in a descendant. 
*/



/*! @fn void sDriver::OnEnterToOperatingState()

@brief Called by the core on enter into the State_Operating().  

In sDriver class this function does nothing, but you can reimplement it in a
descendant. Usually, you should connect the signals of the submissive drivers 
with the signals (or slots) of this device:
@code
  Connect(NewDevice1 ,SIGNAL(NewData(sPhysValue)),SIGNAL(ShowData(sPhysValue)));
@endcode
Note, this could not be done, say, in PostConstructor(), before the link to
the submissive driver got its value.

Note, there is another way to do something on enter into State_Operating(),
namelly, you can add 
@code
    State_Operating()->OnEnterCall(...);
@endcode
into the PostConstructor().  This will call your slot after all actions,
made by the ancestor's class.  Here, however, you can do something BEFORE
the ancestor's actions.

@sa OnExitFromOperatingState().
*/



/*! @fn void sDriver::OnExitFromOperatingState()

@brief Called by the core on exit from the State_Operating().  

In sDriver class this function does nothing, but you can reimplement it in a
descendant. Usually, you should disconnect the signals of the submissive 
drivers, which were connected on enter into this state:
@code
  Disconnect(NewDevice1 ,SIGNAL(NewData(sPhysValue)) 
            ,SIGNAL(ShowData(sPhysValue)));
@endcode

Note, there is another way to do something on exit from State_Operating(),
namelly, you can add 
@code
    State_Operating()->OnExitCall(...);
@endcode
into the PostConstructor().  This will call your slot after all actions,
made by the ancestor's class.  Here, however, you can do something BEFORE
the ancestor's actions.

@sa OnEnterToCoolingState().
*/



/*! @fn void sDriver::OnEnterToCoolingState()

@brief Called by the core on enter into the State_Cooling().

In sDriver class this function does nothing, but you can reimplement it in a
descendant. Usually, you should disconnect all submissive drivers, like this:
@code
  Disconnect(NewDevice1 ,0 ,0);
  NewDevice1.ReleaseExclusiveControl();
  NewDevice1.SetToNULL();
@endcode

Note, there is another way to do something on enter into State_Cooling(),
namelly, you can add 
@code
    State_Cooling()->OnEnterCall(...);
@endcode
into the PostConstructor().  This will call your slot after all actions,
made by the ancestor's class.  Here, however, you can do something BEFORE
the ancestor's actions.

@sa OnExitFromOperatingState().
*/



/*! @fn void sDriver::OnExitFromCoolingState()

@brief Called by the core on exit from the State_Cooling().  

Added mainly for the completeness. In sDriver class this function does nothing,
but you can reimplement it in a descendant. 
*/

//------------------------------------------------------------------------------
} //namespace RA



