////////////////////////////////////////////////////////////////////////////////
/*! @file Driver.h   Драйвер устройства (части экспериментальной установки).
- Part of RAGUI - Research Assistant Graphical User Interface Library.
- Uses  Qt v.5.0.2  - http://qt-project.org/
- Uses  RANet - Research Assistant Net Library.
- Copyright(C) 2010-2014, Viktor E. Bursian, St.Petersburg, Russia.
                          Viktor.Bursian@mail.ioffe.ru
*///////////////////////////////////////////////////////////////////////////////
#ifndef Driver_H
#define Driver_H
#include "DriverLink.h"
#include "DeviceStates.h"
namespace RA {
//------------------------------------------------------------------------------

ANNOUNCE_CLASS(sDriverTag)
ANNOUNCE_ABSTRACT_CLASS(sDriver)

//--------------------------------------------------------------- sDriverTag ---

class RAGUI_EXPORT  sDriverTag : public sDeviceTag
{
  private: // types
    typedef psDriver          fNewDriver (sString,sNodePtr);
    typedef fNewDriver *      pfNewDriver;

  protected:
                              sDriverTag (literal         classname
                                         ,pfPreconfigure  preconf
                                         ,pfNewDriver     createfunc);

  private: // fields
    pfNewDriver               TheNewDriverFunctionPtr;

  friend class sDriverLink;
};


//-------------------------------------------------------- DRIVER(CLASSNAME) ---
/*! Этот макрос должен стоять в начале декларации @b не @b абстрактного класса,
    публично порождённого от RA::sDriver.

@note @b Примечание: Абстрактные классы, публично порождённые от RA::sDriver,
вместо макроса DRIVER используют макрос ABSTRACT_DRIVER().

Декларирует protected конструктор с параметрами
(RA::sString DeviceId ,RA::sNodePtr DeviceCfg).
Он @b должен @b быть @b определён в соответствующем cpp-файле.
Смысл параметров:
@arg @c DeviceId задаваемый в конфигурации уникальный идентификатор
                 реально существующего "в железе" экземпляра устройства
@arg @c DeviceCfg специального вида указатель на узел сети данных,
                  содержащий параметры устройства и, возможно, состояние
                  в котором оно было оставлено после последнего использования
                  (например, текущая длина волны для спектрометра с шаговым
                  управлением развёрткой, но без датчика угол-код)

Также декларирует статическую функцию-член Preconfigure, цель которой
сформировать умолчательный кластер конфигурации устройства.  В соответствующем
cpp-файле необходимо определить эту функцию.

И наконец, декларирует некий статический член - тэг класса, обеспечивающий
динамическое создание объектов класса (т.е. экземпляров устройств), их поиск
и т.п.  В соответствующем cpp-файле необходимо определить этот тэг при помощи
макро DEFINE_CLASS_TAG() с именем класса в качестве аргумента.

Поскольку класс sDriver является наследником
класса QObject, в декларации порождаемых от sDriver классов следует ставить
ещё и макро Q_OBJECT (причём, в @b самом начале).

Использование:
@code
class sMyDriver : public sDriver
{
  Q_OBJECT
  DRIVER(sMyDriver)
  //...в классе не должно быть чистых (pure) виртуальных методов
};
@endcode

@macro
*/
#define  DRIVER(CLASSNAME)                                                     \
  DECLARE_DRIVER_PTR(CLASSNAME)                                                \
  private:                                                                     \
    class sIntrinsicTag : public sDriverTag                                    \
    {                                                                          \
      public:                                                                  \
                                  sIntrinsicTag ()                             \
                                      :sDriverTag(#CLASSNAME                   \
                                                 ,Preconfigure,NewDriver)      \
                                    {};                                        \
        virtual bool              CanCast (psAdam P) const                     \
                                    { return bool(                             \
                                          dynamic_cast<p##CLASSNAME>(P)        \
                                                 );                            \
                                    };                                         \
    };                                                                         \
    static sIntrinsicTag      AdamInheritorTag;                                \
  public:                                                                      \
    static void               Preconfigure (sNodePtr  DevCfg);                 \
  protected:                                                                   \
                              CLASSNAME (sString,sNodePtr);                    \
  private:                                                                     \
    static psDriver           NewDriver (sString   Id  ,sNodePtr  DevCfg)      \
                                { return new CLASSNAME(Id,DevCfg); };          \
  private:



//----------------------------------------------- ABSTRACT_DRIVER(CLASSNAME) ---
/*! Этот макрос должен стоять в начале декларации @b абстрактного класса,
    публично порождённого от sDriver.

@note @b Примечание: Термин "абстрактный" здесь несколько расширен по сравнению
с соответствующим термином из C++. в классе могут быть чистые (pure)
виртуальные методы а могут и не быть, но класс по замыслу абстрактен, т.е.
объекты этого класса никогда не создаются. Невозможность создания гарантируется
приватностью конструктора и специфическим механизмом порождения драйверов.

Декларирует protected конструктор с параметрами
(RA::sString DeviceId ,RA::sNodePtr DeviceCfg).
Он @b должен @b быть @b определён в соответствующем cpp-файле.
Смысл параметров:
@arg @c DeviceId задаваемый в конфигурации уникальный идентификатор
                 реально существующего "в железе" экземпляра устройства
@arg @c DeviceCfg специального вида указатель на узел сети данных,
                  содержащий параметры устройства и, возможно, состояние
                  в котором оно было оставлено после последнего использования
                  (например, текущая длина волны для спектрометра с шаговым
                  управлением развёрткой, но без датчика угол-код)

Также декларирует статическую функцию-член Preconfigure, цель которой
сформировать умолчательный кластер конфигурации устройства.  В соответствующем
cpp-файле необходимо определить эту функцию.

И наконец, декларирует некий статический член - тэг класса, обеспечивающий
динамическое создание объектов класса (т.е. экземпляров устройств), их поиск
и т.п.  В соответствующем cpp-файле необходимо определить этот тэг при помощи
макро DEFINE_CLASS_TAG() с именем класса в качестве аргумента.

Поскольку класс sDriver является наследником
класса QObject, в декларации порождаемых от sDriver классов следует ставить
ещё и макро Q_OBJECT (причём, в @b самом начале).

Использование:
@code
class sMyDriver : public sDriver
{
  Q_OBJECT
  ABSTRACT_DRIVER(sMyDriver)
  //... в классе могут быть чистые (pure) виртуальные методы
  //... а могут и не быть, но класс по замыслу абстрактен,
  //... т.е. объекты этого класса никогда не создаются
};
@endcode

@macro
*/
#define  ABSTRACT_DRIVER(CLASSNAME)                                            \
  ADAM(CLASSNAME)                                                              \
  DECLARE_DRIVER_PTR(CLASSNAME)                                                \
  public:                                                                      \
    static void               Preconfigure (sNodePtr  DevCfg);                 \
  protected:                                                                   \
                              CLASSNAME (sString,sNodePtr);                    \
  private:





//------------------------------------------------------------------ sDriver ---
/*! Абстрактный базовый класс для драйверов устройств.

*/
class RAGUI_EXPORT  sDriver : public sDeviceStateMachine
                            , public sDevice
{
  Q_OBJECT
  ABSTRACT_DRIVER(sDriver)

  public:
    virtual                   ~sDriver ();

    psState                   State_Warming () const
                                { return Warming; }
    psState                   State_Operating () const
                                { return Operating; }
    psState                   State_Cooling () const
                                { return Cooling; }
    psState                   State_Frozen () const
                                { return Frozen; }
    QFinalState *             State_Final () const
                                { return FinalState; }

  signals: //outgoing
    void                      ErrorMessage (QString = QString());

  protected:
    bool                      Connect (rcsDriverLink  sender
                                      ,literal        signal
                                      ,literal        slot_or_signal);
    bool                      Disconnect (rcsDriverLink  sender
                                         ,literal        signal
                                         ,literal        slot_or_signal);
    bool                      SendCommand (rcsDriverLink   receiver
                                          ,rcsCommand      command
                                          ,eControlChange  control_change
                                                                  =cLeaveControl
                                          );
    template<class data_type>
    bool                      SendDirective (rcsDriverLink      receiver
                                            ,const sDirective
                                                 <data_type> &  directive
                                            ,data_type          data
                                            ,eControlChange     control_change
                                                                 =cLeaveControl)
                                {
                                  if( receiver.IsCorrect() ){
                                    return receiver->GetCommand
                                        (new sState::sDirectiveEvent<data_type>
                                          (this,directive,data,control_change));
                                  };
                                  return false;
                                }
    virtual void              PostConstructor ();

  protected slots:
    virtual void              AfterStartInitiation ()
                                {}
    virtual void              OnEnterToWarmingState ()
                                {}
    virtual void              OnExitFromWarmingState ()
                                {}
    virtual void              OnEnterToOperatingState ()
                                {}
    virtual void              OnExitFromOperatingState ()
                                {}
    virtual void              OnEnterToCoolingState ()
                                {}
    virtual void              OnExitFromCoolingState ()
                                {}

  private slots:
    void                      Harakiri ();

  private:
    void                      CallPostConstructor ()
                                { PostConstructor(); }
    bool                      GetCommand (sState::psCommandEvent);
    void                      Suicide ();
    bool                      IsSuiciding () const
                                { return SuicidingFlag; }

  private: //fields
    std::set<psDriverLink>    TheLinkList;
    pcsDevice                 TheExclusiveMaster;
    bool                      SuicidingFlag;

  friend class sDriverLink;
  friend class sState::sCommandTransition;
  friend class sConsole;
};

//------------------------------------------------------------------------------
} //namespace RA
#endif
