////////////////////////////////////////////////////////////////////////////////
/*! @file DriverLink.cpp   Prototype of a smart pointer to a driver.
- Part of RAGUI - Research Assistant Graphical User Interface Library.
- Uses  Qt v.5.0.2  - http://qt-project.org/
- Uses  RANet - Research Assistant Net Library.
- Copyright(C) 2010-2014, Viktor E. Bursian, St.Petersburg, Russia.
                          Viktor.Bursian@mail.ioffe.ru
*///////////////////////////////////////////////////////////////////////////////
#include "DriverLink.h"
#include "BasicNodes.h"
#include "Driver.h"
#include "StringsAndQStrings.h"
#include "Log.h"
namespace RA {
//------------------------------------------------------------------------------
//-------------------------------------------------------------- ssDriverLink---

sDriverLink::~sDriverLink ()
{
  SetToNULL();
}

sDriverLink::sDriverLink (psDevice  link_owner)
    :TheLinkOwner(link_owner)
    ,TheDriverLink(NULL)
{
  ASSERT(link_owner);
}


bool  sDriverLink::IsNULL () const
{
  return  ( TheDriverLink == NULL );
}


void  sDriverLink::SetTo (sPtr<sFolderNode>  links
                         ,rcsString          link_name)
{
  sString                     Id;
  if( ! links.IsCorrect() )
    throw xDeviceLaunch(sString("The '")+link_name
                              +"' link is absent or incorrect.");
  if( ! get_from<sTextNode>(links>>link_name>>"DeviceId",Id) )
    throw xDeviceLaunch(sString("The '")+link_name
                              +"' link has no 'DeviceId' attribute.");
  if( Id.Empty() )
    throw xDeviceLaunch(sString("The '")+link_name
                              +"' link has 'DeviceId' attribute"
                               " with an empty string value.");
  SetTo(Id);
}


void  sDriverLink::SetTo (rcsString device_id)
{
  try{
    if( device_id != DeviceId ){
      SetToNULL();
      DeviceId = device_id;
      GetDriver();
      TheDriverLink->TheLinkList.insert(this);
    }
  }catch(...){
    TheDriverLink = NULL;
    DeviceId = sString();
    throw;
  }
}


void  sDriverLink::GetDriver ()
{
  sPtr<sFolderNode>           DeviceCfg;
  bool                        Disabled = false;
  sString                     ClassName;

  DeviceCfg = sDevice::DeviceList >> DeviceId;
  if( ! DeviceCfg.IsCorrect() )
    throw xDeviceLaunch(
        sString("There is no such device in configuration: '")+DeviceId+"'.");
  if( ! get_from<sTextNode>(DeviceCfg>>"Class",ClassName) )
    throw xDeviceLaunch(
        sString("'Class' attribute is not found for the device '")
        +DeviceId+"'.");
  get_from<sBooleanNode>(DeviceCfg>>"Disabled",Disabled);
  if( Disabled )
    throw xDeviceLaunch(sString("Device '")+DeviceId+"' is disabled.");
  
  map<sString,psDevice>::iterator    D = sDevice::RunningDevicesByName
                                                           .find(DeviceId);
  if( D != sDevice::RunningDevicesByName.end() ){
    //! @todo{Instruments} multitasking: А если решит самоубиться до того,
    //!                    как этот указатель будет окончательно оформлен?
    TheDriverLink = dynamic_cast<psDriver>(D->second);
    if( ! TheDriverLink )
      throw xProgramError("sDriverLink::GetDriver(): #1",__FILE__,__LINE__);
    if( ! TypeMatches() )
      throw xDeviceLaunch(sString("The device '")+DeviceId
                                +"' doesn't match the link's class '"
                                +DriverType()+"'.");
    if( TheDriverLink->IsInvalid() )
      throw xDeviceLaunch(sString("The device '")+DeviceId
            +"' is invalid, probably, because of bad configuration settings.");
    if( TheDriverLink->IsSuiciding() )
      throw xDeviceLaunch(sString("The device '")+DeviceId
            +"' is busy now - it is suiciding.  Try later.");
    TheDriverLink->OrderWarmUp();
  }else{
    pcsTag                      Tag = sTag::ByName(ClassName);
    if( ! Tag )
      throw xDeviceLaunch(sString("Unknown class: ")+ClassName);
    pcsDriverTag                DTag = dynamic_cast<pcsDriverTag>(Tag);
    if( ! DTag )
      throw xDeviceLaunch(sString("Not a driver: ")+ClassName);
    (*(DTag->ThePreconfigureFunctionPtr))(DeviceCfg);
    TheDriverLink = (*(DTag->TheNewDriverFunctionPtr))(DeviceId,DeviceCfg);
    if( ! TheDriverLink )
      throw xProgramError("sDriverLink::GetDriver(): #2",__FILE__,__LINE__);
    if( ! TypeMatches() ){
      TheDriverLink->deleteLater();  TheDriverLink=NULL;
      throw xDeviceLaunch(sString("The device '")+DeviceId
                                +"' doesn't match the link's class '"
                                +DriverType()+"'.");
    }
    if( ! TheDriverLink->IsInvalid() )
      TheDriverLink->CallPostConstructor();
    if( TheDriverLink->IsInvalid() ){
      QString                   Reason(TheDriverLink->InvalidityReason());
      TheDriverLink->deleteLater();  TheDriverLink=NULL;
      throw xDeviceLaunch(sString("The device '")+DeviceId
                                +"' is invalid:\n"+FromQString(Reason));
    }
    TheDriverLink->start();
  }
}


void  sDriverLink::SetToNULL ()
{
  if( TheDriverLink ){
    TheDriverLink->TheLinkList.erase(this);
    if( TheDriverLink->TheLinkList.empty() ){
      if( TheDriverLink->TraceStates() )
        RANet::Log.Put(sLog::Debug,"States"
                ,sString("sDriverLink::SetToNULL: deleting driver for '")
                +DeviceId+"'");
      TheDriverLink->Suicide();
    
    //}else if( TheDriverLink->TheExclusiveMaster == TheLinkOwner ){
      //!@todo{Instruments} Release exclusive control if looses its sense:
      //! необходимо проверить, что среди TheLinkList уже нет таких элементов L, 
      //! у которых L->TheLinkOwner == TheDriverLink->TheExclusiveMaster.
      //! Впрочем, может это и правильно, что отдавание контроля на совести
      //! драйверописателей.
    
    }
    TheDriverLink=NULL;
  }
  DeviceId = sString();
}


bool  sDriverLink::TakeExclusiveControl () const
{
  ASSERT( TheDriverLink );
  if( TheDriverLink->TheExclusiveMaster ){
    return (TheDriverLink->TheExclusiveMaster == TheLinkOwner);
  }else{
    TheDriverLink->TheExclusiveMaster = TheLinkOwner;
    return true;
  }
}


void  sDriverLink::ReleaseExclusiveControl () const
{
  if( TheDriverLink ){
    if( TheDriverLink->TheExclusiveMaster == TheLinkOwner ){
      TheDriverLink->TheExclusiveMaster = NULL;
    }
  }
}

//------------------------------------------------------------------------------
} //namespace RA



