////////////////////////////////////////////////////////////////////////////////
/*! @file DriverLink.h   Prototype of a smart pointer to a driver.
- Part of RAGUI - Research Assistant Graphical User Interface Library.
- Uses  Qt v.5.0.2  - http://qt-project.org/
- Uses  RANet - Research Assistant Net Library.
- Copyright(C) 2010-2014, Viktor E. Bursian, St.Petersburg, Russia.
                          Viktor.Bursian@mail.ioffe.ru
*///////////////////////////////////////////////////////////////////////////////
#ifndef DriverLink_H
#define DriverLink_H
#include "Device.h"
#include "RAGUI_global.h"
namespace RA {
//------------------------------------------------------------------------------

ANNOUNCE_CLASS(sDriverLink)

//--------------------------------------------------------------- sDriverLink---
/*! Прототип указателя, обеспечивающего связь между неким драйвером
    (или консолью) и подчинённым ему (ей) драйвером.

Помнит, кто им владеет, для чего при создании ему указывается владелец
(то есть, во-первых, объект вида sDriverLink обычно является членом класса,
произведённого от sDevice, и, во-вторых, этот член инициируется
конструктором sDriverLink::sDriverLink(psDevice link_owner) с аргументом
@c this).

После этого, его можно установить на драйвер устройства по имени этого
устройства (см. sDriverLink::SetTo(sString DeviceId)) и работать, как с обычным
указателем при помощи операций -> и *.

Описанным выше способом можно создать несколько указателей на один драйвер - с
одним или разными владельцами.  Однако прямое копирование или присваивание
запрещены.

Когда подчинённый драйвер перестаёт быть нужным, указатель на него следует
обнулить (см. sDriverLink::SetToNULL()).  В свою очередь, каждый драйвер помнит
все установленные на него указатели (вида sDriverLink), и когда последний из них
обнуляется, драйвер делает себе харакири.

*/
class RAGUI_EXPORT  sDriverLink
{
  public:
    virtual                   ~sDriverLink ();
                              sDriverLink (psDevice  link_owner);
    psDriver                  operator -> () const
                                { return TheDriverLink; }
    rsDriver                  operator * () const
                                { return *TheDriverLink; }
    bool                      IsNULL () const;
    bool                      IsCorrect () const
                                { return ! IsNULL(); }
    virtual sString           DriverType () const =0;
    void                      SetToNULL ();
    void                      SetTo (sPtr<sFolderNode>  links
                                    ,rcsString          link_name);
    void                      SetTo (rcsString  device_id);
    bool                      TakeExclusiveControl () const;
    void                      ReleaseExclusiveControl () const;

  protected:
    virtual bool              TypeMatches () const =0;

  private: //запрещенные операции
                              sDriverLink (rcsDriverLink);
    void                      operator = (rcsDriverLink);

  private:
    void                      GetDriver ();

  private://fields
    cpcsDevice                TheLinkOwner;
    sString                   DeviceId;
    psDriver                  TheDriverLink;
};

//-------------------------------------------- DECLARE_DRIVER_PTR(CLASSNAME) ---
/*! Этот макрос внутреннего употребления.

Он входит в текст макросов DRIVER() и ABSTRACT_DRIVER()

@macro
*/
#define  DECLARE_DRIVER_PTR(CLASSNAME)                                         \
  public:                                                                      \
    ANNOUNCE_CLASS(sLink)                                                      \
    struct  sLink : public sDriverLink                                         \
    {                                                                          \
        explicit              sLink (psDevice  link_owner)                     \
                                  :sDriverLink(link_owner)                     \
                                {}                                             \
        p##CLASSNAME          operator-> () const                              \
                                {                                              \
                                  return p##CLASSNAME(sDriverLink              \
                                                               ::operator->());\
                                }                                              \
        r##CLASSNAME          operator* () const                               \
                                {                                              \
                                  return r##CLASSNAME(sDriverLink              \
                                                                ::operator*());\
                                }                                              \
        virtual sString       DriverType () const                              \
                                { return sString(#CLASSNAME); }                \
      protected:                                                               \
        virtual bool          TypeMatches () const                             \
                                {                                              \
                                  return (bool) dynamic_cast<p##CLASSNAME>     \
                                                 (sDriverLink::operator ->()); \
                                };                                             \
    };

//------------------------------------------------------------------------------
} //namespace RA
#endif
