////////////////////////////////////////////////////////////////////////////////
/*! @file ExpressionNodeUI.cpp Add-ons to the navigator functionality.
- Part of RAGUI - Research Assistant Graphical User Interface Library.
- Uses  RANet - Research Assistant Net Library.
- Uses  Qt v.5  -  http://qt.io/
- Copyright(C) 2014-2017, Viktor E. Bursian, St.Petersburg, Russia.
                          Viktor.Bursian@mail.ioffe.ru
*///////////////////////////////////////////////////////////////////////////////
#include "ExpressionNodeUI.h"
#include "ExpressionNode.h"
#include "Navigator.h"
#include "EditExpressionDialog.h"
#include "ExportDataDialog.h"
#include "Functions.h"
#include "CurvesUI.h"
#include "NetExceptions.h"
#include "StringsAndQStrings.h"
#include "Globals.h"
#include <QMessageBox>
namespace RA {
//------------------------------------------------------------------------------
//-------------------------------------------------------- sExpressionNodeUI ---

sExpressionNodeUI::sExpressionNodeUI ()
{
  sChartViewport::DrawProcCollection.push_front( & Draw );
  sNavigatorViewport::MenuGenerators.push_front(&MenuGenerator);
}


bool  sExpressionNodeUI::Draw (QPainter *       painter
                              ,psChartViewport  viewport
                              ,psGraphObject    object
                              ,pcsAppearance    /*appearance*/)
{
  psExpressionNode            Expr(dynamic_cast<psExpressionNode>(object));
  if( Expr && Expr->Result() ){
    psFunction                  Function(dynamic_cast<psFunction>
                                                     (Expr->Result()));
//    psCurveAppearance           Appearance(dynamic_cast<psCurveAppearance>
//                                                       (appearance)       );
// appearance может быть уже убита после обращения к Result()
    pcsCurveAppearance          Appearance(dynamic_cast<pcsCurveAppearance>
                                                       (Expr->Appearance()) );
    if( Function )
      return sCurveUI::Draw(painter,viewport,Function,Appearance);
  }
  return false;
}


void  sExpressionNodeUI::MenuGenerator (psNetViewMenu               context_menu
                                       ,sNet2TreeEvolution::pcsItem tree_item)
{
  if( sExpressionNode::Type()->CanCast(tree_item->AttrValue().operator ->()) ){
    context_menu->Insert( new sNetViewAction
        (context_menu->parentWidget()
        ,&EditValue  ,/*checkable*/false ,/*checked*/false
        ,sNetViewMenu::tr("Edit the expression")
        ,sNetViewMenu::tr("View and edit the expression's source text")
        ));
    psExpressionNode   Expr(dynamic_cast<psExpressionNode>
                                        (tree_item->AttrValue().operator ->()));
    if( Expr && Expr->Result()
             && sDataFunction::Type()->CanCast(Expr->Result()) ){
      context_menu->Insert( new sNetViewAction
          (context_menu->parentWidget()
          ,&Export  ,/*checkable*/false ,/*checked*/false
          ,sNetViewMenu::tr("Export")
          ,sNetViewMenu::tr("Export the curve into a separate, "
                            "portable format file")
          ));
    }
    if( Expr && Expr->Result()
             && sFunction::Type()->CanCast(Expr->Result()) ){
      context_menu->Insert( new sNetViewAction
          (context_menu->parentWidget()
          ,&EditCurveAppearance  ,/*checkable*/false ,/*checked*/false
          ,sNetViewMenu::tr("Change appearance")
          ,sNetViewMenu::tr("Change color, thickness, etc.")
          ));
    }
  }
  if( sFolderNode::Type()->CanCast(tree_item->AttrValue().operator ->()) ){
    context_menu->AddNew->addAction( new sNetViewAction
        (context_menu->parentWidget()
        ,&AddNew  ,/*checkable*/false ,/*checked*/false
        ,sNetViewMenu::tr("expression")
        ,sNetViewMenu::tr("Add new expression")
        ));
  }
  if( tree_item->Parent() ){
    context_menu->InsertNew->addAction( new sNetViewAction
        (context_menu->parentWidget()
        ,&InsertNew  ,/*checkable*/false ,/*checked*/false
        ,sNetViewMenu::tr("expression")
        ,sNetViewMenu::tr("Insert new expression")
        ));
  }
}


void  sExpressionNodeUI::EditValue (sNet2TreeEvolution::pcsItem  tree_item
                                   ,sPoint                       /*pos*/)
{
  /*! @todo{UI++} in-place editing */
  sEditExpressionDialog       D;
  D.setWindowTitle( sEditExpressionDialog::tr("Edit the expression") );
  D.SetAttrName(tree_item->AttrName());
  D.SetAttrNameEnabled(false);
  sPtr<sExpressionNode>           C(tree_item->AttrValue());
  ASSERT(C.IsCorrect());
  D.SetDocket(C->Docket());
  if( D.exec() ){
    C->SetDocket(D.Docket());
  }
}


void  sExpressionNodeUI::AddNew (sNet2TreeEvolution::pcsItem  tree_item
                                ,sPoint                       /*pos*/)
{
  sEditExpressionDialog       D;
  D.setWindowTitle( sEditDocketDialog::tr("Add an expression") );
  D.SetDocket("");
  while( D.exec() ){
    try{
      tree_item->AttrValue()
            << sAttribute( D.AttrName()
                         , sNodePtr( new sExpressionNode(D.Docket()) ) );
      break;
    }catch( rxNetInvalidOp ){
      QMessageBox::critical(NULL ,D.windowTitle() ,D.ErrorMsgOnAddAttr() );
    }
  }
}


void  sExpressionNodeUI::InsertNew (sNet2TreeEvolution::pcsItem  tree_item
                                   ,sPoint                       /*pos*/)
{
  sEditExpressionDialog       D;
  D.setWindowTitle( sEditDocketDialog::tr("Insert an expression") );
  D.SetDocket("");
  sPtr<sExpressionNode>           NewNode;
  while( D.exec() ){
    try{
      NewNode = sNodePtr( new sExpressionNode(D.Docket()) );
      tree_item->Parent()->AttrValue()<<sAttribute(D.AttrName(),NewNode);
      break;
    }catch( rxNetInvalidOp ){
      QMessageBox::critical(NULL ,D.windowTitle() ,D.ErrorMsgOnAddAttr() );
    }
  }
  //! @todo{Insert_attribute} move NewNode
}


void  sExpressionNodeUI::EditCurveAppearance
                                   (sNet2TreeEvolution::pcsItem  tree_item
                                   ,sPoint                       /*pos*/)
{
  sPtr<sExpressionNode>       ExprNode(tree_item->AttrValue());
  ASSERT( ExprNode.IsCorrect() );
  pcsCurveAppearance          OldCA( dynamic_cast<pcsCurveAppearance>
                                                 (ExprNode->Appearance()) );
  psCurveAppearance           NewCA( OldCA
                                   ? new sCurveAppearance(*OldCA)
                                   : new sCurveAppearance );
  sEditCurveAppearanceDialog  D(NewCA
                               ,RAGUI::LineColorSet->Value()
                               ,( RAGUI::NightRegime ? sColor::cBlack
                                                     : sColor::cWhite )
                               );
  if( D.exec() ){
    ExprNode->SetAppearance(NewCA);
  }else{
    delete NewCA;
  }
}


void  sExpressionNodeUI::Export (sNet2TreeEvolution::pcsItem  tree_item
                                ,sPoint                       /*pos*/)
{
  try{
    sPtr<sExpressionNode>       Expr(tree_item->AttrValue());
    psDataFunction              Function(dynamic_cast<psDataFunction>
                                                     (Expr->Result()) );
    ASSERT(Function);
    sExportDataDialog           D(tree_item->AttrName().Text(),Expr);
    D.exec();
  }catch(...){
    QMessageBox::critical(NULL,"Error","Error");
  }
}

//------------------------------------------------------------------------------
} //namespace RA
