////////////////////////////////////////////////////////////////////////////////
/*! @file FlexibleFields.cpp   Input fields with useful features.
- Part of RAGUI - Research Assistant Graphical User Interface Library.
- Uses  Qt v.5  - http://qt.io/
- Uses  RANet - Research Assistant Net Library.
- Copyright(C) 2010-2017, Viktor E. Bursian, St.Petersburg, Russia.
                          Viktor.Bursian@mail.ioffe.ru
*///////////////////////////////////////////////////////////////////////////////
#include "FlexibleFields.h"
#include "SymbolsGUI.h"
#include "Trace.h"
#include <QIntValidator>
namespace RA {
//------------------------------------------------------------------------------
//-----------------------------------------------------------  sFlexibleField---

sFlexibleField::sFlexibleField (QWidget *  parent)
    :QLineEdit(parent)
    ,ReadOnlyFlag(false)
    ,TextChangedFlag(false)
{
  connect(this,SIGNAL(textChanged(QString)),this,SLOT(ResizeField()));
}


psFlexibleField  sFlexibleField::SetReadOnly (bool  read_only)
{
  ReadOnlyFlag=read_only;
  setStyleSheet(( read_only ? "color: maroon" : "" ));
  return this;
}


void  sFlexibleField::keyPressEvent (QKeyEvent * event)
{
  if( event->key()  == Qt::Key_Escape ){
    emit EscapeKeyPressed();
  }else if( (event->key()  == Qt::Key_Return)
         || (event->key()  == Qt::Key_Enter) ){
    if( event->modifiers() & Qt::ShiftModifier ){
    }else{
      emit EnterKeyPressed();
    };
  }else{
    QLineEdit::keyPressEvent(event);
  };
}


void  sFlexibleField::focusInEvent (QFocusEvent *  event)
{
  QLineEdit::focusInEvent(event);
  emit GotFocus();
}


void  sFlexibleField::focusOutEvent (QFocusEvent *  event)
{
  QLineEdit::focusOutEvent(event);
  emit LostFocus();
}


void  sFlexibleField::ResizeField ()
{
  setMinimumWidth(fontMetrics().width(text()/*+"00"*/)
                 +10
                 //Не знаю как правильно вычислить размер рамки (зависит
                 //от стиля, меняется при переключении в негатив, при
                 //переходе в другую ОС
                 );
  setMaximumWidth(minimumWidth());

//  parentWidget()-> ?? adjustSize();
}

//---------------------------------------------------------- sSpinDigitField ---

sSpinDigitField::sSpinDigitField (QWidget *  parent)
    :sFlexibleField(parent)
{
}


void  sSpinDigitField::keyPressEvent (QKeyEvent * event)
{
  if( event->key()  == Qt::Key_Up ){
    if( (event->modifiers() & Qt::ShiftModifier) ){
      setText(ToQString(sString::FromReal(real_inf,0)));
    }else{
      int                         CurPos = cursorPosition();
      QString                     T = text();
      Spin(T,CurPos,true);
      setText(T);
      emit ChangedBySpinning();
      setCursorPosition(CurPos);
    };
  }else if( event->key()  == Qt::Key_Down ){
    if( (event->modifiers() & Qt::ShiftModifier) ){
      setText(ToQString(sString::FromReal(real_inf,0)));
    }else{
      int                         CurPos = cursorPosition();
      QString                     T = text();
      Spin(T,CurPos,false);
      setText(T);
      emit ChangedBySpinning();
      setCursorPosition(CurPos);
    };
  }else if( (event->key()  == Qt::Key_Question) ){
    setText(ToQString(sString::FromReal(std::numeric_limits<real>::quiet_NaN()
                                       ,0)));
  }else if( (event->key()  == Qt::Key_Comma) ){
  }else if( (event->key()  == Qt::Key_Space) ){
  }else{
    sFlexibleField::keyPressEvent(event);
  };
}


void  sSpinDigitField::contextMenuEvent (QContextMenuEvent *  /*event*/)
{
//  QMenu                       PopupMenu(parentWidget());
//  PopupMenu.addAction("Clear",this,SLOT(Clear()));
//  PopupMenu.exec(event->globalPos());
}


void  sSpinDigitField::Spin (QString &  Text
                            ,int        CurPos
                            ,bool       upwards)
{
  /*! @bug сколлирование через ноль и правильное направление для отрицательных*/
  if( CurPos > 0 ){
    sString                     DD = FromQString(Text.at(CurPos-1));
    if( DD.Length() == 1 ){
      char                        D;
      D = DD[(size_t)0];
      if( (D == '.') || (D == ',') ){
        Spin(Text,CurPos-1,upwards);
      }else if( (D >= '0') && (D <= '9') ){
        if( upwards ){
          ++D;
          if( D > '9'){
            D = '0';
            Spin(Text,CurPos-1,upwards);
          };
        }else{
          --D;
          if( D < '0'){
            D = '9';
            Spin(Text,CurPos-1,upwards);
          };
        };
        Text[CurPos-1] = QChar(D);
      };
    };
  };
}


//--------------------------------------------------------------- sTextField ---

sTextField::~sTextField ()
{
}


void sTextField::Construct ()
{
  TheAttentionFlag = false;
  UnsetAttentionOnEditFlag = true;
  QHBoxLayout *       Layout = new QHBoxLayout(this);
  Layout->setMargin(0);
  Layout->setSpacing(0);
  Layout->setAlignment(Qt::AlignTop);

  LblLabel = new QLabel(this);
  SetLabel(TheLabel);
  Layout->addWidget(LblLabel);

  FldValue = new sFlexibleField(this);
  SetText(TheDefaultText);
  Layout->addWidget(FldValue);

  connect(FldValue ,SIGNAL(textChanged(QString)) ,this ,SIGNAL(Changed()) );
}


psTextField  sTextField::SetLabel (QString  label)
{
  TheLabel = label;
  if( ! label.isEmpty() )
    label = label + "=";
  LblLabel->setVisible( ! label.isEmpty() );
  LblLabel->setText(label);
  return this;
}


psTextField  sTextField::SetReadOnly (bool  read_only)
{
  FldValue->SetReadOnly(read_only);
  return this;
}


psTextField  sTextField::SetAttention (bool  attentive)
{
  TheAttentionFlag = attentive;
  LblLabel->setStyleSheet(( attentive ? "color: red" : "" ));
  return this;
}


psTextField  sTextField::SetText (sString  a_text)
{
  FldValue->setText(ToQString(a_text));
  FldValue->ResizeField();
  return this;
}


//------------------------------------------------------------------------------
} //namespace RA


