////////////////////////////////////////////////////////////////////////////////
/*! @file PhysRangeField.cpp   Виджет-редактор диапазона физвеличины.
- Part of RAGUI - Research Assistant Graphical User Interface Library.
- Uses  Qt v.5.0.2  - http://qt-project.org/
- Uses  RANet - Research Assistant Net Library.
- Copyright(C) 2010-2015, Viktor E. Bursian, St.Petersburg, Russia.
                          Viktor.Bursian@mail.ioffe.ru
*///////////////////////////////////////////////////////////////////////////////
#include "PhysRangeField.h"
#include "StringsAndQStrings.h"
#include "SymbolsGUI.h"
#include "Trace.h"
#include <QIntValidator>
namespace RA {
//------------------------------------------------------------------------------
//---------------------------------------------------------- sPhysRangeField ---

sPhysRangeField::~sPhysRangeField ()
{
  if( TheValue ){ delete TheValue;  TheValue = NULL; };
}


void  sPhysRangeField::Construct (QString     label
                                 ,sPhysRange  value
                                 ,bool        units_are_editable
                                 ,bool        precision_depends_on_value_digits
                                 )
{
  TheValue = new sPhysRange(value);
  UnitsAreEditableFlag = units_are_editable;
  PrecisionDependsOnValueDigitsFlag = precision_depends_on_value_digits;
  UnsetAttentionOnEditFlag = true;
  ValidRange = NULL;
  ValidationProc = NULL;
    
  QHBoxLayout *       Layout = new QHBoxLayout(this);
  Layout->setMargin(0);
  Layout->setSpacing(0);
  Layout->setAlignment(Qt::AlignTop);

  LblLabel = new QLabel(this);
  SetLabel(label);
  Layout->addWidget(LblLabel);

  FldFrom = new sSpinDigitField(this);
  QDoubleValidator * FromValidator = new QDoubleValidator(FldFrom);
  FromValidator->setNotation(QDoubleValidator::StandardNotation);
  FldFrom->setValidator(FromValidator);
  Layout->addWidget(FldFrom);

  connect(FldFrom,SIGNAL(GotFocus()),this,SLOT(UpdateVisibility()));
  connect(FldFrom,SIGNAL(EnterKeyPressed()),this,SLOT(UpdateValueByFrom()));
  connect(FldFrom,SIGNAL(ChangedBySpinning()),this,SLOT(UpdateValueByFrom()));
  connect(FldFrom,SIGNAL(LostFocus()),this,SLOT(PossiblyLeavingFromFrom()));
                    //!< @todo{PhysValues} restore SLOT(UpdateValueByFrom())?

  LblFromTo = new QLabel(QString::fromUtf8("…"),this);
  Layout->addWidget(LblFromTo);

  FldTo = new sSpinDigitField(this);
  QDoubleValidator * ToValidator = new QDoubleValidator(FldTo);
  ToValidator->setNotation(QDoubleValidator::StandardNotation);
  FldTo->setValidator(ToValidator);
  Layout->addWidget(FldTo);

  connect(FldTo,SIGNAL(GotFocus()),this,SLOT(UpdateVisibility()));
  connect(FldTo,SIGNAL(EnterKeyPressed()),this,SLOT(UpdateValue()));
  connect(FldTo,SIGNAL(ChangedBySpinning()),this,SLOT(UpdateValue()));
  connect(FldTo,SIGNAL(LostFocus()),this,SLOT(PossiblyLeaving()));
                    //!< @todo{PhysValues} restore SLOT(UpdateValue())?

  BtnOrder = new QPushButton(QString::fromUtf8(Symbol::Multiplication),this);
  BtnOrder->setMaximumWidth(12);
  Layout->addWidget(BtnOrder);
  BtnOrder->setVisible(false);

  connect(BtnOrder,SIGNAL(clicked()),this,SLOT(ShowOrder()));

  LblOrder = new QLabel(QString::fromUtf8(Symbol::Multiplication)+"10",this);
  Layout->addWidget(LblOrder);

  FldOrder = new sFlexibleField(this);
  FldOrder->setText("0");
  FldOrder->setValidator(new QIntValidator(-30,30,FldOrder));
  FldOrder->setMinimumHeight(FldOrder->height() / 2 + 1);
  FldOrder->setMaximumHeight(FldOrder->minimumHeight());
  Layout->addWidget(FldOrder,0,Qt::AlignTop);

  connect(FldOrder,SIGNAL(EnterKeyPressed()),this,SLOT(UpdateValue()));
  connect(FldOrder,SIGNAL(LostFocus()),this,SLOT(PossiblyLeaving()));

  LblUnits = new QLabel(this);
  Layout->addWidget(LblUnits);
  LblUnits->setVisible( ! UnitsAreEditableFlag );

  BtnUnits = new QPushButton("[]",this);
  BtnUnits->setMaximumWidth(16);
  Layout->addWidget(BtnUnits);
  BtnUnits->setVisible(false);

  connect(BtnUnits,SIGNAL(clicked()),this,SLOT(ShowUnits()));

  FldUnits = new sFlexibleField(this);
  FldUnits->setText(" ");
  Layout->addWidget(FldUnits);
  FldUnits->setVisible( UnitsAreEditableFlag );

  connect(FldUnits,SIGNAL(EnterKeyPressed()),this,SLOT(UpdateValue()));
  connect(FldUnits,SIGNAL(LostFocus()),this,SLOT(PossiblyLeaving()));

  TheValue->Normalize();
  UpdateImage();
}


psPhysRangeField  sPhysRangeField::SetLabel (QString  label)
{
  TheLabel = label;
  if( ! label.isEmpty() )
    label = label + "=";
  LblLabel->setVisible( ! label.isEmpty() );
  LblLabel->setText(label);
  return this;
}


psPhysRangeField  sPhysRangeField::SetUnitsEditable (bool  editable)
{
  UnitsAreEditableFlag = editable;
  LblUnits->setVisible( ! UnitsAreEditableFlag );
  FldUnits->setVisible( UnitsAreEditableFlag );
  return this;
}


psPhysRangeField  sPhysRangeField::SetPrecisionDependentOnValueDigits 
                                      (bool  /*dependent*/)
{
  //!@patch
  //PrecisionDependsOnValueDigitsFlag=dependent;
  return this;
}


psPhysRangeField  sPhysRangeField::SetReadOnly (bool  read_only)
{
  FldFrom->setStyleSheet(( read_only ? "color: maroon" : "" ));
  FldTo->setStyleSheet(( read_only ? "color: maroon" : "" ));
  return this;
}


psPhysRangeField  sPhysRangeField::SetAttention (bool  attentive)
{
  LblLabel->setStyleSheet(( attentive ? "color: red" : "" ));
  return this;
}


bool  sPhysRangeField::Active () const
{
  return FldFrom->hasFocus()
      || FldTo->hasFocus()
      || BtnOrder->hasFocus()
      || FldOrder->hasFocus()
      || BtnUnits->hasFocus()
      || FldUnits->hasFocus();
}


void  sPhysRangeField::PossiblyLeavingFromFrom ()
{
  if( ! Active() ){
//    TRACE(TheLabel+" leaving");
    UpdateValueByFrom();
    emit LostFocus();
  };
}


void  sPhysRangeField::PossiblyLeaving ()
{
  if( ! Active() ){
//    TRACE(TheLabel+" leaving");
    UpdateValue();
    emit LostFocus();
  };
}


void  sPhysRangeField::UpdateImage ()
{
//  TRACE(TheLabel+" UpdateImage");
  unsigned short int          D = TheValue->Digits();
  FldFrom->setText( ToQString(
                          sString::FromReal(TheValue->FromMantissa(),D)  ));
  FldTo->setText( ToQString(
                          sString::FromReal(TheValue->ToMantissa(),D)  ));
  FldOrder->setText( QString().setNum( TheValue->Order() ) );
  FldUnits->setText(ToQString(TheValue->Units().Text(Plain)));
  LblUnits->setText(ToQString(TheValue->Units().Text(HTML)));
  UpdateVisibility();
}


void  sPhysRangeField::UpdateVisibility ()
{
  bool                        active = Active();
  if( FldOrder->text().simplified().isEmpty()
      || (FldOrder->text().simplified() == "0") ){
    BtnOrder->setVisible(active);
    LblOrder->setVisible(false);
    FldOrder->setVisible(false);
  }else{
    BtnOrder->setVisible(false);
    LblOrder->setVisible(true);
    if( active ){
      LblOrder->setText( QString::fromUtf8(Symbol::Multiplication)+"10" );
    }else{
      LblOrder->setText( QString::fromUtf8(Symbol::Multiplication)
                       + "10<sup>"
                       + FldOrder->text()
                       + "</sup>" );
    };
    FldOrder->setVisible(active);
  };
  if( UnitsAreEditableFlag ){
    if( FldUnits->text().simplified().isEmpty() ){
      BtnUnits->setVisible(active);
      FldUnits->setVisible(false);
      LblUnits->setVisible(false);
    }else{
      BtnUnits->setVisible(false);
      FldUnits->setVisible(active);
      LblUnits->setVisible(!active);
    };
  }else{
    BtnUnits->setVisible(false);
    FldUnits->setVisible(false);
    LblUnits->setVisible( ! FldUnits->text().simplified().isEmpty() );
  };
  adjustSize();
}


void  sPhysRangeField::ShowOrder ()
{
  LblOrder->setVisible(true);
  FldOrder->setVisible(true);
  FldOrder->setFocus();
  BtnOrder->setVisible(false);
  adjustSize();
}


void  sPhysRangeField::ShowUnits ()
{
  FldUnits->setVisible(true);
  FldUnits->setFocus();
  BtnUnits->setVisible(false);
  LblUnits->setVisible(false);
  adjustSize();
}


void  sPhysRangeField::UpdateValueByFrom ()
{
//  TRACE(TheLabel+" UpdateValueByFrom");
  try{
    real                        from = TheValue->FromMantissa();
    real                        to = TheValue->ToMantissa();
    real                        precision = TheValue->PrecisionMantissa();
    int                         order = 0;
    sUnits                      U;
    sString                     temp_str;
    temp_str=FromQString(FldFrom->text().simplified());
    if( ! temp_str.Empty() ){
      sString(temp_str) >> from;
    };
    if( PrecisionDependsOnValueDigitsFlag && ! isinf(from) && ! isnan(from) ){
      temp_str=temp_str.Translate('1','0').Translate('2','0').Translate('3','0')
           .Translate('4','0').Translate('5','0').Translate('6','0')
           .Translate('7','0').Translate('8','0').Translate('9','0');
      if( temp_str.Pos(".") == temp_str.Length() )
        temp_str+=".";
      temp_str+="5";
      temp_str >> precision;
    };
    temp_str=FromQString(FldTo->text().simplified());
    if( ! temp_str.Empty() ){
      sString(temp_str) >> to;
    };
    if( ! FldOrder->text().simplified().isEmpty() )
      FromQString(FldOrder->text()) >> order;
    if( UnitsAreEditableFlag ){
      U = sUnits::Parse(FromQString(FldUnits->text()));
    }else{
      U = TheValue->Units();
    };
    if( to < from )
      to=from;
    psPhysRange               R = new sPhysRange(from*power10(order)
                                               ,to*power10(order)
                                               ,U
                                               ,precision*power10(order)
                                               );
    if( Validate(*R) ){
      delete TheValue;
      TheValue = R;
      TheValue->Normalize();
    }else{
      delete R;
    };
    /*! @todo{UI++} emit Edited signal IF NECESSARY */
    emit Edited();
  }catch(.../*EConvertError& E*/){
    TRACE("CONVERSION ERROR");
  };
  UpdateImage();
}


void  sPhysRangeField::UpdateValue ()
{
//  TRACE(TheLabel+" UpdateValue");
  try{
    real                        from = TheValue->FromMantissa();
    real                        to = TheValue->ToMantissa();
    real                        precision = TheValue->PrecisionMantissa();
    int                         order = 0;
    sUnits                      U;
    sString                     temp_str;
    temp_str=FromQString(FldFrom->text().simplified());
    if( ! temp_str.Empty() ){
      sString(temp_str) >> from;
    };
    temp_str=FromQString(FldTo->text().simplified());
    if( ! temp_str.Empty() ){
      sString(temp_str) >> to;
    };
    if( PrecisionDependsOnValueDigitsFlag && ! isinf(to) && ! isnan(to) ){
      temp_str=temp_str.Translate('1','0').Translate('2','0').Translate('3','0')
           .Translate('4','0').Translate('5','0').Translate('6','0')
           .Translate('7','0').Translate('8','0').Translate('9','0');
      if( temp_str.Pos(".") == temp_str.Length() )
        temp_str+=".";
      temp_str+="5";
      temp_str >> precision;
    };
    if( ! FldOrder->text().simplified().isEmpty() )
      FromQString(FldOrder->text()) >> order;
    if( UnitsAreEditableFlag ){
      U = sUnits::Parse(FromQString(FldUnits->text()));
    }else{
      U = TheValue->Units();
    };
    if( from > to )
      from=to;
    psPhysRange               R = new sPhysRange(from*power10(order)
                                               ,to*power10(order)
                                               ,U
                                               ,precision*power10(order)
                                               );
    if( Validate(*R) ){
      delete TheValue;
      TheValue = R;
      TheValue->Normalize();
    }else{
      delete R;
    };
    /*! @todo{UI++} emit Edited signal IF NECESSARY */
    emit Edited();
  }catch(.../*EConvertError& E*/){
    TRACE("CONVERSION ERROR");
  };
  UpdateImage();
}


sPhysRange  sPhysRangeField::Value ()
{
  ASSERT(TheValue);
  return *TheValue;
}


psPhysRangeField  sPhysRangeField::SetValue (sPhysRange  R)
{
  if( TheValue ) delete TheValue;
  TheValue = new sPhysRange(R);
  TheValue->Normalize();
  UpdateImage();
  return this;
}


psPhysRangeField  sPhysRangeField::SetValidRange (sPhysRange  range)
{
  if( ValidRange )
    delete ValidRange;
  ValidRange = new sPhysRange(range);
  if( ValidRange ){
    UnitsAreEditableFlag=false;
    try{
      (*TheValue) &= (*ValidRange);
    }catch(...){
      delete TheValue;
      TheValue = new sPhysRange(range);
      TheValue->Normalize();
    };
  };
  UpdateImage();
  return this;
}


void  sPhysRangeField::SetValidationProc (tValidationProc *  proc)
{
  ValidationProc=proc;
}


bool  sPhysRangeField::Validate (rsPhysRange  range)
{
  try{
    if( ValidRange ){
      range &= (*ValidRange);
    };
    if( ValidationProc ){
      return (*ValidationProc)(range);
    }else{
      return true;
    };
  }catch(...){
    return false;
  };
}

//------------------------------------------------------------------------------
} //namespace RA


