////////////////////////////////////////////////////////////////////////////////
/*! @file Plot.cpp   A child window class sPlot.
- Part of RAGUI - Research Assistant Graphical User Interface Library.
- Uses  RANet - Research Assistant Net Library.
- Uses  Qt v.5  -  http://qt-project.org/
- Copyright(C) 2010-2015, Viktor E. Bursian, St.Petersburg, Russia.
                          VBursian@gmail.com
*///////////////////////////////////////////////////////////////////////////////
#include "Plot.h"
#include "BasicNodes.h"
#include "attr.h"
#include "PrintPreview.h"
//#include <QMenuBar>
#include <QPainter>
namespace RA {
//------------------------------------------------------------------------------

sPlot::~sPlot()
{
  if( Legend ){ delete Legend;  Legend=NULL; };
  if( Chart ){ delete Chart;  Chart=NULL; };
  if( Evolution ){ delete Evolution;  Evolution=NULL; };
}


sPlot::sPlot (sNodePtr   config
             ,sString    default_title
             ,QWidget *  parent       )
    :sWindow(config,default_title,parent)
{
  Layout = new QHBoxLayout(this);
  Layout->setMargin(0);
  Splitter = new QSplitter(this);
  Splitter->setOrientation(Qt::Vertical);

  sPtr<sFolderNode>           root(WidgetConfig>>"Content");
  ASSERT(root.IsCorrect());
  Evolution = new sNet2TreeEvolution(sAttribute(sAttrName(Title()),root));
  Chart = new sChart(Evolution ,this);
  Legend = new sNavigator(Evolution ,this);

  Splitter->addWidget(Chart);
  Splitter->addWidget(Legend);
  Layout->addWidget(Splitter);

//  Legend->setMinimumSize(80,80);

  connect(Chart ,SIGNAL(CursorPositionChanged(sPhysPair,sPhysPair
                                             ,sString,sString))
         ,this  ,SIGNAL(CursorPositionChanged(sPhysPair,sPhysPair
                                             ,sString,sString)));

  connect(this  ,SIGNAL(CursorPositionChanged(sPhysPair,sPhysPair
                                             ,sString,sString))
         ,MainWindow() ,SLOT(ShowCursorInfo(sPhysPair,sPhysPair
                                           ,sString,sString))
         ,Qt::QueuedConnection);

  connect(Evolution,SIGNAL(Changed()),Legend,SLOT(OnNetChange()));
  connect(Evolution,SIGNAL(Changed()),Chart,SLOT(OnNetChange()));
  Evolution->Initiate();
  UpdatingTimer = new QTimer(this);
  connect(UpdatingTimer, SIGNAL(timeout()), Evolution, SLOT(Update()));
  UpdatingTimer->start(sNavigator::UpdatingPeriod);

  Chart->AllowCurveDrops();

  QActionGroup *              AG = new QActionGroup(this);
  ActnLegendBottom = new QAction(AG);
  ActnLegendBottom->setIcon(QIcon(":/RAGUI/icons/LegendBottom.png"));
  ActnLegendBottom->setCheckable(true);
  ActnLegendBottom->setChecked(true);
  connect(ActnLegendBottom,SIGNAL(toggled(bool))
         ,this,SLOT(PutLegendAtTheBottom(bool)));
  addAction(ActnLegendBottom);
  ActnLegendRight = new QAction(AG);
  ActnLegendRight->setIcon(QIcon(":/RAGUI/icons/LegendRight.png"));
  ActnLegendRight->setCheckable(true);
  addAction(ActnLegendRight);

  ActnPrint = new QAction(this);
  ActnPrint->setShortcuts(QKeySequence::Print);
  ActnPrint->setIcon(QIcon(":/RAGUI/icons/print.png"));
  connect(ActnPrint,SIGNAL(triggered()),this,SLOT(PrintPreview()));
  addAction(ActnPrint);

//  MenuExtension = new QMenu("PLOT",this);
//  MenuExtension->addActions(Chart->actions());
//  QMenu * ML = MenuExtension->addMenu("Legend");
//  ML->addActions(AG->actions());

  Translate();
}


void sPlot::Translate()
{
  ActnLegendBottom->setText(tr("Legend at the bottom"));
  ActnLegendRight->setText(tr("Legend on the right"));
  ActnPrint->setText(tr("Print..."));
}


const QIcon &  sPlot::Icon ()
{
  static QIcon *  MyIcon = new QIcon(":/RAGUI/icons/plot.png");
  return *MyIcon;
}


void  sPlot::RestoreState ()
{
  sWindow::RestoreState();
  if( ! WidgetConfig.IsNULL() ){
    sPtr<sIntegerNode>        I;
    sPtr<sIntegerNode>        I1;
    sPtr<sIntegerNode>        I2;

    I=WidgetConfig>>"Splitter.Orientation";
    if( I.IsCorrect() ){
      Splitter->setOrientation(Qt::Orientation( (int)(*I) ));
      ActnLegendBottom->setChecked(Splitter->orientation() == Qt::Vertical);
      ActnLegendRight->setChecked(Splitter->orientation() == Qt::Horizontal);
    }
    I1=WidgetConfig>>"Splitter.Part1";
    I2=WidgetConfig>>"Splitter.Part2";
    if( I1.IsCorrect() && I2.IsCorrect() ){
      QList<int>    List;
      List.append( (int)(*I1) );
      List.append( (int)(*I2) );
      Splitter->setSizes(List);
    }

    Chart->RestoreScales(WidgetConfig);
  }
}


void  sPlot::StoreState ()
{
  sWindow::StoreState();
  if( ! WidgetConfig.IsNULL() ){
    WidgetConfig << attr("Splitter.Orientation" ,(int)Splitter->orientation());
    WidgetConfig << attr("Splitter.Part1" ,(int)Splitter->sizes().at(0));
    WidgetConfig << attr("Splitter.Part2" ,(int)Splitter->sizes().at(1));
    Chart->StoreScales(WidgetConfig);
  }
}


void  sPlot::focusInEvent (QFocusEvent *  event)
{
  /* во-первых, место занимает;
     во-вторых, тулбары можно показать/спрятать через
     автоматическое контекстное меню (на тулбарах, доках, меню)
  if( MainWindow->ToolBarPlot ){
    MainWindow->ToolBarPlot->clear();
    MainWindow->ToolBarPlot->addActions(actions());
    MainWindow->ToolBarPlot->addSeparator();
    MainWindow->ToolBarPlot->addActions(Chart->actions());
    MainWindow->ToolBarPlot->setVisible(true);
  };
  */
  /*
  PA = MainWindow->menuBar()->addMenu(MenuExtension);
  */
  sWindow::focusInEvent(event);
}


void  sPlot::focusOutEvent (QFocusEvent *  event)
{
  /*
  if( MainWindow->ToolBarPlot ){
    MainWindow->ToolBarPlot->setVisible(false);
  }
  */
  /*
  MainWindow->menuBar()->removeAction(PA);
  */
  sWindow::focusOutEvent(event);
}


void  sPlot::contextMenuEvent (QContextMenuEvent *  event)
{
  QMenu                       PopupMenu/*(this)*/;
  PopupMenu.addActions(actions());
  PopupMenu.exec(event->globalPos());
}


void  sPlot::PutLegendAtTheBottom (bool  AtTheBottom)
{
  if( AtTheBottom ){
    Splitter->setOrientation(Qt::Vertical);
  }else{
    Splitter->setOrientation(Qt::Horizontal);
  }
}


void  sPlot::PrintPreview ()
{
  psPrintPreview              Preview = new sPrintPreview(this);
  Preview->setAttribute(Qt::WA_DeleteOnClose);
  connect(Preview,SIGNAL(paintRequested(QPrinter*))
         ,this, SLOT(Print(QPrinter*)));
  Preview->exec();
}


void  sPlot::Print (QPrinter *  printer)
{
  psChart                     PrintChart;
  psNavigator                 PrintLegend;
  QPainter                    Painter;
  int                         ChartArea = 0;

  Painter.begin(printer);

  if( (int)Splitter->sizes().at(0) > 0 ){
    PrintChart = new sChart(Evolution);
    PrintChart->SetChartScales(Chart->ChartScales());
    if( Splitter->orientation() == Qt::Vertical ){
      ChartArea = Round( (int)Splitter->sizes().at(0) * printer->height()
                        /((int)Splitter->sizes().at(0)
                         +(int)Splitter->sizes().at(1)));
      PrintChart->resize(printer->width(),ChartArea);
      ChartArea = PrintChart->height();
    }else{
      ChartArea = Round( (int)Splitter->sizes().at(0) * printer->width()
                        /((int)Splitter->sizes().at(0)
                         +(int)Splitter->sizes().at(1)));
      PrintChart->resize(ChartArea,printer->height());
      ChartArea = PrintChart->width();
    }
    PrintChart->UpdateAxes();
    PrintChart->Print(&Painter);
    delete PrintChart;
    ChartArea += 10;
  }

  PrintLegend = new sNavigator(Evolution);
  if( (int)Splitter->sizes().at(1) == 0 ){
    PrintLegend->resize(printer->width(),printer->height());
    printer->newPage();
  }else if( Splitter->orientation() == Qt::Vertical ){
    PrintLegend->resize(printer->width(),printer->height()-ChartArea);
    Painter.translate(QPoint(0,ChartArea));
  }else{
    PrintLegend->resize(printer->width()-ChartArea,printer->height());
    Painter.translate(QPoint(ChartArea,0));
  }
  PrintLegend->Print(&Painter);
  delete PrintLegend;

  Painter.end();
}

//------------------------------------------------------------------------------
} //namespace RA

