////////////////////////////////////////////////////////////////////////////////
/*! @file RAMainWindow.cpp   Defines common main window class
                           of the Research Assistant program.
- Part of RAGUI - Research Assistant Graphical User Interface Library.
- Uses  RANet - Research Assistant Net Library.
- Uses  Qt v.5  -  http://qt-project.org/
- Copyright(C) 2010-2015, Viktor E. Bursian, St.Petersburg, Russia.
                          Viktor_dot_Bursian_at_mail_dot_ioffe_dot_ru
*///////////////////////////////////////////////////////////////////////////////
#include "Environment.h"
#include "SymbolsGUI.h"
#include "StringsAndQStrings.h"
#include "Net.h"
#include "attr.h"
#include "ActionSignallingText.h"
#include "UnitSystemIO.h"
#include "ColorNodes.h"
#include "Record.h"
#include "Helper.h"
#include "RAMainWindow.h"
#include "Window.h"
#include "RADock.h"
#include "Workbook.h"
#include "Plot.h"
#include "Console.h"
#include "ColorsUI.h"
#include "ConfigureDeviceDialog.h"
#include "AboutRA.h"
#include "Restyling.h"
#include "Version.h"
#include "Log.h"
#include "Trace.h"
#include <QMessageBox>
#include <QDir>
#include <QPluginLoader>
#include <QMenuBar>
#include <QStatusBar>
#include <QApplication>
#include <QLibraryInfo>
#include <QFontDialog>

#include "NetExceptions.h"

namespace RA {
//------------------------------------------------------------------------------

csAttrName           sRAMainWindow::ChildrenAttrName ("Children");
csAttrName           sRAMainWindow::UserAccessLevelAttrName ("UserAccessLevel");
csAttrName           sRAMainWindow::NightRegimeAttrName ("NightRegime");
csAttrName           sRAMainWindow::DockNestingAttrName ("DockNesting");
csAttrName           sRAMainWindow::DesktopGeometryAttrName ("Geometry");
csAttrName           sRAMainWindow::DesktopStateAttrName ("State");

//------------------------------------------------------------ sRAMainWindow ---

sRAMainWindow::~sRAMainWindow ()
{
  RANet::Log.Put(sLog::Debug,"Quitting","sRAMainWindow destructor");
  TheRoot.SetToNULL();
  Desktop.SetToNULL();
  DataToBrowse.SetToNULL();
  Plugins.SetToNULL();
  FontFamily.SetToNULL();
  FontSize.SetToNULL();
  Styles.SetToNULL();
  Language.SetToNULL();
  UnitSystem.SetToNULL();
  RecordsToBrowse.SetToNULL();
  SamplesToBrowse.SetToNULL();
  TopicsToBrowse.SetToNULL();
  Children.SetToNULL();
  /*! @todo{Net} Who will release sDevice::DeviceList
                 and sRecord::RecordNumberingData
                 and sRecord::ChronologicalStorage? */
}


sRAMainWindow::sRAMainWindow (sNodePtr       root
                             ,psByeScreen    bye_screen)
    :QMainWindow(NULL)
    ,TheRoot(root)
    ,SuccessfullyConstructed(false)
    ,QuittingStarted(false)
    ,QuittingAllowed(false)
    ,ByeScreen(bye_screen)
    ,ClosingTimer(new QTimer(this))
//    ,IgnoreLanguageChangeEvent(true)
    ,QtTranslator(new QTranslator(this))
    ,Translator(new QTranslator(this))
//    ,ToolBarPlot(NULL)
    ,MenuFile(NULL)
//    ,MenuEdit(NULL)
//    ,MenuWorkbook(NULL)
//    ,MenuPlot(NULL)
    ,MenuInstrument(NULL)
    ,MenuSettings(NULL)
    ,MenuLanguage(NULL)
    ,MenuUser(NULL)
    ,MenuInstallNewDevice(NULL)
    ,MenuInstallNewDriver(NULL)
    ,MenuInstallNewConsole(NULL)
    ,MenuInstallNewGadget(NULL)
    ,MenuConfigureDevice(NULL)
//    ,MenuWindow(NULL)
    ,MenuHelp(NULL)
    ,MenuAboutPlugins(NULL)
{
  try{
    setAttribute(Qt::WA_DeleteOnClose);

    CentralWidget = new QFrame(this);
    CentralWidget->setObjectName(QString::fromUtf8("CentralWidget"));
    setCentralWidget(CentralWidget);

    setWindowIcon(QIcon(":/RAGUI/icons/ra.png"));

    qApp->installTranslator(QtTranslator);//each posts QEvent::LanguageChange,
    qApp->installTranslator(Translator);  //only one of which remains in the queue

    CreateActions();
    CreateMenus();
    CreateToolBars();
    CreateStatusBar();

  }catch( rcxException E ){
    E.Report2cerr();
    throw;
  }
}


void sRAMainWindow::showEvent (QShowEvent *  event)
{
  static bool                 FirstTime = true;
  QMainWindow::showEvent(event);
  if( FirstTime ){
    FirstTime = false;
    connect(this,SIGNAL(WindowOpened()),this,SLOT(CompleteConstruction())
           ,Qt::QueuedConnection);
    QTimer::singleShot(50,this,SIGNAL(WindowOpened()));
  }
}


void  sRAMainWindow::CompleteConstruction ()
{
  try{
    if( ! SetMainDataPtrs() ){
      QMessageBox::critical(this,tr("RA launching error")
          ,tr("Some attributes are missing.  Check it with RAdmin."));
      qApp->postEvent(this,new QEvent(QEvent::Close));
    }else{
      SetLanguage();
      FillMenuLanguage();
      Retranslate();
      RestoreLayout();

      UnitSystemLoad(UnitSystem);
      LoadPlugins ();

      RANet::Log.Put(sLog::Debug,"locale","RAMainWindow: restoring desktop");
      RestoreDesktop();
      SuccessfullyConstructed = true;
    }
  }catch( rcxException E ){
    E.Report2cerr();
    throw;
  }
}


void  sRAMainWindow::CheckClosingChildren ()
{
//  RANet::Log.Put(sLog::Debug,"Quitting","sRAMainWindow::CheckClosingChildren");
  if( QuittingStarted && findChildren<psRADock>().empty()
                      && sDevice::NoneIsRunning() ){
    ClosingTimer->stop();
    RANet::Log.Put(sLog::Debug,"Quitting","sRAMainWindow::CheckClosingChildren -- OK");
    emit QuittingIsAllowed();
  }
}


void  sRAMainWindow::FinishQuitting ()
{
  if( QuittingStarted ){
    RANet::Log.Put(sLog::Debug,"Quitting","sRAMainWindow::FinishQuitting");
    QuittingAllowed = true;
    close();
  }
}


void  sRAMainWindow::closeEvent (QCloseEvent *  event)
{
  if( QuittingAllowed ){
    RANet::Log.Put(sLog::Debug,"Quitting"
                  ,"sRAMainWindow::closeEvent accepted");
    event->accept();
  }else if( ! QuittingStarted ){
    bool                            CanClose=true;
    QList<psRADock>                 Children(findChildren<psRADock>());
    // before exit, ask children if they agree
    foreach( psRADock Dock, Children ){
      if( ! Dock->Window()->CanClose() ){
        CanClose=false;
      }
    }
    if( ! CanClose ){
      foreach( psRADock Dock, Children )
        Dock->Window()->DonotClose();
    }else{
      QuittingStarted = true;
      setEnabled(false);
      connect(this,SIGNAL(QuittingIsAllowed())
             ,this,SLOT(FinishQuitting())
             ,Qt::QueuedConnection);
      if( ByeScreen ){
        ByeScreen->show();
        connect(this,SIGNAL(QuittingIsAllowed())
               ,ByeScreen,SLOT(close())
               ,Qt::QueuedConnection);
      }
      if( SuccessfullyConstructed ){
        StoreDesktop();
        UnitSystemStore(UnitSystem);
      }
      foreach( psRADock Dock, Children )
        Dock->StartQuitting();
      connect(ClosingTimer,SIGNAL(timeout())
             ,this,SLOT(CheckClosingChildren()));
      ClosingTimer->start(TIME_PERIOD_FOR_sRAMainWindow_closing_check);
    }
    event->ignore();
  }else{
    RANet::Log.Put(sLog::Debug,"Quitting"
                  ,"sRAMainWindow::closeEvent ignored");
    event->ignore();
  }
}


void  sRAMainWindow::focusInEvent (QFocusEvent *  event)
{
  sString sRAMainWindow_focusInEvent(FromQString(focusWidget()->objectName()));
  TRACE(sRAMainWindow_focusInEvent);
  QMainWindow::focusInEvent(event);
}


void  sRAMainWindow::focusOutEvent (QFocusEvent *  event)
{
  sString sRAMainWindow_focusOutEvent;
  TRACE(sRAMainWindow_focusOutEvent);
  QMainWindow::focusOutEvent(event);
}


bool  sRAMainWindow::SetMainDataPtrs ()
{
  try{
    Desktop                       = TheRoot >> GlobalAttr::Desktop;
    DataToBrowse                  = TheRoot >> GlobalAttr::Data;
    Plugins                       = TheRoot >> GlobalAttr::Plugins;
    sDevice::DeviceList           = TheRoot >> GlobalAttr::Devices;
    FontFamily                    = TheRoot >> GlobalAttr::FontFamily;
    FontSize                      = TheRoot >> GlobalAttr::FontSize;
    Styles                        = TheRoot >> GlobalAttr::Styles;
    RAGUI::LineColorSet           = TheRoot >> GlobalAttr::LineColorSet;
    RAGUI::CustomColors           = TheRoot >> GlobalAttr::CustomColors;
    Language                      = TheRoot >> GlobalAttr::Language;
    UnitSystem                    = TheRoot >> GlobalAttr::UnitSystem;
    sRecord::RecordNumberingData  = TheRoot >> GlobalAttr::RecordNumbering;
    sRecord::ChronologicalStorage = TheRoot >> GlobalAttr::RecordHistory;
    RecordsToBrowse               = TheRoot >> GlobalAttr::RecordsToBrowse;
    SamplesToBrowse               = TheRoot >> GlobalAttr::Samples;
    TopicsToBrowse                = TheRoot >> GlobalAttr::Topics;
    RAGUI::CSVExportSettings      = TheRoot >> GlobalAttr::CSVExportSettings;
    RAGUI::CSVImportSettings      = TheRoot >> GlobalAttr::CSVImportSettings;
    RAGUI::SOSImportSettings      = TheRoot >> GlobalAttr::SOSImportSettings;
  }catch(...){
    return false;
  }
  return Desktop.IsCorrect()
      && DataToBrowse.IsCorrect()
      && Plugins.IsCorrect()
      && sDevice::DeviceList.IsCorrect()
      && FontFamily.IsCorrect()
      && FontSize.IsCorrect()
      // Styles is optional
      && RAGUI::LineColorSet.IsCorrect()
      && RAGUI::CustomColors.IsCorrect()
      && Language.IsCorrect()
      && UnitSystem.IsCorrect()
      && sRecord::RecordNumberingData.IsCorrect()
      && sRecord::ChronologicalStorage.IsCorrect()
      && RecordsToBrowse.IsCorrect()
      && SamplesToBrowse.IsCorrect()
      && TopicsToBrowse.IsCorrect()
      //!@patch{UI} RAGUI::CSVExportSettings is optional
      // RAGUI::CSVImportSettings is optional
      // RAGUI::SOSImportSettings is optional
      ;
}


void  sRAMainWindow::RestoreLayout ()
{
  bool                        DockNesting = isDockNestingEnabled();
  get_from<sBooleanNode>(Desktop>>DockNestingAttrName,DockNesting);
  setDockNestingEnabled(DockNesting);
  ActnSetDockNesting->setChecked(DockNesting);

  std::vector<byte>           DesktopGeometryData;
  std::vector<byte>           DesktopStateData;
  if( (  get_from<sByteArrayNode>(Desktop>>DesktopGeometryAttrName
                                          ,DesktopGeometryData)
      && get_from<sByteArrayNode>(Desktop>>DesktopStateAttrName
                                          ,DesktopStateData) ) ){
    restoreGeometry(QByteArray( (const char *)(DesktopGeometryData.data())
                              , DesktopGeometryData.size() ));
    restoreState(QByteArray( (const char *)(DesktopStateData.data())
                           , DesktopStateData.size() ));
  }
//  Desktop |= "MainWindow";
//  (Desktop>>"MainWindow") |= attr("Top",0);
//  (Desktop>>"MainWindow") |= attr("Left",0);
//  (Desktop>>"MainWindow") |= attr("Width",600);
//  (Desktop>>"MainWindow") |= attr("Height",460);
//  (Desktop>>"MainWindow") |= attr("WindowState",0/*int(wsNormal)*/);
//  resize(rsIntegerNode(Desktop>>"MainWindow">>"Width")
//        ,rsIntegerNode(Desktop>>"MainWindow">>"Height") );
//  move(rsIntegerNode(Desktop>>"MainWindow">>"Left")
//      ,rsIntegerNode(Desktop>>"MainWindow">>"Top")  );
// //  WindowState=TWindowState(int( rsIntegerNode(Desktop>>"MainWindow">>"WindowState") ));
}


void  sRAMainWindow::StoreLayout ()
{
//  if( WindowState == wsMinimized )  WindowState=wsNormal;
//  (Desktop>>"MainWindow") << attr("WindowState",int(WindowState));
//  if( WindowState == wsMaximized )  WindowState=wsNormal;
//  (Desktop>>"MainWindow") << attr("Left",pos().x())
//                          << attr("Top",pos().y())
//                          << attr("Width",size().width())
//                          << attr("Height",size().height());
  Desktop << attr(DockNestingAttrName,isDockNestingEnabled());
  QByteArray                  BA(saveGeometry());
  Desktop << attr(DesktopGeometryAttrName
                 ,sNodePtr(new sByteArrayNode(BA.constData(),BA.size())));
  BA = saveState();
  Desktop << attr(DesktopStateAttrName
                 ,sNodePtr(new sByteArrayNode(BA.constData(),BA.size())));
}


void sRAMainWindow::RestoreDesktop ()
{
  QFont                       Font(font());
  Font.setFamily(ToQString(FontFamily->Value()));
  Font.setPointSize(FontSize->Value());
  //!@bug При запуске размер шрифта (и сам шрифт?) системный.
  qApp->setFont(Font);

  Children = (Desktop >= ChildrenAttrName);
  RestoreWindows(Children);
  RestoreLayout();
  foreach (QDockWidget * Dock, findChildren<QDockWidget *>()) {
//    restoreDockWidget(Dock);
    psWindow Window = qobject_cast<psWindow>(Dock->widget());
    if( Window ){
      //!@bug Second call of ResoreState() for each child
      Window->RestoreState();
    }
  }
  Desktop |= attr(UserAccessLevelAttrName,(int)sUser::cNewbie);
  sPtr<sIntegerNode>          U(Desktop >> UserAccessLevelAttrName);
  if( U.IsCorrect() ){
    ChangeUserAccessLevel( (sUser::eAccessLevel)(int)(*U) );
  }
  get_from<sBooleanNode>(Desktop>>NightRegimeAttrName,RAGUI::NightRegime);
  SetNightRegime(RAGUI::NightRegime);
  ActnSetNightRegime->setChecked(RAGUI::NightRegime);
  RANet::Log.Put(sLog::Debug,"locale","RAMainWindow::RestoreDesktop - done");
}


void sRAMainWindow::StoreDesktop ()
{
  StoreLayout();
  foreach (QDockWidget * Dock, findChildren<QDockWidget *>()) {
    psWindow Window = qobject_cast<psWindow>(Dock->widget());
    if( Window ){
      Window->StoreState();
    }
  }
  ClearClosedWindows(Children);
  Desktop << attr(UserAccessLevelAttrName ,(int)sEnvironment::User.AccessLevel);
}


void sRAMainWindow::LoadPlugins ()  //!@bug move LoadPlugins() out to main()?
{
  QDir pluginsDir;
  pluginsDir = QDir(qApp->applicationDirPath());
  #if defined(Q_OS_LINUX)
    pluginsDir.setNameFilters( QStringList() << "librap*.so" );
  #elif defined(Q_OS_WIN)
    pluginsDir.setNameFilters( QStringList() << "rap*.dll" );
  #endif
  foreach( QString fileName, pluginsDir.entryList(QDir::Files) ){
    QPluginLoader loader(pluginsDir.absoluteFilePath(fileName));
    QObject *plugin = loader.instance();
    if( ! plugin ){
      RANet::Log.Put(sLog::Warning,"Plugins"
                  ,sString("sRAMainWindow::LoadPlugins: file found: '")
                  +FromQString(fileName)
                  +sString("', but cannot be loaded.")
                  );
    }else{
      psInstrumentalPackage  IP = qobject_cast<psInstrumentalPackage>(plugin);
      if( IP ){
        RANet::Log.Put(sLog::Debug,"Plugins"
            ,sString("sRAMainWindow::LoadPlugins: InstrumentalPackage: ")
            +IP->TrueType()->Name() );
        if( RANet::Version.Principal() < IP->CoreVersionMin().Principal() ){
          QMessageBox::critical(this,tr("Version mismatch")
              ,tr("Plugin '%1' version %2 is incompatible with the RA core "
              "program version %3.  Please upgrade RA core to at least version "
              "%4.  The program will now terminate.")
              .arg(ToQString(IP->TrueType()->Name().Tail(1))
                  ,ToQString(IP->Version().Text())
                  ,ToQString(RANet::Version.Principal().Text())
                  ,ToQString(IP->CoreVersionMin().Principal().Text())  )
                               );
          throw xProgramError("Version mismatch",__FILE__,__LINE__);
        }
        if( IP->CoreVersionUsed().Principal() < RANet::Version.Principal() ){
          QMessageBox::critical(this,tr("Version mismatch")
              ,tr("Plugin '%1' version %2 is not known to be compatible with "
              "the RA core program version %3.  Please upgrade the plugin.  "
              "The program will now terminate.")
              .arg(ToQString(IP->TrueType()->Name().Tail(1))
                  ,ToQString(IP->Version().Text())
                  ,ToQString(RANet::Version.Principal().Text())  )
                               );
          throw xProgramError("Version mismatch",__FILE__,__LINE__);
        }
        LoadedPackages.push_back(IP);
        if( IP->HasDefaults() ){
          InstallPluginDefaults(IP);
        }
        psActionSignallingText   Actn = new sActionSignallingText(this);
        Actn->setText(ToQString(IP->TrueType()->Name().Tail(1)));
        Actn->setStatusTip(IP->UserFriendlyName());
        Actn->SetDataString(ToQString(IP->TrueType()->Name().Tail(1))
                           +QString(" version ")
                           +ToQString(IP->Version().Text())
                           +QString(".\n\"")
                           +IP->UserFriendlyName()
                           +QString("\".\n\n")
                           +IP->About()
                           );
        MenuAboutPlugins->addAction(Actn);
        connect(Actn,SIGNAL(DataStringSignal(QString))
               ,this,SLOT(AboutPlugin(QString)));
      }
    }
  }
//    brushMenu->setEnabled(!brushActionGroup->actions().isEmpty());
//    shapesMenu->setEnabled(!shapesMenu->actions().isEmpty());
}


void sRAMainWindow::CreateActions ()
{
//  ActnNewWorkbook = new QAction(this);
//  ActnNewWorkbook->setShortcut(QKeySequence("Ctrl+B"));
//  ActnNewWorkbook->setIcon(QIcon(":/RAGUI/icons/book.png"));
//  connect(ActnNewWorkbook,SIGNAL(triggered()),this,SLOT(NewWorkbook()));

  ActnNewPlot = new QAction(this);
  ActnNewPlot->setShortcut(QKeySequence("Ctrl+O"));
  ActnNewPlot->setIcon(QIcon(":/RAGUI/icons/plot.png"));
  connect(ActnNewPlot,SIGNAL(triggered()),this,SLOT(NewPlot()));

  ActnOpenTopics = new QAction(this);
  ActnOpenTopics->setShortcut(QKeySequence("Ctrl+T"));
  ActnOpenTopics->setIcon(QIcon(":/RAGUI/icons/book.png"));
  connect(ActnOpenTopics,SIGNAL(triggered()),this,SLOT(OpenTopics()));

  ActnOpenSamples = new QAction(this);
  ActnOpenSamples->setShortcut(QKeySequence("Ctrl+S"));
  connect(ActnOpenSamples,SIGNAL(triggered()),this,SLOT(OpenSamples()));

  ActnOpenRecords = new QAction(this);
  connect(ActnOpenRecords,SIGNAL(triggered()),this,SLOT(OpenRecords()));

  ActnOpenData = new QAction(this);
  connect(ActnOpenData,SIGNAL(triggered()),this,SLOT(OpenData()));

//  ActnPrint = new QAction(this);
//  ActnPrint->setShortcuts(QKeySequence::Print);
//  ActnPrint->setIcon(QIcon(":/RAGUI/icons/print.png"));
//  connect(ActnPrint,SIGNAL(triggered()),this,SLOT

  ActnExit = new QAction(this);
  ActnExit->setShortcuts(QKeySequence::Quit);
//  connect(ActnExit, SIGNAL(triggered()), qApp, SLOT(closeAllWindows()));
  connect(ActnExit, SIGNAL(triggered()), this, SLOT(close()));

//  ActnUndo = new QAction(QIcon(":/RAGUI/icons/undo.png"), tr("Undo"), this);
//  ActnUndo->setShortcuts(QKeySequence::Undo);
//  ActnUndo->setStatusTip(tr("Undo"));
//  connect(actionUndo, SIGNAL(triggered()), this, SLOT(Undo()));

//  ActnRedo = new QAction(QIcon(":/RAGUI/icons/redo.png"), tr("Redo"), this);
//  ActnRedo->setShortcuts(QKeySequence::Redo);
//  ActnRedo->setStatusTip(tr("Redo"));
//  connect(actionRedo, SIGNAL(triggered()), this, SLOT(Redo()));

//  ActnCut = new QAction(QIcon(":/RAGUI/icons/cut.png"), tr("Cu&t"), this);
//  actionCut->setShortcuts(QKeySequence::Cut);
//  actionCut->setStatusTip(tr("Cut the current selection's contents to the "
//                             "clipboard"));
//  connect(actionCut, SIGNAL(triggered()), this, SLOT(cut()));

//  ActnCopy = new QAction(QIcon(":/RAGUI/icons/copy.png"), tr("&Copy"), this);
//  actionCopy->setShortcuts(QKeySequence::Copy);
//  actionCopy->setStatusTip(tr("Copy the current selection's contents to the "
//                              "clipboard"));
//  connect(actionCopy, SIGNAL(triggered()), this, SLOT(copy()));

//  ActnPaste = new QAction(QIcon(":/RAGUI/icons/paste.png"), tr("&Paste"), this);
//  actionPaste->setShortcuts(QKeySequence::Paste);
//  actionPaste->setStatusTip(tr("Paste the clipboard's contents into the current "
//                               "selection"));
//  connect(actionPaste, SIGNAL(triggered()), this, SLOT(paste()));

//  ActnSeparator = new QAction(this);
//  ActnSeparator->setSeparator(true);


  QActionGroup *              UserActions;
  UserActions = new QActionGroup(this);
  UserActions->setExclusive(true);

  ActnNewbieLevel = new sActionSignallingText(this);
  ActnNewbieLevel->SetDataString("cNewbie");
  ActnNewbieLevel->setCheckable(true);
  connect(ActnNewbieLevel,SIGNAL(DataStringSignal(QString))
         ,this,SLOT(ChangeUser(QString)));
  UserActions->addAction(ActnNewbieLevel);

  ActnExpertLevel = new sActionSignallingText(this);
  ActnExpertLevel->SetDataString("cExpert");
  ActnExpertLevel->setCheckable(true);
  connect(ActnExpertLevel,SIGNAL(DataStringSignal(QString))
         ,this,SLOT(ChangeUser(QString)));
  UserActions->addAction(ActnExpertLevel);

  ActnAdminLevel = new sActionSignallingText(this);
  ActnAdminLevel->SetDataString("cAdmin");
  ActnAdminLevel->setCheckable(true);
  connect(ActnAdminLevel,SIGNAL(DataStringSignal(QString))
         ,this,SLOT(ChangeUser(QString)));
  UserActions->addAction(ActnAdminLevel);

  ActnSetNightRegime = new QAction(this);
  ActnSetNightRegime->setCheckable(true);
  ActnSetNightRegime->setShortcut(QKeySequence("Ctrl+L"));
  connect(ActnSetNightRegime,SIGNAL(triggered(bool))
         ,this,SLOT(SetNightRegime(bool)));

  ActnSetDockNesting = new QAction(this);
  ActnSetDockNesting->setCheckable(true);
  connect(ActnSetDockNesting,SIGNAL(triggered(bool))
         ,this,SLOT(setDockNestingEnabled(bool)));

  ActnChangeBasicLineColors = new QAction(this);
  connect(ActnChangeBasicLineColors,SIGNAL(triggered())
         ,this,SLOT(ChangeBasicLineColors()));

  ActnChangeFont = new QAction(this);
  connect(ActnChangeFont,SIGNAL(triggered()),this,SLOT(ChangeFont()));

  ActnHelpContents = new QAction(this);
  ActnHelpContents->setShortcut(QKeySequence::HelpContents);
  connect(ActnHelpContents,SIGNAL(triggered()),this,SLOT(ShowHelpContents()));

  ActnAbout = new QAction(this);
  connect(ActnAbout, SIGNAL(triggered()), this, SLOT(About()));

  ActnAboutQt = new QAction(this);
  connect(ActnAboutQt, SIGNAL(triggered()), qApp, SLOT(aboutQt()));

  WindowListActions = new QActionGroup(this);
  WindowListActions->setExclusive(true);
}


void sRAMainWindow::CreateMenus ()
{
  MenuFile = menuBar()->addMenu("");
  MenuFile->addAction(ActnNewPlot);
  MenuFile->addSeparator();
  MenuFile->addAction(ActnOpenTopics);
  MenuFile->addAction(ActnOpenSamples);
  if( sEnvironment::User.MaxAccessLevel != sUser::cNewbie ){
    MenuFile->addSeparator();
    MenuFile->addAction(ActnOpenRecords);
    MenuFile->addAction(ActnOpenData);
  }
  //MenuFile->addSeparator();
  //MenuFile->addAction(ActnPrint);
  MenuFile->addSeparator();
  MenuFile->addAction(ActnExit);

//  MenuEdit = menuBar()->addMenu("");
//  MenuEdit->addAction(ActnUndo);
//  MenuEdit->addAction(ActnRedo);
//  MenuEdit->addSeparator();
//  MenuEdit->addAction(ActnCut);
//  MenuEdit->addAction(ActnCopy);
//  MenuEdit->addAction(ActnPaste);

//  MenuWorkbook = menuBar()->addMenu("");

//  MenuPlot = menuBar()->addMenu("");
//  QActionGroup * CurrentPlotActions;
//  CurrentPlotActions = new QActionGroup(this);
//  CurrentPlotActions->addAction("1");
//  CurrentPlotActions->addAction("2");
//  MenuPlot->addActions(CurrentPlotActions->actions());

  MenuInstrument = menuBar()->addMenu("");
  connect(MenuInstrument,SIGNAL(aboutToShow())
         ,this,SLOT(UpdateMenuInstrument()));

  MenuSettings = menuBar()->addMenu("");
  MenuLanguage = new QMenu(this);
  MenuSettings->addMenu(MenuLanguage);
  MenuUser = new QMenu(this);
  MenuSettings->addMenu(MenuUser);
  MenuUser->addAction(ActnNewbieLevel);
  MenuUser->addAction(ActnExpertLevel);
  MenuUser->addAction(ActnAdminLevel);
  MenuSettings->addAction(ActnSetNightRegime);
  MenuSettings->addAction(ActnSetDockNesting);
  MenuSettings->addAction(ActnChangeBasicLineColors);
  MenuSettings->addAction(ActnChangeFont);
  MenuSettings->addSeparator();
  MenuInstallNewDevice = new QMenu(this);
  MenuInstallNewDriver = new QMenu(this);
  MenuInstallNewDevice->addMenu(MenuInstallNewDriver);
  MenuInstallNewConsole = new QMenu(this);
  MenuInstallNewDevice->addMenu(MenuInstallNewConsole);
  MenuInstallNewGadget = new QMenu(this);
  MenuInstallNewDevice->addMenu(MenuInstallNewGadget);
  MenuSettings->addMenu(MenuInstallNewDevice);
  connect(MenuSettings,SIGNAL(aboutToShow())
         ,this,SLOT(UpdateMenuInstallNewDevice()));
  MenuConfigureDevice = new QMenu(this);
  MenuSettings->addMenu(MenuConfigureDevice);
  connect(MenuSettings,SIGNAL(aboutToShow())
         ,this,SLOT(UpdateMenuConfigureDevice()));

//  MenuWindow = menuBar()->addMenu("");

  MenuHelp = menuBar()->addMenu("");
  MenuHelp->addAction(ActnHelpContents);
  MenuHelp->addSeparator();
  MenuHelp->addAction(ActnAbout);
  MenuAboutPlugins = new QMenu(this);
  MenuHelp->addMenu(MenuAboutPlugins);
  MenuHelp->addAction(ActnAboutQt);
}


void sRAMainWindow::UpdateMenuInstrument ()
{
  if( ! MenuInstrument )
    return;
  MenuInstrument->clear();
  for_each_attribute( AF , sDevice::DeviceList ){
    sPtr<sBooleanNode>          Disabled(AF->Value() >> "Disabled");
    sPtr<sTextNode>             ClassName(AF->Value() >> "Class");
    if( Disabled.IsCorrect() && ClassName.IsCorrect() ){
      if( ! bool(*Disabled) && sInstrument::IsConsole((sString)(*ClassName)) ){
        psActionSignallingText     Actn = new sActionSignallingText(this);
        sString                    DeviceId(AF->Name().Text());
        Actn->setText(ToQString(DeviceId));
        MenuInstrument->addAction(Actn);
        if( sDevice::IsRunning(DeviceId) ){
          Actn->setEnabled(false);
        }else{
          connect(Actn,SIGNAL(TitleSignal(QString))
                 ,this,SLOT(OpenConsole(QString)));
        }
      }
    }
  }end_attr_loop()
}


void sRAMainWindow::OpenInstrument (QString Id)
{
  sNodePtr                    Cfg;
//  sNet::Manager->Undo.GroupOpen();
  Children << (Cfg = sNodePtr(new sFolderNode()) );
  Cfg << attr("Type","Instrument")
      << attr("Title",FromQString(Id))
      << attr("DockArea",int(Qt::TopDockWidgetArea));
  CreateWindow(Cfg);
//  sNet::Manager->Undo.GroupFlush();
}


void sRAMainWindow::OpenConsole (QString Id)
{
  sNodePtr                    Cfg;
//  sNet::Manager->Undo.GroupOpen();
  Children << (Cfg = sNodePtr(new sFolderNode()) );
  Cfg << attr("Type","Console")
      << attr("Title",FromQString(Id))
      << attr("DockArea",int(Qt::TopDockWidgetArea));
  CreateWindow(Cfg);
//  sNet::Manager->Undo.GroupFlush();
}


void sRAMainWindow::UpdateMenuInstallNewDevice ()
{
  MenuInstallNewDriver->clear();
  MenuInstallNewConsole->clear();
  MenuInstallNewGadget->clear();
  for( map<sString,eInstrumentType>::const_iterator
                                i = sInstrumentTag::AvailableClassList.begin()
                              ; i != sInstrumentTag::AvailableClassList.end()
                              ; ++i                                    ){
    psActionSignallingText     Actn = new sActionSignallingText(this);
    sString                       ClassName = i->first;
    Actn->setText(ToQString(ClassName));
    switch( i->second ){
      case cDriver:
        MenuInstallNewDriver->addAction(Actn);
        break;
      case cConsole:
        MenuInstallNewConsole->addAction(Actn);
        break;
      case cGadget:
        MenuInstallNewGadget->addAction(Actn);
        break;
      default:
        break;
    }
    connect(Actn,SIGNAL(TitleSignal(QString))
           ,this,SLOT(InstallNewDevice(QString)));
  }
  MenuInstallNewDriver
      ->setEnabled( ! MenuInstallNewDriver->actions().empty()
                    && ( sEnvironment::User.AccessLevel > sUser::cNewbie ) );
  MenuInstallNewConsole
      ->setEnabled( ! MenuInstallNewConsole->actions().empty()
                    && ( sEnvironment::User.AccessLevel > sUser::cNewbie ) );
  MenuInstallNewGadget
      ->setEnabled( ! MenuInstallNewGadget->actions().empty() );
  MenuInstallNewDevice
      ->setEnabled(   MenuInstallNewDriver->isEnabled()
                   || MenuInstallNewConsole->isEnabled()
                   || MenuInstallNewGadget->isEnabled() );
}


void sRAMainWindow::InstallNewDevice (QString ClassName)
{
  sString                 DeviceId
                              = sDevice::InstallNew(FromQString(ClassName));
  sConfigureDeviceDialog *    MaintainDeviceDlg
                              = new sConfigureDeviceDialog(DeviceId,this);
  MaintainDeviceDlg->setAttribute(Qt::WA_DeleteOnClose);
  MaintainDeviceDlg->exec();
}


void sRAMainWindow::UpdateMenuConfigureDevice ()
{
  if( ! MenuConfigureDevice )
    return;
  MenuConfigureDevice->clear();
  for_each_attribute( AF , sDevice::DeviceList ){
    psActionSignallingText      Actn = new sActionSignallingText(this);
    sString                     DeviceId(AF->Name().Text());
    Actn->setText(ToQString(DeviceId));
    MenuConfigureDevice->addAction(Actn);
    if( sDevice::IsRunning(DeviceId) ){
      Actn->setEnabled(false);
    }else{
      connect(Actn,SIGNAL(TitleSignal(QString))
             ,this,SLOT(ConfigureDevice(QString)));
    }
  }end_attr_loop()
  MenuConfigureDevice
      ->setEnabled( ! MenuConfigureDevice->actions().empty()
                    && (sEnvironment::User.AccessLevel > sUser::cNewbie ) );
}


void sRAMainWindow::ConfigureDevice (QString Id)
{
  psConfigureDeviceDialog       Dlg = new sConfigureDeviceDialog(FromQString(Id)
                                                          ,this);
  Dlg->setAttribute(Qt::WA_DeleteOnClose);
  Dlg->exec();
}


void  sRAMainWindow::InstallPluginDefaults (psInstrumentalPackage  IP)
{
  bool                        Done = false;
  sString                     LastVerText;
  if( get_from<sTextNode>(Plugins >> IP->TrueType()->Name() >> "Version"
                         ,LastVerText) ){
    sVersion                    LastVersion(LastVerText);
    if( LastVersion.Valuable() < IP->Version().Valuable() ){
      IP->ModifyConfig(LastVersion,Plugins);
    }else if( IP->Version().Valuable() < LastVersion.Valuable() ){
      QMessageBox::critical(this,tr("Plugin version downshift")
          ,tr("Plugin '%1' version %2 may not work with configuration settings "
              "that seem to be done for plugin version %3. "
              "The program will now terminate.")
              .arg(ToQString(IP->TrueType()->Name().Tail(1))
                  ,ToQString(IP->Version().Text())
                  ,ToQString(LastVersion.Text())  )
                            );
      throw xProgramError("Plugin version downshift",__FILE__,__LINE__);
    }
  }else{
    QMessageBox               D;
    D.setWindowTitle(tr("New instrumental plugin"));
    D.setIcon(QMessageBox::Question);
    D.setText(tr("New instrumental plugin has been found:")
             +QString("\n\n")
             +ToQString(IP->TrueType()->Name())
             +QString("\n")
             +IP->UserFriendlyName()
             +QString("\n\n")
             +tr("Whould you like to perform a default installation "
                 "for this plugin?"));
    D.setDetailedText(tr("The plugin may have a default installation "
                         "instructions.  These instructions will "
                         "install one or more devices with default "
                         "settings.  You will be able to review and "
                         "change any of these settings, as well as "
                         "uninstall any device through the "
                         "Settings menu at any time.  Also, you can "
                         "install your devices manually one-by-one.\n\n"
                         "It is recommended, however, to start with "
                         "the default installation provided with this "
                         "plugin."));
//    QAbstractButton *  AskBtn;
    QAbstractButton *  NoBtn;
    QAbstractButton *  YesBtn;
    /*AskBtn=*/D.addButton(tr("Ask me next time"),QMessageBox::NoRole);
    NoBtn=D.addButton(tr("No, I will install every device manually")
                     ,QMessageBox::NoRole);
    YesBtn=D.addButton(tr("Yes, install default devices")
                      ,QMessageBox::NoRole);
//    qApp->sendEvent(&D,new QKeyEvent(QEvent::KeyPress,Qt::Key_Space,Qt::NoModifier));
//    qApp->sendEvent(&D,new QKeyEvent(QEvent::KeyRelease,Qt::Key_Space,Qt::NoModifier));
    D.exec();
    if( D.clickedButton() == YesBtn ){
      IP->InstallDefaults(sDevice::DeviceList);
      Done=true;
    }else if( D.clickedButton() == NoBtn ){
      Done=true;
    }
  }
  if( Done ){
    sPtr<sFolderNode>          P = (Plugins >= IP->TrueType()->Name());
    P << attr("Version",IP->Version().Valuable().Text());
  }
}


void  sRAMainWindow::AboutPlugin (QString AboutText)
{
  QMessageBox                 A(this);
  A.setWindowTitle(tr("About the RA Plugin"));
  A.setText(AboutText);
  A.exec();
}


void sRAMainWindow::CreateToolBars ()
{
//    fileToolBar = addToolBar(tr("File"));
//    fileToolBar->addAction(newAct);
//    fileToolBar->addAction(openAct);
//    fileToolBar->addAction(saveAct);
//
//    editToolBar = addToolBar(tr("Edit"));
//    editToolBar->addAction(cutAct);
//    editToolBar->addAction(copyAct);
//    editToolBar->addAction(pasteAct);

//  ToolBarPlot = new QToolBar("Plot Toolbar",this);
//  ToolBarPlot->setObjectName(QString::fromUtf8("ToolBarPlot"));
//  addToolBar(Qt::TopToolBarArea, ToolBarPlot);
//  ToolBarPlot->setFloatable(false);
}


void sRAMainWindow::CreateStatusBar ()
{
//  statusBar()->showMessage(tr("Ready"));
  statusBar()
      ->setStyleSheet("QLabel {background: white; border: 1px solid grey}");
  CursorStatusX = new QLabel(this);
  CursorStatusY = new QLabel(this);
  CursorStatusDX = new QLabel(this);
  CursorStatusDY = new QLabel(this);
  CursorStatusX->setVisible(false);
  CursorStatusY->setVisible(false);
  CursorStatusDX->setVisible(false);
  CursorStatusDY->setVisible(false);
  CursorStatusDX->setStyleSheet("color: navy");
  CursorStatusDY->setStyleSheet("color: navy");
  statusBar()->addPermanentWidget(CursorStatusX);
  statusBar()->addPermanentWidget(CursorStatusY);
  statusBar()->addPermanentWidget(CursorStatusDX);
  statusBar()->addPermanentWidget(CursorStatusDY);
}


void  sRAMainWindow::ShowCursorInfo (sPhysPair  position
                                    ,sPhysPair  reference
                                    ,sString    x_title
                                    ,sString    y_title)
{
  sPhysPair                   Difference(position);
  bool                        OK = true;
  if( x_title.Empty() ){
    CursorStatusX->setText(ToQString( position.X.Text(HTML,Laconic) ));
  }else{
    CursorStatusX->setText(ToQString( sString("<i>") + x_title + "</i>="
                                    + position.X.Text(HTML,Laconic)));
  }
  if( y_title.Empty() ){
    CursorStatusY->setText(ToQString( position.Y.Text(HTML,Laconic) ));
  }else{
    CursorStatusY->setText(ToQString( sString("<i>") + y_title + "</i>="
                                    + position.Y.Text(HTML,Laconic) ));
  }
  if( ! CursorStatusX->isVisible() )  CursorStatusX->setVisible(true);
  if( ! CursorStatusY->isVisible() )  CursorStatusY->setVisible(true);
  try{
    Difference = position - reference;
    if( x_title.Empty() ){
      CursorStatusDX->setText(ToQString( sString(Symbol::DeltaCapital) + "="
                                       + Difference.X.Text(HTML,Laconic) ));
    }else{
      CursorStatusDX->setText(ToQString(
          sString(Symbol::DeltaCapital) + "<sub><i>" + x_title + "</i></sub>="
          + Difference.X.Text(HTML,Laconic)
          ));
    }
    if( y_title.Empty() ){
      CursorStatusDY->setText(ToQString( sString(Symbol::DeltaCapital) + "="
                                       + Difference.Y.Text(HTML,Laconic) ));
    }else{
      CursorStatusDY->setText(ToQString(
          sString(Symbol::DeltaCapital) + "<sub><i>" + y_title + "</i></sub>="
          + Difference.Y.Text(HTML,Laconic)
          ));
    }
  }catch(rxRAlgebra){
    OK = false;
  }
  if( CursorStatusDX->isVisible() != OK )  CursorStatusDX->setVisible(OK);
  if( CursorStatusDY->isVisible() != OK )  CursorStatusDY->setVisible(OK);
}


void  sRAMainWindow::ShowHelpContents ()
{
  sHelper::ShowIndex();
}


void  sRAMainWindow::About ()
{
  sAboutRA *                 AboutRA = new sAboutRA(this);
  AboutRA->setAttribute(Qt::WA_DeleteOnClose);
  AboutRA->exec();
}


void  sRAMainWindow::RestoreWindows (sNodePtr  SetOfCfg)
{
  for_each_nameless_attribute( V , SetOfCfg ){
    sPtr<sFolderNode>           Cfg ( V );
    if( Cfg.IsCorrect() )
      CreateWindow(Cfg);
  }end_attr_loop()
}


void  sRAMainWindow::ClearClosedWindows (sNodePtr SetOfCfg)
{
  bool                        NothingToDestroy = false;
  while( ! NothingToDestroy ){
    sPtr<sFolderNode>           Cfg2BeDestroyed;
    for_each_nameless_attribute( V , SetOfCfg ){
      sNodePtr                    D ( V >> "Destroyed" );
      if( ! D.IsNULL() ){
        Cfg2BeDestroyed = V;
        break;
      }
    }end_attr_loop();
    if( Cfg2BeDestroyed.IsNULL() ){
      NothingToDestroy = true;
    }else{
      SetOfCfg.DelAttr(Cfg2BeDestroyed);
    }
  }
}


sString  sRAMainWindow::GetObjectName4Qt (sNodePtr   SetOfCfg
                                         ,sNodePtr   Cfg
                                         ,rcsString  Type)
{
  sString                     ObjectName4Qt;
  sPtr<sTextNode>             N (Cfg >> "ObjectName4Qt");
  if( N.IsCorrect() ){
    ObjectName4Qt = N->Value();
  }else{
    int                         No = 0;
    bool                        Free = false;
    while( ! Free ){
      No++;
      ObjectName4Qt=Type;
      ObjectName4Qt<<No;

      Free=true;
      for_each_nameless_attribute( V , SetOfCfg ){
        N = V >> "ObjectName4Qt";
        if( N.IsCorrect() )
          Free = ! ( sString(*N) == ObjectName4Qt );
        if( ! Free )
          break;
      }end_attr_loop()
    }
    Cfg << attr("ObjectName4Qt",ObjectName4Qt);
  }
  return ObjectName4Qt;
}


void  sRAMainWindow::CreateWindow (sNodePtr  Cfg)
{
  psWindow                    Window = NULL;
  sString                     Type;
  sString                     ObjectName4Qt;
  try{
    Cfg |= attr("Type","Workbook");
    Type=rsTextNode( * (Cfg>>"Type"));
    ObjectName4Qt=GetObjectName4Qt(Children,Cfg,Type);
    if(       Type == "Workbook" ){
      Window= new sWorkbook(Cfg,ObjectName4Qt,this);
    }else if( Type == "Plot" ){
      Window= new sPlot(Cfg,ObjectName4Qt,this);
    }else if( Type == "Console" ){
      sPtr<sTextNode>           ConsoleUID (Cfg >> "Title");
      if( ConsoleUID.IsCorrect() ){
        Window=sConsole::CreateConsole(*ConsoleUID,Cfg,this);
      }
    }
  }catch( rxProgramError E ){
    QMessageBox::critical(this ,tr("Instrument launching error")
                         ,ToQString(E.SorryText())   );
  }
  if( ! Window ){
    Cfg |= attr("Destroyed",true);
  }else{
    /*psRADock  Dock = */new sRADock(this,ObjectName4Qt,Cfg,Window);
    Window->RestoreState();
//    if( MenuWindow ){
//      QAction *  WindowSelectAction
//          = MenuWindow->addAction(Dock->windowIcon(),Dock->windowTitle());
//      WindowSelectAction->setCheckable(true);
//      WindowListActions->addAction(WindowSelectAction);
//      WindowSelectAction->setChecked(true);
//      connect(WindowSelectAction,SIGNAL(triggered())
//             ,Dock,SLOT(raise()));
//      connect(WindowSelectAction,SIGNAL(triggered())
//             ,Window,SLOT(setFocus()));
//      connect(Window,SIGNAL(FocusReceived(bool))
//             ,WindowSelectAction,SLOT(setChecked(bool)));
//      connect(Dock,SIGNAL(destroyed())
//             ,WindowSelectAction,SLOT(deleteLater()));
//    }
  }
}


void sRAMainWindow::NewWorkbook ()
{
  sNodePtr                    Cfg;
  sNodePtr                    Content( new sFolderNode );
                              //! @todo{UI++} new sTopic?
//  sNet::Manager->Undo.GroupOpen();
  Children << (Cfg = sNodePtr(new sFolderNode()));
  Cfg << attr("Type","Workbook")
      << attr("Content",Content)
      << attr("DockArea",int(Qt::LeftDockWidgetArea));
  CreateWindow(Cfg);
//  sNet::Manager->Undo.GroupFlush();
}


void sRAMainWindow::NewPlot ()
{
  sNodePtr                    Cfg;
  sNodePtr                    Content( new sFolderNode );
                              //! @todo{UI++} new sPicture?
//  sNet::Manager->Undo.GroupOpen();
  Children << (Cfg = sNodePtr(new sFolderNode()));
  Cfg << attr("Type","Plot")
      << attr("Content",Content)
      << attr("DockArea",int(Qt::RightDockWidgetArea));
  CreateWindow(Cfg);
//  sNet::Manager->Undo.GroupFlush();
}


void sRAMainWindow::OpenData ()
{
  sNodePtr                    Cfg;
//  sNet::Manager->Undo.GroupOpen();
  Children << (Cfg = sNodePtr(new sFolderNode()));
  Cfg << attr("Type","Workbook")
      << attr("Title",FromQString(tr("Data")))
      << attr("Content" ,DataToBrowse)
      << attr("DockArea",int(Qt::LeftDockWidgetArea));
  CreateWindow(Cfg);
//  sNet::Manager->Undo.GroupFlush();
}


void sRAMainWindow::OpenRecords ()
{
  sNodePtr                    Cfg;
//  sNet::Manager->Undo.GroupOpen();
  Children << (Cfg = sNodePtr(new sFolderNode()));
  Cfg << attr("Type","Workbook")
      << attr("Title",FromQString(tr("Records")))
      << attr("Content" ,RecordsToBrowse)
      << attr("DockArea",int(Qt::LeftDockWidgetArea));
  CreateWindow(Cfg);
//  sNet::Manager->Undo.GroupFlush();
}


void sRAMainWindow::OpenSamples ()
{
  sNodePtr                    Cfg;
//  sNet::Manager->Undo.GroupOpen();
  Children << (Cfg = sNodePtr(new sFolderNode()));
  Cfg << attr("Type","Workbook")
      << attr("Title",FromQString(tr("Samples")))
      << attr("Content" ,SamplesToBrowse)
      << attr("DockArea",int(Qt::LeftDockWidgetArea));
  CreateWindow(Cfg);
//  sNet::Manager->Undo.GroupFlush();
}


void sRAMainWindow::OpenTopics ()
{
  sNodePtr                    Cfg;
//  sNet::Manager->Undo.GroupOpen();
  Children << (Cfg = sNodePtr(new sFolderNode()));
  Cfg << attr("Type","Workbook")
      << attr("Title",FromQString(tr("Topics")))
      << attr("Content" ,TopicsToBrowse)
      << attr("DockArea",int(Qt::LeftDockWidgetArea));
  CreateWindow(Cfg);
//  sNet::Manager->Undo.GroupFlush();
}


void  sRAMainWindow::ChangeUser (QString  U)
{
  if( U == "cAdmin" ){
    ChangeUserAccessLevel(sUser::cAdmin);
  }else if( U == "cExpert" ){
    ChangeUserAccessLevel(sUser::cExpert);
  }else{
    ChangeUserAccessLevel(sUser::cNewbie);
  }
}


void  sRAMainWindow::ChangeUserAccessLevel (sUser::eAccessLevel  level)
{
  sEnvironment::User.AccessLevel = level;
  if( sEnvironment::User.AccessLevel > sEnvironment::User.MaxAccessLevel ){
    sEnvironment::User.AccessLevel = sEnvironment::User.MaxAccessLevel;
  }
  switch( sEnvironment::User.AccessLevel ){
    case sUser::cAdmin:
        ActnAdminLevel->setChecked(true);
        break;
    case sUser::cExpert:
        ActnExpertLevel->setChecked(true);
        break;
    case sUser::cNewbie:
        ActnNewbieLevel->setChecked(true);
        break;
    default:
        ActnAdminLevel->setChecked(false);
        ActnExpertLevel->setChecked(false);
        ActnNewbieLevel->setChecked(false);
  }
  ActnExpertLevel->setEnabled(sEnvironment::User.MaxAccessLevel >= sUser::cExpert);
  ActnAdminLevel->setVisible(sEnvironment::User.MaxAccessLevel >= sUser::cAdmin);
  ActnOpenRecords->setEnabled(sEnvironment::User.AccessLevel > sUser::cNewbie);
  ActnOpenData->setEnabled(sEnvironment::User.AccessLevel > sUser::cNewbie);
}


void  sRAMainWindow::SetNightRegime (bool  night)
{
  literal                     Regime( night ? "NightStyle" : "NormalStyle" );
  sString                     Shapes( CSS::Shapes );
  sString                     Style( night ? CSS::NightStyle
                                           : CSS::NormalStyle );
  try{
    get_from<sTextNode>(Styles>>"Shapes",Shapes);
    get_from<sTextNode>(Styles>>Regime,Style);
    qApp->setStyleSheet(ToQString(Shapes)+ToQString(Style));
    RAGUI::NightRegime = night;
    Desktop << attr(NightRegimeAttrName,night);
  }catch(...){}
}


void  sRAMainWindow::ChangeBasicLineColors ()
{
  sColorSetEdit               D(RAGUI::LineColorSet->Value()
                               ,( RAGUI::NightRegime ? sColor::cBlack
                                                     : sColor::cWhite )
                               );
  if( D.exec() ){
    RAGUI::LineColorSet->Assign( D.Colors() );
  }
}


void  sRAMainWindow::ChangeFont ()
{
  bool                        OK = false;
  QFont                       Font = QFontDialog::getFont(&OK,font(),this);
  if( OK ){
    if( Font.family() != font().family() )
      FontFamily->Assign(FromQString(Font.family()));
    if( Font.pointSize() != font().pointSize() )
      FontSize->Assign(Font.pointSize());
    qApp->setFont(QFont(Font.family(),Font.pointSize()));
  }
}


void sRAMainWindow::FillMenuLanguage ()
{
  QActionGroup *              LanguageActions;
  LanguageActions = new QActionGroup(this);
  LanguageActions->setExclusive(true);

  psActionSignallingText      LA;
  QString                     L;

  //source code locale (C)
  {
    L = "C";
    LA = new sActionSignallingText(this);
    LA->SetDataString(L);
    LA->setText("default (English)");
    LA->setCheckable(true);
    connect(LA,SIGNAL(DataStringSignal(QString))
           ,this,SLOT(ChangeLanguage(QString)));
    LanguageActions->addAction(LA);
    MenuLanguage->addAction(LA);
    LA->setChecked( L == ToQString(Language->Value()) );
  }

  QDir                        LocalesDir(qApp->applicationDirPath()
                                        +QDir::separator()+".."
                                        +QDir::separator()+"locales");
  QMap<QString,bool>          Languages;
  LocalesDir.setNameFilters( QStringList() << "*.qm" );
  foreach( QString FileName , LocalesDir.entryList(QDir::Files) ){
//    RANet::Log.Put(sLog::Debug,"locale"
//                  ,sString("Translator: ")+FromQString(FileName));
    L = QFileInfo(QFileInfo(FileName).completeBaseName()).completeSuffix();
    if( ! Languages.contains(L) || ! Languages.value(L) ){
      Languages.insert( L , L == ToQString(Language->Value()) );
    }
  }
  QMapIterator<QString,bool>    i(Languages);
  while( i.hasNext() ){
    i.next();
    LA = new sActionSignallingText(this);
    LA->SetDataString(i.key());
    LA->setText(i.key()+" :  "+QLocale(i.key()).nativeLanguageName());
    LA->setCheckable(true);
    connect(LA,SIGNAL(DataStringSignal(QString))
           ,this,SLOT(ChangeLanguage(QString)));
    LanguageActions->addAction(LA);
    MenuLanguage->addAction(LA);
    LA->setChecked(i.value());
  }
}


void sRAMainWindow::changeEvent (QEvent *  e)
{
  QMainWindow::changeEvent(e);
  switch (e->type()) {
    case QEvent::LanguageChange:
        Retranslate();
        break;
    default:
        break;
  }
}


void  sRAMainWindow::ChangeLanguage (QString  L)
{
  SetLanguage(L);
  qApp->postEvent(this,new QEvent(QEvent::LanguageChange));
}


void  sRAMainWindow::SetLanguage (QString  language)
{
  RANet::Log.Put(sLog::Debug,"locale"
      ,sString("SetLanguage(")+FromQString(language)+")");

  if( language.isEmpty() ){
    if( Language->Value().Empty() ){
      language = QLocale::system().name(); //gives ru_RU
    }else{
      language = ToQString(Language->Value());
    }
  }
  if( language != ToQString(Language->Value()) ){
    Language->Assign(FromQString(language));
  }

//  setLocale(QLocale(ToQString(Language->Value())));
//  setLocale(QLocale("C"));
  //Этого мало. Например, QDoubleValidator по умолчанию ставит умолчательную
  //локаль, которая берется из системных настроек.  Ее можно изменить
  //(QLocale::setDefault(QLocale("C"))), но только до старта любых ниток, что
  //мне здесь не подходит (сделано в main()).
  //Кроме того, не все ясно с правой точкой - она не всегда превращается в то
  //во что должна бы.

  if( QtTranslator ){
    if( QtTranslator->load(QString("qt_")+ToQString(Language->Value())
                 ,QLibraryInfo::location(QLibraryInfo::TranslationsPath)) ){
//      RANet::Log.Put(sLog::Debug,"locale","Qt translator loaded from Qt");
    }else if( QtTranslator->load(QString("qt_")+ToQString(Language->Value())
                                ,qApp->applicationDirPath()+QDir::separator()
                                +".."+QDir::separator()+"locales"           ) ){
      RANet::Log.Put(sLog::Debug,"locale","Qt translator loaded from locales");
    }
  }
  if( Translator->load(QString("RA.")+ToQString(Language->Value())
                      ,qApp->applicationDirPath()+QDir::separator()
                      +".."+QDir::separator()+"locales"              )  ){
//    RANet::Log.Put(sLog::Debug,"locale","RA translator loaded");
  }
  for( std::list<psInstrumentalPackage>::const_iterator
        IP = LoadedPackages.begin() ; IP != LoadedPackages.end() ; ++IP ){
    if( (*IP)->Translator->load(ToQString((*IP)->TrueType()->Name().Tail(1))
                            +"."+ToQString(Language->Value())
                            ,qApp->applicationDirPath()+QDir::separator()
                            +".."+QDir::separator()+"locales"            )  ){
//      RANet::Log.Put(sLog::Debug,"locale"
//                    ,(*IP)->TrueType()->Name()+" translator loaded");
    }
  }
//  QTextCodec::setCodecForTr(QTextCodec::codecForName("utf8"));
}


void  sRAMainWindow::Retranslate ()
{
  RANet::Log.Put(sLog::Debug,"locale"
                ,sString("Retranslate ")+FromQString(locale().name()));
  setWindowTitle(tr("Research Assistant"));
//  ActnNewWorkbook->setText(tr("New work&book"));
//  ActnNewWorkbook->setStatusTip(tr("Open a new empty workbook"));
  ActnNewPlot->setText(tr("New plo&t"));
  ActnNewPlot->setStatusTip(tr("Open a new empty plot"));
  ActnOpenTopics->setText(tr("Open &Topics"));
  ActnOpenTopics->setStatusTip(tr("Open general catalogue of research topics"));
  ActnOpenSamples->setText(tr("Open &Samples"));
  ActnOpenSamples->setStatusTip(tr("Open general catalogue of samples"));
  ActnOpenRecords->setText(tr("Open Records"));
  ActnOpenRecords->setStatusTip(tr("Open general catalogue of records"));
  ActnOpenData->setText(tr("Open Data"));
//  ActnOpenData->setStatusTip(tr(""));
//  ActnPrint->setText(tr("Print ..."));
//  ActnPrint->setStatusTip(tr("Print ..."));
  ActnExit->setText(tr("E&xit"));
  ActnExit->setStatusTip(tr("Exit the application"));
  ActnNewbieLevel->setText(tr("Newbie"));
//  ActnNewbieLevel->setStatusTip(tr(""));
  ActnExpertLevel->setText(tr("Expert"));
//  ActnExpertLevel->setStatusTip(tr(""));
  ActnAdminLevel->setText(tr("Admin"));
//  ActnAdminLevel->setStatusTip(tr(""));
  ActnSetNightRegime->setText(tr("Reduce light pollution"));
  ActnSetNightRegime->setStatusTip(tr("Save the World!  Reduce the "
                                      "light pollution of our planet."));
  ActnSetDockNesting->setText(tr("Smarter child windows"));
  ActnSetDockNesting->setStatusTip(tr("Allow more complicated arrangement "
                                      "of the child windows (may become "
                                      "a bit tricky in use)."));
  ActnChangeBasicLineColors->setText(tr("Basic colors"));
  ActnChangeBasicLineColors->setStatusTip(tr("Change the basic set of colors"
                                             "for curves, marks etc." ));
  ActnChangeFont->setText(tr("Change font"));
//  ActnChangeFont->setStatusTip(tr(""));
  ActnHelpContents->setText(tr("Contents"));
  ActnHelpContents->setStatusTip(tr("Show the program's interactive manual"));
  ActnAbout->setText(tr("&About"));
  ActnAbout->setStatusTip(tr("Show the application's About box"));
  ActnAboutQt->setText(tr("About &Qt"));
  ActnAboutQt->setStatusTip(tr("Show the Qt library's About box"));

  MenuFile->setTitle(tr("&File"));
  MenuInstrument->setTitle(tr("&Instrument"));
  MenuInstrument
      ->setToolTip(tr("Run one of the installed experimental techniques"));
  MenuSettings->setTitle(tr("&Settings"));
  MenuLanguage->setTitle(tr("Language"));
  MenuUser->setTitle(tr("User"));
  MenuInstallNewDevice->setTitle(tr("Install new device"));
  MenuInstallNewDriver->setTitle(tr("Driver"));
  MenuInstallNewConsole->setTitle(tr("Console"));
  MenuInstallNewGadget->setTitle(tr("Gadget"));
  MenuConfigureDevice->setTitle(tr("Configure an installed device"));
  MenuHelp->setTitle(tr("&Help"));
  MenuAboutPlugins->setTitle(tr("About plugins"));
}

//------------------------------------------------------------------------------
} //namespace RA
