////////////////////////////////////////////////////////////////////////////////
/*! @file InstallJob.cpp
- Uses selected files from RANet - Research Assistant Net Library.
- Uses  Qt v.5  -  http://qt-project.org/
- Copyright(C) 2011-2015, Viktor E. Bursian, St.Petersburg, Russia.
                          VBursian@gmail.com
*///////////////////////////////////////////////////////////////////////////////
#include "InstallJob.h"
#include "StringsAndQStrings.h"
#include <QStandardPaths>
#include <QDebug>
namespace RA {
//------------------------------------------------------------------------------

//-------------------------------------------------------------- sInstallJob ---

sInstallJob::sInstallJob (bool  testing)
    :sApplicationJob(testing)
{
  DistFilesDir = AppDir;
  ZipProgDir = ( System == "win32" ? "" : "./" );
  if( testing ){
    DistFilesDir.cd("distfiles");
    ZipProgDir = QString("distfiles") + QDir::separator();
  }
  if( System == "win32" ){
    ProgDirDefaultLocation = "C:";
  }else if( System == "unix" ){
    ProgDirDefaultLocation = QStandardPaths::writableLocation(
                                                  QStandardPaths::HomeLocation);
  }
  if( Testing )  qDebug() << "TESTING MODE";
}


bool  sInstallJob::AnalizeDistributive ()
{
  QString                     VersionStr;
  sVersion                    Version;
  foreach( QString FName, DistFilesDir.entryList(QDir::Files) ){
    if( FName.endsWith(QString(".")+System+".zip") ){
      if( FName.startsWith("RA.") ){
        int                         P;
        QString                     Part;

        Part=FName.mid(QString("RA.").length());
        P = Part.indexOf(".");
        if( P < 0 )  return false;
        VersionStr=Part.mid(P+1);
        Part = Part.mid(0,P);

        P = VersionStr.indexOf(QString(".")+System+".zip");
        if( P < 0 )  return false;
        VersionStr = VersionStr.mid(0,P);
        Version = sVersion(FromQString(VersionStr));
        if( ! PrincipalVersion.IsValid() ){
          PrincipalVersion = Version.Principal();
        }else if( ! (Version.Principal() == PrincipalVersion) ){
          return false;
        }

        if( Part == "core" ){
          Core = new sProjectInfo::sPartInfo(Version,QString());
        }else{
          Plugins << Part;
          Plugin[Part] = new sProjectInfo::sPartInfo(Version,QString());
        }
      }else if( FName.startsWith("fonts.") ){
        FontsZip = FName;
      }else{
        OtherProgZips << FName;
      }
    }else if( FName.endsWith(".src.zip") ){
      Sources << FName;
    }
  }
  ProgDir.setPath( ProgDirDefaultLocation
                 + QDir::separator()
                 + "ResearchAssistant"
                 + QDir::separator()
                 + PrincipalVersion.Text() );
  QDir                        DocsDir(QStandardPaths::writableLocation(
                                            QStandardPaths::DocumentsLocation));
  FontsDir.setPath(DocsDir.absoluteFilePath("optional_fonts_for_RA"));
  SandboxDir.setPath(DocsDir.absoluteFilePath("RA/sandbox"));
  SrcCopyDir.setPath(DocsDir.absoluteFilePath(QString("Programming/RA/")
                                       +PrincipalVersion.Text()));
  return true;
}


QString  sInstallJob::Plan ()
{
  QString                     R;
  Report->clear();
  R = "<font color=navy> <p>You have chosen:</p> <li>";
  if( Install_Prog ){
    if( Core && Core->Selected ){
      R += "<ul>%1 RA program %2 <b>&quot;%3&quot;</b>"
           " with the following plugins:";
    }else{
      R += "<ul>%1 the following RA plugins %2 <b>&quot;%3&quot;</b>:";
    }
    R = R.arg( ( Updating ? "Update" : "Install" ) )
         .arg( ( Updating ? "in the" : "into" ) )
         .arg( ProgDir.absolutePath() );
    for( QStringList::const_iterator  P = Plugins.constBegin()
                                    ; P != Plugins.constEnd(); ++P ){
      if( Plugin[*P]->Selected ){
        R += "<ul>";
        R += *P + "</ul>";
      }
    }
    R += "</ul>";
  }
  if( Install_Fonts ){
    R += "<ul>Fonts  will be copied into <b>&quot;";
    R += FontsDir.absolutePath();
    R += "&quot;</b> folder.  You will have to install them manually.</ul>";
  }
  if( Install_Sources ){
    R += "<ul>Program source texts will be copied into <b>&quot;";
    R += SrcCopyDir.absolutePath();
    R += "&quot;</b> folder.</ul>";
  }
  R += "</li>"
          "<p>___________________________________________________________</p>"
          "</font>\n\n";
  return R;
}


bool  sInstallJob::MakeDirs ()
{
  bool                        OK = true;
  OK = OK && ProgDir.mkpath(".");
  OK = OK && FontsDir.mkpath(".");
  OK = OK && SrcCopyDir.mkpath(".");

  return OK;
}


bool sInstallJob::Install ()
{
  bool                        OK = true;
  OK = OK && ( ! Install_Prog || InstallProg() );
  OK = OK && ( ! Install_Fonts || InstallFonts() );
  OK = OK && ( ! Install_Sandbox || InstallSandbox() );
  OK = OK && ( ! Install_Launcher || InstallLauncher() );
  OK = OK && ( ! Install_Sources || InstallSources() );
  return OK;
}


bool  sInstallJob::InstallProg ()
{
  bool                        OK = true;
  if( Core && Core->Selected ){
    OK = OK && Unzip( "RA.core."
                    + ToQString(Core->Version.Valuable().Text())
                    + "." + System + ".zip"
                    ,ProgDir.absolutePath()
                    );
    OK = OK && Copy(DistFilesDir ,QDir(ProgDir.absoluteFilePath("bin"))
                   ,QStringList()
                       << QtLibraries
                       << GNULibraries
                       << "platforms"
                   );
    foreach( QString ZF , OtherProgZips  ){
      OK = OK && Unzip(ZF
                      ,ProgDir.absolutePath()+QDir::separator()+"bin"
                      );
    }
  }
  for( QStringList::const_iterator  P = Plugins.constBegin()
                                  ; P != Plugins.constEnd(); ++P ){
    if( Plugin[*P]->Selected ){
      OK = OK && Unzip( "RA." + *P + "."
                      + ToQString(Plugin[*P]->Version.Valuable().Text())
                      + "." + System + ".zip"
                      ,ProgDir.absolutePath()
                      );
    }
  }
  return OK;
}


bool sInstallJob::InstallFonts()
{
  return Unzip(DistFilesDir.absoluteFilePath(FontsZip)
              ,FontsDir.absolutePath());
}


bool  sInstallJob::InstallSources ()
{
  bool                        OK = true;
  foreach( QString Src , Sources ){
    OK = OK && Unzip(DistFilesDir.absoluteFilePath(Src)
                    ,SrcCopyDir.absolutePath()
                    );
  }
  return OK;
}


bool  sInstallJob::InstallSandbox ()
{
  bool                        OK = true;

  return OK;
}


bool  sInstallJob::InstallLauncher ()
{
  bool                        OK = true;
  if( System == "unix" ){
    /*OK = OK && */Execute("sed"
                      ,QStringList()
                        << "-i"
                        << "s@RA_PROG_DIR@"
                            +ProgDir.canonicalPath()
                            +"@g"
                        << "RA.desktop"
                      ,ProgDir.absoluteFilePath("desktop") );
    /*OK = OK && */Execute("sed"
                      ,QStringList()
                        << "-i"
                        << "s@RA_ROOT_DIR@"
//                            +RootDir.canonicalPath()
                            +ProgDir.canonicalPath()+"/sandbox"
                            +"@g"
                        << "RA.desktop"
                      ,ProgDir.absoluteFilePath("desktop") );
    /*OK = OK && */Execute("sed"
                      ,QStringList()
                        << "-i"
                        << "s@SANDBOX_DIR@"
                            +ProgDir.canonicalPath()+"/sandbox"
                            +"@g"
                        << "install-desktop.sh"
                      ,ProgDir.absoluteFilePath("desktop") );
    OK = OK && Execute("./install-desktop.sh"
                      ,QStringList()
                      ,ProgDir.absoluteFilePath("desktop") );

    //drag from Dash and drop onto Launcher
  }
  return OK;
}


//------------------------------------------------------------------------------
} //namespace RA
