////////////////////////////////////////////////////////////////////////////////
/*! @file LonelyApplication.cpp QApplication subclass, which prevents
                                running of another instance of the same program.
- Uses  Qt v.5  -  http://qt-project.org/
- Copyright(C) 2014, Viktor E. Bursian, St.Petersburg, Russia.
                     VBursian@gmail.com
*///////////////////////////////////////////////////////////////////////////////
#include "LonelyApplication.h"
#include <QSystemSemaphore>
//------------------------------------------------------------------------------
//------------------------------------------------------- sLonelyApplication ---

/*! Checks whether this is the first instance of the program.

@param argc
@param argv
@param unique_name

Useful ideas have been taken from
  http://habrahabr.ru/post/173281/
and from
  http://stackoverflow.com/questions/25135678/qt-single-instance-application-with-qsharedmemory

QSystemSemaphore is utilized in order to solve the so-called race problem,
when two or more instances start simultaneously.  Thus, the constructor body
executions are separated in time.

QSharedMemory is taken by the first instance and is kept during its life
as a proof of a primogeniture.

There is a problem under Linux:
Contrary to mustdie this system doesn't release shared memory, if the
program crashes.  However, a dummy sequence attach() -- detach() destroyes
the shared memory block, if there are no actual pointers to it,
or leaves it alive, if there are some.

@note
An odd feature appeared under Ubuntu 14.04.1 with Qt 5.3.2 and g++ 4.8.2:
It looks like, when subclassing QApplication, one must have a constructor
with argc and argv as the FIRST two parameters.  Otherwise the program
crashes (segmentation fault) at some unpredictable moment later (supposingly,
when the first window or dialog is about to show). Having an additional third
parameter is OK.

*/
sLonelyApplication::sLonelyApplication (int      argc
                                       ,char*    argv[]
                                       ,QString  unique_name
                                       )
    :QApplication(argc, argv)
    ,FirstInstanceFlag(false)
    ,SharedMemory(NULL)
{
  QSystemSemaphore  LaunchingSeparationSemaphore(unique_name+"SEM",1);
  LaunchingSeparationSemaphore.acquire();
  //crash problem workaround:
  {
    QSharedMemory         ShMem(unique_name);
    ShMem.attach();
  }
  SharedMemory = new QSharedMemory(unique_name,this);
  if( SharedMemory->create(1) ){
    FirstInstanceFlag = true;
  }
  LaunchingSeparationSemaphore.release();
}

//------------------------------------------------------------------------------
