////////////////////////////////////////////////////////////////////////////////
/*! @file Primogeniture.cpp Checking primogeniture, in order to prevent
                            running of another instance of the same program.
- Uses  Qt v.5  - http://qt.io/
- Copyright(C) 2020, Viktor E. Bursian, St.Petersburg, Russia.
                     VBursian AT gmail DOT com
*///////////////////////////////////////////////////////////////////////////////
#include "Primogeniture.h"
#include <QSharedMemory>
#include <QSystemSemaphore>
namespace RA {
//------------------------------------------------------------------------------

static QSharedMemory *        PrimogenitureSharedMemory = nullptr;


/*!

Useful ideas have been taken from
  http://habrahabr.ru/post/173281/
and from
  http://stackoverflow.com/questions/25135678/qt-single-instance-application-with-qsharedmemory

QSystemSemaphore is utilized in order to solve the so-called race problem,
when two or more instances start simultaneously.  Thus, the constructor body
executions are separated in time.

QSharedMemory is taken by the first instance and is kept during its life
as a proof of a primogeniture.

There is a problem under Linux:
Contrary to mustdie this system doesn't release shared memory, if the
program crashes.  However, a dummy sequence attach() -- detach() destroyes
the shared memory block, if there are no actual pointers to it,
or leaves it alive, if there are some.

*/
bool  GetPrimogeniture (const QString &  unique_name)
{
  QSystemSemaphore  LaunchingSeparationSemaphore(unique_name+"SEM",1);
  LaunchingSeparationSemaphore.acquire();
  //crash problem workaround:
  {
    QSharedMemory         ShMem(unique_name);
    ShMem.attach();
  }
  PrimogenitureSharedMemory = new QSharedMemory(unique_name);
  bool                FirstInstanceFlag = PrimogenitureSharedMemory->create(1);
  LaunchingSeparationSemaphore.release();
  return FirstInstanceFlag;
}

//------------------------------------------------------------------------------
}//namespace RA
