////////////////////////////////////////////////////////////////////////////////
/*! @file RA.cpp  Research Assistant: the main program.
- Uses  RAGUI - Research Assistant Graphical User Interface Library.
- Uses  RANet - Research Assistant Net Library.
- Uses  Qt v.5  - http://qt.io/
- Copyright(C) 2010-2020, Viktor E. Bursian, St.Petersburg, Russia.
                          VBursian AT gmail DOT com
*///////////////////////////////////////////////////////////////////////////////
#include "Log.h"
#include "RAGUI_global.h"
#include "Environment.h"
#include "Primogeniture.h"
#include "DB_SQL_Net.h"
#include "ByeScreen.h"
#include "RAMainWindow.h"
#include "Helper.h"
#include "TestMemUsage.h"
#include <QApplication>
#include <QDir>
#include <QLocale>
#include <QMessageBox>

namespace RA {
extern void  SetTracing ();
}//namespace RA

using namespace RA;



int main(int argc, char *argv[])
{
  RANet::Log.OpenBlock(sLog::Conditionless,"RA"
                      ,"#####################################################");
  SetTracing();

  int                         ReturnCode = 255;
  std::set_terminate (__gnu_cxx::__verbose_terminate_handler);
  TestMemoryUsage("initially");

  InitiateRAGUI(); // initiates the resources, checks RANet version match,

  try{

    //initiates sEnvironment's static members
    { sEnvironment(argc,argv); };


    {
      QLocale::setDefault(QLocale::c());

      QApplication          Application(argc,argv);
      if( ! GetPrimogeniture("ResearchAssistant") ){
        QMessageBox * M(new QMessageBox(QMessageBox::Icon::Information
            ,"Research Assistant"
            ,"One instance of the Research Assistant program is already "
             "running on this computer.  You cannot start another one."));
        M->show();
        Application.exec();
        return 1;
      }

      //Setting libraryPaths
      {
        //At this point, it is too late for the dlls which are linked
        //directly to RA.  However, it is necessary for the plugins.
        QCoreApplication::addLibraryPath(QCoreApplication::applicationDirPath()
                                        +QDir::separator()
                                        +"plugins");
        QCoreApplication::addLibraryPath(QCoreApplication
                                                       ::applicationDirPath());
        RANet::Log.Put(sLog::Admin,"Environment"
                      ,sString("libraryPaths=")
                      +FromQString(QCoreApplication::libraryPaths().join(';')));
      }

      try{
        sDB_SQL_Net                 Net;
        if( ! Net.Open() ){
          QMessageBox * M(new QMessageBox(QMessageBox::Icon::Critical
              ,"Research Assistant"
              ,QString("Cannot open the root database\n'")
              +Net.DesktopInfoText()
              +"\nWorking Dir:\n'" + QDir::currentPath() + "'"
              ));
          M->show();
          Application.exec();
          return 12;
        };

        sHelper::Initiate(sEnvironment::HelpFilePath.PathStr());

        psByeScreen                 ByeScreen = new sByeScreen(&Net);
        psRAMainWindow              MainWindow(new sRAMainWindow(Net.Root()
                                                                ,ByeScreen  ));
        MainWindow->show();

//        if( sNet::Manager->Undo.Size() ){
//          RANet::Log.Put(sLog::Debug,"RA"
//                        ,sString("Undo.Size() before Application exec ")
//                        +sString::FromInteger(sNet::Manager->Undo.Size()));
//        };
//        QObject::connect(&Application ,SIGNAL(ShowUp())
//                        ,MainWindow  ,SLOT(hide()));
//        QObject::connect(&Application ,SIGNAL(ShowUp())
//                        ,MainWindow  ,SLOT(showMaximized()));
        Application.setQuitOnLastWindowClosed(false);
        ReturnCode = Application.exec();
        sHelper::Terminate();
        RANet::Log.Put(sLog::Debug,"Quitting","correct quitting from GUI");
      }catch( rxException E ){
//!@todo{elegancy} Надо переделывать древо xException:
//!                Title вообще не нужен? Зато нужно имя класса exception?
//!             Под большим вопросом:
//!                Text в виде списка строк?
//!                (Чтобы можно было отсечь абракадабру для юзера)
//!                Автосклейка списка.
//!                Подставляемые в сообщения аргументы отдельно?
//!                (Чтобы выше их можно было вставить в другой текст)
        QMessageBox * M(new QMessageBox(QMessageBox::Icon::Critical
            ,"Research Assistant Necrology"
            ,QString("Oops!!!")
            +"\n\n"+ToQString(E.SorryText())
            ));
        M->show();
        Application.exec();
        throw;
      }
    }

//    TestMemoryUsage("1");

//    RANet::Log.Put(sLog::Debug,"RA"
//                  ,sString("Undo.Size() at the end ")
//                  +sString::FromInteger(sNet::Manager->Undo.Size()));
//    sNet::Manager->Undo.Clear();
//    TestMemoryUsage("2");
//    sNet::Manager->Root=NULL;
//    TestMemoryUsage("3");

  }catch( rxException E ){
    RANet::Log.Put(sLog::ProgramError,"RA",E.SorryText());
    E.Report2cerr();
    ReturnCode = 230;
  }catch( const std::exception & E ){
    RANet::Log.Put(sLog::ProgramError,"RA",E.what());
    xException::Report2cerr(E);
    ReturnCode = 253;
  }catch(...){
    RANet::Log.Put(sLog::ProgramError,"RA","unknown exception");
    ReturnCode = 254;
  };
  TestMemoryUsage("finally");
  RANet::Log.Put(sLog::Debug,"Quitting"
                ,"######################## Bye-bye ##########################");
  RANet::Log.CloseBlock(sLog::Conditionless,"RA");
  RANet::Log.Close(ReturnCode);
  return ReturnCode;
}
