////////////////////////////////////////////////////////////////////////////////
/*! @file BasicMath.cpp   Basic expression interpreter.
- Part of RANet - Research Assistant Net Library.
- Copyright(C) 2014, Viktor E. Bursian, St.Petersburg, Russia.
                     Viktor.Bursian@mail.ioffe.ru
*///////////////////////////////////////////////////////////////////////////////
#include "BasicMath.h"
#include <math.h>
#include "Trace.h"
#include "Log.h"
namespace RA {
//------------------------------------------------------------------------------

//see in StaticInitiators.cpp:
//sMath::sDefinitions       sMath::Definitions;
const real                  sMath::pi = 3.14159265358979323846L;

//--------------------------------------------------------------- sMathValue ---

sMathValue::sMathValue (std::istream &  a_stream
                       ,rcsVersion      version)
    :sStorable(a_stream,version)
{
}


void  sMathValue::Store (std::ostream & a_stream) const
{
  sStorable::Store(a_stream);
}


sMathValue::sMathValue (rcsMathValue  other)
    :sStorable()
{
  RANet::Log.Put(sLog::Debug,"Math"
                ,sString("sMathValue::sMathValue(")+other.Text()+")");
}


rsMathValue  sMathValue::operator = (rcsMathValue  other)
{
  RANet::Log.Put(sLog::Debug,"Math"
                ,sString("sMathValue::operator=")+other.Text()+";");
  return *this;
}


//sString  sMathValue::Text () const
sString  sMathValue::Text (eTextFormat,eTextDetalization) const
{
  return sString("undefined");
}

//--------------------------------------------------------------- sMathError ---

sMathError::sMathError (std::istream &  a_stream
                       ,rcsVersion      version)
    :sMathValue(a_stream,version)
{
  Message.Read(a_stream);
  a_stream.READ_INTO(Cursor1);
  a_stream.READ_INTO(Cursor2);
}


void  sMathError::Store (std::ostream & a_stream) const
{
  sMathValue::Store(a_stream);
  Message.Write(a_stream);
  a_stream.WRITE_FROM(Cursor1);
  a_stream.WRITE_FROM(Cursor2);
}


sMathError::sMathError (rcsMathError  other)
    :sStorable()
    ,sMathValue(other)
    ,Message(other.Message)
    ,Cursor1(other.Cursor1)
    ,Cursor2(other.Cursor2)
{
}


sString  sMathError::Text (eTextFormat,eTextDetalization) const
{
  return Message;
}

//--------------------------------------------------------------- sRealValue ---

sRealValue::sRealValue (std::istream &  a_stream
                       ,rcsVersion      version)
    :sMathValue(a_stream,version)
{
  if( version < sVersion(1,5) ){
    realIOv14 R;
    a_stream.READ_INTO(R);  Real=R;
  }else{
    realIO R;
    a_stream.READ_INTO(R);  Real=R;
  }
}


void  sRealValue::Store (std::ostream &  a_stream) const
{
  sMathValue::Store(a_stream);
  realIO R;
  R=Real;  a_stream.WRITE_FROM(R);
}


sRealValue::sRealValue (rcsRealValue  other)
    :sStorable()
    ,sMathValue(other)
    ,Real(other.Real)
{
}


//sString  sRealValue::Text () const
sString  sRealValue::Text (eTextFormat,eTextDetalization) const
{
  return sString() << Real;
//  return sString::FromReal(Real);
}

//------------------------------------------------------------ sIntegerValue ---

sIntegerValue::sIntegerValue (std::istream &  a_stream
                             ,rcsVersion      version)
    :sRealValue(a_stream,version)
{
  a_stream.READ_INTO(Integer);
}


void  sIntegerValue::Store (std::ostream & a_stream) const
{
  sRealValue::Store(a_stream);
  a_stream.WRITE_FROM(Integer);
}


sIntegerValue::sIntegerValue (rcsIntegerValue  other)
    :sStorable()
    ,sRealValue(other)
    ,Integer(other.Integer)
{
}


//sString  sIntegerValue::Text () const
sString  sIntegerValue::Text (eTextFormat,eTextDetalization) const
{
  return sString() << Integer;
}

//-------------------------------------------------------------------- sMath ---

//.......................................... sMath::sDefinitions ...

void  sMath::sDefinitions::Add (pcsMathValue  value
                               ,rcsString     identifier)
{
  ConstList.push_front(sConst(identifier,value));
}


void  sMath::sDefinitions::Add (pfFunc0Arg  function_ptr
                               ,rcsString   identifier  )
{
  sFunc  Func(identifier);
  Func.Func0Arg = function_ptr;
  FuncList.push_front(Func);
}


void  sMath::sDefinitions::Add (pfFunc1Arg  function_ptr
                               ,rcsString   identifier
                               ,cpcsTag     arg_type    )
{
  sFunc  Func(identifier,arg_type);
  Func.Func1Arg = function_ptr;
  FuncList.push_front(Func);
}


void  sMath::sDefinitions::Add (pfFunc2Arg  function_ptr
                               ,rcsString   identifier
                               ,cpcsTag     arg1_type
                               ,cpcsTag     arg2_type   )
{
  sFunc  Func(identifier,arg1_type,arg2_type);
  Func.Func2Arg = function_ptr;
  FuncList.push_front(Func);
}


void  sMath::sDefinitions::Add (pfFunc3Arg  function_ptr
                               ,rcsString   identifier
                               ,cpcsTag     arg1_type
                               ,cpcsTag     arg2_type
                               ,cpcsTag     arg3_type   )
{
  sFunc  Func(identifier,arg1_type,arg2_type,arg3_type);
  Func.Func3Arg = function_ptr;
  FuncList.push_front(Func);
}


void  sMath::sDefinitions::Add (pfFunc4Arg  function_ptr
                               ,rcsString   identifier
                               ,cpcsTag     arg1_type
                               ,cpcsTag     arg2_type
                               ,cpcsTag     arg3_type
                               ,cpcsTag     arg4_type   )
{
  sFunc  Func(identifier,arg1_type,arg2_type,arg3_type,arg4_type);
  Func.Func4Arg = function_ptr;
  FuncList.push_front(Func);
}


sMath::sDefinitions::sDefinitions ()
{
  Add(&UnaryMinusReal  ,"-" ,sRealValue::Type());
  Add(&RealPlusReal    ,"+" ,sRealValue::Type() ,sRealValue::Type());
  Add(&RealMinusReal   ,"-" ,sRealValue::Type() ,sRealValue::Type());
  Add(&RealMultReal    ,"*" ,sRealValue::Type() ,sRealValue::Type());
  Add(&RealDivReal     ,"/" ,sRealValue::Type() ,sRealValue::Type());
  Add(&RealPowerReal   ,"^" ,sRealValue::Type() ,sRealValue::Type());

  Add(&UnaryMinusInteger    ,"-" ,sIntegerValue::Type());
  Add(&IntegerPlusInteger   ,"+" ,sIntegerValue::Type() ,sIntegerValue::Type());
  Add(&IntegerMinusInteger  ,"-" ,sIntegerValue::Type() ,sIntegerValue::Type());
  Add(&IntegerMultInteger   ,"*" ,sIntegerValue::Type() ,sIntegerValue::Type());

  Add(&lg  ,"lg"  ,sRealValue::Type());
  Add(&ln  ,"ln"  ,sRealValue::Type());
  Add(&log ,"log" ,sRealValue::Type() ,sRealValue::Type());
  Add(&exp ,"exp" ,sRealValue::Type());
  Add(&sin ,"sin" ,sRealValue::Type());
  Add(&cos ,"cos" ,sRealValue::Type());
  Add(&tg  ,"tg"  ,sRealValue::Type());
  Add(&ctg ,"ctg" ,sRealValue::Type());

  Add( new sRealValue(pi) ,"pi");
}

//......................................................................

bool  sMath::Valid (psMathValue  value_ptr)
{
  return (bool)value_ptr && ! sMathError::Type()->CanCast(value_ptr);
}


sMath::sMath (rcsString  source)
    :Source(source)
    ,SourceLength(Source.Length())
    ,Cursor(0)
    ,OKCursor(0)
{
}


psMathValue  sMath::Calculate ()
{
  psMathValue                 Result = Expression();
  if( Valid(Result) && ! End() ){
    delete Result;
    Result = new sMathError("wrong tail",Cursor,SourceLength);
  }
  return Result;
}


void  sMath::AddConstant (real       value
                         ,rcsString  identifier)
{
  Definitions.Add(new sRealValue(value),identifier);
}


/*! @class sMath

@par Syntax

Expression ::= spaces [UnaryOperator] spaces Addend spaces
               [AddingOperator spaces Addend spaces]...

UnaryOperator ::= + | -

AddingOperator :: = + | -

Addend ::= Factor spaces [MultiplyingOperator spaces Factor spaces]...

MultiplyingOperator ::= * | /

Factor ::= Term spaces [PowerSign spaces Term]

PowerSign ::= ** | ^

@note
Sequential power operators are not allowed unless separated with braces.

Term ::= ( Expression )
       | { Expression }
       | Value
       | @ AttributeName [ @ AttributeName ]...
       | # RecordReference [ @ AttributeName ]...
       | FunctionName ( spaces [Expression [, Expression  [, Expression [, Expression]]]] )
       | ConstantName

@note
Spaces after the function name are not allowed.
@par
A function may have 0-4 arguments.

Value ::= Number [LeftBracket Units RightBracket]

AttributeName ::= Identifier | 'AnyCharSequence' | Digits | -Digits

LeftBracket ::= [

RightBracket ::= ]

Number ::= Digits [ . [ Digits ] ] [ e [ UnaryOperator ] Digits ]
         |  . Digits [ e [ UnaryOperator ] Digits ]

Digits ::= Digit [ Digit ]...

Digit ::= 0 | 1 | 2 | 3 | 4 | 5 | 6 | 7 | 8 | 9

FunctionName ::= ln | lg | log | exp | sin | cos | tg | ctg
               | any_other_registered_identifier

ConstantName ::= pi | any_other_registered_identifier

*/

psMathValue  sMath::Expression ()
{
//  int                         StartCursor = Cursor;
  bool                        ThereIsUnaryMinus = false;
  sString                     Op;
  psMathValue                 Result(NULL);
  psMathValue                 A1(NULL);
  psMathValue                 A2(NULL);
  PassSpaces();
  OKCursor = Cursor;
  if( End() ){
    //expression expected
    return NULL;
  }
  if( Source[Cursor] == '+' ){
    Cursor++;
  }else if( Source[Cursor] == '-' ){
    ThereIsUnaryMinus=true;
    Cursor++;
  }
  PassSpaces();
  Result = Addend();
  if( Valid(Result) ){
    if( ThereIsUnaryMinus ){
      A1 = Result;
      Result = Call("-",OKCursor,A1);
      delete A1;  A1=NULL;
    }
    PassSpaces();
    OKCursor = Cursor;
    while( Valid(Result) && AddingOperator(Op) ){
      PassSpaces();
      A1 = Result;
      Result = NULL;
      if( ! Valid(A2 = Addend()) ){
        Result = A2;
      }else{
        Result = Call(Op,OKCursor,A1,A2);
        delete A2;  A2=NULL;
        PassSpaces();
        OKCursor = Cursor;
      }
      delete A1;  A1=NULL;
    }
  }
//  if( ! Valid(Result) && Result ){
//    psMathError(Result)->Cursor1 = StartCursor;
//    psMathError(Result)->Cursor2 = Cursor;
//  }
  if( ! Result )
    RANet::Log.Put(sLog::Critical,"Math","sMath::Expression - NULL");
  return Result;
}


void  sMath::PassSpaces ()
{
  while( (Cursor < SourceLength) && (Source[Cursor] == ' ') )
    Cursor++;
}


bool  sMath::AddingOperator (rsString  op)
{
  op = sString();
  if( ! End() && ((Source[Cursor] == '+') || (Source[Cursor] == '-')) ){
    op += Source[Cursor];
    Cursor++;
  }
  return ! op.Empty();
}


bool  sMath::MultiplyingOperator (rsString  op)
{
  op = sString();
  if( ! End() && ((Source[Cursor] == '*') || (Source[Cursor] == '/')) ){
    op += Source[Cursor];
    Cursor++;
  }
  return ! op.Empty();
}


bool  sMath::PowerOperator (rsString  op)
{
  op = sString();
  if( (Cursor < SourceLength) && (Source[Cursor] == '^') ){
    op += '^';
    Cursor++;
  }else if( (Cursor < SourceLength-1) && (Source[Cursor] == '*')
                                      && (Source[Cursor+1] == '*')  ){
    op += '^';
    Cursor += 2;
  }
  return ! op.Empty();
}


psMathValue  sMath::Addend ()
{
//  int                         StartCursor = Cursor;
  psMathValue                 Result = NULL;
  sString                     Op;
  psMathValue                 A1 = NULL;
  psMathValue                 A2 = NULL;
  Result = Factor();
  if( Valid(Result) ){
    PassSpaces();
    OKCursor = Cursor;
  }
  while( Valid(Result) && MultiplyingOperator(Op) ){
    PassSpaces();
    A1 = Result;
    Result = NULL;
    if( ! Valid(A2 = Factor()) ){
      Result = A2;
    }else{
      Result = Call(Op,OKCursor,A1,A2);
      delete A2;  A2=NULL;
      PassSpaces();
      OKCursor = Cursor;
    }
    delete A1;  A1=NULL;
  }
  if( ! Result )
    RANet::Log.Put(sLog::Critical,"Math","sMath::Addend - NULL");
  return Result;
}


psMathValue  sMath::Factor ()
{
//  int                         StartCursor = Cursor;
  psMathValue                 Result = NULL;
  sString                     Op;
  psMathValue                 A1 = NULL;
  psMathValue                 A2 = NULL;
  Result = Term();
  if( Valid(Result) ){
    PassSpaces();
    OKCursor = Cursor;
    if( PowerOperator(Op) ){
      PassSpaces();
      A1 = Result;
      Result = NULL;
      if( ! Valid(A2 = Term()) ){
        Result = A2;
      }else{
        Result = Call(Op,OKCursor,A1,A2);
        delete A2;  A2=NULL;
        PassSpaces();
        OKCursor = Cursor;
      }
      delete A1;  A1=NULL;
    }
  }
  if( ! Result )
    RANet::Log.Put(sLog::Critical,"Math","sMath::Factor - NULL");
  return Result;
}


psMathValue  sMath::Term ()
{
  int                         StartCursor = Cursor;
  psMathValue                 Result = NULL;
  sString                     Id;
  if( End() ){
    Result = new sMathError("missing term",OKCursor,Cursor);
  }else if( Source[Cursor] == '(' ){
    Cursor++;
    Result = Expression();
    if( Valid(Result) ){
      if( ! End() && (Source[Cursor] == ')') ){
        Cursor++;
      }else{
        delete Result;
        Result = new sMathError("wrong term or missed ')'",StartCursor,Cursor);
      }
    }
  }else if( Source[Cursor] == '{' ){
    Cursor++;
    Result = Expression();
    if( Valid(Result) ){
      if( ! End() && (Source[Cursor] == '}') ){
        Cursor++;
      }else{
        delete Result;
        Result = new sMathError("wrong term or missed '}'",StartCursor,Cursor);
      }
    }
  }else if( Valid(Result = Value()) ){
    //do nothing
  }else if( Identifier(Id) ){
    if( ! End() && (Source[Cursor] == '(') ){
      Cursor++;
      Result = Function(Id,StartCursor);
//      if( Valid(Result) ){
//        if( ! End() && (Source[Cursor] == ')') ){
//          Cursor++;
//        }else{
//          delete Result;
//          Result = new sMathError("wrong argument or missed ')'"
//                                 ,StartCursor,Cursor);
//        }
//      }
    }else{
      Result = Constant(Id,StartCursor);
    }
  }else{
    Result = new sMathError("unrecognized term",StartCursor,SourceLength);
  }
//  if( ! Valid(Result) )
//    Cursor = StartCursor;
  if( ! Result )
    RANet::Log.Put(sLog::Critical,"Math","sMath::Term - NULL");
  return Result;
}


psMathValue  sMath::Function (rcsString  identifier
                             ,int        start_pos )
{
  int                         StartCursor = Cursor;
  psMathValue                 Result(NULL);
  psMathValue                 A1(NULL);
  psMathValue                 A2(NULL);
  psMathValue                 A3(NULL);
  psMathValue                 A4(NULL);
  PassSpaces();
  if( End() ){
  }else if( Source[Cursor] == ')' ){
    Cursor++;
    Result = Call(identifier,start_pos);
  }else{
    StartCursor = Cursor;
    A1 = Expression();
    if( ! A1 ){
      Result = new sMathError(sString("wrong argument for ")+identifier
                             ,StartCursor,Cursor);
    }else if( ! Valid(A1) ){
      Result = A1;
      A1 = NULL;
    }else if( End() ){
    }else if( Source[Cursor] == ')' ){
      Cursor++;
      Result = Call(identifier,start_pos,A1);
    }else if( Source[Cursor] == ',' ){
      Cursor++;
      StartCursor = Cursor;
      A2 = Expression();
      if( ! A2 ){
        Result = new sMathError(sString("wrong argument for ")+identifier
                               ,StartCursor,Cursor);
      }else if( ! Valid(A2) ){
        Result = A2;
        A2 = NULL;
      }else if( End() ){
      }else if( Source[Cursor] == ')' ){
        Cursor++;
        Result = Call(identifier,start_pos,A1,A2);
      }else if( Source[Cursor] == ',' ){
        Cursor++;
        StartCursor = Cursor;
        A3 = Expression();
        if( ! A3 ){
          Result = new sMathError(sString("wrong argument for ")+identifier
                                 ,StartCursor,Cursor);
        }else if( ! Valid(A3) ){
          Result = A3;
          A3 = NULL;
        }else if( End() ){
        }else if( Source[Cursor] == ')' ){
          Cursor++;
          Result = Call(identifier,start_pos,A1,A2,A3);
        }else if( Source[Cursor] == ',' ){
          Cursor++;
          StartCursor = Cursor;
          A4 = Expression();
          if( ! A4 ){
            Result = new sMathError(sString("wrong argument for ")+identifier
                                   ,StartCursor,Cursor);
          }else if( ! Valid(A4) ){
            Result = A4;
            A4 = NULL;
          }else if( End() ){
          }else if( Source[Cursor] == ')' ){
            Cursor++;
            Result = Call(identifier,start_pos,A1,A2,A3,A4);
          }else if( Source[Cursor] == ',' ){
            Result = new sMathError("there is no known functions with more "
                                    "than 4 arguments",Cursor,Cursor+1);
          }else{
          }
        }else{
        }
      }else{
      }
    }else{
    }
  }
  if( A1 )  delete A1;
  if( A2 )  delete A2;
  if( A3 )  delete A3;
  if( A4 )  delete A4;
  if( ! Result )
    RANet::Log.Put(sLog::Critical,"Math","sMath::Function - NULL");
  return Result;
}


psMathValue  sMath::Constant (rcsString  identifier
                             ,int        start_pos ) const
{
  pcsMathValue                C = FindConst(identifier);
  if( C && (C->TrueType() == sRealValue::Type()) ){
    pcsRealValue  RV = dynamic_cast<pcsRealValue>(C);
    return  RV->Replica();
  }
  return  new sMathError(sString("undefined '")+identifier+"'"
                        ,start_pos,Cursor);
}


psMathValue  sMath::Value ()
{
  integer                     IntegerResult;
  real                        RealResult;
  real                        Precision;
  psMathValue                 Result(NULL);
  if( End() ){
  }else if( IntegerNumber(IntegerResult) ){
    Result = new sIntegerValue(IntegerResult);
  }else if( RealNumber(RealResult,Precision) ){
    Result = new sRealValue( RealResult );
  }
  if( ! Result )
    RANet::Log.Put(sLog::Critical,"Math","sMath::Value - NULL");
  return Result;
}


bool  sMath::Identifier (rsString  result)
{
  result = sString();
  if( (Cursor < SourceLength) && IsLetter(Source[Cursor]) ){
    while( (Cursor < SourceLength) && (   IsLetter(Source[Cursor])
                                       || IsDigit(Source[Cursor]) ) ){
      result += Source[Cursor];
      Cursor++;
    }
  }
  return ! result.Empty();
}


bool  sMath::IntegerNumber (integer &  result)
{
  int                         C = Cursor;
  while( (C < SourceLength) && (Source[C] >= '0') && (Source[C] <= '9') )
    C++;
  if( C == Cursor ){
    return false; //not an integer (an integer must have at least one digit)
  }else if( (C < SourceLength) && (   (Source[C] == '.')
                             || (Source[C] == 'e')
                             || (Source[C] == 'E') ) ){
    return false; //not an integer ()
  }else{
    Source.Substr(Cursor,C-Cursor) >> result;
    Cursor = C;
    return true;
  }
}


bool  sMath::RealNumber (real &  result
                              ,real &  precision)
{
  int                         C = Cursor;
  int                         CC;
  bool                        NoIntPart;
  sString                     PrecisionString = "0.";
  while( (C < SourceLength) && (Source[C] >= '0') && (Source[C] <= '9') )
    C++;
  NoIntPart = (C == Cursor);
  if( (C < SourceLength) && (Source[C] == '.') ){
    C++;
    while( (C < SourceLength) && (Source[C] >= '0') && (Source[C] <= '9') ){
      PrecisionString += '0';
      C++;
    }
  }
  if( NoIntPart && (C < Cursor+2) )
    return false;
  PrecisionString += '5';
  if( (C < SourceLength) && ((Source[C] == 'e') || (Source[C] == 'E')) ){
    PrecisionString += Source[C];
    C++;
    if( (C < SourceLength) && ((Source[C] == '-') || (Source[C] == '+')) ){
      PrecisionString += Source[C];
      C++;
    }
    CC = C;
    while( (C < SourceLength) && (Source[C] >= '0') && (Source[C] <= '9') ){
      PrecisionString += Source[C];
      C++;
    }
    if( C == CC )
      return false;
  }
  if( C == Cursor ){
    return false;
  }else{
    Source.Substr(Cursor,C-Cursor) >> result;
    PrecisionString >> precision;
    Cursor = C;
    return true;
  }
}


bool  sMath::IsLetter (const char c) const
{
  return ((c >= 'a') && (c <= 'z')) || ((c >= 'A') && (c <= 'Z'))
      || (c == '_');
}


bool  sMath::IsDigit (const char c) const
{
  return (c >= '0') && (c <= '9');
}


psMathValue  sMath::Call (rcsString  identifier
                         ,int        start_pos ) const
{
  psMathValue                 Result(NULL);
  pfFunc0Arg                  FuncPtr;
  if( (bool)(FuncPtr = FindFunc(identifier)) ){
    Result = (*FuncPtr)();
  }else{
    Result = new sMathError(sString("undefined '")+identifier+"()'"
                           ,start_pos,Cursor);
  }
  return Result;
}


psMathValue  sMath::Call (rcsString    identifier
                         ,int          start_pos
                         ,psMathValue  arg1      ) const
{
  psMathValue                 Result = NULL;
  pfFunc1Arg                  FuncPtr = NULL;
  if( (bool)(FuncPtr = FindFunc(identifier,arg1)) ){
    Result = (*FuncPtr)(arg1);
  }else{
    Result = new sMathError(sString("undefined '")+identifier
                           +"("+arg1->TrueType()->Name()+")'"
                           ,start_pos,Cursor);
  }
  return Result;
}


psMathValue  sMath::Call (rcsString    identifier
                         ,int          start_pos
                         ,psMathValue  arg1
                         ,psMathValue  arg2      ) const
{
  psMathValue                 Result(NULL);
  pfFunc2Arg                  FuncPtr;
  if( (bool)(FuncPtr = FindFunc(identifier,arg1,arg2)) ){
    Result = (*FuncPtr)(arg1,arg2);
  }else{
    Result = new sMathError(sString("undefined '")+identifier
                           +"("+arg1->TrueType()->Name()
                           +","+arg2->TrueType()->Name()+")'"
                           ,start_pos,Cursor);
  }
  return Result;
}


psMathValue  sMath::Call (rcsString    identifier
                         ,int          start_pos
                         ,psMathValue  arg1
                         ,psMathValue  arg2
                         ,psMathValue  arg3      ) const
{
  psMathValue                 Result(NULL);
  pfFunc3Arg                  FuncPtr;
  if( (bool)(FuncPtr = FindFunc(identifier,arg1,arg2,arg3)) ){
    Result = (*FuncPtr)(arg1,arg2,arg3);
  }else{
    Result = new sMathError(sString("undefined '")+identifier
                           +"("+arg1->TrueType()->Name()
                           +","+arg2->TrueType()->Name()
                           +","+arg3->TrueType()->Name()+")'"
                           ,start_pos,Cursor);
  }
  return Result;
}


psMathValue  sMath::Call (rcsString    identifier
                         ,int          start_pos
                         ,psMathValue  arg1
                         ,psMathValue  arg2
                         ,psMathValue  arg3
                         ,psMathValue  arg4      ) const
{
  psMathValue                 Result(NULL);
  pfFunc4Arg                  FuncPtr;
  if( (bool)(FuncPtr = FindFunc(identifier,arg1,arg2,arg3,arg4)) ){
    Result = (*FuncPtr)(arg1,arg2,arg3,arg4);
  }else{
    Result = new sMathError(sString("undefined '")+identifier
                           +"("+arg1->TrueType()->Name()
                           +","+arg2->TrueType()->Name()
                           +","+arg3->TrueType()->Name()
                           +","+arg4->TrueType()->Name()+")'"
                           ,start_pos,Cursor);
  }
  return Result;
}


sMath::pfFunc0Arg  sMath::FindFunc (rcsString    identifier) const
{
  pfFunc0Arg                  FuncPtr = NULL;
  for( sDefinitions::sFuncList::const_iterator
      O = Definitions.FuncList.begin() ; O != Definitions.FuncList.end()
                                       ; O++ ){
    if(   ( (*O).Id == identifier )
        && ! (*O).Type1 && ! (*O).Type2 && ! (*O).Type3 && ! (*O).Type4 ){
      FuncPtr = (*O).Func0Arg;
      RANet::Log.Put(sLog::Debug,"Math"
                ,sString("sMath::FindFunc(")+identifier+"()");
      break;
    }
  }
  return FuncPtr;
}


sMath::pfFunc1Arg  sMath::FindFunc (rcsString    identifier
                                   ,psMathValue  arg1      ) const
{
  pfFunc1Arg                  FuncPtr = NULL;
  for( sDefinitions::sFuncList::const_iterator
      O = Definitions.FuncList.begin() ; O != Definitions.FuncList.end()
                                       ; O++ ){
    if(   ( (*O).Id == identifier )
        && (*O).Type1 && (*O).Type1->CanCast(arg1)
        && ! (*O).Type2 && ! (*O).Type3 && ! (*O).Type4 ){
      FuncPtr = (*O).Func1Arg;
      RANet::Log.Put(sLog::Debug,"Math"
                ,sString("sMath::FindFunc(")+identifier+","
                +arg1->TrueType()->Name()+"->"
                +(*O).Type1->Name()+")");
      break;
    }
  }
  return FuncPtr;
}


sMath::pfFunc2Arg  sMath::FindFunc (rcsString    identifier
                                   ,psMathValue  arg1
                                   ,psMathValue  arg2      ) const
{
  pfFunc2Arg                  FuncPtr = NULL;
  for( sDefinitions::sFuncList::const_iterator
      O = Definitions.FuncList.begin() ; O != Definitions.FuncList.end()
                                       ; O++ ){
    if(   ( (*O).Id == identifier )
        && (*O).Type1 && (*O).Type1->CanCast(arg1)
        && (*O).Type2 && (*O).Type2->CanCast(arg2)
        && ! (*O).Type3 && ! (*O).Type4 ){
      FuncPtr = (*O).Func2Arg;
      RANet::Log.Put(sLog::Debug,"Math"
                ,sString("sMath::FindFunc(")+identifier+","
                +arg1->TrueType()->Name()+"->"
                +(*O).Type1->Name()+","
                +arg2->TrueType()->Name()+"->"
                +(*O).Type2->Name()+")");
      break;
    }
  }
  return FuncPtr;
}


sMath::pfFunc3Arg  sMath::FindFunc (rcsString    identifier
                                   ,psMathValue  arg1
                                   ,psMathValue  arg2
                                   ,psMathValue  arg3      ) const
{
  pfFunc3Arg                  FuncPtr = NULL;
  for( sDefinitions::sFuncList::const_iterator
      O = Definitions.FuncList.begin() ; O != Definitions.FuncList.end()
                                       ; O++ ){
    if(   ( (*O).Id == identifier )
        && (*O).Type1 && (*O).Type1->CanCast(arg1)
        && (*O).Type2 && (*O).Type2->CanCast(arg2)
        && (*O).Type3 && (*O).Type3->CanCast(arg3)
        && ! (*O).Type4 ){
      FuncPtr = (*O).Func3Arg;
      break;
    }
  }
  return FuncPtr;
}


sMath::pfFunc4Arg  sMath::FindFunc (rcsString    identifier
                                   ,psMathValue  arg1
                                   ,psMathValue  arg2
                                   ,psMathValue  arg3
                                   ,psMathValue  arg4      ) const
{
  pfFunc4Arg                  FuncPtr = NULL;
  for( sDefinitions::sFuncList::const_iterator
      O = Definitions.FuncList.begin() ; O != Definitions.FuncList.end()
                                       ; O++ ){
    if(   ( (*O).Id == identifier )
        && (*O).Type1 && (*O).Type1->CanCast(arg1)
        && (*O).Type2 && (*O).Type2->CanCast(arg2)
        && (*O).Type3 && (*O).Type3->CanCast(arg3)
        && (*O).Type4 && (*O).Type4->CanCast(arg4) ){
      FuncPtr = (*O).Func4Arg;
      break;
    }
  }
  return FuncPtr;
}


pcsMathValue  sMath::FindConst (rcsString identifier) const
{
  for( sDefinitions::sConstList::const_iterator
      C = Definitions.ConstList.begin() ; C != Definitions.ConstList.end()
                                       ; C++ ){
    if( (*C).Id == identifier ){
      return (*C).Value;
    }
  }
  return NULL;
}


psMathValue  sMath::UnaryMinusReal (pcsMathValue  arg)
{
  pcsRealValue  Arg(dynamic_cast<pcsRealValue>(arg));
  ASSERT( Arg );
  return new sRealValue( - real(*Arg) );
}


psMathValue  sMath::RealPlusReal (pcsMathValue  arg1
                                 ,pcsMathValue  arg2)
{
  pcsRealValue  Arg1(dynamic_cast<pcsRealValue>(arg1));
  pcsRealValue  Arg2(dynamic_cast<pcsRealValue>(arg2));
  ASSERT( Arg1 && Arg2 );
  return new sRealValue( real(*Arg1) + real(*Arg2) );
}


psMathValue  sMath::RealMinusReal (pcsMathValue  arg1
                                  ,pcsMathValue  arg2)
{
  pcsRealValue  Arg1(dynamic_cast<pcsRealValue>(arg1));
  pcsRealValue  Arg2(dynamic_cast<pcsRealValue>(arg2));
  ASSERT( Arg1 && Arg2 );
  return new sRealValue( real(*Arg1) - real(*Arg2) );
}


psMathValue  sMath::RealMultReal (pcsMathValue  arg1
                                 ,pcsMathValue  arg2)
{
  pcsRealValue  Arg1(dynamic_cast<pcsRealValue>(arg1));
  pcsRealValue  Arg2(dynamic_cast<pcsRealValue>(arg2));
  ASSERT( Arg1 && Arg2 );
  return new sRealValue( real(*Arg1) * real(*Arg2) );
}


psMathValue  sMath::RealDivReal (pcsMathValue  arg1
                                ,pcsMathValue  arg2)
{
  pcsRealValue  Arg1(dynamic_cast<pcsRealValue>(arg1));
  pcsRealValue  Arg2(dynamic_cast<pcsRealValue>(arg2));
  ASSERT( Arg1 && Arg2 );
  return new sRealValue( real(*Arg1) / real(*Arg2) );
}


psMathValue  sMath::RealPowerReal (pcsMathValue  arg1
                                  ,pcsMathValue  arg2)
{
  pcsRealValue  Arg1(dynamic_cast<pcsRealValue>(arg1));
  pcsRealValue  Arg2(dynamic_cast<pcsRealValue>(arg2));
  ASSERT( Arg1 && Arg2 );
  return new sRealValue( ::exp( ::log(real(*Arg1)) * real(*Arg2) ) );
}


psMathValue  sMath::UnaryMinusInteger (pcsMathValue  arg)
{
  pcsIntegerValue  Arg(dynamic_cast<pcsIntegerValue>(arg));
  ASSERT( Arg );
  return new sIntegerValue( - Arg->Integer);
}


psMathValue  sMath::IntegerPlusInteger (pcsMathValue  arg1
                                       ,pcsMathValue  arg2)
{
  pcsIntegerValue  Arg1(dynamic_cast<pcsIntegerValue>(arg1));
  pcsIntegerValue  Arg2(dynamic_cast<pcsIntegerValue>(arg2));
  ASSERT( Arg1 && Arg2 );
  return new sIntegerValue( Arg1->Integer + Arg2->Integer );
}


psMathValue  sMath::IntegerMinusInteger (pcsMathValue  arg1
                                        ,pcsMathValue  arg2)
{
  pcsIntegerValue  Arg1(dynamic_cast<pcsIntegerValue>(arg1));
  pcsIntegerValue  Arg2(dynamic_cast<pcsIntegerValue>(arg2));
  ASSERT( Arg1 && Arg2 );
  return new sIntegerValue( Arg1->Integer - Arg2->Integer );
}


psMathValue  sMath::IntegerMultInteger (pcsMathValue  arg1
                                       ,pcsMathValue  arg2)
{
  pcsIntegerValue  Arg1(dynamic_cast<pcsIntegerValue>(arg1));
  pcsIntegerValue  Arg2(dynamic_cast<pcsIntegerValue>(arg2));
  ASSERT( Arg1 && Arg2 );
  return new sIntegerValue( Arg1->Integer * Arg2->Integer );
}


psMathValue  sMath::lg (pcsMathValue  arg)
{
  pcsRealValue  Arg(dynamic_cast<pcsRealValue>(arg));
  ASSERT( Arg );
  return new sRealValue( log10(real(*Arg)) );
}


psMathValue  sMath::ln (pcsMathValue  arg)
{
  pcsRealValue  Arg(dynamic_cast<pcsRealValue>(arg));
  ASSERT( Arg );
  return new sRealValue( ::log(real(*Arg)) );
}


psMathValue  sMath::log (pcsMathValue  arg1  ,pcsMathValue  arg2)
{
  pcsRealValue  Arg1(dynamic_cast<pcsRealValue>(arg1));
  pcsRealValue  Arg2(dynamic_cast<pcsRealValue>(arg2));
  ASSERT( Arg1 && Arg2 );
  return new sRealValue( ::log(real(*Arg2)) / ::log(real(*Arg1)) );
}


psMathValue  sMath::exp (pcsMathValue  arg)
{
  pcsRealValue  Arg(dynamic_cast<pcsRealValue>(arg));
  ASSERT( Arg );
  return new sRealValue( ::exp(real(*Arg)) );
}


psMathValue  sMath::sin (pcsMathValue  arg)
{
  pcsRealValue  Arg(dynamic_cast<pcsRealValue>(arg));
  ASSERT( Arg );
  return new sRealValue( ::sin(real(*Arg)) );
}


psMathValue  sMath::cos (pcsMathValue  arg)
{
  pcsRealValue  Arg(dynamic_cast<pcsRealValue>(arg));
  ASSERT( Arg );
  return new sRealValue( ::cos(real(*Arg)) );
}


psMathValue  sMath::tg (pcsMathValue  arg)
{
  pcsRealValue  Arg(dynamic_cast<pcsRealValue>(arg));
  ASSERT( Arg );
  return new sRealValue( ::tan(real(*Arg)) );
}


psMathValue  sMath::ctg (pcsMathValue  arg)
{
  pcsRealValue  Arg(dynamic_cast<pcsRealValue>(arg));
  ASSERT( Arg );
  return new sRealValue( 1.0 / ::tan(real(*Arg)) );
}

//------------------------------------------------------------------------------
} //namespace RA



