////////////////////////////////////////////////////////////////////////////////
/*! @file BasicMath.h   Basic expression interpreter.
- Part of RANet - Research Assistant Net Library.
- Copyright(C) 2014, Viktor E. Bursian, St.Petersburg, Russia.
                     Viktor.Bursian@mail.ioffe.ru
*///////////////////////////////////////////////////////////////////////////////
#ifndef BasicMath_H
#define BasicMath_H
#include "Storable.h"
#include <list>
namespace RA {
//------------------------------------------------------------------------------

ANNOUNCE_CLASS(sMath)
ANNOUNCE_CLASS(sMathValue)
ANNOUNCE_CLASS(sMathError)
ANNOUNCE_CLASS(sRealValue)
ANNOUNCE_CLASS(sIntegerValue)

//--------------------------------------------------------------- sMathValue ---

class RANet_EXPORT  sMathValue : virtual public sStorable
{
  ABSTRACT_STORABLE(sMathValue)
  public:
//    virtual sString           Text () const;
    virtual sString           Text (eTextFormat        format=Plain
                                   ,eTextDetalization  detalization=Casual)
                                   const;
                                //!< Visual representation for user
    virtual psMathValue       Replica () const =0;

  protected:
                              sMathValue ()
                                  :sStorable()
                                {}
                              sMathValue (rcsMathValue);
  private:
    rsMathValue               operator = (rcsMathValue);
};

//--------------------------------------------------------------- sMathError ---

class RANet_EXPORT  sMathError : public sMathValue
{
  STORABLE(sMathError)
  public:
                              sMathError (rcsString  message
                                         ,int        cursor1 = -1
                                         ,int        cursor2 = -1)
                                  :sMathValue(),Message(message)
                                  ,Cursor1(cursor1),Cursor2(cursor2)
                                {}
    virtual sString           Text (eTextFormat        format=Plain
                                   ,eTextDetalization  detalization=Casual)
                                   const;
                                //!< Visual representation for user
    virtual psMathValue       Replica () const
                                { return new sMathError(*this); }

  protected:
                              sMathError (rcsMathError);

  public: //fields
    sString                   Message;
    int                       Cursor1;
    int                       Cursor2;
};

//--------------------------------------------------------------- sRealValue ---

class RANet_EXPORT  sRealValue : public sMathValue
{
  STORABLE(sRealValue)
  public:
                              sRealValue (real  value)
                                  :sMathValue(),Real(value)
                                {}
    virtual sString           Text (eTextFormat        format=Plain
                                   ,eTextDetalization  detalization=Casual)
                                   const;
                                //!< Visual representation for user
    virtual psMathValue       Replica () const
                                { return new sRealValue(*this); }
                              operator real () const
                                { return Real; }

  protected:
                              sRealValue (rcsRealValue);

  protected: //fields
    real                      Real;
};

//------------------------------------------------------------ sIntegerValue ---

class RANet_EXPORT  sIntegerValue : public sRealValue
{
  STORABLE(sIntegerValue)
  public:
                              sIntegerValue (integer  value)
                                  :sRealValue(value),Integer(value)
                                {}
    virtual sString           Text (eTextFormat        format=Plain
                                   ,eTextDetalization  detalization=Casual)
                                   const;
                                //!< Visual representation for user
    virtual psMathValue       Replica () const
                                { return new sIntegerValue(*this); }
                              operator integer () const
                                { return Integer; }

  protected:
                              sIntegerValue (rcsIntegerValue);

  private: //fields
    integer                   Integer;
  friend class sMath;
};

//-------------------------------------------------------------------- sMath ---
/*! Performs calculation of an expression over phys.values and functions.
*/
class RANet_EXPORT  sMath
{
  public://types
    typedef psMathValue       fFunc0Arg ();
    typedef psMathValue       fFunc1Arg (pcsMathValue);
    typedef psMathValue       fFunc2Arg (pcsMathValue,pcsMathValue);
    typedef psMathValue       fFunc3Arg (pcsMathValue,pcsMathValue
                                        ,pcsMathValue);
    typedef psMathValue       fFunc4Arg (pcsMathValue,pcsMathValue
                                        ,pcsMathValue,pcsMathValue);
    typedef fFunc0Arg *       pfFunc0Arg;
    typedef fFunc1Arg *       pfFunc1Arg;
    typedef fFunc2Arg *       pfFunc2Arg;
    typedef fFunc3Arg *       pfFunc3Arg;
    typedef fFunc4Arg *       pfFunc4Arg;

    //.......................................... sMath::sDefinitions ...
    class RANet_EXPORT  sDefinitions
    {
      public:
        void                      Add (pcsMathValue,rcsString);
        void                      Add (pfFunc0Arg,rcsString);
        void                      Add (pfFunc1Arg,rcsString,cpcsTag);
        void                      Add (pfFunc2Arg,rcsString,cpcsTag,cpcsTag);
        void                      Add (pfFunc3Arg,rcsString
                                      ,cpcsTag,cpcsTag,cpcsTag);
        void                      Add (pfFunc4Arg,rcsString
                                      ,cpcsTag,cpcsTag,cpcsTag,cpcsTag);
      private:
        //............................ sMath::sDefinitions::sFunc ...
        struct RANet_EXPORT  sFunc
        {
                                      sFunc (rcsString  id
                                            ,cpcsTag  type1 = NULL
                                            ,cpcsTag  type2 = NULL
                                            ,cpcsTag  type3 = NULL
                                            ,cpcsTag  type4 = NULL
                                            )
                                           :Id(id),Type1(type1),Type2(type2)
                                           ,Type3(type3),Type4(type4)
                                         {}
            sString                   Id;
            cpcsTag                   Type1;
            cpcsTag                   Type2;
            cpcsTag                   Type3;
            cpcsTag                   Type4;
            union
            {
                pfFunc0Arg                Func0Arg;
                pfFunc1Arg                Func1Arg;
                pfFunc2Arg                Func2Arg;
                pfFunc3Arg                Func3Arg;
                pfFunc4Arg                Func4Arg;
            };
        };
        //............................ sMath::sDefinitions::sConst ...
        struct RANet_EXPORT  sConst
        {
                                      sConst (rcsString     id
                                             ,pcsMathValue  value)
                                           :Id(id),Value(value)
                                         {}
            sString                   Id;
            pcsMathValue              Value;
        };
        //..............................................................
        typedef std::list<sFunc>  sFuncList;
        typedef std::list<sConst> sConstList;

                                  sDefinitions ();
        sFuncList                 FuncList;
        sConstList                ConstList;
      friend class sMath;
    };

    //......................................................................

  public: //static member
    static sDefinitions       Definitions;
    static const real         pi;

  public: //static methods
    static bool               Valid (psMathValue);
    static void               AddConstant (real value, rcsString identifier);
//    static psMathValue             Calculate (rcsString  source);

  public:
                              sMath (rcsString  source);
    virtual psMathValue       Calculate ();


  protected:
    virtual bool              End () const
                                { return Cursor > SourceLength-1; }
    virtual void              PassSpaces ();
    virtual psMathValue       Expression ();
    virtual bool              AddingOperator (rsString  op);
    virtual bool              MultiplyingOperator (rsString  op);
    virtual bool              PowerOperator (rsString  op);
    virtual psMathValue       Addend ();
    virtual psMathValue       Factor ();
    virtual psMathValue       Term ();
    virtual psMathValue       Function (rcsString  identifier
                                       ,int        start_pos );
    virtual psMathValue       Constant (rcsString  identifier
                                       ,int        start_pos ) const;
    virtual psMathValue       Value ();
    virtual bool              Identifier (rsString  result);
    virtual bool              IntegerNumber (integer &  result);
    virtual bool              RealNumber (real &  result
                                         ,real &  precision);
    virtual bool              IsLetter (const char  c) const;
    virtual bool              IsDigit (const char  c) const;
    virtual psMathValue       Call (rcsString    identifier
                                   ,int          start_pos ) const;
    virtual psMathValue       Call (rcsString    identifier
                                   ,int          start_pos
                                   ,psMathValue  arg1      ) const;
    virtual psMathValue       Call (rcsString    identifier
                                   ,int          start_pos
                                   ,psMathValue  arg1
                                   ,psMathValue  arg2      ) const;
    virtual psMathValue       Call (rcsString    identifier
                                   ,int          start_pos
                                   ,psMathValue  arg1
                                   ,psMathValue  arg2
                                   ,psMathValue  arg3      ) const;
    virtual psMathValue       Call (rcsString    identifier
                                   ,int          start_pos
                                   ,psMathValue  arg1
                                   ,psMathValue  arg2
                                   ,psMathValue  arg3
                                   ,psMathValue  arg4      ) const;
    pfFunc0Arg                FindFunc (rcsString    identifier) const;
    pfFunc1Arg                FindFunc (rcsString    identifier
                                       ,psMathValue  arg1      ) const;
    pfFunc2Arg                FindFunc (rcsString    identifier
                                       ,psMathValue  arg1
                                       ,psMathValue  arg2      ) const;
    pfFunc3Arg                FindFunc (rcsString    identifier
                                       ,psMathValue  arg1
                                       ,psMathValue  arg2
                                       ,psMathValue  arg3      ) const;
    pfFunc4Arg                FindFunc (rcsString    identifier
                                       ,psMathValue  arg1
                                       ,psMathValue  arg2
                                       ,psMathValue  arg3
                                       ,psMathValue  arg4      ) const;
    pcsMathValue              FindConst (rcsString    identifier) const;

  protected://fields
    sString                   Source;
    int                       SourceLength;
    int                       Cursor;
    int                       OKCursor;

  private:
    static psMathValue        UnaryMinusReal (pcsMathValue);
    static psMathValue        RealPlusReal (pcsMathValue,pcsMathValue);
    static psMathValue        RealMinusReal (pcsMathValue,pcsMathValue);
    static psMathValue        RealMultReal (pcsMathValue,pcsMathValue);
    static psMathValue        RealDivReal (pcsMathValue,pcsMathValue);
    static psMathValue        RealPowerReal (pcsMathValue,pcsMathValue);

    static psMathValue        UnaryMinusInteger (pcsMathValue);
    static psMathValue        IntegerPlusInteger (pcsMathValue,pcsMathValue);
    static psMathValue        IntegerMinusInteger (pcsMathValue,pcsMathValue);
    static psMathValue        IntegerMultInteger (pcsMathValue,pcsMathValue);

    static psMathValue        lg (pcsMathValue);
    static psMathValue        ln (pcsMathValue);
    static psMathValue        log (pcsMathValue,pcsMathValue);
    static psMathValue        exp (pcsMathValue);
    static psMathValue        sin (pcsMathValue);
    static psMathValue        cos (pcsMathValue);
    static psMathValue        tg (pcsMathValue);
    static psMathValue        ctg (pcsMathValue);
};

//------------------------------------------------------------------------------
} //namespace RA
#endif
