////////////////////////////////////////////////////////////////////////////////
/*! @file BasicNodes.cpp   Базовые типы терминальных объектов сети.
- Part of RANet - Research Assistant Net Library (based on ANSI C++).
- Copyright(C) 1999-2013, Viktor E. Bursian, St.Petersburg, Russia.
                     Viktor.Bursian@mail.ioffe.ru
*///////////////////////////////////////////////////////////////////////////////
#include "BasicNodes.h"
#include "Symbols.h"
#include <cstring>
#include <QMutexLocker>
namespace RA {
//------------------------------------------------------------------------------

//DEFINE_CLASS_TAG(sUndoAssignValue2BooleanNode)
//DEFINE_CLASS_TAG(sUndoAssignValue2IntegerNode)
//DEFINE_CLASS_TAG(sUndoAssignValue2TextNode)
//DEFINE_CLASS_TAG(sUndoAssignValue2PhysValueNode)
//DEFINE_CLASS_TAG(sUndoAssignValue2PhysRangeNode)

//------------------------------------------------------------ sTerminalNode ---

sTerminalNode::sTerminalNode (std::istream &  a_stream ,rcsVersion  version)
  :sNode(a_stream,version)
{
}


void  sTerminalNode::Store (std::ostream &  a_stream) const
{
  sNode::Store(a_stream);
}

//------------------------------------------------------------- sBooleanNode ---

sBooleanNode::sBooleanNode (std::istream &  a_stream ,rcsVersion  version)
    :sTerminalNode(a_stream,version)
{
  a_stream.READ_INTO(TheValue);
}


void sBooleanNode::Store (std::ostream &  a_stream) const
{
  sTerminalNode::Store(a_stream);
  a_stream.WRITE_FROM(TheValue);
}


sString  sBooleanNode::Text (eTextFormat        /*F*/
                            ,eTextDetalization  /*D*/)
{
  sString                     T( (Value() ? "true" : "false") );
  return T;
}


bool  sBooleanNode::Value ()
{
  QMutexLocker                BodyLocker(&BodyMutex);
  return TheValue;
}


void  sBooleanNode::Assign (bool  new_value)
{
  QMutexLocker                BodyLocker(&BodyMutex);
//  sNet::Manager->Undo.Push( new sUndoAssignValue2BooleanNode(sNodePtr(this),B) );
  //!@todo{Net} Assign нигде не проверяет, что значение другое.  Нехорошо.
  if( new_value !=  TheValue ){
    TheValue = new_value;
    MarkAsEdited();
  }
}


void  sBooleanNode::CopyFrom (psNode  other)
{
  //!@todo{Net} CopyFrom нигде не берёт мьютекса.  Нехорошо?
  sTerminalNode::CopyFrom(other);
  psBooleanNode               Other = dynamic_cast<psBooleanNode>(other);
  if( Other ){
    bool                        V = Other->Value();
    if( TheValue != V ){
      TheValue = V;
      MarkAsEdited();
    }
  }
}

//------------------------------------------------------------- sIntegerNode ---

sIntegerNode::sIntegerNode (std::istream &  a_stream ,rcsVersion  version)
    :sTerminalNode(a_stream,version)
{
  a_stream.READ_INTO(TheValue);
}


void sIntegerNode::Store (std::ostream &  a_stream) const
{
  sTerminalNode::Store(a_stream);
  a_stream.WRITE_FROM(TheValue);
}


sString  sIntegerNode::Text (eTextFormat        /*F*/
                            ,eTextDetalization  /*D*/)
{
  return sString::FromInteger(Value());
}


integer  sIntegerNode::Value()
{
  QMutexLocker                BodyLocker(&BodyMutex);
  return TheValue;
}


void  sIntegerNode::Assign (integer  new_value)
{
  QMutexLocker                BodyLocker(&BodyMutex);
//  sNet::Manager->Undo.Push( new sUndoAssignValue2IntegerNode(sNodePtr(this),I) );
  if( new_value !=  TheValue ){
    TheValue = new_value;
    MarkAsEdited();
  }
}


void  sIntegerNode::CopyFrom (psNode  other)
{
  sTerminalNode::CopyFrom(other);
  psIntegerNode               Other = dynamic_cast<psIntegerNode>(other);
  if( Other ){
    integer                     V = Other->Value();
    if( TheValue != V ){
      TheValue = V;
      MarkAsEdited();
    }
  }
}

//----------------------------------------------------------- sByteArrayNode ---

sByteArrayNode::sByteArrayNode (const byte *  data_ptr
                               ,size_t        data_size)
    :sTerminalNode(),TheValue(data_size)
{
  std::memcpy(TheValue.data(),data_ptr,data_size);
}


sByteArrayNode::sByteArrayNode (const char *  data_ptr
                               ,size_t        data_size)
    :sTerminalNode(),TheValue(data_size)
{
  std::memcpy(TheValue.data(),data_ptr,data_size);
}


sByteArrayNode::sByteArrayNode (std::istream &  a_stream ,rcsVersion  version)
    :sTerminalNode(a_stream,version)
{
//  size_t                      Size(TheValue.size());
  uint32_t                    Size(TheValue.size());
  a_stream.READ_INTO(Size);
  TheValue.resize(Size);
  a_stream.read( (char *)(TheValue.data()) , Size );
}


void sByteArrayNode::Store (std::ostream &  a_stream) const
{
  sTerminalNode::Store(a_stream);
//  size_t                      Size(TheValue.size());
  uint32_t                    Size(TheValue.size());
  a_stream.WRITE_FROM(Size);
  a_stream.write( (const char *)(TheValue.data()) , Size );
}


sString  sByteArrayNode::Text (eTextFormat        /*F*/
                              ,eTextDetalization  /*D*/)
{
  return sString(Symbol::BinaryData) + Symbol::BinaryData;
}


std::vector<byte>  sByteArrayNode::Value()
{
  QMutexLocker                BodyLocker(&BodyMutex);
  return TheValue;
}


void  sByteArrayNode::Assign (const std::vector<byte> &  new_value)
{
  QMutexLocker                BodyLocker(&BodyMutex);
//  sNet::Manager->Undo.Push(
  if( new_value !=  TheValue ){
    TheValue = new_value;
    MarkAsEdited();
  }
}


void  sByteArrayNode::CopyFrom (psNode  other)
{
  sTerminalNode::CopyFrom(other);
  psByteArrayNode             Other = dynamic_cast<psByteArrayNode>(other);
  if( Other ){
    std::vector<byte>           V = Other->Value();
    if( TheValue != V ){
      TheValue = V;
      MarkAsEdited();
    }
  }
}

//---------------------------------------------------------------- sTextNode ---

sTextNode::sTextNode (std::istream &  a_stream ,rcsVersion  version)
    :sTerminalNode(a_stream,version)
{
  TheValue.Read(a_stream);
}


void sTextNode::Store (std::ostream &  a_stream) const
{
  sTerminalNode::Store(a_stream);
  TheValue.Write(a_stream);
}


sString  sTextNode::Text (eTextFormat        /*F*/
                         ,eTextDetalization  D)
{
  sString                     T(Value());
  /*! @todo{UI} implement HTML -> Plain*/
  if( (D == Laconic) && (T.Length() > CutOffLengthForLaconic) ){
    T=T.Substr(0,CutOffLengthForLaconic-strlen(Symbol::Ellipsis));
    /*! @todo{sString} Substr for UTF-8*/
    while( T[T.Length()-1] < 0 )  T=T.Substr(0,T.Length()-1);
    T+=Symbol::Ellipsis;
  };
  return T;
}


sString  sTextNode::Value ()
{
  QMutexLocker                BodyLocker(&BodyMutex);
  return TheValue;
}


void  sTextNode::Assign (rcsString  new_value)
{
  QMutexLocker                BodyLocker(&BodyMutex);
//  sNet::Manager->Undo.Push( new sUndoAssignValue2TextNode(sNodePtr(this)
//                                                     ,value) );
  if( new_value !=  TheValue ){
    TheValue = new_value;
    MarkAsEdited();
  }
}


void  sTextNode::CopyFrom (psNode  other)
{
  sTerminalNode::CopyFrom(other);
  psTextNode                  Other = dynamic_cast<psTextNode>(other);
  if( Other ){
    sString                     V = Other->Value();
    if( TheValue != V ){
      TheValue = V;
      MarkAsEdited();
    }
  }
}

//----------------------------------------------------------- sPhysValueNode ---

sPhysValueNode::sPhysValueNode (std::istream &  a_stream ,rcsVersion  version)
  :sTerminalNode(a_stream,version)
  ,TheValue(a_stream,version)
{
}


void sPhysValueNode::Store (std::ostream &  a_stream) const
{
  sTerminalNode::Store(a_stream);
  TheValue.Store(a_stream);
}


sString  sPhysValueNode::Text (eTextFormat        F
                              ,eTextDetalization  D)
{
  return Value().Text(F,D);
}


sPhysValue  sPhysValueNode::Value ()
{
  QMutexLocker                BodyLocker(&BodyMutex);
  return TheValue;
}


void  sPhysValueNode::Assign (sPhysValue V)
{
  QMutexLocker                BodyLocker(&BodyMutex);
//  sNet::Manager->Undo.Push( new sUndoAssignValue2PhysValueNode(sNodePtr(this),V) );
  TheValue = V;
    /*! @todo{Net} Для sPhysValue и sPhysRange нет возможности проверить
    изменилось ли Value. Приходится помечать узел как отредактированный - это
    сообщается всем sNodePtr и происходит (возможно лишнее) обновление,
    например, навигатора.
    */
  MarkAsEdited();
}


void  sPhysValueNode::CopyFrom (psNode  other)
{
  sTerminalNode::CopyFrom(other);
  psPhysValueNode             Other = dynamic_cast<psPhysValueNode>(other);
  if( Other ){
    TheValue = Other->Value();
    /*! @todo{Net} Здесь, как и в Assign, не всегда обоснованный вызов
    MarkAsEdited.  При этом, поскольку пока CopyFrom() вызывается только
    из процесса подкачки, сам узел сразу же получает статус cSynchronized
    (обратно, как и было, всё в критической секции).  То есть пока не страшно,
    но нехорошо.
    */
    MarkAsEdited();
  }
}


//----------------------------------------------------------- sPhysRangeNode ---

sPhysRangeNode::sPhysRangeNode (std::istream &  a_stream ,rcsVersion  version)
  :sTerminalNode(a_stream,version)
  ,TheValue(a_stream,version)
{
}


void sPhysRangeNode::Store (std::ostream &  a_stream) const
{
  sTerminalNode::Store(a_stream);
  TheValue.Store(a_stream);
}


sString  sPhysRangeNode::Text (eTextFormat        F
                              ,eTextDetalization  /*D*/)
{
  return Value().Text(F);
}


sPhysRange  sPhysRangeNode::Value ()
{
  QMutexLocker                BodyLocker(&BodyMutex);
  return TheValue;
}


void  sPhysRangeNode::Assign (sPhysRange V)
{
  QMutexLocker                BodyLocker(&BodyMutex);
//  sNet::Manager->Undo.Push( new sUndoAssignValue2PhysRangeNode(sNodePtr(this),V) );
  TheValue=V;
  MarkAsEdited();
}


void  sPhysRangeNode::CopyFrom (psNode  other)
{
  sTerminalNode::CopyFrom(other);
  psPhysRangeNode             Other = dynamic_cast<psPhysRangeNode>(other);
  if( Other ){
    TheValue = Other->Value();
    MarkAsEdited();
  }
}

//--------------------------------------------- sUndoAssignValue2BooleanNode ---

//void  sUndoAssignValue2BooleanNode::UndoIt ()
//{
//  rsBooleanNode(*NodePtr).Assign(OldValue);
//};


//void  sUndoAssignValue2BooleanNode::RedoIt ()
//{
//  rsBooleanNode(*NodePtr).Assign(NewValue);
//};


//sString  sUndoAssignValue2BooleanNode::Hint ()
//{
//  sString                     Old( (OldValue ? "true" : "false") );
//  sString                     New( (NewValue ? "true" : "false") );
//  return  sString("change value '") + Old
//        + sString("' -> '") + New
//        + sString("'");
//};

////--------------------------------------------- sUndoAssignValue2IntegerNode ---

//void  sUndoAssignValue2IntegerNode::UndoIt ()
//{
//  rsIntegerNode(*NodePtr).Assign(OldValue);
//};


//void  sUndoAssignValue2IntegerNode::RedoIt ()
//{
//  rsIntegerNode(*NodePtr).Assign(NewValue);
//};


//sString  sUndoAssignValue2IntegerNode::Hint ()
//{
//  sString                     Old;
//  sString                     New;
//  Old<<OldValue;
//  New<<NewValue;
//  return  sString("change value '") + Old
//        + sString("' -> '") + New
//        + sString("'");
//};

////------------------------------------------------ sUndoAssignValue2TextNode ---

//void  sUndoAssignValue2TextNode::UndoIt ()
//{
//  rsTextNode(*NodePtr).SetValue( literal(OldValue) );
//};


//void  sUndoAssignValue2TextNode::RedoIt ()
//{
//  rsTextNode(*NodePtr).SetValue( literal(NewValue) );
//};


//sString  sUndoAssignValue2TextNode::Hint ()
//{
//  return  sString("change value '") + OldValue
//        + sString("' -> '") + NewValue + sString("'");
//};

////------------------------------------------- sUndoAssignValue2PhysValueNode ---

//void  sUndoAssignValue2PhysValueNode::UndoIt ()
//{
//  rsPhysValueNode(*NodePtr).Assign(OldValue);
//};


//void  sUndoAssignValue2PhysValueNode::RedoIt ()
//{
//  rsPhysValueNode(*NodePtr).Assign(NewValue);
//};


//sString  sUndoAssignValue2PhysValueNode::Hint ()
//{
//  return  sString("change value '") + OldValue.Text()
//        + sString("' -> '") + NewValue.Text()
//        + sString("'");
//};


////------------------------------------------- sUndoAssignValue2PhysRangeNode ---

//void  sUndoAssignValue2PhysRangeNode::UndoIt ()
//{
//  rsPhysRangeNode(*NodePtr).Assign(OldValue);
//};


//void  sUndoAssignValue2PhysRangeNode::RedoIt ()
//{
//  rsPhysRangeNode(*NodePtr).Assign(NewValue);
//};


//sString  sUndoAssignValue2PhysRangeNode::Hint ()
//{
//  return  sString("change range '") + OldValue.Text()
//        + sString("' -> '") + NewValue.Text()
//        + sString("'");
//};

//------------------------------------------------------------------------------
} //namespace RA

