////////////////////////////////////////////////////////////////////////////////
/*! @file BasicNodes.h   Базовые типы терминальных объектов сети.
- Part of RANet - Research Assistant Net Library (based on ANSI C++).
- Copyright(C) 1999-2013, Viktor E. Bursian, St.Petersburg, Russia.
                     Viktor.Bursian@mail.ioffe.ru
*///////////////////////////////////////////////////////////////////////////////
#ifndef BasicNodes_H
#define BasicNodes_H
#include "Nodes.h"
#include "Net.h"
#include "PhysValue.h"
#include "Ranges.h"
#include <vector>
namespace RA {
//------------------------------------------------------------------------------

ANNOUNCE_ABSTRACT_CLASS(sTerminalNode)
ANNOUNCE_CLASS(sBooleanNode)
ANNOUNCE_CLASS(sIntegerNode)
ANNOUNCE_CLASS(sByteArrayNode)
ANNOUNCE_CLASS(sTextNode)
ANNOUNCE_CLASS(sPhysValueNode)
ANNOUNCE_CLASS(sPhysRangeNode)

//template<class node_class ,class variable_type>
//bool  get_from (rcsNodePtr  node_ptr_expression ,variable_type &  variable)


//ANNOUNCE_CLASS(sUndoAssignValue2BooleanNode)
//ANNOUNCE_CLASS(sUndoAssignValue2IntegerNode)
//ANNOUNCE_CLASS(sUndoAssignValue2TextNode)
//ANNOUNCE_CLASS(sUndoAssignValue2PhysValueNode)
//ANNOUNCE_CLASS(sUndoAssignValue2PhysRangeNode)

//-------------------------------------- get_from <node_class,variable_type> ---
/*! Convenience function template for extracting data (usually the optional
    configuration settings) from the net.

*/
template<class node_class ,class variable_type>
bool  get_from (rcsNodePtr  node_ptr_expression ,variable_type &  variable)
{
  sPtr<node_class>            ValPtr (node_ptr_expression);
  if( ValPtr.IsCorrect() ){
    variable = ValPtr->Value();
    return true;
  }
  return false;
}


//------------------------------------------------------------ sTerminalNode ---
/*! Абстрактный базовый класс для терминальных объектов сети, то есть для
    объектов неспособных иметь атрибуты.

*/
class RANet_EXPORT  sTerminalNode : public sNode
{
  ABSTRACT_STORABLE(sTerminalNode)
  protected:
                              sTerminalNode ()
                                  :sNode()
                                {}
};


//------------------------------------------------------------- sBooleanNode ---

class RANet_EXPORT  sBooleanNode : public sTerminalNode
{
  STORABLE(sBooleanNode)

  public:
                              sBooleanNode ()
                                  :sTerminalNode(),TheValue(false)
                                {}
                              sBooleanNode (bool V)
                                  :sTerminalNode(),TheValue(V)
                                {}
    virtual sString           Text (eTextFormat        /*F=Plain*/
                                   ,eTextDetalization  /*D=Casual*/
                                   ); //!< visual representation for user
                              operator bool ()
                                { return Value(); }
    virtual bool              Value ();
    virtual void              Assign (bool);

  protected:
    virtual void              CopyFrom (psNode);

  private:
    bool                      TheValue;

//  friend class sUndoAssignValue2BooleanNode;
};

//------------------------------------------------------------- sIntegerNode ---

class RANet_EXPORT  sIntegerNode : public sTerminalNode
{
  STORABLE(sIntegerNode)

  public:
                              sIntegerNode ()
                                  :sTerminalNode(),TheValue(0)
                                {}
                              sIntegerNode (integer V)
                                  :sTerminalNode(),TheValue(V)
                                {}
    virtual sString           Text (eTextFormat        /*F=Plain*/
                                   ,eTextDetalization  /*D=Casual*/
                                   ); //!< visual representation for user
                              operator integer ()
                                { return Value(); }
    virtual integer           Value ();
    virtual void              Assign (integer);

  protected:
    virtual void              CopyFrom (psNode);

  private:
    integer                   TheValue;

//  friend class sUndoAssignValue2IntegerNode;
};

//----------------------------------------------------------- sByteArrayNode ---

class RANet_EXPORT  sByteArrayNode : public sTerminalNode
{
  STORABLE(sByteArrayNode)

  public:
                              sByteArrayNode ()
                                  :sTerminalNode()
                                {}
                              sByteArrayNode (const std::vector<byte> &  V)
                                  :sTerminalNode(),TheValue(V)
                                {}
                              sByteArrayNode (const byte *  data_ptr
                                             ,size_t        data_size);
                              sByteArrayNode (const char *  data_ptr
                                             ,size_t        data_size);
    virtual sString           Text (eTextFormat        /*F=Plain*/
                                   ,eTextDetalization  /*D=Casual*/
                                   ); //!< visual representation for user
//                              operator integer ()
//                                { return Value(); }
    virtual std::vector<byte> Value ();
    virtual void              Assign (const std::vector<byte> &);

  protected:
    virtual void              CopyFrom (psNode);

  protected:
    std::vector<byte>         TheValue;

//  friend class sUndoAssignValue2ByteArrayNode;
};

//---------------------------------------------------------------- sTextNode ---

class RANet_EXPORT  sTextNode : public sTerminalNode
{
  STORABLE(sTextNode)

  public:
                              sTextNode ()
                                  :sTerminalNode(),TheValue()
                                {}
                              sTextNode (rcsString S)
                                  :sTerminalNode(),TheValue(S)
                                {}
                              sTextNode (literal V)
                                  :sTerminalNode(),TheValue(V)
                                {}
    virtual sString           Text (eTextFormat        /*F=Plain*/
                                   ,eTextDetalization  /*D=Casual*/
                                   ); //!< visual representation for user
                              operator sString ()
                                { return Value(); }
    virtual sString           Value ();
    virtual void              Assign (rcsString);

  protected:
    virtual void              CopyFrom (psNode);

  private:
    sString                   TheValue;

//  friend class sUndoAssignValue2TextNode;
};

//----------------------------------------------------------- sPhysValueNode ---
/*! Объект сети данных типа физ-величина
    (квазивещественное число с размерностью и погрешностью).

*/
class RANet_EXPORT  sPhysValueNode : public sTerminalNode
{
  STORABLE(sPhysValueNode)

  public:
                              sPhysValueNode ()
                                  :sTerminalNode(),TheValue()
                                {}
                              sPhysValueNode (sPhysValue  V)
                                  :sTerminalNode(),TheValue(V)
                                {}
    virtual sString           Text (eTextFormat        F
                                   ,eTextDetalization  D
                                   ); //!< visual representation for user
                              operator sPhysValue ()
                                { return Value(); }
    virtual sPhysValue        Value ();
    virtual void              Assign (sPhysValue);

  protected:
    virtual void              CopyFrom (psNode);

  private:
    sPhysValue                TheValue;

//  friend class sUndoAssignValue2PhysValueNode;
};


//----------------------------------------------------------- sPhysRangeNode ---
/*! Объект сети данных типа диапазон физ-величины.

*/
class RANet_EXPORT  sPhysRangeNode : public sTerminalNode
{
  STORABLE(sPhysRangeNode)

  public:
                              sPhysRangeNode ()
                                  :sTerminalNode(),TheValue()
                                {}
                              sPhysRangeNode (sPhysRange  V)
                                  :sTerminalNode(),TheValue(V)
                                {}
    virtual sString           Text (eTextFormat        F
                                   ,eTextDetalization  /*D*/
                                   ); //!< visual representation for user
                              operator sPhysRange ()
                                { return Value(); }
    virtual sPhysRange        Value ();
    virtual void              Assign (sPhysRange);

  protected:
    virtual void              CopyFrom (psNode);

  private:
    sPhysRange                TheValue;

//  friend class sUndoAssignValue2PhysRangeNode;
};

//--------------------------------------------- sUndoAssignValue2BooleanNode ---

//class RANet_EXPORT  sUndoAssignValue2BooleanNode  : public sUndoNodeCorrectAction
//{
//  public:
//                              sUndoAssignValue2BooleanNode (sNodePtr  NP
//                                                           ,bool      new_value)
//                                  :sUndoNodeCorrectAction(NP)
//                                  ,OldValue(rsBooleanNode(*NP).Value)
//                                  ,NewValue(new_value)
//                                {};
//    virtual void              UndoIt ();
//    virtual void              RedoIt ();
//    virtual sString           Hint ();

//  protected: // fields
//    bool                      OldValue;
//    bool                      NewValue;
//};

//--------------------------------------------- sUndoAssignValue2IntegerNode ---

//class RANet_EXPORT  sUndoAssignValue2IntegerNode  : public sUndoNodeCorrectAction
//{
//  public:
//                              sUndoAssignValue2IntegerNode (sNodePtr  NP
//                                                    ,integer   new_value)
//                                  :sUndoNodeCorrectAction(NP)
//                                  ,OldValue(rsIntegerNode(*NP).Value)
//                                  ,NewValue(new_value)
//                                {};
//    virtual void              UndoIt ();
//    virtual void              RedoIt ();
//    virtual sString           Hint ();

//  protected: // fields
//    integer                   OldValue;
//    integer                   NewValue;
//};

//------------------------------------------------ sUndoAssignValue2TextNode ---

//class RANet_EXPORT  sUndoAssignValue2TextNode  : public sUndoNodeCorrectAction
//{
//  public:
//                              sUndoAssignValue2TextNode (sNodePtr   NP
//                                                        ,rcsString  new_value)
//                                  :sUndoNodeCorrectAction(NP)
//                                  ,OldValue(rsString(rsTextNode(*NP)))
//                                  ,NewValue(new_value)
//                                {};
//    virtual void              UndoIt ();
//    virtual void              RedoIt ();
//    virtual sString           Hint ();

//  protected: // fields
//    sString                   OldValue;
//    sString                   NewValue;
//};

//------------------------------------------- sUndoAssignValue2PhysValueNode ---

//class RANet_EXPORT  sUndoAssignValue2PhysValueNode  : public sUndoNodeCorrectAction
//{
//  public:
//                              sUndoAssignValue2PhysValueNode
//                                          (sNodePtr    NP
//                                          ,sPhysValue  new_value)
//                                  :sUndoNodeCorrectAction(NP)
//                                  ,OldValue(rsPhysValueNode(*NP).Value)
//                                  ,NewValue(new_value)
//                                {};
//    virtual void              UndoIt ();
//    virtual void              RedoIt ();
//    virtual sString           Hint ();

//  protected: // fields
//    sPhysValue                OldValue;
//    sPhysValue                NewValue;
//};


//------------------------------------------- sUndoAssignValue2PhysRangeNode ---

//class RANet_EXPORT  sUndoAssignValue2PhysRangeNode  : public sUndoNodeCorrectAction
//{
//  public:
//                              sUndoAssignValue2PhysRangeNode
//                                      (sNodePtr   NP
//                                      ,sPhysRange new_value)
//                                  :sUndoNodeCorrectAction(NP)
//                                  ,OldValue(rsPhysRangeNode(*NP).Value)
//                                  ,NewValue(new_value)
//                                {};
//    virtual void              UndoIt ();
//    virtual void              RedoIt ();
//    virtual sString           Hint ();

//  protected: // fields
//    sPhysRange                OldValue;
//    sPhysRange                NewValue;
//};

//------------------------------------------------------------------------------
} //namespace RA
#endif

