////////////////////////////////////////////////////////////////////////////////
/*! @file Curves.cpp
- Part of RANet - Research Assistant Net Library.
- Copyright(C) 2000-2015, Viktor E. Bursian, St.Petersburg, Russia.
                          Viktor_dot_Bursian_at_mail_dot_ioffe_dot_ru
*///////////////////////////////////////////////////////////////////////////////
#include "Curves.h"
#include "Symbols.h"
#include "Excepts.h"
#include "Log.h"
#include "Trace.h"
#include <math.h>
#include <limits>
namespace RA {
//------------------------------------------------------------------------------
//--------------------------------------------------------- sCurveAppearance ---

sCurveAppearance::sCurveAppearance ()
    :sAppearance()
    ,sLineAppearance()
    ,sPointAppearance()
    ,ShowLine(true)
    ,ShowPoints(false)
{
}

sCurveAppearance::sCurveAppearance(rcsColor  color)
    :sAppearance()
    ,sLineAppearance(color)
    ,sPointAppearance(color)
    ,ShowLine(true)
    ,ShowPoints(false)
{
}

sCurveAppearance::sCurveAppearance(rcsColor    color
                                  ,byte        line_thickness
                                  ,eLineStyle  line_style)
    :sAppearance()
    ,sLineAppearance(color,line_thickness,line_style)
    ,sPointAppearance(color)
    ,ShowLine(true)
    ,ShowPoints(false)
{
}

sCurveAppearance::sCurveAppearance (byte     point_size
                                   ,rcsColor point_color
                                   ,eShape   point_shape
                                   )
    :sAppearance()
    ,sLineAppearance(point_color)
    ,sPointAppearance(point_color,point_size,point_shape)
    ,ShowLine(false)
    ,ShowPoints(true)
{
}


sCurveAppearance::sCurveAppearance (std::istream & a_stream ,rcsVersion  version)
    :sStorable()
    ,sAppearance()
    ,sLineAppearance()
    ,sPointAppearance()
{
  if( version < sVersion(1,1) ){
    byte                        U,R,G,B;
    a_stream.READ_INTO(U);
    a_stream.READ_INTO(R);
    a_stream.READ_INTO(G);
    a_stream.READ_INTO(B);
    LineColor = sColor(R,G,B);
    for( int i = 0 ; i < 17 ; i++ )  a_stream.READ_INTO(U);
    Hidden = false;
    LineThickness=DefaultLineThickness;
    LineStyle=DefaultLineStyle;
    FillColor=LineColor;
    ContourColor=LineColor;
    Filled=true;
    Contoured=false;
    ContourThickness=LineThickness;
    PointSize=DefaultSize;
    PointShape=DefaultShape;
    ShowLine=true;
    ShowPoints=false;
  }else{
    {
      sLineAppearance LA(a_stream,version);
      Hidden = LA.Hidden;
      LineColor = LA.LineColor;
      LineThickness = LA.LineThickness;
      LineStyle = LA.LineStyle;
    }
    { sColor C(a_stream,version); FillColor=C; }
    { sColor C(a_stream,version); ContourColor=C; }
    a_stream.READ_INTO(Filled);
    a_stream.READ_INTO(Contoured);
    a_stream.READ_INTO(ContourThickness);
    a_stream.READ_INTO(PointSize);
    a_stream.READ_INTO(PointShape);
    a_stream.READ_INTO(ShowLine);
    a_stream.READ_INTO(ShowPoints);
  }
}


void  sCurveAppearance::Store (std::ostream & a_stream) const
{
  sLineAppearance::Store(a_stream);
  /*! @todo{bomb} Be careful with storing/loading classes with virtual bases:
                  this should be unified, say, as a call to
  sPointAppearance::Store(a_stream,sStorable::cOmitVirtualBases); */
  {
    FillColor.Store(a_stream);
    ContourColor.Store(a_stream);
    a_stream.WRITE_FROM(Filled);
    a_stream.WRITE_FROM(Contoured);
    a_stream.WRITE_FROM(ContourThickness);
    a_stream.WRITE_FROM(PointSize);
    a_stream.WRITE_FROM(PointShape);
  }
  a_stream.WRITE_FROM(ShowLine);
  a_stream.WRITE_FROM(ShowPoints);
}


//psAppearance  sCurveAppearance::AssignFrom (pcsAppearance  other)
//{
//  sLineAppearance::AssignFrom(other);
//  sPointAppearance::AssignFrom(other);
//  if( pcsCurveAppearance A = dynamic_cast<pcsCurveAppearance>(other) ){
//    ShowLine = A->ShowLine;
//    ShowPoints = A->ShowPoints;
//  }
//  return this;
//}

//------------------------------------------------------------------- sCurve ---

sCurve::sCurve (std::istream & a_stream ,rcsVersion  version)
    :sGraphObject(a_stream,version)
{
}


void  sCurve::Store (std::ostream & a_stream) const
{
  sGraphObject::Store(a_stream);
}


int  sCurve::Distance (rcsIntPoint  position
                      ,rcsScales    scales  )
{


  sCurve::psTracer            T = GetCurveTracer(scales);
  int                         D_min = std::numeric_limits<int>::max();

                                  //  P -- точка с координатами position
  sPoint                      R;  //вектор, идущий из точки кривой в точку P
  sPoint                      L;       //вектор отрезка кривой
  real                        LengthL; //его длина
  sPoint                      PP;  //вектор, идущий из начала отрезка кривой
                                   //в точку P
  sPoint                      PPP; //координаты точки P в системе координат,
                                   //повёрнутой так, что ось х совмещена с
                                   //с вектором L
  real                        DD_min ,DD;
  DD_min= real(std::numeric_limits<int>::max())
        * real(std::numeric_limits<int>::max());
  while( T->NextPoint() ){
    R = position - scales.T(T->Point);
    DD=R.X*R.X+R.Y*R.Y;
    if( DD < DD_min ){
      DD_min=DD;
    }
  }
  while( T->NextLineSegment() ){
    PP = position - scales.T(T->LineSegment.Start);
    L = scales.T(T->LineSegment.End) - scales.T(T->LineSegment.Start);
    LengthL = sqrt(L.X*L.X+L.Y*L.Y);
    PPP.X = (L.X/LengthL)*PP.X + (L.Y/LengthL)*PP.Y;
    PPP.Y = (L.Y/LengthL)*PP.X - (L.X/LengthL)*PP.Y;
    if( PPP.X < 0 ){
      DD = PPP.X*PPP.X + PPP.Y*PPP.Y;
    }else if( PPP.X > LengthL ){
      DD = (PPP.X-LengthL)*(PPP.X-LengthL) + PPP.Y*PPP.Y;
    }else{
      DD = PPP.Y*PPP.Y;
    }
    if( DD < DD_min ){
      DD_min=DD;
    }
  }
  try{
    D_min = std::min( D_min , Round(sqrt(DD_min)) );
  }catch(...){
  }

  if( T ){ delete T;  T=NULL; }
  return D_min;
}

//------------------------------------------------------------------------------
} //namespace RA




