////////////////////////////////////////////////////////////////////////////////
/*! @file DB_Link.cpp Classes sDB_Info, sDB_UUID and sDB_Link.
- Part of RANet - Research Assistant Net Library (based on ANSI C++).
- Uses  Qt v.5.0  -  http://qt-project.org/
- Copyright(C) 2013-2020, Viktor E. Bursian, St.Petersburg, Russia.
                          VBursian@gmail.com
*///////////////////////////////////////////////////////////////////////////////
#include "DB_Link.h"
#include "Environment.h"
#include "StringsAndQStrings.h"
#include <QUrl>
#include <QDir>
//#include "Log.h"
namespace RA {
//------------------------------------------------------------------------------

sDB_Info::fModifyInfo *     sDB_Info::ModifyInfo = NULL;
const sString               sDB_Info::TypeName []     ={ "SQLite"
                                                       , "MySQL"
                                                       , "PostgreSQL"
                                                       };
const QString               sDB_Info::QtDriverName [] ={ "QSQLITE"
                                                       , "QMYSQL"
                                                       , "QPSQL"
                                                       };

const sString               sDB_Link::MyselfNickname( "__this_database__" );

//----------------------------------------------------------------- sDB_Info ---
/*! @class sDB_Info

@par RANet URL syntax

@include "RANetURI_syntax.txt"

*/

sDB_Info::sDB_Info (rcsString  partition_URL)
    :Type(cSQLite)
    ,Port(-1)
{
  SetByURL(partition_URL);
}


/*! Creates in accordance with the command line option --root
    (which is a bit weird).
*/
sDB_Info::sDB_Info ()
    :Type(cSQLite)
    ,Port(-1)
{
  SetByURL(sEnvironment::StringOption("--root"
                                     ,sEnvironment::StringOption("-r"
                                                                ,"root") ) );
}


/*! Creates with default settings for the given type.
*/
sDB_Info::sDB_Info (eType  type)
    :Type(type)
    ,SchemaName()
    ,DBName()
    ,Host()
    ,Port(-1)
    ,Login()
    ,Password()
    ,ConnectOptions()
{
  SetDefaults();
}


void sDB_Info::SetByURL (rcsString  partition_URL)
{
  QUrl                        URL (ToQString(partition_URL));
  Type = cSQLite;
  TableNamePrefix = "";
  SchemaName = "";
  DBName = "";
  Host = "";
  Port = -1;
  Login = "";
  Password = "";
  ConnectOptions = "";
  if( URL.scheme().toLower() == "sqlite" ){
    Type = cSQLite;
  }else if( URL.scheme().toLower() == "mysql" ){
    Type = cMySQL;
  }else if( URL.scheme().mid(0,7).toLower() == "postgre" ){
    Type = cPostgreSQL;
  }
  SetDefaults();
  if( ! URL.host().isEmpty() ){
    Host = FromQString(URL.host());
  }
  //!@todo{SQL} Временно port=0 считается незаданным.
  //if( URL.port() >= 0 ){
  if( URL.port() > 0 ){
    Port = URL.port();
  }
  if( ! URL.path().isEmpty() ){
    DBName = FromQString(URL.path());
  }
  switch( Type ){
    case cSQLite:
      if( ! Host.Empty() ){
        DBName = sString("//") + Host + DBName;
        Host = "";
      }
      break;
    default:
      if( ! DBName.Empty() ){
        if( DBName[0] == '/' ){
          DBName = DBName.Tail(1);
        }
      }
      break;
  }
  if( URL.hasQuery() ){
    sString                   QueryPart(FromQString(URL.query()));
    SchemaName = QueryPart.Substr(0,QueryPart.Pos(";"));
    QueryPart = QueryPart.Tail(QueryPart.Pos(";")+1);
    TableNamePrefix = QueryPart.Substr(0,QueryPart.Pos(";"));
    ConnectOptions = QueryPart.Tail(QueryPart.Pos(";")+1);
  }
  if( ! URL.userName().isEmpty() )
    Login = FromQString(URL.userName());
  if( ! URL.password().isEmpty() )
    Password = FromQString(URL.password());
//  DEBUG_LOG( Environment , sString("Type=")<<Type );
//  DEBUG_LOG( Environment , sString("Login:")+Login );
//  DEBUG_LOG( Environment , sString("Password:")+Password );
//  DEBUG_LOG( Environment , sString("Host:")+Host );
//  DEBUG_LOG( Environment , sString("Port=")<<Port );
//  DEBUG_LOG( Environment , sString("DBName:")+DBName );
//  DEBUG_LOG( Environment , sString("SchemaName:")+SchemaName );
}


void  sDB_Info::SetDefaults ()
{
  switch( Type ){
    case cSQLite:
        DBName = "root.RANet.sqlite";
    break;
    case cMySQL:
        Host = "127.0.0.1";
        Port = 3306;
    break;
    case cPostgreSQL:
        Host = "";
        Port = 5432;
    break;
  }
}


sDB_Info::sDB_Info (rcsDB_Info  other)
    :Type(other.Type)
    ,TableNamePrefix(other.TableNamePrefix)
    ,SchemaName(other.SchemaName)
    ,DBName(other.DBName)
    ,Host(other.Host)
    ,Port(other.Port)
    ,Login(other.Login)
    ,Password(other.Password)
    ,ConnectOptions(other.ConnectOptions)
{
}


rsDB_Info  sDB_Info::operator = (rcsDB_Info  other)
{
  Type = other.Type;
  TableNamePrefix = other.TableNamePrefix;
  SchemaName = other.SchemaName;
  DBName = other.DBName;
  Host = other.Host;
  Port = other.Port;
  Login = other.Login;
  Password = other.Password;
  ConnectOptions = other.ConnectOptions;
  return *this;
}


sString  sDB_Info::URL (bool  with_password) const
{
  sString                     R(sDB_Info::TypeName[Type]+":");
  if( Type == sDB_Info::cSQLite ){
    R += DBName;
  }else{
    R += "//";
    if( ! Login.Empty() ){
      R += Login;
      if( ! Password.Empty() ){
        R += ":";
      }
      if( with_password ){
        R += Password;
      }else if( Password == "___" ){
        R += "___";
      }else if( ! Password.Empty() ){
        R += "***";
      }
      R += "@";
    }
    R += Host;
    if( Port >= 0 ){
      R += ":";
      R << Port;
    }
    if( ! DBName.Empty() ){
      R += "/";
      R += DBName;
    }
  }
  sString                     QueryPart(ConnectOptions);
  if( ! QueryPart.Empty() )
    QueryPart = sString(";") + QueryPart;
  QueryPart = TableNamePrefix + QueryPart;
  if( ! QueryPart.Empty() )
    QueryPart = sString(";") + QueryPart;
  QueryPart = SchemaName + QueryPart;
  if( ! QueryPart.Empty() ){
    R += "?";
    R += QueryPart;
  }
  return R;
}


sString  sDB_Info::TableFullName (literal  table_name) const
{
  sString                   FN;
  switch( Type ){
    case cMySQL:
      FN = sString("`")+SchemaName+"`.`"+TableNamePrefix+table_name+"`";
      break;
    default:
      if( ! SchemaName.Empty() )
        FN = sString("""")+SchemaName+""".";
      FN += sString("""")+TableNamePrefix+table_name+"""";
      break;
  }
  return FN;
}


sString  sDB_Info::ColumnNotation (literal col_name) const
{
  sString                   CN;
  switch( Type ){
    case cMySQL:
      CN = sString("`")+col_name+"`";
      break;
    default:
      CN = sString("""")+col_name+"""";
      break;
  }
  return CN;
}

//----------------------------------------------------------------- sDB_UUID ---

sDB_UUID  sDB_UUID::NewUUID ()
{
  return sDB_UUID( QUuid::createUuid() );
}


rsDB_UUID  sDB_UUID::operator = (rcsDB_UUID  other)
{
  QUuid::operator =(other);
  return *this;
}


sDB_UUID::sDB_UUID (rcsString  string)
    :QUuid( QByteArray(literal(string)) )
{
}


/*! Visial representation in common 38-length textual form, like
"{550e8400-e29b-41d4-a716-446655440000}".
*/
sString  sDB_UUID::ToString () const
{
  return sString(toByteArray().data());
}

//----------------------------------------------------------------- sDB_Link ---

sDB_Link::sDB_Link (rcsDB_Info  info)
    :sDB_Info(info)
    ,Id()
{
}


sDB_Link::sDB_Link (rcsDB_Link  other)
    :sDB_Info(other)
    ,Id(other.Id)
    ,Version(other.Version)
{
}


rsDB_Link  sDB_Link::operator = (rcsDB_Link  other)
{
  sDB_Info::operator =(other);
  Id = other.Id;
  Version = other.Version;
  return *this;
}

//------------------------------------------------------------------------------
} //namespace RA


