////////////////////////////////////////////////////////////////////////////////
/*! @file DB_SQL_Net.cpp RANet back-end, based on storage in a relational DB
                         accessed via SQL.
- Part of RANet - Research Assistant Net Library.
- Uses  Qt v.5  - http://qt.io/
- Copyright(C) 2013-2020, Viktor E. Bursian, St.Petersburg, Russia.
                          VBursian AT gmail DOT com
*///////////////////////////////////////////////////////////////////////////////
#include "DB_SQL_Net.h"
#include "Nodes.h"
#include "AttrName.h"
#include "FolderNode.h"
#include "StringsAndQStrings.h"
#include "NetExceptions.h"
#include <QUuid>
#include <QMutexLocker>
#include <QTimer>
#include <sstream>
#include <string>

//!@todo{NetIO} Убрать отсюда все сообщения
#include <QMessageBox>

#include "Log.h"
namespace RA {
//------------------------------------------------------------------------------

csString  sDB_SQL_Net::DBIdFieldDecl  ( "VARCHAR(38)" );
csString  sDB_SQL_Net::NodeIdFieldDecl( "VARCHAR(38)" );

namespace DB_SQL_Net {
const int sSynchronizer::StartDelay  = 10000;//ms
const int sSynchronizer::OnIdlePause = 10000;//ms
}

//-------------------------------------------------------------- sDB_SQL_Net ---

    //........................................... sDB_SQL_Net::sNodeUUID ...

/*! @class sDB_SQL_Net::sNodeUUID


*/


sDB_SQL_Net::sNodeUUID  sDB_SQL_Net::sNodeUUID::ForTheRoot
                                                     (psNet       net
                                                     ,sDB_UUID    database_UUID
                                                     ,rcsVersion  version)
{
  return ( version < sVersion(1,5)
         ? sNodeUUID(net,database_UUID,QUuid())
         : sNodeUUID(net,database_UUID,database_UUID) );
}

sDB_SQL_Net::sNodeUUID::sNodeUUID (rcsString  formatted_db_id
                                  ,rcsString  formatted_node_id
                                  ,psNet      net              )
    :sNodeLocation(net)
    ,TheDBId(formatted_db_id)
    ,TheNodeId(QUuid(QByteArray(literal(formatted_node_id))))
{
}


sString  sDB_SQL_Net::sNodeUUID::FormattedDBId () const
{
  return  TheDBId.ToString();
}


sString  sDB_SQL_Net::sNodeUUID::FormattedNodeId () const
{
  return sString( TheNodeId.toByteArray().data() );
}


bool  sDB_SQL_Net::sNodeUUID::operator == (rcsNodeUUID  other) const
{
  return (TheNodeId == other.TheNodeId) && (TheDBId == other.TheDBId);
}


bool  sDB_SQL_Net::sNodeUUID::operator < (rcsNodeUUID  other) const
{
  return (TheNodeId < other.TheNodeId)
      || ((TheNodeId == other.TheNodeId) && (TheDBId < other.TheDBId));
}


bool  sDB_SQL_Net::sNodeUUID::operator == (pcsNodeLocation  other) const
{
  pcsNodeUUID                 TheOther (dynamic_cast<pcsNodeUUID>(other));
  ASSERT(TheOther);
  return (*this == *TheOther);
}


psNodeLocation  sDB_SQL_Net::sNodeUUID::Replica () const
{
  return new sNodeUUID(*this);
}


psNodeLocation  sDB_SQL_Net::sNodeUUID::LocationForAttr (rcsAttrName) const
{
  return new sNodeUUID(Net
                      ,TheDBId
                      ,QUuid::createUuid().toString().toLatin1()
                      );

}

    //......................................................................

sDB_SQL_Net::~sDB_SQL_Net()
{
  Close();
}


sDB_SQL_Net::sDB_SQL_Net ()
    :RootDB_Info()
    ,TheRoot()
    ,SynchronizationThread(NULL)
{
}


bool sDB_SQL_Net::SaveAll()
{
  RANet::Log.OpenBlock(sLog::Debug,"Saving","sDB_SQL_Net::SaveAll");
  bool                        AllDone = false;
  int                         PassNo = 0;
  while( ! AllDone ){
    AllDone = true;
    PassNo++;
    RANet::Log.OpenBlock(sLog::Debug,"Saving",sString("PassNo=")<<PassNo);
    for( std::map<sNodeUUID,psNode>::const_iterator  P = LoadedNodes.begin()
                                                   ; P != LoadedNodes.end()
                                                   ; P++ ){
      psNode     N = P->second;
      //TRACING
      {
        #ifdef  RANET_DEBUG
          sNodeUUID  UID = P->first;
          sString    LateSaving("LateSaving");
          LateSaving += UID.FormattedNodeId();
          LateSaving += N->TrueType()->Name();
          if( N->BodyHasToBeSaved() || N->AttrsHaveToBeSaved() )
            DEBUG_LOG(Saving,LateSaving);
        #endif
      }
      if( N->BodyHasToBeSaved() ){
        Save(N);
        DEBUG_LOG(Saving,sString("body - ")
                      +(N->BodyHasToBeSaved() ? "FAILURE" : "done"));
      }
      if( N->AttrsHaveToBeSaved() ){
        SaveAttributes(N);
        DEBUG_LOG(Saving,sString("attributes - ")
                      +(N->AttrsHaveToBeSaved() ? "FAILURE" : "done"));
        AllDone = false;
        break;  //new nodes could be allocated during SaveAttributes(N);
                //their allocation extends the LoadedNodes map;
                //another pass is necessary
      }
      AllDone = AllDone && ! N->BodyHasToBeSaved() && ! N->AttrsHaveToBeSaved();
    }
    RANet::Log.CloseBlock(sLog::Debug,"Saving");
  }
  RANet::Log.CloseBlock(sLog::Debug,"Saving");
  return AllDone;
}


void  sDB_SQL_Net::Close ()
{
  RANet::Log.OpenBlock(sLog::Debug,"Saving","sDB_SQL_Net::Close");
  if( SynchronizationThread ){
    SynchronizationThread->quit();
    SynchronizationThread->wait();
    DEBUG_LOG(Saving,"SynchronizationThread stopped");
    delete SynchronizationThread;  SynchronizationThread=NULL;
  }
  TheRoot.SetToNULL();
  ASSERT(SaveAll());
  for( std::map<sDB_UUID,psDatabase>::iterator  D = Databases.begin()
                                              ; D != Databases.end() ; ++D ){
    delete D->second;  D->second=NULL;
  };
  Databases.clear();
  LoadedNodes.clear();
  RANet::Log.CloseBlock(sLog::Debug,"Saving");
}


bool  sDB_SQL_Net::Preset (sString  textual_URL)
{
  RootDB_Info = sDB_Info(textual_URL);
  return true;
}


bool  sDB_SQL_Net::CreateNew ()
{
  return CreateNewPartition(RootDB_Info);
}


bool  sDB_SQL_Net::PreOpenSetUp ()
{
  ASSERT( ! IsOpen() );
  if( sDB_Info::ModifyInfo )
    return (*sDB_Info::ModifyInfo)(&RootDB_Info);
  return true;
}


bool  sDB_SQL_Net::Open (sString  textual_URL)
{
  Preset(textual_URL);
  return Open();
}


bool  sDB_SQL_Net::Open ()
{
  if(  LoadDatabases()  /*&& CheckPartitionVersions()*/ && LoadRoot() ){
    DB_SQL_Net::psSynchronizer  Synchronizer
                                         = new DB_SQL_Net::sSynchronizer(this);
    SynchronizationThread = new QThread;
    Synchronizer->moveToThread(SynchronizationThread);
    //!@todo{?} Интересно заданы аргументы connect (взято из Help). К чему бы?
    QObject::connect(SynchronizationThread ,&QThread::finished
                    ,Synchronizer          ,&QObject::deleteLater);
    QObject::connect(SynchronizationThread ,&QThread::started
                    ,Synchronizer     ,&DB_SQL_Net::sSynchronizer::StartSaving);
    SynchronizationThread->start();
    return true;
  }else{
    TheRoot.SetToNULL();
    for( std::map<sDB_UUID,psDatabase>::iterator  D = Databases.begin()
                                                ; D != Databases.end() ; ++D ){
      delete D->second;  D->second=NULL;
    };
    Databases.clear();
    LoadedNodes.clear();
  }
  return false;
}


bool  sDB_SQL_Net::IsOpen ()
{
  return ! TheRoot.IsNULL();
}


sString  sDB_SQL_Net::DesktopInfoText ()
{
  return RootDB_Info.URL();
}


rsNodePtr  sDB_SQL_Net::Root ()
{
  return TheRoot;
}


void  sDB_SQL_Net::RegisterNode (pcsNodeLocation  location
                                ,psNode           node    )
{
  pcsNodeUUID                 uuid( dynamic_cast<pcsNodeUUID>(location) );
  ASSERT(uuid);
//  std::pair<tNodeIndex::iterator,bool>  P;
//  P=LoadedNodes.insert(std::pair<sNodeUUID,psNode>(*uuid,node));
//  ASSERT(P.second);

  //можно проще, поскольку здесь проверка не нужна:
  LoadedNodes[*uuid]=node;
}


psNode  sDB_SQL_Net::FindNode (pcsNodeLocation  location)
{
  pcsNodeUUID                 Location = dynamic_cast<pcsNodeUUID>(location);
  std::map<sNodeUUID,psNode>::const_iterator  P = LoadedNodes.find(*Location);
  return ( P != LoadedNodes.end()  ?  P->second  :  NULL );
}


sNodePtr  sDB_SQL_Net::GetNode (pcsNodeLocation  location)
{
  QMutexLocker                Locker(&LoadedNodesMutex);
  pcsNodeUUID                 Location = dynamic_cast<pcsNodeUUID>(location);
  ASSERT( Location );
  psNode                      Node = FindNode(Location);
  if( ! Node ){
    Node = LoadNode(Location);
    if( Node ){
      RegisterNode(Location,Node);
    }
  }
  return sNodePtr(Node);
}


psNode  sDB_SQL_Net::LoadNode (pcsNodeLocation  location)
{
  pcsNodeUUID                 Location = dynamic_cast<pcsNodeUUID>(location);
  ASSERT( Location );
  psNode                      Node;
  psDatabase                  DB(Databases[Location->DBId()]);
  try{
    sSQLQuery                   SQL(DB->Connection());
    sString                     NodeClass;
    QByteArray                  NodeBody;
    if( DB->Version < sVersion(1,5) ){
      SQL.prepare(sString("SELECT class, body FROM ")
                 +DB->TableFullName("nodes")
                 +" WHERE id = ?");
    }else{
      SQL.prepare(sString("SELECT node_class, node_body FROM ")
                 +DB->TableFullName("nodes")
                 +" WHERE node_id = ?");
    }
    SQL.addBindString( Location->FormattedNodeId() );
    SQL.exec();
    if( ! SQL.first() )
      throw xNetIO(sString("cannot find node ")+Location->FormattedNodeId());
    NodeClass = SQL.StringValue(0);
    NodeBody = SQL.value(1).toByteArray();
    if( SQL.next() )
      throw xNetIO(sString("ambiguos node id=")/*+Location->FormattedNodeId()*/);
    std::stringstream           Stream(std::string(NodeBody.data()
                                                  ,NodeBody.size()+1   )
                                      ,std::ios_base::in);
    Node = dynamic_cast<psNode>
                       ( sStorable::PolymorphicLoad(NodeClass,Stream) );
    SetLocation(Node,Location);
  }catch( rxSQL E ){
    throw xNetIO(sString("SQL error in sDB_SQL_Net::LoadNode:\n")
                +"\nDB->URL()="+DB->URL()
                +"\nDB->Version="+DB->Version.Text()
                +"\nLocation->FormattedNodeId()="+Location->FormattedNodeId()
                +"\n\n"+E.Error);
  }
  return Node;
}


bool  sDB_SQL_Net::GetNodeStorageTime (pcsNodeLocation  location
                                      ,rsTime           when     )
{
  pcsNodeUUID                 Location = dynamic_cast<pcsNodeUUID>(location);
  ASSERT( Location );
  try{
    psDatabase                  DB(Databases[Location->DBId()]);
    sSQLQuery                   SQL(DB->Connection());
    if( DB->Version < sVersion(1,5) ){
      SQL.prepare(sString("SELECT time FROM ")
                 +DB->TableFullName("nodes")
                 +" WHERE id = ?");
    }else{
      SQL.prepare(sString("SELECT node_save_time FROM ")
                 +DB->TableFullName("nodes")
                 +" WHERE node_id = ?");
    }
    SQL.addBindString( Location->FormattedNodeId() );
    SQL.exec();
    if( ! SQL.first() )
      throw xSQL(sString("cannot find node ")/*+Location->FormattedNodeId()*/);
    ///@todo{SQL} more effective time storage for Postgres, MySQL
    when = sTime::FromISO_8601(SQL.StringValue(0));
    if( SQL.next() )
      throw xSQL(sString("ambiguos node id=")/*+Location->FormattedNodeId()*/);
    return true;
  }catch( rxSQL E ){
    throw xNetIO(sString("SQL error in sDB_SQL_Net::GetNodeStorageTime\n\n")
                +E.Error);
  }
  return false;
}


void  sDB_SQL_Net::GetAttrList (std::list<sAttribute> &  attributes
                               ,pcsNodeLocation          location)
{
  pcsNodeUUID                 Location = dynamic_cast<pcsNodeUUID>(location);
  ASSERT( Location );
  psDatabase                  DB(Databases[Location->DBId()]);
  try{
    sSQLQuery                   SQL(DB->Connection());
    quint8                      Type;
    sString                     ValueNodeId;
    sString                     DBId;

    ///@todo{NetIO} Время сохранения атрибута пишется, но не используется
    if( DB->Version < sVersion(1,5) ){
      SQL.prepare(sString("SELECT type,value,value_db,")
                 +DB->ColumnNotation("numeric")
                 +",symbolic FROM "
                 +DB->TableFullName("attributes")
                 +" WHERE host = ?");
    }else{
      SQL.prepare(sString("SELECT attr_type,value_node,value_db,")
                 +"numeric_name,symbolic_name FROM "
                 +DB->TableFullName("attrs")
                 +" WHERE host_node = ?");
    }
    SQL.addBindString( Location->FormattedNodeId() );
    /*! @bug Happens at first start: looks like the '?' char is not replaced
             by the node id.  Probably because it is not located yet.
             See crash.txt.
             Вероятно, связано с ошибкой в
             RA::sFolderNode::AttrsHaveToBeRefreshed() и/или в
             RA::sFolderNode::AttrsHaveToBeSaved()
             для новорожденных и не размещённых
    */
    SQL.exec();
    while( SQL.next() ){
      Type = SQL.value(0).toUInt();
      ValueNodeId = SQL.StringValue(1);
      if( SQL.value(2).isNull() ){
        DBId = Location->FormattedDBId(); //the same DB
      }else{
        DBId = SQL.StringValue(2);
      }
      switch( Type ){
        case sAttrName::cNameless:
              attributes.push_back(
                    sAttribute(sAttrName()
                              ,new sNodeUUID(DBId,ValueNodeId,this)
                              ,sAttribute::cSynchronized));
              break;
        case sAttrName::cNumeric:
              attributes.push_back(
                    sAttribute(sAttrName(SQL.value(3).toInt())
                              ,new sNodeUUID(DBId,ValueNodeId,this)
                              ,sAttribute::cSynchronized));
              break;
        case sAttrName::cSymbolic:
              attributes.push_back(
                    sAttribute(sAttrName(SQL.StringValue(4))
                              ,new sNodeUUID(DBId,ValueNodeId,this)
                              ,sAttribute::cSynchronized));
              break;
      }
    }
  }catch( rxSQL E ){
    throw xNetIO(sString("SQL error in sDB_SQL_Net::GetAttrList:\n")
                +"\nDB->URL()="+DB->URL()
                +"\nDB->Version="+DB->Version.Text()
                +"\nLocation->FormattedNodeId()="+Location->FormattedNodeId()
                +"\n\n"+E.Error);
  }
}


bool  sDB_SQL_Net::StoreNewNode (psNode  node
                                ,rsTime  when)
{
  ASSERT( node );
  pcsNodeUUID                 Location = dynamic_cast<pcsNodeUUID>
                                                     (GetLocation(node));
  ASSERT( Location );
  try{
    std::stringstream           Stream (std::ios_base::out);

    node->Store(Stream);
    QByteArray                  NodeBody (Stream.str().data()
                                         ,Stream.str().size());
    psDatabase                  DB (Databases[Location->DBId()]);
    sSQLQuery                   SQL (DB->Connection());
    if( DB->Version < sVersion(1,5) ){
      SQL.prepare(sString("INSERT INTO ")+DB->TableFullName("nodes")
                 +" (id,class,body,time) VALUES(?,?,?,?)");
    }else{
      SQL.prepare(sString("INSERT INTO ")+DB->TableFullName("nodes")
                 +" (node_id,node_class,node_body,node_save_time)"
                  " VALUES(?,?,?,?)");
    }
    SQL.addBindString( Location->FormattedNodeId() );
    SQL.addBindString( node->VersionAndClass() );
    SQL.addBindValue ( NodeBody , QSql::In | QSql::Binary  );
    SQL.addBindString( (when = sTime()).ToISO_8601() );
    SQL.exec();
  }catch( rxSQL E ){
    throw xNetIO(sString("SQL error in sDB_SQL_Net::StoreNewNode\n\n")+E.Error);
  }
  return true;
}


bool  sDB_SQL_Net::UpdateNode (psNode  node
                              ,rsTime  when)
{
  ASSERT( node );
  pcsNodeUUID                 Location = dynamic_cast<pcsNodeUUID>
                                                     (GetLocation(node));
  ASSERT( Location );
  try{
    std::stringstream           Stream (std::ios_base::out);

    node->Store(Stream);
    QByteArray                  NodeBody (Stream.str().data()
                                         ,Stream.str().size());
    psDatabase                  DB (Databases[Location->DBId()]);
    sSQLQuery                   SQL (DB->Connection());
    if( DB->Version < sVersion(1,5) ){
      SQL.prepare(sString("UPDATE ")+DB->TableFullName("nodes")
                 +" SET class=?, body=?, time=? WHERE id=?");
    }else{
      SQL.prepare(sString("UPDATE ")+DB->TableFullName("nodes")
                 +" SET node_class=?, node_body=?, node_save_time=?"
                  " WHERE node_id=?");
    }
    SQL.addBindString( node->VersionAndClass() );
    SQL.addBindValue( NodeBody , QSql::In | QSql::Binary );
    SQL.addBindString( (when = sTime()).ToISO_8601() );
    SQL.addBindString( Location->FormattedNodeId() );
    SQL.exec();
    if( SQL.numRowsAffected() != 1 )
      throw xNetIO(sString("Multiple rows affected in sDB_SQL_Net::UpdateNode")
                  +sString("  rows=") << SQL.numRowsAffected());
  }catch( rxSQL E ){
    throw xNetIO(sString("SQL error in sDB_SQL_Net::UpdateNode\n\n")+E.Error);
  }
  return true;
}


bool  sDB_SQL_Net::StoreNewAttribute (pcsNodeLocation   host
                                     ,rcsAttrName       name
                                     ,pcsNodeLocation   value
                                     ,rsTime            when)
{
  pcsNodeUUID                 HostLocation = dynamic_cast<pcsNodeUUID>(host);
  pcsNodeUUID                 ValueLocation = dynamic_cast<pcsNodeUUID>(value);
  sString                     Fields;
  sString                     Placeholders;
  ASSERT( HostLocation );
  ASSERT( ValueLocation );
  ///@todo{NetIO} Время сохранения атрибута пишется, но не используется
  try{
    psDatabase                  DB (Databases[HostLocation->DBId()]);
    sSQLQuery                   SQL (DB->Connection());
    if( DB->Version < sVersion(1,5) ){
      Fields = "time,host,type,value";
      Placeholders = "?,?,?,?";
    }else{
      Fields = "birth_time,host_node,ordering_index,attr_type,value_node";
      Placeholders = "?,?,?,?,?";
    }
    if( ValueLocation->FormattedDBId() != HostLocation->FormattedDBId() ){
      Fields += ",value_db";
      Placeholders += ",?";
    }
    switch( name.Type() ){
      case sAttrName::cNameless:
            break;
      case sAttrName::cNumeric:
            if( DB->Version < sVersion(1,5) ){
              Fields += ",";
              Fields += DB->ColumnNotation("numeric");
            }else{
              Fields += ",numeric_name";
            }
            Placeholders += ",?";
            break;
      case sAttrName::cSymbolic:
            if( DB->Version < sVersion(1,5) ){
              Fields += ",symbolic";
            }else{
              Fields += ",symbolic_name";
            }
            Placeholders += ",?";
            break;
    }
    if( DB->Version < sVersion(1,5) ){
      SQL.prepare(sString("INSERT INTO ")+DB->TableFullName("attributes")
                 +" (" + Fields + ") VALUES (" + Placeholders +")" );
    }else{
      SQL.prepare(sString("INSERT INTO ")+DB->TableFullName("attrs")
                 +" (" + Fields + ") VALUES (" + Placeholders +")" );
    }
    SQL.addBindString( (when = sTime()).ToISO_8601() );
    SQL.addBindString( HostLocation->FormattedNodeId() );
    if( DB->Version < sVersion(1,5) ){
    }else{
      SQL.addBindValue ( qint32(0) );
    }
    SQL.addBindValue ( quint8(name.Type()) );
    SQL.addBindString( ValueLocation->FormattedNodeId() );
    if( ValueLocation->DBId() != HostLocation->DBId() ){
      SQL.addBindString( ValueLocation->FormattedDBId() );
    }
    switch( name.Type() ){
      case sAttrName::cNameless:
            break;
      case sAttrName::cNumeric:
            SQL.addBindValue( name.ToNumber() );
            break;
      case sAttrName::cSymbolic:
            SQL.addBindString( name.ToString() );
            break;
    }
    SQL.exec();
  }catch( rxSQL E ){
    throw xNetIO(sString("SQL error in sDB_SQL_Net::StoreNewAttribute\n\n")
                +E.Error);
  }
  return true;
}


bool  sDB_SQL_Net::RemoveAttribute (pcsNodeLocation host, rcsAttrName name)
{
  pcsNodeUUID                 HostLocation = dynamic_cast<pcsNodeUUID>(host);
  ASSERT( HostLocation );
  try{
    psDatabase                  DB (Databases[HostLocation->DBId()]);
    sSQLQuery                   SQL (DB->Connection());
    switch( name.Type() ){
      case sAttrName::cNumeric:
            if( DB->Version < sVersion(1,5) ){
              SQL.prepare(sString("DELETE FROM ")
                         +DB->TableFullName("attributes")
                         +" WHERE (host = ?) AND (type = ?) AND ("
                         +DB->ColumnNotation("numeric")+" = ?)");
            }else{
              SQL.prepare(sString("DELETE FROM ")+DB->TableFullName("attrs")
                         +" WHERE (host_node = ?) AND (attr_type = ?)"
                          " AND (numeric_name = ?)");
            }
            SQL.addBindString( HostLocation->FormattedNodeId() );
            SQL.addBindValue ( quint8(name.Type()) );
            SQL.addBindValue ( name.ToNumber() );
            break;
      case sAttrName::cSymbolic:
            if( DB->Version < sVersion(1,5) ){
              SQL.prepare(sString("DELETE FROM ")
                         +DB->TableFullName("attributes")
                         +" WHERE (host = ?) AND (type = ?)"
                          " AND (symbolic = ?)");
            }else{
              SQL.prepare(sString("DELETE FROM ")+DB->TableFullName("attrs")
                         +" WHERE (host_node = ?) AND (attr_type = ?)"
                          " AND (symbolic_name = ?)");
            }
            SQL.addBindString( HostLocation->FormattedNodeId() );
            SQL.addBindValue ( quint8(name.Type()) );
            SQL.addBindString( name.Text() );
            break;
      default:
            throw xProgramError("sDB_SQL_Net::RemoveAttribute"
                               ,__FILE__,__LINE__);
            break;
    }
    SQL.exec();
    if( SQL.numRowsAffected() != 1 )
      throw xNetIO(sString("Multiple rows affected in sDB_SQL_Net::")
                  +sString("RemoveAttribute  rows=") << SQL.numRowsAffected());
  }catch( rxSQL E ){
    throw xNetIO(sString("SQL error in sDB_SQL_Net::RemoveAttribute\n\n")
                +E.Error);
  }
  return true;
}


bool  sDB_SQL_Net::RemoveAttribute (pcsNodeLocation host, pcsNodeLocation value)
{
  pcsNodeUUID                 HostLocation = dynamic_cast<pcsNodeUUID>(host);
  pcsNodeUUID                 ValueLocation = dynamic_cast<pcsNodeUUID>(value);
  ASSERT( HostLocation );
  ASSERT( ValueLocation );
  try{
    psDatabase                  DB (Databases[HostLocation->DBId()]);
    sSQLQuery                   SQL (DB->Connection());
    if( DB->Version < sVersion(1,5) ){
      SQL.prepare(sString("DELETE FROM ")+DB->TableFullName("attributes")
                 +" WHERE (host = ?) AND (type = ?) AND (value = ?)"
                 );
    }else{
      SQL.prepare(sString("DELETE FROM ")+DB->TableFullName("attrs")
                 +" WHERE (host_node = ?) AND (attr_type = ?)"
                  " AND (value_node = ?)"
                 );
    }
    SQL.addBindString( HostLocation->FormattedNodeId() );
    SQL.addBindValue ( quint8(sAttrName::cNameless) );
    SQL.addBindString( ValueLocation->FormattedNodeId() );
    SQL.exec();
    if( SQL.numRowsAffected() != 1 )
      throw xNetIO(sString("Multiple rows affected in sDB_SQL_Net::")
                  +sString("RemoveAttribute  rows=") << SQL.numRowsAffected());
  }catch( rxSQL E ){
    throw xNetIO(sString("SQL error in sDB_SQL_Net::RemoveAttribute\n\n")
                +E.Error);
  }
  return true;
}


bool  sDB_SQL_Net::LoadRoot ()
{
  DEBUG_LOG(NetOpen,sString("LoadRoot version ")+RootDB_Version.Text());
  TheRoot = GetNode( sNodeUUID::ForTheRoot(this,RootDB_UUID,RootDB_Version)
                     .Replica() );
  return true;
}


bool  sDB_SQL_Net::LoadDatabases (bool  recursive)
{
  DEBUG_LOG(NetOpen,sString("LoadDatabases(")<<recursive);
  bool                        OK = false;
  bool                        HasVersionField = false;
  psDatabase                  RootDB = new sDatabase(RootDB_Info);
  try{
    sSQLQuery                   SQL(RootDB->Connection());
    try{
      SQL.exec(sString("SELECT parval FROM ")
              +RootDB->TableFullName("params")
              +" WHERE parkey = 'part_id'");
      if( ! SQL.first() )
        throw xNetOpen("cannot find part_id in 'params' of the root partition");
      RootDB_UUID = sDB_UUID( SQL.StringValue(0) );
      SQL.exec(sString("SELECT parval FROM ")
              +RootDB->TableFullName("params")
              +" WHERE parkey = 'version'");
      if( ! SQL.first() )
        throw xNetOpen("cannot find version in 'params' of the root partition");
      RootDB_Version = sVersion( SQL.StringValue(0) );
    }catch( rxSQL ){ ///@todo{1_6} Убрать чтение старого формата таблиц.
      try{
        SQL.exec(sString("SELECT id, version FROM ")
                +RootDB->TableFullName("databases")
                +" WHERE nickname = '"+sDatabase::MyselfNickname+"'");
        HasVersionField = true;
      }catch( rxSQL ){
        SQL.exec(sString("SELECT id FROM ")
                +RootDB->TableFullName("databases")
                +" WHERE nickname = '"+sDatabase::MyselfNickname+"'");
      }
      if( ! SQL.first() )
        throw xNetOpen(sString("cannot find '")+sDatabase::MyselfNickname
                      +"' in databases of the root partition");
      RootDB_UUID = sDB_UUID( SQL.StringValue(0) );
      if( HasVersionField )
        RootDB_Version = sVersion( SQL.StringValue(1) );
      if( SQL.next() )
        throw xNetOpen(sString("ambiguos '")+sDatabase::MyselfNickname
                      +"' in databases of the root partition");
    }
    RootDB->Id = RootDB_UUID;
    RootDB->Version = RootDB_Version;
    Databases[RootDB->Id] = RootDB;
    OK = LoadDatabases(RootDB,recursive);
  }catch( rxSQL_OpenUnexistingSqlite E ){
    QMessageBox::critical(NULL,"Error opening root partition"
        ,QString("\n\n%1")
            .arg(ToQString(E.Error))
        ,QMessageBox::Close);
  }catch( rxSQL_Open E ){
    QMessageBox::critical(NULL,"Error opening root partition"
        ,QString("\n\n%1")
            .arg(ToQString(E.Error))
        ,QMessageBox::Close);
  }catch( rxSQL ){
    if( RootDB ){ delete RootDB; RootDB=NULL; }
  }catch( rxNetIO ){
    //!@todo{NetIO} странная логика; а если её уже вставили в Databases?
    if( RootDB ){ delete RootDB; RootDB=NULL; }
  }
  return OK;
}


bool  sDB_SQL_Net::LoadDatabases (psDatabase  database, bool  recursive)
{
  DEBUG_LOG(NetOpen,sString("LoadDatabases(")<<recursive);
  bool                        OK = true;
  bool                        HasVersionField = false;
  try{
    try{
      sSQLQuery                   SQL(database->Connection());
      try{
        SQL.exec(sString("SELECT parval FROM ")
                +database->TableFullName("params")
                +" WHERE parkey = 'part_id'");
        if( ! SQL.first() )
          throw xNetOpen(sString("Cannot find part_id in 'params' of ")
                        +database->URL());
        sDB_UUID                          this_part_id( SQL.StringValue(0) );
        if( this_part_id != database->Id ){
          throw xNetOpen(sString("Wrong partition reference: ")+database->URL()
                        +"\nexpected part_id="+database->Id.ToString()
                        +",\nin fact, part_id="+this_part_id.ToString()
                        );
        }else{
          // get version
          {
            SQL.exec(sString("SELECT parval FROM ")
                    +database->TableFullName("params")
                    +" WHERE parkey = 'version'");
            if( ! SQL.first() )
              throw xNetOpen(sString("cannot find version in 'params' of ")
                            +database->URL());
            database->Version = sVersion( SQL.StringValue(0) );
          }
          //!@todo{NetIO} get oldest_node_version, newest_node_version
          // load related partitions
          {
            SQL.exec(sString("SELECT part_id, part_url, pcode FROM ")
                    +database->TableFullName("parts"));
            while( SQL.next() ){
              sDB_Link        reference( SQL.StringValue(1) );
              reference.Id = sDB_UUID( SQL.StringValue(0) );
              reference.Password  = FromQString( QString(
                             QByteArray::fromBase64(SQL.value(2).toByteArray())
                                             ));
              DEBUG_LOG(NetOpen,sString("LoadDatabases::passw=")
                               +reference.Password+".");
              std::map<sDB_UUID,psDatabase>
                            ::const_iterator  D =Databases.find(reference.Id);
              if( D != Databases.end() ){
                //!@todo{NetIO} duplicate info in linked database: check consistance
              }else{
                Databases[reference.Id] = new sDatabase(reference);
                if( recursive )
                  OK = LoadDatabases(Databases[reference.Id]) && OK;
              }
              //!@todo{NetIO} Databases[reference.Id]->SetReferencedFrom(database->Id)
            }
          }
        }
      }catch( rxSQL ){ ///@todo{1_6} Убрать чтение старого формата таблиц.
        try{
          SQL.exec(sString("SELECT id ,nickname ,comment ,type ,name "
                           ",host ,port ,login ,password ,version FROM ")
                  +database->TableFullName("databases"));
          HasVersionField = true;
        }catch( rxSQL ){
          SQL.exec(sString("SELECT id ,nickname ,comment ,type ,name "
                           ",host ,port ,login ,password FROM ")
                  +database->TableFullName("databases"));
        }
        while( SQL.next() ){
          psDatabase                DB = new sDatabase(sDB_Info());
          DB->Id        = sDB_UUID( SQL.StringValue(0) );
          sString DB_Nickname( SQL.StringValue(1) );
          DB->Type      = sDB_Info::eType( SQL.value(3).toUInt() );
          DB->DBName    = SQL.StringValue(4);
          if( DB->DBName.Pos("#") < DB->DBName.Length() ){
            DB->SchemaName = DB->DBName.Tail(DB->DBName.Pos("#")+1);
            DB->DBName = DB->DBName.Substr(0,DB->DBName.Pos("#"));
          }else if( DB->DBName.Pos("?") < DB->DBName.Length() ){
            DB->SchemaName = DB->DBName.Tail(DB->DBName.Pos("?")+1);
            DB->DBName = DB->DBName.Substr(0,DB->DBName.Pos("?"));
          }
          DB->TableNamePrefix = DB->SchemaName.Tail(DB->SchemaName.Pos(";")+1);
          DB->SchemaName = DB->SchemaName.Substr(0,DB->SchemaName.Pos(";"));
          DB->Host      = SQL.StringValue(5);
          DB->Port      = SQL.value(6).toInt();
          DB->Login     = SQL.StringValue(7);
          DB->Password  = SQL.StringValue(8);
          if( DB_Nickname == sDatabase::MyselfNickname ){
            if( DB->Id == database->Id ){
              if( HasVersionField )
                database->Version = sVersion( SQL.StringValue(9) );
            }else{
              QMessageBox::critical(NULL ,"Error reading database info"
                  ,QString("Wrong database reference: '%1',\n"
                           "expected id='%2',\n"
                           "in fact, id='%3'.")
                        .arg(ToQString(database->URL()))
                        .arg(ToQString(database->Id.ToString()))
                        .arg(ToQString(DB->Id.ToString()))
                  ,QMessageBox::Cancel);
              OK = false;
            }
            delete DB;
          }else{
            std::map<sDB_UUID,psDatabase>
                          ::const_iterator  D =Databases.find(DB->Id);
            if( D != Databases.end() ){
              //!@todo{NetIO} duplicate info in linked database: check consistance
            }else{
              Databases[DB->Id] = DB;
              if( recursive )
                OK = LoadDatabases(DB) && OK;
            }
          }
        }
      }
    }catch( rxSQL E ){
      OK = false;
      throw xNetIO(sString("SQL error in sDB_SQL_Net::LoadDatabases\n\n")
                  +E.Error);
    }
  }catch( rxNetIO E ){
    QMessageBox::critical(NULL,"Error opening RANet related databases"
        ,QString("\n\n%1")
            .arg(ToQString(E.MsgText()))
        ,QMessageBox::Close);
  }
  return OK;
}


void  sDB_SQL_Net::AddDBLink (rsDB_Link  link)
{
  //---- read database Id [and version] for the link
  psDatabase                  DB = new sDatabase(link);
  try{
    sSQLQuery                   SQL(DB->Connection());
    try{
      SQL.exec(sString("SELECT parval FROM ")+DB->TableFullName("params")
              +" WHERE parkey = 'part_id'");
      if( ! SQL.first() )
        throw xNetOpen("cannot find part_id in params of the given partition");
      DB->Id = sDB_UUID( SQL.StringValue(0) );
      SQL.exec(sString("SELECT parval FROM ")
              +DB->TableFullName("params")
              +" WHERE parkey = 'version'");
      if( ! SQL.first() )
        throw xNetOpen("cannot find version in params of the given partition");
      DB->Version = sVersion( SQL.StringValue(0) );
    }catch( rxSQL ){ ///@todo{1_6} Убрать чтение старого формата таблиц.
      bool                        HasVersionField = false;
      try{
        SQL.exec(sString("SELECT id, version FROM ")
                +DB->TableFullName("databases")
                +" WHERE nickname = '"+sDB_Link::MyselfNickname+"'");
        HasVersionField = true;
      }catch( rxSQL ){
        SQL.exec(sString("SELECT id FROM ")+DB->TableFullName("databases")
                +" WHERE nickname = '"+sDB_Link::MyselfNickname+"'");
      }
      if( ! SQL.first() )
        throw xSQL("Cannot find info on this database");
      DB->Id = sDB_UUID( SQL.StringValue(0) );
      if( HasVersionField )
        DB->Version = sVersion( SQL.StringValue(1) );
      if( SQL.next() )
        throw xSQL("Ambiguos info on this database");
    }
    link.Id = DB->Id;
    link.Version = DB->Version;
  }catch( rxSQL E ){
    delete DB;
    throw xNetIO(sString("SQL error in sDB_SQL_Net::AddDBLink\n")
                +"while reading info on the given database\n\n"
                +E.Error);
  }catch( ... ){
    delete DB;
    throw xNetIO("Unknown error in sDB_SQL_Net::AddDBLink\n"
                 "while reading info on the given database.");
  }
  //---- add database link into root database
  try{
    psDatabase                  RootDB (Databases[RootDB_UUID]);
    sSQLQuery                   InQ (RootDB->Connection());
    Databases[RootDB_UUID]->Connection().transaction();
    if( RootDB->Version < sVersion(1,5) ){
      InQ.prepare(sString("INSERT INTO ")+RootDB->TableFullName("databases")
                 +" (id,type,name,host,port,login,password)"
                 +" VALUES(?,?,?,?,?,?,?)");
      InQ.addBindString( link.Id.ToString() );
      InQ.addBindValue ( quint8(link.Type) );
      if( link.TableNamePrefix.Empty() ){
        InQ.addBindString( link.DBName + "?" + link.SchemaName );
      }else{
        InQ.addBindString( link.DBName + "?" + link.SchemaName
                                       + ";" + link.TableNamePrefix );
      }
      InQ.addBindString( link.Host );
      InQ.addBindValue( link.Port );
      InQ.addBindString( link.Login );
      InQ.addBindString( link.Password );
    }else{
      InQ.prepare(sString("INSERT INTO ")+RootDB->TableFullName("parts")
                 +" VALUES(?,?,?)");
      InQ.addBindString( link.Id.ToString() );
      InQ.addBindString( link.URL() );
      InQ.addBindValue( QByteArray((literal)(link.Password)).toBase64()
                      , QSql::In | QSql::Binary );
    }
    InQ.exec();
    Databases[RootDB_UUID]->Connection().commit();
  }catch( rxSQL E ){
    delete DB;
    try{ Databases[RootDB_UUID]->Connection().rollback(); }catch(...){};
    throw xNetIO(sString("SQL error in sDB_SQL_Net::AddDBLink\n\n")
                +E.Error);
  }catch( ... ){
    delete DB;
    throw xNetIO("Unknown error in sDB_SQL_Net::AddDBLink\n"
                 "while storing link.");
  }
  Databases[link.Id] = DB;
  LoadDatabases(DB);
}


void  sDB_SQL_Net::EditLink (rsDB_Link  link)
{
  try{
    psDatabase                  RootDB(Databases[RootDB_UUID]);
    sSQLQuery                   UpdateQ (RootDB->Connection());
    if( RootDB->Version < sVersion(1,5) ){
      UpdateQ.prepare(sString("UPDATE ")+RootDB->TableFullName("databases")
                 +" SET type=?,name=?,host=?,port=?,login=?,password=?"
                 +" WHERE id=?");
      UpdateQ.addBindValue ( quint8(link.Type) );
      if( link.TableNamePrefix.Empty() ){
        UpdateQ.addBindString( link.DBName + "?" + link.SchemaName );
      }else{
        UpdateQ.addBindString( link.DBName + "?" + link.SchemaName
                                       + ";" + link.TableNamePrefix );
      }
      UpdateQ.addBindString( link.Host );
      UpdateQ.addBindValue( link.Port );
      UpdateQ.addBindString( link.Login );
      UpdateQ.addBindString( link.Password );
    }else{
      UpdateQ.prepare(sString("UPDATE ")+RootDB->TableFullName("parts")
                 +" SET part_url = ? , pcode = ? WHERE part_id = ?");
      UpdateQ.addBindString( link.URL() );
      UpdateQ.addBindValue( QByteArray((literal)(link.Password)).toBase64()
                      , QSql::In | QSql::Binary );
    }
    UpdateQ.addBindString( link.Id.ToString() );
    UpdateQ.exec();
  }catch( rxSQL E ){
    throw xNetIO(sString("SQL error in sDB_SQL_Net::EditLink\n")
                +E.Error);
  }catch( ... ){
    throw xNetIO("Unknown error in sDB_SQL_Net::EditLink\n"
                 "while updating link.");
  }
}


void  sDB_SQL_Net::RemovePartition (psDatabase  partition)
{
  try{
    psDatabase                  RootDB(Databases[RootDB_UUID]);
    sSQLQuery                   SQL (RootDB->Connection());
    if( RootDB->Version < sVersion(1,5) ){
      SQL.prepare(sString("DELETE FROM ")+RootDB->TableFullName("databases")
                 +" WHERE id=?");
    }else{
      SQL.prepare(sString("DELETE FROM ")+RootDB->TableFullName("parts")
                 +" WHERE part_id = ?");
    }
    SQL.addBindString( partition->Id.ToString() );
    SQL.exec();
    Databases.erase(partition->Id);
  }catch( rxSQL E ){
    throw xNetIO(sString("SQL error in sDB_SQL_Net::RemovePartition\n")
                +E.Error);
  }catch( ... ){
    throw xNetIO("Unknown error in sDB_SQL_Net::RemovePartition\n");
  }
}


bool  sDB_SQL_Net::CreateTables (rcsDB_Info  info
                                ,bool        initiate)
{
  bool                        OK = false;
//  sDB_UUID                    DBUUID( sDB_UUID::NewUUID() );
  //do it with a temporary connection
  {
    sDatabase                    DB(info);
    sDBConnection                DBC = DB.Connection("creation");
    try{
      DBC.transaction();
      sSQLQuery                   SQL(DBC);
      sString                     NewUUID;

      //_______ create and fill table "params"
      {
        SQL.exec(sString("CREATE TABLE ")+info.TableFullName("params")
                +"(parkey VARCHAR(128) UNIQUE NOT NULL"
                +",parval VARCHAR(1024)"
                +",PRIMARY KEY (parkey)"
                +")"
                );
        if( initiate ){
          NewUUID = FromQString(QUuid::createUuid().toString());
          SQL.SetParamValue( info.TableFullName("params")
                           , "part_id"
//                           , sNodeUUID::ForTheRoot(this,DBUUID
//                                                  ,RANet::Version
//                                                  ).FormattedDBId()
                           , NewUUID
                           );
          SQL.SetParamValue( info.TableFullName("params")
                           , "version"
                           , RANet::Version.Text() );
          SQL.SetParamValue( info.TableFullName("params")
                           , "created at"
                           , sTime().ToISO_8601() );
          SQL.SetParamValue( info.TableFullName("params")
                           , "oldest_node_version"
                           , RANet::Version.Principal().Text() );
          SQL.SetParamValue( info.TableFullName("params")
                           , "newest_node_version"
                           , RANet::Version.Principal().Text() );
        }
      }
      //_______ create table "parts"
      {
        SQL.exec(sString("CREATE TABLE ")+info.TableFullName("parts")
                +"(part_id  "+DBIdFieldDecl+" UNIQUE NOT NULL"
                +",part_url VARCHAR(1024) NOT NULL"
                +",pcode "+( info.Type==sDB_Info::cPostgreSQL
                           ? "bytea"
                           : "MEDIUMBLOB"
                           )
                +")"
                );
      }
      //_______ create and fill table "nodes"
      {
        SQL.exec(sString("CREATE TABLE ")+info.TableFullName("nodes")
                +"(node_id  "+NodeIdFieldDecl+" UNIQUE NOT NULL"
                +",node_class  VARCHAR(80) NOT NULL"
                +",node_body  "+( info.Type==sDB_Info::cPostgreSQL
                                ? "bytea"
                                : "MEDIUMBLOB"
                                )
                +",node_save_time  VARCHAR(30) NOT NULL"
                +",PRIMARY KEY ("+info.ColumnNotation("node_id")+")"
                +( info.Type==sDB_Info::cMySQL
                 ? ",UNIQUE INDEX `node_id_UNIQUE` (`node_id` ASC)"
                 : ""
                 )
                +")"
                );
        if( initiate ){
          SQL.prepare(sString("INSERT INTO ")+info.TableFullName("nodes")
                     +" (node_id,node_class,node_body,node_save_time)"
                      " VALUES(?,?,?,?)");
          SQL.addBindString( NewUUID );
//          SQL.addBindString( sNodeUUID::ForTheRoot(this,DBUUID
//                                                  ,RANet::Version
//                                                  ).FormattedNodeId() );
          SQL.addBindString( RANet::Version.Principal().Text()
                           + sFolderNode::Type()->Name() );
          SQL.addBindValue ( QByteArray(4,'\0') );
          SQL.addBindString( sTime().ToISO_8601() );
          SQL.exec();
        }
      }
      //_______ create table "attrs"
      {
        SQL.exec(sString("CREATE TABLE ")+info.TableFullName("attrs")
                +"(host_node      "+NodeIdFieldDecl+" NOT NULL"
                +",ordering_index INT NOT NULL"
                +",attr_type      INT NOT NULL"
                +",value_node     "+NodeIdFieldDecl+" NOT NULL"
                +",value_db       "+DBIdFieldDecl
                +",numeric_name   INT"
                +",symbolic_name  TEXT"
                ///@todo{NetIO} Время атрибута пишется, но не используется
                +",birth_time     TEXT NOT NULL"
              ///@todo{SQL}  +",PRIMARY KEY (host_node,ordering_index)"
              /// Не надо ли индексировать отдельно и по host_node?
                +")"
                );
      }
      OK = DBC.commit();
      OK = true;
    }catch( rxSQL E ){
      try{
        DBC.rollback();
        throw xNetIO(sString("SQL error in sDB_SQL_Net::CreateTables\n\n")
                    +E.Error);
      }catch( rxSQL RBE ){
        throw xNetIO(sString("SQL error in sDB_SQL_Net::CreateTables\n\n")
                    +E.Error
                    +"\nИ откатить не удалось из-за\n"
                    +RBE.Error);
      }
    }
  }
  return OK;
}


void  sDB_SQL_Net::ReplicatePartition (rcsDB_Info  source
                                      ,rcsDB_Info  destination
                                      ,int *       progress)
{
  DEBUG_LOG(NetTools,sString("ReplicatePartition(")+source.URL()
                     +","+destination.URL()+")");

  try{
    CreateNewPartition(destination,false);
  }catch( rxNetIO ){
    throw;
  }catch(...){
    throw xException("Uknown error while creating new partition for replica.");
  }
  sDatabase                   Source(source);
  sDatabase                   Destination(destination);
  bool                        HasParamsTable = false;
  sSQLQuery                   SQL(Source.Connection("replication"));
  sSQLQuery                   InQ(Destination.Connection("replication"));
  try{ // --- replicate params ---
    try{
      SQL.exec(sString("SELECT parkey, parval FROM ")
              +source.TableFullName("params"));
      HasParamsTable = true;
    }catch( rxSQL ){
      SQL.exec(sString("SELECT id FROM ")
              +source.TableFullName("databases")
              +" WHERE nickname = '"+sDatabase::MyselfNickname+"'");
    }
    if( HasParamsTable ){
      while( SQL.next() ){
        if(       SQL.StringValue(0) == "part_id" ){
          Source.Id = sDB_UUID( SQL.StringValue(1) );
        }else if( SQL.StringValue(0) == "version" ){
          Source.Version = sVersion( SQL.StringValue(1) );
        //!@todo{NetIO} get oldest_node_version, newest_node_version
        //}else if( SQL.StringValue(0) == "oldest_node_version" ){
        //}else if( SQL.StringValue(0) == "newest_node_version" ){
        }
        InQ.prepare(sString("INSERT INTO ")
                   +destination.TableFullName("params")
                   +" VALUES(?,?)");
        InQ.addBindString( SQL.StringValue(0) );
        InQ.addBindString( SQL.StringValue(1) );
        InQ.exec();
      }
    }else{
      if( ! SQL.first() )
        throw xNetIO(sString("ReplicatePartition: cannot find '")
                      +sDatabase::MyselfNickname
                      +"' in databases of the source partition");
      Source.Id = sDB_UUID( SQL.StringValue(0) );
      if( SQL.next() )
        throw xNetOpen(sString("ReplicatePartition: ambiguos '")
                      +sDatabase::MyselfNickname
                      +"' in 'databases' table of the source partition");
      InQ.SetParamValue( destination.TableFullName("params")
                       , "part_id" , Source.Id.ToString() );
      InQ.SetParamValue( destination.TableFullName("params")
                       , "version" , RANet::Version.Text() );
    }
    InQ.SetParamValue( destination.TableFullName("params")
                     , "replicated at" , sTime().ToISO_8601() );
    InQ.SetParamValue( destination.TableFullName("params")
                     , "replicated from" , Source.URL() );
  }catch( rxSQL E ){
    throw xNetIO(sString("SQL error in sDB_SQL_Net::ReplicatePartition\n\n")
                +E.Error);
  }catch( ... ){
    throw xException("Uknown error while replicating 'params' table.");
  }
  try{ // --- replicate parts ---
    if( HasParamsTable ){
      SQL.exec(sString("SELECT part_id, part_url, pcode FROM ")
              +source.TableFullName("parts"));
      while( SQL.next() ){
        InQ.prepare(sString("INSERT INTO ")
                   +destination.TableFullName("parts")
                   +" VALUES(?,?,?)");
        InQ.addBindString( SQL.StringValue(0) );
        InQ.addBindString( SQL.StringValue(1) );
        InQ.addBindString( SQL.StringValue(2) );
        InQ.exec();
      }
    }else{
      sDB_Link                    Link;
      SQL.exec(sString("SELECT id ,type ,name "
                       ",host ,port ,login ,password FROM ")
              +source.TableFullName("databases")
              +" WHERE nickname != '"+sDatabase::MyselfNickname+"'");
      while( SQL.next() ){
        Link.Id        = sDB_UUID( SQL.StringValue(0) );
        Link.Type      = sDB_Info::eType( SQL.value(1).toUInt() );
        Link.DBName    = SQL.StringValue(2);
        if( Link.DBName.Pos("#") < Link.DBName.Length() ){
          Link.SchemaName = Link.DBName.Tail(Link.DBName.Pos("#")+1);
          Link.DBName = Link.DBName.Substr(0,Link.DBName.Pos("#"));
        }else if( Link.DBName.Pos("?") < Link.DBName.Length() ){
          Link.SchemaName = Link.DBName.Tail(Link.DBName.Pos("?")+1);
          Link.DBName = Link.DBName.Substr(0,Link.DBName.Pos("?"));
        }
        Link.Host      = SQL.StringValue(3);
        Link.Port      = SQL.value(4).toInt();
        if( Link.Port == 0 )  Link.Port = -1;
        Link.Login     = SQL.StringValue(5);
        Link.Password  = SQL.StringValue(6);
        InQ.prepare(sString("INSERT INTO ")+destination.TableFullName("parts")
                   +" (part_id, part_url, pcode) VALUES(?,?,?)");
        InQ.addBindString( Link.Id.ToString() );
        InQ.addBindString( Link.URL() );
        InQ.addBindValue( QByteArray((literal)(Link.Password)).toBase64()
                        , QSql::In | QSql::Binary );
        InQ.exec();
      }
    }
  }catch( rxSQL E ){
    throw xNetIO(sString("SQL error in sDB_SQL_Net::ReplicatePartition\n\n")
                +E.Error);
  }catch( ... ){
    throw xException("Uknown error while replicating 'parts' table.");
  }

  ReplicatePartNodes( Source , Destination , progress );
  ReplicatePartAttrs( Source , Destination , progress );
}


void  sDB_SQL_Net::ReplicatePartNodes (rcsDB_Link  source
                                      ,rcsDB_Link  destination
                                      ,int *       progress)
{
  sDatabase                   Source(source);
  sDatabase                   Destination(destination);
  sSQLQuery                   SQL(Source.Connection("replication_nodes"));
  sSQLQuery                   InQ(Destination.Connection("replication_nodes"));
  bool                        SourceVersionIsLessThan_1_5 = false;

  try{ // --- replicate nodes ---
    sString                     NodeId;
    sString                     NodeClass;
    QByteArray                  OldNodeBody;
    sString                     TimeText;
    if( Source.Version < sVersion(1,5) ){
      SourceVersionIsLessThan_1_5 = true;
      SQL.exec(sString("SELECT id,class,body,time")
              +" FROM "+Source.TableFullName("nodes"));
    }else{
      SQL.exec(sString("SELECT node_id,node_class,node_body,node_save_time")
              +" FROM "+Source.TableFullName("nodes"));
    }
    while( SQL.next() ){
      NodeId = SQL.StringValue(0);
      if( SourceVersionIsLessThan_1_5 ){
        if( NodeId == sString(QUuid().toByteArray().data()) ){
          NodeId = Source.Id.ToString();
        }
      }
      NodeClass = SQL.StringValue(1);
      OldNodeBody = SQL.value(2).toByteArray();
      psNode                      Node;
      std::stringstream           Stream(std::string(OldNodeBody.data()
                                                    ,OldNodeBody.size()+1   )
                                        ,std::ios_base::in);
      Node = dynamic_cast<psNode>
                         ( sStorable::PolymorphicLoad(NodeClass,Stream) );
      Node->Store(Stream);
      QByteArray                  NewNodeBody (Stream.str().data()
                                              ,Stream.str().size());
      delete Node;
      TimeText = SQL.StringValue(3);
      InQ.prepare(sString("INSERT INTO ")+Destination.TableFullName("nodes")
               +" VALUES(?,?,?,?)" );
      InQ.addBindString( NodeId );
      InQ.addBindString( NodeClass );
      InQ.addBindValue( NewNodeBody , QSql::In | QSql::Binary );
      InQ.addBindString( TimeText );
      InQ.exec();
      if( progress ){
        (*progress)++;
        if( (*progress) % 100 == 0 ){
          std::cout << "№№№ nodes: " << (*progress) << std::endl;
        }
      }
    }
    InQ.SetParamValue( destination.TableFullName("params")
                     , "oldest_node_version"
                     , RANet::Version.Principal().Text() );
    InQ.SetParamValue( destination.TableFullName("params")
                     , "newest_node_version"
                     , RANet::Version.Principal().Text() );
  }catch( rxSQL E ){
    throw xNetIO(sString("SQL error in sDB_SQL_Net::ReplicatePartNodes\n\n")
                +E.Error);
  }catch( rxException E ){
    throw xException(E.MsgText()+" \n while replicating 'nodes' table.");
  }catch( ... ){
    throw xException("Uknown error while replicating 'nodes' table.");
  }
}


void  sDB_SQL_Net::ReplicatePartAttrs (rcsDB_Link  source
                                      ,rcsDB_Link  destination
                                      ,int *       progress)
{
  sDatabase                   Source(source);
  sDatabase                   Destination(destination);
  sSQLQuery                   SQL(Source.Connection("replication_attrs"));
  sSQLQuery                   InQ(Destination.Connection("replication_attrs"));
  bool                        SourceVersionIsLessThan_1_5 = false;

  try{ // --- replicate attrs ---
    sString                     HostNodeId;
    sString                     PrevHostNodeId;
    int32_t                     OrderingIndex =0;//to avoid compiler warning
    quint8                      Type;
    sString                     ValueNodeId;
    bool                        ValueDBIdIsNull;
    sString                     ValueDBId;
    bool                        NumericIsNull;
    int                         Numeric =0;//to avoid compiler warning
    bool                        SymbolicIsNull;
    sString                     Symbolic;
    sString                     TimeText;
    sString                     Fields;
    sString                     Placeholders;
    if( Source.Version < sVersion(1,5) ){
      SourceVersionIsLessThan_1_5 = true;
      SQL.exec(sString("SELECT host,type,value,value_db,")
              +Source.ColumnNotation("numeric")
              +",symbolic,time"
              +" FROM "+Source.TableFullName("attributes"));
    }else{
      SQL.exec(sString("SELECT host_node,attr_type,value_node,value_db,")
              +"numeric_name,symbolic_name,birth_time"
              +" FROM "+Source.TableFullName("attrs")
              +" ORDER BY host_node, ordering_index");
    }
    while( SQL.next() ){
      Fields = "birth_time,host_node,ordering_index,attr_type,value_node";
      Placeholders = "?,?,?,?,?";
      HostNodeId = SQL.StringValue(0);
      if( HostNodeId != PrevHostNodeId ){
        OrderingIndex = sAttribute::OrderingIndexMin;
        PrevHostNodeId = HostNodeId;
      }else{
        OrderingIndex += sAttribute::OrderingIndexInc;
      }
      Type = SQL.value(1).toUInt();
      ValueNodeId = SQL.StringValue(2);
      ValueDBIdIsNull = SQL.value(3).isNull();
      if( ! ValueDBIdIsNull ){
        ValueDBId = SQL.StringValue(3);
        Fields += ",value_db";
        Placeholders += ",?";
      }
      NumericIsNull = SQL.value(4).isNull();
      if( ! NumericIsNull ){
        Numeric = SQL.value(4).toInt();
        Fields += ",numeric_name";
        Placeholders += ",?";
      }
      SymbolicIsNull = SQL.value(5).isNull();
      if( ! SymbolicIsNull ){
        Symbolic = SQL.StringValue(5);
        Fields += ",symbolic_name";
        Placeholders += ",?";
      }
      TimeText = SQL.StringValue(6);
      if( TimeText.Empty() ){
        TimeText = sTime().ToISO_8601();
      }
      if( SourceVersionIsLessThan_1_5 ){
        if( HostNodeId == sString(QUuid().toByteArray().data()) ){
          HostNodeId = Source.Id.ToString();
        }
        if( ValueNodeId == sString(QUuid().toByteArray().data()) ){
          if( ValueDBIdIsNull ){
            ValueNodeId = Source.Id.ToString();
          }else{
            ValueNodeId = ValueDBId;
          }
        }
      }
      InQ.prepare(sString("INSERT INTO ")+Destination.TableFullName("attrs")
               +" (" + Fields + ") VALUES (" + Placeholders +")" );
      InQ.addBindString( TimeText );
      InQ.addBindString( HostNodeId );
      InQ.addBindValue ( OrderingIndex );
      InQ.addBindValue ( Type );
      InQ.addBindString( ValueNodeId );
      if( ! ValueDBIdIsNull ){
        InQ.addBindString( ValueDBId );
      }
      if( ! NumericIsNull ){
        InQ.addBindValue( Numeric );
      }
      if( ! SymbolicIsNull ){
        InQ.addBindString( Symbolic );
      }
      InQ.exec();
      if( progress ){
        (*progress)++;
        if( (*progress) % 100 == 0 ){
          std::cout << "@@@ attrs: " << (*progress) << std::endl;
        }
      }
    }
  }catch( rxSQL E ){
    throw xNetIO(sString("SQL error in sDB_SQL_Net::ReplicatePartAttrs\n\n")
                +E.Error);
  }catch( rxException E ){
    throw xException(E.MsgText()+" \n while replicating 'attrs' table.");
  }catch( ... ){
    throw xException("Uknown error while replicating 'attrs' table.");
  }
}


void  sDB_SQL_Net::RunSQL (rcsDB_Info  part_info
                          ,rcsString   sql_text  )
{
  sDatabase                   DB (part_info);
  try{
    sSQLQuery                   SQL(DB.Connection("extra_SQL"));
    std::cout << "@@@         SQL: " << (literal)sql_text
        << ";" << std::endl;
    SQL.exec(sql_text);
    std::cout << "@@@           numRowsAffected=" << SQL.numRowsAffected()
        << std::endl;
    SQL.finish();
  }catch( rxSQL E ){
    throw xNetIO(sString("SQL error in sDB_SQL_Net::RunSQL\n\n")
                +E.Error);
  }catch( ... ){
    throw xNetIO("Unknown error in sDB_SQL_Net::RunSQL\n\n");
  }
}


void  sDB_SQL_Net::SetParamValue (rcsDB_Link  partition
                                 ,rcsString   param_key
                                 ,rcsString   param_value)
{
  try{
    sDatabase                   Partition(partition);
    sSQLQuery                   SQL(Partition.Connection("set_param"));
    SQL.SetParamValue( partition.TableFullName("params")
                     , param_key , param_value );
  }catch( rxSQL E ){
    throw xNetIO(sString("SQL error in sDB_SQL_Net::SetParamValue\n\n")
                +E.Error);
  }catch( ... ){
    throw xException("Uknown error in sDB_SQL_Net::SetParamValue.");
  }
}


bool  sDB_SQL_Net::CreateNewPartition (rcsDB_Info  info
                                      ,bool        initiate)
{
  if( info.Type == sDB_Info::cSQLite ){
    sPath  FP(info.DBName);
    if( ! FP.Exists() ){
      sFile  F;
      F.Assign(FP,sFile::Binary);
      F.Open(sFile::Writing);
      F.Close();
    }
  }
  return CreateTables(info,initiate);
}




namespace DB_SQL_Net {
//------------------------------------------------------------------------------
//------------------------------------------------ DB_SQL_Net::sSynchronizer ---

sSynchronizer::~sSynchronizer ()
{
  DEBUG_LOG(Saving,"~sSynchronizer");
  for( std::map<sDB_UUID,psDatabase>::iterator  D = Net->Databases.begin()
                                              ; D != Net->Databases.end()
                                              ; ++D ){
    D->second->CloseBackgroundConnection();
  };
}


sSynchronizer::sSynchronizer (psDB_SQL_Net  net)
    :QObject()
    ,Net(net)
    ,LastSaved(NULL)
    ,LastRefreshed(NULL)
{
  DEBUG_LOG(Saving,"sSynchronizer::sSynchronizer");
  connect(this,SIGNAL(SaveMore())   ,SLOT(SaveSome())   ,Qt::QueuedConnection);
  connect(this,SIGNAL(RefreshMore()),SLOT(RefreshSome()),Qt::QueuedConnection);
}


void  sSynchronizer::StartSaving ()
{
  QTimer::singleShot(StartDelay,this,SLOT(SaveSome()));
}


void  sSynchronizer::SaveSome ()
{
  QMutexLocker                LoadedNodesLocker(&(Net->LoadedNodesMutex));
  if( Net->LoadedNodes.empty() ){
    DEBUG_LOG(Saving
        ,"sSynchronizer::SaveSome: Net->LoadedNodes.empty()!!!!!!!!!!!!");
    QTimer::singleShot(OnIdlePause,this,SLOT(SaveSome()));
  }else{
    std::map<sDB_SQL_Net::sNodeUUID,psNode>::iterator  P;
    //set P to the next element that has to be saved:
    {
      if( ! LastSaved )
        P = Net->LoadedNodes.begin();
      else
        P = Net->LoadedNodes.upper_bound(*LastSaved);
      if( P != Net->LoadedNodes.end() )
        P = find_if( P ,Net->LoadedNodes.end() ,sSynchronizer::HasToBeSaved);
    }
    if( P == Net->LoadedNodes.end() ){
      delete LastSaved;  LastSaved=NULL;
      emit RefreshMore();
    }else{
      //TRACING
      {
        #ifdef  RANET_DEBUG
          sDB_SQL_Net::sNodeUUID  UID = P->first;
          psNode     N = P->second;
          sString    Saving("Saving");
          Saving += UID.FormattedNodeId();
          Saving += N->TrueType()->Name();
          if( N->BodyHasToBeSaved() || N->AttrsHaveToBeSaved() )
            DEBUG_LOG(Saving,Saving);
        #endif
      }
      if( P->second->BodyHasToBeSaved() ){
        Net->Save(P->second);
        DEBUG_LOG(Saving,sString("body - ")
                      +(P->second->BodyHasToBeSaved() ? "FAILURE" : "done"));
      }
      if( P->second->AttrsHaveToBeSaved() ){
        Net->SaveAttributes(P->second);
        DEBUG_LOG(Saving,sString("attributes - ")
                      +(P->second->AttrsHaveToBeSaved() ? "FAILURE" : "done"));
      }
      if( ! LastSaved ){
        LastSaved = new sDB_SQL_Net::sNodeUUID(P->first);
      }else{
        (*LastSaved) = P->first;
      }
      emit SaveMore();
    }
  }
}


void  sSynchronizer::RefreshSome ()
{
  QMutexLocker                LoadedNodesLocker(&(Net->LoadedNodesMutex));
  if( Net->LoadedNodes.empty() ){
    RANet::Log.Put(sLog::Warning,"Refreshing"
        ,"sSynchronizer::RefreshSome: Net->LoadedNodes.empty()!!!!!!!!!!!!!!");
    QTimer::singleShot(OnIdlePause,this,SLOT(SaveSome()));
  }else{
    std::map<sDB_SQL_Net::sNodeUUID,psNode>::iterator  P;
    //set P to the next element that has to be refreshed:
    {
      if( ! LastRefreshed )
        P = Net->LoadedNodes.begin();
      else
        P = Net->LoadedNodes.upper_bound(*LastRefreshed);
      if( P != Net->LoadedNodes.end() )
        P = find_if(P ,Net->LoadedNodes.end() ,sSynchronizer::HasToBeRefreshed);
    }
    if( P == Net->LoadedNodes.end() ){
      delete LastRefreshed;  LastRefreshed=NULL;
      DEBUG_LOG(Refreshing,"synchronization paused");
      QTimer::singleShot(OnIdlePause,this,SLOT(SaveSome()));
    }else{
      //TRACING
      {
        #ifdef  RANET_DEBUG
          sDB_SQL_Net::sNodeUUID  UID = P->first;
          psNode     N = P->second;
          sString    Refreshing("Refreshing");
          Refreshing += UID.FormattedNodeId();
          Refreshing += N->TrueType()->Name();
          if( N->BodyHasToBeRefreshed() || N->AttrsHaveToBeRefreshed() )
            DEBUG_LOG(Refreshing,Refreshing);
        #endif
      }
      if( P->second->BodyHasToBeRefreshed() ){
        Net->Refresh(P->second);
        DEBUG_LOG(Refreshing,sString("body - ")
                      +(P->second->BodyHasToBeSaved() ? "FAILURE" : "done"));
      }
      if( P->second->AttrsHaveToBeRefreshed() ){
        Net->RefreshAttributes(P->second);
        DEBUG_LOG(Refreshing,sString("attributes - ")
                      +(P->second->AttrsHaveToBeSaved() ? "FAILURE" : "done"));
      }
      if( ! LastRefreshed ){
        LastRefreshed = new sDB_SQL_Net::sNodeUUID(P->first);
      }else{
        (*LastRefreshed) = P->first;
      }
      emit SaveMore();
    }
  }
}


bool sSynchronizer::HasToBeSaved (const
                                   std::pair<sDB_SQL_Net::sNodeUUID,psNode>
                                   &  item)
{
  return item.second->BodyHasToBeSaved()
      || item.second->AttrsHaveToBeSaved();
}


bool sSynchronizer::HasToBeRefreshed (const
                                      std::pair<sDB_SQL_Net::sNodeUUID,psNode>
                                      &  item)
{
  return item.second->BodyHasToBeRefreshed()
      || item.second->AttrsHaveToBeRefreshed();
}

//------------------------------------------------------------------------------
}//namespace DB_SQL_Net


//------------------------------------------------------------------------------
} //namespace RA


