////////////////////////////////////////////////////////////////////////////////
/*! @file Excepts.cpp   Basic classes for specific exception handling.
- Part of RANet - Research Assistant Net Library.
- Copyright(C) 1994-2017, Viktor E. Bursian, St.Petersburg, Russia.
                          VBursian@gmail.com
*///////////////////////////////////////////////////////////////////////////////
#include "Excepts.h"
#include <iostream>
#include <typeinfo>
/*
 *  @brief  Throwing exceptions on errors.
 *  @param  except  The new exceptions mask.
 *
 *  By default, error flags are set silently.  You can set an
 *  exceptions mask for each stream; if a bit in the mask becomes set
 *  in the error flags, then an exception of type
 *  std::ios_base::failure is thrown.
 *
 *  If the error flage is already set when the exceptions mask is
 *  added, the exception is immediately thrown.  Try running the
 *  following under GCC 3.1 or later:
 *  @code
 *  #include <iostream>
 *  #include <fstream>
 *  #include <exception>
 *
 *  int main()
 *  {
 *      std::set_terminate (__gnu_cxx::__verbose_terminate_handler);
 *
 *      std::ifstream f ("/etc/motd");
 *
 *      std::cerr << "Setting badbit\n";
 *      f.setstate (std::ios_base::badbit);
 *
 *      std::cerr << "Setting exception mask\n";
 *      f.exceptions (std::ios_base::badbit);
 *  }
 *  @endcode
*/

namespace RA {
//------------------------------------------------------------------------------
//--------------------------------------------------------------- xException ---

//xException::xException(rcxException other)
//    :TheFileName(other.TheFileName)
//    ,TheLineNo(other.TheLineNo)
//    ,TheClassName(other.TheClassName)
//    ,TheMsgText(other.TheMsgText)
//{
//  std::cerr << (literal)
//               (sString("+++++++++++++++++++++++++++++++++++++++++++\n+++ ")
//               +sString(__func__)            +sString(" +++ ")
//               +sString(__FUNCTION__)        +sString(" +++ ")
//               +sString(__PRETTY_FUNCTION__) +sString(" +++ ")
//               +TheClassName
//               +WhereText()
//               +( TheMsgText.Empty() ? "" : sString(": ") + TheMsgText )
//               +sString("\n+++++++++++++++++++++++++++++++++++++++++++\n")
//               );
//}


sString  xException::SorryText () const
{
  return sString("Sorry, there was an unhandled exception ") + TheClassName
         +WhereText()
         +( TheMsgText.Empty() ? "" : ":\n") + TheMsgText;
}

sString  xException::WhereText () const
{
  return ( TheLineNo
         ? sString(" thrown in ") + TheFileName + (sString(":") << TheLineNo)
         : ""
         );
}


void xException::Report2cerr(const std::exception &  e)
{
  std::cerr << (literal)
               (sString("***********************************************\n*** ")
               +sString(typeid(e).name())
               +sString(": ") + sString(e.what())
               +sString("\n***********************************************\n")
               );
}


void  xException::Report2cerr () const
{
  std::cerr << (literal)
               (sString("***********************************************\n*** ")
               +TheClassName
               +WhereText()
               +( TheMsgText.Empty() ? "" : sString(": ") + TheMsgText )
               +sString("\n***********************************************\n")
               );
}

//------------------------------------------------------------ xAssertFailed ---

xAssertFailed::xAssertFailed (literal  assertion
                             ,literal  src_file_name
                             ,int      line_no      )
    :xProgramError(sString("ASSERT( ")+assertion+" ) failure"
                  ,src_file_name,line_no,"xAssertFailed")
{
}

//---------------------------------------------------------- xNotImplemented ---

xNotImplemented::xNotImplemented (rcsString  what)
    :xProgramError(sString("Not yet implemented: ")
                  +what
                  ,"",0,"xNotImplemented")
{
}

//------------------------------------------------------------------------------
} //namespace RA

