////////////////////////////////////////////////////////////////////////////////
/*! @file FolderNode.cpp A node of RANet Knowledgebase that can have attributes.
- Part of RANet - Research Assistant Net Library (based on ANSI C++).
- Copyright(C) 1994-2013, Viktor E. Bursian, St.Petersburg, Russia.
                          Viktor.Bursian@mail.ioffe.ru
*///////////////////////////////////////////////////////////////////////////////
#include "FolderNode.h"
#include "Net.h"
#include "Symbols.h"
#include "Log.h"
#include <QMutexLocker>
#include <cstring>
namespace RA {
//------------------------------------------------------------------------------
//-------------------------------------------------------------- sFolderNode ---
/*! @class sFolderNode


@par sFolderNode::AttrListIOStatus lifecycle:
-  (node bearing constructor)--> cRefreshed
-  (node loading constructor)--> cToBeLoaded
-  cToBeLoaded ---(HasAttributes)--> cRefreshed
-  cToBeLoaded ---(GetAttr)--> cRefreshed
-  cToBeLoaded ---(net ops)-->                 !!!!!!!!!!
-  cRefreshed  ---(net ops)--> cChanged
-  cSaved      ---(net ops)--> cChanged
-  cChanged    ---(net ops)--> cChanged
-  cChanged ---(SaveAttributes)--> cSaved
-  cSaved ---(RefreshAttributes)--> cRefreshed

The cNewborn status is for sNode::IOStatus only.

@todo{--Qt} see mutex in C++11 or Boost.Interprocess for mutexes (which,
            by the way, have try_to_lock and recursive features)

*/



sFolderNode::~sFolderNode ()
{
  //!@todo{Net} allocate, save, check ---  what if not?!
  TheAttributes.clear();
}


sFolderNode::sFolderNode (sString  docket)
    :sNode()
    ,AttrListMutex()
    ,AttrListIOStatus(cRefreshed)
    ,AttrListStatusTimeStamp()
    ,TheAttributes()
    ,TheDocket(docket)
{
}


sFolderNode::sFolderNode (std::istream &  a_stream ,rcsVersion  version)
    :sNode(a_stream,version)
    ,AttrListMutex()
    ,AttrListIOStatus(cToBeLoaded)
    ,AttrListStatusTimeStamp()
    ,TheAttributes()
{
  TheDocket.Read(a_stream);
}


void  sFolderNode::Store (std::ostream &  a_stream) const
{
  sNode::Store(a_stream);
  TheDocket.Write(a_stream);
}


sString  sFolderNode::Text (eTextFormat        /*F*/
                           ,eTextDetalization  D)
{
  sString                     T;
  {
    QMutexLocker                BodyLocker(&BodyMutex);
    T = TheDocket;
  }
  /*! @todo{UI} implement HTML -> Plain*/
  if( (D == Laconic) && (T.Length() > CutOffLengthForLaconic) ){
    T=T.Substr(0,CutOffLengthForLaconic-strlen(Symbol::Ellipsis));
    /*! @todo{sString} Substr for UTF-8*/
    while( T[T.Length()-1] < 0 )  T=T.Substr(0,T.Length()-1);
    T+=Symbol::Ellipsis;
  };
  return T;
}


sString  sFolderNode::Docket ()
{
  QMutexLocker                BodyLocker(&BodyMutex);
  return TheDocket;
}


void  sFolderNode::SetDocket (rcsString  new_docket)
{
  QMutexLocker                BodyLocker(&BodyMutex);
  if( TheDocket != new_docket ){
//    sNet::Manager->Undo.Push( new sUndoAssignDocket2Cluster(sNodePtr(this)
//                                                           ,new_docket) );
    TheDocket=new_docket;
    MarkAsEdited();
  };
}


void sFolderNode::CopyFrom (psNode  other)
{
  sNode::CopyFrom(other);
  psFolderNode                Other = dynamic_cast<psFolderNode>(other);
  if( Other ){
    if( TheDocket != Other->TheDocket ){
      TheDocket=Other->TheDocket;
      MarkAsEdited();
    }
  }
}


bool  sFolderNode::HasAttributes ()
{
  QMutexLocker                AttrListLocker(&AttrListMutex);
  if( AttrListIOStatus == cToBeLoaded ){
    Location()->Net->GetAttrList( TheAttributes , Location());
    AttrListIOStatus = cRefreshed;
    AttrListStatusTimeStamp = sTime();
  }
  return  ! TheAttributes.empty();
}


bool  sFolderNode::AttrsHaveToBeSaved () const
{
  //!@todo{NetIO} Location() && ! (IOStatus == cNewborn) && ...
  return  AttrListIOStatus == cChanged;
}


bool  sFolderNode::AttrsHaveToBeRefreshed () const
{
  bool R = (AttrListIOStatus == cSaved)
      || (   (AttrListIOStatus == cRefreshed)
          && (sTime() - AttrListStatusTimeStamp > 10.0/*s*/)   );

  //!@todo{NetIO} Should not return true, if (IOStatus == cNewborn) ?
  if( R && (IOStatus == cNewborn) ){
    RANet::Log.Put(sLog::ProgramError,"Refreshing"
                  ,"sFolderNode::AttrsHaveToBeRefreshed newborn refreshing");
  }

  return R;
}


sNodePtr  sFolderNode::GetAttr (sAttrName  attr_name)
{
  QMutexLocker                AttrListLocker(&AttrListMutex);
  if( AttrListIOStatus == cToBeLoaded ){
    Location()->Net->GetAttrList( TheAttributes , Location());
    AttrListIOStatus = cRefreshed;
    AttrListStatusTimeStamp = sTime();
  }
  for( std::list<sAttribute>::const_iterator  A = TheAttributes.begin()
                                            ; A != TheAttributes.end() ; ++A ){
    if( A->IOStatus != sAttribute::cDeleted )
      if( A->Name() == attr_name )
        return  A->Value();
  }
  return sNodePtr();
}


void  sFolderNode::SetAttr (sAttribute  attribute)
{
  QMutexLocker                AttrListLocker(&AttrListMutex);
  if( AttrListIOStatus == cToBeLoaded ){
    Location()->Net->GetAttrList( TheAttributes , Location());
    AttrListIOStatus = cRefreshed;
    AttrListStatusTimeStamp = sTime();
  }
  SetAttribute(attribute);
}


sNodePtr  sFolderNode::AscribeAttr (sAttribute  attribute)
{
  QMutexLocker                AttrListLocker(&AttrListMutex);
  if( AttrListIOStatus == cToBeLoaded ){
    Location()->Net->GetAttrList( TheAttributes , Location());
    AttrListIOStatus = cRefreshed;
    AttrListStatusTimeStamp = sTime();
  }
  return AscribeAttribute(attribute);
}


void  sFolderNode::DelAttr (sAttrName  attr_name)
{
  QMutexLocker                AttrListLocker(&AttrListMutex);
  if( AttrListIOStatus == cToBeLoaded ){
    Location()->Net->GetAttrList( TheAttributes , Location());
    AttrListIOStatus = cRefreshed;
    AttrListStatusTimeStamp = sTime();
  }
  for(std::list<sAttribute>::iterator A = TheAttributes.begin()
                                    ; A != TheAttributes.end()  ; A++ ){
    if( (A->IOStatus != sAttribute::cDeleted) && (A->Name() == attr_name) ){
      if( A->IOStatus == sAttribute::cNewborn ){
        TheAttributes.erase(A);
      }else{
        A->IOStatus = sAttribute::cDeleted;
        A->StatusTimeStamp = sTime();
      };
      AttrListIOStatus = cChanged;
      AttrListStatusTimeStamp = sTime();
      InformThePointers();
      break;
    }
  }
}


void  sFolderNode::DelAttr (sNodePtr  value_ptr)
{
  QMutexLocker                AttrListLocker(&AttrListMutex);
  if( AttrListIOStatus == cToBeLoaded ){
    Location()->Net->GetAttrList( TheAttributes , Location());
    AttrListIOStatus = cRefreshed;
    AttrListStatusTimeStamp = sTime();
  }
  for(std::list<sAttribute>::iterator A = TheAttributes.begin()
                                    ; A != TheAttributes.end()  ; A++ ){
    if(    (A->IOStatus != sAttribute::cDeleted)
        && (A->Name().Type() == sAttrName::cNameless)
        && (A->Value() == value_ptr) ){
      if( A->IOStatus == sAttribute::cNewborn ){
        TheAttributes.erase(A);
      }else{
        A->IOStatus = sAttribute::cDeleted;
        A->StatusTimeStamp = sTime();
      };
      AttrListIOStatus = cChanged;
      AttrListStatusTimeStamp = sTime();
      InformThePointers();
      break;
    }
  }
}


void  sFolderNode::DeleteAllAttributes ()
{
  bool                        SomethingDone = false;
  sTime                       When;
  QMutexLocker                AttrListLocker(&AttrListMutex);
  if( AttrListIOStatus == cToBeLoaded ){
    Location()->Net->GetAttrList( TheAttributes , Location());
    AttrListIOStatus = cRefreshed;
    AttrListStatusTimeStamp = sTime();
  }
  for(std::list<sAttribute>::iterator A = TheAttributes.begin()
                                    ; A != TheAttributes.end()  ; A++ ){
    if( A->IOStatus != sAttribute::cDeleted ){
      if( A->IOStatus == sAttribute::cNewborn ){
        std::list<sAttribute>::iterator AA = A;
        A--;
        TheAttributes.erase(AA);
      }else{
        A->IOStatus = sAttribute::cDeleted;
        A->StatusTimeStamp = sTime();
      };
      SomethingDone = true;
    }
  }
  if( SomethingDone ){
    AttrListIOStatus = cChanged;
    AttrListStatusTimeStamp = When;
    InformThePointers();
  }
}


void  sFolderNode::SetAttribute (sAttribute  attribute)
{
  if( attribute.NameType() == sAttrName::cNameless ){
    SetNamelessAttribute(attribute.Value());
  }else{ // Numeric || Atomic || Symbolic
    SetNamedAttribute(attribute);
  }
}


void  sFolderNode::SetNamelessAttribute (sNodePtr  attr_value)
{
  if( attr_value.IsNULL() ){
    //вообще говоря, идиотизм - надо бы в лог писать такие вещи
    //но запрещать не будем - может какой-нибудь алгоритм станет проще
  }else{
    bool                        Found = false;
    for( std::list<sAttribute>::const_iterator  A = TheAttributes.begin()
                                              ; A != TheAttributes.end()
                                              ; A++ ){
      if(    (A->IOStatus != sAttribute::cDeleted)
          && ( A->Value() == attr_value)              ){
        Found = true;
        break;
      }
    }
    if( ! Found ){
      sAttribute                NewA (sAttrName(),attr_value);
      AppendAttribute(NewA);
    }
  }
}


void  sFolderNode::SetNamedAttribute (sAttribute  attribute)
{
  std::list<sAttribute>::iterator  OldAttr = TheAttributes.begin();
  while( (OldAttr != TheAttributes.end())
      && (   (OldAttr->Name() != attribute.Name())
          || (OldAttr->IOStatus == sAttribute::cDeleted))  )
    OldAttr++;
  if( OldAttr != TheAttributes.end() ){
    if( attribute.Value() == OldAttr->Value() )
      return; //уже все сделано ранее (предотвращает бесконечную рекурсию?)

    if( OldAttr->IOStatus == sAttribute::cNewborn ) {
      TheAttributes.erase(OldAttr);
    }else{
      OldAttr->IOStatus = sAttribute::cDeleted;
      OldAttr->StatusTimeStamp = sTime();
    };
    AttrListIOStatus = cChanged;
    AttrListStatusTimeStamp = sTime();
    InformThePointers();
  }
  if( ! attribute.Value().IsNULL() ){
    AppendAttribute(attribute);
  }
}


sNodePtr  sFolderNode::AscribeAttribute (sAttribute  attribute)
{
  sNodePtr                    NewValue = attribute.Value();
  if( ! (NewValue.IsNULL()) ){
    //!@todo{Net}    NewValue=NewValue->Copy();
    if( attribute.NameType() == sAttrName::cNameless ){
      SetNamelessAttribute(NewValue);
    }else{// Numeric || Atomic || Symbolic
      sNodePtr                    OldValue;
      for( std::list<sAttribute>::const_iterator  A = TheAttributes.begin()
                                                ; A != TheAttributes.end()
                                                ; A++ ){
        if(    (A->IOStatus != sAttribute::cDeleted)
            && (A->Name() == attribute.Name())       )
          OldValue = A->Value();
      }
      if( OldValue.IsNULL() ){
        AppendAttribute(attribute);
      }else{
        //!@todo{Net} check type, merge if folders
        NewValue = OldValue;
      }
    }
  }
  return NewValue;
}


void  sFolderNode::AppendAttribute (sAttribute  attribute)
{
  TheAttributes.push_back(attribute);
  AttrListIOStatus = cChanged;
  AttrListStatusTimeStamp = sTime();
  attribute.Value()->Allocate(this,attribute.Name());
  InformThePointers();
}


void  sFolderNode::MergeAttrList (std::list<sAttribute> &  loaded_list)
{
  //!@todo{NetIO} Это что, патч?!  Как же оно работало?  Не нарывались?
  if( TheAttributes.size() != loaded_list.size() ){
    RANet::Log.Put(sLog::ProgramError,"Refreshing",
        (sString("sFolderNode::MergeAttrList: has ")
         << integer(TheAttributes.size()))
        +(sString(" attributes, loaded ") << integer(loaded_list.size()))
        );
  }
  TheAttributes = loaded_list;
}


bool  sFolderNode::AllocateAttributes ()
{
  bool                        OK = true;
  for( std::list<sAttribute>::iterator  A = TheAttributes.begin()
                                      ; A != TheAttributes.end()  ; ++A ){
    if( A->IOStatus != sAttribute::cDeleted ){
      if( ! A->Value().NodeLocation() ){
        bool  Allocated = A->Value()->Allocate(this,A->Name());
        OK = OK && Allocated;
      }
    }
  }
  return OK;
}

//------------------------------------------------------------------------------
} //namespace RA
