////////////////////////////////////////////////////////////////////////////////
/*! @file Functions.cpp
- Part of RANet - Research Assistant Net Library.
- Copyright(C) 2000-2014, Viktor E. Bursian, St.Petersburg, Russia.
                          Viktor.Bursian@mail.ioffe.ru
*///////////////////////////////////////////////////////////////////////////////
#include "Functions.h"
#include "Symbols.h"
#include "Excepts.h"
#include "TableDataFunction.h"
#include <math.h>
#include <limits>
namespace RA {
//------------------------------------------------------------------------------
//---------------------------------------------------------------- sFunction ---

sFunction::sFunction ()
    :sCurve()
    ,sMathValue()
{
}


sFunction::sFunction (std::istream & a_stream ,rcsVersion  version)
    :sCurve(a_stream,version)
    ,sMathValue(a_stream,version)
{
}


void  sFunction::Store (std::ostream & a_stream) const
{
  sCurve::Store(a_stream);
  sMathValue::Store(a_stream);
}


sString  sFunction::Text (eTextFormat        format
                         ,eTextDetalization  /*detalization*/) const
{
  sString                     T;
  if( format == HTML ){
    psCurveAppearance           A = dynamic_cast<psCurveAppearance>
                                                          (TheAppearance);
    if( A )
      T = sString("<font color=") + A->LineColor.HexARGB() + ">";
    T += Symbol::CurveIcon;
    if( A )
      T +=  "</font>";
  }
//  if( detalization > Laconic ){
//    T += sString(" [") + TheMultiplier.Y.Units().Text(F)
//        +sString("(") + TheMultiplier.X.Units().Text(F) + sString(")]");
//  }
  return T;
}

//------------------------------------------------------------ sDataFunction ---

sDataFunction::sDataFunction ()
    :sFunction()
    ,TheMultiplier()
    ,RecalculateBoundaries(true)
{
  TheMultiplier.X = sPhysValue(real_nan);
  TheMultiplier.Y = sPhysValue(real_nan);
}


sDataFunction::sDataFunction (rcsPhysPair  multiplier)
    :sFunction()
    ,TheMultiplier()
    ,RecalculateBoundaries(true)
{
  TheMultiplier.X = multiplier.X;
  TheMultiplier.Y = multiplier.Y;
}


psMathValue  sDataFunction::Replica () const
{
  psTableDataFunction         R = new sTableDataFunction(TheMultiplier
                                                        ,NumberOfPoints());
  sCurve::psTracer            T = NULL;
  int                         K = 0;
  T = GetCurveTracer(sScales(sScale(Multiplier().X)
                            ,sScale(Multiplier().Y)));
  while( T->NextPoint() ){
    R->SetValue(K,T->Point);
    K++;
  }
  if( T ){ delete T;  T=NULL; };
  return R;
}


sDataFunction::sDataFunction (std::istream & a_stream ,rcsVersion  version)
    :sFunction(a_stream,version)
    ,RecalculateBoundaries(true)
{
  if( version <= sVersion(0,9) ){
  }else{
    sPhysPair                 M(a_stream,version);
    TheMultiplier = M;
  }
}


void  sDataFunction::Store (std::ostream & a_stream) const
{
  sFunction::Store(a_stream);
  TheMultiplier.Store(a_stream);
}


sBoundaries  sDataFunction::Boundaries ()
{
  if( RecalculateBoundaries ){
    LastBoundaries=CalculateBoundaries();
    RecalculateBoundaries=false;
  };
  return LastBoundaries;
}


sBoundaries  sDataFunction::Boundaries () const
{
  if( RecalculateBoundaries )
    return CalculateBoundaries();
  return LastBoundaries;
}


sString  sDataFunction::Text (eTextFormat        format
                             ,eTextDetalization  detalization)
{
  sString                     T(sFunction::Text(format,detalization));
  if( detalization > Laconic ){
    T += sString(" [") + TheMultiplier.Y.Units().Text(format)
        +sString("(") + TheMultiplier.X.Units().Text(format) + sString(")]");
  }
  return T;
}


sString  sDataFunction::Text (eTextFormat        format
                             ,eTextDetalization  detalization) const
{
  sString                     T(sFunction::Text(format,detalization));
  if( detalization > Laconic ){
    T += sString(" [") + TheMultiplier.Y.Units().Text(format)
        +sString("(") + TheMultiplier.X.Units().Text(format) + sString(")]");
  }
  return T;
}


int  sDataFunction::NumberOfPoints () const
{
  sCurve::psTracer            T = NULL;
  int                         N = 0;
  T = GetCurveTracer(sScales(sScale(Multiplier().X)
                            ,sScale(Multiplier().Y)));
  while( T->NextPoint() )
    N++;
  if( T ){ delete T;  T=NULL; };
  return N;
}


sPhysRange  sDataFunction::AbscissaRange () const
{
  return Boundaries().X;
}


sUnits  sDataFunction::OrdinateUnits () const
{
  return TheMultiplier.Y.Units();
}


sPhysValue  sDataFunction::Value (rcsPhysValue  x) const
{
  sCurve::psTracer            T=GetCurveTracer(sScales(sScale(Multiplier().X)
                                                      ,sScale(Multiplier().Y)));
  sPhysValue                  V(real_nan,Multiplier().Y.Units());
  try{
    real                        X = (x/Multiplier().X).ToReal();
    sPoint                      P1;
    sPoint                      P2;
    bool                        FoundMatch = false;
    bool                        Found1 = false;
    bool                        Found2 = false;
    while( T->NextPoint() && ! FoundMatch ){
      if( T->Point.X == X ){
        V = T->Point.Y * Multiplier().Y;
        FoundMatch = true;
      }else if( T->Point.X < X ){
        if( ! Found1 ){
          P1 = T->Point;
          Found1 = true;
        }else if( P1.X < T->Point.X ){
          P1 = T->Point;
        }
      }else{
        if( ! Found2 ){
          P2 = T->Point;
          Found2 = true;
        }else if( P2.X > T->Point.X ){
          P2 = T->Point;
        }
      }
    }
    if( ! FoundMatch && Found1 && Found2 )
      V = ( P1.Y + (P2.Y-P1.Y)*(X-P1.X)/(P2.X-P1.X) ) * Multiplier().Y;
  }catch( rxRAlgebra ){
  }
  if( T ){ delete T;  T=NULL; };
  return V;
}


sBoundaries  sDataFunction::CalculateBoundaries () const
{
  sRectangle                  B;
  sCurve::psTracer            T=GetCurveTracer(sScales(sScale(Multiplier().X)
                                                      ,sScale(Multiplier().Y)));
  real                        Xmin;
  real                        Xmax;
  real                        Ymin;
  real                        Ymax;

  B = sRectangle( sRange(real_nan,real_nan)  , sRange(real_nan,real_nan) );
  if( T->NextPoint() ){
    Xmin = Xmax = T->Point.X;
    Ymin = Ymax = T->Point.Y;
    while( T->NextPoint() ){
      if( T->Point.X < Xmin ){
        Xmin=T->Point.X;
      }else if( Xmax < T->Point.X ){
        Xmax=T->Point.X;
      };
      if( T->Point.Y < Ymin ){
        Ymin=T->Point.Y;
      }else if( Ymax < T->Point.Y ){
        Ymax=T->Point.Y;
      };
    };
    B = sRectangle( sRange(Xmin,Xmax) , sRange(Ymin,Ymax) );
  };
/*
  while( T->NextLineSegment() ){
    if( T->LineSegment.Start.X < B.X.From ){
      B.X.From=T->LineSegment.Start.X;
    }else if( T->LineSegment.Start.X > B.X.To ){
      B.X.To=T->LineSegment.Start.X;
    };
    if( T->LineSegment.Start.Y < B.Y.From ){
      B.Y.From=T->LineSegment.Start.Y;
    }else if( T->LineSegment.Start.Y > B.Y.To ){
      B.Y.To=T->LineSegment.Start.Y;
    };
    if( T->LineSegment.End.X < B.X.From ){
      B.X.From=T->LineSegment.End.X;
    }else if( T->LineSegment.End.X > B.X.To ){
      B.X.To=T->LineSegment.End.X;
    };
    if( T->LineSegment.End.Y < B.Y.From ){
      B.Y.From=T->LineSegment.End.Y;
    }else if( T->LineSegment.End.Y > B.Y.To ){
      B.Y.To=T->LineSegment.End.Y;
    };
  };
*/

//  RANet::Log.Put(sLog::Debug,"Graph"
//      ,sString("CalculateBoundaries: ")
//      +" MX="+Multiplier().X.Text(Plain,Verbose)
//      +" MY="+Multiplier().Y.Text(Plain,Verbose)
//      +(sString(" X.From=") << B.X.From)
//      +(sString(" X.To=") << B.X.To)
//      +(sString(" Y.From=") << B.Y.From)
//      +(sString(" Y.To=") << B.Y.To)
//      );

  if( T ){ delete T;  T=NULL; };

  return sBoundaries(sPhysRange(B.X.From*Multiplier().X
                               ,B.X.To*Multiplier().X   )
                    ,sPhysRange(B.Y.From*Multiplier().Y
                               ,B.Y.To*Multiplier().Y   )
                    );
}

//-------------------------------------------------------- sAnalyticFunction ---

  //......................................... sAnalyticFunction::sTracer ...

sAnalyticFunction::sTracer::sTracer (pcsAnalyticFunction  host_curve
                                    ,rcsScales            scales    )
    :TheFunction(host_curve)
    ,Scales(scales)
    ,PointNo(0)
    ,LineSegmentNo(1)
{
  try{
    (sPhysValue(1,host_curve->AbscissaRange().Units()) / Scales.X.Unit())
                                                                    .ToReal();
    (sPhysValue(1,host_curve->OrdinateUnits()) / Scales.Y.Unit()).ToReal();
    Step = int(scales.X.Size() / 100);
    if( Step < 1 )  Step = 1;
    LastPointNo = int(scales.X.Size() / Step)+1;
  }catch(rxRAlgebra){
    LastPointNo = -1;
  }
}


bool  sAnalyticFunction::sTracer::NextPoint ()
{
  if( PointNo < LastPointNo ){
    Point.X= Scales.X.Tr2real( PointNo*Step );
    Point.Y= Scales.Y.Tr2real( TheFunction->Value( Scales.X.T(PointNo*Step) ) );
    PointNo++;
    return true;
  }else{
    return false;
  }
}


bool  sAnalyticFunction::sTracer::NextLineSegment ()
{
  if( LineSegmentNo < LastPointNo ){
    if( LineSegmentNo == 1 ){
      LineSegment.Start.X = Scales.X.Tr2real( 0 );
      LineSegment.Start.Y = Scales.Y.Tr2real( TheFunction->Value(
                                                              Scales.X.T(0) ) );
    }else{
      LineSegment.Start.X = LineSegment.End.X;
      LineSegment.Start.Y = LineSegment.End.Y;
    }
    LineSegment.End.X = Scales.X.Tr2real( LineSegmentNo*Step );
    LineSegment.End.Y = Scales.Y.Tr2real( TheFunction->Value(
                                             Scales.X.T(LineSegmentNo*Step) ) );
    LineSegmentNo++;
    return true;
  }else{
    return false;
  }
}

  //........................................................................

sAnalyticFunction::sAnalyticFunction ()
    :sFunction()
{
}


sAnalyticFunction::sAnalyticFunction (std::istream &  a_stream
                                     ,rcsVersion      version )
    :sFunction(a_stream,version)
{
}


void  sAnalyticFunction::Store (std::ostream & a_stream) const
{
  sFunction::Store(a_stream);
}


sCurve::psTracer  sAnalyticFunction::GetCurveTracer (rcsScales  scales) const
{
  return new sAnalyticFunction::sTracer(this,scales);
}


//------------------------------------------------------------------------------
} //namespace RA




