////////////////////////////////////////////////////////////////////////////////
/*! @file Graphicals.cpp   Графические примитивы.
- Part of RANet - Research Assistant Net Library.
- Copyright(C) 2000-2015, Viktor E. Bursian, St.Petersburg, Russia.
                          Viktor_dot_Bursian_at_mail_dot_ioffe_dot_ru
*///////////////////////////////////////////////////////////////////////////////
#include "Graphicals.h"
#include "Excepts.h"
#include "Log.h"
#include <math.h>
#include <limits>
namespace RA {
//------------------------------------------------------------------------------

const
sLineAppearance::eLineStyle   sLineAppearance::DefaultLineStyle
                                               = sLineAppearance::cLineSolid;
const byte                    sLineAppearance::DefaultLineThickness = 1;
const byte                    sLineAppearance::MinLineThickness = 1;
const byte                    sLineAppearance::MaxLineThickness = 9;

const byte                    sSpotAppearance::DefaultContourThickness = 1;
const byte                    sSpotAppearance::MinContourThickness = 1;
const byte                    sSpotAppearance::MaxContourThickness = 3;

const
sPointAppearance::eShape      sPointAppearance::DefaultShape
                                                = sPointAppearance::cCircle;
const byte                    sPointAppearance::DefaultSize  = 6;
const byte                    sPointAppearance::MinSize  = 1;
const byte                    sPointAppearance::MaxSize  = 50;

//---------------------------------------------------------------- sPhysPair ---

sPhysPair::sPhysPair (std::istream & a_stream ,rcsVersion  version)
    :sStorable(a_stream,version)
    ,X(a_stream,version)
    ,Y(a_stream,version)
{
}


void  sPhysPair::Store (std::ostream & a_stream) const
{
  sStorable::Store(a_stream);
  X.Store(a_stream);
  Y.Store(a_stream);
}

//-------------------------------------------------------------- sBoundaries ---

sBoundaries::sBoundaries (std::istream & a_stream ,rcsVersion  version)
    :sStorable(a_stream,version)
    ,X(a_stream,version)
    ,Y(a_stream,version)
{
}


void  sBoundaries::Store (std::ostream & a_stream) const
{
  sStorable::Store(a_stream);
  X.Store(a_stream);
  Y.Store(a_stream);
}


/*!
@return the left operand, which is (if necessary) expanded in order
        to contain also the right one.

If the units aren't comparable, then xRAlgebra is thrown.

@note There is a specific behaviour, when one of operands IsNaN().

*/
rsBoundaries  sBoundaries::operator |= (rcsBoundaries  boundaries)
{
  X |= boundaries.X;
  Y |= boundaries.Y;
  return *this;
}


/*!
@return the left operand, which is (if necessary) expanded in order
        to contain also the value of the right operand.

If the units aren't comparable, then xRAlgebra is thrown.

*/
rsBoundaries  sBoundaries::operator |= (rcsPhysPair  point)
{
  X |= point.X;
  Y |= point.Y;
  return *this;
}

//------------------------------------------------------------------- sScale ---

sScale::sScale (bool  upside_down)
    :TheUnit()
    ,TheRange()
    ,TheSize(10)
    ,Offset(0)
    ,UpsideDown(upside_down)
    ,NoTr(false)
{
}


sScale::sScale (rcsPhysValue  unit)
    :TheUnit(unit)
    ,TheRange()
    ,TheSize(10)
    ,Offset(0)
    ,UpsideDown(false)
    ,NoTr(true)
{
}


sScale::sScale (rcsScale  scale)
    :TheUnit(scale.TheUnit)
    ,TheRange(scale.TheRange)
    ,TheSize(scale.TheSize)
    ,Offset(scale.Offset)
    ,UpsideDown(scale.UpsideDown)
    ,NoTr(scale.NoTr)
{
}


rsScale  sScale::operator = (rcsScale  scale)
{
  TheUnit = scale.TheUnit;
  TheRange = scale.TheRange;
  TheSize = scale.TheSize;
  Offset = scale.Offset;
  UpsideDown = scale.UpsideDown;
  NoTr = scale.NoTr;
  return *this;
}


void  sScale::SetRange (sPhysRange  range)
{
//  RANet::Log.Put(sLog::Debug,"Graph",sString("SetRange: ")+range.Text());
  range.Normalize();
//  RANet::Log.Put(sLog::Debug,"Graph",sString("normalized: ")+range.Text());
//  RANet::Log.Put(sLog::Debug,"Graph",sString("size: ")+range.Size().Text());
  real                        From = range.FromMantissa();
  real                        To = range.ToMantissa();
  real                        Precision = std::max(range.PrecisionMantissa()
//                                            ,power10(-5));
                                            ,power10(-sPhysValue::MaxDigits-1));
  if( To-From < 2*Precision ){
    RANet::Log.Put(sLog::Debug,"Graph",sString("small size: ")
        +" To-From="+sString::FromReal(To-From,12)
        +" P="+sString::FromReal(Precision,12));
    To =   From + Precision; //sic!
    From = From - Precision;
  }
  if( isinf(From) || isnan(From) ){
    if( isinf(To) || isnan(To) ){
      From = 0.0;
      To = 10.0;
    }else{
      From = To - 10.0;
    }
  }else if( isinf(To) || isnan(To) ){
    To = From + 10.0;
  }
  TheRange = sRange( From , To );
  TheUnit = sPhysValue(range.Units()) << range.Order();
  if( ! TheUnit.IsPowerOfTen() ){
    RANet::Log.Put(sLog::Debug,"Graph",sString("not a power of ten: ")+TheUnit.Text());
  }
}


void  sScale::SetRange (rcsRange  range)
{
  SetRange( sPhysRange( range.From * TheUnit
                      , range.To * TheUnit   ));
}


void  sScale::SetRange (int  from ,int  to)
{
  if( ! UpsideDown ){
    SetRange(sPhysRange(TheUnit * (TheRange.From + (from+Offset)*PixelValue())
                       ,TheUnit * (TheRange.From + (to  +Offset)*PixelValue())
                       ));
  }else{
    SetRange(sPhysRange(TheUnit * (TheRange.To - (from+Offset)*PixelValue())
                       ,TheUnit * (TheRange.To - (to  +Offset)*PixelValue())
                       ));
  }
}


void  sScale::ShiftRange (int  shift)
{
  if( ! UpsideDown ){
    SetRange(sPhysRange(TheUnit * (TheRange.From + shift*PixelValue())
                       ,TheUnit * (TheRange.To   + shift*PixelValue())
                       ));
  }else{
    SetRange(sPhysRange(TheUnit * (TheRange.From - shift*PixelValue())
                       ,TheUnit * (TheRange.To   - shift*PixelValue())
                       ));
  }
}


void  sScale::SetSize (int  size)
{
  TheSize = size;
}


int  sScale::T (real  value) const
{
  int                         P;
  try{
    P = Round((value-TheRange.From)/(TheRange.To-TheRange.From)*TheSize);
    if( ! UpsideDown ){
      P -= Offset;
    }else{
      P += Offset;
    };
  }catch(...){
    P = 0;
  };
  if( UpsideDown )  P = TheSize - P;
  return P;
}


int  sScale::T (sPhysValue  value) const
{
  return T( (value/TheUnit).ToReal() );
}


sPhysValue  sScale::T (int  position) const
{
  if( ! UpsideDown ){
    return TheUnit
         * sPhysValue( TheRange.From + (position + Offset) * PixelValue()
                     , 0.5 * PixelValue() );
  }else{
    return TheUnit
         * sPhysValue( TheRange.To - (position + Offset) * PixelValue()
                     , 0.5 * PixelValue() );
  };
}


real  sScale::Tr2real (int  position) const
{
  if( ! UpsideDown ){
    position += Offset;
  }else{
    position -= Offset;
  };
  return TheRange.From + position*(TheRange.To-TheRange.From)/TheSize;
}


real  sScale::Tr2real (sPhysValue  value) const
{
  return (value/TheUnit).ToReal();
}


//-------------------------------------------------------------- sAppearance ---

sAppearance::~sAppearance ()
{
}


sAppearance::sAppearance (bool  hidden)
    :Hidden(hidden)
{
}


sAppearance::sAppearance (std::istream & a_stream ,rcsVersion  version)
    :sStorable(a_stream,version)
{
  a_stream.READ_INTO(Hidden);
}


void  sAppearance::Store (std::ostream & a_stream) const
{
  sStorable::Store(a_stream);
  a_stream.WRITE_FROM(Hidden);
}


// /*! Creates a copy in the heap (the caller takes the ownership).
//*/
//psAppearance  sAppearance::Replica () const
//{
//  return (new sAppearance)->AssignFrom(this);
//}


// /*! Gets what is possible from another sAppearance instance, returns @a this.
//*/
//psAppearance  sAppearance::AssignFrom (pcsAppearance  other)
//{
//  Hidden = other->Hidden;
//  return this;
//}

//---------------------------------------------------------- sLineAppearance ---

sLineAppearance::~sLineAppearance ()
{
}


sLineAppearance::sLineAppearance ()
    :LineColor()
    ,LineThickness(DefaultLineThickness)
    ,LineStyle(DefaultLineStyle)
{
}


sLineAppearance::sLineAppearance (rcsColor  color)
    :LineColor(color)
    ,LineThickness(DefaultLineThickness)
    ,LineStyle(DefaultLineStyle)
{
}


sLineAppearance::sLineAppearance (rcsColor    color
                                 ,byte        line_thickness
                                 ,eLineStyle  line_style    )
    :LineColor(color)
    ,LineThickness( line_thickness > MaxLineThickness
                  ? MaxLineThickness
                  : ( line_thickness < MinLineThickness
                    ? MinLineThickness
                    : line_thickness
                    )
                  )
    ,LineStyle(line_style)
{
}


sLineAppearance::sLineAppearance (std::istream & a_stream ,rcsVersion  version)
    :sStorable(a_stream,version)
    ,sAppearance(a_stream,version)
    ,LineColor(a_stream,version)
{
  a_stream.READ_INTO(LineThickness);
  a_stream.READ_INTO(LineStyle);
}

void  sLineAppearance::Store (std::ostream & a_stream) const
{
  sAppearance::Store(a_stream);
  LineColor.Store(a_stream);
  a_stream.WRITE_FROM(LineThickness);
  a_stream.WRITE_FROM(LineStyle);
}


//psAppearance  sLineAppearance::AssignFrom (pcsAppearance  other)
//{
//  sAppearance::AssignFrom(other);
//  if( pcsLineAppearance A = dynamic_cast<pcsLineAppearance>(other) ){
//    LineColor = A->LineColor;
//    LineThickness = A->LineThickness;
//    LineStyle = A->LineStyle;
//  }
//  return this;
//}

//---------------------------------------------------------- sSpotAppearance ---

sSpotAppearance::sSpotAppearance ()
    :FillColor()
    ,ContourColor()
    ,Filled(true)
    ,Contoured(false)
    ,ContourThickness(DefaultContourThickness)
{
}


sSpotAppearance::sSpotAppearance (rcsColor  color)
    :FillColor(color)
    ,ContourColor(color)
    ,Filled(true)
    ,Contoured(false)
    ,ContourThickness(DefaultContourThickness)
{
}


sSpotAppearance::sSpotAppearance (std::istream &  a_stream
                                 ,rcsVersion      version)
    :sStorable(a_stream,version)
    ,sAppearance(a_stream,version)
    ,FillColor(a_stream,version)
    ,ContourColor(a_stream,version)
{
  a_stream.READ_INTO(Filled);
  a_stream.READ_INTO(Contoured);
  a_stream.READ_INTO(ContourThickness);
}

void  sSpotAppearance::Store (std::ostream & a_stream) const
{
  sAppearance::Store(a_stream);
  FillColor.Store(a_stream);
  ContourColor.Store(a_stream);
  a_stream.WRITE_FROM(Filled);
  a_stream.WRITE_FROM(Contoured);
  a_stream.WRITE_FROM(ContourThickness);
}


//psAppearance  sSpotAppearance::AssignFrom (pcsAppearance  other)
//{
//  sAppearance::AssignFrom(other);
//  if( pcsSpotAppearance A = dynamic_cast<pcsSpotAppearance>(other) ){
//    FillColor = A->FillColor;
//    ContourColor = A->ContourColor;
//    Filled = A->Filled;
//    Contoured = A->Contoured;
//    ContourThickness = A->ContourThickness;
//  }
//  return this;
//}

//--------------------------------------------------------- sPointAppearance ---

sPointAppearance::sPointAppearance ()
    :sSpotAppearance()
    ,PointSize(DefaultSize)
    ,PointShape(DefaultShape)
{
}


sPointAppearance::sPointAppearance (rcsColor  color)
    :sSpotAppearance(color)
    ,PointSize(DefaultSize)
    ,PointShape(DefaultShape)
{
}


sPointAppearance::sPointAppearance (rcsColor  color
                                   ,byte      point_size
                                   ,eShape    point_shape)
    :sSpotAppearance(color)
    ,PointSize( point_size > MaxSize
              ? MaxSize
              : ( point_size < MinSize
                ? MinSize
                : point_size
                )
              )
    ,PointShape(point_shape)
{
}


sPointAppearance::sPointAppearance (std::istream &  a_stream
                                   ,rcsVersion      version)
    :sStorable(a_stream,version)
    ,sAppearance(a_stream,version)
    ,sSpotAppearance(a_stream,version)
{
  a_stream.READ_INTO(PointSize);
  a_stream.READ_INTO(PointShape);
}

void  sPointAppearance::Store (std::ostream & a_stream) const
{
  sSpotAppearance::Store(a_stream);
  a_stream.WRITE_FROM(PointSize);
  a_stream.WRITE_FROM(PointShape);
}


//psAppearance  sPointAppearance::AssignFrom (pcsAppearance  other)
//{
//  sSpotAppearance::AssignFrom(other);
//  if( pcsPointAppearance A = dynamic_cast<pcsPointAppearance>(other) ){
//    PointSize = A->PointSize;
//    PointShape = A->PointShape;
//  }
//  return this;
//}

//------------------------------------------------------------- sGraphObject ---

sGraphObject::sGraphObject (std::istream & a_stream ,rcsVersion  version)
    :sStorable(a_stream,version)
    ,TheAppearance(dynamic_cast<psAppearance>(PolymorphicLoad(a_stream)))
{
}


void  sGraphObject::Store (std::ostream & a_stream) const
{
  sStorable::Store(a_stream);
  StorePolymorphic(a_stream,TheAppearance);
}


sGraphObject::~sGraphObject()
{
  if( TheAppearance )
    delete TheAppearance;
}


void  sGraphObject::SetAppearance (psAppearance  appearance)
{
  psAppearance  Old = TheAppearance;
  TheAppearance = appearance;
  if( Old )
    delete Old;
}


sBoundaries  sGraphObject::Boundaries ()
{
  return sBoundaries();
}


sBoundaries  sGraphObject::Boundaries () const
{
  return sBoundaries();
}


int  sGraphObject::Distance (rcsIntPoint  /*position*/
                            ,rcsScales    /*scales*/)
{
  return std::numeric_limits<int>::max();
}

//------------------------------------------------------------------------------
} //namespace RA




