////////////////////////////////////////////////////////////////////////////////
/*! @file Net.cpp   Абстракция сети RANet и примитивнейшая реализация.
- Part of RANet - Research Assistant Net Library (based on ANSI C++).
- Copyright(C) 2006-2010, Viktor E. Bursian, St.Petersburg, Russia.
                     Viktor.Bursian@mail.ioffe.ru
*///////////////////////////////////////////////////////////////////////////////
#include "Net.h"
#include "Nodes.h"
#include "Attribute.h"
#include "FolderNode.h"
#include "Log.h"
namespace RA {
//------------------------------------------------------------------------------

const int          sNet::BackgroundLockingTimeout = 10;//ms

//--------------------------------------------------------------------- sNet ---

sNet::sNet ()
{
}


void  sNet::Save (psNode  node)
{
  sTime                       When;
  if( node->BodyMutex.tryLock(BackgroundLockingTimeout) ){
    try{
      if( node->IOStatus == sNode::cNewborn ){
        if( node->Allocate() )
          if( StoreNewNode(node,When) ){
            node->IOStatus = sNode::cSynchronized;
            node->StatusTimeStamp = When;
          }
      }else if( node->IOStatus == sNode::cEdited ){
        if( UpdateNode(node,When) ){
          node->IOStatus = sNode::cSynchronized;
          node->StatusTimeStamp = When;
        }
      }
    }catch(...){
      node->BodyMutex.unlock();
      throw;
    }
    node->BodyMutex.unlock();
  }
}





/*!

*/
void  sNet::SaveAttributes (psNode  node)
{
  if( node->Allocate() ){
    psFolderNode                FolderNode( dynamic_cast<psFolderNode>(node) );
    ASSERT( FolderNode );
    if( FolderNode->AttrListMutex.tryLock(BackgroundLockingTimeout) ){
      try{
        if( FolderNode->AllocateAttributes()  ){
          bool                        AllDone = true;
          sTime                       When;
          //removing from DB, then deleting the attributes, marked as cDeleted:
          {
            std::list<sAttribute>::iterator  A;
            while(  ( A = find_if( FolderNode->TheAttributes.begin()
                                 , FolderNode->TheAttributes.end()
                                 , sAttribute::Deleted )
                    , A != FolderNode->TheAttributes.end()            )  ){
              bool                        Removed = false;
              if( A->Name().Type() == sAttrName::cNameless ){
                Removed = RemoveAttribute(FolderNode->TheLocation
                                         ,A->Value().NodeLocation());
              }else{
                Removed = RemoveAttribute(FolderNode->TheLocation,A->Name());
              }
              if( Removed )
                FolderNode->TheAttributes.erase(A);
              AllDone = AllDone && Removed;
            }
          }
          //storing the newborn attributes into DB:
          {
            for( std::list<sAttribute>::iterator
                                 A = FolderNode->TheAttributes.begin()
                               ; A != FolderNode->TheAttributes.end()  ; ++A ){
              if( A->IOStatus == sAttribute::cNewborn ){
                bool                        Stored = false;
                Stored =  FolderNode->TheLocation
                    && A->Value().NodeLocation()
                    && StoreNewAttribute( FolderNode->TheLocation
                                        , A->Name()
                                        , A->Value().NodeLocation()
                                        , When );
                if( Stored ){
                  A->IOStatus = sAttribute::cSynchronized;
                  A->StatusTimeStamp = When;
                }
                AllDone = AllDone && Stored;
              }
            }
          }
          if( AllDone ){
            FolderNode->AttrListIOStatus = sFolderNode::cSaved;
            FolderNode->AttrListStatusTimeStamp = When;
          }
        }
      }catch(...){
        FolderNode->AttrListMutex.unlock();
        throw;
      }
      FolderNode->AttrListMutex.unlock();
    }
  }
}


void  sNet::Refresh (psNode  node)
{
  sTime                       When;
  psNode                      NodeCopy;
  if( node->BodyMutex.tryLock(BackgroundLockingTimeout) ){
    try{
      if( node->BodyHasToBeRefreshed() ){
        if( node->LocationMutex.tryLock(BackgroundLockingTimeout) ){
          try{
            //---
            if( GetNodeStorageTime(node->TheLocation,When) ){
              if( node->StatusTimeStamp < When ) {
                When = sTime();
                NodeCopy = LoadNode(node->TheLocation);
                if( NodeCopy ){
                  node->CopyFrom(NodeCopy);
                  node->IOStatus = sNode::cSynchronized;
                  node->StatusTimeStamp = When;
                  delete NodeCopy;
                  RANet::Log.Put(sLog::Debug,"Refreshing"
                                ,"sNet::Refresh(psNode): reloaded");
                }
              }else{
                node->StatusTimeStamp = sTime();
              }
            }
            //---
          }catch(...){
            node->LocationMutex.unlock();
            throw;
          }
          node->LocationMutex.unlock();
        }
      }
    }catch(...){
      node->BodyMutex.unlock();
      throw;
    }
    node->BodyMutex.unlock();
  }
}


void  sNet::RefreshAttributes (psNode  node)
{
  psFolderNode                FolderNode( dynamic_cast<psFolderNode>(node) );
  ASSERT( FolderNode );
  if( FolderNode->AttrListMutex.tryLock(BackgroundLockingTimeout) ){
    try{
      if( FolderNode->AttrsHaveToBeRefreshed() ){
        sTime                       When;
        std::list<sAttribute>       StoredAttributes;
        GetAttrList( StoredAttributes , FolderNode->Location());
        FolderNode->MergeAttrList(StoredAttributes);
        FolderNode->AttrListIOStatus = sFolderNode::cRefreshed;
        FolderNode->AttrListStatusTimeStamp = When;
      }
    }catch(...){
      FolderNode->AttrListMutex.unlock();
      throw;
    }
    FolderNode->AttrListMutex.unlock();
  }
}


pcsNodeLocation  sNet::GetLocation (pcsNode  node)
{
  return node->TheLocation;
}


void sNet::SetLocation (psNode node, pcsNodeLocation location)
{
  //!@todo{Net} mutex?
  node->TheLocation = location->Replica();
  node->PrimaryHostForDelayedAllocation.SetToNULL();
  node->AttrNameForDelayedAllocation = sAttrName();
}

//------------------------------------------------------------------------------
} //namespace RA
