////////////////////////////////////////////////////////////////////////////////
/*! @file Nodes.cpp   Абстракция объекта сети.
- Part of RANet - Research Assistant Net Library (based on ANSI C++).
- Copyright(C) 1994-2010, Viktor E. Bursian, St.Petersburg, Russia.
                     Viktor.Bursian@mail.ioffe.ru
*///////////////////////////////////////////////////////////////////////////////
#include "Nodes.h"
#include "Net.h"
#include "Attribute.h"
#include "NetExceptions.h"

#include "Trace.h"
#include "DB_SQL_Net.h"

#include <QMutexLocker>
namespace RA {
//------------------------------------------------------------------------------

#ifdef RANET_DEBUG
long int                      sNode::Counter(0);
#endif

float                         sNode::RefreshingPeriod = 10.00;/*s*/
size_t                        sNode::CutOffLengthForLaconic = 40;
//DEFINE_CLASS_TAG( sUndoNodeDirectAction )
//DEFINE_CLASS_TAG( sUndoNodeCorrectAction )
//DEFINE_CLASS_TAG( sRedoNetOperation )

//-------------------------------------------------------------------- sNode ---
/*! @class sNode


@par sNode instance lifecycle:
-  (bearing constructor)--> cNewborn
-  cNewborn ---(MarkAsEdited)--> cNewborn
-  cNewborn ---(Save)--> cSynchronized
-  (loading constructor)--> cSynchronized
-  cSynchronized ---(MarkAsEdited)--> cEdited
-  cEdited ---(Save)--> cSynchronized


@todo{--Qt} see mutex in C++11 or Boost.Interprocess for mutexes (which,
            by the way, have try_to_lock and recursive features)

*/



sNode::~sNode ()
{
  IF_DEBUG(Counter--;)
  ABORT_IF_NOT( Pointers.empty() );
}


sNode::sNode ()
    :TheLocation(NULL)
    ,IOStatus(cNewborn)
    ,StatusTimeStamp()
{
  IF_DEBUG(Counter++;)
}


sNode::sNode (std::istream &  stream ,rcsVersion  version)
    :sStorable(stream,version)
    ,TheLocation(NULL)
    ,IOStatus(cSynchronized)
    ,StatusTimeStamp()
{
  IF_DEBUG(Counter++;)
}


void  sNode::Store (std::ostream &  stream) const
{
  sStorable::Store(stream);
}


bool  sNode::BodyHasToBeSaved () const
{
  return (IOStatus==cNewborn) || (IOStatus==cEdited);
}


bool  sNode::BodyHasToBeRefreshed () const
{
  return (IOStatus==cSynchronized)
      && (sTime()-StatusTimeStamp > RefreshingPeriod);
}


bool  sNode::AttrsHaveToBeSaved () const
{
  return false;
}


bool  sNode::AttrsHaveToBeRefreshed () const
{
  return false;
}


pcsNodeLocation  sNode::Location ()
{
  QMutexLocker                LocationLocker(&LocationMutex);
  return TheLocation;
}


void  sNode::MarkAsEdited ()
{
  if( IOStatus != cNewborn ){
    IOStatus = cEdited;
    StatusTimeStamp = sTime();
  }
  InformThePointers();
}


void  sNode::InformThePointers ()
{
  QMutexLocker                PtrListLocker(&PtrListMutex);
  for( tListOfPtrs::const_iterator  P = Pointers.begin()
                                  ; P != Pointers.end()   ; P++ ){
    (*P)->NodeIsRecentlyEditedFlag = true;
  }
}


void sNode::CopyFrom (psNode)
{
}


sString  sNode::Text (eTextFormat        /*F*/
                     ,eTextDetalization  /*D*/)
{
  return sString();
}


sNodePtr  sNode::GetAttr (sAttrName)
{
  return sNodePtr();
}


void  sNode::SetAttr (sAttribute)
{
  throw xNetInvalidOp("SetAttr on a terminal node",__FILE__,__LINE__);
}


sNodePtr  sNode::AscribeAttr (sAttribute)
{
  return sNodePtr();
}


void  sNode::DelAttr (sAttrName)
{
}


void  sNode::DelAttr (sNodePtr)
{
}


void  sNode::DeleteAllAttributes ()
{
}


bool  sNode::Allocate ()
{
  QMutexLocker                LocationLocker(&LocationMutex);
  if( ! TheLocation ){
    if( ! PrimaryHostForDelayedAllocation.IsNULL()
        && PrimaryHostForDelayedAllocation->TheLocation ){
      TheLocation = PrimaryHostForDelayedAllocation->TheLocation
                              ->LocationForAttr(AttrNameForDelayedAllocation);
      TheLocation->Net->RegisterNode(TheLocation,this);
      if( TrueType()->Name() == "sDataArrayRecord" ){
        TRACE(sString("Allocated (1)")
            +sDB_SQL_Net::psNodeUUID(TheLocation)->FormattedNodeId());
      }
    }
  }
  return bool( TheLocation );
}


bool  sNode::Allocate (psNode            host
                      ,rcsAttrName  attr_name)
{
  QMutexLocker                Locker(&LocationMutex);
  if( ! TheLocation ){
    if( ! PrimaryHostForDelayedAllocation.IsNULL()
        && PrimaryHostForDelayedAllocation->TheLocation ){
      TheLocation = PrimaryHostForDelayedAllocation->TheLocation
                              ->LocationForAttr(AttrNameForDelayedAllocation);
      PrimaryHostForDelayedAllocation.SetToNULL();
      AttrNameForDelayedAllocation = sAttrName();
      TheLocation->Net->RegisterNode(TheLocation,this);
      if( TrueType()->Name() == "sDataArrayRecord" ){
        TRACE(sString("Allocated (2)")
            +sDB_SQL_Net::psNodeUUID(TheLocation)->FormattedNodeId());
      }
    }else if( host->TheLocation ){
      TheLocation = host->TheLocation->LocationForAttr(attr_name);
      PrimaryHostForDelayedAllocation.SetToNULL();
      AttrNameForDelayedAllocation = sAttrName();
      TheLocation->Net->RegisterNode(TheLocation,this);
      if( TrueType()->Name() == "sDataArrayRecord" ){
        TRACE(sString("Allocated (3)")
            +sDB_SQL_Net::psNodeUUID(TheLocation)->FormattedNodeId());
      }
    }else/* if( PrimaryHostForDelayedAllocation.IsNULL()
          || (  (AttrNameForDelayedAllocation.Type()
                                            == sAttrName::cNameless)
             && (attr_name.Type() != sAttrName::cNameless)          ) )*/{
      PrimaryHostForDelayedAllocation = sNodePtr(host);
      AttrNameForDelayedAllocation = attr_name;
    }
  }
  return bool( TheLocation );
}


void  sNode::RegisterPointer (psNodePtr  a_ptr_to_me)
{
  QMutexLocker                Locker(&PtrListMutex);
  Pointers.push_front(a_ptr_to_me);
}


void  sNode::UnregisterPointer (psNodePtr a_ptr_to_me)
{
  QMutexLocker                Locker(&PtrListMutex);
  tListOfPtrs::iterator       P = Pointers.begin();
  while( (P != Pointers.end()) && ((*P) != a_ptr_to_me))  P++;
  ASSERT( P != Pointers.end() );
  Pointers.erase(P);
}


///*! returns true if any of
//unstripable nodes down the net
//have to be saved (i.e. are Edited).
//*/
//bool  sNode::Strip ()
//{
//  return EditedFlag;
//};


//void  sNode::Preallocate (sNodePtr Host ,sAttrName  AN)
//{
//  sNet::Manager->PreallocateNode(this,Host,AN);
//};


//void  sNode::AllocateMandatory (sNodeLocation::eType  LType
//                               ,rcsString             LPath)
//{
//  sNet::Manager->AllocateMandatory(this,LType,LPath);
//};


//void  sNode::Save ()
//{
//  if( EditedFlag ){
//    /*/////////DEBUG
//    tNodeUID    DEBUG_ID=Location.NodeUID;
//    sLongString DEBUG_LP=Location.Path+Location.DefaultSuffix;
//    bool        DEBUG_EDITED=Edited;
//    int         DEBUG_NUMOFHOSTS=NumberOfHosts;
//    int         DEBUG_NUMOFPTRS=NumberOfPointers;
//    */
//    EditedFlag= ! sNet::Manager->SaveNode(this);
//  };
//};


//void  sNode::AssignIntrinsicContent (sNodePtr /*NP*/)
//{
//  //! @todo{NetIO} ASSERT( OwnType() <= NP->OwnType() );
//  //NumberOfHosts - is not copied
//  ASSERT( Redirector.IsNULL() );
////  Location.Type;//=NP->Location.Type;     //DEBUG!!!!!!!!!!!!!!!!!!!!!!!!!!!!
////  Location.NodeUID;//=NP->Location.NodeUID;
////  Location.Path;//=NP->Location.Path;
////  Location.DefaultSuffix;//=NP->Location.DefaultSuffix;
////  Location.Offset;//=NP->Location.Offset;
////  LocationEdited;//=NP->LocationEdited;
//};


//void  sNode::DisconnectAllPtrs ()    //////move it to its place
//{
//  ASSERT( Location.NodeUID != NoID );
//  ASSERT( NumberOfPointers > 0 );
//  ASSERT( PtrRing );
//  while( PtrRing != PtrRing->Next ){
//    psNodePtr                 N = PtrRing->Next;
//    psNodeLocation            NL = new sNodeLocation(N->TheNodePtr->Location);
//    N->TheNodePtr->NumberOfPointers--;
//    N->InMemory=false;
//    N->TheNodeLocationPtr=NL;
//    PtrRing->Next=N->Next;
//  };
//  psNodeLocation              NL = new sNodeLocation(Location);
//  NumberOfPointers--;
//  PtrRing->InMemory=false;
//  PtrRing->TheNodeLocationPtr=NL;
//  PtrRing=NULL;
//  ASSERT( NumberOfPointers == 0 );
//};


//psClusterNavigator  sNode::new_Navigator ()
//{
//  return NULL;
//};


//void  sNode::IncHosts ()
//{
//  ++NumberOfHosts;
//  SetEditedFlag();
//};


//void  sNode::DecHosts ()
//{
//  ASSERT( NumberOfHosts );
//  --NumberOfHosts;
//  SetEditedFlag();
//};


//sString  sNode::DebugInfo ()
//{
//  sString                  Txt;
//  (Txt+="UID=")<<Location.NodeUID;
//  (Txt+=" #hosts=")<<NumberOfHosts;
//  (Txt+=" #ptrs=")<<NumberOfPointers;
//  (Txt+=" \n loc='")+=Location.Path;
//  (Txt+="'")+=Location.DefaultSuffix;
//  Txt+=(EditedFlag ? "edited" : "");
//  return Txt;
//};

//------------------------------------------------------------------------------
} //namespace RA

