////////////////////////////////////////////////////////////////////////////////
/*! @file PhysMath.cpp   Basic expression interpreter.
- Part of RANet - Research Assistant Net Library.
- Copyright(C) 2014, Viktor E. Bursian, St.Petersburg, Russia.
                     Viktor.Bursian@mail.ioffe.ru
*///////////////////////////////////////////////////////////////////////////////
#include "PhysMath.h"
#include "Symbols.h"
#include "Log.h"
namespace RA {
//------------------------------------------------------------------------------
//---------------------------------------------------------------- sPhysMath ---

sPhysMath::sMathOpRegistrator::sMathOpRegistrator ()
{
  Definitions.Add(&UnaryMinusPhysValue
      ,"-" ,sPhysValue::Type());
  Definitions.Add(&PhysValuePlusPhysValue
      ,"+" ,sPhysValue::Type() ,sPhysValue::Type());
  Definitions.Add(&PhysValueMinusPhysValue
      ,"-" ,sPhysValue::Type() ,sPhysValue::Type());
  Definitions.Add(&PhysValueMultPhysValue
      ,"*" ,sPhysValue::Type() ,sPhysValue::Type());
  Definitions.Add(&PhysValueDivPhysValue
      ,"/" ,sPhysValue::Type() ,sPhysValue::Type());
  Definitions.Add(&PhysValuePowerInteger
      ,"^" ,sPhysValue::Type() ,sIntegerValue::Type());

  Definitions.Add(&sin  ,"sin" ,sPhysValue::Type());
  Definitions.Add(&cos  ,"cos" ,sPhysValue::Type());
  Definitions.Add(&tg   ,"tg"  ,sPhysValue::Type());
  Definitions.Add(&ctg  ,"ctg" ,sPhysValue::Type());

  Definitions.Add(&seven   ,"seven");

  sPhysMath::AddConstant(sPhysValue(299792458.0,0.5,"m/s"),"c");

  sUnit::Register(Symbol::Degree ,sMetricPrefix::NoPrefixAllowed);
  sUnit::Register("degree",sMetricPrefix::NoPrefixAllowed
                          ,sPhysValue(Symbol::Degree));
  sUnit::Register("&deg;" ,sMetricPrefix::NoPrefixAllowed
                          ,sPhysValue(Symbol::Degree));
}


void  sPhysMath::AddConstant (rcsPhysValue  value
                             ,rcsString     identifier)
{
  Definitions.Add(new sPhysValue(value),identifier);
}


psMathValue  sPhysMath::Constant (rcsString  identifier
                                 ,int        start_pos ) const
{
  pcsMathValue                C = FindConst(identifier);
  if( C && (C->TrueType() == sPhysValue::Type()) ){
    pcsPhysValue  PV = dynamic_cast<pcsPhysValue>(C);
    return  new sPhysValue(*PV);
  }else if( C && (C->TrueType() == sRealValue::Type()) ){
    pcsRealValue  RV = dynamic_cast<pcsRealValue>(C);
    return  RV->Replica();
  }
  return  new sMathError(sString("undefined '")+identifier+"'"
                        ,start_pos,Cursor);
}

/*!

@note Real numbers are autoconverted into unitsless physvalues.
*/
psMathValue  sPhysMath::Value ()
{
  integer                     IntegerResult;
  real                        RealResult;
  real                        Precision;
  bool                        GotIntNumber = false;
  bool                        GotRealNumber = false;
  sString                     UnitsString;
  if( End() ){
  }else if( IntegerNumber(IntegerResult) ){
    GotIntNumber = true;
  }else if( RealNumber(RealResult,Precision) ){
    GotRealNumber = true;
  }
  if( GotIntNumber || GotRealNumber ){
    if( (Cursor < SourceLength-1) && (Source[Cursor] == '[') ){
      UnitsString = Source.Tail(Cursor+1);
      size_t                  C = UnitsString.Pos("]");
      if( C < UnitsString.Length() ){
        UnitsString = UnitsString.Substr(0,C);
        Cursor += (C + 2);
      }else{
        return new sMathError("missing ']'",Cursor);
      }
    }else if( Identifier(UnitsString) ){
    }
  }
  if( GotIntNumber ){
    if( UnitsString.Empty() ){
      return new sIntegerValue(IntegerResult);
    }else{
      return new sPhysValue(IntegerResult,UnitsString);
    }
  }else if( GotRealNumber ){
    return new sPhysValue(RealResult,Precision,UnitsString);
  }
  return NULL;
}


psMathValue  sPhysMath::Call (rcsString    identifier
                             ,int          start_pos
                             ,psMathValue  arg1      ) const
{
  psMathValue                 Result(NULL);
  pfFunc1Arg                  FuncPtr;
  sString                     FuncDecl(identifier
                                      +"("+arg1->TrueType()->Name()
                                      +")");
  if( (bool)(FuncPtr = FindFunc(identifier,arg1)) ){
    Result = (*FuncPtr)(arg1);
  }
  if( ! Result && sRealValue::Type()->CanCast(arg1) ){
    psMathValue                 A1;
    A1 = new sPhysValue(real(*dynamic_cast<psRealValue>(arg1)));
    if( (bool)(FuncPtr = FindFunc(identifier,A1)) )
      Result = (*FuncPtr)(A1);
    delete A1;
  }
  if( ! Result && sPhysValue::Type()->CanCast(arg1) ){
    try{
      psMathValue                 A1;
      A1 = new sRealValue(dynamic_cast<psPhysValue>(arg1)->ToReal());
      if( (bool)(FuncPtr = FindFunc(identifier,A1)) )
        Result = (*FuncPtr)(A1);
      delete A1;
    }catch( rxRAlgebra ){
//      Result = new sMathError(FuncDecl+" requires unitsless argument"
//                             ,start_pos,Cursor);
    }
  }
  if( ! Result ){
    Result = new sMathError(FuncDecl+"is undefined",start_pos,Cursor);
  }
  return Result;
}



psMathValue  sPhysMath::Call (rcsString    identifier
                             ,int          start_pos
                             ,psMathValue  arg1
                             ,psMathValue  arg2      ) const
{
  psMathValue                 Result(NULL);
  pfFunc2Arg                  FuncPtr;
  sString                     FuncDecl(identifier
                                      +"("+arg1->TrueType()->Name()
                                      +","+arg2->TrueType()->Name()
                                      +")");
  if( (bool)(FuncPtr = FindFunc(identifier,arg1,arg2)) ){
    Result = (*FuncPtr)(arg1,arg2);
  }
  //if not found, but either one or both arguments are real, then try to apply
  //a function with physvalue argument(s)
  {
    if( ! Result && sRealValue::Type()->CanCast(arg1) ){
      psMathValue                 A1;
      A1 = new sPhysValue(real(*dynamic_cast<psRealValue>(arg1)));
      if( (bool)(FuncPtr = FindFunc(identifier,A1,arg2)) )
        Result = (*FuncPtr)(A1,arg2);
      delete A1;
    }
    if( ! Result && sRealValue::Type()->CanCast(arg2) ){
      psMathValue                 A2;
      A2 = new sPhysValue(real(*dynamic_cast<psRealValue>(arg2)));
      if( (bool)(FuncPtr = FindFunc(identifier,arg1,A2)) )
        Result = (*FuncPtr)(arg1,A2);
      delete A2;
    }
    if( ! Result && sRealValue::Type()->CanCast(arg1)
                 && sRealValue::Type()->CanCast(arg2) ){
      psMathValue                 A1;
      psMathValue                 A2;
      A1 = new sPhysValue(real(*dynamic_cast<psRealValue>(arg1)));
      A2 = new sPhysValue(real(*dynamic_cast<psRealValue>(arg2)));
      if( (bool)(FuncPtr = FindFunc(identifier,A1,A2)) )
        Result = (*FuncPtr)(A1,A2);
      delete A1;
      delete A2;
    }
  }
  //if not found, but either one or both arguments are unitsless physvalues,
  //then try to apply a function with real argument(s)
  {
    if( ! Result && sPhysValue::Type()->CanCast(arg1) ){
      try{
        psMathValue                 A1;
        A1 = new sRealValue(dynamic_cast<psPhysValue>(arg1)->ToReal());
        if( (bool)(FuncPtr = FindFunc(identifier,A1,arg2)) )
          Result = (*FuncPtr)(A1,arg2);
        delete A1;
      }catch( rxRAlgebra ){
//        Result = new sMathError(FuncDecl+" requires unitsless 1st argument"
//                               ,start_pos,Cursor);
      }
    }
    if( ! Result && sPhysValue::Type()->CanCast(arg2) ){
      try{
        psMathValue                 A2;
        A2 = new sRealValue(dynamic_cast<psPhysValue>(arg2)->ToReal());
        if( (bool)(FuncPtr = FindFunc(identifier,arg1,A2)) )
          Result = (*FuncPtr)(arg1,A2);
        delete A2;
      }catch( rxRAlgebra ){
//        Result = new sMathError(FuncDecl+" requires unitsless 2nd argument"
//                               ,start_pos,Cursor);
      }
    }
    if( ! Result && sPhysValue::Type()->CanCast(arg1)
                 && sPhysValue::Type()->CanCast(arg2) ){
      try{
        psMathValue                 A1;
        psMathValue                 A2;
        A1 = new sRealValue(dynamic_cast<psPhysValue>(arg1)->ToReal());
        A2 = new sRealValue(dynamic_cast<psPhysValue>(arg2)->ToReal());
        if( (bool)(FuncPtr = FindFunc(identifier,A1,A2)) )
          Result = (*FuncPtr)(A1,A2);
        delete A1;
        delete A2;
      }catch( rxRAlgebra ){
//        Result = new sMathError(FuncDecl+" requires unitsless arguments"
//                               ,start_pos,Cursor);
      }
    }
  }
  if( ! Result ){
    Result = new sMathError(FuncDecl+"is undefined",start_pos,Cursor);
  }
  return Result;
}



psMathValue  sPhysMath::Call (rcsString    identifier
                             ,int          start_pos
                             ,psMathValue  arg1
                             ,psMathValue  arg2
                             ,psMathValue  arg3      ) const
{
  psMathValue                 Result(NULL);
  pfFunc3Arg                  FuncPtr;
  sString                     FuncDecl(identifier
                                      +"("+arg1->TrueType()->Name()
                                      +","+arg2->TrueType()->Name()
                                      +","+arg3->TrueType()->Name()
                                      +")");
  if( (bool)(FuncPtr = FindFunc(identifier,arg1,arg2,arg3)) ){
    Result = (*FuncPtr)(arg1,arg2,arg3);
  }
  //if not found, but either one, or two, or three arguments are real, then try
  //to apply a function with physvalue argument(s)
  {
    if( ! Result && sRealValue::Type()->CanCast(arg1) ){
      psMathValue                 A1;
      A1 = new sPhysValue(real(*dynamic_cast<psRealValue>(arg1)));
      if( (bool)(FuncPtr = FindFunc(identifier,A1,arg2,arg3)) )
        Result = (*FuncPtr)(A1,arg2,arg3);
      delete A1;
    }
    if( ! Result && sRealValue::Type()->CanCast(arg2) ){
      psMathValue                 A2;
      A2 = new sPhysValue(real(*dynamic_cast<psRealValue>(arg2)));
      if( (bool)(FuncPtr = FindFunc(identifier,arg1,A2,arg3)) )
        Result = (*FuncPtr)(arg1,A2,arg3);
      delete A2;
    }
    if( ! Result && sRealValue::Type()->CanCast(arg3) ){
      psMathValue                 A3;
      A3 = new sPhysValue(real(*dynamic_cast<psRealValue>(arg3)));
      if( (bool)(FuncPtr = FindFunc(identifier,arg1,arg2,A3)) )
        Result = (*FuncPtr)(arg1,arg2,A3);
      delete A3;
    }
    if( ! Result && sRealValue::Type()->CanCast(arg1)
                 && sRealValue::Type()->CanCast(arg2) ){
      psMathValue                 A1;
      psMathValue                 A2;
      A1 = new sPhysValue(real(*dynamic_cast<psRealValue>(arg1)));
      A2 = new sPhysValue(real(*dynamic_cast<psRealValue>(arg2)));
      if( (bool)(FuncPtr = FindFunc(identifier,A1,A2,arg3)) )
        Result = (*FuncPtr)(A1,A2,arg3);
      delete A1;
      delete A2;
    }
    if( ! Result && sRealValue::Type()->CanCast(arg1)
                 && sRealValue::Type()->CanCast(arg3) ){
      psMathValue                 A1;
      psMathValue                 A3;
      A1 = new sPhysValue(real(*dynamic_cast<psRealValue>(arg1)));
      A3 = new sPhysValue(real(*dynamic_cast<psRealValue>(arg3)));
      if( (bool)(FuncPtr = FindFunc(identifier,A1,arg2,A3)) )
        Result = (*FuncPtr)(A1,arg2,A3);
      delete A1;
      delete A3;
    }
    if( ! Result && sRealValue::Type()->CanCast(arg2)
                 && sRealValue::Type()->CanCast(arg3) ){
      psMathValue                 A2;
      psMathValue                 A3;
      A2 = new sPhysValue(real(*dynamic_cast<psRealValue>(arg2)));
      A3 = new sPhysValue(real(*dynamic_cast<psRealValue>(arg3)));
      if( (bool)(FuncPtr = FindFunc(identifier,arg1,A2,A3)) )
        Result = (*FuncPtr)(arg1,A2,A3);
      delete A2;
      delete A3;
    }
    if( ! Result && sRealValue::Type()->CanCast(arg1)
                 && sRealValue::Type()->CanCast(arg2)
                 && sRealValue::Type()->CanCast(arg3) ){
      psMathValue                 A1;
      psMathValue                 A2;
      psMathValue                 A3;
      A1 = new sPhysValue(real(*dynamic_cast<psRealValue>(arg1)));
      A2 = new sPhysValue(real(*dynamic_cast<psRealValue>(arg2)));
      A3 = new sPhysValue(real(*dynamic_cast<psRealValue>(arg3)));
      if( (bool)(FuncPtr = FindFunc(identifier,A1,A2,A3)) )
        Result = (*FuncPtr)(A1,A2,A3);
      delete A1;
      delete A2;
      delete A3;
    }
  }
  //if not found, but either one, or two, or three arguments are
  //unitsless physvalues, then try to apply a function with real argument(s)
  {
    if( ! Result && sPhysValue::Type()->CanCast(arg1) ){
      try{
        psMathValue                 A1;
        A1 = new sRealValue(dynamic_cast<psPhysValue>(arg1)->ToReal());
        if( (bool)(FuncPtr = FindFunc(identifier,A1,arg2,arg3)) )
          Result = (*FuncPtr)(A1,arg2,arg3);
        delete A1;
      }catch( rxRAlgebra ){
//        Result = new sMathError(FuncDecl+" requires unitsless 1st argument"
//                               ,start_pos,Cursor);
      }
    }
    if( ! Result && sPhysValue::Type()->CanCast(arg2) ){
      try{
        psMathValue                 A2;
        A2 = new sRealValue(dynamic_cast<psPhysValue>(arg2)->ToReal());
        if( (bool)(FuncPtr = FindFunc(identifier,arg1,A2,arg3)) )
          Result = (*FuncPtr)(arg1,A2,arg3);
        delete A2;
      }catch( rxRAlgebra ){
//        Result = new sMathError(FuncDecl+" requires unitsless 2nd argument"
//                               ,start_pos,Cursor);
      }
    }
    if( ! Result && sPhysValue::Type()->CanCast(arg3) ){
      try{
        psMathValue                 A3;
        A3 = new sRealValue(dynamic_cast<psPhysValue>(arg3)->ToReal());
        if( (bool)(FuncPtr = FindFunc(identifier,arg1,arg2,A3)) )
          Result = (*FuncPtr)(arg1,arg2,A3);
        delete A3;
      }catch( rxRAlgebra ){
//        Result = new sMathError(FuncDecl+" requires unitsless 3rd argument"
//                               ,start_pos,Cursor);
      }
    }
    if( ! Result && sPhysValue::Type()->CanCast(arg1)
                 && sPhysValue::Type()->CanCast(arg2) ){
      try{
        psMathValue                 A1;
        psMathValue                 A2;
        A1 = new sRealValue(dynamic_cast<psPhysValue>(arg1)->ToReal());
        A2 = new sRealValue(dynamic_cast<psPhysValue>(arg2)->ToReal());
        if( (bool)(FuncPtr = FindFunc(identifier,A1,A2,arg3)) )
          Result = (*FuncPtr)(A1,A2,arg3);
        delete A1;
        delete A2;
      }catch( rxRAlgebra ){
//        Result = new sMathError(FuncDecl
//                               +" requires unitsless 1st and 2nd arguments"
//                               ,start_pos,Cursor);
      }
    }
    if( ! Result && sPhysValue::Type()->CanCast(arg1)
                 && sPhysValue::Type()->CanCast(arg3) ){
      try{
        psMathValue                 A1;
        psMathValue                 A3;
        A1 = new sRealValue(dynamic_cast<psPhysValue>(arg1)->ToReal());
        A3 = new sRealValue(dynamic_cast<psPhysValue>(arg3)->ToReal());
        if( (bool)(FuncPtr = FindFunc(identifier,A1,arg2,A3)) )
          Result = (*FuncPtr)(A1,arg2,A3);
        delete A1;
        delete A3;
      }catch( rxRAlgebra ){
//        Result = new sMathError(FuncDecl
//                               +" requires unitsless 1st and 3rd arguments"
//                               ,start_pos,Cursor);
      }
    }
    if( ! Result && sPhysValue::Type()->CanCast(arg2)
                 && sPhysValue::Type()->CanCast(arg3) ){
      try{
        psMathValue                 A2;
        psMathValue                 A3;
        A2 = new sRealValue(dynamic_cast<psPhysValue>(arg2)->ToReal());
        A3 = new sRealValue(dynamic_cast<psPhysValue>(arg3)->ToReal());
        if( (bool)(FuncPtr = FindFunc(identifier,arg1,A2,A3)) )
          Result = (*FuncPtr)(arg1,A2,A3);
        delete A2;
        delete A3;
      }catch( rxRAlgebra ){
//        Result = new sMathError(FuncDecl
//                               +" requires unitsless 2nd and 3rd arguments"
//                               ,start_pos,Cursor);
      }
    }
    if( ! Result && sPhysValue::Type()->CanCast(arg1)
                 && sPhysValue::Type()->CanCast(arg2)
                 && sPhysValue::Type()->CanCast(arg3) ){
      try{
        psMathValue                 A1;
        psMathValue                 A2;
        psMathValue                 A3;
        A1 = new sRealValue(dynamic_cast<psPhysValue>(arg1)->ToReal());
        A2 = new sRealValue(dynamic_cast<psPhysValue>(arg2)->ToReal());
        A3 = new sRealValue(dynamic_cast<psPhysValue>(arg3)->ToReal());
        if( (bool)(FuncPtr = FindFunc(identifier,A1,A2,A3)) )
          Result = (*FuncPtr)(A1,A2,A3);
        delete A1;
        delete A2;
        delete A3;
      }catch( rxRAlgebra ){
//        Result = new sMathError(FuncDecl+" requires unitsless arguments"
//                               ,start_pos,Cursor);
      }
    }
  }
  if( ! Result ){
    Result = new sMathError(FuncDecl+"is undefined",start_pos,Cursor);
  }
  return Result;
}



psMathValue  sPhysMath::Call (rcsString    identifier
                             ,int          start_pos
                             ,psMathValue  arg1
                             ,psMathValue  arg2
                             ,psMathValue  arg3
                             ,psMathValue  arg4      ) const
{
  psMathValue                 Result(NULL);
  pfFunc4Arg                  FuncPtr;
  sString                     FuncDecl(identifier
                                      +"("+arg1->TrueType()->Name()
                                      +","+arg2->TrueType()->Name()
                                      +","+arg3->TrueType()->Name()
                                      +","+arg4->TrueType()->Name()
                                      +")");
  if( (bool)(FuncPtr = FindFunc(identifier,arg1,arg2,arg3,arg4)) ){
    Result = (*FuncPtr)(arg1,arg2,arg3,arg4);
  }
  //if not found, but either one, or two, or three arguments are real, then try
  //to apply a function with physvalue argument(s)
  {
    if( ! Result && sRealValue::Type()->CanCast(arg1) ){
      psMathValue                 A1;
      A1 = new sPhysValue(real(*dynamic_cast<psRealValue>(arg1)));
      if( (bool)(FuncPtr = FindFunc(identifier,A1,arg2,arg3,arg4)) )
        Result = (*FuncPtr)(A1,arg2,arg3,arg4);
      delete A1;
    }
    if( ! Result && sRealValue::Type()->CanCast(arg2) ){
      psMathValue                 A2;
      A2 = new sPhysValue(real(*dynamic_cast<psRealValue>(arg2)));
      if( (bool)(FuncPtr = FindFunc(identifier,arg1,A2,arg3,arg4)) )
        Result = (*FuncPtr)(arg1,A2,arg3,arg4);
      delete A2;
    }
    if( ! Result && sRealValue::Type()->CanCast(arg3) ){
      psMathValue                 A3;
      A3 = new sPhysValue(real(*dynamic_cast<psRealValue>(arg3)));
      if( (bool)(FuncPtr = FindFunc(identifier,arg1,arg2,A3,arg4)) )
        Result = (*FuncPtr)(arg1,arg2,A3,arg4);
      delete A3;
    }
    if( ! Result && sRealValue::Type()->CanCast(arg1)
                 && sRealValue::Type()->CanCast(arg2) ){
      psMathValue                 A1;
      psMathValue                 A2;
      A1 = new sPhysValue(real(*dynamic_cast<psRealValue>(arg1)));
      A2 = new sPhysValue(real(*dynamic_cast<psRealValue>(arg2)));
      if( (bool)(FuncPtr = FindFunc(identifier,A1,A2,arg3,arg4)) )
        Result = (*FuncPtr)(A1,A2,arg3,arg4);
      delete A1;
      delete A2;
    }
    if( ! Result && sRealValue::Type()->CanCast(arg1)
                 && sRealValue::Type()->CanCast(arg3) ){
      psMathValue                 A1;
      psMathValue                 A3;
      A1 = new sPhysValue(real(*dynamic_cast<psRealValue>(arg1)));
      A3 = new sPhysValue(real(*dynamic_cast<psRealValue>(arg3)));
      if( (bool)(FuncPtr = FindFunc(identifier,A1,arg2,A3,arg4)) )
        Result = (*FuncPtr)(A1,arg2,A3,arg4);
      delete A1;
      delete A3;
    }
    if( ! Result && sRealValue::Type()->CanCast(arg2)
                 && sRealValue::Type()->CanCast(arg3) ){
      psMathValue                 A2;
      psMathValue                 A3;
      A2 = new sPhysValue(real(*dynamic_cast<psRealValue>(arg2)));
      A3 = new sPhysValue(real(*dynamic_cast<psRealValue>(arg3)));
      if( (bool)(FuncPtr = FindFunc(identifier,arg1,A2,A3,arg4)) )
        Result = (*FuncPtr)(arg1,A2,A3,arg4);
      delete A2;
      delete A3;
    }
    if( ! Result && sRealValue::Type()->CanCast(arg1)
                 && sRealValue::Type()->CanCast(arg2)
                 && sRealValue::Type()->CanCast(arg3) ){
      psMathValue                 A1;
      psMathValue                 A2;
      psMathValue                 A3;
      A1 = new sPhysValue(real(*dynamic_cast<psRealValue>(arg1)));
      A2 = new sPhysValue(real(*dynamic_cast<psRealValue>(arg2)));
      A3 = new sPhysValue(real(*dynamic_cast<psRealValue>(arg3)));
      if( (bool)(FuncPtr = FindFunc(identifier,A1,A2,A3,arg4)) )
        Result = (*FuncPtr)(A1,A2,A3,arg4);
      delete A1;
      delete A2;
      delete A3;
    }
  }
  //if not found, but either one, or two, or three arguments are
  //unitsless physvalues, then try to apply a function with real argument(s)
  {
    if( ! Result && sPhysValue::Type()->CanCast(arg1) ){
      try{
        psMathValue                 A1;
        A1 = new sRealValue(dynamic_cast<psPhysValue>(arg1)->ToReal());
        if( (bool)(FuncPtr = FindFunc(identifier,A1,arg2,arg3,arg4)) )
          Result = (*FuncPtr)(A1,arg2,arg3,arg4);
        delete A1;
      }catch( rxRAlgebra ){
//        Result = new sMathError(FuncDecl+" requires unitsless 1st argument"
//                               ,start_pos,Cursor);
      }
    }
    if( ! Result && sPhysValue::Type()->CanCast(arg2) ){
      try{
        psMathValue                 A2;
        A2 = new sRealValue(dynamic_cast<psPhysValue>(arg2)->ToReal());
        if( (bool)(FuncPtr = FindFunc(identifier,arg1,A2,arg3,arg4)) )
          Result = (*FuncPtr)(arg1,A2,arg3,arg4);
        delete A2;
      }catch( rxRAlgebra ){
//        Result = new sMathError(FuncDecl+" requires unitsless 2nd argument"
//                               ,start_pos,Cursor);
      }
    }
    if( ! Result && sPhysValue::Type()->CanCast(arg3) ){
      try{
        psMathValue                 A3;
        A3 = new sRealValue(dynamic_cast<psPhysValue>(arg3)->ToReal());
        if( (bool)(FuncPtr = FindFunc(identifier,arg1,arg2,A3,arg4)) )
          Result = (*FuncPtr)(arg1,arg2,A3,arg4);
        delete A3;
      }catch( rxRAlgebra ){
//        Result = new sMathError(FuncDecl+" requires unitsless 3rd argument"
//                               ,start_pos,Cursor);
      }
    }
    if( ! Result && sPhysValue::Type()->CanCast(arg1)
                 && sPhysValue::Type()->CanCast(arg2) ){
      try{
        psMathValue                 A1;
        psMathValue                 A2;
        A1 = new sRealValue(dynamic_cast<psPhysValue>(arg1)->ToReal());
        A2 = new sRealValue(dynamic_cast<psPhysValue>(arg2)->ToReal());
        if( (bool)(FuncPtr = FindFunc(identifier,A1,A2,arg3,arg4)) )
          Result = (*FuncPtr)(A1,A2,arg3,arg4);
        delete A1;
        delete A2;
      }catch( rxRAlgebra ){
//        Result = new sMathError(FuncDecl
//                               +" requires unitsless 1st and 2nd arguments"
//                               ,start_pos,Cursor);
      }
    }
    if( ! Result && sPhysValue::Type()->CanCast(arg1)
                 && sPhysValue::Type()->CanCast(arg3) ){
      try{
        psMathValue                 A1;
        psMathValue                 A3;
        A1 = new sRealValue(dynamic_cast<psPhysValue>(arg1)->ToReal());
        A3 = new sRealValue(dynamic_cast<psPhysValue>(arg3)->ToReal());
        if( (bool)(FuncPtr = FindFunc(identifier,A1,arg2,A3,arg4)) )
          Result = (*FuncPtr)(A1,arg2,A3,arg4);
        delete A1;
        delete A3;
      }catch( rxRAlgebra ){
//        Result = new sMathError(FuncDecl
//                               +" requires unitsless 1st and 3rd arguments"
//                               ,start_pos,Cursor);
      }
    }
    if( ! Result && sPhysValue::Type()->CanCast(arg2)
                 && sPhysValue::Type()->CanCast(arg3) ){
      try{
        psMathValue                 A2;
        psMathValue                 A3;
        A2 = new sRealValue(dynamic_cast<psPhysValue>(arg2)->ToReal());
        A3 = new sRealValue(dynamic_cast<psPhysValue>(arg3)->ToReal());
        if( (bool)(FuncPtr = FindFunc(identifier,arg1,A2,A3,arg4)) )
          Result = (*FuncPtr)(arg1,A2,A3,arg4);
        delete A2;
        delete A3;
      }catch( rxRAlgebra ){
//        Result = new sMathError(FuncDecl
//                               +" requires unitsless 2nd and 3rd arguments"
//                               ,start_pos,Cursor);
      }
    }
    if( ! Result && sPhysValue::Type()->CanCast(arg1)
                 && sPhysValue::Type()->CanCast(arg2)
                 && sPhysValue::Type()->CanCast(arg3) ){
      try{
        psMathValue                 A1;
        psMathValue                 A2;
        psMathValue                 A3;
        A1 = new sRealValue(dynamic_cast<psPhysValue>(arg1)->ToReal());
        A2 = new sRealValue(dynamic_cast<psPhysValue>(arg2)->ToReal());
        A3 = new sRealValue(dynamic_cast<psPhysValue>(arg3)->ToReal());
        if( (bool)(FuncPtr = FindFunc(identifier,A1,A2,A3,arg4)) )
          Result = (*FuncPtr)(A1,A2,A3,arg4);
        delete A1;
        delete A2;
        delete A3;
      }catch( rxRAlgebra ){
//        Result = new sMathError(FuncDecl+" requires unitsless arguments"
//                               ,start_pos,Cursor);
      }
    }
  }
  if( ! Result ){
    Result = new sMathError(FuncDecl+"is undefined",start_pos,Cursor);
  }
  return Result;
}



psMathValue  sPhysMath::UnaryMinusPhysValue (pcsMathValue  arg)
{
  pcsPhysValue  Arg(dynamic_cast<pcsPhysValue>(arg));
  ASSERT( Arg );
  return new sPhysValue( -(*Arg) );
}


psMathValue  sPhysMath::PhysValuePlusPhysValue (pcsMathValue  arg1
                                               ,pcsMathValue  arg2)
{
  pcsPhysValue  Arg1(dynamic_cast<pcsPhysValue>(arg1));
  pcsPhysValue  Arg2(dynamic_cast<pcsPhysValue>(arg2));
  ASSERT( Arg1 && Arg2 );
  try{
    return new sPhysValue( (*Arg1) + (*Arg2) );
  }catch( rxRAlgebra ){
    return new sMathError("adding phys.value with uncompatible units");
  }
}


psMathValue  sPhysMath::PhysValueMinusPhysValue (pcsMathValue  arg1
                                                ,pcsMathValue  arg2)
{
  pcsPhysValue  Arg1(dynamic_cast<pcsPhysValue>(arg1));
  pcsPhysValue  Arg2(dynamic_cast<pcsPhysValue>(arg2));
  ASSERT( Arg1 && Arg2 );
  try{
    return new sPhysValue( (*Arg1) - (*Arg2) );
  }catch( rxRAlgebra ){
    return new sMathError("subtracting phys.value with uncompatible units");
  }
}


psMathValue  sPhysMath::PhysValueMultPhysValue (pcsMathValue  arg1
                                               ,pcsMathValue  arg2)
{
  pcsPhysValue  Arg1(dynamic_cast<pcsPhysValue>(arg1));
  pcsPhysValue  Arg2(dynamic_cast<pcsPhysValue>(arg2));
  ASSERT( Arg1 && Arg2 );
  return new sPhysValue( (*Arg1) * (*Arg2) );
}


psMathValue  sPhysMath::PhysValueDivPhysValue (pcsMathValue  arg1
                                              ,pcsMathValue  arg2)
{
  pcsPhysValue  Arg1(dynamic_cast<pcsPhysValue>(arg1));
  pcsPhysValue  Arg2(dynamic_cast<pcsPhysValue>(arg2));
  ASSERT( Arg1 && Arg2 );
  return new sPhysValue( (*Arg1) / (*Arg2) );
}


psMathValue  sPhysMath::PhysValuePowerInteger (pcsMathValue  arg1
                                              ,pcsMathValue  arg2)
{
  pcsPhysValue      Arg1(dynamic_cast<pcsPhysValue>(arg1));
  pcsIntegerValue   Arg2(dynamic_cast<pcsIntegerValue>(arg2));
  ASSERT( Arg1 && Arg2 );
  sPhysValue        R;
  R = sPhysValue(1.0);
  if( integer(*Arg2) > 0 ){
    for( int i = integer(*Arg2) ; i > 0 ; i-- )
      R = R * (*Arg1);
  }else if( integer(*Arg2) < 0 ){
    for( int i = integer(*Arg2) ; i < 0 ; i++ )
      R = R / (*Arg1);
  }
  return new sPhysValue(R);
}


psMathValue  sPhysMath::sin (pcsMathValue  arg)
{
  pcsPhysValue  Arg(dynamic_cast<pcsPhysValue>(arg));
  ASSERT( Arg );
  try{
    return new sPhysValue( ::sin( Arg->ToReal() ) );
  }catch( rxRAlgebra ){
    try{
      return new sPhysValue( ::sin( (((real)3.14159265358979323846)
                                       *(*Arg)
                                       /sPhysValue(180,"degree")).ToReal() ));
    }catch( rxRAlgebra ){
      return new sMathError(sString("argument for sin(...) should be unitsless"
                                    ", or in [")
                                   +Symbol::Degree
                                   +"], or in [degree], or in [&amp;deg;].");
    }
  }
}


psMathValue  sPhysMath::cos (pcsMathValue  arg)
{
  pcsPhysValue  Arg(dynamic_cast<pcsPhysValue>(arg));
  ASSERT( Arg );
  try{
    return new sPhysValue( ::cos( Arg->ToReal() ) );
  }catch( rxRAlgebra ){
    try{
      return new sPhysValue( ::cos( (((real)3.14159265358979323846)
                                       *(*Arg)
                                       /sPhysValue(180,"degree")).ToReal() ));
    }catch( rxRAlgebra ){
      return new sMathError(sString("argument for cos(...) should be unitsless"
                                    ", or in [")
                                   +Symbol::Degree
                                   +"], or in [degree], or in [&amp;deg;].");
    }
  }
}


psMathValue  sPhysMath::tg (pcsMathValue  arg)
{
  pcsPhysValue  Arg(dynamic_cast<pcsPhysValue>(arg));
  ASSERT( Arg );
  try{
    return new sPhysValue( ::tan( Arg->ToReal() ) );
  }catch( rxRAlgebra ){
    try{
      return new sPhysValue( ::tan( (((real)3.14159265358979323846)
                                       *(*Arg)
                                       /sPhysValue(180,"degree")).ToReal() ));
    }catch( rxRAlgebra ){
      return new sMathError(sString("argument for tg(...) should be unitsless"
                                    ", or in [")
                                   +Symbol::Degree
                                   +"], or in [degree], or in [&amp;deg;].");
    }
  }
}


psMathValue  sPhysMath::ctg (pcsMathValue  arg)
{
  pcsPhysValue  Arg(dynamic_cast<pcsPhysValue>(arg));
  ASSERT( Arg );
  try{
    return new sPhysValue( 1.0 / ::tan( Arg->ToReal() ) );
  }catch( rxRAlgebra ){
    try{
      return new sPhysValue(1.0/ ::tan( (((real)3.14159265358979323846)
                                       *(*Arg)
                                       /sPhysValue(180,"degree")).ToReal() ));
    }catch( rxRAlgebra ){
      return new sMathError(sString("argument for ctg(...) should be unitsless"
                                    ", or in [")
                                   +Symbol::Degree
                                   +"], or in [degree], or in [&amp;deg;].");
    }
  }
}


psMathValue  sPhysMath::seven ()
{
  return new sIntegerValue( 7 );
}

//------------------------------------------------------------------------------
} //namespace RA



