////////////////////////////////////////////////////////////////////////////////
/*! @file QtSQLWrapper.hpp Thin wrapper over the Qt Sql classes.
- Part of RANet - Research Assistant Net Library.
- Copyright(C) 2013-2020, Viktor E. Bursian, St.Petersburg, Russia.
                          VBursian AT gmail DOT com
*///////////////////////////////////////////////////////////////////////////////
#ifndef QtSQLWrapper_H
#define QtSQLWrapper_H
#include <exception>
#include "Excepts.h"
#include "StringsAndQStrings.h"
#include <QFileInfo>
#include <QSqlDatabase>
#include <QSqlQuery>
#include <QSqlError>
#include <QSqlDriver>
#include <QVariant>
namespace RA {
//------------------------------------------------------------------------------

ANNOUNCE_CLASS(xSQL)
ANNOUNCE_CLASS(xSQL_Open)
ANNOUNCE_CLASS(xSQL_OpenUnexistingSqlite)
ANNOUNCE_CLASS(sDBConnection)
ANNOUNCE_CLASS(sSQLQuery)

//--------------------------------------------------------------------- xSQL ---

//class RANet_EXPORT  xSQL : public xException
class RANet_EXPORT  xSQL : public std::exception
{
  public:
                              xSQL ()
                                {}
                              xSQL (rcsString  error)
                                  :Error(error)
                                {}
    const sString             Error;
  public:
                              xSQL (QSqlError  error)
                                  :SQLError(error)
                                {}
    const char *              what() const noexcept
                                {
                                  return (QString("SQL error: ")
                                         +SQLError.text()
                                         ).toUtf8().data();
                                }
    QSqlError                 SQLError;
};

//---------------------------------------------------------------- xSQL_Open ---

class RANet_EXPORT  xSQL_Open : public xSQL
{
  public:
                              xSQL_Open (rcsString  error)
                                  :xSQL(error)
                                {}
  public:
                              xSQL_Open (QString    connection_name
                                        ,QSqlError  error)
                                  :xSQL(error)
                                  ,ConnectionName(connection_name)
                                {}
    const char *              what() const noexcept
                                {
                                  return (QString("SQL error: ")
                                         +SQLError.text()
                                         ).toUtf8().data();
                                }
    QString                   ConnectionName;
};

//---------------------------------------------------------------- xSQL_Open ---

class RANet_EXPORT  xSQL_OpenUnexistingSqlite : public xSQL_Open
{
  public:
                              xSQL_OpenUnexistingSqlite (rcsString  error)
                                  :xSQL_Open(error)
                                {}
};

//------------------------------------------------------------ sDBConnection ---

class  RANet_EXPORT  sDBConnection : public QSqlDatabase
{
  public:
    sDBConnection () {}

    sDBConnection (rcsDBConnection  other) :QSqlDatabase(other) {}

    sDBConnection (const QSqlDatabase &  other) :QSqlDatabase(other) {}

    rsDBConnection operator= (rcsDBConnection  other)
      { return rsDBConnection(QSqlDatabase::operator=(other)); }

    rsDBConnection operator= (const QSqlDatabase &  other)
      { return rsDBConnection(QSqlDatabase::operator=(other)); }

    bool open ()
      {
        if( driverName() == "QSQLITE" ){
          if( ! QFileInfo( databaseName() ).exists() ){
            throw xSQL_OpenUnexistingSqlite(sString("in sDBConnection::open()")
                           +"\ndatabaseName='"+FromQString(databaseName())
                           +"' - file is not found");
          }
        }
        bool OK = QSqlDatabase::open();
        if( ! OK ){
          throw xSQL_Open(sString("in sDBConnection::open():")
                         +"\nDatabase name='"+FromQString(databaseName())+"'"
                         +"\nDatabase error: "
                         +FromQString(lastError().databaseText())
                         +"\nDriver error: "
                         +FromQString(lastError().driverText())
                         );
        };
        return OK;
      }

    bool transaction ()
      {
        bool OK = QSqlDatabase::transaction();
        if( ! OK ){ throw xSQL(); };
        return OK;
      }

    bool commit ()
      {
        bool OK = QSqlDatabase::commit();
        if( ! OK ){ throw xSQL(); };
        return OK;
      }

    bool rollback ()
      {
        bool OK = QSqlDatabase::rollback();
        if( ! OK ){ throw xSQL(); };
        return OK;
      }

};

//---------------------------------------------------------------- sSQLQuery ---

class  RANet_EXPORT  sSQLQuery : public QSqlQuery
{
  public:
    explicit sSQLQuery (QSqlResult *  r) :QSqlQuery(r) {}

    explicit sSQLQuery (const QString &  query = QString()
                       ,sDBConnection     db = QSqlDatabase() )
        :QSqlQuery(query,db) {}

    explicit sSQLQuery (const QString &  query = QString()
                       ,QSqlDatabase     db = QSqlDatabase() )
        :QSqlQuery(query,db) {}

    explicit sSQLQuery (sDBConnection db)
        :QSqlQuery(db) {}

    explicit sSQLQuery (QSqlDatabase db)
        :QSqlQuery(db) {}

    sSQLQuery (rcsSQLQuery  other) :QSqlQuery(other) {}

    sSQLQuery (const QSqlQuery &  other) :QSqlQuery(other) {}

    rsSQLQuery  operator= (rcsSQLQuery  other)
      { return rsSQLQuery(QSqlQuery::operator=(other)); }

    rsSQLQuery  operator= (const QSqlQuery &  other)
      { return rsSQLQuery(QSqlQuery::operator=(other)); }

    bool  exec (literal  query)
      {
        return exec( QString(query) );
      }

    bool  exec (rcsString  query)
      {
        return exec( ToQString(query) );
      }

    bool exec (const QString &  query)
      {
        if( ! QSqlQuery::exec(query) ){
          throw xSQL(sString("in sSQLQuery::exec(")
                    +FromQString(query)+")\n"
                    +"\nDatabase error: "
                    +FromQString(lastError().databaseText())
                    +"\nDriver error: "
                    +FromQString(lastError().driverText())
                    );
        };
        return true;
      }

    bool  exec ()
      {
        if( ! QSqlQuery::exec() ){
          throw xSQL(sString("in sSQLQuery::exec()  ")
                    +"SQL:("+FromQString(executedQuery())+")\n"
                    +"\nDatabase error: "
                    +FromQString(lastError().databaseText())
                    +"\nDriver error: "
                    +FromQString(lastError().driverText())
                    );
        };
        return true;
      }

    bool  prepare (literal  query)
      {
        return QSqlQuery::prepare( QString(query) );
      }

    bool  prepare (rcsString  query)
      {
        return QSqlQuery::prepare( ToQString(query) );
      }

    bool  prepare (const QString &  query)
      {
        if( ! QSqlQuery::prepare(query) ){
          throw xSQL(sString("in sSQLQuery::prepare(")
                    +FromQString(query)+")\n"
                    +"\nDatabase error: "
                    +FromQString(lastError().databaseText())
                    +"\nDriver error: "
                    +FromQString(lastError().driverText())
                    );
        };
        return true;
      }

    void  addBindString (rcsString  string)
      {
        QSqlQuery::addBindValue( literal(string) );
      }

    void  BindString (int pos , rcsString  string)
      {
        QSqlQuery::bindValue( pos , literal(string) );
      }

    sString  StringValue (int index) const
      {
        return QSqlQuery::value(index).toByteArray().data();
      }

    bool  SetParamValue (rcsString  table_name
                        ,rcsString  param_key
                        ,rcsString  param_value
                        ,rcsString  parkey_column_name = sString("parkey")
                        ,rcsString  parval_column_name = sString("parval") )
      {
        sString             QueryText;
        bool                OK1,OK2;
        QueryText = sString("UPDATE ")+table_name
            +" SET "+parval_column_name+"='"+param_value+"'"
            +" WHERE "+parkey_column_name+"='"+param_key+"'";
        OK1 = exec( ToQString(QueryText) );
        QueryText = sString("INSERT INTO ")+table_name
            +" SELECT '"+param_key+"','"+param_value+"'"
            +" WHERE NOT EXISTS (SELECT 1 FROM "+table_name
                              +" WHERE "+parkey_column_name+"='"+param_key+"')";
        OK2 = exec( ToQString(QueryText) );
        return OK1 && OK2;
      }

};

//------------------------------------------------------------------------------
} //namespace RA
#endif


