////////////////////////////////////////////////////////////////////////////////
/*! @file Ranges.cpp   Диапазон изменения физвеличины.
- Part of RANet - Research Assistant Net Library (based on ANSI C++).
- Copyright(C) 2010, Viktor E. Bursian, St.Petersburg, Russia.
                     Viktor.Bursian@mail.ioffe.ru
*///////////////////////////////////////////////////////////////////////////////
#include "Ranges.h"
#include "Symbols.h"
#include "Log.h"
namespace RA {
//------------------------------------------------------------------------------

//------------------------------------------------------------------- sRange ---

rsRange sRange::Invert()
{
  real                        F = From;
  From = To;
  To = F;
  return *this;
}

//--------------------------------------------------------------- sPhysRange ---
/*! Конструирует диапазон изменения безразмерной физвеличины
    от минус до плюс бесконечности.

Но при этом размерность не объявляется фиксированной. Т.е. переменная или
член класса, инициированная этим "пустым" конструктором, сможет принимать
в качестве значения диапазоны любой размерности.
Этот конструктор должен применяться лишь в весьма специфических случаях!

*/
sPhysRange::sPhysRange ()
    :TheUnits()
    ,UnitsAreFixedFlag(false)
    ,TheOrder(0)
    ,TheFromMantissa( - real_inf )
    ,TheToMantissa( real_inf )
    ,ThePrecisionMantissa(0.0)
    ,TheDigits(sPhysValue::MaxDigits-1)
    ,NormalizationLevel(99)
{
}


sPhysRange::sPhysRange (std::istream &  a_stream ,rcsVersion  version)
    :sStorable(a_stream,version)
    ,TheUnits(a_stream,version)
{
  a_stream.READ_INTO(UnitsAreFixedFlag);
  a_stream.READ_INTO(TheOrder);
  if( version < sVersion(1,5) ){
    realIOv14 R;
    a_stream.READ_INTO(R);  TheFromMantissa=R;
    a_stream.READ_INTO(R);  TheToMantissa=R;
    a_stream.READ_INTO(R);  ThePrecisionMantissa=R;
  }else{
    realIO R;
    a_stream.READ_INTO(R);  TheFromMantissa=R;
    a_stream.READ_INTO(R);  TheToMantissa=R;
    a_stream.READ_INTO(R);  ThePrecisionMantissa=R;
  }
  a_stream.READ_INTO(TheDigits);
  a_stream.READ_INTO(NormalizationLevel);
}


void  sPhysRange::Store (std::ostream &  a_stream) const
{
  sStorable::Store(a_stream);
  TheUnits.Store(a_stream);
  a_stream.WRITE_FROM(UnitsAreFixedFlag);
  a_stream.WRITE_FROM(TheOrder);
  realIO R;
  R=TheFromMantissa;       a_stream.WRITE_FROM(R);
  R=TheToMantissa;         a_stream.WRITE_FROM(R);
  R=ThePrecisionMantissa;  a_stream.WRITE_FROM(R);
  a_stream.WRITE_FROM(TheDigits);
  a_stream.WRITE_FROM(NormalizationLevel);
}


/*! Конструирует физдиапазон, измеряемый в указанных единицах,
    со значением от минус до плюс бесконечности и погрешностью 0.

Размерность объявляется фиксированной. Т.е. переменная или член класса,
инициированная этим конструктором, не сможет принимать значения в единицах, не
сопоставимых с заданными.

Заведомо безразмерные физдиапазоны тоже следует инициализировать этим
конструктором с константой RA::_Unitsless_ в качестве аргумента.

@sa sPhysRange::sPhysRange(sString), sPhysValue::sPhysValue(rcsUnits).
*/
sPhysRange::sPhysRange (rcsUnits  units)
    :TheUnits(units)
    ,UnitsAreFixedFlag(true)
    ,TheOrder(0)
    ,TheFromMantissa( - real_inf )
    ,TheToMantissa( real_inf )
    ,ThePrecisionMantissa(0.0)
    ,TheDigits(sPhysValue::MaxDigits-1)
    ,NormalizationLevel(1)
{
  Normalize(sPhysValue::DraftNormalization);
}


/*! Конструирует физдиапазон, измеряемый в указанных единицах,
    со значением от @a from до @a to и погрешностью @a precision.

Остальное - как в sPhysRange::sPhysRange(rcsUnits).

@sa sPhysRange::sPhysRange(real,real,sString,real).
*/
sPhysRange::sPhysRange (real      from
                       ,real      to
                       ,rcsUnits  units
                       ,real      precision)
    :TheUnits(units)
    ,UnitsAreFixedFlag(true)
    ,TheOrder(0)
    ,TheFromMantissa(from)
    ,TheToMantissa(to)
    ,ThePrecisionMantissa(std::abs(precision))
    ,TheDigits(sPhysValue::MaxDigits-1)
    ,NormalizationLevel(0)
{
  Normalize(sPhysValue::DraftNormalization);
}


/*! То же, что sPhysRange::sPhysRange(rcsUnits),
    только единицы задаются в виде текстовой строки.
*/
sPhysRange::sPhysRange (sString  unitsstr)
    :TheUnits(unitsstr)
    ,UnitsAreFixedFlag(true)
    ,TheOrder(0)
    ,TheFromMantissa( - real_inf )
    ,TheToMantissa( real_inf )
    ,ThePrecisionMantissa(0.0)
    ,TheDigits(sPhysValue::MaxDigits-1)
    ,NormalizationLevel(1)
{
  Normalize(sPhysValue::DraftNormalization);
}


/*! То же, что sPhysRange::sPhysRange(real,real,rcsUnits,real),
    только единицы задаются в виде текстовой строки.
*/
sPhysRange::sPhysRange (real     from
                       ,real     to
                       ,sString  unitsstr
                       ,real     precision)
    :TheUnits(unitsstr)
    ,UnitsAreFixedFlag(true)
    ,TheOrder(0)
    ,TheFromMantissa(from)
    ,TheToMantissa(to)
    ,ThePrecisionMantissa(std::abs(precision))
    ,TheDigits(sPhysValue::MaxDigits-1)
    ,NormalizationLevel(0)
{
  Normalize(sPhysValue::DraftNormalization);
}


/*! Конструирует физдиапазон из двух физвеличин, задающих его границы.

Разумеется, физвеличины @a from и @a to должны иметь значения в сопоставимых
единицах.  В противном случае, генерируется xRAlgebra.

Физдиапазон получает единицы от @a from, а погрешность максимальную из двух.
*/
sPhysRange::sPhysRange (rcsPhysValue from ,rcsPhysValue to)
    :TheUnits(from.Units())
    ,UnitsAreFixedFlag(true)
    ,TheOrder(0)
    ,ThePrecisionMantissa(0.0)
    ,TheDigits(sPhysValue::MaxDigits-1)
    ,NormalizationLevel(0)
{
  real                        Ratio;
  if( ! from.Units().Comparable(to.Units(),Ratio) )
    throw xRAlgebra("Incomparable units of arguments in sPhysRange constructor"
                   ,__FILE__,__LINE__);
  //! @todo{efficiency} reduce to one call of power10:
  //! Ratio *= power10( - );
  TheFromMantissa=from.TheValueMantissa*power10(from.TheOrder);
  TheToMantissa=to.TheValueMantissa*power10(to.TheOrder) / Ratio;
  ThePrecisionMantissa =
      rmax( std::abs( from.TheErrorMantissa*power10(from.TheOrder)     )
          , std::abs( to.TheErrorMantissa*power10(to.TheOrder) / Ratio )
          );
  Normalize(sPhysValue::DraftNormalization);
//  RANet::Log.Put(sLog::Debug,"sPhysRange::sPhysRange"
//      ,sString("sPhysRange::sPhysRange(")+from.Text()+","+to.Text()+")="+Txt()
//      +(sString("         Ratio=")<<Ratio)
//      );
}


/*!  Конструктор копирования.

Делает то, что от него ожидается. В том числе копируется и UnitsAreFixedFlag.

*/
sPhysRange::sPhysRange (rcsPhysRange range)
    :sStorable()
    ,TheUnits(range.TheUnits)
    ,UnitsAreFixedFlag(range.UnitsAreFixedFlag)
    ,TheOrder(range.TheOrder)
    ,TheFromMantissa(range.TheFromMantissa)
    ,TheToMantissa(range.TheToMantissa)
    ,ThePrecisionMantissa(range.ThePrecisionMantissa)
    ,TheDigits(range.TheDigits)
    ,NormalizationLevel(range.NormalizationLevel)
{
}


/*! Оператор присваивания.

Если в левой части стоит переменная-физдиапазон с фиксированными единицами, то
результат вычисления выражения в правой части будет преобразован к этим
единицам. А если преобразование невозможно, случится исключение вида xRAlgebra.
*/
rsPhysRange  sPhysRange::operator = (rcsPhysRange range)
{
  real                        Ratio=1.0;
  if( ! UnitsAreFixedFlag ){
    TheUnits = range.TheUnits;
    TheOrder = range.TheOrder;
    TheFromMantissa = range.TheFromMantissa;
    TheToMantissa = range.TheToMantissa;
    ThePrecisionMantissa = range.ThePrecisionMantissa;
    TheDigits = range.TheDigits;
    NormalizationLevel = range.NormalizationLevel;
  }else if( TheUnits.Comparable(range.TheUnits,Ratio) ){
    TheOrder = range.TheOrder;
    TheFromMantissa = range.TheFromMantissa / Ratio;
    TheToMantissa = range.TheToMantissa / Ratio;
    ThePrecisionMantissa = std::abs(range.ThePrecisionMantissa / Ratio);
    NormalizationLevel = 0;
    Normalize(sPhysValue::DraftNormalization);
  }else{
    throw xRAlgebra("Incomparable units sPhysRange::operator="
                   ,__FILE__,__LINE__);
  };
  return *this;
}


/*!
@return the left operand, which is (if necessary) narrowed in order
        to fit inside the right one.

If the units aren't comparable, then xRAlgebra is thrown.

*/
rsPhysRange  sPhysRange::operator &= (sPhysRange  range)
{
  real                        Ratio=1.0;
  if( ! TheUnits.Comparable(range.TheUnits,Ratio) )
    throw xRAlgebra("sPhysRange::operator &=: incomparable units"
                   ,__FILE__,__LINE__);
  Ratio *= power10(TheOrder-range.TheOrder);
  if( TheFromMantissa < range.TheFromMantissa / Ratio ){
    TheFromMantissa=range.TheFromMantissa / Ratio;
    ThePrecisionMantissa=std::max(ThePrecisionMantissa
                                 ,range.ThePrecisionMantissa / Ratio);
  }else if( TheToMantissa > range.TheToMantissa / Ratio ){
    TheToMantissa=range.TheToMantissa / Ratio;
    ThePrecisionMantissa=std::max(ThePrecisionMantissa
                                 ,std::abs(range.ThePrecisionMantissa / Ratio));
  };
  NormalizationLevel = 0;
  Normalize(sPhysValue::DraftNormalization);
  return *this;
}


/*!
@return the left operand, which is (if necessary) expanded in order
        to contain also the right one.

If the units aren't comparable, then xRAlgebra is thrown.

@note There is a specific behaviour, when one of operands IsNaN().

*/
rsPhysRange  sPhysRange::operator |= (sPhysRange  range)
{
  if( range.IsNaN() ){ //do nothing
  }else if( IsNaN() ){
    bool F = UnitsAreFixedFlag;
    UnitsAreFixedFlag = false;
    operator =(range);
    UnitsAreFixedFlag = F;
  }else{
    real                        Ratio=1.0;
    bool                        Changed = false;
    if( ! TheUnits.Comparable(range.TheUnits,Ratio) )
    throw xRAlgebra("sPhysRange::operator |=: incomparable units"
                   ,__FILE__,__LINE__);
    Ratio *= power10(TheOrder-range.TheOrder);
    if( TheFromMantissa > range.TheFromMantissa / Ratio ){
      TheFromMantissa=range.TheFromMantissa / Ratio;
      Changed = true;
    }
    if( TheToMantissa < range.TheToMantissa / Ratio ){
      TheToMantissa=range.TheToMantissa / Ratio;
      Changed = true;
    }
    if(  Changed  ){
      ThePrecisionMantissa=std::max(ThePrecisionMantissa
                                   ,range.ThePrecisionMantissa / Ratio);
    }
    NormalizationLevel = 0;
    Normalize(sPhysValue::DraftNormalization);
  }
  return *this;
}


/*!
@return the left operand, which is (if necessary) expanded in order
        to contain also the value of the right operand.

If the units aren't comparable, then xRAlgebra is thrown.

*/
rsPhysRange  sPhysRange::operator |= (sPhysValue  value)
{
  if( ! value.IsNaN() ){
    real                        Ratio=1.0;
    bool                        Changed = false;
    if( ! TheUnits.Comparable(value.TheUnits,Ratio) )
    throw xRAlgebra("sPhysRange::operator |=: incomparable units"
                   ,__FILE__,__LINE__);
    Ratio *= power10(TheOrder-value.TheOrder);
    if( TheFromMantissa > value.TheValueMantissa / Ratio ){
      TheFromMantissa = value.TheValueMantissa / Ratio;
      Changed = true;
    }
    if( TheToMantissa < value.TheValueMantissa / Ratio ){
      TheToMantissa = value.TheValueMantissa / Ratio;
      Changed = true;
    }
    if(  Changed  ){
      NormalizationLevel = 0;
      Normalize(sPhysValue::DraftNormalization);
    }
  }
  return *this;
}


/*!
@return  true, if the left operand range is equal or inside the right operand
         range within the precision of the right operand.

If the units aren't comparable, then xRAlgebra is thrown.

*/
bool  sPhysRange::operator <= (rcsPhysRange  range) const
{
  real                        Ratio=1.0;
  if( ! TheUnits.Comparable(range.TheUnits,Ratio) )
    throw xRAlgebra("sPhysRange::operator <=: incomparable units"
                   ,__FILE__,__LINE__);
  Ratio *= power10(TheOrder-range.TheOrder);
  return (TheFromMantissa
          >= (range.TheFromMantissa-std::abs(range.ThePrecisionMantissa)) / Ratio)
      && (TheToMantissa
          <= (range.TheToMantissa+std::abs(range.ThePrecisionMantissa)) / Ratio);
}


void  sPhysRange::Normalize (unsigned short int  level)
{
  if( level <= NormalizationLevel )
    return;
  RANet::Log.OpenBlock(sLog::Debug,"sPhysRange::Normalize",Txt());
  real                        MaxMantissa;
  if(     ( isnan(TheFromMantissa) || isinf(TheFromMantissa) )
       && ( isnan(TheToMantissa) || isinf(TheToMantissa) ) ){
    if( level > 1 )  TheUnits.DissolveOrder(level);
    TheOrder = 0;
  }else{
    if( isnan(TheFromMantissa) || isinf(TheFromMantissa) ){
      MaxMantissa = std::abs(TheToMantissa);
    }else if( isnan(TheToMantissa) || isinf(TheToMantissa) ){
      MaxMantissa = std::abs(TheFromMantissa);
    }else{
      MaxMantissa = std::max( std::abs(TheFromMantissa) , std::abs(TheToMantissa) );
    };
    if( (MaxMantissa == 0) && (ThePrecisionMantissa == 0) ){
      if( level > 1 )  TheUnits.DissolveOrder(level);
      TheOrder = 0;
      TheDigits = sPhysValue::MaxDigits;
    }else{
      int                         Shift = 0;
      bool                        TryToIncreaseExpectedRest = false;
      int                         ExpectedRest = 0;
      real                        C = 1.0;
      //приведение мантисс: V к [0.1,1) или если |V|<E, то E к [0.1,1)
      //и вычисление максимального сдвига,
      //допустимого после приведения единиц
      {
        if( std::abs(MaxMantissa) < ThePrecisionMantissa ){
//          TRACENOCR("|V|<E");
          real                        LgE = log10(ThePrecisionMantissa);
          if( LgE == - real_inf ){
            RANet::Log.Put(sLog::Debug,"sPhysRange::Normalize","LgE=-inf");
          }else if( LgE == real_inf ){
            RANet::Log.Put(sLog::Debug,"sPhysRange::Normalize","LgE=+inf");
          }else if( isnan(LgE) ){
            RANet::Log.Put(sLog::Debug,"sPhysRange::Normalize","LgE=NaN");
          }else{
            Shift = floor(LgE)+1;
          };
        }else{
          real                        LgV = log10(std::abs(MaxMantissa));
          if( LgV == -real_inf ){
            RANet::Log.Put(sLog::Debug,"sPhysRange::Normalize","LgV=-inf");
          }else if( LgV == real_inf ){
            RANet::Log.Put(sLog::Debug,"sPhysRange::Normalize","LgV=+inf");
          }else if( isnan(LgV) ){
            RANet::Log.Put(sLog::Debug,"sPhysRange::Normalize","LgV=NaN");
          }else{
            Shift = floor(LgV)+1;
          };
          TryToIncreaseExpectedRest = true;
        };
      };
      if( Shift != 0 ){
        C = power10(Shift);
        TheFromMantissa /= C;
        TheToMantissa /= C;
        ThePrecisionMantissa /= C;
        TheOrder += Shift;
      };
      if( TryToIncreaseExpectedRest ){
        if( ThePrecisionMantissa < 0.1 )  ExpectedRest = 1;
        if( ThePrecisionMantissa < 0.01 )  ExpectedRest = 2;
        if( ThePrecisionMantissa < 0.001 )  ExpectedRest = 3;
      };
      if( level > 1 ){
        TheOrder = TheUnits.DissolveOrder(level,TheOrder
                                         ,ExpectedRest-2,ExpectedRest);
      };
      if( TheOrder < 0 ){
        Shift = TheOrder - floor(TheOrder/3)*3;
      }else{
        Shift = TheOrder - floor((TheOrder-1)/3)*3;
      };
      if( Shift != 0 ){
        TheOrder = TheOrder - Shift;
        C = power10(Shift);
        TheFromMantissa *= C;
        TheToMantissa *= C;
        ThePrecisionMantissa *= C;
      };
      real                        D = log10(0.3/ThePrecisionMantissa);
      TheDigits = ( D > sPhysValue::MaxDigits ? sPhysValue::MaxDigits
                                              : ( D < 0 ? 0 : ceil(D) ) );
    };
  };
  NormalizationLevel = level;
  RANet::Log.Put(sLog::Debug,"sPhysRange::Normalize",sString("finally ")+Txt());
  RANet::Log.CloseBlock(sLog::Debug,"sPhysRange::Normalize");
}


/*! @details
Может иметь побочный эффект - нормализацию.

Сравни с sPhysRange::Text(eTextFormat)const.
С++ умеет подставлять ту или иную в зависимости от объявления объекта.
*/
sString  sPhysRange::Text (eTextFormat  F)
{
//  RANet::Log.Put(sLog::Debug,"sPhysRange::Text","");
  if( NormalizationLevel < sPhysValue::BriefNormalization ){
    Normalize(sPhysValue::BriefNormalization);
  };
  return Txt(F);
}


/*! @details
Без побочного эффекта, так как для нормализации использует копию. К сожалению,
при этом полезная работа по нормализации пропадает.

Сравни с sPhysRange::Text(eTextFormat).
С++ умеет подставлять ту или иную в зависимости от объявления объекта.
*/
sString  sPhysRange::Text (eTextFormat  F) const
{
//  RANet::Log.Put(sLog::Debug,"sPhysRange::Text","const");
  if( NormalizationLevel < sPhysValue::BriefNormalization ){
    sPhysRange                  NormalizedForm(*this);
    NormalizedForm.Normalize(sPhysValue::BriefNormalization);
    return NormalizedForm.Txt(F);
  }else{
    return Txt(F);
  };
}


sString  sPhysRange::Txt (eTextFormat  F) const
{
  sString                     T("[");
  sString                     UT;
  T += sString::FromReal(TheFromMantissa,TheDigits);
  T += "…";
  T += sString::FromReal(TheToMantissa,TheDigits);
  T += "]";
  if( TheOrder ){
    if( F == HTML ){
      T += Symbol::Multiplication;
      T += "10<sup>";
    }else{
      T += "*1e";
    };
    T << TheOrder;
    if( F == HTML ){
      T += "</sup> ";
    };
  };
  UT = TheUnits.Text(F);
  if( ! UT.Empty() ){
    T += " ";
    T += UT;
  };
  return T;
}

//------------------------------------------------------------------------------
} //namespace RA
