////////////////////////////////////////////////////////////////////////////////
/*! @file TableDataFunction.cpp Function given by data set in X-Y-table form.
- Part of RANet - Research Assistant Net Library.
- Copyright(C) 2010-2015, Viktor E. Bursian, St.Petersburg, Russia.
                          Viktor.Bursian@mail.ioffe.ru
*///////////////////////////////////////////////////////////////////////////////
#include "TableDataFunction.h"
//#include "Log.h"
namespace RA {
//------------------------------------------------------------------------------
//---------------------------------------------- sTableDataFunction::sTracer ---

sTableDataFunction::sTracer::sTracer (pcsTableDataFunction  host_curve
                                     ,rcsScales             scales    )
    :TheFunction(host_curve)
    ,Scaling(1.0,1.0)
    ,LastPointNo(host_curve->NumberOfPoints())
    ,PointNo(0)
    ,LineSegmentNo(1)
{
  try{
    Scaling.X = (host_curve->Multiplier().X / scales.X.Unit()).ToReal();
    Scaling.Y = (host_curve->Multiplier().Y / scales.Y.Unit()).ToReal();
//    RANet::Log.Put(sLog::Debug,"Graph"
//        ,sString("sTableDataFunction::sTracer::sTracer: Scaling X:")
//        +SX.Text()+" Y:"+SY.Text());
  }catch(rxRAlgebra){
//    RANet::Log.Put(sLog::Debug,"Graph"
//                  ,"sTableDataFunction::sTracer::sTracer: WrongUnits");
    PointNo = LastPointNo;
  }
  LineSegmentNo = PointNo + 1;
}


bool  sTableDataFunction::sTracer::NextPoint ()
{
  if( PointNo < LastPointNo ){
    Point.X= TheFunction->XValues[PointNo] * Scaling.X;
    Point.Y= TheFunction->YValues[PointNo] * Scaling.Y;
    PointNo++;
    return true;
  }else{
    return false;
  }
}


bool  sTableDataFunction::sTracer::NextLineSegment ()
{
  if( LineSegmentNo < LastPointNo ){
    LineSegment.Start.X = TheFunction->XValues[LineSegmentNo-1] * Scaling.X;
    LineSegment.Start.Y = TheFunction->YValues[LineSegmentNo-1] * Scaling.Y;
    LineSegment.End.X   = TheFunction->XValues[LineSegmentNo] * Scaling.X;
    LineSegment.End.Y   = TheFunction->YValues[LineSegmentNo] * Scaling.Y;
    LineSegmentNo++;
    return true;
  }else{
    return false;
  }
}

//------------------------------------------------------- sTableDataFunction ---

sTableDataFunction::~sTableDataFunction ()
{
  if( XValues ){
    delete [] XValues;  XValues=NULL;
  }
  if( YValues ){
    delete [] YValues;  YValues=NULL;
  }
}


sTableDataFunction::sTableDataFunction (sPhysPair  multiplier
                                       ,int        number_of_points)
    :sDataFunction(multiplier)
    ,TheNumberOfPoints(number_of_points)
    ,XValues(NULL)
    ,YValues(NULL)
{
  ASSERT( TheNumberOfPoints > 0 );
  XValues= new real [TheNumberOfPoints];
  YValues= new real [TheNumberOfPoints];
  for( int PointNo=0 ; PointNo < TheNumberOfPoints ; PointNo++ ){
    XValues[PointNo]=0.0;
    YValues[PointNo]=0.0;
  }
}


sTableDataFunction::sTableDataFunction (int number_of_points)
    :sDataFunction()
    ,TheNumberOfPoints(number_of_points)
    ,XValues(NULL)
    ,YValues(NULL)
{
  ASSERT( TheNumberOfPoints > 0 );
  XValues= new real [TheNumberOfPoints];
  YValues= new real [TheNumberOfPoints];
  for( int PointNo=0 ; PointNo < TheNumberOfPoints ; PointNo++ ){
    XValues[PointNo]=0.0;
    YValues[PointNo]=0.0;
  }
}


sTableDataFunction::sTableDataFunction (std::istream &  a_stream
                                       ,rcsVersion      version)
    :sDataFunction(a_stream,version)
{
  if( version <= sVersion(0,9) ){
  }else if( version < sVersion(1,5) ){
    a_stream.READ_INTO(TheNumberOfPoints);
    ASSERT( TheNumberOfPoints > 0 );
    XValues= new real [TheNumberOfPoints];
    YValues= new real [TheNumberOfPoints];
    realIOv14 R;
    for( int PointNo=0 ; PointNo < TheNumberOfPoints ; PointNo++ ){
      a_stream.READ_INTO(R);  XValues[PointNo]=R;
      a_stream.READ_INTO(R);  YValues[PointNo]=R;
    }
  }else{
    a_stream.READ_INTO(TheNumberOfPoints);
    ASSERT( TheNumberOfPoints > 0 );
    XValues= new real [TheNumberOfPoints];
    YValues= new real [TheNumberOfPoints];
    realIO R;
    for( int PointNo=0 ; PointNo < TheNumberOfPoints ; PointNo++ ){
      a_stream.READ_INTO(R);  XValues[PointNo]=R;
      a_stream.READ_INTO(R);  YValues[PointNo]=R;
    }
  }
}


void  sTableDataFunction::Store (std::ostream & a_stream) const
{
  sDataFunction::Store(a_stream);
  a_stream.WRITE_FROM(TheNumberOfPoints);
  realIO R;
  for( int PointNo=0 ; PointNo < TheNumberOfPoints ; PointNo++ ){
    R=XValues[PointNo];  a_stream.WRITE_FROM(R);
    R=YValues[PointNo];  a_stream.WRITE_FROM(R);
  }
}


sCurve::psTracer  sTableDataFunction::GetCurveTracer (rcsScales  scales) const
{
  return new sTableDataFunction::sTracer(this,scales);
}


sPhysValue  sTableDataFunction::Value (rcsPhysValue  x) const
{
  try{
    real                        X = (x/Multiplier().X).ToReal();
    sPoint                      P1;
    sPoint                      P2;
    bool                        Found = false;
    for( int PointNo=0 ; PointNo < TheNumberOfPoints-1 ; PointNo++ ){
      if( (XValues[PointNo] <= X) && (X <= XValues[PointNo+1]) ){
        P1 = Point(PointNo);
        P2 = Point(PointNo+1);
        Found = true;
        break;
      }
    }
    if( Found )
      return ( P1.Y + (P2.Y-P1.Y)*(X-P1.X)/(P2.X-P1.X) ) * Multiplier().Y;
  }catch( rxRAlgebra ){
  }
  return sPhysValue(real_nan,Multiplier().Y.Units());
}


sPoint  sTableDataFunction::Point (const int N) const
{
  ASSERT( N >= 0 );
  ASSERT( N < TheNumberOfPoints );
  return sPoint(XValues[N],YValues[N]);
}


real  sTableDataFunction::PointX (const int N) const
{
  ASSERT( N >= 0 );
  ASSERT( N < TheNumberOfPoints );
  return XValues[N];
}


real  sTableDataFunction::PointY (const int N) const
{
  ASSERT( N >= 0 );
  ASSERT( N < TheNumberOfPoints );
  return YValues[N];
}


void  sTableDataFunction::SetValue (int N ,rcsPoint P)
{
  if( (N >= 0) && (N < TheNumberOfPoints) ){
    XValues[N]=P.X;
    YValues[N]=P.Y;
//    if( ! RecalculateBoundaries )
//      LastBoundaries |= sPhysPair( P.X*Multiplier().X , P.Y*Multiplier().Y );
    RecalculateBoundaries = true;
  };
}


void  sTableDataFunction::SetXValue (int N ,real  X)
{
  if( (N >= 0) && (N < TheNumberOfPoints) ){
    XValues[N]=X;
//    if( ! RecalculateBoundaries )
//      LastBoundaries.X |= X * Multiplier().X;
    RecalculateBoundaries = true;
  };
}


void  sTableDataFunction::SetYValue (int N ,real  Y)
{
  if( (N >= 0) && (N < TheNumberOfPoints) ){
    YValues[N]=Y;
//    if( ! RecalculateBoundaries )
//      LastBoundaries.Y |= Y * Multiplier().Y;
    RecalculateBoundaries = true;
  };
}

//------------------------------------------------------------------------------
} //namespace RA




