////////////////////////////////////////////////////////////////////////////////
/*! @file Undo.cpp   Механизм откатов для сети и вообще Undo-Redo.
- Part of RANet - Research Assistant Net Library (based on ANSI C++).
- Copyright(C) 2006-2010, Viktor E. Bursian, St.Petersburg, Russia.
                     Viktor.Bursian@mail.ioffe.ru
*///////////////////////////////////////////////////////////////////////////////
#include "Undo.h"
#include "Excepts.h"
//------------------------------------------------------------------------------
namespace RA {

//---------------------------------------------------------------- sUndoList ---

sUndoList::~sUndoList ()
{
  if( RedoSubstitution ){
    delete RedoSubstitution;  RedoSubstitution=NULL;
  };
  Clear();
};


void  sUndoList::UndoIt ()
{
  psUndoItem                  I;
  ASSERT( Owner() );
  while( ! Items.empty() ){
    I=Items.back();
    Items.pop_back();
    Owner()->ExecuteUndoOn(I);
  };
};


void  sUndoList::RedoIt ()
{
  ASSERT( RedoSubstitution == NULL );
  psUndoItem                  I;
  ASSERT( Owner() );
  while( ! Items.empty() ){
    I=Items.back();
    Items.pop_back();
    ASSERT( dynamic_cast<psRedoItem>(I) || dynamic_cast<psUndoRedoItem>(I) );
    Owner()->ExecuteRedoOn(I);
  };
};


void  sUndoList::Push (psUndoItem I)
{
  if( I ) Items.push_back(I);
};


psUndoItem  sUndoList::Pop ()
{
  if( ! Items.empty() ){
    psUndoItem                  I=Items.back();
    Items.pop_back();
    return I;
  };
  return NULL;
};


psUndoItem  sUndoList::Top ()
{
  if( ! Items.empty() ){
    return Items.back();
  };
  return NULL;
};


void  sUndoList::Clear ()
{
  psUndoItem                  I;
  while( ! Items.empty() ){
    I=Items.back();
    Items.pop_back();
    delete I;
  };
};

//--------------------------------------------------------- sUndoGroupedList ---

void  sUndoGroupedList::Push (psUndoItem I)
{
  if( ! OpenGroups.empty() ){
    OpenGroups.top()->Push(I);
  }else{
    sUndoList::Push(I);
  };
};


psUndoItem  sUndoGroupedList::Pop ()
{
  if( ! OpenGroups.empty() ){
    return OpenGroups.top()->Pop();
    //If a group is started this method only pops from the group
    //until the group is empty
    //but not from the main undo list or any upper-level groups
    //Consequently, RollBack() only rolls back the current group
  }else{
    return sUndoList::Pop();
  };
};


psUndoItem  sUndoGroupedList::Top ()
{
  if( ! OpenGroups.empty() ){
    return OpenGroups.top()->Top();
  }else{
    return sUndoList::Top();
  };
};


void  sUndoGroupedList::GroupBegin (rcsString  hint ,psRedoItem RS)
{
  psUndoList                  G = new sUndoList(hint,RS);
  Push(G);
  OpenGroups.push(G);
};


void  sUndoGroupedList::GroupEnd ()
{
  if( ! OpenGroups.empty() ){
    OpenGroups.pop();
  }else{
    throw xProgramError("sUndoGroupedList::GroupEnd");
  };
};

//---------------------------------------------------------------- sUndoRedo ---

sUndoRedo::~sUndoRedo ()
{
  Clear();
};


sUndoRedo::sUndoRedo (int limit)
    :Undoing(false),Redoing(false),UndoLimit(limit)
{
};


void  sUndoRedo::Push (psUndoItem I)
{
  I->SetOwner(this);
  sUndoGroupedList::Push(I);
  if( ! (Undoing || Redoing || ThereIsUnclosedGroup()) ){
    REDO.Clear();
    if( UndoLimit >= 0 ){
      if( Size() > UndoLimit ){
        psUndoItem              II=Items.front();
        Items.pop_front();
        delete II;
      };
    };
  };
};


void  sUndoRedo::Clear ()
{
  REDO.Clear();
  sUndoGroupedList::Clear();
};


void  sUndoRedo::SetLimit (int limit)
{
  UndoLimit=limit;
  if( ! (Undoing || Redoing)
      & (UndoLimit >= 0)
      & CanUndo()              ){
    while( Size() > UndoLimit ){
      psUndoItem              II=Items.front();
      Items.pop_front();
      delete II;
    };
  };
};


void  sUndoRedo::ExecuteUndoOn (psUndoItem I)
{
  if( psUndoList  UL=dynamic_cast<psUndoList>(I) ){
    GroupOpen();
    UL->UndoIt();
    if( UL->RedoSubstitution ){
      GroupFlush();
      Push(UL->RedoSubstitution);
      UL->SetRedoSubstitution(NULL);
    }else{
      GroupClose(UL->Hint());
    };
    delete UL;
  }else if( psUndoRedoItem  URI=dynamic_cast<psUndoRedoItem>(I) ){
    GroupOpen();
    URI->UndoIt();
    GroupFlush();
    Push(URI);
  }else{
    I->UndoIt();
    delete I;
  };
};


void  sUndoRedo::ExecuteRedoOn (psUndoItem I)  //psRedoItem
{
  if( psUndoList  UL=dynamic_cast<psUndoList>(I) ){
    ASSERT( UL->RedoSubstitution == NULL );
    GroupOpen();
    UL->RedoIt();
    GroupClose(UL->Hint());
  }else if( psUndoRedoItem URI=dynamic_cast<psUndoRedoItem>(I) ){
    URI->RedoIt();
  }else if( psRedoItem  RI=dynamic_cast<psRedoItem>(I) ){
    RI->RedoIt();
  };
  delete I;
};


void  sUndoRedo::Undo ()
{
  if( CanUndo() ){
    Undoing=true;
    psUndoItem                  I = Pop();
    int                         S = Size();
    ASSERT( I );
    ExecuteUndoOn(I);
    while( Size() > S ){
      REDO.Push(Pop());
    };
    Undoing=false;
  };
};


void  sUndoRedo::Redo ()
{
  if( CanRedo() ){
    Redoing=true;
    psUndoItem                  I = psUndoItem(REDO.Pop());
    ASSERT( I );
    ASSERT( dynamic_cast<psRedoItem>(I) || dynamic_cast<psUndoRedoItem>(I) );
    ExecuteRedoOn(I);
    Redoing=false;
  };
};


void  sUndoRedo::GroupClose ()
{
  GroupEnd();
  if( ! (Undoing || Redoing || ThereIsUnclosedGroup()) ){
    REDO.Clear();
  };
};


void  sUndoRedo::GroupClose (rcsString  hint)
{
  if( ThereIsUnclosedGroup() ){
    GroupOnTop()->SetHintText(hint);
  };
  GroupClose();
};


void  sUndoRedo::GroupClose (psRedoItem RS)
{
  if( ThereIsUnclosedGroup() ){
    GroupOnTop()->SetRedoSubstitution(RS);
  };
  GroupClose();
};


void  sUndoRedo::GroupClose (rcsString  hint ,psRedoItem RS)
{
  if( ThereIsUnclosedGroup() ){
    GroupOnTop()->SetHintText(hint);
    GroupOnTop()->SetRedoSubstitution(RS);
  };
  GroupClose();
};


void  sUndoRedo::GroupFlush ()
{
  if( ThereIsUnclosedGroup() ){
    GroupEnd();
    delete Pop();
  }else{
    throw xProgramError("sUndoGroupedList::GroupFlush");
  };
};


void  sUndoRedo::RollBack ()
{
  if( ThereIsUnclosedGroup() ){
    psUndoItem                  I=Pop();
    while( I ){
      GroupOpen();
      I->UndoIt();
      GroupFlush();
      delete I;
      I=Pop();
    };
    GroupFlush();
  };
};

/*
void  sUndoRedo::RollBackSelective (rcsTag class_type)
{
  if( ThereIsUnclosedGroup() ){
    stack<psUndoItem>           Temp;
    psUndoItem                  I=Pop();
    while( I ){
      if( I <= class_type ){
        GroupOpen();
        I->UndoIt();
        GroupFlush();
        delete I;
      }else{
        Temp.push(I);
      };
      I=Pop();
    };
    while( ! Temp.empty() ){
      I=Temp.top();
      Temp.pop();
      Push(I);
    };
  };
};
*/

void  sUndoRedo::RollBackSelective (pfSelected Selected)
{
  if( ThereIsUnclosedGroup() ){
    stack<psUndoItem>           Temp;
    psUndoItem                  I=Pop();
    while( I ){
      if( (*Selected)(I) ){
        GroupOpen();
        I->UndoIt();
        GroupFlush();
        delete I;
      }else{
        Temp.push(I);
      };
      I=Pop();
    };
    while( ! Temp.empty() ){
      I=Temp.top();
      Temp.pop();
      Push(I);
    };
  };
};

bool  sUndoRedo::CanUndo ()
{
  return  ! ThereIsUnclosedGroup()  && ! Empty();
};


bool  sUndoRedo::CanRedo ()
{
  return  ! ThereIsUnclosedGroup()  && ! RedoEmpty();
};


sString  sUndoRedo::Hint ()
{
  return ( CanUndo() ? Top()->Hint() : sString() );
};


sString  sUndoRedo::RedoHint ()
{
  return ( CanRedo() ? REDO.Top()->Hint() : sString() );
};

//------------------------------------------------------------------------------
}; //namespace RA
