////////////////////////////////////////////////////////////////////////////////
/*! @file CCDDriver.cpp   Абстрактный базовый класс драйвера ПЗС-линейки.
- Part of BasicDevices - Research Assistant Plugin package.
- Uses  RAGUI - Research Assistant Graphical User Interface Library.
- Uses  Qt v.5.0.2  - http://qt-project.org/
- Uses  RANet - Research Assistant Net Library.
- Copyright(C) 2017, Dmitrii Nelson, St.Petersburg, Russia.
                          d.nelson@mail.ioffe.ru
*///////////////////////////////////////////////////////////////////////////////
#include "CCDDriver.h"
#include "attr.h"
#include <QDebug>
#include "Log.h"

namespace RA {
//------------------------------------------------------------------------------

DEFINE_CLASS_TAG(sCCDDriver)

csCommand                     sCCDDriver::cStart;
csCommand                     sCCDDriver::cStop;
const sDirective<sCCDDriver::sParameters>         sCCDDriver::cSetParameters;

//-------------------------------------------------------- sCCDDriver ----------

void  sCCDDriver::Preconfigure (sNodePtr  DeviceCfg)
{
  sPtr<sFolderNode>              DeviceSettings;
  DeviceCfg |= "Settings";
  DeviceSettings = DeviceCfg >> "Settings";
  if( DeviceSettings.IsCorrect() ){
    DeviceSettings |= attr("NumberOfPixels",0);
    DeviceSettings |= attr("IntegrationTime",sPhysValue(0.1,0.001,"s"));
  };
  sDriver::Preconfigure(DeviceCfg);
}

sCCDDriver::~sCCDDriver ()
{
  DataArray.SetSize(0);
  Wavelengths.SetSize(0);
}

sCCDDriver::sCCDDriver (sString   DevicetId
                       ,sNodePtr  DeviceCfg)
    :sDriver(DevicetId,DeviceCfg)
    ,NumOfPixels(0)
{
}


void  sCCDDriver::PostConstructor ()
{
  sDriver::PostConstructor();
  try{ //get config parameters
    get_from<sIntegerNode>(DeviceSettings >> "NumberOfPixels"   ,NumOfPixels);
    get_from<sPhysValueNode>(DeviceSettings >> "IntegrationTime" ,Parameters.IntegrationTime);
  }catch(...){
    SetInvalid(tr("Something wrong in configuration."));
  }
  DataArray.SetSize(NumOfPixels);
  Wavelengths.SetSize(NumOfPixels);
  Parameters.Gain = 0;
  Parameters.TriggerEdge = 0;
  Parameters.TriggerMode = 0;
  //evolution of the inherited state machine (behaviour)
  {
    { // evolution of the State_Warming
      State_W_Connect = new sState(State_Warming(),"State_W_Connect");
      State_W_GetWavelength = new sState(State_Warming(),"State_W_GetWavelength");
      State_W_Connect
          ->SetAsInitial()
          ->OnEnterCall( SLOT(ConnectUSBDevice()) )
          ->AddTransition(SIGNAL(Connected()),State_W_GetWavelength)
          ->AddTransition(SIGNAL(ErrorMessage()),State_Cooling());
      State_W_GetWavelength
          ->OnEnterCall( SLOT(CalculateWavelengths()) )
          ->AddTransition(State_Warming()->Fin());
    }
    { // evolution of the State_Operating
      State_O_Stopped = new sState(State_Operating(),"State_O_Stopped");
      State_O_Measuring = new sState(State_Operating(),1,"State_O_Measuring");
      State_O_GetMeasured = new sState(State_Operating(),"State_O_GetMeasured");
      State_O_Stopped
          ->SetAsInitial()
          ->AddTransition(cStart,SLOT(StartCCD ()) ,State_O_Measuring)
          ->AddReaction<sCCDDriver::sParameters>(cSetParameters
                       ,SLOT(SetCCD_Parameters(sCCDDriver::sParameters)));
      State_O_Measuring
          ->AddTransition(State_O_Measuring,SIGNAL(Timeout())
                         ,State_O_GetMeasured)
          ->AddTransition(cStop,SLOT(StopCCD ()) ,State_O_Stopped);
      State_O_GetMeasured
          ->OnEnterCall( SLOT(GetCCD_Data()) )
          ->AddTransition(State_O_Measuring)
          ->AddTransition(cStop,SLOT(StopCCD ()) ,State_O_Stopped);
    };
  };
}


//------------------------------------------------------------------------------
} //namespace RA
