////////////////////////////////////////////////////////////////////////////////
/*! @file CCDScaner.h  Базовый класс спектрометра с ПЗС-линейкой.
- Part of BasicDevices - a Research Assistant Plugin package.
- Uses  RAGUI - the Research Assistant Graphical User Interface.
- Uses  QtGui v.5.0  -  http://qt.nokia.com/
- Uses  RANet - Research Assistant Net Library (based on ANSI C++).
- Copyright(C) 2017, Dmitry K. Nelson, St.Petersburg, Russia.
                          D.Nelson@mail.ioffe.ru
*///////////////////////////////////////////////////////////////////////////////
#include "CCDScaner.h"
#include "attr.h"
#include "StringsAndQStrings.h"
#include "Trace.h"
#include <QMessageBox>
#include <QDebug>
#include "Log.h"

#include "NetMath.h"
#include "ExpressionNode.h"
#include "DataArrayFolder.h"
#include "DataListFolder.h"
#include "DataTableFolder.h"
#include "DataArrayRecord.h"
#include "DataListRecord.h"
#include "DataTableRecord.h"

namespace RA {
//------------------------------------------------------------------------------

DEFINE_CLASS_TAG(sCCDScaner)

//--------------------------------------------------------------- sCCDScaner ---

void  sCCDScaner::Preconfigure (sNodePtr  DeviceCfg)
{
  sPtr<sFolderNode>              DeviceSettings;
  sPtr<sFolderNode>              DeviceLinks;

  DeviceCfg |= "Links";
  DeviceLinks = DeviceCfg >> "Links";
  if( DeviceLinks.IsCorrect() ){
    DeviceLinks |= "CCD-Device";
    DeviceLinks >> "CCD-Device" |= attr("Class","sCCDDriver");
  };
  DeviceCfg |= "Settings";
  DeviceSettings = DeviceCfg >> "Settings";
  if( DeviceSettings.IsCorrect() ){
    //изменение умолчательных шкал для sScanner
    sChartScales      ChartScales(sChartScale("Wavelength"
                                              ,sPhysRange(200.0,1000.0,"nm"))
                                 ,sChartScale("Intensity"
                                             ,sPhysRange(0.0,70000.0,_Unitsless_))
                                 );
    ChartScales.StoreSettings(DeviceSettings);

    DeviceSettings |= attr("IntegrationTime",sPhysValue(0.1,0.001,"s"));
    DeviceSettings |= attr("IntegrationTimeRange",sPhysRange(0.01,1.000,"s"));
    DeviceSettings |= attr("NumOfSamples",10);
    DeviceSettings |= attr("NumOfAcquisitions",100);
  };
  sScanner::Preconfigure(DeviceCfg);
}

sCCDScaner::~sCCDScaner()
{
/*  Эти параметры не нужно каждый раз сохранять в установки устройства
  if( DeviceSettings.IsCorrect() ){
    DeviceSettings << attr("IntegrationTime",FldIntegrationTime->value())
                   << attr("NumOfSamples",FldNumOfSamples->value())
                   << attr("NumOfAcquisitions",FldNumOfAcquisitions->value());
  };
*/
}

sCCDScaner::sCCDScaner (sString    DevicetId
                       ,sNodePtr   DeviceCfg
                       ,sNodePtr   WidgetCfg
                       ,QWidget *  parent   )
    :sScanner(DevicetId,DeviceCfg,WidgetCfg,parent)
    ,CCDDevice(this)
    ,ScanMode(cMonitoring)
    ,NumOfSamples(0)
    ,CurrentSamplingNo(0)
    ,NumOfAcquis(0)
    ,CurrentAcquisNo(0)
    ,NumOfPixels(0)
    ,SubtractConstant(false)
    ,BackConstant(0)
    ,Data(NULL)
    ,BackgroundData(NULL)
    ,CorrectionData(NULL)
{
    Multiplier.X = sPhysValue(1,"nm");
    //evolution of the inherited state machine
    {
      //evolution of the State_Warming
      {
        State_W_CCD_Connecting = new sState(State_Warming(),20000,"State_W_CCD_Connecting");
        State_W_CCD_Connecting
            ->SetAsInitial()
            ->OnEnterCall(SLOT(LinkToCCDDevice()))
            ->AddTransition<sCCDDriver::sDoubleArray>(SIGNAL(CCDWavelengths(sCCDDriver::sDoubleArray))
                                                     ,SLOT(AcceptWavelengths(sCCDDriver::sDoubleArray))
                                                     ,State_Warming()->Fin())
            ->AddTransition(SIGNAL(ErrorMessage()),State_Cooling())
            ->AddTransition(SIGNAL(CCDNoAnswer()),SLOT(ReportNoAnswerError()),State_Cooling());
      };
      // evolution of State_O_Scanning
      {
        State_O_Scanning->OnExitCall(SLOT(ReleaseControl()));
        State_O_Idle
            ->AddReaction<sCCDDriver::sDoubleArray>(SIGNAL(CCDWavelengths(sCCDDriver::sDoubleArray))
                                                   ,SLOT(AcceptWavelengths(sCCDDriver::sDoubleArray)));

        State_O_Sc_Prepare = new sState(State_O_Scanning,"State_O_Sc_Prepare");
        State_O_Sc_Prepare
            ->SetAsInitial()
            ->OnEnterCall(SLOT(TakeControlAndSetParameters()))
            ->AddTransition(SIGNAL(DevicesBusy()),State_O_Idle)
            ->AddTransition(SIGNAL(CCDParamsSet()),State_O_Sc_Acquisition);
      };
      connect(this ,SIGNAL(CCDDeviceError(QString))
             ,this ,SLOT(ReportCCDDeviceError(QString)) );
      connect(State_W_CCD_Connecting,SIGNAL(Timeout()),this,SIGNAL(CCDNoAnswer()));
    };
}

void  sCCDScaner::AppearanceMakeUp ()
{
  sScanner::AppearanceMakeUp();

  sPhysRange _ITimeRange(0.01,1.000,"s");
  sPhysValue _ITime(0.1,"s");
  try{ //get config parameters
    get_from<sPhysValueNode>(DeviceSettings >> "IntegrationTime" ,_ITime);
    get_from<sPhysRangeNode>(DeviceSettings >> "IntegrationTimeRange" ,_ITimeRange);
    get_from<sIntegerNode>(DeviceSettings >> "NumOfSamples" ,NumOfSamples);
    get_from<sIntegerNode>(DeviceSettings >> "NumOfAcquisitions" ,NumOfAcquis);
  }catch(...){
    SetInvalid(tr("Something wrong in configuration."));
  }
  //add new widgets
  {
    LblRangeTitle = new QLabel(ParamArea);
    LblRangeTitle->setObjectName("LblRangeTitle");
    ParamLayout->addWidget(LblRangeTitle,Row,0,1,1);
    LblWLRange = new QLabel(ParamArea);
    LblWLRange->setObjectName("LblWLRange");
    ParamLayout->addWidget(LblWLRange,Row++,1,1,-1);

    LblMode = new QLabel(ParamArea);
    LblMode->setObjectName("LblMode");
    ParamLayout->addWidget(LblMode,Row,0,1,1);
    FldMode = new QComboBox(ParamArea);
    FldMode->setObjectName("FldMode");
    FldMode->insertItem(cMonitoring,"Monitor");
    FldMode->insertItem(cAcquisition,"Acquisition");
    FldMode->insertItem(cEvolution,"Evolution");
    FldMode->setCurrentIndex(cMonitoring);
    ParamLayout->addWidget(FldMode,Row++,1,1,1);
    LblIntegrationTime = new QLabel(ParamArea);
    LblIntegrationTime->setObjectName("LblIntegrationTime");
    ParamLayout->addWidget(LblIntegrationTime,Row,0,1,1);
    FldIntegrationTime = new sPhysValueField(ParamArea
                                  ,sPhysValue(0.1,0.001,"s"),false,false,false);
    FldIntegrationTime->SetValidRange(_ITimeRange);
    ParamLayout->addWidget(FldIntegrationTime,Row++,1,1,3,Qt::AlignLeft);
    LblNumOfSamples = new QLabel(ParamArea);
    LblNumOfSamples->setObjectName("LblNumOfSamples");
    ParamLayout->addWidget(LblNumOfSamples,Row,0,1,1);
    FldNumOfSamples = new QSpinBox(ParamArea);
    FldNumOfSamples->setObjectName("FldNumOfSamples");
    FldNumOfSamples->setRange(1,1000);
    ParamLayout->addWidget(FldNumOfSamples,Row++,1,1,1);
    LblNumOfAcquisitions = new QLabel(ParamArea);
    LblNumOfAcquisitions->setObjectName("LblNumOfAcquisitions");
    ParamLayout->addWidget(LblNumOfAcquisitions,Row,0,1,1);
    FldNumOfAcquisitions = new QSpinBox(ParamArea);
    FldNumOfAcquisitions->setObjectName("FldNumOfAcquisitions");
    FldNumOfAcquisitions->setRange(1,1000000);
    ParamLayout->addWidget(FldNumOfAcquisitions,Row++,1,1,1);

    ChkBackConstant = new QCheckBox(ParamArea);
    ChkBackConstant->setObjectName("ChkBackConstant");
    ParamLayout->addWidget(ChkBackConstant,Row,0,1,1);
    FldBackConstant = new QSpinBox(ParamArea);
    FldBackConstant->setObjectName("FldBackConstant");
    FldBackConstant->setRange(0,65535);
    ParamLayout->addWidget(FldBackConstant,Row++,1,1,1);

    ChkBackground = new QCheckBox(ParamArea);
    ChkBackground->setObjectName("ChkBackground");
    ParamLayout->addWidget(ChkBackground,Row,0,1,1);
    FldBackground = new sDropNodeWidget(ParamArea);
    FldBackground->setObjectName("FldBackground");
    ParamLayout->addWidget(FldBackground,Row++,1,1,8);
    FldBackground->Prompt=FromQString(tr(
                      "In the workbook find the relevant object "
                      "and drag and drop it here"));

    ChkCorrection = new QCheckBox(ParamArea);
    ChkCorrection->setObjectName("ChkCorrection");
    ParamLayout->addWidget(ChkCorrection,Row,0,1,1);
    FldCorrection = new sDropNodeWidget(ParamArea);
    FldCorrection->setObjectName("FldCorrection");
    ParamLayout->addWidget(FldCorrection,Row++,1,1,8);
    FldCorrection->Prompt=FromQString(tr(
                      "In the workbook find the relevant object "
                      "and drag and drop it here"));
  }

  FldIntegrationTime->SetValue(_ITime);
  FldNumOfSamples->setValue(NumOfSamples);
  FldNumOfAcquisitions->setValue(NumOfAcquis);
  Translate();
}


void  sCCDScaner::Translate ()
{
  LblRangeTitle->setText(tr("WL Range"));
  LblMode->setText(tr("Mode"));
  LblIntegrationTime->setText(tr("Integration time"));
  LblNumOfSamples->setText(tr("Number of samples"));
  LblNumOfAcquisitions->setText(tr("Number of acquisitions"));
  ChkBackConstant->setText(tr("Background value"));
  ChkBackground->setText(tr("Background curve"));
  ChkCorrection->setText(tr("Correction curve"));
}


void  sCCDScaner::RestoreState ()
{
  sScanner::RestoreState();
  if( WidgetConfig.IsCorrect() ){
    sPhysValue _ITime(0.1,"s");
    try{ //get stored parameters
      get_from<sPhysValueNode>(WidgetConfig >> "IntegrationTime" ,_ITime);
      get_from<sIntegerNode>(WidgetConfig >> "NumOfSamples" ,NumOfSamples);
      get_from<sIntegerNode>(WidgetConfig >> "NumOfAcquisitions" ,NumOfAcquis);

      bool Chkd = false;
      get_from<sBooleanNode>(WidgetConfig >> "BackConstantChkd" ,Chkd);
      ChkBackConstant->setChecked(Chkd);
      int BackValue = 0;
      get_from<sIntegerNode>(WidgetConfig >> "BackConstantValue" ,BackValue);
      FldBackConstant->setValue(BackValue);
      Chkd = false;
      get_from<sBooleanNode>(WidgetConfig >> "BackgroundChkd" ,Chkd);
      ChkBackground->setChecked(Chkd);
      FldBackground->SetValue(WidgetConfig >> "Background");
      Chkd = false;
      get_from<sBooleanNode>(WidgetConfig >> "CorrectionChkd" ,Chkd);
      ChkCorrection->setChecked(Chkd);
      FldCorrection->SetValue(WidgetConfig >> "Correction curve");
    }catch(...){
      SetInvalid(tr("Something wrong in desktop stored."));
    }

    FldIntegrationTime->SetValue(_ITime);
    FldNumOfSamples->setValue(NumOfSamples);
    FldNumOfAcquisitions->setValue(NumOfAcquis);
  };
}

void  sCCDScaner::StoreState ()
{
  sScanner::StoreState();
  if( WidgetConfig.IsCorrect() ){
    WidgetConfig << attr("IntegrationTime",FldIntegrationTime->Value())
                 << attr("NumOfSamples",FldNumOfSamples->value())
                 << attr("NumOfAcquisitions",FldNumOfAcquisitions->value());
    WidgetConfig << attr("BackConstantChkd", ChkBackConstant->isChecked ());
    WidgetConfig << attr("BackConstantValue",FldBackConstant->value());
    WidgetConfig << attr("BackgroundChkd", ChkBackground->isChecked ());
    WidgetConfig.DelAttr("Background");
    if( ! FldBackground->Value.IsNULL() ){
      WidgetConfig << attr("Background",FldBackground->Value);
    };
    WidgetConfig << attr("CorrectionChkd", ChkCorrection->isChecked ());
    WidgetConfig.DelAttr("Correction curve");
    if( ! FldCorrection->Value.IsNULL() ){
      WidgetConfig << attr("Correction curve",FldCorrection->Value);
    };
  };
}

void sCCDScaner::ReportCCDDeviceError (QString  error_text)
{
  emit ErrorMessage( tr("Got error from '%1':\n\n%2")
                       .arg(ToQString(CCDDevice->DeviceIdentifier()))
                       .arg(error_text)     );
}

void sCCDScaner::LinkToCCDDevice ()
{
  try{
    CCDDevice.SetTo(DeviceLinks,"CCD-Device");
    Connect(CCDDevice ,SIGNAL(ErrorMessage(QString)) ,SIGNAL(CCDDeviceError(QString)) );
    Connect(CCDDevice ,SIGNAL(IsBusy()),SIGNAL(DevicesBusy()) );
    Connect(CCDDevice ,SIGNAL(ParamsSet()),SIGNAL(CCDParamsSet()) );
    Connect(CCDDevice ,SIGNAL(NewWavelengths(sCCDDriver::sDoubleArray))
                      ,SIGNAL(CCDWavelengths(sCCDDriver::sDoubleArray)) );
  }catch( rxDeviceLaunch E ){
    emit ErrorMessage( tr("Error while linking to a submissive device.\n"
                          "%1\nCheck configuration.")
                         .arg(ToQString(E.MsgText()))   );
  }
}

void sCCDScaner::ReportNoAnswerError ()
{
    emit ErrorMessage( tr("No wavelengths got from '%1'")
                         .arg(ToQString(CCDDevice->DeviceIdentifier())));
}

void sCCDScaner::AcceptWavelengths(sCCDDriver::sDoubleArray _WLs)
{
  NumOfPixels = _WLs.Size();
  Wavelengths = _WLs;
  LblWLRange->setText(ToQString(sPhysRange(sPhysValue(Wavelengths[0],"nm")
                               ,sPhysValue(Wavelengths[NumOfPixels-1],"nm")).Text()));
//  qDebug() << "sCCDScaner::AcceptWavelengths";
}

void sCCDScaner::TakeControlAndSetParameters()
{
  if( ! CCDDevice.TakeExclusiveControl() ){
    QMessageBox::critical(this,tr("Denied")
                         ,tr("Device '%1' is busy.\nTry again later.")
                         .arg(ToQString(CCDDevice->DeviceIdentifier())) );
    emit DevicesBusy();
  }else{
    sCCDDriver::sParameters Params;
    Params.IntegrationTime = FldIntegrationTime->Value();
    Params.Gain = 0;
    Params.TriggerEdge = 0;
    Params.TriggerMode = 0;
    SendDirective<sCCDDriver::sParameters>(CCDDevice,sCCDDriver::cSetParameters,Params);
  };
}

void  sCCDScaner::ReleaseControl ()
{
  CCDDevice.ReleaseExclusiveControl();
}

void  sCCDScaner::InitializeScanning ()
{
  ScanMode = (eScanMode)( FldMode->currentIndex() );
  sScanner::InitializeScanning ();
}

void  sCCDScaner::CreateNewDependence ()
{
  if( Multiplier.X.IsNaN() ){
      Multiplier.X = sPhysValue("nm");
  };
  if( Multiplier.Y.IsNaN() ){
    Multiplier.Y = sPhysValue(_Unitsless_);
  };
  if( ScanMode == cEvolution ){
    Dependence = sNodePtr(new sRecord());
    SubSpectrum =
        sNodePtr(new sDataTableFolder(""/*docket*/
                                    ,sPhysPair(Multiplier.X,Multiplier.Y)
                                    ,NumOfPixels));
  }else{
    SpectrumDependence = sNodePtr(new sDataTableRecord(NumOfPixels));
    SpectrumDependence->SetMultiplier(sPhysPair(Multiplier.X,Multiplier.Y));
    Dependence = SpectrumDependence;
  }
}

void  sCCDScaner::StartScanning ()
{
  qDebug() << "Begin starting CCD aq.";
  SubtractConstant = false;
  if (ChkBackConstant->isChecked()){
    BackConstant=FldBackConstant->value();
    SubtractConstant = true;
  }
  if( ChkBackground->isChecked() ){
    pcsFunction      Func = sNetMath::GetMathFunctionFrom(FldBackground->Value);
    if( FldBackground->Value.IsNULL() ){
      QMessageBox::critical(this
          ,tr("CCD scanner: Background subtraction won't be made")
          ,tr("To make it work, drag a suitable object into "
              "the corresponding field."));
    }else if( ! Func ){
      QMessageBox::critical(this
          ,tr("CCD scanner: Background subtraction won't be made")
          ,tr("Given object doesn't look like a function."));
    }else{
      try{
        BackgroundData = new double[NumOfPixels];
        for (int i = 0; i < NumOfPixels; i++) {
          double YY;
          YY = ( Func->Value(Wavelengths[i]*Multiplier.X)
               / Multiplier.Y
               ).ToReal();
          BackgroundData[i] = YY;
        }
      }catch( rxRAlgebra ){
        if( BackgroundData ){ delete [] BackgroundData; BackgroundData=NULL; }
        QMessageBox::critical(this
            ,tr("CCD scanner: Background subtraction won't be made")
            ,tr("Perhaps, the given function has wrong physical units."));
      }
    }
  }
//! @todo{math} Почему-то не получается доступиться к виртуальным функциям
//!             путём прямого приведения:
//  if( ChkCorrection->isChecked() ){
//    qDebug() << "Correction Dependence - forming started";
//    qDebug() << "NumOfPixels" << NumOfPixels;
//    sPtr<sFunction> GivenFunc;
//    GivenFunc = FldCorrection->Value;
//    if (GivenFunc.IsCorrect()) {
//      qDebug() << "Correction Dependence - GivenFunc is correct";
//      pcsFunction  Func(dynamic_cast<pcsFunction>(GivenFunc.operator ->()));
//      qDebug() << "Func is null" << (Func==NULL);
//      qDebug() << "Func->AbscissaRange"
//               << ToQString(Func->AbscissaRange().Text());
//      CorrectionData = new double[NumOfPixels];
//      for (int i = 0; i < NumOfPixels; i++) {
//        double YY;
//        YY = ( Func->Value(Wavelengths[i]*Multiplier.X)
//             / Multiplier.Y
//             ).ToReal();
//        qDebug() << "Y = " << YY;
//        CorrectionData[i] = YY;
//      }
//      qDebug() << "Correction Dependence is formed";
//    }
//  }
  if( ChkCorrection->isChecked() ){
    pcsFunction      Func = sNetMath::GetMathFunctionFrom(FldCorrection->Value);
    if( FldCorrection->Value.IsNULL() ){
      QMessageBox::critical(this,tr("CCD scanner: Correction won't be made")
          ,tr("To make it work, drag a suitable object into "
              "the corresponding field."));
    }else if( ! Func ){
      QMessageBox::critical(this,tr("CCD scanner: Correction won't be made")
          ,tr("Given object doesn't look like a function."));
    }else{
      try{
        CorrectionData = new double[NumOfPixels];
        for (int i = 0; i < NumOfPixels; i++) {
          double YY;
          YY = ( Func->Value(Wavelengths[i]*Multiplier.X)
               / Multiplier.Y
               ).ToReal();
          if (isnan(YY)) {YY=0;}
          CorrectionData[i] = YY;
        }
      }catch( rxRAlgebra ){
        if( CorrectionData ){ delete [] CorrectionData; CorrectionData=NULL; }
        QMessageBox::critical(this,tr("CCD scanner: Correction won't be made")
            ,tr("Perhaps, the given function has wrong physical units."));
      }
    }
  }
  CurrentSamplingNo = 0;
  CurrentAcquisNo = 0;
  NumOfSamples = FldNumOfSamples->value();
  NumOfAcquis = FldNumOfAcquisitions->value();

  Data = new ulong[NumOfPixels];
  for (int i = 0; i < NumOfPixels; i++) {
    Data[i] = 0;
    if( ScanMode == cEvolution ){
      SubSpectrum->SetValue(i, sPoint(Wavelengths[i],0.0));
    }else{
      SpectrumDependence->SetValue(i, sPoint(Wavelengths[i],0.0));
    }
  }
  StartTime = sTime();
  Connect(CCDDevice ,SIGNAL(NewData (sCCDDriver::sUShortArray))
         ,SLOT(AddNewData(sCCDDriver::sUShortArray)) );
  SendCommand(CCDDevice,sCCDDriver::cStart);

//  qDebug() << "CCD Scan is started";
}

void  sCCDScaner::StopScanning ()
{
  qDebug() << "Scan is stopping 1";
  SendCommand(CCDDevice,sCCDDriver::cStop);
  Disconnect(CCDDevice ,SIGNAL(NewData (sCCDDriver::sUShortArray))
            ,SLOT(AddNewData(sCCDDriver::sUShortArray)) );
  if( Data ){ delete [] Data; Data=NULL; }
  if( BackgroundData ){ delete [] BackgroundData; BackgroundData=NULL; }
  if( CorrectionData ){ delete [] CorrectionData; CorrectionData=NULL; }
  qDebug() << "Scan stoped";
}

void  sCCDScaner::EarlySetAttributes ()
{
  sScanner::EarlySetAttributes ();
  Dependence << attr("Mode",FromQString(FldMode->currentText()))
             << attr("Integration time",FldIntegrationTime->Value())
             << attr("Number of samples", FldNumOfSamples->value());
  if( ScanMode == cEvolution ){
    Dependence << attr(1,SubSpectrum);
    if( ! LineColorSet->Empty() ){
      LineColorSet->SetToFirst();
      SubSpectrum
          ->SetAppearance(new sCurveAppearance(LineColorSet->CurrentColor()
                                              ,LineThickness));
    }
  }
}

void  sCCDScaner::LateSetAttributes ()
{
  sScanner::LateSetAttributes ();
  if (ScanMode == cAcquisition){
      Dependence << attr("Number of acquisitions", CurrentAcquisNo);
  }
}

void  sCCDScaner::AddNewData (sCCDDriver::sUShortArray _aSample)
{
//  RANet::Log.Put(sLog::Info,"CCDScaner","Begin Add New Data");
  if((ScanMode == cMonitoring) && ! SpectrumDependence.IsCorrect() ){
    return;
  }
  if((ScanMode == cAcquisition) && ! SpectrumDependence.IsCorrect() ){
    return;
  }
  if((ScanMode == cEvolution) && ! SubSpectrum.IsCorrect() ){
    return;
  }
  if( Multiplier.X.IsNaN() ){
    return;
  };
  if( Multiplier.Y.IsNaN() ){
    return;
  };
  if( CurrentSamplingNo < NumOfSamples){
    for (int i = 0; i < NumOfPixels; i++) {
      Data[i] = Data[i] + _aSample[i];
    }
    CurrentSamplingNo++;
  }else{
    CurrentSamplingNo = 0;
//    double N = NumOfSamples;
//    N = N*IntegrationTime/100000; Деление на IntegrationTime кажется разумным,
//                                  но на деле оказывается вредным так как
//                                  приводит к странному результату.
    switch (ScanMode)
    {
      case cMonitoring : {
        for (int i = 0; i < NumOfPixels; i++) {
          double Y = Data[i];
          Y = Y / NumOfSamples;
          if (SubtractConstant) {
            Y = Y - BackConstant;
          };
          if( BackgroundData ){
            Y = Y - BackgroundData[i];
          };
          if( CorrectionData ){
            Y = Y * CorrectionData[i];
          }
          SpectrumDependence->SetYValue(i,Y);
          Data[i] = 0;
        }
        break;
      }
      case cAcquisition : {
        CurrentAcquisNo++;
        for (int i = 0; i < NumOfPixels; i++) {
          double Y = Data[i];
          Y = Y / NumOfSamples;
          if (SubtractConstant) {
            Y = Y - BackConstant;
          };
          if( BackgroundData ){
            Y = Y - BackgroundData[i];
          };
          if( CorrectionData ){
            Y = Y * CorrectionData[i];
          }
          Y = (SpectrumDependence->PointY(i)*(CurrentAcquisNo-1)
                                        + Y)/CurrentAcquisNo;
          SpectrumDependence->SetYValue(i, Y);
          Data[i] = 0;
        }
        if (!(CurrentAcquisNo < NumOfAcquis)){
          emit ScanningHasCompleted();
        }
        break;
      }
      case cEvolution : {
        CurrentAcquisNo++;
        for (int i = 0; i < NumOfPixels; i++) {
          double Y = Data[i];
          Y = Y / NumOfSamples;
          if (SubtractConstant) {
            Y = Y - BackConstant;
          };
          if( BackgroundData ){
            Y = Y - BackgroundData[i];
          };
          if( CorrectionData ){
            Y = Y * CorrectionData[i];
          }
          SubSpectrum->SetYValue(i,Y);
          Data[i] = 0;
        }
        SubSpectrum << attr("&Delta;t",sPhysValue((sTime()- StartTime),"s"));
        SubSpectrum << attr("series",Dependence);
        if( ! PreviousSubSpectrum.IsNULL() )
          Content.DelAttr(PreviousSubSpectrum);
        PreviousSubSpectrum = SubSpectrum;
        Content << PreviousSubSpectrum;
        if (!(CurrentAcquisNo < NumOfAcquis)){
          emit ScanningHasCompleted();
        } else {
          SubSpectrum =
              sNodePtr(new sDataTableFolder(""/*docket*/
                                           ,sPhysPair(Multiplier.X
                                                     ,Multiplier.Y)
                                           ,NumOfPixels));
          for (int i = 0; i < NumOfPixels; i++)
            SubSpectrum->SetValue(i, sPoint(Wavelengths[i],0.0));
          Dependence << attr(CurrentAcquisNo+1,SubSpectrum);
          if( ! LineColorSet->Empty() ){
            LineColorSet->ShiftToNext();
            SubSpectrum
                ->SetAppearance(new sCurveAppearance(LineColorSet
                                                            ->CurrentColor()
                                                    ,LineThickness));
          }
          StartTime = sTime();
        }
        break;
      }
    }
  };
//  RANet::Log.Put(sLog::Info,"CCDScaner","Finish Add New Data");
}

//------------------------------------------------------------------------------
} //namespace RA
