////////////////////////////////////////////////////////////////////////////////
/*! @file EffectorConsole.cpp Консоль управления эффектором.
- Part of BasicDevices - Research Assistant Plugin package.
- Uses  RAGUI - Research Assistant Graphical User Interface Library.
- Uses  Qt v.5.0.2  - http://qt-project.org/
- Uses  RANet - Research Assistant Net Library.
- Copyright(C) 2013-2014, Viktor E. Bursian, St.Petersburg, Russia.
                          Viktor.Bursian@mail.ioffe.ru
*///////////////////////////////////////////////////////////////////////////////
#include "EffectorConsole.h"
#include "attr.h"
#include "StringsAndQStrings.h"
#include "SymbolsGUI.h"
#include <QVBoxLayout>
#include <QHBoxLayout>
#include <QMessageBox>
namespace RA {
//------------------------------------------------------------------------------

DEFINE_CLASS_TAG(sEffectorConsole)

//--------------------------------------------------------- sEffectorConsole ---

void  sEffectorConsole::Preconfigure (sNodePtr  DeviceCfg)
{
  sPtr<sFolderNode>              DeviceLinks;
  DeviceCfg |= "Links";
  DeviceLinks = DeviceCfg >> "Links";
  if( DeviceLinks.IsCorrect() ){
    DeviceLinks |= "Effector";
    DeviceLinks >> "Effector" |= attr("Class","sEffector");
  };
  sPtr<sFolderNode>              DeviceSettings;
  DeviceCfg |= "Settings";
  DeviceSettings = DeviceCfg >> "Settings";
  if( DeviceSettings.IsCorrect() ){
    DeviceSettings |= attr("DoSetWithButton",false);
  };
  sConsole::Preconfigure(DeviceCfg);
}


sEffectorConsole::~sEffectorConsole ()
{
  Effector.SetToNULL();
}


sEffectorConsole::sEffectorConsole (sString   DeviceId
                                   ,sNodePtr  DeviceCfg
                                   ,sNodePtr  WidgetCfg
                                   ,QWidget *  parent    )
    :sConsole(DeviceId,DeviceCfg,WidgetCfg,parent)
    ,DoSetWithButton(false)
    ,Effector(this)
{
  //evolution of the inherited state machine
  {
    //evolution of the State_Warming
    {
      State_W_Effector 
          = (new sState(State_Warming(),"State_W_Effector"))
          ->SetAsInitial()
          ->AddTransition(State_Warming()->Fin());
      //State_W_Effector internal structure and behaviour
      {
        State_W_Effector_Linking 
            = (new sState(State_W_Effector ,"State_W_Effector_Linking"))
            ->SetAsInitial()
            ->OnEnterCall( SLOT(LinkToEffector()) )
            ->AddTransition( SIGNAL(EffectorLinked()) ,State_W_Effector->Fin() )
            ->AddTransition( SIGNAL(ErrorMessage()) ,State_Cooling() );
      }
    }
    //State_Operating internal structure and behaviour
    {
      State_O_Watching = new sState(State_Operating(),"State_O_Watching");
      State_O_Controling = new sState(State_Operating(),"State_O_Controling");
      
      State_O_Watching
          ->SetAsInitial()
          ->AddReaction(SIGNAL(SetValue()) ,SLOT(OnSetValue())   )
          ->AddTransition(SIGNAL(EffectorControlTaken()) ,State_O_Controling);

      State_O_Controling
          ->AddTransition(State_O_Watching)
          ->OnExitCall(SLOT(ReleaseEffectorControl()))
          ;
      
      //State_O_Controling internal structure and behaviour
      {
        State_O_C_Setting = new sState(State_O_Controling,"State_O_C_Waiting");
        State_O_C_Breaking= new sState(State_O_Controling,"State_O_C_Breaking");
        
        State_O_C_Setting 
            ->SetAsInitial()
            ->AddTransition(SIGNAL(Break()) ,SLOT(OnBreak())
                           ,State_O_C_Breaking )
            ->AddTransition(SIGNAL(EffectorSaysDone())
                           ,State_O_Controling->Fin() )
            ;
        State_O_C_Breaking
            ->AddTransition(SIGNAL(EffectorSaysDone())
                           ,State_O_Watching)
            ;
      }
    }
    connect(this ,SIGNAL(EffectorError(QString))
           ,this ,SLOT(ReportEffectorError(QString)) );
  }
}


void  sEffectorConsole::AppearanceMakeUp ()
{
  sConsole::AppearanceMakeUp();
  try{ //get config parameters
    get_from<sBooleanNode>(DeviceSettings>>"DoSetWithButton",DoSetWithButton);
  }catch(...){
    SetInvalid(tr("Something wrong in configuration."));
  }

  Screen = new QLabel(this);
  SetValueField = new sPhysValueField(this,sPhysValue());
  BtnSet = new QPushButton(this);
  BtnBreak = new QPushButton(this);
  
  QBoxLayout *                Layout = new QVBoxLayout();
  QBoxLayout *                CtrlsLayout = new QHBoxLayout();
  setLayout(Layout);
  Layout->addWidget(Screen);
  Layout->addItem(CtrlsLayout);
  CtrlsLayout->setSpacing(3);
  CtrlsLayout->addWidget(SetValueField);
  CtrlsLayout->addWidget(BtnSet);
  CtrlsLayout->addWidget(BtnBreak);
  CtrlsLayout->addStretch(1);
  Layout->addStretch(1);
  
  ShowNaN();
  BtnSet->setText(Symbol::Play);
  BtnSet->setMaximumWidth(BtnSet->height());
  BtnSet->setVisible(DoSetWithButton);
  BtnBreak->setText(Symbol::Stop);
  BtnBreak->setMaximumWidth(BtnBreak->height());
  BtnBreak->setVisible(false);
  
  connect(BtnBreak ,SIGNAL(clicked()) ,this ,SIGNAL(Break()));
  State_O_Controling
      ->OnEnterCall(SetValueField ,SLOT(SetReadOnly()) )
      ->OnEnterCall(BtnBreak ,SLOT(show()) )
      ->OnExitCall(BtnBreak ,SLOT(hide()) )
      ->OnExitCall(SetValueField ,SLOT(UnsetReadOnly()) )
      ;
  State_O_C_Breaking
      ->OnEnterCall(BtnBreak ,SLOT(hide()) )
      ;
  if( DoSetWithButton ){
    connect(BtnSet ,SIGNAL(clicked()) ,this ,SIGNAL(SetValue()));
    State_O_Controling
        ->OnEnterCall(BtnSet ,SLOT(hide()) )
        ->OnExitCall(BtnSet ,SLOT(show()) )
        ;
  }else{
    connect(SetValueField ,SIGNAL(Edited()) ,this ,SIGNAL(SetValue()));
  }
}


void  sEffectorConsole::RestoreState ()
{
  sConsole::RestoreState();
  if( WidgetConfig.IsCorrect() ){
    sPhysValue                  SetValue;
    if( get_from<sPhysValueNode>(WidgetConfig >> "SetValue" ,SetValue) )
      SetValueField->SetValue(SetValue);
  }
}


void  sEffectorConsole::StoreState ()
{
  sConsole::StoreState();
  if( WidgetConfig.IsCorrect() ){
    WidgetConfig << attr("SetValue",SetValueField->Value());
  }
}


void  sEffectorConsole::ShowData (sPhysValue  V)
{
  Screen->setText( QString::fromUtf8("<font size=20 color=blue>")
                 + ToQString(V.Text(HTML))
                 );
}


void  sEffectorConsole::ShowNaN ()
{
  Screen->setText( QString::fromUtf8("<font size=20 color=red>")
                   + "--------");
}


void  sEffectorConsole::ReportEffectorError (QString  error_text)
{
  emit ErrorMessage( tr("Got error from '%1':\n\n%2")
                       .arg(ToQString(Effector->DeviceIdentifier()))
                       .arg(error_text)     );
}


void  sEffectorConsole::LinkToEffector ()
{
  try{
    Effector.SetTo(DeviceLinks,"Effector");
    Connect(Effector ,SIGNAL(Warmed())       ,SIGNAL(EffectorLinked()) );
    Connect(Effector ,SIGNAL(ErrorMessage(QString))
           ,SIGNAL(EffectorError(QString)) );
    Connect(Effector ,SIGNAL(RangeChanged(sPhysRange))
           ,SLOT(ApplyRange(sPhysRange)) );
  }catch( rxDeviceLaunch E ){
    emit ErrorMessage( tr("Error while linking to a submissive device.\n"
                          "%1\nCheck configuration.")
                         .arg(ToQString(E.MsgText()))   );
  }
}


void  sEffectorConsole::OnEnterToCoolingState ()
{
  Disconnect(Effector,0,0);      //disconnect all signals from Effector,
                                 //which were connected to any of my slots
  Effector.ReleaseExclusiveControl();
  Effector.SetToNULL();
  sConsole::OnEnterToCoolingState();
}


void  sEffectorConsole::OnEnterToOperatingState()
{
  sConsole::OnEnterToOperatingState();
  Connect(Effector ,SIGNAL(NewData(sPhysValue))
         ,SLOT(ShowData(sPhysValue)) );
  Connect(Effector ,SIGNAL(Done(sPhysValue))
         ,SLOT(ShowData(sPhysValue)) );
  Connect(Effector ,SIGNAL(Done()) ,SIGNAL(EffectorSaysDone()) );
  SetValueField->SetValidRange(Effector->Range());
}


void  sEffectorConsole::OnExitFromOperatingState()
{
  Disconnect(Effector ,SIGNAL(NewData(sPhysValue))
            ,SLOT(ShowData(sPhysValue)) );
  Disconnect(Effector ,SIGNAL(Done(sPhysValue))
            ,SLOT(ShowData(sPhysValue)) );
  Disconnect(Effector ,SIGNAL(Done()) ,SIGNAL(EffectorSaysDone()) );
  ShowNaN();
  sConsole::OnExitFromOperatingState();
}


void  sEffectorConsole::OnSetValue()
{
  if( SendDirective(Effector ,sEffector::cSetTo ,SetValueField->Value()
                   ,cTakeControl) ){
    emit EffectorControlTaken();
  };
}


void  sEffectorConsole::OnBreak()
{
  SendCommand(Effector ,sEffector::cBreak);
}


void  sEffectorConsole::ReleaseEffectorControl ()
{
  Effector.ReleaseExclusiveControl();
}


void  sEffectorConsole::ApplyRange (sPhysRange  new_range)
{
  SetValueField->SetValidRange(new_range);
}

//------------------------------------------------------------------------------
} //namespace RA
