////////////////////////////////////////////////////////////////////////////////
/*! @file FixedUnitsLCD.cpp Жидко-кристаллический дисплей с фиксированными единицами.
- Part of BasicDevices - Research Assistant Plugin package.
- Uses  RAGUI - Research Assistant Graphical User Interface.
- Uses  QtGui v.4.6  -  http://qt.nokia.com/
- Uses  RANet - Research Assistant Net Library (based on ANSI C++).
- Copyright(C) 2010-2014, Dmitry K. Nelson, St.Petersburg, Russia.
                          D.Nelson@mail.ioffe.ru
- Copyright(C) 2013,      Viktor E. Bursian, St.Petersburg, Russia.
                          Viktor.Bursian@mail.ioffe.ru
*///////////////////////////////////////////////////////////////////////////////
#include "FixedUnitsLCD.h"
#include "attr.h"
#include "StringsAndQStrings.h"
#include <QHBoxLayout>
#include <QVBoxLayout>
#include <QSpacerItem>
#include <QMessageBox>
namespace RA {
//------------------------------------------------------------------------------

DEFINE_CLASS_TAG(sFixedUnitsLCD)

//----------------------------------------------------------- sFixedUnitsLCD ---

void  sFixedUnitsLCD::Preconfigure (sNodePtr  DeviceCfg)
{
  sPtr<sFolderNode>              DeviceSettings;
  DeviceCfg |= "Settings";
  DeviceSettings = DeviceCfg >> "Settings";
  if( DeviceSettings.IsCorrect() ){
    DeviceSettings |= attr("DisplayUnits","photon/s");
    DeviceSettings |= attr("DigitsAfterTheDot",0);
  };
  sGaugeDisplay::Preconfigure(DeviceCfg);
}


sFixedUnitsLCD::sFixedUnitsLCD (sString   DeviceId
                               ,sNodePtr  DeviceCfg
                               ,sNodePtr  WidgetCfg
                               ,QWidget *  parent    )
    :sGaugeDisplay(DeviceId,DeviceCfg,WidgetCfg,parent)
    ,UnitsToShow("")
    ,DigitsAfterTheDot(0)
    ,Multiplier(1.0)
{
}


void  sFixedUnitsLCD::AppearanceMakeUp ()
{
  try{ //get config parameters
    get_from<sTextNode>(DeviceSettings >> "DisplayUnits" ,UnitsToShow);
    get_from<sIntegerNode>(DeviceSettings >> "DigitsAfterTheDot" 
                                             ,DigitsAfterTheDot);
    if( DigitsAfterTheDot < 0 ){
      DigitsAfterTheDot = 0;
    }else if( DigitsAfterTheDot > 9 ){
      DigitsAfterTheDot = 9;
    }
  }catch(...){
    SetInvalid(tr("Something wrong in configuration."));
  }

  Multiplier = power10(DigitsAfterTheDot);

  setLayout(new QVBoxLayout);

  ValueLCD = new QLCDNumber(this);
  ValueLCD->setSegmentStyle(QLCDNumber::Flat);
  layout()->addWidget(ValueLCD);

  QHBoxLayout *               L( new QHBoxLayout );
  layout()->addItem(L);
  UnitsDisplay = new QLabel(this);
  L->addStretch();
  L->addWidget(UnitsDisplay);
  UnitsDisplay->setSizePolicy(QSizePolicy::Fixed,QSizePolicy::Fixed);

  ShowNaN();
}


void sFixedUnitsLCD::ShowData (sPhysValue  V)
{
  integer                     VI;
  QString                     T;
  try{
    VI = Round( (V/sPhysValue(1.0,UnitsToShow)).ToReal() * Multiplier );
  }catch(rxRAlgebra){
    emit Failure(tr("Wrong units in the configuration settings.\n"
                    "Cannot convert '%1' into '%2'.")
                .arg(ToQString(V.Text()))
                .arg(ToQString(UnitsToShow))    );
    return;
  };
  if(   ( VI == std::numeric_limits<integer>::min() )
      ||( VI == std::numeric_limits<integer>::max() ) ){
    T = "Error";
  }else{
    T.setNum(VI);
    for( int i = 3+DigitsAfterTheDot ; i < T.length()-( VI<0 ? 1 : 0 ) ; i+=4 )
      T = T.left(T.length()-i) + ' ' + T.right(i);
    if( DigitsAfterTheDot > 0 ){
      T = T.left(T.length()-DigitsAfterTheDot)
          + '.' + T.right(DigitsAfterTheDot);
      for( int i = 3 ; i < DigitsAfterTheDot ; i+=3 )
        T = T.left(T.length()-DigitsAfterTheDot+i)
            + ' ' + T.right(DigitsAfterTheDot-i);
    }
  }
  ValueLCD->setDigitCount(T.length());
  ValueLCD->display(T);
  UnitsDisplay->setText(ToQString( sUnits(UnitsToShow).Text(HTML) ));
}


void sFixedUnitsLCD::ShowNaN ()
{
  ValueLCD->display(QString("-----"));
  UnitsDisplay->setText(ToQString( UnitsToShow ));
}

//------------------------------------------------------------------------------
} //namespace RA


