////////////////////////////////////////////////////////////////////////////////
/*! @file Gauge.cpp Abstract prototype of a gauge.
- Part of BasicDevices - Research Assistant Plugin package.
- Uses  RAGUI - Research Assistant Graphical User Interface Library.
- Uses  Qt v.5.0.2  - http://qt-project.org/
- Uses  RANet - Research Assistant Net Library.
- Copyright(C) 2010-2014, Viktor E. Bursian, St.Petersburg, Russia.
                          Viktor.Bursian@mail.ioffe.ru
*///////////////////////////////////////////////////////////////////////////////
#include "Gauge.h"
#include "attr.h"
namespace RA {
//------------------------------------------------------------------------------

DEFINE_CLASS_TAG(sGauge)

csCommand                     sGauge::cStart("sGauge::cStart");

csCommand                     sGauge::cStop("sGauge::cStop");

/*! Команда настройки прибора, в частности,начального состояния и т. д.
*/
csCommand                     sGauge::cTune("sGauge::cTune");

/*! Sets or unsets the auto-repeat mode.

If set to @a false, then each cStart command will produce @b one 
new measurement.

If set to @a true, then measurements are taken periodically until 
it gets cStop command.

May or may not work, depending on implementation.
May work in some states, being ignored in the others.

*/
const sDirective<bool>        sGauge::cSetAutoRepeatMode
                                           ("sGauge::cSetAutoRepeatMode");

/*! Changes the time interval between measurements.

Attached data: the new period in time units 
("s", "min", "hour", or self-defined).

May or may not work, depending on implementation.
May work in some states, being ignored in the others.

*/
const sDirective<sPhysValue>  sGauge::cSetPeriod("sGauge::cSetPeriod");

//------------------------------------------------------------------ sGauge ---

void  sGauge::Preconfigure (sNodePtr  DeviceCfg)
{
  sPtr<sFolderNode>           DeviceSettings;
  DeviceCfg |= "Settings";
  DeviceSettings = DeviceCfg >> "Settings";
  if( DeviceSettings.IsCorrect() ){
    DeviceSettings |= attr("AutoStart",true);
    DeviceSettings |= attr("AutoRepeat",true);
    DeviceSettings |= attr("Period",sPhysValue(1.0,"s"));
  };
  sDriver::Preconfigure(DeviceCfg);
}


sGauge::sGauge (sString   DevicetId
               ,sNodePtr  DeviceCfg)
    :sDriver(DevicetId,DeviceCfg)
    ,AutoRepeatFlag(true)
    ,AutoStartFlag(true)
{
}


void  sGauge::PostConstructor ()
{
  sDriver::PostConstructor();
  try{ //get config parameters
    get_from<sBooleanNode>(DeviceSettings >> "AutoStart"  ,AutoStartFlag);
    get_from<sBooleanNode>(DeviceSettings >> "AutoRepeat" ,AutoRepeatFlag);
    get_from<sPhysValueNode>(DeviceSettings >> "Period"   ,ThePeriod);
  }catch(...){
    SetInvalid(tr("Something wrong in configuration."));
  }
  //evolution of the inherited state machine (behaviour)
  {
    // evolution of the State_Operating
    {
      State_O_Measuring = new sState(State_Operating(),"State_O_Measuring");
      State_O_Measuring
          ->SetAsInitial()
          ;
      //State_O_Measuring internal structure and behaviour
      {
        State_O_M_Stopped 
            = new sState(State_O_Measuring,"State_O_M_Stopped");
        State_O_M_Acquiring 
            = new sState(State_O_Measuring,"State_O_M_Acquiring");
        State_O_M_Cycling 
            = new sState(State_O_Measuring,"State_O_M_Cycling");
        State_O_M_SettingAutoRepeatMode 
            = new sState(State_O_Measuring,"State_O_M_SettingAutoRepeatMode");
        State_O_M_SettingPeriod 
            = new sState(State_O_Measuring,"State_O_M_SettingPeriod");
        
        if( AutoStartFlag ){
          State_O_M_Acquiring->SetAsInitial();
        }else{
          State_O_M_Stopped->SetAsInitial();
        };
  
        State_O_M_Stopped
            ->OnEnterCall(SLOT(UnsetStartedFlag()))
            ->AddTransition(cStart ,State_O_M_Acquiring)
            ->AddTransition(cSetAutoRepeatMode
                           ,SLOT(MemorizeWantedAutoRepeatMode(bool))
                           ,State_O_M_SettingAutoRepeatMode)
            ->AddTransition(cSetPeriod
                           ,SLOT(MemorizeWantedPeriod(sPhysValue))
                           ,State_O_M_SettingPeriod)
            ;
        State_O_M_Acquiring
            ->OnEnterCall(SLOT(SetStartedFlag()))
            ->AddTransition(cStop ,State_O_M_Stopped)
            ;
        State_O_M_Cycling
            ->OnEnterCall(SLOT(OnCycling()))
            ->AddTransition( SIGNAL(RepeatAcquiring()) ,State_O_M_Acquiring )
            ->AddTransition( SIGNAL(StopAcquiring()) ,State_O_M_Stopped )
            ;
        State_O_M_SettingAutoRepeatMode 
            ->OnEnterCall(SLOT(SetWantedAutoRepeatMode()))
            ->AddTransition(SIGNAL(AutoRepeatModeChanged(bool))
                           ,State_O_M_Stopped)
            ;
        State_O_M_SettingPeriod 
            ->OnEnterCall(SLOT(SetWantedPeriod()))
            ->AddTransition(SIGNAL(PeriodChanged(sPhysValue))
                           ,State_O_M_Stopped)
            ;
      }
      /*! The descendants must provide the rest of functionality. 
          At least, it is necessary to add a transition from State_O_M_Acquiring
          to State_O_M_Cycling, and do not forget to emit NewData signal:
          @code
            State_O_M_Acquiring
                ->AddTransition<sPhysValue>(SIGNAL(GotMagneticField(sPhysValue)) 
                                           ,SIGNAL(NewData(sPhysValue)) 
                                           ,State_O_M_Cycling);
          @endcode
          It is also possible to move the whole State_O_Measuring state 
          to another hosting state:
          @code
            State_O_Measuring->setParent(State_O_Parallel);
          @endcode
      */
    };
  };
}


bool  sGauge::AutoRepeat () const
{
  return AutoRepeatFlag;
}


sPhysValue  sGauge::Period () const
{
  return ThePeriod;
}


void  sGauge::SetAutoRepeat (bool  auto_repeat)
{
  AutoRepeatFlag = auto_repeat;
}


void  sGauge::SetWantedAutoRepeatMode ()
{
  AutoRepeatFlag = WantedAutoRepeatMode;
  emit AutoRepeatModeChanged(AutoRepeatFlag);
}


void  sGauge::SetWantedPeriod ()
{
  ThePeriod = WantedPeriod;
  emit PeriodChanged(ThePeriod);
}


void  sGauge::OnCycling ()
{
  if( StartedFlag && AutoRepeatFlag ){
    emit RepeatAcquiring();
  }else{
    emit StopAcquiring();
  }
}

//------------------------------------------------------------------------------
} //namespace RA
