////////////////////////////////////////////////////////////////////////////////
/*! @file Gauge.h   Abstract prototype of a gauge.
- Part of BasicDevices - Research Assistant Plugin package.
- Uses  RAGUI - Research Assistant Graphical User Interface Library.
- Uses  Qt v.5.0.2  - http://qt-project.org/
- Uses  RANet - Research Assistant Net Library.
- Copyright(C) 2010-2014, Viktor E. Bursian, St.Petersburg, Russia.
                          Viktor.Bursian@mail.ioffe.ru
*///////////////////////////////////////////////////////////////////////////////
#ifndef Gauge_H
#define Gauge_H
#include "BasicDevices.h"
#include "Driver.h"
#include "PhysValue.h"
namespace RA {
//------------------------------------------------------------------------------

ANNOUNCE_ABSTRACT_CLASS(sGauge)

//------------------------------------------------------------------ sGauge ---
/*! Abstract prototype of a gauge.

Выдаёт отсчёт некой физвеличины
- сразу после запуска (см. AutoStart()) или по команде cStart;
- однократно или периодически, т.е. по внутреннему таймеру (см. AutoRepeat())

В последнем случае, может быть остановлен командой cStop,
а затем снова запущен командой cStart.

Наследники могут реализовывать только часть этого функционала,
например, только однократные измерения.

@par The state machine
@dotfile "Gauge.state_graph"

*/
class BasicDevices_EXPORT  sGauge : public sDriver
{
  Q_OBJECT
  ABSTRACT_DRIVER(sGauge)

  public: //commands & directives
    static csCommand          cStart;
    static csCommand          cStop;
    static csCommand          cTune; // Команда настройки прибора, в частности,
                                     // начального состояния и т. д.
    static const
      sDirective<bool>        cSetAutoRepeatMode;
    static const
      sDirective<sPhysValue>  cSetPeriod;

  public:
    virtual bool              AutoRepeat () const;
    virtual void              SetAutoRepeat (bool auto_repeat = true);
    virtual sPhysValue        Period () const;
    bool                      AutoStart () const
                                { return AutoStartFlag; }
    bool                      Started () const
                                { return StartedFlag; }

  signals: //outgoing
    void                      AutoRepeatModeChanged (bool);
    void                      PeriodChanged (sPhysValue);
    void                      NewData (sPhysValue);
    void                      TuneFinished(); //Сигнал, что настройка завершена

  public: //evolution of the inherited state machine
    psState                   State_O_Measuring;
    psState                   State_O_M_Stopped;
    psState                   State_O_M_Acquiring;
    psState                   State_O_M_Cycling;
    psState                   State_O_M_SettingPeriod;
    psState                   State_O_M_SettingAutoRepeatMode;

  protected: //reimplementing inherited
    virtual void              PostConstructor ();

  signals: //for internal needs
    void                      RepeatAcquiring ();
    void                      StopAcquiring ();
    
  protected slots:
// ///////////////////////////
//    virtual void              SetAutoRepeat (bool auto_repeat);
//    virtual void              SetPeriod (sPhysValue  period);
// ///////////////////////////
    virtual void              SetWantedAutoRepeatMode ();
    virtual void              SetWantedPeriod ();

  protected: //fields
    bool                      AutoRepeatFlag;
    sPhysValue                ThePeriod;
    bool                      WantedAutoRepeatMode;
    sPhysValue                WantedPeriod;
    
  private slots:
    void                      SetStartedFlag () { StartedFlag = true; }
    void                      UnsetStartedFlag () { StartedFlag = false; }
    void                      MemorizeWantedAutoRepeatMode (bool  mode)
                                { WantedAutoRepeatMode = mode; }
    void                      MemorizeWantedPeriod (sPhysValue  period)
                                { WantedPeriod = period; }
    void                      OnCycling ();

  private: //fields
    bool                      AutoStartFlag;
    bool                      StartedFlag;
};

//------------------------------------------------------------------------------
} //namespace RA
#endif
