////////////////////////////////////////////////////////////////////////////////
/*! @file PeriodicGaugeRecorder.cpp  Самописец.
- Part of BasicDevices - Research Assistant Plugin package.
- Uses  RAGUI - Research Assistant Graphical User Interface.
- Uses  QtGui v.4.6  -  http://qt.nokia.com/
- Uses  RANet - Research Assistant Net Library (based on ANSI C++).
- Copyright(C) 2010-2013, Viktor E. Bursian, St.Petersburg, Russia.
                          Viktor.Bursian@mail.ioffe.ru
*///////////////////////////////////////////////////////////////////////////////
#include "PeriodicGaugeRecorder.h"
#include "attr.h"
#include "StringsAndQStrings.h"
#include "Log.h"
#include <QMessageBox>
namespace RA {
//------------------------------------------------------------------------------

DEFINE_CLASS_TAG(sPeriodicGaugeRecorder)

//--------------------------------------------------- sPeriodicGaugeRecorder ---
/*! @class sPeriodicGaugeRecorder

@par The state machine

The upper-level state machine is inherited from sScanner.
Below is the evolution of some states.
@dotfile "PeriodicGaugeRecorder.state_graph"

*/


void  sPeriodicGaugeRecorder::Preconfigure (sNodePtr  DeviceCfg)
{
  sPtr<sFolderNode>              DeviceSettings;
  sPtr<sFolderNode>              DeviceLinks;
  DeviceCfg |= "Links";
  DeviceLinks = DeviceCfg >> "Links";
  if( DeviceLinks.IsCorrect() ){
    DeviceLinks |= "Gauge";
    DeviceLinks >> "Gauge" |= attr("Class","sGauge");
  }
  DeviceCfg |= "Settings";
  DeviceSettings = DeviceCfg >> "Settings";
  if( DeviceSettings.IsCorrect() ){
    DeviceSettings |= attr("ValidScanRange"
                          ,sPhysRange()=sPhysRange(0.0,real_inf,"s"));
    DeviceSettings |= attr("ValidScanSteps"
                          ,sPhysRange()=sPhysRange(0.01,1000,"s"));
    DeviceSettings |= attr("ScanRange"
                          ,sPhysRange()=sPhysRange(0.0,real_inf,"s",0.005));
    DeviceSettings |= attr("ScanStep"
                          ,sPhysValue()=sPhysValue(0.3,0.005,"s"));
  }
  sScanner::Preconfigure(DeviceCfg);
}


sPeriodicGaugeRecorder::~sPeriodicGaugeRecorder()
{
//  if( DeviceSettings.IsCorrect() ){
//    DeviceSettings << attr("ScanRange",ScanRangeField->Value())
//                   << attr("ScanStep",ScanStepField->Value());
//  }
  Gauge.SetToNULL();
  DataListRecord.SetToNULL();
}


sPeriodicGaugeRecorder::sPeriodicGaugeRecorder (sString    DeviceId
                                               ,sNodePtr   DeviceCfg
                                               ,sNodePtr   WidgetCfg
                                               ,QWidget *  parent   )
    :sScanner(DeviceId,DeviceCfg,WidgetCfg,parent)
    ,Gauge(this)
    ,ArbFrom(_Unitsless_)
    ,ArbStep(_Unitsless_)
    ,NumOfPointsEvaluation(_Unitsless_)
{
  Speed = sPhysValue(1);
  //evolution of the inherited state machine
  {
    //evolution of the State_Warming
    {
      State_W_Parallel
          = (new sParallelStateGroup(State_Warming()))
          ->SetAsInitial()
          ->AddTransition(State_Warming()->Fin());
      //State_W_Parallel internal structure and behaviour
      {
        State_W_Gauge = new sState(State_W_Parallel,"State_W_Gauge");
        //State_W_Gauge structure and behaviour
        {
          State_W_Gauge_Linking
              = (new sState(State_W_Gauge,"State_W_Gauge_Linking"))
              ->SetAsInitial()
              ->OnEnterCall( SLOT(LinkToGauge()) )
              ->AddTransition( SIGNAL(GaugeLinked()) ,State_W_Gauge->Fin() )
              ->AddTransition( SIGNAL(ErrorMessage())  ,State_Cooling()  );
        }
      }
    }
    //evolution of the State_O_Scanning
    {
      State_O_Scanning->OnExitCall(SLOT(ReleaseGaugeControl()));

      State_O_Sc_SettingGauge = new sState(State_O_Scanning
                                          ,"State_O_Sc_SettingGauge");
      State_O_Sc_ResettingGauge = new sState(State_O_Scanning
                                            ,"State_O_Sc_ResettingGauge");

      State_O_Sc_SettingGauge
          ->SetAsInitial()
          ->AddTransition( SIGNAL(Break()) ,State_O_Sc_ResettingGauge )
          ->AddTransition(State_O_Sc_Acquisition);

      State_O_Sc_Deciding
          ->RemoveOnFinishTransition()
          ->AddTransition(State_O_Sc_ResettingGauge);

      State_O_Sc_ResettingGauge
          ->AddTransition(State_O_Scanning->Fin());

      //State_O_Sc_SettingGauge internal structure and behavior
      {
        State_O_Sc_SG_Waiting = new sState(State_O_Sc_SettingGauge
                                            ,"State_O_Sc_SG_Waiting");
        State_O_Sc_SG_Waiting
            ->SetAsInitial()
            ->OnEnterCall(SLOT(SetGaugeRegime()))
            ->AddTransition(SIGNAL(GaugeBusy()) ,State_O_Idle)
            ->AddTransition<sPhysValue>(SIGNAL(GaugePeriodChanged(sPhysValue))
                                       ,SLOT(CorrectStep(sPhysValue))
                                       ,State_O_Sc_SettingGauge->Fin());
      }
      //State_O_Sc_ResettingGauge internal structure and behavior
      {
        State_O_Sc_RG_Waiting = new sState(State_O_Sc_ResettingGauge
                                            ,"State_O_Sc_RG_Waiting");

        State_O_Sc_RG_Waiting
            ->SetAsInitial()
            ->OnEnterCall(SLOT(ResetGaugeRegime()))
            ->AddTransition(SIGNAL(GaugePeriodChanged(sPhysValue))
                           ,SLOT(RestartGauge())
                           ,State_O_Sc_ResettingGauge->Fin());
      }
    }
    connect(this ,SIGNAL(GaugeError(QString))
           ,this ,SLOT(ReportGaugeError(QString)) );
  };
}


void  sPeriodicGaugeRecorder::AppearanceMakeUp ()
{
  sScanner::AppearanceMakeUp();
  try{ //get config parameters
    get_from<sPhysRangeNode>(DeviceSettings >> "ValidScanRange",ValidScanRange);
    get_from<sPhysRangeNode>(DeviceSettings >> "ValidScanSteps",ValidScanSteps);
    get_from<sPhysRangeNode>(DeviceSettings >> "ScanRange" ,ScanRange);
    get_from<sPhysValueNode>(DeviceSettings >> "ScanStep" ,ScanStep);
    get_from<sPhysValueNode>(DeviceSettings >> "Speed" ,Speed);
  }catch(...){
    SetInvalid(tr("Something wrong in configuration."));
  }
  //add new widgets
  {
    ScanRangeField = new sPhysRangeField(ParamArea
                                        ,sPhysRange(),false,false);
    ParamLayout->addWidget(ScanRangeField,Row++,0,1,3,Qt::AlignLeft);

    ScanStepField = new sPhysValueField(ParamArea,sPhysValue());
    ParamLayout->addWidget(ScanStepField,Row++,0,1,3,Qt::AlignLeft);
  }
  ScanRangeField->SetValidRange(ValidScanRange);
  ScanRangeField->SetValue(ScanRange);
  ScanStepField->SetValidRange(ValidScanSteps);
  ScanStep.FixTheUnits();
  ScanStepField->SetValue(ScanStep);
  Translate();
  connect(ScanStepField,SIGNAL(Edited())
         ,this         ,SLOT(EraseTrueStepToolTip()));
}


void  sPeriodicGaugeRecorder::Translate ()
{
  ScanRangeField->SetLabel(tr("Range"));
  ScanStepField->SetLabel(tr("Step"));
}


void  sPeriodicGaugeRecorder::RestoreState ()
{
  sScanner::RestoreState();
  if( WidgetConfig.IsCorrect() ){
    get_from<sPhysRangeNode>(WidgetConfig >> "ScanRange" ,ScanRange);
    get_from<sPhysValueNode>(WidgetConfig >> "ScanStep" ,ScanStep);
  }
  ScanRangeField->SetValidRange(ValidScanRange);
  ScanRangeField->SetValue(ScanRange);
  ScanStepField->SetValidRange(ValidScanSteps);
  ScanStepField->SetValue(ScanStep);
}


void  sPeriodicGaugeRecorder::StoreState ()
{
  sScanner::StoreState();
  if( WidgetConfig.IsCorrect() ){
    WidgetConfig << attr("ScanRange",ScanRangeField->Value())
                 << attr("ScanStep",ScanStepField->Value());
  }
}


void sPeriodicGaugeRecorder::ReportGaugeError (QString  error_text)
{
  emit ErrorMessage( tr("Got error from '%1':\n\n%2")
                       .arg(ToQString(Gauge->DeviceIdentifier()))
                       .arg(error_text)     );
}


void sPeriodicGaugeRecorder::LinkToGauge ()
{
  try{
    Gauge.SetTo(DeviceLinks,"Gauge");
    Connect(Gauge ,SIGNAL(Warmed())       ,SIGNAL(GaugeLinked()) );
    Connect(Gauge ,SIGNAL(ErrorMessage(QString)) ,SIGNAL(GaugeError(QString)) );
    Connect(Gauge ,SIGNAL(PeriodChanged(sPhysValue))
           ,SIGNAL(GaugePeriodChanged(sPhysValue)) );
  }catch( rxDeviceLaunch E ){
    emit ErrorMessage( tr("Error while linking to a submissive device.\n"
                          "%1\nCheck configuration.")
                         .arg(ToQString(E.MsgText()))   );
  }
}


void sPeriodicGaugeRecorder::OnEnterToCoolingState ()
{
  Disconnect(Gauge ,0 ,0);
  Gauge.ReleaseExclusiveControl();
  Gauge.SetToNULL();
  sScanner::OnEnterToCoolingState();
}


void  sPeriodicGaugeRecorder::CreateNewDependence ()
{
  if( Multiplier.X.IsNaN() ){
    Multiplier.X = sPhysValue(ScanStep.Units());
    sPhysRange                  RX(0*Multiplier.X,100*Multiplier.X);
    SetChartScales(sChartScales(sChartScale(ChartScales().X.Name,RX)
                               ,ChartScales().Y
                               ));
  };
  NumOfPointsEvaluation = ScanRange.Size() / ScanStep;
  ArbFrom = ScanRange.From() / Multiplier.X;
  ArbStep = ScanStep / Multiplier.X;

  Dependence = sNodePtr(new sDataListRecord);
  DataListRecord = Dependence;
  DataListRecord->SetMultiplier(sPhysPair(Multiplier.X,Multiplier.Y));
}


void  sPeriodicGaugeRecorder::StartScanning ()
{
  CurrentPointNo=0;
  Connect(Gauge ,SIGNAL(NewData(sPhysValue)) ,SLOT(AddNewData(sPhysValue)) );
  Gauge->SetAutoRepeat(true);
  SendCommand(Gauge,sGauge::cStart);
}


void  sPeriodicGaugeRecorder::StopScanning ()
{
  SendCommand(Gauge,sGauge::cStop);
  Disconnect(Gauge,SIGNAL(NewData(sPhysValue)) ,SLOT(AddNewData(sPhysValue)));
}


void  sPeriodicGaugeRecorder::SetGaugeRegime ()
{
  try{
    ScanStep = ScanStepField->Value();
    ScanRange = ScanRangeField->Value();
    if( ! Gauge.TakeExclusiveControl() ){
      QMessageBox::critical(this
          ,tr("'%1' cannot scan")
          .arg(ToQString(DeviceIdentifier()))
          ,tr("Gauge '%1' is busy now.\n\n  Try again later.")
          .arg(ToQString(Gauge->DeviceIdentifier()))
          );
      emit GaugeBusy();
    }else{
      //вычисляем желательное время на шаг:
      sPhysValue                  TimePerStep("s");

//!@test
//      ScanStep.FixTheUnits();

      TimePerStep = ScanStep / Speed;
      GaugeWasStarted = Gauge->Started();
      GaugeWasInAutoRepeatMode = Gauge->AutoRepeat();
      OldPeriod = Gauge->Period();
      SendCommand(Gauge ,sGauge::cStop);
      SendDirective(Gauge ,sGauge::cSetPeriod ,TimePerStep);
    }
  }catch(rxRAlgebra){
    emit Failure(tr("Wrong units in the configuration settings."));
  }
}


void sPeriodicGaugeRecorder::EraseTrueStepToolTip()
{
  ScanStepField->setToolTip(QString());
}


void  sPeriodicGaugeRecorder::CorrectStep (sPhysValue  gauge_period)
{
  try{
    //правим ScanStep и поле ScanStepField:
    ScanStep = Speed * gauge_period;
    ScanStepField->setToolTip(ToQString( ScanStep.Text() ));
    //правим поле ScanRangeField (его часть To):
    if( ScanRange.To() < ScanRange.From() + ScanStep ){
      ScanRange = sPhysRange( ScanRange.From() , ScanRange.From() + ScanStep );
      ScanRangeField->SetValue(ScanRange);
    }
  }catch(rxRAlgebra){
    emit Failure(tr("Wrong units in the configuration settings."));
  }
}


void  sPeriodicGaugeRecorder::ResetGaugeRegime ()
{
  if( GaugeWasInAutoRepeatMode != Gauge->AutoRepeat() ){
    Gauge->SetAutoRepeat(GaugeWasInAutoRepeatMode);
  }
  SendDirective(Gauge ,sGauge::cSetPeriod ,OldPeriod);
}


void  sPeriodicGaugeRecorder::RestartGauge ()
{
  if( GaugeWasStarted ){
    SendCommand(Gauge ,sGauge::cStart ,cReleaseControl);
  }
}


void  sPeriodicGaugeRecorder::ReleaseGaugeControl ()
{
  Gauge.ReleaseExclusiveControl();
}


void  sPeriodicGaugeRecorder::AddNewData (sPhysValue  Data)
{
  if( ! DataListRecord.IsCorrect() ){
    return;
  }
  if( Multiplier.X.IsNaN() ){
    return;
  }
  //!@todo{devices} Some gauges could change their units to incompatible
  //! either between the scans, or on the fly; something has to be done
  //! (@sa @ref sDNSpectroScanner::AddNewData).
  if( Multiplier.Y.IsNaN() ){
    Multiplier.Y = sPhysValue(Data.Units());
    sPhysRange                  RY(-Abs(Data),3*Abs(Data));
    SetChartScales(sChartScales(ChartScales().X
                               ,sChartScale(ChartScales().Y.Name,RY)
                               ));
  }
  if( DataListRecord->Multiplier().Y.IsNaN() ){
    DataListRecord->SetMultiplier(sPhysPair(Multiplier.X,Multiplier.Y));
  };
  sPhysValue                  X(_Unitsless_);
  sPhysValue                  Y(_Unitsless_);
  X = ArbFrom + ArbStep/2 + ArbStep * CurrentPointNo;
  Y = Data / Multiplier.Y;
  DataListRecord->AddNewPoint(X.ToReal(),Y.ToReal());
  if( CurrentPointNo < Round(NumOfPointsEvaluation.ToReal()+0.5)-1 ){
    CurrentPointNo++;
  }else{
    emit ScanningHasCompleted();
  }
}

//------------------------------------------------------------------------------
} //namespace RA
