////////////////////////////////////////////////////////////////////////////////
/*! @file ProtoAccumulator.cpp   Прообраз накопителя данных,
                                 поступающих в виде массива.
- Part of BasicDevices - a Research Assistant Plugin package.
- Uses  RAGUI - the Research Assistant Graphical User Interface.
- Uses  QtGui v.4.6  -  http://qt.nokia.com/
- Uses  RANet - Research Assistant Net Library (based on ANSI C++).
- Copyright(C) 2010, Dmitry K. Nelson, St.Petersburg, Russia.
                     D.Nelson@mail.ioffe.ru
*///////////////////////////////////////////////////////////////////////////////
#include "ProtoAccumulator.h"
#include "attr.h"
#include "StringsAndQStrings.h"
#include "Helper.h"
#include "Trace.h"
#include "Log.h"
#include <QMessageBox>
#include <QDebug>
namespace RA {
//------------------------------------------------------------------------------

DEFINE_CLASS_TAG(sProtoAccumulator)

//-------------------------------------------------------------- sProtoAccumulator ---

void  sProtoAccumulator::Preconfigure (sNodePtr  DeviceCfg)
{
  sPtr<sFolderNode>              DeviceSettings;
  DeviceCfg |= "Settings";
  DeviceSettings = DeviceCfg >> "Settings";
  if( DeviceSettings.IsCorrect() ){
    DeviceSettings |= attr("ValidScanRange",sPhysRange(0.0,real_inf,"s"));
    DeviceSettings |= attr("ValidScanSteps",sPhysRange(1.0e-8,100,"s"));
    DeviceSettings |= attr("ScanRange",sPhysRange(0.0,real_inf,"s",0.005));
    DeviceSettings |= attr("ScanStep",sPhysValue(5.0e-6,0.005,"s"));
  };
  sAScanner::Preconfigure(DeviceCfg);
};


sProtoAccumulator::~sProtoAccumulator()
{
  if( DeviceSettings.IsCorrect() ){
    DeviceSettings << attr("ScanRange",ScanRangeField->Value())
                   << attr("ScanStep",ScanStepField->Value());
  };
};

sProtoAccumulator::sProtoAccumulator(sString   DeviceId
                        ,sNodePtr  DeviceCfg
                        ,sNodePtr  WidgetCfg)
    :sProtoScanner(DeviceId,DeviceCfg,WidgetCfg)
    ,NumOfPoints(0)
    ,ArbFrom(_Unitsless_)
    ,ArbStep(_Unitsless_)
    ,CurrentSamplingNo(0)
{
  if( DeviceSettings.IsCorrect()){
    sPtr<sPhysValueNode>        PV;
    sPtr<sPhysRangeNode>        PR;

    PR=DeviceSettings >> "ValidScanRange";
    if( PR.IsCorrect() )  ValidScanRange = (sPhysRange)(*PR);

    PR=DeviceSettings >> "ValidScanSteps";
    if( PR.IsCorrect() )  ValidScanSteps = (sPhysRange)(*PR);

    PR=DeviceSettings >> "ScanRange";
    if( PR.IsCorrect() )  ScanRange = (sPhysRange)(*PR);

    PV=DeviceSettings >> "ScanStep";
    if( PV.IsCorrect() )  ScanStep = (sPhysValue)(*PV);
  };

  ScanRangeField = new sPhysRangeField(ParamArea,FromQString(tr("Range"))
                                      ,sPhysRange(),false,false);
  ParamLayout->addWidget(ScanRangeField,Row++,0,1,3,Qt::AlignLeft);

  ScanStepField = new sPhysValueField(ParamArea,FromQString(tr("Step"))
                                     ,sPhysValue(),false,false,false);
  ParamLayout->addWidget(ScanStepField,Row++,0,1,3,Qt::AlignLeft);
};


void  sProtoAccumulator::RestoreState ()
{
  sAScanner::RestoreState();
  if( ! WidgetConfig.IsNULL() ){
    sPtr<sPhysValueNode>        PV;
    sPtr<sPhysRangeNode>        PR;

    PR=WidgetConfig >> "ScanRange";
    if( PR.IsCorrect() )  ScanRange = (sPhysRange)(*PR);

    PV=WidgetConfig >> "ScanStep";
    if( PV.IsCorrect() )  ScanStep = (sPhysValue)(*PV);

  };
  ScanRangeField->SetValidRange(ValidScanRange);
  ScanRangeField->SetValue(ScanRange);
  ScanStepField->SetValidRange(ValidScanSteps);
  ScanStepField->SetValue(ScanStep);
};


void  sProtoAccumulator::StoreState ()
{
  sAScanner::StoreState();
  if( ! WidgetConfig.IsNULL() ){
    WidgetConfig << attr("ScanRange",ScanRangeField->Value())
                 << attr("ScanStep",ScanStepField->Value());
//    WidgetConfig << attr("Speed",SpeedField->Value());
  };
  if( DeviceSettings.IsCorrect() ){
    DeviceSettings << attr("ScanRange",ScanRangeField->Value())
                   << attr("ScanStep",ScanStepField->Value());
  };
};


void  sProtoAccumulator::StartScan ()
{
  if( DriverState() != Standby )
    return;
  ScanRange = ScanRangeField->Value();
  ScanStep = ScanStepField->Value();

  if( Multiplier.X.IsNaN() ){
    Multiplier.X = sPhysValue(ScanStep.Units());
    sPhysRange                  RX(0*Multiplier.X,100*Multiplier.X);
    SetChartScales(sChartScales(sChartScale(ChartScales().X.Name,RX),ChartScales().Y));
  };
  sPhysValue NOP = ScanRange.Size() / ScanStep;
  NumOfPoints = Round (NOP.ArbValue());
  ArbFrom = ScanRange.From() / Multiplier.X;
  ArbStep = ScanStep / Multiplier.X;

  Dependence = new sArrayTypeDependence(NumOfPoints,ArbFrom.ArbValue(),ArbStep.ArbValue());
  ArrayTypeDependence = Dependence;
  Dependence->SetMultiplier(sPhysPair(Multiplier.X,Multiplier.Y));
  CurrentSamplingNo=0;

  sAScanner::StartScan();
  Dependence << attr("range", ScanRange)
             << attr("step",ScanStep);
};


void  sProtoAccumulator::FinishScan ()
{
  qDebug() << "Registered " << CurrentSamplingNo << " times";
  sAScanner::FinishScan();
};


void  sProtoAccumulator::AddNewData (real* Value)
{
    if( Multiplier.X.IsNaN() ){
      return;
    };
    if( Multiplier.Y.IsNaN() ){
      Multiplier.Y = sPhysValue(_Unitsless_);
    //  sPhysRange                  RY(-Abs(Data),3*Abs(Data));
    //  SetChartScales(sChartScales(ChartScales().X,sChartScale(ChartScales().Y.Name,RY)));
    };
    if( Dependence->Multiplier().Y.IsNaN() ){
      Dependence->SetMultiplier(sPhysPair(Multiplier.X,Multiplier.Y));
    }
    CurrentSamplingNo++;
    for (int PointNo = 0; PointNo < NumOfPoints; PointNo++ ){
        if (CurrentSamplingNo == 1){
            ArrayTypeDependence->SetValue(PointNo, Value[PointNo]);
        }else{
            ArrayTypeDependence->SetValue(PointNo, (ArrayTypeDependence->PointY(PointNo)
               *(CurrentSamplingNo-1)+ Value[PointNo])/CurrentSamplingNo);
        }
    };
};

//------------------------------------------------------------------------------
}; //namespace RA
