////////////////////////////////////////////////////////////////////////////////
/*! @file Scanner.cpp   Довольно-таки универсальный прототип
                             сканирующего прибора.
- Part of BasicDevices - Research Assistant Plugin package.
- Uses  RAGUI - Research Assistant Graphical User Interface Library.
- Uses  RANet - Research Assistant Net Library.
- Uses  Qt v.5  -  http://qt-project.org/
- Copyright(C) 2010-2015, Viktor E. Bursian, St.Petersburg, Russia.
                          Viktor_dot_Bursian_at_mail_dot_ioffe_dot_ru
*///////////////////////////////////////////////////////////////////////////////
#include "Scanner.h"
#include "attr.h"
#include "StringsAndQStrings.h"
#include "Helper.h"
#include "Trace.h"
#include "Log.h"
#include <QMessageBox>
namespace RA {
//------------------------------------------------------------------------------

DEFINE_CLASS_TAG(sScanner)

//---------------------------------------------------------------- sScanner ---
/*! @class sScanner

@par The state machine

The upper-level state machine is inherited from sConsole.  
Below is the evolution of the state Operating.
@dotfile "Scanner.state_graph"

Legend: 
- thick solid black arrow - unconditional transition (when the state finishes);
- thin solid black arrow - transition on signal or command or directive;
- dotted black circular arrow - reaction on signal or command or directive
  (place the mouse over the line to bring up a hint);
- dashed black arrow - an alternative transition
  (place the mouse over the line to bring up a hint);
- dashed violet arrow - an additional transition, which appears later, when
  the state machine is working
  (place the mouse over the line to bring up a hint).

@todo{Instruments} Hierarchic layouts instead of dummy widgets?
                   See sGaugeTestboard

@todo{Instruments} Use virtual method(s) here? It is not important
                   where the controls are!
*/

void  sScanner::Preconfigure (sNodePtr  DeviceCfg)
{
  sPtr<sFolderNode>              DeviceSettings;
  DeviceCfg |= "Settings";
  DeviceSettings = DeviceCfg >> "Settings";
  if( DeviceSettings.IsCorrect() ){
    DeviceSettings |= attr("HistorySize",0);
    DeviceSettings |= attr("LineThickness",1);
    DeviceSettings |= attr("LineColorSet",RAGUI::LineColorSet);
  }
  sConsole::Preconfigure(DeviceCfg);
}


sScanner::~sScanner()
{
  if( DeviceSettings.IsCorrect() ){
    Chart->StoreScales(DeviceSettings);
  }
  if( Chart ){ delete Chart;  Chart=NULL; };
  if( Evolution ){ delete Evolution;  Evolution=NULL; };
}


sScanner::sScanner(sString   DeviceId
                  ,sNodePtr  DeviceCfg
                  ,sNodePtr  WidgetCfg
                  ,QWidget *  parent    )
    :sConsole(DeviceId,DeviceCfg,WidgetCfg,parent)
    ,HistorySize(0)
    ,LineThickness(1)
    ,Evolution(NULL)
    ,Chart(NULL)
{
  Multiplier = sPhysPair(sPhysValue(real_nan),sPhysValue(real_nan));

  try{ //get config parameters
    get_from<sIntegerNode> (DeviceSettings >> "HistorySize" , HistorySize);
    get_from<sIntegerNode> (DeviceSettings >> "LineThickness" , LineThickness);
    LineColorSet = DeviceSettings >> "LineColorSet";
    ASSERT(LineColorSet.IsCorrect());
  }catch(...){
    SetInvalid(tr("Something wrong in configuration."));
    //!@bug RA crashes after showing this message
  }
  //evolution of the inherited state machine
  {
    //evolution of the State_Operating
    {
      State_O_Idle = new sState(State_Operating(),"State_O_Idle");
      State_O_Scanning = new sState(State_Operating(),"State_O_Scanning");

      State_O_Idle
          ->SetAsInitial()
          ->AddTransition( SIGNAL(Scan()) ,State_O_Scanning );
          
      State_O_Scanning          
            ->AddReaction( SIGNAL(CoolDown()) ,SLOT(OnCoolDownWhileScanning()) )
            ->AddTransition( SIGNAL(Break()) ,State_O_Idle )
            ->AddTransition(State_O_Idle);
      
      //State_O_Scanning internal structure and behavior
      {
        State_O_Sc_Acquisition = new sState(State_O_Scanning
                                           ,"State_O_Sc_Acquisition");
        State_O_Sc_Deciding = new sState(State_O_Scanning
                                        ,"State_O_Sc_Deciding");
      
        State_O_Sc_Acquisition
            ->SetAsInitial()
            ->OnEnterCall( SLOT(InitializeScanning()) )
            ->AddTransition(SIGNAL(Break()) ,State_O_Sc_Deciding )
            ->AddTransition(SIGNAL(ScanningHasCompleted()) ,State_O_Sc_Deciding)
            ->OnExitCall( SLOT(FinalizeScanning()) );
                         
        State_O_Sc_Deciding
            ->AddTransition(State_O_Scanning->Fin());

        //State_O_Sc_Deciding internal structure and behavior
        {
          State_O_Sc_D_Waiting = new sState(State_O_Sc_Deciding
                                           ,"State_O_Sc_D_Waiting");
          State_O_Sc_D_Waiting
              ->SetAsInitial()
              ->AddTransition(SIGNAL(Accept())
                             ,SLOT(OnAccept())
                             ,State_O_Sc_Deciding->Fin())
              ->AddTransition(SIGNAL(Disregard())
                             ,SLOT(OnDisregard())
                             ,State_O_Sc_Deciding->Fin());
        }
      }
    }
  }
  //!@bug RA often crashes if Accept button pressed immediately after Break
}


void  sScanner::OnCoolDownWhileScanning ()
{
  RANet::Log.Put(sLog::Debug,"States","sScanner::OnCoolDownWhileScanning");
  State_O_Idle
      ->AddTransition( State_Cooling() );
  emit Break();
}


void  sScanner::AppearanceMakeUp ()
{
  sConsole::AppearanceMakeUp();
  //add new widgets
  {
    Layout = new QHBoxLayout();
    setLayout(Layout);
    Layout->setObjectName("Layout");
    Layout->setMargin(0);

    Splitter = new QSplitter(this);
    Splitter->setObjectName("Splitter");
    Layout->addWidget(Splitter);

    SideFrame = new QFrame(this);
    SideFrame->setObjectName("SideFrame");

    SideLayout = new QVBoxLayout(/*SideFrame*/);
    SideLayout->setObjectName("SideLayout");
    SideFrame->setLayout(SideLayout);
    SideLayout->setSpacing(0);
    SideLayout->setMargin(0);

    ButtonArea = new QWidget(SideFrame);
    ButtonArea->setObjectName("ButtonArea");

    ButtonLayout = new QGridLayout(/*ButtonArea*/);
    ButtonLayout->setObjectName("ButtonLayout");
    ButtonArea->setLayout(ButtonLayout);
    ButtonLayout->setSpacing(0);
    ButtonLayout->setMargin(0);

    ButtonsMakeUp(ButtonArea,ButtonLayout);

    SideLayout->addWidget(ButtonArea);

    ParamScroll = new QScrollArea(SideFrame);
    ParamScroll->setObjectName("ParamScroll");
    SideLayout->addWidget(ParamScroll);

    ParamArea = new QFrame(ParamScroll);
    ParamArea->setObjectName("ParamFrame");

    ParamLayout = new QGridLayout(/*ParamArea*/);
    ParamLayout->setObjectName("ParamLayout");
    ParamArea->setLayout(ParamLayout);
    ParamLayout->setSpacing(0);
    ParamLayout->setMargin(0);

    Row=0;

    LblSample = new QLabel(ParamArea);
    LblSample->setObjectName("LblSample");
    ParamLayout->addWidget(LblSample,95,0,1,1);
    FldSample = new sDropNodeWidget(ParamArea);
    FldSample->setObjectName("FldSample");
    ParamLayout->addWidget(FldSample,95,1,1,8);

    LblTopic = new QLabel(ParamArea);
    LblTopic->setObjectName("LblTopic");
    ParamLayout->addWidget(LblTopic,96,0,1,9);
    FldTopic = new sDropNodeWidget(ParamArea);
    FldTopic->setObjectName("FldTopic");
    ParamLayout->addWidget(FldTopic,97,0,1,9);

    LblConditions = new QLabel(ParamArea);
    LblConditions->setObjectName("LblConditions");
    ParamLayout->addWidget(LblConditions,98,0,1,9);
    FldConditions = new QTextEdit(ParamArea);
    FldConditions->setObjectName("FldConditions");
    ParamLayout->addWidget(FldConditions,99,0,1,9);

    ParamScroll->setWidget(ParamArea);
    ParamScroll->setWidgetResizable(true);

    WidgetConfig |= "Content";
    Content = WidgetConfig >> "Content";
    ASSERT(Content.IsCorrect());
    Evolution = new sNet2TreeEvolution(sAttribute(sAttrName(),Content));
    Chart = new sChart(Evolution ,this);

    connect(Chart ,SIGNAL(CursorPositionChanged(sPhysPair,sPhysPair
                                               ,sString,sString))
           ,MainWindow() ,SLOT(ShowCursorInfo(sPhysPair,sPhysPair
                                             ,sString,sString))
           ,Qt::QueuedConnection);

    connect(Evolution,SIGNAL(Changed()),Chart,SLOT(OnNetChange()));
    Evolution->Initiate();

    connect(FldSample ,SIGNAL(Changed()) ,this ,SLOT(OnSampleChange()));
    connect(FldTopic ,SIGNAL(Changed()) ,this ,SLOT(OnTopicChange()));
  }
  //add new widgets
  {//! @todo{Instruments} review controls creation, connection, layout
//    MainSplitter = new QSplitter(this);

  }
  //lay the widgets out
  {

    Splitter->addWidget(SideFrame);
    Splitter->addWidget(Chart);
  }

  if( DeviceSettings.IsCorrect() ){
    Chart->RestoreScales(DeviceSettings);
  }
  Translate();
}


void  sScanner::Translate ()
{
  LblState->setText(TextState);
  LblSample->setText(tr("Sample"));
  FldSample->Prompt=FromQString(tr(
                      "In the Samples workbook, "
                      "find (or create) an object, representing the sample.\n "
                      "Then drag and drop it here"));
  LblTopic->setText(tr("Research topic"));
  FldTopic->Prompt=FromQString(tr(
                      "Find (or create) an object in the Topics workbook, "
                      "with the docket indicating the aim and sense of study.\n"
                      "Then drag and drop it here"));
  LblConditions->setText(tr("Conditions"));
  BtnTurnOn->setText(tr("Turn ON"));
  BtnScan->setText(tr("Scan"));
  BtnBreak->setText(tr("Break"));
  BtnAccept->setText(tr("Accept"));
  BtnDisregard->setText(tr("Disregard"));
}


void  sScanner::ButtonsMakeUp (QWidget * parent ,QLayout * button_layout)
{
  QGridLayout *               TargetLayout = dynamic_cast<QGridLayout *>
                                                        (button_layout);
  if( TargetLayout ){
    TextState = TextFrozen;
    LblState = new QLabel(TextState,parent);
    LblState->setObjectName("LblState");
    TargetLayout->addWidget(LblState,0,0,1,1);
    
    connect(this,SIGNAL(NewUpperLevelState(sDeviceStateMachine::eState))
           ,this,SLOT(OnConsoleStateChanged(sDeviceStateMachine::eState)));

    BtnTurnOn = new QPushButton(parent);
    BtnTurnOn->setObjectName("BtnTurnOn");
    TargetLayout->addWidget(BtnTurnOn,0,1,1,1);
    BtnTurnOn->setVisible(false);
    connect(BtnTurnOn ,SIGNAL(clicked()) ,this ,SIGNAL(WarmUp()));
    State_Frozen()->OnEnterCall( BtnTurnOn ,SLOT(show()) )
                  ->OnExitCall( BtnTurnOn ,SLOT(hide()) );

    BtnScan = new QPushButton(parent);
    BtnScan->setObjectName("BtnScan");
    TargetLayout->addWidget(BtnScan,1,0,1,1);
    BtnScan->setVisible(false);
    connect(BtnScan ,SIGNAL(clicked()) ,this ,SIGNAL(Scan()));
    State_O_Idle->OnEnterCall( BtnScan ,SLOT(show()) )
                   ->OnExitCall( BtnScan ,SLOT(hide()) );

    BtnBreak = new QPushButton(parent);
    BtnBreak->setObjectName("BtnBreak");
    TargetLayout->addWidget(BtnBreak,1,1,1,1);
    BtnBreak->setVisible(false);
    connect(BtnBreak ,SIGNAL(clicked()) ,this ,SIGNAL(Break()));
    State_O_Scanning->OnEnterCall( BtnBreak ,SLOT(show()) )
                    ->OnExitCall( BtnBreak ,SLOT(hide()) );
    State_O_Sc_Deciding->OnEnterCall( BtnBreak ,SLOT(hide()) )
                       ->OnExitCall( BtnBreak ,SLOT(show()) );

    BtnAccept = new QPushButton(parent);
    BtnAccept->setObjectName("BtnAccept");
    TargetLayout->addWidget(BtnAccept,1,2,1,1);
    BtnAccept->setVisible(false);
    connect(BtnAccept ,SIGNAL(clicked()) ,this ,SIGNAL(Accept()));
    State_O_Sc_D_Waiting->OnEnterCall( BtnAccept ,SLOT(show()) )
                        ->OnExitCall( BtnAccept ,SLOT(hide()) );

    BtnDisregard = new QPushButton(parent);
    BtnDisregard->setObjectName("BtnDisregard");
    TargetLayout->addWidget(BtnDisregard,1,3,1,1);
    BtnDisregard->setVisible(false);
    connect(BtnDisregard ,SIGNAL(clicked()) ,this ,SIGNAL(Disregard()));
    State_O_Sc_D_Waiting->OnEnterCall( BtnDisregard ,SLOT(show()) )
                        ->OnExitCall( BtnDisregard ,SLOT(hide()) );
  }
}


void  sScanner::AfterStartInitiation ()
{
  sConsole::AfterStartInitiation();
  RefreshTimer = new QTimer(this);
  connect(RefreshTimer, SIGNAL(timeout()), Evolution, SLOT(Update()));
  RefreshTimer->start(100);
}


void  sScanner::SetChartScales (sChartScales  scales)
{
  Chart->SetChartScales(scales);
}


void  sScanner::RestoreState ()
{
  sConsole::RestoreState();
  if( WidgetConfig.IsCorrect() ){
    int                       Orientation;
    if( get_from<sIntegerNode>(WidgetConfig>>"Splitter.",Orientation) )
      Splitter->setOrientation(Qt::Orientation(Orientation));
    int                       Part1;
    int                       Part2;
    if( get_from<sIntegerNode>(WidgetConfig>>"Splitter.Part1",Part1) 
        && get_from<sIntegerNode>(WidgetConfig>>"Splitter.Part2",Part2) ){
      QList<int>                List;
      List.append(Part1);
      List.append(Part2);
      Splitter->setSizes(List);
    }
    
    Chart->RestoreScales(WidgetConfig);

    get_from<sTextNode>(WidgetConfig>>"Conditions",Conditions);
    Sample =  WidgetConfig >> "Sample";
    Topic =  WidgetConfig >> "Topic";
  }
  FldConditions->setPlainText(ToQString(Conditions));
  FldSample->SetValue(Sample);
  disconnect(FldTopic,SIGNAL(Changed()),0,0);
  FldTopic->SetValue(Topic);
  connect(FldTopic,SIGNAL(Changed()),this,SLOT(ResetColorSequence()));
  Content.DeleteAllAttributes();
}


void  sScanner::StoreState ()
{
  sConsole::StoreState();
  if( WidgetConfig.IsCorrect() ){
    WidgetConfig << attr("Splitter.Orientation" ,(int)Splitter->orientation());
    WidgetConfig << attr("Splitter.Part1" ,(int)Splitter->sizes().at(0));
    WidgetConfig << attr("Splitter.Part2" ,(int)Splitter->sizes().at(1));

    Chart->StoreScales(WidgetConfig);

    Sample=FldSample->Value;
    Topic=FldTopic->Value;
    Conditions=FromQString(FldConditions->toPlainText());
    WidgetConfig << attr("Conditions",Conditions);
    WidgetConfig.DelAttr("Sample");
    if( ! Sample.IsNULL() ){
      WidgetConfig << attr("Sample",Sample);
    }
    WidgetConfig.DelAttr("Topic");
    if( ! Topic.IsNULL() ){
      WidgetConfig << attr("Topic",Topic);
    }
  }
}


bool  sScanner::CanClose ()
{
  if( States->configuration().contains(State_O_Sc_Deciding) ){
      QMessageBox::information(this ,tr("Cannot close")
          ,tr("The device %1 is waiting for your decision.")
          .arg(ToQString(DeviceIdentifier()))
          );
    return false;
  }else if( States->configuration().contains(State_O_Scanning) ){
      QMessageBox::information(this ,tr("Cannot close")
          ,tr("The device %1 is scanning.  Please, wait or break it.")
          .arg(ToQString(DeviceIdentifier()))
          );
    return false;
  }
  return true;
}


void sScanner::OnConsoleStateChanged (sDeviceStateMachine::eState  new_state)
{
//  sConsole::OnConsoleStateChanged(new_state);

  switch( new_state ){
      case sDeviceStateMachine::cWarming:
          LblState->setVisible(true);
          TextState = TextWarming;
          LblState->setText(TextState);
          break;
      case sDeviceStateMachine::cOperating:
          LblState->setVisible(false);
          TextState = TextOperating;
          LblState->setText(TextState);
          break;
      case sDeviceStateMachine::cCooling:
          LblState->setVisible(true);
          TextState = TextCooling;
          LblState->setText(TextState);
          break;
      case sDeviceStateMachine::cFrozen:
          LblState->setVisible(true);
          TextState = TextFrozen;
          LblState->setText(TextState);
          break;
  }
}


void  sScanner::InitializeScanning ()
{
  CreateNewDependence();
  if( ! Dependence.IsCorrect() )
    return;
  Dependence->AllocateRecord();
  EarlySetAttributes();
  Content << Dependence;
  PutSpectrumIntoSecondaryDestinations();
  Evolution->Update();

  StartScanning();
}


void  sScanner::EarlySetAttributes ()
{
  {
    psGraphObject  GrOb = dynamic_cast<psGraphObject>(Dependence.operator->());
    if( GrOb && ! LineColorSet->Empty() )
      GrOb->SetAppearance(new sCurveAppearance(LineColorSet->CurrentColor()
                                              ,LineThickness));
  }
  Sample=FldSample->Value;
  Topic=FldTopic->Value;
  Conditions=FromQString(FldConditions->toPlainText());
  //! @todo{sString} Trim
  //if( ! Conditions.Trim().Empty() )
  if( ! Conditions.Empty() )
    Dependence << attr(Conditions);
  if( ! Sample.IsNULL() )
    Dependence << attr("sample",Sample);
  Dependence << attr("technique",DeviceIdentifier()+" scan");
}


void  sScanner::LateSetAttributes ()
{
  if( Dependence.IsCorrect() )
    Dependence->SetFinishTime();
}


void  sScanner::FinalizeScanning ()
{
  StopScanning();
  LateSetAttributes();
  Evolution->Update();
}


void  sScanner::OnAccept ()
{
  UpdateHistory();
  LineColorSet->ShiftToNext();
  Evolution->Update();
}


void  sScanner::OnDisregard ()
{
  RemoveSpectrumFromSecondaryDestinations();
  if( Dependence.IsCorrect() )
    Dependence->DisregardRecord("immediately after recording");
  UpdateHistory();
  Evolution->Update();
}


void  sScanner::ResetColorSequence ()
{
  LineColorSet->SetToFirst();
}


void  sScanner::OnSampleChange ()
{
  if( FldSample->Value.IsNULL() )
    return;
  sPtr<sFolderNode>           S (FldSample->Value);
  if( S.IsCorrect() ){
    if( ! S->Docket().Empty() ){
      return;
    }
  }
  if( QMessageBox::Help ==
      QMessageBox::critical(this
          ,tr("Unacceptable object for the Sample field")
          ,tr("The object you dropped into the Sample field "
              "does not seem to be a valid description of a "
              "sample.\n\n"
              "It should be an object and its docket "
              "should unambigously identify this particular "
              "sample.")
          ,QMessageBox::Ok | QMessageBox::Help) ){
    sHelper::ShowKeyId("sScanner_OnSampleChange");
  }
  FldSample->SetValue(sNodePtr());
}


void  sScanner::OnTopicChange ()
{
  if( FldTopic->Value.IsNULL() )
    return;
  sPtr<sFolderNode>           T (FldTopic->Value);
  if( T.IsCorrect() ){
    if( ! T->Docket().Empty() ){
      return;
    }
  }
  if( QMessageBox::Help ==
      QMessageBox::critical(this
          ,tr("Unacceptable object for the Topic field")
          ,tr("The object you dropped into the Topic field "
              "does not look like a topic of an insvestigation.\n\n"
              "It should be an object and its docket "
              "should contain a more or less informative "
              "description of what is the sense of this "
              "particular study (i.e. 'Angular dependence' "
              "or 'Overview of new samples').")
          ,QMessageBox::Ok | QMessageBox::Help) ){
    sHelper::ShowKeyId("sScanner_OnTopicChange");
  }
  FldTopic->SetValue(sNodePtr());
}


void  sScanner::UpdateHistory ()
{
  Content.DelAttr(Dependence);
  Dependence.SetToNULL();
}


void  sScanner::PutSpectrumIntoSecondaryDestinations ()
{
  if( ! Sample.IsNULL() ){
    sPtr<sFolderNode>     Allocation(Sample);
    Allocation |= "records";
    Allocation = Allocation >> "records";
    Allocation |= Dependence->StartTime().Year();
    Allocation = Allocation >> Dependence->StartTime().Year();
    Allocation |= Dependence->StartTime().Month();
    Allocation = Allocation >> Dependence->StartTime().Month();
    Allocation |= Dependence->StartTime().Day();
    Allocation = Allocation >> Dependence->StartTime().Day();
    if( Allocation.IsCorrect() )
      Allocation << Dependence;
  }
  if( ! Topic.IsNULL() ){
    Topic << Dependence;
  }
}


void  sScanner::RemoveSpectrumFromSecondaryDestinations ()
{
  if( ! Sample.IsNULL() ){
    sPtr<sFolderNode>     Allocation(Sample);
    Allocation = Allocation >> "records"
                            >> Dependence->StartTime().Year()
                            >> Dependence->StartTime().Month()
                            >> Dependence->StartTime().Day();
    if( Allocation.IsCorrect() )
      Allocation.DelAttr(Dependence);
  }
  if( ! Topic.IsNULL() ){
    Topic.DelAttr(Dependence);
  }
}

//------------------------------------------------------------------------------
} //namespace RA
