////////////////////////////////////////////////////////////////////////////////
/*! @file StepByStepScanner.cpp  Прибор пошагового сканирования.
- Part of BasicDevices - Research Assistant Plugin package.
- Uses  RAGUI - Research Assistant Graphical User Interface Library.
- Uses  Qt v.5 - http://qt-project.org/
- Uses  RANet - Research Assistant Net Library.
- Copyright(C) 2014-2015, Dmitry K. Nelson, St.Petersburg, Russia.
                          D.Nelson@mail.ioffe.ru
*///////////////////////////////////////////////////////////////////////////////
#include "StepByStepScanner.h"
#include "attr.h"
#include "StringsAndQStrings.h"
#include <QMessageBox>
namespace RA {
//------------------------------------------------------------------------------

DEFINE_CLASS_TAG(sStepByStepScanner)

//--------------------------------------------------- sStepByStepScanner ---

void  sStepByStepScanner::Preconfigure (sNodePtr  DeviceCfg)
{
  sPtr<sFolderNode>              DeviceSettings;
  sPtr<sFolderNode>              DeviceLinks;
  DeviceCfg |= "Links";
  DeviceLinks = DeviceCfg >> "Links";
  if( DeviceLinks.IsCorrect() ){
    DeviceLinks |= "Gauge";
    DeviceLinks >> "Gauge" |= attr("Class","sGauge");
    DeviceLinks |= "Effector";
    DeviceLinks >> "Effector" |= attr("Class","sEffector");
  }
  DeviceCfg |= "Settings";
  DeviceSettings = DeviceCfg >> "Settings";
  if( DeviceSettings.IsCorrect() ){
    DeviceSettings |= attr("ValidScanRange",sPhysRange()
                                           =sPhysRange(200,1000,"nm"));
    DeviceSettings |= attr("ValidScanSteps",sPhysRange()
                                           =sPhysRange(0.01,100,"nm"));
    DeviceSettings |= attr("ScanRange",sPhysRange()
                                      =sPhysRange(300,700,"nm",0.05));
    DeviceSettings |= attr("ScanStep",sPhysValue()
                                     =sPhysValue(0.5,0.05,"nm"));
    DeviceSettings |= attr("MeasureTime",sPhysValue(1.0,"s"));
  }
  sScanner::Preconfigure(DeviceCfg);
}


sStepByStepScanner::~sStepByStepScanner()
{
  Gauge.SetToNULL();
  Effector.SetToNULL();
  DataArrayRecord.SetToNULL();
}


sStepByStepScanner::sStepByStepScanner (sString    DeviceId
                                       ,sNodePtr   DeviceCfg
                                       ,sNodePtr   WidgetCfg
                                       ,QWidget *  parent   )
    :sScanner(DeviceId,DeviceCfg,WidgetCfg,parent)
    ,Gauge(this)
    ,Effector(this)
    ,MeasureTime()
    ,ValidScanRange()
    ,ValidScanSteps()
    ,ScanRange()
    ,ScanStep()
    ,NumOfPoints(0)
    ,CurrentPointNo(0)
{
  //evolution of the inherited state machine
  {
    //evolution of the State_Warming
    {
      State_W_Parallel 
          = (new sParallelStateGroup(State_Warming()))
          ->SetAsInitial()
          ->AddTransition(State_Warming()->Fin());
      //State_W_Parallel internal structure and behaviour
      {
        State_W_Gauge = new sState(State_W_Parallel,"State_W_Gauge");
        //State_W_Gauge structure and behaviour
        {
          State_W_Gauge_Linking 
              = (new sState(State_W_Gauge,"State_W_Gauge_Linking"))
              ->SetAsInitial()
              ->OnEnterCall( SLOT(LinkToGauge()) )
              ->AddTransition( SIGNAL(GaugeLinked()) ,State_W_Gauge->Fin() )
              ->AddTransition( SIGNAL(ErrorMessage())  ,State_Cooling()  );
        }
        State_W_Effector = new sState(State_W_Parallel,"State_W_Effector");
        //State_W_Effector structure and behaviour
        {
          State_W_Effector_Linking
              = (new sState(State_W_Effector,"State_W_Effector_Linking"))
              ->SetAsInitial()
              ->OnEnterCall( SLOT(LinkToEffector()) )
              ->AddTransition( SIGNAL(EffectorLinked()) ,State_W_Effector->Fin() )
              ->AddTransition( SIGNAL(ErrorMessage())  ,State_Cooling()  );
        }
      }
    }
    //evolution of the State_O_Scanning
    {
      State_O_Scanning->OnExitCall(SLOT(ReleaseControl()));
        
      State_O_Sc_PrepareDevices = new sState(State_O_Scanning
                                          ,"State_O_Sc_PrepareDevices");
      State_O_Sc_SettingGauge = new sState(State_O_Scanning
                                          ,"State_O_Sc_SettingGauge");
      State_O_Sc_ResettingGauge = new sState(State_O_Scanning
                                            ,"State_O_Sc_ResettingGauge");
      
      State_O_Sc_PrepareDevices
          ->SetAsInitial()
          ->OnEnterCall(SLOT(TakeDevicesControl()))
          ->AddTransition(SIGNAL(DevicesBusy()),State_O_Idle)
          ->AddTransition(SIGNAL(DevicesOK()),State_O_Sc_SettingGauge);

      State_O_Sc_SettingGauge
          ->AddTransition( SIGNAL(Break()) ,State_O_Sc_ResettingGauge )
          ->AddTransition(State_O_Sc_Acquisition);
      
      State_O_Sc_Deciding
          ->RemoveOnFinishTransition()
          ->AddTransition(State_O_Sc_ResettingGauge);
      
      State_O_Sc_ResettingGauge
          ->AddTransition(State_O_Scanning->Fin());
      
      //State_O_Sc_SettingGauge internal structure and behavior
      {
        State_O_Sc_SG_Waiting = new sState(State_O_Sc_SettingGauge
                                            ,"State_O_Sc_SG_Waiting");
        State_O_Sc_SG_Waiting
            ->SetAsInitial()
            ->OnEnterCall(SLOT(SetGaugeRegime()))
            ->AddTransition(SIGNAL(GaugeBusy()) ,State_O_Idle)
            ->AddTransition(SIGNAL(GaugePeriodChanged(sPhysValue))
                                       ,State_O_Sc_SettingGauge->Fin());
      }
      //State_O_Sc_ResettingGauge internal structure and behavior
      {
        State_O_Sc_RG_Waiting = new sState(State_O_Sc_ResettingGauge
                                            ,"State_O_Sc_RG_Waiting");
                                            
        State_O_Sc_RG_Waiting
            ->SetAsInitial()
            ->OnEnterCall(SLOT(ResetGaugeRegime()))
            ->AddTransition(SIGNAL(GaugePeriodChanged(sPhysValue))
                           ,SLOT(RestartGauge())
                           ,State_O_Sc_ResettingGauge->Fin());
      }
      //State_O_Sc_Acquisition internal structure and behavior
      {
        State_O_Sc_Acquisition
            ->AddTransition(SIGNAL(Break()),SLOT(BreakScanning()) ,State_O_Sc_Deciding );

        State_O_Sc_Ac_SettingX = new sState(State_O_Sc_Acquisition
                                         ,"State_O_Sc_Ac_SettingX");
        State_O_Sc_Ac_GettingY = new sState(State_O_Sc_Acquisition
                                       ,"State_O_Sc_Ac_GettingY");
        State_O_Sc_Ac_CheckingFin = new sState(State_O_Sc_Acquisition
                                       ,"State_O_Sc_Ac_CheckingFin");
        State_O_Sc_Ac_SettingX
            ->SetAsInitial()
            ->AddTransition(SIGNAL(EffectorIsAtPos(sPhysValue))
                           ,State_O_Sc_Ac_GettingY);

        State_O_Sc_Ac_GettingY
            ->OnEnterCall(SLOT(GaugeStartMeasure()))
            ->AddTransition<sPhysValue>(SIGNAL(GaugeMeasured(sPhysValue))
                                       ,SLOT(AddNewData(sPhysValue))
                                       ,State_O_Sc_Ac_CheckingFin);

        State_O_Sc_Ac_CheckingFin
            ->OnEnterCall(SLOT(CheckScanningFinished()))
            ->AddTransition(SIGNAL(ScanContinue())
                           ,State_O_Sc_Ac_SettingX);

      }
    }
    connect(this ,SIGNAL(GaugeError(QString))
           ,this ,SLOT(ReportGaugeError(QString)) );
    connect(this ,SIGNAL(EffectorError(QString))
           ,this ,SLOT(ReportEffectorError(QString)) );
  };
}

void  sStepByStepScanner::AppearanceMakeUp ()
{
  sScanner::AppearanceMakeUp();
  try{ //get config parameters
    get_from<sPhysRangeNode>(DeviceSettings >> "ValidScanRange",ValidScanRange);
    get_from<sPhysRangeNode>(DeviceSettings >> "ValidScanSteps",ValidScanSteps);
    get_from<sPhysRangeNode>(DeviceSettings >> "ScanRange" ,ScanRange);
    get_from<sPhysValueNode>(DeviceSettings >> "ScanStep" ,ScanStep);
    get_from<sPhysValueNode>(DeviceSettings >> "MeasureTime" ,MeasureTime);
  }catch(...){
    SetInvalid(tr("Something wrong in configuration."));
  }
  //add new widgets
  {
    ScanRangeField = new sPhysRangeField(ParamArea
                                        ,sPhysRange(),false,false);
    ParamLayout->addWidget(ScanRangeField,Row++,0,1,3,Qt::AlignLeft);

    ScanStepField = new sPhysValueField(ParamArea
                                       ,sPhysValue(),false,false,false);
    ParamLayout->addWidget(ScanStepField,Row++,0,1,3,Qt::AlignLeft);

    MeasureTimeField = new sPhysValueField(ParamArea,sPhysValue(),false,false,false);
    ParamLayout->addWidget(MeasureTimeField,Row++,0,1,3,Qt::AlignLeft);
  }
  ScanRangeField->SetValidRange(ValidScanRange);
  ScanRangeField->SetValue(ScanRange);
  ScanStepField->SetValidRange(ValidScanSteps);
  ScanStepField->SetValue(ScanStep);
  MeasureTimeField->SetValue(MeasureTime);
  Translate();
}


void  sStepByStepScanner::Translate ()
{
  ScanRangeField->SetLabel(tr("Range"));
  ScanStepField->SetLabel(tr("Step"));
  MeasureTimeField->SetLabel(tr("MeasureTime"));
}


void  sStepByStepScanner::RestoreState ()
{
  sScanner::RestoreState();
  if( WidgetConfig.IsCorrect() ){
    try{ //get stored parameters
      get_from<sPhysRangeNode>(WidgetConfig >> "ScanRange" ,ScanRange);
      get_from<sPhysValueNode>(WidgetConfig >> "ScanStep" ,ScanStep);
      get_from<sPhysValueNode>(WidgetConfig >> "MeasureTime" ,MeasureTime);
    }catch(...){
      SetInvalid(tr("Something wrong in desktop stored."));
    }
  }
  ScanRangeField->SetValidRange(ValidScanRange);
  ScanRangeField->SetValue(ScanRange);
  ScanStepField->SetValidRange(ValidScanSteps);
  ScanStepField->SetValue(ScanStep);
  MeasureTimeField->SetValue(MeasureTime);
}


void  sStepByStepScanner::StoreState ()
{
  sScanner::StoreState();
  if( WidgetConfig.IsCorrect() ){
    WidgetConfig << attr("ScanRange",ScanRangeField->Value())
                 << attr("ScanStep",ScanStepField->Value())
                 << attr("MeasureTime",MeasureTimeField->Value());
  }
}


void sStepByStepScanner::ReportGaugeError (QString  error_text)
{
  emit ErrorMessage( tr("Got error from '%1':\n\n%2")
                       .arg(ToQString(Gauge->DeviceIdentifier()))
                       .arg(error_text)     );
}

void sStepByStepScanner::ReportEffectorError (QString  error_text)
{
  emit ErrorMessage( tr("Got error from '%1':\n\n%2")
                       .arg(ToQString(Effector->DeviceIdentifier()))
                       .arg(error_text)     );
}

void sStepByStepScanner::LinkToGauge ()
{
  try{
    Gauge.SetTo(DeviceLinks,"Gauge");
    Connect(Gauge ,SIGNAL(Warmed())       ,SIGNAL(GaugeLinked()) );
    Connect(Gauge ,SIGNAL(ErrorMessage(QString)) ,SIGNAL(GaugeError(QString)) );
    Connect(Gauge ,SIGNAL(PeriodChanged(sPhysValue)) 
           ,SIGNAL(GaugePeriodChanged(sPhysValue)) );
  }catch( rxDeviceLaunch E ){
    emit ErrorMessage( tr("Error while linking to a submissive device.\n"
                          "%1\nCheck configuration.")
                         .arg(ToQString(E.MsgText()))   );
  }
}

void sStepByStepScanner::LinkToEffector ()
{
  try{
    Effector.SetTo(DeviceLinks,"Effector");
    Connect(Effector ,SIGNAL(Warmed())       ,SIGNAL(EffectorLinked()) );
    Connect(Effector ,SIGNAL(ErrorMessage(QString)),SIGNAL(EffectorError(QString)) );
    Connect(Effector ,SIGNAL(Done(sPhysValue)),SIGNAL(EffectorIsAtPos(sPhysValue)) );
  }catch( rxDeviceLaunch E ){
    emit ErrorMessage( tr("Error while linking to a submissive device.\n"
                          "%1\nCheck configuration.")
                         .arg(ToQString(E.MsgText()))   );
  }
}

void sStepByStepScanner::OnEnterToCoolingState ()
{
  Disconnect(Effector ,0 ,0);
  Effector.ReleaseExclusiveControl();
  Effector.SetToNULL();
  Disconnect(Gauge ,0 ,0);
  Gauge.ReleaseExclusiveControl();
  Gauge.SetToNULL();
  sScanner::OnEnterToCoolingState();
}


void  sStepByStepScanner::CreateNewDependence ()
{
  try{
    ScanStep = ScanStepField->Value();
    ScanRange = ScanRangeField->Value();
  }catch(rxRAlgebra){
    emit Failure(tr("Wrong units in the configuration settings."));
  }
  if( Multiplier.X.IsNaN() ){
    Multiplier.X = sPhysValue(ScanStep.Units());
    sPhysRange                  RX(0*Multiplier.X,100*Multiplier.X);
    SetChartScales(sChartScales(sChartScale(ChartScales().X.Name,RX)
                               ,ChartScales().Y
                               ));
  };
  try{
    NumOfPoints = Round((ScanRange.Size()/ScanStep).ToReal()+1);
    real ArbFrom = (ScanRange.From() / Multiplier.X).ToReal();
    real ArbStep = (ScanStep / Multiplier.X).ToReal();

    DataArrayRecord =
        sNodePtr(new sDataArrayRecord(NumOfPoints,ArbFrom,ArbStep));
    DataArrayRecord->SetMultiplier(sPhysPair(Multiplier.X,Multiplier.Y));
    Dependence = DataArrayRecord;
  }catch(rxRAlgebra){
    emit Failure(tr("Wrong units in the console fields."));
  }
}


void  sStepByStepScanner::StartScanning ()
{
  CurrentPointNo=0;
  Connect(Gauge ,SIGNAL(NewData(sPhysValue)) ,SIGNAL(GaugeMeasured(sPhysValue)) );
  Gauge->SetAutoRepeat(false);
  SendDirective<sPhysValue>(Effector,sEffector::cSetTo,ScanRange.From());
}


void  sStepByStepScanner::StopScanning ()
{
  SendCommand(Gauge,sGauge::cStop);
  Disconnect(Gauge,SIGNAL(NewData(sPhysValue)) ,SIGNAL(GaugeMeasured(sPhysValue)));
}

void  sStepByStepScanner::BreakScanning ()
{
  SendCommand(Effector,sEffector::cBreak);
  SendCommand(Gauge,sGauge::cStop);
  Disconnect(Gauge,SIGNAL(NewData(sPhysValue)) ,SIGNAL(GaugeMeasured(sPhysValue)));
}


void  sStepByStepScanner::TakeDevicesControl ()
{
  if( ! Gauge.TakeExclusiveControl() ){
    QMessageBox::critical(this
        ,tr("'%1' cannot scan")
        .arg(ToQString(DeviceIdentifier()))
        ,tr("Gauge '%1' is busy now.\n\n  Try again later.")
        .arg(ToQString(Gauge->DeviceIdentifier()))
        );
    emit DevicesBusy();
  }else if( ! Effector.TakeExclusiveControl() ){
    Gauge.ReleaseExclusiveControl();
    QMessageBox::critical(this,tr("Denied")
                         ,tr("Device '%1' is busy.\nTry again later.")
                         .arg(ToQString(Effector->DeviceIdentifier())) );
    emit DevicesBusy();
  }else{
    emit DevicesOK();
  };
}

void  sStepByStepScanner::SetGaugeRegime ()
{
  if( ! Gauge.TakeExclusiveControl() ){
    QMessageBox::critical(this
        ,tr("'%1' cannot scan")
        .arg(ToQString(DeviceIdentifier()))
        ,tr("Gauge '%1' is busy now.\n\n  Try again later.")
        .arg(ToQString(Gauge->DeviceIdentifier()))
        );
    emit GaugeBusy();
  }else{
    GaugeWasStarted = Gauge->Started();
    GaugeWasInAutoRepeatMode = Gauge->AutoRepeat();
    OldPeriod = Gauge->Period();
    SendCommand(Gauge ,sGauge::cStop);
    SendDirective(Gauge ,sGauge::cSetPeriod ,MeasureTime);
  }
}


void  sStepByStepScanner::ResetGaugeRegime ()
{
  if( GaugeWasInAutoRepeatMode != Gauge->AutoRepeat() ){
    Gauge->SetAutoRepeat(GaugeWasInAutoRepeatMode);
  }
  SendDirective(Gauge ,sGauge::cSetPeriod ,OldPeriod);
}


void  sStepByStepScanner::RestartGauge ()
{
  if( GaugeWasStarted ){
    SendCommand(Gauge ,sGauge::cStart ,cReleaseControl);
  }
}

void  sStepByStepScanner::GaugeStartMeasure ()
{
  SendCommand(Gauge ,sGauge::cStart);
}

void  sStepByStepScanner::ReleaseControl ()
{
  Effector.ReleaseExclusiveControl();
  Gauge.ReleaseExclusiveControl();
}


void  sStepByStepScanner::AddNewData (sPhysValue  Data)
{
  if( ! DataArrayRecord.IsCorrect() ){
    return;
  };
  if( Multiplier.X.IsNaN() ){
    return;
  };
  if( Multiplier.Y.IsNaN() ){
    Multiplier.Y = sPhysValue(Data.Units());
    sPhysRange                  RY(-Abs(Data),3*Abs(Data));
    SetChartScales(sChartScales(ChartScales().X
                               ,sChartScale(ChartScales().Y.Name,RY)
                               ));
  };
  if( DataArrayRecord->Multiplier().Y.IsNaN() ){
    DataArrayRecord->SetMultiplier(sPhysPair(Multiplier.X,Multiplier.Y));
  };
  sPhysValue                  Y(_Unitsless_);
  Y = Data / Multiplier.Y;
  DataArrayRecord->SetValue(CurrentPointNo,Y.ToReal());
  CurrentPointNo++;
}

void  sStepByStepScanner::CheckScanningFinished ()
{
  if (CurrentPointNo >= NumOfPoints)
  {
    emit ScanningHasCompleted();
  } else
  {
    SendDirective<sPhysValue>(Effector,sEffector::cSetTo
                             ,ScanRange.From()+ScanStep*CurrentPointNo);
    emit ScanContinue();
  }
}

//------------------------------------------------------------------------------
} //namespace RA
