////////////////////////////////////////////////////////////////////////////////
/*! @file LomoCCDDriver.cpp Драйвер ПЗС-линейки ЛОМО.
- Part of CCDLomo - Research Assistant Plugin package.
- Uses  RAGUI - Research Assistant Graphical User Interface Library.
- Uses  Qt v.5.0.2  - http://qt-project.org/
- Uses  RANet - Research Assistant Net Library.
- Copyright(C) 2010-2017, Dmitrii Nelson, St.Petersburg, Russia.
                          d.nelson@mail.ioffe.ru
*///////////////////////////////////////////////////////////////////////////////
#include "LomoCCDDriver.h"
#include "attr.h"
#include <QDebug>
#include "Log.h"

namespace RA {
//------------------------------------------------------------------------------

DEFINE_CLASS_TAG(sLomoCCDDriver)

//---------------------------------------------------- sLomoCCDDriver ----------

void  sLomoCCDDriver::Preconfigure (sNodePtr  DeviceCfg)
{
  sPtr<sFolderNode>           DeviceSettings;
  DeviceCfg |= "Settings";
  DeviceSettings = DeviceCfg >> "Settings";
  if( DeviceSettings.IsCorrect() ){
    //! @todo{Instruments} Why SetAttr?
    DeviceSettings |= attr("NumberOfPixels",3648);
    DeviceSettings |= attr("CCD_SH",64);
    DeviceSettings |= attr("CCD_ICG",64);
    DeviceSettings |= attr("MiddlePixel",1733);
    DeviceSettings |= attr("WL_0",sPhysValue(0.02389,"nm"));
    DeviceSettings |= attr("WL_k",sPhysValue(-7.8e-6,_Unitsless_));
    DeviceSettings |= attr("CCDDeviceName","CCD Utils");
  };
  sTunableCCDDriver::Preconfigure(DeviceCfg);
}

sLomoCCDDriver::~sLomoCCDDriver ()
{
}

sLomoCCDDriver::sLomoCCDDriver (sString   DevicetId
                               ,sNodePtr  DeviceCfg)
  :sTunableCCDDriver(DevicetId,DeviceCfg)
  ,WL_0(sPhysValue(1.0,"nm"))
  ,WL_k(sPhysValue(1,_Unitsless_))
{
}


void  sLomoCCDDriver::PostConstructor ()
{
  sTunableCCDDriver::PostConstructor();
  try{ //get config parameters
    get_from<sIntegerNode>(DeviceSettings >> "CCD_SH"   ,CCD_SH);
    get_from<sIntegerNode>(DeviceSettings >> "CCD_ICG"   ,CCD_ICG);
    get_from<sTextNode>(DeviceSettings >> "CCDDeviceName" ,CCDDeviceName);
    get_from<sPhysValueNode>(DeviceSettings >> "WL_0" ,WL_0);
    get_from<sPhysValueNode>(DeviceSettings >> "WL_k" ,WL_k);
  }catch(...){
    SetInvalid(tr("Something wrong in configuration."));
  }
}

void  sLomoCCDDriver::ConnectUSBDevice ()
{
    FT_STATUS CCDDeviceStatus;
    CCDDeviceStatus = FT_OpenEx((void*)ptr_to_const_char(CCDDeviceName)
                               ,FT_OPEN_BY_DESCRIPTION,&CCDDeviceHandle);
    //qDebug()<<"CCDDeviceName - "<<CCDDeviceName;
    //qDebug()<<"FT_OpenEx - "<<CCDDeviceStatus;
    if(CCDDeviceStatus != FT_OK){
        emit ErrorMessage(FormCCDErrorMessage(CCDDeviceStatus));
        return;
    }
    CCDDeviceStatus = FT_ResetDevice(CCDDeviceHandle);
    //qDebug()<<"FT_ResetDevice - "<<CCDDeviceStatus;
    if(CCDDeviceStatus != FT_OK){
        emit ErrorMessage(FormCCDErrorMessage(CCDDeviceStatus));
        return;
    }
    CCDDeviceStatus = FT_Purge(CCDDeviceHandle,FT_PURGE_RX | FT_PURGE_TX);
    //qDebug()<<"FT_Purge - "<<CCDDeviceStatus;
    if(CCDDeviceStatus != FT_OK){
        emit ErrorMessage(FormCCDErrorMessage(CCDDeviceStatus));
        return;
    }
    if(!SendParam(0x90,CCD_ICG)) return;
    if(!SendParam(0xB0,CCD_SH)) return;
    emit Connected();
}

void  sLomoCCDDriver::GetCCD_Data ()
{
    DWORD BytesToRead = 0;
    FT_STATUS CCDDeviceStatus;
    CCDDeviceStatus = FT_GetQueueStatus(CCDDeviceHandle,&BytesToRead);
    if(CCDDeviceStatus != FT_OK)
    {
        emit ErrorMessage(FormCCDErrorMessage(CCDDeviceStatus));
        return;
    }
    if(BytesToRead > 0)
    {
//        RANet::Log.Put(sLog::Info,"CCDLomo",sString("BytesToRead = ") + sString::FromInteger(BytesToRead));
        DWORD ReallyRead;
        BYTE *TmpBuffer;
        TmpBuffer = new BYTE[BytesToRead];
        CCDDeviceStatus = FT_Read(CCDDeviceHandle,TmpBuffer,BytesToRead,&ReallyRead);
        if(CCDDeviceStatus != FT_OK)
        {
            emit ErrorMessage(FormCCDErrorMessage(CCDDeviceStatus));
            delete[] TmpBuffer;
            return;
        }
        for(DWORD i = 0;i < ReallyRead; i ++) {
            AddByte(TmpBuffer[i]);
//            qDebug() << "Read i="<<i<<"of "<<ReallyRead<<" data=" << TmpBuffer[i];
        }
        delete[] TmpBuffer;
//        RANet::Log.Put(sLog::Info,"CCDLomo",sString("Bytes Ttreated - ") + sString::FromInteger(BytesToRead));
    }
}

void  sLomoCCDDriver::CalculateWavelengths ()
{
  sPhysValue dWL = WL_0 + WL_k * WLPosition;
  sPhysValue LeftPos = WLPosition - dWL*MiddlePixel;
  sPhysValue RightPos = WLPosition + dWL*((NumOfPixels -1) - MiddlePixel);
  real dX_0 = ((WL_0 + WL_k * LeftPos)/(sPhysValue(1.0,"nm"))).ToReal();
  real dX_1 = ((WL_0 + WL_k * RightPos)/(sPhysValue(1.0,"nm"))).ToReal();
  real dX = (dWL/(sPhysValue(1.0,"nm"))).ToReal(); // in "nm/pixel"
  real d2X = (dX_1 - dX_0)/NumOfPixels;
  real  XM = (WLPosition/(sPhysValue(1.0,"nm"))).ToReal(); // in "nm"
  for (int i = 0; i < NumOfPixels; i++) {
    Wavelengths.ArrBody[i] = XM + (dX + d2X *(i - MiddlePixel))*(i - MiddlePixel);
  }
  emit NewWavelengths (Wavelengths);
}

void  sLomoCCDDriver::StartCCD ()
{
    BYTE byteToSend = 0x80;
    ResetTriadeBuffer();
    PointNo = 0;
    FT_STATUS CCDDeviceStatus;
    CCDDeviceStatus = FT_Purge(CCDDeviceHandle,FT_PURGE_RX | FT_PURGE_TX);
    if(CCDDeviceStatus != FT_OK){
        emit ErrorMessage(FormCCDErrorMessage(CCDDeviceStatus));
        return;
    }
    DWORD bytesWritten;
    CCDDeviceStatus = FT_Write(CCDDeviceHandle,&byteToSend,1,&bytesWritten);
    if(CCDDeviceStatus != FT_OK)
        emit ErrorMessage(FormCCDErrorMessage(CCDDeviceStatus));
    State_O_Measuring->SetTimeout(1);
}

void  sLomoCCDDriver::StopCCD ()
{
    BYTE byteToSend = 0xA0;
    FT_STATUS CCDDeviceStatus;
    DWORD bytesWritten;
    CCDDeviceStatus = FT_Write(CCDDeviceHandle,&byteToSend,1,&bytesWritten);
    if(CCDDeviceStatus != FT_OK)
        emit ErrorMessage(FormCCDErrorMessage(CCDDeviceStatus));
    WLDriver.ReleaseExclusiveControl();
}

void  sLomoCCDDriver::SetCCD_Parameters (sCCDDriver::sParameters Params)
{
  Parameters = Params;
  if( ! WLDriver.TakeExclusiveControl() ){
    emit IsBusy();
  }else{
    int _IT = 1;
    try{
      _IT = Round((Parameters.IntegrationTime/(sPhysValue(1.0e-5,"s"))).ToReal());
      if (_IT < 1) _IT = 1;
      else if (_IT > 65535) _IT = 65535;
    }catch(rxRAlgebra){
      SetInvalid(tr("Something wrong with units."));
    }
    WORD _ITime = _IT;
    if(SendParam(0xC0,_ITime))
      emit ParamsSet();
  }
}

QString  sLomoCCDDriver::FormCCDErrorMessage(FT_STATUS CCDStatus)
{
    QString ErMsg = "";
    switch (CCDStatus){
    case 1:
            ErMsg = "Ошибка USB - Неверный дескриптор";
            break;
    case 2:
            ErMsg = "Ошибка USB - Устройство не найдено";
            break;
    case 3:
            ErMsg = "Ошибка USB - Канал не инициализирован";
            break;
    case 4:
            ErMsg = "Ошибка USB - Ошибка ввода-вывода";
            break;
    case 5:
            ErMsg = "Ошибка USB - Передавать нечего";
            break;
    case 6:
            ErMsg = "Ошибка USB - Недоступный ресурс";
            break;
    case 7:
            ErMsg = "Ошибка USB - Неверная скорость обмена";
            break;
    case 8:
            ErMsg = "Ошибка USB - Стирание недоступно";
            break;
    case 9:
            ErMsg = "Ошибка USB - Запись невозможна";
            break;
    case 10:
            ErMsg = "Ошибка USB - Ошибка записи,Ошибка USB";
            break;
    case 11:
            ErMsg = "Ошибка USB - Чтение ЭСППЗУ невозможно";
            break;
    case 12:
            ErMsg = "Ошибка USB - Запись ЭСППЗУ невозможна";
            break;
    case 13:
            ErMsg = "Ошибка USB - Стирание ЭСППЗУ невозможно";
            break;
    case 14:
            ErMsg = "Ошибка USB - ЭСППЗУ отсутствует";
            break;
    case 15:
            ErMsg = "Ошибка USB - ЭСППЗУ не программируется";
            break;
    case 16:
            ErMsg = "Ошибка USB - Неверные аргументы";
            break;
    default:
            ErMsg = "Ошибка USB - Неизвестная ошибка";
            break;
    };
    return ErMsg;
}

bool sLomoCCDDriver::SendParam(BYTE Command,WORD Param)
{
//  qDebug() << "sLomoCCDDriver, Param = " << Param;
    BYTE bytesToSend[3];
    DWORD bytesWritten;
    FT_STATUS CCDDeviceStatus;
    bytesToSend[0] = Command;
    bytesToSend[1] = ((BYTE*)&Param)[0];
    bytesToSend[2] = ((BYTE*)&Param)[1];
//  qDebug() << "sLomoCCDDriver, bytesToSend = " << bytesToSend[1] << " ; "<< bytesToSend[2];
    if(bytesToSend[1] & 0x80) bytesToSend[0] |= 0x01;
    if(bytesToSend[2] & 0x80) bytesToSend[0] |= 0x02;
    bytesToSend[1] &= 0x7F;
    bytesToSend[2] &= 0x7F;
    CCDDeviceStatus = FT_Write(CCDDeviceHandle,&bytesToSend,3,&bytesWritten);
    if(CCDDeviceStatus != FT_OK) {
        emit ErrorMessage(FormCCDErrorMessage(CCDDeviceStatus));
        return false;
    }
    return true;
}

// ------ Это взято у ЛОМО и слегка модифицировано -----------------

void  sLomoCCDDriver::AddByte(BYTE  DataByte )
{
   TriadeBuffer[NumInTriade ++] =  DataByte ;
   if(( DataByte & 0x80) == 0) {    //Промежуточный байт
      if(NumInTriade == 3) {        // Лишний байт
         TriadeBuffer[0] = TriadeBuffer[1];
         TriadeBuffer[1] = TriadeBuffer[2];
         NumInTriade = 2;
      }
      return;
   }else {                       //Командный байт
      if(NumInTriade < 2) {
         NumInTriade = 0;
         PointNo = 0;
         DataArray.Reset();
         return;
      }
      BYTE CommandCode =  DataByte & 0xF0;
      if(CommandCode == FRAME_START) { // Первая точка
         if(PointNo != 0){
            DataArray.Reset();
            PointNo = 0;
         }
         DataArray.ArrBody[NumOfPixels - PointNo - 1] = ConvertTriadeBuffer();
         PointNo ++;
         NumInTriade = 0;
         return;
      }else if(CommandCode == FRAME_MIDDLE) {
         if(PointNo == NumOfPixels - 1) {
            DataArray.Reset();
            PointNo = 0;
            NumInTriade = 0;
            return;
         }else if(PointNo == 0) {
            NumInTriade = 0;
            return;
         }else {
            DataArray.ArrBody[NumOfPixels - PointNo - 1] = ConvertTriadeBuffer();
            PointNo ++;
            NumInTriade = 0;
            return;
         }
      }else if(CommandCode == FRAME_END) {
         if(PointNo == NumOfPixels - 1) { //Конец фрейма
            DataArray.ArrBody[NumOfPixels - PointNo - 1] = ConvertTriadeBuffer();
            PointNo ++;
            NumInTriade = 0;
            PointNo = 0;
            emit NewData(DataArray);
//            RANet::Log.Put(sLog::Info,"CCDLomo","New Data Array Formed");
            return;
         }else if(PointNo == 0) {
            NumInTriade = 0;
            return;
         }else {
            NumInTriade = 0;
            DataArray.Reset();
            PointNo = 0;
            return;
         }
      }else {
         NumInTriade = 0;
         DataArray.Reset();
         PointNo = 0;
         return;
      };
   }
}
// -----------------------------------------------------------------

void sLomoCCDDriver::ResetTriadeBuffer()
{
    TriadeBuffer[0] = 0x00;
    TriadeBuffer[1] = 0x00;
    TriadeBuffer[2] = 0x00;
    NumInTriade = 0;
}

WORD sLomoCCDDriver::ConvertTriadeBuffer()
{
    WORD Result = ~TriadeBuffer[0] & 0x7F;
    Result = (Result << 8) + (~TriadeBuffer[1] & 0x7F);
    if (TriadeBuffer[2] & 0x01) Result |= 0x80;
    if (TriadeBuffer[2] & 0x02) Result |= 0x8000;
    return Result;
}

//------------------------------------------------------------------------------
} //namespace RA
