////////////////////////////////////////////////////////////////////////////////
/*! @file COMPortTeleCom.cpp   Short message service via serial port.
- Part of COMPort - Research Assistant Plugin package.
- Uses  RAGUI - Research Assistant Graphical User Interface.
- Uses  QtGui v.4.6  -  http://qt.nokia.com/
- Uses  RANet - Research Assistant Net Library (based on ANSI C++).
- Copyright(C) 2010-2013, Viktor E. Bursian, St.Petersburg, Russia.
                          Viktor.Bursian@mail.ioffe.ru
- Copyright(C) 2010-2013, Dmitry K. Nelson, St.Petersburg, Russia.
                          D.Nelson@mail.ioffe.ru
*///////////////////////////////////////////////////////////////////////////////
#include "COMPortTeleCom.h"
#include "attr.h"
#include "StringsAndQStrings.h"
#include <QDebug>
#include <QtAddOnSerialPort/serialportinfo.h>
namespace RA {
//------------------------------------------------------------------------------

DEFINE_CLASS_TAG(sCOMPortTeleCom)

//---------------------------------------------------------- sCOMPortTeleCom ---

void  sCOMPortTeleCom::Preconfigure (sNodePtr  DeviceCfg)
{
  sPtr<sFolderNode>              DeviceSettings;
  DeviceCfg |= "Settings";
  DeviceSettings = DeviceCfg->GetAttr("Settings");
  if( DeviceSettings.IsCorrect() ){
    DeviceSettings |= attr("Name","COM1");
    DeviceSettings |= attr("NameIsDeviceName",0);
    DeviceSettings |= attr("BaudRate",115200);
  };
  sTeleCom::Preconfigure(DeviceCfg);
}


sCOMPortTeleCom::~sCOMPortTeleCom ()
{
  if( ThePort ){
    ThePort->close();
    delete ThePort;  ThePort = NULL;
  };
}


sCOMPortTeleCom::sCOMPortTeleCom (sString   DevicetId
                                 ,sNodePtr  DeviceCfg)
    :sTeleCom(DevicetId,DeviceCfg)
    ,Name("COM1")
    ,NameIsDeviceName(false)
    ,BaudRate(115200)
    ,TerminalSymbols("\x0D\x0A")
    ,ThePort(new SerialPort())
{
}


void  sCOMPortTeleCom::PostConstructor ()
{
  sTeleCom::PostConstructor();

  if( DeviceSettings.IsCorrect() ){//get config parameters
    sPtr<sTextNode>             S;
    sPtr<sIntegerNode>          I;

    S=DeviceSettings->GetAttr("Name");  //берём указатель на значение атрибута
    if( S.IsCorrect() )     //проверяем, что значение есть и оно строковое
      Name = ToQString( (sString)(*S) ); //преобразуем его в QString

    I=DeviceSettings->GetAttr("NameIsDeviceName");
    if( I.IsCorrect() )
      NameIsDeviceName = (int)(*I) > 0;

    I=DeviceSettings->GetAttr("BaudRate");
    if( I.IsCorrect() )
      BaudRate = (int)(*I);
  };

  //evolution of the inherited state machine (behaviour)
  {
    //evolution of the State_Warming
    {
      //create new substates and set initial one
      {
        State_W_OpeningPort = new QState(State_Warming());
        State_Warming()->setInitialState(State_W_OpeningPort);
      };
      //State_W_OpeningPort behaviour
      {
        AddTransition(State_W_OpeningPort ,SIGNAL(PortOK())
                     ,State_Warming_Success());
      };
    };
  };
}


void sCOMPortTeleCom::OnEnterToOperatingState()
{
  sTeleCom::OnEnterToOperatingState();
  connect(ThePort, SIGNAL(readyRead()), this, SLOT(ReadAvailableBytes()));
}


void sCOMPortTeleCom::OnExitFromOperatingState()
{
  disconnect(ThePort, SIGNAL(readyRead()), this, SLOT(ReadAvailableBytes()));
  sTeleCom::OnExitFromOperatingState();
}


void  sCOMPortTeleCom::OnEnterToCoolingState ()
{
  ThePort->close();
}


void sCOMPortTeleCom::PutTelex (sTelex  Telex)
{
  QByteArray             TelexWithTerminal = Telex + TerminalSymbols;
  long                   BytesWritten  = ThePort->write(TelexWithTerminal);
  if( TelexWithTerminal.size() !=  BytesWritten ){
    emit IOError("I/O Error: Writing: '"+((QString)TelexWithTerminal)+"' ("
                +QString::number(TelexWithTerminal.size())+" bytes), written "
                +QString::number(BytesWritten)+" bytes."
                );
  };
}


void sCOMPortTeleCom::ReadAvailableBytes ()
{
  while( qint64  AvailableBytes = ThePort->bytesAvailable() ){
    QByteArray                  BA = ThePort->read(AvailableBytes);
    InputBuffer.append(BA);
    int                         TerminalPos;
    while( (TerminalPos = InputBuffer.indexOf(TerminalSymbols)) >= 0 ){
      sTelex                    Telex( InputBuffer.left(TerminalPos) );
      InputBuffer.remove(0, TerminalPos + TerminalSymbols.size());
      emit GotTelex(Telex);
    };
  };
}


void  sCOMPortTeleCom::OpenPort ()
{

}


/*
void  sCOMPortTeleCom::Connect ()
{
  if( DriverState() == Offline ){
    SetDriverState(Warming);
    SerialPort::Rate      BR=SerialPort::Rate115200;
    if( BaudRate >= 115200 ){
      BaudRate=115200;  BR=SerialPort::Rate115200;
    }else if( BaudRate >= 9600 ){
      BaudRate=9600;  BR=SerialPort::Rate9600;
    }else if( BaudRate >= 1200 ){
      BaudRate=1200;  BR=SerialPort::Rate1200;
    };

    bool OK = true;
    if (NameIsDeviceName){
      OK = FindDeviceAndSetPortName();
    } else {
      Port->setPort(Name);
    };

    if ((OK)&& (Port->open(QIODevice::ReadWrite | QIODevice::Unbuffered))) {
      if (!Port->setRate(BR)) {
          OK = false;
          qDebug() << "Set baud rate " <<  BaudRate << " error.";
          Port->close();
      };
      if (!Port->setDataBits(SerialPort::Data8)) {
          OK = false;
          qDebug() << "Set data bits " <<  SerialPort::Data8 << " error.";
          Port->close();
      };
    };
    if (OK){
      qDebug() << "Serial device " << Port->portName() << " open in " << Port->openMode();
      qDebug() << "= Default parameters =";
      qDebug() << "Device name            : " << Port->portName();
      qDebug() << "Baud rate              : " << Port->rate();
      qDebug() << "Data bits              : " << Port->dataBits();
      qDebug() << "Parity                 : " << Port->parity();
      qDebug() << "Stop bits              : " << Port->stopBits();
      qDebug() << "Flow                   : " << Port->flowControl();
      SetDriverState(Standby);

      Port->setDtr(true);  // This is set to switch on
      Port->setRts(true);  // power supply to ML327
      Port->flush();
      Port->reset();
    }
    else {
      qDebug() << "Error opened serial device " << Port->portName();
      SetDriverState(Offline);
    }

  };
}


bool sCOMPortTeleCom::FindDeviceAndSetPortName()
{
    bool OK = false;
    int numBytes = 0;
    QString ThePortName = "";
    foreach (const SerialPortInfo &info, SerialPortInfo::availablePorts()) {
        qDebug() << "Port name: " << info.portName();
        InputString.clear();
        Port->setPort(info);
        if ((!OK) && Port->open(QIODevice::ReadWrite)){
            Port->setRate(SerialPort::Rate115200);
            Port->setDataBits(SerialPort::Data8);
            Port->write("?"+TerminalSymbols);
            qDebug() << "Sending <?> to port " << info.portName();
            while (Port->waitForReadyRead(100)){
               numBytes = Port->bytesAvailable();
               QByteArray ba = Port->read(numBytes);
               InputString.append(ba);
            }
            qDebug() << "InputString = " << InputString;
            int Pos = InputString.indexOf(TerminalSymbols);
            if (InputString.left(Pos) == Name){
                ThePortName = info.portName();
                InputString.clear();
                OK = true;
            }
            Port->close();
        }
    }
    Port->setPort(ThePortName);
    return OK;
}
*/

//------------------------------------------------------------------------------
} //namespace RA

