////////////////////////////////////////////////////////////////////////////////
/*! @file TeleCOMPort.cpp   Short message service via serial port.
- Part of COMPort - Research Assistant Plugin package.
- Uses  RAGUI - Research Assistant Graphical User Interface.
- Uses  QtGui v.4.6  -  http://qt.nokia.com/
- Uses  RANet - Research Assistant Net Library (based on ANSI C++).
- Copyright(C) 2010-2013, Viktor E. Bursian, St.Petersburg, Russia.
                          Viktor.Bursian@mail.ioffe.ru
- Copyright(C) 2010-2013, Dmitry K. Nelson, St.Petersburg, Russia.
                          D.Nelson@mail.ioffe.ru
*///////////////////////////////////////////////////////////////////////////////
#include "TeleCOMPort.h"
#include "attr.h"
#include "StringsAndQStrings.h"
#include <QDebug>
#include <QtSerialPort/QSerialPortInfo>
namespace RA {
//------------------------------------------------------------------------------

DEFINE_CLASS_TAG(sTeleCOMPort)

csString  sTeleCOMPort::DefaultTerminalSymbols("0d0a");
                        //which means 2 bytes - CR and LF

//--------------------------------------------------------- sSerialSequencer ---

sSerialSequencer::~sSerialSequencer ()
{
  if( Port ){
    disconnect(Port, 0, this, 0);
    Port->close();
    delete  Port;  Port = NULL;
  };
}


sSerialSequencer::sSerialSequencer (QObject *          parent
                                   ,sPtr<sFolderNode>  settings
                                   ,const QString &    port_name)
    :QObject(parent)
    ,DeviceSettings(settings)
    ,Port( new QSerialPort(port_name,this) )
{
  GetConfigParameters();
}


sSerialSequencer::sSerialSequencer (QObject *                parent
                                   ,sPtr<sFolderNode>        settings
                                   ,const QSerialPortInfo &  info)
    :QObject(parent)
    ,DeviceSettings(settings)
    ,Port( new QSerialPort(info,this) )
{
  GetConfigParameters();
}


void sSerialSequencer::GetConfigParameters()
{
  BaudRate = 115200;
  Parity = QSerialPort::NoParity;
  DataBits = QSerialPort::Data8;
  StopBits = QSerialPort::OneStop;
  TerminalSymbols = QByteArray::fromHex
                    ((literal)sTeleCOMPort::DefaultTerminalSymbols);
  try{
    integer                     I;
    sString                     S;
    if( get_from<sIntegerNode>(DeviceSettings>>"BaudRate",BaudRate) ){
//      if( BaudRate >= 115200 ){
//        BaudRate = 115200;
//      }else if( BaudRate >= 57600 ){
//        BaudRate=57600;
//      }else if( BaudRate >= 38400 ){
//        BaudRate=38400;
//      }else if( BaudRate >= 19200 ){
//        BaudRate=19200;
//      }else if( BaudRate >= 9600 ){
//        BaudRate=9600;
//      }else if( BaudRate >= 4800 ){
//        BaudRate=4800;
//      }else if( BaudRate >= 2400 ){
//        BaudRate=2400;
//      }else{
//        BaudRate=1200;
//      }
    }
    if( get_from<sTextNode>(DeviceSettings>>"Parity",S) ){
      if( S.UpperCase() == "N" ){
        Parity = QSerialPort::NoParity;
      }else if( S.UpperCase() == "E" ){
        Parity = QSerialPort::EvenParity;
      }else if( S.UpperCase() == "O" ){
        Parity = QSerialPort::OddParity;
      }
    }
    if( get_from<sIntegerNode>(DeviceSettings>>"DataBits",I) ){
      switch( I ){
        case 5:  DataBits = QSerialPort::Data5;  break;
        case 6:  DataBits = QSerialPort::Data6;  break;
        case 7:  DataBits = QSerialPort::Data7;  break;
        case 8:  DataBits = QSerialPort::Data8;  break;
        default: ;
      }
    }
    if( get_from<sIntegerNode>(DeviceSettings>>"StopBits",I) ){
      switch( I ){
        case 1:  StopBits = QSerialPort::OneStop;  break;
        case 3:  StopBits = QSerialPort::OneAndHalfStop;  break;
        case 2:  StopBits = QSerialPort::TwoStop;  break;
        default: ;
      }
    }
    if( get_from<sTextNode>(DeviceSettings>>"TerminalSymbols",S) ){
      TerminalSymbols = QByteArray::fromHex((literal)S);
      #pragma GCC diagnostic push
      #pragma GCC diagnostic ignored "-Wsign-compare"
      if( 2*TerminalSymbols.length() < S.Length()  ){
        TerminalSymbols = QByteArray::fromHex
                          ((literal)sTeleCOMPort::DefaultTerminalSymbols);
      }
      #pragma GCC diagnostic pop
    }
  }catch(...){
  }
}


void  sSerialSequencer::Open ()
{
  qDebug()<< "Opening COM port " << Port->portName();

  if( ! Port->open(QIODevice::ReadWrite /*| QIODevice::Unbuffered*/) ){
    emit OpeningError( tr("Cannot open '%1'.").arg(Port->portName()) );
    return;
  };
  if( ! Port->setBaudRate(BaudRate) ){
    emit OpeningError( tr("Port '%1': set baud rate %2 error.")
                         .arg(Port->portName()).arg(BaudRate) );
    return;
  };
  if( ! Port->setParity(Parity) ){
    emit OpeningError( tr("Port '%1': set parity %2 error.")
                         .arg(Port->portName()).arg(Parity) );
    return;
  };
  if( ! Port->setDataBits(DataBits) ){
    emit OpeningError( tr("Port '%1': set data bits %2 error.")
                         .arg(Port->portName()).arg(DataBits) );
    return;
  };
  if( ! Port->setStopBits(StopBits) ){
    emit OpeningError( tr("Port '%1': set stop bits %2 error.")
                         .arg(Port->portName()).arg(StopBits) );
    return;
  };

  qDebug() << Port->portName() << "is open:";
  qDebug() << "Mode     " << Port->openMode();
  qDebug() << "Baud rate" << Port->baudRate();
  qDebug() << "Data bits" << Port->dataBits();
  qDebug() << "Parity   " << Port->parity();
  qDebug() << "Stop bits" << Port->stopBits();
  qDebug() << "Flow ctrl" << Port->flowControl();

/*! @todo{Instruments} Если восстанавливать Rigaku, то для неё это было нужно.
                       Другим, возможно, мешает.
  Port->setDataTerminalReady(true);  // This is set to switch on
  Port->setRequestToSend(true);  // power supply to ML327
*/

  Port->flush();
  Port->reset();

  connect(Port, SIGNAL(readyRead()), this, SLOT(ReadAvailableBytes()));
  emit Ready();
}


void sSerialSequencer::PutTelex (sTelex  Telex)
{
  QByteArray             TelexWithTerminal = Telex + TerminalSymbols;
  long                   BytesWritten  = Port->write(TelexWithTerminal);
  if( TelexWithTerminal.size() !=  BytesWritten ){
    emit IOError("I/O Error: Writing: '"+((QString)TelexWithTerminal)+"' ("
                +QString::number(TelexWithTerminal.size())+" bytes), written "
                +QString::number(BytesWritten)+" bytes."
                );
  };
}


void sSerialSequencer::ReadAvailableBytes ()
{
  while( qint64  AvailableBytes = Port->bytesAvailable() ){
    QByteArray                  BA = Port->read(AvailableBytes);
    InputBuffer.append(BA);
    int                         TerminalPos;
    while( (TerminalPos = InputBuffer.indexOf(TerminalSymbols)) >= 0 ){
      sTelex                    Telex( InputBuffer.left(TerminalPos) );
      InputBuffer.remove(0, TerminalPos + TerminalSymbols.size());
      emit GotTelex(Telex);
    };
  };
}

//------------------------------------------------------------- sTeleCOMPort ---

void  sTeleCOMPort::Preconfigure (sNodePtr  DeviceCfg)
{
  sPtr<sFolderNode>              DeviceSettings;
  DeviceCfg |= "Settings";
  DeviceSettings = DeviceCfg >> "Settings";
  if( DeviceSettings.IsCorrect() ){
    DeviceSettings |= attr("SearchThroughAvailablePorts",false);
    DeviceSettings |= attr("PortName","COM1");
    DeviceSettings |= attr("Description","");
    DeviceSettings |= attr("Manufacturer","");
//    DeviceSettings |= attr("VendorId","");
//    DeviceSettings |= attr("ProductId","");
    DeviceSettings |= attr("IntroducementPrefix","");
    DeviceSettings |= attr("DeviceName","");
//    DeviceSettings |= attr("RequestToBeSent","");
    DeviceSettings |= attr("BaudRate",115200);
    DeviceSettings |= attr("Parity","N");
    DeviceSettings |= attr("DataBits",8);
    DeviceSettings |= attr("StopBits",1);
    DeviceSettings |= attr("TerminalSymbols",DefaultTerminalSymbols);
  };
  sTeleCom::Preconfigure(DeviceCfg);
}


sTeleCOMPort::~sTeleCOMPort ()
{
  if( TheSequencer ){
    disconnect(TheSequencer, 0, 0, 0);
    delete  TheSequencer;  TheSequencer = NULL;
  };
}


sTeleCOMPort::sTeleCOMPort (sString   DevicetId
                           ,sNodePtr  DeviceCfg)
    :sTeleCom(DevicetId,DeviceCfg)
    ,State_W_Parallel(NULL)
    ,SearchThroughAvailablePorts(false)
    ,PortName("COM1")
    ,VendorId(-1)
    ,ProductId(-1)
    ,TheSequencer(NULL)
{
}


void  sTeleCOMPort::PostConstructor ()
{
  sTeleCom::PostConstructor();

  try{ //get config parameters
    get_from<sBooleanNode>(DeviceSettings >> "SearchThroughAvailablePorts"
                                             ,SearchThroughAvailablePorts);
    get_from<sTextNode>(DeviceSettings >> "PortName" ,PortName);
    get_from<sTextNode>(DeviceSettings >> "Description" ,Description);
    get_from<sTextNode>(DeviceSettings >> "Manufacturer" ,Manufacturer);
    get_from<sIntegerNode>(DeviceSettings >> "VendorId" ,VendorId);
    get_from<sIntegerNode>(DeviceSettings >> "ProductId" ,ProductId);
    get_from<sTextNode>(DeviceSettings >> "IntroducementPrefix"
                                          ,IntroducementPrefix);
    get_from<sTextNode>(DeviceSettings >> "DeviceName" ,DeviceName);
//    get_from<sTextNode>(DeviceSettings >> "RequestToBeSent" ,RequestToBeSent);
  }catch(...){
    SetInvalid(tr("Something wrong in configuration."));
  }

  if( ! SearchThroughAvailablePorts ){
    State_W_Opening = new sState(State_Warming());
    TheSequencer= new sSerialSequencer(this,DeviceSettings,ToQString(PortName));
    State_W_Opening
        ->SetAsInitial()
        ->OnEnterCall(TheSequencer ,SLOT(Open()))
        ->AddTransition(SIGNAL(PortOK()) ,State_Warming()->Fin());
    connect(TheSequencer ,SIGNAL(OpeningError(QString))
           ,this      ,SLOT(OnOpeningError(QString)));
    connect(TheSequencer ,SIGNAL(Ready())
           ,this      ,SIGNAL(PortOK()));
  }else{
    State_W_Parallel = new sParallelStateGroup(State_Warming());
    State_W_Parallel
        ->SetAsInitial()
        ->AddTransition<psSerialSequencer>(SIGNAL(PortOK(psSerialSequencer))
                                          ,SLOT(OnPortOK(psSerialSequencer))
                                          ,State_Warming()->Fin());
    foreach (const QSerialPortInfo & Info, QSerialPortInfo::availablePorts()) {
      qDebug() << "Port found: name" << Info.portName()
               << "at (systemLocation)" << Info.systemLocation();
      qDebug() << "            description" << Info.description();
      qDebug() << "            manufacturer" << Info.manufacturer();
      qDebug() << "            serialNumber" << Info.serialNumber();
      qDebug() << "            VID" << Info.vendorIdentifier()
               << "PID" << Info.productIdentifier()
               << "          isNull" << Info.isNull()
               << "isBusy" << Info.isBusy()
               << "isValid" << Info.isValid();
      if( (Description.Empty() || (Info.description()
                                   == ToQString(Description)))
          && (Manufacturer.Empty() || (Info.manufacturer()
                                       == ToQString(Manufacturer)))
          && ((VendorId < 0) || (Info.vendorIdentifier() == VendorId))
          && ((ProductId < 0) || (Info.productIdentifier() == ProductId))
      ){
        psSerialSequencer     Sequencer = new sSerialSequencer(this
                                                              ,DeviceSettings
                                                              ,Info  );
        psPortOpeningState    OpeningState
                                    = new sPortOpeningState(State_W_Parallel
                                                           ,IntroducementPrefix
                                                               +DeviceName
                                                           ,Sequencer);
        connect(OpeningState ,SIGNAL(Success(psSerialSequencer))
               ,this         ,SIGNAL(PortOK(psSerialSequencer))  );
        OpeningStates.push_front(OpeningState);
      };
    };

    //!@test Замена испускания SIGNAL(CoolDown())на переход в State_Cooling
    //AddReaction(State_W_Parallel ,SIGNAL(finished())
    //           ,SIGNAL(CoolDown()));
    State_W_Parallel->AddTransition(State_Cooling());
  };
}


void sTeleCOMPort::OnExitFromWarmingState()
{
  while( ! OpeningStates.empty() ){
    psPortOpeningState  OS = OpeningStates.front();
    if( OS->Sequencer == TheSequencer ){
      OS->Sequencer = NULL;
    };
    OpeningStates.pop_front();
  };
  State_Warming()->setInitialState(State_Warming()->Fin());
  if( State_W_Parallel ){
    State_W_Parallel->deleteLater();  State_W_Parallel = NULL;
  };
}


void sTeleCOMPort::OnEnterToOperatingState()
{
  sTeleCom::OnEnterToOperatingState();
  connect(TheSequencer ,SIGNAL(GotTelex(sTelex))
         ,this      ,SIGNAL(GotTelex(sTelex)));
}


void sTeleCOMPort::OnExitFromOperatingState()
{
  disconnect(TheSequencer ,SIGNAL(GotTelex(sTelex))
            ,this      ,SIGNAL(GotTelex(sTelex)));
  sTeleCom::OnExitFromOperatingState();
}


void  sTeleCOMPort::OnEnterToCoolingState ()
{
  if( TheSequencer ){
    disconnect(TheSequencer, 0, this, 0);
    delete  TheSequencer;  TheSequencer = NULL;
  };
  sTeleCom::OnEnterToCoolingState();
}


void sTeleCOMPort::PutTelex (sTelex  T)
{
  if( TheSequencer )
    TheSequencer->PutTelex(T);
}


void  sTeleCOMPort::OnOpeningError (QString  text)
{
  emit ErrorMessage(tr("%1").arg(text));

  //!@todo{Instruments} What should it do on errors?
  emit CoolDown();
}


void sTeleCOMPort::OnPortOK (psSerialSequencer  S)
{
  qDebug()<<"sTeleCOMPort::OnPortOK";
  TheSequencer = S;
}


//-------------------------------------------------------- sPortOpeningState ---

sPortOpeningState::~sPortOpeningState ()
{
  qDebug()<<"sPortOpeningState destructor";
  if( Sequencer ){
    disconnect(Sequencer, 0, 0, 0);
    delete  Sequencer;  Sequencer = NULL;
  };
//  ->deleteLater();
}


sPortOpeningState::sPortOpeningState (psState            parent_state
                                     ,sString            introducement
                                     ,psSerialSequencer  sequencer)
    :sState(parent_state)
    ,Sequencer(sequencer)
    ,Introducement(introducement)
{
  State_Opening = new sState(this);
  State_WaitingForIntroducement = new sState(this);
  State_Success = new sState(this);
  State_Opening
      ->SetAsInitial()
      ->OnEnterCall(Sequencer ,SLOT(Open()))
      ->AddTransition(Sequencer ,SIGNAL(OpeningError(QString)) ,Fin())
      ->AddTransition(Sequencer ,SIGNAL(Ready())
                     ,State_WaitingForIntroducement);
  State_WaitingForIntroducement
//!@todo{Instruments} Not implemented slots OnEnterTo_State_WaitingForDeviceName
//!                   and OnExitFrom_State_WaitingForDeviceName.
      ->OnEnterCall(SLOT(OnEnterTo_State_WaitingForDeviceName()))
      ->OnExitCall(SLOT(OnExitFrom_State_WaitingForDeviceName()));
}


void sPortOpeningState::CheckDeviceName(sTelex  T)
{
  if( ((QString)T) == Introducement )
    emit Success(Sequencer);
}

//------------------------------------------------------------------------------

/*
bool sTeleCOMPort::FindDeviceAndSetPortName()
{
    bool OK = false;
    int numBytes = 0;
    QString ThePortName = "";
    foreach (const SerialPortInfo &info, SerialPortInfo::availablePorts()) {
        qDebug() << "Port name: " << info.portName();
        InputString.clear();
        Port->setPort(info);
        if ((!OK) && Port->open(QIODevice::ReadWrite)){
            Port->setRate(SerialPort::Rate115200);
            Port->setDataBits(SerialPort::Data8);
            Port->write("?"+TerminalSymbols);
            qDebug() << "Sending <?> to port " << info.portName();
            while (Port->waitForReadyRead(100)){
               numBytes = Port->bytesAvailable();
               QByteArray ba = Port->read(numBytes);
               InputString.append(ba);
            }
            qDebug() << "InputString = " << InputString;
            int Pos = InputString.indexOf(TerminalSymbols);
            if (InputString.left(Pos) == Name){
                ThePortName = info.portName();
                InputString.clear();
                OK = true;
            }
            Port->close();
        }
    }
    Port->setPort(ThePortName);
    return OK;
}
*/

//------------------------------------------------------------------------------
} //namespace RA

