////////////////////////////////////////////////////////////////////////////////
/*! @file MathAn.cpp MathAn on functions.
- Part of CommonMath - Research Assistant Plugin package.
- Uses  RAGUI - Research Assistant Graphical User Interface.
- Uses  RANet - Research Assistant Net Library.
- Copyright(C) 2015, Viktor E. Bursian, St.Petersburg, Russia.
                     Viktor.Bursian@mail.ioffe.ru
*///////////////////////////////////////////////////////////////////////////////
#include "MathAn.h"
#include "TableDataFunction.h"
namespace RA {
//------------------------------------------------------------------------------

DEFINE_CLASS_TAG(sLorentzFunction)
DEFINE_CLASS_TAG(sGaussFunction)
const real sGaussFunction::WidthVsSigma = 2.0L * sqrt(2.0L*std::log(2.0L));
const real sGaussFunction::Sqrt2pi = sqrt(2.0L*sMath::pi);

//------------------------------------------------------------------------------


psMathValue  minX (pcsMathValue  function)
{
  pcsFunction                 Function(dynamic_cast<pcsFunction>(function));
  ASSERT( Function );
  return  new sPhysValue( Function->AbscissaRange().From() );
}



psMathValue  maxX (pcsMathValue  function)
{
  pcsFunction                 Function(dynamic_cast<pcsFunction>(function));
  ASSERT( Function );
  return  new sPhysValue( Function->AbscissaRange().To() );
}


psMathValue  ValueAt (pcsMathValue  position
                     ,pcsMathValue  function)
{
  pcsPhysValue                Position(dynamic_cast<pcsPhysValue>(position));
  pcsFunction                 Function(dynamic_cast<pcsFunction>(function));
  ASSERT( Position && Function );
  try{
    ((*Position)/sPhysValue(Function->AbscissaRange().Units())).ToReal();
    return new sPhysValue(Function->Value(*Position));
  }catch( rxRAlgebra ){
    return new sMathError("1st argument must have comparable units");
  }
}


psMathValue  Maximum (pcsMathValue  data_function)
{
  pcsDataFunction             Arg(dynamic_cast<pcsDataFunction>(data_function));
  ASSERT( Arg );
  sCurve::psTracer            T = NULL;
  T = Arg->GetCurveTracer(sScales(sScale(Arg->Multiplier().X)
                                 ,sScale(Arg->Multiplier().Y)));
  T->NextPoint();
  sPoint                      P = T->Point;
  while( T->NextPoint() ){
    if( T->Point.Y > P.Y )
      P = T->Point;
  }
  delete T;
  return new sPhysValue( P.Y * Arg->Multiplier().Y );
}


psMathValue  MaximumPos (pcsMathValue  data_function)
{
  pcsDataFunction             Arg(dynamic_cast<pcsDataFunction>(data_function));
  ASSERT( Arg );
  sCurve::psTracer            T = NULL;
  T = Arg->GetCurveTracer(sScales(sScale(Arg->Multiplier().X)
                                 ,sScale(Arg->Multiplier().Y)));
  T->NextPoint();
  sPoint                      P = T->Point;
  while( T->NextPoint() ){
    if( T->Point.Y > P.Y )
      P = T->Point;
  }
  delete T;
  return new sPhysValue( P.X * Arg->Multiplier().X );
}


psMathValue  RunningIntegral (pcsMathValue  data_function)
{
  pcsDataFunction             Arg(dynamic_cast<pcsDataFunction>(data_function));
  ASSERT( Arg );
  int                         N = Arg->NumberOfPoints();
  if( N <= 1 ){
    return new sMathError("data function should have more than 1 point");
  }else{
    psTableDataFunction         R = NULL;
    real                        Integral = 0.0;
    real                        DX = 0.0;
    sCurve::psTracer            T = NULL;
    int                         Sorting = 0;
    int                         PointNo = 0;
    sPoint                      Prev;
    R = new sTableDataFunction(N);
    R->SetMultiplier(sPhysPair(Arg->Multiplier().X
                              ,Arg->Multiplier().Y*Arg->Multiplier().X));
    T = Arg->GetCurveTracer(sScales(sScale(Arg->Multiplier().X)
                                   ,sScale(Arg->Multiplier().Y)));
    T->NextPoint();
    R->SetValue(PointNo ,sPoint( T->Point.X , Integral ) );
    Prev = T->Point;
    PointNo++;
    while( T->NextPoint() ){
      DX = T->Point.X - Prev.X;
      if( DX != 0 ){
        if( Sorting == 0 ){
          Sorting = ( DX > 0 ? 1 : -1 );
        }else if( Sorting*DX < 0 ){
          delete T;
          delete R;
          return new sMathError("data points are not sorted");
        }
      }
      Integral += ( T->Point.Y + Prev.Y ) / 2 * std::abs(DX);
      R->SetValue(PointNo ,sPoint( T->Point.X , Integral ) );
      Prev = T->Point;
      PointNo++;
    }
    delete T;
    return R;
  }
}


psMathValue  FullIntegral (pcsMathValue  data_function)
{
  pcsDataFunction             Arg(dynamic_cast<pcsDataFunction>(data_function));
  ASSERT( Arg );
  int                         N = Arg->NumberOfPoints();
  if( N <= 1 ){
    return new sMathError("data function should have more than 1 point");
  }else{
    real                        Integral = 0.0;
    sCurve::psTracer            T = NULL;
    T = Arg->GetCurveTracer(sScales(sScale(Arg->Multiplier().X)
                                   ,sScale(Arg->Multiplier().Y)));
    T->NextPoint();
    sPoint                      Prev = T->Point;
    while( T->NextPoint() ){
      Integral += ( T->Point.Y + Prev.Y ) / 2 * (T->Point.X - Prev.X);
      Prev = T->Point;
    }
    delete T;
    return new sPhysValue(Integral * Arg->Multiplier().X
                                   * Arg->Multiplier().Y );
  }
}

/*
psMathValue  FullIntegral (pcsMathValue  data_function)
{
  pcsDataFunction             Arg(dynamic_cast<pcsDataFunction>(data_function));
  ASSERT( Arg );
  int                         N = Arg->NumberOfPoints();
  if( N <= 1 ){
    return new sMathError("data function should have more than 1 point");
  }else{
    real                        Integral = 0.0;
    real                        DX = 0.0;
    sCurve::psTracer            T = NULL;
    T = Arg->GetCurveTracer(sScales(sScale(Arg->Multiplier().X)
                                   ,sScale(Arg->Multiplier().Y)));
    int                         Sorting = 0;
    T->NextPoint();
    sPoint                      Prev = T->Point;
    while( T->NextPoint() ){
      DX = T->Point.X - Prev.X;
      if( DX != 0 ){
        if( Sorting == 0 ){
          Sorting = ( DX > 0 ? 1 : -1 );
        }else if( Sorting*DX < 0 ){
          delete T;
          return new sMathError("data points are not sorted");
        }
      }
      Integral += ( T->Point.Y + Prev.Y ) / 2 * std::abs(DX);
      Prev = T->Point;
    }
    delete T;
    return new sPhysValue(Integral * Arg->Multiplier().X
                                   * Arg->Multiplier().Y );
  }
}
*/

psMathValue  lorentz (pcsMathValue  position
                     ,pcsMathValue  width   )
{
  pcsPhysValue                Position(dynamic_cast<pcsPhysValue>(position));
  pcsPhysValue                Width(dynamic_cast<pcsPhysValue>(width));
  ASSERT( Position && Width );
  try{
    ((*Position)/(*Width)).ToReal();
  }catch( rxRAlgebra ){
    return new sMathError("both arguments must have comparable units");
  }
  return new sLorentzFunction(*Position,*Width);
}



psMathValue  gauss (pcsMathValue  position
                   ,pcsMathValue  width   )
{
  pcsPhysValue                Position(dynamic_cast<pcsPhysValue>(position));
  pcsPhysValue                Width(dynamic_cast<pcsPhysValue>(width));
  ASSERT( Position && Width );
  try{
    ((*Position)/(*Width)).ToReal();
  }catch( rxRAlgebra ){
    return new sMathError("both arguments must have comparable units");
  }
  return new sGaussFunction(*Position,*Width);
}

//--------------------------------------------------------- sLorentzFunction ---

sLorentzFunction::sLorentzFunction (rcsPhysValue  position
                                   ,rcsPhysValue  width   )
    :sAnalyticFunction()
{
  try{
    TheMultiplier.X = sPhysValue(position.Units());
    TheMultiplier.Y = 1 / TheMultiplier.X;
    ThePosition = (position/TheMultiplier.X).ToReal();
    TheGamma = (width/TheMultiplier.X).ToReal() / 2;
  }catch( rxRAlgebra ){
    TheMultiplier.X = sPhysValue(1.0,_Unitsless_);
    TheMultiplier.Y = sPhysValue(1.0,_Unitsless_);
    ThePosition = 0.0;
    TheGamma = 0.5;
  }
}


sLorentzFunction::sLorentzFunction (std::istream &  a_stream
                                   ,rcsVersion      version )
    :sAnalyticFunction(a_stream,version)
    ,TheMultiplier(a_stream,version)
{
  a_stream.READ_INTO(ThePosition);
  a_stream.READ_INTO(TheGamma);
}


void  sLorentzFunction::Store (std::ostream & a_stream) const
{
  sAnalyticFunction::Store(a_stream);
  TheMultiplier.Store(a_stream);
  a_stream.WRITE_FROM(ThePosition);
  a_stream.WRITE_FROM(TheGamma);
}


psMathValue sLorentzFunction::Replica() const
{
  return new sLorentzFunction(Position(),Width());
}


sBoundaries sLorentzFunction::Boundaries() const
{
  return sBoundaries(sPhysRange(-real_inf ,real_inf ,TheMultiplier.X.Units())
                    ,sPhysRange(0, 1.0/sMath::pi/TheGamma
                                                    ,TheMultiplier.Y.Units()));
}


sPhysRange  sLorentzFunction::AbscissaRange () const
{
  return sPhysRange(-real_inf,real_inf,TheMultiplier.X.Units());
}


sUnits  sLorentzFunction::OrdinateUnits () const
{
  return  TheMultiplier.Y.Units();
}


sPhysValue  sLorentzFunction::Value (rcsPhysValue  x) const
{
  try{
    real   X = (x/TheMultiplier.X).ToReal();
    return ( 1.0 / sMath::pi / TheGamma / (1.0 + ((X-ThePosition)/TheGamma)
                                                *((X-ThePosition)/TheGamma))
           ) * TheMultiplier.Y;
  }catch( rxRAlgebra ){
  }
  return sPhysValue(real_nan,TheMultiplier.X.Units());
}

//--------------------------------------------------------- sGaussFunction ---

sGaussFunction::sGaussFunction (rcsPhysValue  position
                               ,rcsPhysValue  width   )
    :sAnalyticFunction()
{
  try{
    TheMultiplier.X = sPhysValue(position.Units());
    TheMultiplier.Y = 1 / TheMultiplier.X;
    ThePosition = (position/TheMultiplier.X).ToReal();
    TheSigma = (width/TheMultiplier.X).ToReal() / WidthVsSigma;
  }catch( rxRAlgebra ){
    TheMultiplier.X = sPhysValue(1.0,_Unitsless_);
    TheMultiplier.Y = sPhysValue(1.0,_Unitsless_);
    ThePosition = 0.0;
    TheSigma = 1.0 / WidthVsSigma;
  }
}


sGaussFunction::sGaussFunction (std::istream &  a_stream
                               ,rcsVersion      version )
    :sAnalyticFunction(a_stream,version)
    ,TheMultiplier(a_stream,version)
{
  a_stream.READ_INTO(ThePosition);
  a_stream.READ_INTO(TheSigma);
}


void  sGaussFunction::Store (std::ostream & a_stream) const
{
  sAnalyticFunction::Store(a_stream);
  TheMultiplier.Store(a_stream);
  a_stream.WRITE_FROM(ThePosition);
  a_stream.WRITE_FROM(TheSigma);
}


psMathValue sGaussFunction::Replica() const
{
  return new sGaussFunction(Position(),Width());
}


sBoundaries sGaussFunction::Boundaries() const
{
  return sBoundaries(sPhysRange(-real_inf ,real_inf ,TheMultiplier.X.Units())
                    ,sPhysRange(0, 1.0/Sqrt2pi/TheSigma
                                                    ,TheMultiplier.Y.Units()));
}


sPhysRange  sGaussFunction::AbscissaRange () const
{
  return sPhysRange(-real_inf,real_inf,TheMultiplier.X.Units());
}


sUnits  sGaussFunction::OrdinateUnits () const
{
  return  TheMultiplier.Y.Units();
}


sPhysValue  sGaussFunction::Value (rcsPhysValue  x) const
{
  try{
    real   X = (x/TheMultiplier.X).ToReal();
    return ( 1.0 / Sqrt2pi / TheSigma
           * exp( - (X-ThePosition)*(X-ThePosition)/2/TheSigma/TheSigma)
           ) * TheMultiplier.Y;
  }catch( rxRAlgebra ){
  }
  return sPhysValue(real_nan,TheMultiplier.X.Units());
}

//------------------------------------------------------------------------------
} //namespace RA
