////////////////////////////////////////////////////////////////////////////////
/*! @file Redigitizing.cpp Redigitizing on functions.
- Part of CommonMath - Research Assistant Plugin package.
- Uses  RAGUI - Research Assistant Graphical User Interface.
- Uses  RANet - Research Assistant Net Library.
- Copyright(C) 2015, Viktor E. Bursian, St.Petersburg, Russia.
                     Viktor.Bursian@mail.ioffe.ru
*///////////////////////////////////////////////////////////////////////////////
#include "Redigitizing.h"
#include "TableDataFunction.h"
namespace RA {
//------------------------------------------------------------------------------


psMathValue  digitize (pcsMathValue  from
                      ,pcsMathValue  to
                      ,pcsMathValue  step
                      ,pcsMathValue  function)
{
  pcsPhysValue                From(dynamic_cast<pcsPhysValue>(from));
  pcsPhysValue                To(dynamic_cast<pcsPhysValue>(to));
  pcsPhysValue                Step(dynamic_cast<pcsPhysValue>(step));
  pcsFunction                 Function(dynamic_cast<pcsFunction>(function));
  ASSERT( From && To && Step && Function );
  try{
    sPhysValue                  MX(1.0,Function->AbscissaRange().Units());
    sPhysValue                  MY(1.0,Function->OrdinateUnits());
    real                        ArbFrom = ((*From)/MX).ToReal();
    real                        ArbStep = ((*Step)/MX).ToReal();
    const int                   N  = Round(( ((*To)/MX).ToReal() - ArbFrom )
                                           / ArbStep);
    if( N < 2 )
      return new sMathError("wrong interval");
    psTableDataFunction         DF = new sTableDataFunction(sPhysPair(MX,MY)
                                                           ,N);
    real                        X;
    for( int PointNo=0 ; PointNo < DF->NumberOfPoints() ; PointNo++ ){
      X = ArbFrom + ArbStep*PointNo;
      DF->SetValue(PointNo
                  ,sPoint( X , (Function->Value(X*MX)/MY).ToReal() ) );
    }
    return DF;
  }catch( rxRAlgebra ){
    return new sMathError("arguments must have comparable units");
  }
}


psMathValue  CutDataFunction (pcsMathValue  from
                             ,pcsMathValue  to
                             ,pcsMathValue  data_function)
{
  pcsPhysValue                From(dynamic_cast<pcsPhysValue>(from));
  pcsPhysValue                To(dynamic_cast<pcsPhysValue>(to));
  pcsDataFunction             F(dynamic_cast<pcsDataFunction>(data_function));
  ASSERT( From && To && F );
  try{
    real                        ArbFrom = ((*From)/F->Multiplier().X).ToReal();
    real                        ArbTo = ((*To)/F->Multiplier().X).ToReal();
    int                         N  = 0;
    sCurve::psTracer            T = NULL;
    T = F->GetCurveTracer(sScales(sScale(F->Multiplier().X)
                                 ,sScale(F->Multiplier().Y)));
    while( T->NextPoint() ){
      if( (ArbFrom <= T->Point.X) && (T->Point.X <= ArbTo) )
        N++;
    }
    delete T;
    if( N < 2 )
      return new sMathError("wrong interval");
    psTableDataFunction         RF = new sTableDataFunction(F->Multiplier()
                                                           ,N);
    int                         RPointNo = 0;
    T = F->GetCurveTracer(sScales(sScale(F->Multiplier().X)
                                 ,sScale(F->Multiplier().Y)));
    while( T->NextPoint() ){
      if( (ArbFrom <= T->Point.X) && (T->Point.X <= ArbTo) ){
        RF->SetValue(RPointNo ,sPoint( T->Point.X , T->Point.Y ) );
        RPointNo++;
      }
    }
    delete T;
    return RF;
  }catch( rxRAlgebra ){
    return new sMathError("arguments must have comparable units");
  }
}


psMathValue  ShiftDataFunction (pcsMathValue  by
                               ,pcsMathValue  data_function)
{
  pcsPhysValue                By(dynamic_cast<pcsPhysValue>(by));
  pcsDataFunction             F(dynamic_cast<pcsDataFunction>(data_function));
  ASSERT( By && F );
  real                        DX;
  try{
    DX = ((*By)/F->Multiplier().X).ToReal();
  }catch( rxRAlgebra ){
    return new sMathError("1st argument must have comparable units"
                          " with abscissa of the 2nd");
  }
  psTableDataFunction         Result = NULL;
  sCurve::psTracer            T = NULL;
  int                         PointNo = 0;
  Result = new sTableDataFunction(F->Multiplier(),F->NumberOfPoints());
  T = F->GetCurveTracer(sScales(sScale(F->Multiplier().X)
                               ,sScale(F->Multiplier().Y)));
  while( T->NextPoint() ){
    Result->SetValue(PointNo ,sPoint( T->Point.X + DX , T->Point.Y ) );
    PointNo++;
  }
  if( T ){ delete T;  T=NULL; };
  return Result;
}


psMathValue  ReverseFunction (pcsMathValue  arg1)
{
  pcsDataFunction             Arg(dynamic_cast<pcsDataFunction>(arg1));
  ASSERT( Arg );
  psMathValue                 Result = NULL;
  psTableDataFunction         R = NULL;
  sCurve::psTracer            T = NULL;
  T = Arg->GetCurveTracer(sScales(sScale(Arg->Multiplier().X)
                                  ,sScale(Arg->Multiplier().Y)));
  R = new sTableDataFunction(sPhysPair(Arg->Multiplier().Y,Arg->Multiplier().X)
                            ,Arg->NumberOfPoints());
  int                         PointNo = 0;
  while( T->NextPoint() ){
    R->SetValue(PointNo ,sPoint( T->Point.Y , T->Point.X ) );
    PointNo++;
  }
  Result = R;
  if( T ){ delete T;  T=NULL; };
  return Result;
}


psMathValue  Fold2v (pcsMathValue  data_function)
{
  pcsDataFunction             F(dynamic_cast<pcsDataFunction>(data_function));
  ASSERT( F );
  real                        P;
  real                        X;
  try{
    P = (sPhysValue(180,"degree")/F->Multiplier().X).ToReal();
  }catch( rxRAlgebra ){
    return new sMathError("the abscissa of the argument must be in [degree]"
                          " or comparable units");
  }
  psTableDataFunction         Result = NULL;
  sCurve::psTracer            T = NULL;
  T = F->GetCurveTracer(sScales(sScale(F->Multiplier().X)
                               ,sScale(F->Multiplier().Y)));
  Result = new sTableDataFunction(F->Multiplier(),F->NumberOfPoints());
  int                         PointNo = 0;
  while( T->NextPoint() ){
    X = T->Point.X;
    X = ( X/P - floor(X/P) ) * P;
    X = rmin( X , P-X );
    Result->SetValue(PointNo ,sPoint( X , T->Point.Y ) );
    PointNo++;
  }
  if( T ){ delete T;  T=NULL; };
  return Result;
}


psMathValue  Fold4v (pcsMathValue  data_function)
{
  pcsDataFunction             F(dynamic_cast<pcsDataFunction>(data_function));
  ASSERT( F );
  real                        P;
  real                        X;
  try{
    P = (sPhysValue(90,"degree")/F->Multiplier().X).ToReal();
  }catch( rxRAlgebra ){
    return new sMathError("the abscissa of the argument must be in [degree]"
                          " or comparable units");
  }
  psTableDataFunction         Result = NULL;
  sCurve::psTracer            T = NULL;
  T = F->GetCurveTracer(sScales(sScale(F->Multiplier().X)
                               ,sScale(F->Multiplier().Y)));
  Result = new sTableDataFunction(F->Multiplier(),F->NumberOfPoints());
  int                         PointNo = 0;
  while( T->NextPoint() ){
    X = T->Point.X;
    X = ( X/P - floor(X/P) ) * P;
    X = rmin( X , P-X );
    Result->SetValue(PointNo ,sPoint( X , T->Point.Y ) );
    PointNo++;
  }
  if( T ){ delete T;  T=NULL; };
  return Result;
}

//------------------------------------------------------------------------------
} //namespace RA
