////////////////////////////////////////////////////////////////////////////////
/*! @file DummyTunableCCDDriver.cpp Имитатор драйвера ПЗС-линейки с перестройкой длины волны.
- Part of DemoDevices - Research Assistant Plugin package.
- Uses  RAGUI - Research Assistant Graphical User Interface Library.
- Uses  Qt v.5.0.2  - http://qt-project.org/
- Uses  RANet - Research Assistant Net Library.
- Copyright(C) 2017, Dmitrii Nelson, St.Petersburg, Russia.
                          d.nelson@mail.ioffe.ru
*///////////////////////////////////////////////////////////////////////////////
#include "DummyTunableCCDDriver.h"
#include "attr.h"
#include <QDebug>
#include "Log.h"
#include <cstdlib>

namespace RA {
//------------------------------------------------------------------------------

DEFINE_CLASS_TAG(sDummyTunableCCDDriver)

//-------------------------------------------- sDummyTunableCCDDriver ----------

void  sDummyTunableCCDDriver::Preconfigure (sNodePtr  DeviceCfg)
{
  sPtr<sFolderNode>              DeviceSettings;
  DeviceCfg |= "Settings";
  DeviceSettings = DeviceCfg >> "Settings";
  if( DeviceSettings.IsCorrect() ){
    DeviceSettings |= attr("NumberOfPixels",3648);
    DeviceSettings |= attr("MiddlePixel",1733);
    DeviceSettings |= attr("WL_0",sPhysValue(0.02389,"nm"));
    DeviceSettings |= attr("WL_k",sPhysValue(-7.8e-6,_Unitsless_));
  };
  sTunableCCDDriver::Preconfigure(DeviceCfg);
}

sDummyTunableCCDDriver::~sDummyTunableCCDDriver ()
{
}

sDummyTunableCCDDriver::sDummyTunableCCDDriver (sString   DevicetId
                               ,sNodePtr  DeviceCfg)
  :sTunableCCDDriver(DevicetId,DeviceCfg)
  ,Amplitude(50000)
  ,WL_0(sPhysValue(1.0,"nm"))
  ,WL_k(sPhysValue(1,_Unitsless_))
{
}

void  sDummyTunableCCDDriver::PostConstructor ()
{
  sTunableCCDDriver::PostConstructor();
  try{ //get config parameters
    get_from<sPhysValueNode>(DeviceSettings >> "WL_0" ,WL_0);
    get_from<sPhysValueNode>(DeviceSettings >> "WL_k" ,WL_k);
  }catch(...){
    SetInvalid(tr("Something wrong in configuration."));
  }
}

void  sDummyTunableCCDDriver::ConnectUSBDevice ()
{
  emit Connected();
}

void  sDummyTunableCCDDriver::GetCCD_Data ()
{
    ushort                    NoiseMagnitude = 320;
    Amplitude = Amplitude - 1000;
    if (Amplitude < 2000) Amplitude = 50000;
    ushort Width = 1000;
    for( int i = 0 ; i < NumOfPixels ; i ++ ){
        DataArray.ArrBody[i] =
            Amplitude/(1+(i-NumOfPixels/2)*(i-NumOfPixels/2)/Width)
            +Round(std::rand()*real(NoiseMagnitude)/RAND_MAX);
    }
    emit NewData(DataArray);
}

void  sDummyTunableCCDDriver::CalculateWavelengths ()
{
  sPhysValue dWL = WL_0 + WL_k * WLPosition;
  sPhysValue LeftPos = WLPosition - dWL*MiddlePixel;
  sPhysValue RightPos = WLPosition + dWL*((NumOfPixels -1) - MiddlePixel);
  real dX_0 = ((WL_0 + WL_k * LeftPos)/(sPhysValue(1.0,"nm"))).ToReal();
  real dX_1 = ((WL_0 + WL_k * RightPos)/(sPhysValue(1.0,"nm"))).ToReal();
  real dX = (dWL/(sPhysValue(1.0,"nm"))).ToReal(); // in "nm/pixel"
  real d2X = (dX_1 - dX_0)/NumOfPixels;
  real  XM = (WLPosition/(sPhysValue(1.0,"nm"))).ToReal(); // in "nm"
  for (int i = 0; i < NumOfPixels; i++) {
    Wavelengths.ArrBody[i] = XM + (dX + d2X *(i - MiddlePixel))*(i - MiddlePixel);
  }
  emit NewWavelengths (Wavelengths);
}

void  sDummyTunableCCDDriver::StartCCD ()
{
  try{
    int _IT = Round((Parameters.IntegrationTime/(sPhysValue(1.0e-3,"s"))).ToReal());
    if (_IT < 1) _IT = 1;
    State_O_Measuring->SetTimeout(_IT);
  }catch(rxRAlgebra){
    SetInvalid(tr("Something wrong with units."));
  }
}

void  sDummyTunableCCDDriver::StopCCD ()
{
  WLDriver.ReleaseExclusiveControl();
}

void  sDummyTunableCCDDriver::SetCCD_Parameters (sCCDDriver::sParameters Params)
{
  if( ! WLDriver.TakeExclusiveControl() ){
    emit IsBusy();
  }else{
    Parameters =  Params;
    emit ParamsSet();
  }
}

//------------------------------------------------------------------------------
} //namespace RA
