////////////////////////////////////////////////////////////////////////////////
/*! @file RadiofictivityGauge.cpp  Эмулятор.
- Part of DemoDevices - Research Assistant Plugin package.
- Uses  RAGUI - Research Assistant Graphical User Interface Library.
- Uses  Qt v.5.0.2  - http://qt-project.org/
- Uses  RANet - Research Assistant Net Library.
- Copyright(C) 2010-2014, Viktor E. Bursian, St.Petersburg, Russia.
                          Viktor.Bursian@mail.ioffe.ru
*///////////////////////////////////////////////////////////////////////////////
#include "RadiofictivityGauge.h"
#include "attr.h"
#include "StringsAndQStrings.h"
#include "PhysValue.h"
#include "Log.h"
#include <cstdlib>
namespace RA {
//------------------------------------------------------------------------------

DEFINE_CLASS_TAG(sRadiofictivityGauge)

/*! Директива, вынуждающая датчик однократно выдать ложный отсчёт
    (attached data: значение ложного отсчёта).
*/
const sDirective<sPhysValue>  sRadiofictivityGauge::cCheat("cCheat");

/*! Директива, вынуждающая датчик выдать сигнал Say(QString) и остановиться
    (attached data: what to say).
*/
const sDirective<QString>     sRadiofictivityGauge::cSay("cSay");

//----------------------------------------------------- sRadiofictivityGauge ---

void  sRadiofictivityGauge::Preconfigure (sNodePtr  DeviceCfg)
{
  sPtr<sFolderNode>              DeviceSettings;
  DeviceCfg |= "Settings";
  DeviceSettings = DeviceCfg >> "Settings";
  if( DeviceSettings.IsCorrect() ){
    DeviceSettings |= attr("CurrentActivity",sPhysValue()
                                            =sPhysValue(1e4,"photon/s"));
    DeviceSettings |= attr("DecayTau",sPhysValue(30.0,"s"));
    // изменение умолчаний для sGauge:
    DeviceSettings |= attr("Period",sPhysValue(0.1,"s"));
  }
  sGauge::Preconfigure(DeviceCfg);
}


sRadiofictivityGauge::~sRadiofictivityGauge ()
{
  if( DeviceSettings.IsCorrect() ){
    // Не все конфигурационные параметры следует сохранять после работы.
    // В данном случае, я хочу, чтобы "сохранялось" только это:
    DeviceSettings << attr("CurrentActivity",CurrentActivity*100);
  }
}


sRadiofictivityGauge::sRadiofictivityGauge (sString   DevicetId
                                           ,sNodePtr  DeviceCfg)
    :sGauge(DevicetId,DeviceCfg)
    ,Clock(new QTimer(this))
    ,Interval(100)
    ,CurrentActivity()
    ,DecayTau("s")
    ,noise(5e-2)
{
}


void  sRadiofictivityGauge::PostConstructor ()
{
  sGauge::PostConstructor();
  try{ //get config parameters
    get_from<sPhysValueNode>(DeviceSettings >> "CurrentActivity"
                                               ,CurrentActivity);
    get_from<sPhysValueNode>(DeviceSettings >> "DecayTau"
                                               ,DecayTau);
  }catch(...){
    SetInvalid(tr("Something wrong in configuration."));
  }
  // evolution of the inherited state machine
  {
    // evolution of the State_Warming
    {
      State_W_Pending = new sState(State_Warming());
      State_W_Check = new sState(State_Warming());
      State_W_Pending
          ->SetAsInitial()
          ->OnEnterCall( SLOT(SetUpThePendingTimer()) )
          ->AddTransition(Clock ,SIGNAL(timeout()) ,State_W_Check);
      State_W_Check
          ->OnEnterCall( SLOT(CheckMood()) )
          ->AddTransition( SIGNAL(NotInTheMood()) ,State_Frozen() )
          ->AddTransition( SIGNAL(OK()) ,State_Warming()->Fin() );
    }
    // evolution of the State_Operating
    {
      State_O_M_Acquiring
          ->OnEnterCall(SLOT(RunTheClock()))
          ->AddTransition(SIGNAL(ClockSignal()),SLOT(ExtractData())
                         ,State_O_M_Cycling)
          ->AddTransition(cCheat ,SIGNAL(NewData(sPhysValue))
                         ,State_O_M_Cycling)
          ->AddTransition(cSay ,SIGNAL(Say(QString))
                         ,State_O_M_Stopped)
          ->OnExitCall(Clock ,SLOT(stop()))
          ;
      connect(Clock,SIGNAL(timeout()),this,SIGNAL(ClockSignal()));
    }
    //evolution of the State_Cooling
    {
      State_C_Pausing = new sState(State_Cooling());
      State_C_Pausing
          ->SetAsInitial()
          ->OnEnterCall(SLOT(RunCoolingClock()))
          ->AddTransition( SIGNAL(CoolingClockAlarm()) ,State_Cooling()->Fin());
    }
  }
}


void  sRadiofictivityGauge::SetWantedPeriod ()
{
  try{
    Interval = 100 * Round( (WantedPeriod / sPhysValue(100.0,"ms")).ToReal() );
    if( Interval < 100 )  Interval=100;
    ThePeriod = sPhysValue(Interval,0.5,"ms");
    emit PeriodChanged(ThePeriod);
  }catch(rxRAlgebra){
    //!@todo{Instruments} define Failure signal for drivers?
    emit ErrorMessage(tr("Cannot set period, period=%1")
                     .arg(ToQString(WantedPeriod.Text())));
    emit CoolDown();
  }
}


void sRadiofictivityGauge::SetUpThePendingTimer ()
{
  Clock->setSingleShot(true);
  Clock->start(5000);
}


void sRadiofictivityGauge::CheckMood ()
{
  int   Mood = std::rand(); //бросаем монетку
  if( i_am_lasy() && (Mood < 16000) ){  //я ленивый и сейчас не в духах
    emit ErrorMessage(tr("I am lasy and not in the mood.  Bye-bye!"));
    emit NotInTheMood();
  }else{                             //ладно, поработаем
    WantedPeriod = ThePeriod;
    SetWantedPeriod();
    emit OK();
  }
}


void sRadiofictivityGauge::RunTheClock()
{
  Clock->setSingleShot(true);
  Clock->start(Interval);
}


void  sRadiofictivityGauge::ExtractData ()
{
  sPhysValue                  MeasuredData;
  MeasuredData = CurrentActivity;
  MeasuredData *= 1.0 + (noise/RAND_MAX)*std::rand();
  emit NewData(MeasuredData);
  CurrentActivity -= CurrentActivity*Period()/DecayTau;
  CurrentActivity.SetError(sPhysValue(0,CurrentActivity.Units()));
  if( shit_happens() ){
    Clock->stop();   //типа "контакт отвалился"
  }
}


bool  sRadiofictivityGauge::i_am_lasy ()
{
  return  DeviceIdentifier().Pos("broken") < DeviceIdentifier().Length();
}


bool  sRadiofictivityGauge::shit_happens ()
{
  if( i_am_lasy() ){
    int   PieceOfLuck = std::rand() % 1000;
    return  PieceOfLuck==0;
  }else{
    return false;
  }
}


void  sRadiofictivityGauge::RunCoolingClock ()
{
  QTimer::singleShot(4000,this,SIGNAL(CoolingClockAlarm()));
}

//------------------------------------------------------------------------------
} //namespace RA
