////////////////////////////////////////////////////////////////////////////////
/*! @file StepperFantastic317.cpp  Эмулятор устойства управления развёрткой
                                   спектрометра с шаговым двигателем.
- Part of DemoDevices - Research Assistant Plugin package.
- Uses  RAGUI - Research Assistant Graphical User Interface.
- Uses  QtGui v.4.6  -  http://qt.nokia.com/
- Uses  RANet - Research Assistant Net Library (based on ANSI C++).
- Copyright(C) 2010, Viktor E. Bursian, St.Petersburg, Russia.
                     Viktor.Bursian@mail.ioffe.ru
*///////////////////////////////////////////////////////////////////////////////
#include "StepperFantastic317.h"
#include "attr.h"
#include <QApplication>
namespace RA {
//------------------------------------------------------------------------------

DEFINE_CLASS_TAG(sStepperFantastic317)

const int      sStepperFantastic317::MaxShift( 100 );

//----------------------------------------------------- sStepperFantastic317 ---

void  sStepperFantastic317::Preconfigure (sNodePtr  DeviceCfg)
{
  sPtr<sFolderNode>              DeviceSettings;
  DeviceCfg |= "Settings";
  DeviceSettings = DeviceCfg >> "Settings";
  if( DeviceSettings.IsCorrect() ){
    DeviceSettings |= attr("Range",sPhysRange()
                                  =sPhysRange(360.0,780.0,"nm",0.1));
    DeviceSettings |= attr("StepTimeQuantum",sPhysValue()=sPhysValue(0.2,"s"));
    DeviceSettings |= attr("Step",sPhysValue()=sPhysValue(0.1,0.05,"nm"));
  }
  sStepper::Preconfigure(DeviceCfg);
}


sStepperFantastic317::~sStepperFantastic317 ()
{
  if( DeviceSettings.IsCorrect() ){
    DeviceSettings << attr("CurrentPositionIndex",CurrentIndex);
  }
}


sStepperFantastic317::sStepperFantastic317 (sString   DevicetId
                                           ,sNodePtr  DeviceCfg)
    :sStepper(DevicetId,DeviceCfg)
    ,StepTimeQuantum(sPhysValue(0.3,"s"))
    ,CurrentIndex(0)
    ,Zhzhzh(this)
    ,Schelk(this)
{
  Step = sPhysValue(0.1,0.05,"nm");
  Zhzhzh.setSource(QUrl::fromLocalFile(qApp->applicationDirPath()
                                      +"/../extrafiles/DemoDevices/zh.wav"));
//  Zhzhzh.setLoopCount(QSoundEffect::Infinite); - не работает почему-то
  Schelk.setSource(QUrl::fromLocalFile(qApp->applicationDirPath()
                                    +"/../extrafiles/DemoDevices/ClickOn.wav"));
//  Zhzhzh.setVolume(0.25);
//  Schelk.setVolume(0.25f);
}


void  sStepperFantastic317::PostConstructor ()
{
  sStepper::PostConstructor();
  try{ //get config parameters
    get_from<sPhysValueNode>(DeviceSettings >> "Step",Step);
    get_from<sIntegerNode>(DeviceSettings >> "CurrentPositionIndex"
                                                     ,CurrentIndex);
    get_from<sPhysValueNode>(DeviceSettings >> "StepTimeQuantum"
                                                     ,StepTimeQuantum);
    {
      sPhysValue  Pos = Position(CurrentIndex);
      if( Pos < Range().From() ){
        Pos = Range().From();
        CurrentIndex = Steps(Pos);
      }
      if( Range().To() < Pos ){
        Pos = Range().To();
        CurrentIndex = Steps(Pos);
      }
    }
    {
      int Quantum = Round((StepTimeQuantum/sPhysValue(100,"ms")).ToReal());
      if( Quantum < 5 ) Quantum = 5;
      StepTimeQuantum = sPhysValue(Quantum*100,"ms");
    }
  }catch(...){
    SetInvalid(tr("Something wrong in configuration."));
  }

  //evolution of the inherited state machine (behaviour)
  {
    //evolution of the State_Operating
    {
      State_O_Parallel = new sParallelStateGroup(State_Operating());
      State_O_Parallel->SetAsInitial();

      //State_O_Parallel internal structure and behaviour
      {
        State_OP_Measuring = new sState(State_O_Parallel);
        State_OP_Control = new sState(State_O_Parallel);

        //State_OP_Measuring internal structure and behaviour
        {
          State_OPM_Hanging = new sState(State_OP_Measuring,ThePeriod);
          State_OPM_Hanging
              ->SetAsInitial()
              ->AddTransition(State_OPM_Hanging,SIGNAL(Timeout())
                             ,State_OPM_Hanging)
              ->OnExitCall(SLOT(ReportPosition()));
        }

        //State_OP_Control internal structure and behaviour
        {
          State_OPC_Idle = new sState(State_OP_Control);
          State_OPC_SettingTo = new sState(State_OP_Control);
          State_OPC_MovingBy = new sState(State_OP_Control);

          State_OPC_Idle
              ->SetAsInitial()
              ->AddTransition(cSetTo ,SLOT(SetDestination(sPhysValue))
                             ,State_OPC_SettingTo)
              ->AddReaction(cSetTimePerStep
                           ,SLOT(SetScanTimePerStep(sPhysValue)) )
              ->AddTransition(cMoveBy
                             ,SLOT(SetNumberOfStepsToMoveBy(int))
                             ,State_OPC_MovingBy)
              ;
          State_OPC_SettingTo
              ->AddTransition(cBreak ,SLOT(SayDone()) ,State_OPC_Idle)
              ->AddTransition(State_OPC_Idle)
              ->OnFinishCall(SLOT(SayDone()))
              ;
          State_OPC_MovingBy
              ->AddTransition(cBreak ,SLOT(SayDone()) ,State_OPC_Idle)
              ->AddTransition(State_OPC_Idle)
              ->OnFinishCall(SLOT(SayDone()))
              ;
          //State_OPC_SettingTo internal structure and behaviour
          {
            State_OPC_SettingTo
                ->OnExitCall(&Zhzhzh,SLOT(stop()))
                ;
            State_OPC_SettingTo_Check = new sState(State_OPC_SettingTo);
            State_OPC_SettingTo_Zh = new sState(State_OPC_SettingTo,500);

            State_OPC_SettingTo_Check
                ->SetAsInitial()
                ->OnEnterCall(SLOT(CalculateShift()))
                ->AddTransition(SIGNAL(SettingDone())
                               ,State_OPC_SettingTo->Fin())
                ->AddTransition(State_OPC_SettingTo_Zh)
                ;
            State_OPC_SettingTo_Zh
                ->OnEnterCall(&Zhzhzh,SLOT(play()))
                ->AddTransition(State_OPC_SettingTo_Zh,SIGNAL(Timeout())
                               ,State_OPC_SettingTo_Check)
                ->OnExitCall(SLOT(ShiftPosition()))
                ;
          }
          //State_OPC_ScanningTo internal structure and behaviour
          {
            State_OPC_MovingBy_MakeStep = new sState(State_OPC_MovingBy);

            State_OPC_MovingBy_MakeStep
                ->SetAsInitial()
                ->OnEnterCall(SLOT(DoStep()))
                ->AddTransition(State_OPC_MovingBy_MakeStep,SIGNAL(Timeout())
                               ,State_OPC_MovingBy_MakeStep)
                ->AddTransition(SIGNAL(MovingDone())
                               ,State_OPC_MovingBy->Fin())
                ;
          }
        }
      }
    }
  }
}


sPhysValue  sStepperFantastic317::Position (int  steps) const
{
  sPhysValue                  P = Step*steps;
  P.SetError(Step/2);
  return P;
}


int  sStepperFantastic317::Steps (sPhysValue  position) const
{
  return Round( ( position / Step ).ToReal() );
}


void  sStepperFantastic317::SetWantedPeriod ()
{
  int                         P;
  P = Round( (WantedPeriod/sPhysValue(1,"ms")).ToReal() );
  ThePeriod = P * sPhysValue(1,"ms");
  State_OPM_Hanging->SetTimeout(P);
}


void  sStepperFantastic317::ReportPosition ()
{
  emit NewData(Position(CurrentIndex));
}


void  sStepperFantastic317::SetDestination (sPhysValue  destination)
{
  Destination = Steps(destination);
}


void  sStepperFantastic317::SayDone ()
{
  emit Done(Position(CurrentIndex));
}


void  sStepperFantastic317::SayZh ()
{
  Zhzhzh.play();
}


void  sStepperFantastic317::CalculateShift ()
{
  Shift = Destination - CurrentIndex;
  if( Shift == 0 ){
    emit SettingDone();
  }else if( Shift > MaxShift ){
    Shift = MaxShift;
  }else if( Shift < -MaxShift ){
    Shift = -MaxShift;
  }
}


void  sStepperFantastic317::ShiftPosition ()
{
  CurrentIndex += Shift;
}


void  sStepperFantastic317::SetScanTimePerStep (sPhysValue  time_per_step)
{
  int Quanta = Round((time_per_step/StepTimeQuantum).ToReal());
  if( Quanta < 1 ) Quanta = 1;
  StepTime = Quanta * StepTimeQuantum;
  emit TimePerStepSet(StepTime);
}


void  sStepperFantastic317::SetNumberOfStepsToMoveBy (int  num_of_steps)
{
  Destination = CurrentIndex + num_of_steps;
  State_OPC_MovingBy_MakeStep->SetTimeout(StepTime);
}


void sStepperFantastic317::DoStep()
{
  if( CurrentIndex == Destination ){
    emit MovingDone();
  }else{
    CurrentIndex  += ( Destination > CurrentIndex ? 1 : -1);
    Schelk.play();
  }
}

//------------------------------------------------------------------------------
} //namespace RA
