////////////////////////////////////////////////////////////////////////////////
/*! @file MSDN2041.c  Firmware for Electronic Hart - a demo firmware for
                      Microsin AVR-USB162 board without extensions.
- Part of the MSDN firmware.
- Part of the Research Assistant project.
- Uses  LUFA library written by
   - Copyright(C) 2012, Dean Camera, dean [at] fourwalledcubicle [dot] com,
                                     www.lufa-lib.org.

- Copyright(C) 2014, Dmitry K. Nelson, St.Petersburg, Russia.
                     D.Nelson@mail.ioffe.ru
- Copyright(C) 2016, Viktor E. Bursian, St.Petersburg, Russia.
                     Viktor.Bursian@mail.ioffe.ru
*///////////////////////////////////////////////////////////////////////////////
#include "Microsin_USB_IO.h"
#include "CommandParser.h"
#include "ClocksSimple.h"
#include <avr/wdt.h>
//#include <avr/power.h>
#include <avr/iousb162.h>
#include <avr/interrupt.h>
#include <stdlib.h>
//#include <avr/io.h>
//#include <stdio.h>
//#include <string.h>
//! @todo{MSDN} Неясная роль inttypes здесь
#include <inttypes.h>
//! @todo{MSDN} Неясная роль math
#include <math.h>
//#include <util/delay.h>
//------------------------------------------------------------------------------
#define CHECK_PERIOD 156 //!< Периодичность вывода в тактах таймера 0 (~10ms)

const uint16_t                MeasuringPeriod = 10000; //!< in microseconds

uint8_t                       PulseFrequency = 60; //!< in 1/min
//uint8_t                       Noise =           5; //!< in %
uint8_t                       Health =        100; //!< in %,
                                                   //!< 100% - never fails,
                                                   //!<   0% - never starts.

int8_t                        State = 0; //!< -1 - not being measured,
                                         //!<  0 - dead,
                                         //!<  1 - alive,
                                         //!<  2 - just shocked.
int32_t                       PulsePeriod; //!< in microseconds
int32_t                       Time = 0;    //!< in microseconds
int8_t                        GotTimerSignal = 0;

const int32_t                 ShockAmplitude = 10000;
const int32_t                 ShockWidth =      3000; //in microseconds
const int32_t                 PeakAmplitude =   1000;
const int32_t                 SubAmplitude =     200;

static int32_t                PeakShift;  //in microseconds
static int32_t                SubShift;   //in microseconds
static int32_t                PeakWidth;  //in microseconds
static int32_t                SubWidth;   //in microseconds


void  CalculatePeriod (void)
{
  const uint32_t              Minute = 60000000;
  PulsePeriod = Minute / PulseFrequency;
  SubShift = PulsePeriod / 2;
  SubWidth = PulsePeriod / 10;
  PeakWidth = SubWidth / 10;
  PeakShift = SubShift - SubWidth/2;
}


int8_t  ExecuteCommand0 (const char  command)
{
  switch( command ){
	case 's':
        State = 1;
        Time = 0;
        GotTimerSignal = 0;
        Clock0_SetToZero();
        Clock0_AllowInterrupts();
        break;
	case '/':
        State = 0;
        GotTimerSignal = 0;
        Clock0_ForbidInterrupts();
        break;
	case 'e':
        State = 2;
        Time = 0;
        GotTimerSignal = 0;
        Clock0_SetToZero();
        Clock0_AllowInterrupts();
        break;
	case '+':
        PulseFrequency+=5;
        CalculatePeriod();
        break;
	case '-':
        PulseFrequency-=5;
        CalculatePeriod();
        break;
	default:
        return 1;
  }
  return 0;
}


int8_t  ExecuteCommand2 (const char     command
                        ,const uint8_t  argument)
{
  switch( command ){
	case 'f':
        PulseFrequency = argument;
        CalculatePeriod();
        SendStringToUSB("pulse=");
        SendNumberToUSB(PulseFrequency);
        Send_EOL_ToUSB();
        SendStringToUSB("period=");
        SendNumberToUSB(PulsePeriod);
        Send_EOL_ToUSB();
        break;
	case 'H':
        Health = argument;
        break;
	default:
        return 1;
  }
  return 0;
}


int8_t  ExecuteCommand4 (const char      command
                        ,const uint16_t  argument)
{
  return 1;
}


int8_t  ExecuteCommand8 (const char      command
                        ,const uint32_t  argument)
{
  return 1;
}

void  SetupHardware (void)
{
  /* Disable watchdog if enabled by bootloader/fuses */
  MCUSR &= ~(1 << WDRF);
  wdt_disable();

  Clock0_Init();
  CalculatePeriod();
}


void  BeforeLoop (void)
{
}


void  InLoop (void)
{
  if( GotTimerSignal ){
    GotTimerSignal = 0;
    int32_t                     V = 0;
    int32_t                     W;
    if( State == 2 ){
      if( Time >= PulsePeriod - PeakShift ){
        if( rand() < RAND_MAX / 3 ){ //getting alive with 1/3 probability
          Time = 0;
          State = 1;
        }else{
          State = 0;
        }
      }else{
        W = ShockWidth >> 6;
        W *= W;
        V = Time >> 6;
        V *= V;
        V += W;
        V = (ShockAmplitude * W)  / V;
      }
    }
    if( State == 1 ){
      int32_t                     V1;
      int32_t                     V2;
      W = PeakWidth >> 6;
      W *= W;
      V1 = (Time-PeakShift) >> 6;
      V1 *= V1;
      V1 += W;
      V1 = (PeakAmplitude * W)  / V1;
      W = SubWidth >> 6;
      W *= W;
      V2 = (Time-SubShift) >> 6;
      V2 *= V2;
      V2 += W;
      V2 = (SubAmplitude * W)  / V2;
      V = V1 - V2;
    }
    SendCharToUSB('V');
    SendNumberToUSB(V);
    Send_EOL_ToUSB();
    Time += MeasuringPeriod;
    if( Time >= PulsePeriod ){
      Time = 0;
      if( rand() < ( ( ( ((int32_t)100) - Health ) * RAND_MAX) / 100 )
                                                          / PulseFrequency ){
        State = 0; //dying with (1-Health/100)/minute probability
      }
    }
  }
}


//------------------------------------------------------------------------------
/*! Обработчик прерывания от таймера 0 (COMPA).
*/
ISR(TIMER0_COMPA_vect){
  sei();                  // Дадим возможность обработать прерывания от USB...
  GotTimerSignal = 1;
}
