////////////////////////////////////////////////////////////////////////////////
/*! @file GaugeTestboard.cpp A testboard for any gauge (sGauge inheritor).
- Part of DevelopmentTools - Research Assistant Plugin package.
- Uses  RAGUI - Research Assistant Graphical User Interface Library.
- Uses  Qt v.5.0.2  - http://qt-project.org/
- Uses  RANet - Research Assistant Net Library.
- Copyright(C) 2012-2014, Viktor E. Bursian, St.Petersburg, Russia.
                          Viktor.Bursian@mail.ioffe.ru
*///////////////////////////////////////////////////////////////////////////////
#include "GaugeTestboard.h"
#include "attr.h"
#include "StringsAndQStrings.h"
#include <QHBoxLayout>
#include <QPushButton>
#include <QMessageBox>
namespace RA {
//------------------------------------------------------------------------------

DEFINE_CLASS_TAG(sGaugeTestboard)

//---------------------------------------------------------- sGaugeTestboard ---

void  sGaugeTestboard::Preconfigure (sNodePtr  DeviceCfg)
{
  sGaugeDisplay::Preconfigure(DeviceCfg);
}


sGaugeTestboard::sGaugeTestboard (sString   DeviceId
                                 ,sNodePtr  DeviceCfg
                                 ,sNodePtr  WidgetCfg
                                 ,QWidget *  parent    )
    :sGaugeDisplay(DeviceId,DeviceCfg,WidgetCfg,parent)
{
  //evolution of the inherited state machine (behaviour)
  {
    //State_Operating internal structure and behaviour
    {
      State_O_Watching = new sState(State_Operating());
      State_O_Controling = new sState(State_Operating());
      
      State_O_Watching
          ->SetAsInitial()
          ->AddTransition(SIGNAL(ControlTaken()) ,State_O_Controling)
          ->AddReaction(SIGNAL(StartExclusive()) ,SLOT(OnStartExclusive()) );

      State_O_Controling
          ->AddTransition(SIGNAL(ControlReleased()),State_O_Watching);
    }
    //evolution of State_Operating (behaviour)
    {
      State_Operating()
          ->AddReaction( SIGNAL(Start()) ,SLOT(OnStart()) )
          ->AddReaction( SIGNAL(Stop())  ,SLOT(OnStop())  )
            //The following won't work in State_O_Watching (because there is 
            //another reaction for the same signal), otherwise it will:
          ->AddReaction( SIGNAL(StartExclusive()) ,SLOT(BerateTheUser())   );
    }
  }
}


void  sGaugeTestboard::AppearanceMakeUp ()
{
  sGaugeDisplay::AppearanceMakeUp();

  QHBoxLayout *             ParamLayout;
  QHBoxLayout *             ButtonLayout;
  QPushButton *             Button;
  QLabel *                  Label;

  (ParamLayout = new QHBoxLayout)->setObjectName("ParamLayout");
  ParamLayout->setSpacing(10);
  ParamLayout->setMargin(0);
  ((QBoxLayout*)(layout()))->insertLayout(0,ParamLayout);

  PeriodField = new sPhysValueField(this,"Period",sPhysValue("s"),true);
  ParamLayout->addWidget(PeriodField);
  PeriodField->setVisible(false);
  State_Operating()->OnEnterCall(PeriodField,SLOT(show()))
                   ->OnExitCall(PeriodField,SLOT(hide()));

  AutoRepeatCheckBox = new QCheckBox("autorepeat",this);
  AutoRepeatCheckBox->setObjectName("AutoRepeatCheckBox");
  ParamLayout->addWidget(AutoRepeatCheckBox);
  AutoRepeatCheckBox->setVisible(false);
//  AutoRepeatCheckBox->setTristate(true);
  State_Operating()->OnEnterCall(AutoRepeatCheckBox,SLOT(show()))
                   ->OnExitCall(AutoRepeatCheckBox,SLOT(hide()));

  ParamLayout->addStretch(100);

  (ButtonLayout = new QHBoxLayout)->setObjectName("ButtonLayout3");
  ButtonLayout->setSpacing(0);
  ButtonLayout->setMargin(0);
  ((QBoxLayout*)(layout()))->insertLayout(0,ButtonLayout);

  Button = new QPushButton("START (exclusive)",this);
  ButtonLayout->addWidget(Button);
  Button->setVisible(false);
  connect(Button ,SIGNAL(clicked()) ,this ,SIGNAL(StartExclusive()));
  State_Operating()->OnEnterCall(Button,SLOT(show()))
                   ->OnExitCall(Button,SLOT(hide()));

  (ButtonLayout = new QHBoxLayout)->setObjectName("ButtonLayout2");
  ButtonLayout->setSpacing(0);
  ButtonLayout->setMargin(0);
  ((QBoxLayout*)(layout()))->insertLayout(0,ButtonLayout);

  Button = new QPushButton("START",this);
  ButtonLayout->addWidget(Button);
  Button->setVisible(false);
  connect(Button ,SIGNAL(clicked()) ,this ,SIGNAL(Start()));
  State_Operating()->OnEnterCall(Button,SLOT(show()))
                   ->OnExitCall(Button,SLOT(hide()));

  Button = new QPushButton("STOP",this);
  ButtonLayout->addWidget(Button);
  Button->setVisible(false);
  connect(Button ,SIGNAL(clicked()) ,this ,SIGNAL(Stop()));
  State_Operating()->OnEnterCall(Button,SLOT(show()))
                   ->OnExitCall(Button,SLOT(hide()));

  ButtonLayout = new QHBoxLayout;
  ButtonLayout->setObjectName("ButtonLayout1");
  ButtonLayout->setSpacing(0);
  ButtonLayout->setMargin(0);
  ((QBoxLayout*)(layout()))->insertLayout(0,ButtonLayout);

  Button = new QPushButton("Turn ON",this);
  ButtonLayout->addWidget(Button);
  Button->setVisible(false);
  connect(Button ,SIGNAL(clicked()) ,this ,SIGNAL(WarmUp()));
  State_Frozen()->OnEnterCall(Button,SLOT(show()))
                ->OnExitCall(Button,SLOT(hide()));

  Button = new QPushButton("Turn OFF",this);
  ButtonLayout->addWidget(Button);
  Button->setVisible(false);
  connect(Button ,SIGNAL(clicked()) ,this ,SIGNAL(CoolDown()));
  State_Operating()->OnEnterCall(Button,SLOT(show()))
                   ->OnExitCall(Button,SLOT(hide()));

  Label = new QLabel(this);
  layout()->addWidget(Label);
  connect(this  ,SIGNAL(NewUpperLevelState(QString))
         ,Label ,SLOT(setText(QString)) );

  Label = new QLabel("gauge is exclusively controlled",this);
  layout()->addWidget(Label);
  Label->setVisible(false);
  State_O_Controling->OnEnterCall(Label ,SLOT(show()) )
      ->OnExitCall(Label ,SLOT(hide()) );
}


void  sGaugeTestboard::OnEnterToOperatingState ()
{
  sGaugeDisplay::OnEnterToOperatingState();
  PeriodField->SetValue(Gauge->Period());
  Connect(Gauge,SIGNAL(PeriodChanged(sPhysValue))
               ,SLOT(OnPeriodChanged(sPhysValue)));
  connect(PeriodField ,SIGNAL(Edited()) ,this ,SLOT(OnPeriodFieldEdited()));
  AutoRepeatCheckBox->setChecked(Gauge->AutoRepeat());
  Connect(Gauge,SIGNAL(AutoRepeatModeChanged(bool))
               ,SLOT(OnAutoRepeatModeChanged(bool)));
  connect(AutoRepeatCheckBox ,SIGNAL(toggled(bool))
         ,this ,SLOT(OnAutoRepeatCheckBoxEdited(bool)));
}


void  sGaugeTestboard::OnExitFromOperatingState ()
{
  disconnect(AutoRepeatCheckBox ,SIGNAL(toggled(bool))
            ,this ,SLOT(OnAutoRepeatCheckBoxEdited(bool)));
  Disconnect(Gauge,SIGNAL(AutoRepeatModeChanged(bool)),0);
  disconnect(PeriodField ,SIGNAL(Edited()) ,this ,SLOT(OnPeriodFieldEdited()));
  Disconnect(Gauge,SIGNAL(PeriodChanged(sPhysValue)),0);
  sGaugeDisplay::OnExitFromOperatingState();
}


void  sGaugeTestboard::OnStart ()
{
  if( ! SendCommand(Gauge,sGauge::cStart) ){
    QMessageBox::critical(this,"Denied"
                         ,QString("Device '%1' is busy.\nTry again later.")
                                 .arg(ToQString(Gauge->DeviceIdentifier()))  );
  };
}


void  sGaugeTestboard::OnStop ()
{
  if( ! SendCommand(Gauge,sGauge::cStop,cReleaseControl) ){
    QMessageBox::critical(this,"Denied"
                         ,QString("Device '%1' is busy.\nTry again later.")
                                 .arg(ToQString(Gauge->DeviceIdentifier()))  );
  }else{
    emit ControlReleased();
  }
}


void  sGaugeTestboard::OnStartExclusive ()
{
  if( ! SendCommand(Gauge,sGauge::cStart,cTakeControl) ){
    QMessageBox::critical(this,"Denied"
                         ,QString("Device '%1' is busy.\nTry again later.")
                                 .arg(ToQString(Gauge->DeviceIdentifier()))  );
  }else{
    emit ControlTaken();
  }
}


void  sGaugeTestboard::BerateTheUser ()
{
  QMessageBox::critical(this,"Denied"
                       ,"Watch your hands, you, bloody bastard!!!\n"
                        "The gauge is under exclusive control ALREADY.");
}


void  sGaugeTestboard::OnPeriodFieldEdited ()
{
  if( ! SendDirective(Gauge,sGauge::cSetPeriod,PeriodField->Value()) ){
    QMessageBox::critical(this,"Denied"
                         ,QString("Device '%1' is busy.\nTry again later.")
                                 .arg(ToQString(Gauge->DeviceIdentifier()))  );
    PeriodField->SetValue(Gauge->Period());
  }else{
    PeriodField->SetAttention();
  }
}


void  sGaugeTestboard::OnPeriodChanged (sPhysValue  period)
{
  PeriodField->SetValue(period);
  PeriodField->UnsetAttention();
}


void  sGaugeTestboard::OnAutoRepeatCheckBoxEdited (bool  mode)
{
  if( ! SendDirective(Gauge,sGauge::cSetAutoRepeatMode,mode) ){
    QMessageBox::critical(this,"Denied"
                         ,QString("Device '%1' is busy.\nTry again later.")
                                 .arg(ToQString(Gauge->DeviceIdentifier()))  );
    AutoRepeatCheckBox->setChecked(Gauge->AutoRepeat());
  }else{
    setStyleSheet("#AutoRepeatCheckBox {color:red; } "
                  "#AutoRepeatCheckBox::indicator {} " //sic!
                  );
  }
}


void  sGaugeTestboard::OnAutoRepeatModeChanged (bool  mode)
{
  AutoRepeatCheckBox->setChecked(mode);
  setStyleSheet("");
}

//------------------------------------------------------------------------------
} //namespace RA


