////////////////////////////////////////////////////////////////////////////////
/*! @file MSDNGauge.cpp  A gauge, which takes readings from periodically comming
                         telexes (for instance, generated by an MSDN device).
- Part of DevelopmentTools - Research Assistant Plugin package.
- Uses  RAGUI - Research Assistant Graphical User Interface Library.
- Uses  Qt v.5 - http://qt-project.org/
- Uses  RANet - Research Assistant Net Library.
- Copyright(C) 2016, Viktor E. Bursian, St.Petersburg, Russia.
                          Viktor.Bursian@mail.ioffe.ru
*///////////////////////////////////////////////////////////////////////////////
#include "MSDNGauge.h"
#include "attr.h"
#include "StringsAndQStrings.h"
#include "PhysValue.h"
#include "Log.h"
#include <limits>
#include <cstdlib>
namespace RA {
//------------------------------------------------------------------------------

DEFINE_CLASS_TAG(sMSDNGauge)

//--------------------------------------------------------------- sMSDNGauge ---

void  sMSDNGauge::Preconfigure (sNodePtr  DeviceCfg)
{
  sPtr<sFolderNode>              DeviceLinks;
  DeviceCfg |= "Links";
  DeviceLinks = DeviceCfg >> "Links";
  if( DeviceLinks.IsCorrect() ){
    DeviceLinks |= "MSDN Port";
    DeviceLinks >> "MSDN Port" |= attr("Class","sTeleCom");
  }
  sPtr<sFolderNode>              DeviceSettings;
  DeviceCfg |= "Settings";
  DeviceSettings = DeviceCfg >> "Settings";
  if( DeviceSettings.IsCorrect() ){
    DeviceSettings |= attr("ReadingsPrefix","V");
    DeviceSettings |= attr("SetPeriodPrefix","P");
    DeviceSettings |= attr("GetPeriodPrefix","P");
    DeviceSettings |= attr("StartCommandChar","S");
    DeviceSettings |= attr("StopCommandChar","s");
    DeviceSettings |= attr("ToSendOnConnect","?\nS");
    DeviceSettings |= attr("Calibration",sPhysValue() =
                                         sPhysValue(1,_Unitsless_));
    DeviceSettings |= attr("ZeroCorrection",sPhysValue() =
                                            sPhysValue(0,_Unitsless_));
    DeviceSettings |= attr("PeriodConversion",sPhysValue() =
                                              sPhysValue(1e-6,"s"));
    // изменение умолчаний для sGauge:
    DeviceSettings |= attr("Period",sPhysValue(0.01,"s"));
  }
  sGauge::Preconfigure(DeviceCfg);
}


sMSDNGauge::~sMSDNGauge ()
{
  MSDNPort.SetToNULL();
}


sMSDNGauge::sMSDNGauge (sString   DevicetId
                                           ,sNodePtr  DeviceCfg)
    :sGauge(DevicetId,DeviceCfg)
    ,MSDNPort(this)
    ,ReadingsPrefix('V')
    ,SetPeriodPrefix('P')
    ,GetPeriodPrefix('P')
    ,StartCommandChar('S')
    ,StopCommandChar('s')
    ,ToSendOnConnect()
    ,Calibration()
    ,ZeroCorrection()
    ,PeriodConversion()
{
  Calibration = sPhysValue(1,_Unitsless_);
  ZeroCorrection = sPhysValue(0,_Unitsless_);
  PeriodConversion = sPhysValue(1e-6,"s");
}


void  sMSDNGauge::PostConstructor ()
{
  sGauge::PostConstructor();
  try{ //get config parameters
    sString                   S;
    if( get_from<sTextNode>(DeviceSettings>>"ReadingsPrefix",S) ){
      if( S.Length() == 1 ){
        ReadingsPrefix = S[0];
      }
    }
    if( get_from<sTextNode>(DeviceSettings>>"SetPeriodPrefix",S) ){
      if( S.Length() == 1 )  SetPeriodPrefix = S[0];
      else                   SetPeriodPrefix = 0x0;
    }
    if( get_from<sTextNode>(DeviceSettings>>"GetPeriodPrefix",S) ){
      if( S.Length() == 1 )  GetPeriodPrefix = S[0];
      else                   GetPeriodPrefix = 0x0;
    }
    if( get_from<sTextNode>(DeviceSettings>>"StartCommandChar",S) ){
      if( S.Length() == 1 )  StartCommandChar = S[0];
      else                   StartCommandChar = 0x0;
    }
    if( get_from<sTextNode>(DeviceSettings>>"StopCommandChar",S) ){
      if( S.Length() == 1 )  StopCommandChar = S[0];
      else                   StopCommandChar = 0x0;
    }
    if( get_from<sTextNode>(DeviceSettings>>"ToSendOnConnect",S) ){
      if( ! S.Empty() ){
        ToSendOnConnect = S;
      }
    }
    get_from<sPhysValueNode>(DeviceSettings>>"Calibration",Calibration);
    get_from<sPhysValueNode>(DeviceSettings>>"ZeroCorrection",ZeroCorrection);
    (Calibration/ZeroCorrection).ToReal();
    get_from<sPhysValueNode>(DeviceSettings>>"PeriodConversion"
                                             ,PeriodConversion);
  }catch(...){
    SetInvalid(tr("Something wrong in configuration."));
  }
  // evolution of the inherited state machine
  {
    //evolution of the State_Warming
    {
      State_W_MSDNPort
          =( new sState(State_Warming()) )
          ->SetAsInitial()
          ->AddTransition(State_Warming()->Fin());
      //State_W_MSDNPort internal structure and behaviour
      {
        State_W_MSDNPort_Linking
            =( new sState(State_W_MSDNPort) )
            ->SetAsInitial()
            ->OnEnterCall(SLOT(LinkToMSDNPort()))
            ->AddTransition(SIGNAL(MSDNPortLinked()) ,State_W_MSDNPort->Fin())
            ->AddTransition(SIGNAL(ErrorMessage())  ,State_Cooling());
      }
    }
    // evolution of the State_Operating
    {
      //evolution of the State_O_Measuring
      {
        State_O_M_Acquiring
            ->AddTransition(SIGNAL(NewData(sPhysValue)), State_O_M_Cycling);
      }
    }
    connect(this ,SIGNAL(GotIncommingMessage(sTelex))
           ,this ,SLOT(ProcessIncommingMessage(sTelex)) );
  }
}


void sMSDNGauge::ReportMSDNPortError (QString  error_text)
{
  emit ErrorMessage( tr("Got error from '%1':\n\n%2")
                       .arg(ToQString(MSDNPort->DeviceIdentifier()))
                       .arg(error_text)     );
}


void sMSDNGauge::LinkToMSDNPort ()
{
  try{
    MSDNPort.SetTo(DeviceLinks,"MSDN Port");
    Connect(MSDNPort ,SIGNAL(Warmed()) ,SIGNAL(MSDNPortLinked()) );
    Connect(MSDNPort ,SIGNAL(ErrorMessage(QString))
           ,SIGNAL(MSDNPortError(QString)) );
  }catch( rxDeviceLaunch E ){
    emit ErrorMessage( tr("Error while linking to a submissive device.\n"
                          "%1\nCheck configuration.")
                         .arg(ToQString(E.MsgText()))   );
  }
}


void sMSDNGauge::OnEnterToCoolingState ()
{
  Disconnect(MSDNPort,0,0);//all its signals connected to any of my slots
  MSDNPort.ReleaseExclusiveControl();
  MSDNPort.SetToNULL();
  sGauge::OnEnterToCoolingState();
}


void sMSDNGauge::OnEnterToOperatingState()
{
  sGauge::OnEnterToOperatingState();
  Connect(MSDNPort ,SIGNAL(GotTelex(sTelex))
         ,SIGNAL(GotIncommingMessage(sTelex)) );
  SendDirective(MSDNPort ,sTeleCom::cSendTelex
               ,sTelex(ToSendOnConnect) );
}


void sMSDNGauge::OnExitFromOperatingState()
{
  Disconnect(MSDNPort ,SIGNAL(GotTelex(sTelex))
            ,SIGNAL(GotIncommingMessage(sTelex)) );
  sGauge::OnExitFromOperatingState();
}


void  sMSDNGauge::SetWantedPeriod ()
{
  if( GetPeriodPrefix ){
    real                          Period;
    quint32                       P;
    sTelex                        T;
    try{
      Period = (WantedPeriod/PeriodConversion).ToReal();
      if( Period < 1 ){
        P = 1;
      }else if( Period > std::numeric_limits<quint32>::max() ){
        P = std::numeric_limits<quint32>::max();
      }else{
        P = Period;
      }
      T.append(GetPeriodPrefix);
      T.append(QByteArray::number(P, 16).toUpper());
      SendDirective(MSDNPort ,sTeleCom::cSendTelex ,T );
    }catch( rxRAlgebra ){
      SetInvalid(tr("Cannot convert period.  Wrong settings?"));
    }
  }else{
    ThePeriod = WantedPeriod;
    emit PeriodChanged(ThePeriod);
  }
}


void  sMSDNGauge::ProcessIncommingMessage (sTelex  text)
{
  QString                     ConversionErr(tr("Got '%1'. Cannot convert."));
  bool                        OK = true;
  qint32                      Reading;
  if( text.startsWith(ReadingsPrefix) ){
    real                        V = 1.0;
    int                         P;
    if( (P = text.indexOf('/',2)) > 0 ){
      Reading = text.mid(P+1).toLongLong(&OK,16);
      if( ! OK ){
        emit MSDNPortError(ConversionErr.arg(QString::fromLatin1(text)));
      }else{
        V /= Reading;
      }
    }else{
      P = -1;
    }
    if( OK ){
      Reading = text.mid(1,P-1).toLongLong(&OK,16);
      if( ! OK ){
        emit MSDNPortError(ConversionErr.arg(QString::fromLatin1(text)));
      }else{
        V *= Reading;
        emit NewData( V * Calibration - ZeroCorrection );
      }
    }

  }else if( text.startsWith('D') ){
    real                        D = 1.0;
    int                         P;
    if( (P = text.indexOf('/',2)) > 0 ){
      Reading = text.mid(P+1).toLongLong(&OK,16);
      if( ! OK ){
        emit MSDNPortError(ConversionErr.arg(QString::fromLatin1(text)));
      }else{
        D /= Reading;
      }
    }else{
      P = -1;
    }
    if( OK ){
      Reading = text.mid(1,P-1).toLongLong(&OK,16);
      if( ! OK ){
        emit MSDNPortError(ConversionErr.arg(QString::fromLatin1(text)));
      }else{
        D *= Reading;
        if( D < 0 ){
          emit ErrorMessage("negative D");
//        }else{
//          D = sqrt(D);
//          if( D < 4 )
        }
      }
    }

  }else if( text.startsWith(GetPeriodPrefix) ){
    Reading = text.mid(1).toLongLong(&OK,16);
    if( ! OK ){
      emit MSDNPortError(ConversionErr.arg(QString::fromLatin1(text)));
    }else{
      ThePeriod = real(Reading) * PeriodConversion;
      emit PeriodChanged(ThePeriod);
    }

  }
}


//------------------------------------------------------------------------------
} //namespace RA
