////////////////////////////////////////////////////////////////////////////////
/*! @file NewConsole.cpp A new console pattern for the developers.
- Part of Examples - Research Assistant Plugin package.
- Uses  RAGUI - Research Assistant Graphical User Interface.
- Uses  RANet - Research Assistant Net Library.
- Uses  Qt v.5  - http://qt-project.org/
- Copyright(C) 2015, Viktor E. Bursian, St.Petersburg, Russia.
                     Viktor_dot_Bursian_at_mail_dot_ioffe_dot_ru
*///////////////////////////////////////////////////////////////////////////////
#include "NewConsole.h"
#include "attr.h"
#include "StringsAndQStrings.h"
#include "Log.h"
#include <QHBoxLayout>
#include <QMessageBox>
namespace RA {
//------------------------------------------------------------------------------

DEFINE_CLASS_TAG(sNewConsole)

//-------------------------------------------------------------- sNewConsole ---
/*! @class sNewConsole

@par The state machine

The upper-level state machine is inherited from sConsole.  
Below is the evolution of some states.
@dotfile "NewConsole.state_graph"

*/


void  sNewConsole::Preconfigure (sNodePtr  DeviceCfg)
{
  sPtr<sFolderNode>              DeviceLinks;
  DeviceCfg |= "Links";
  DeviceLinks = DeviceCfg >> "Links";
  if( DeviceLinks.IsCorrect() ){
    DeviceLinks |= "Device1";
    DeviceLinks >> "Device1" |= attr("Class","sNewDriver");
    DeviceLinks |= "Device2";
    DeviceLinks >> "Device2" |= attr("Class","sNewDriver");
  }
  sPtr<sFolderNode>              DeviceSettings;
  DeviceCfg |= "Settings";
  DeviceSettings = DeviceCfg >> "Settings";
  if( DeviceSettings.IsCorrect() ){
    DeviceSettings |= attr("IntParameter",317);
    DeviceSettings |= attr("PhysParameter",sPhysValue(38,"parot"));
  }
  sConsole::Preconfigure(DeviceCfg);
}


sNewConsole::~sNewConsole()
{
  if( DeviceSettings.IsCorrect() ){
    // Не все конфигурационные параметры следует сохранять после работы.
    DeviceSettings << attr("PhysParameter",PhysParameter);
  }
  Device1.SetToNULL();
  Device2.SetToNULL();
}



/*! Constructor.

It is declared by CONSOLE() macro or ABSTRACT_CONSOLE() macro.  
You must provide it's definition.
*/
sNewConsole::sNewConsole (sString    DeviceId
                         ,sNodePtr   DeviceCfg
                         ,sNodePtr   WidgetCfg
                         ,QWidget *  parent   )
    :sConsole(DeviceId,DeviceCfg,WidgetCfg,parent)
    ,Device1(this)
    ,Device2(this)
{
  try{ //get config parameters
    get_from<sPhysValueNode>(DeviceSettings>>"PhysParameter",PhysParameter);
  }catch(...){
    SetInvalid(tr("Something wrong in configuration."));
  }
  //evolution of the inherited state machine
  {
    //evolution of the State_Warming
    {
      State_W_Parallel 
          = (new sParallelStateGroup(State_Warming()))
          ->SetAsInitial()
          ->AddTransition(State_Warming()->Fin());
      //State_W_Parallel internal structure and behaviour
      {
        State_W_Device1 = new sState(State_W_Parallel,"State_W_Device1");
        State_W_Device2 = new sState(State_W_Parallel,"State_W_Device2");
        //State_W_Device1 structure and behaviour
        {
          State_W_Device1_Linking 
              = (new sState(State_W_Device1 ,"State_W_Device1_Linking"))
              ->SetAsInitial()
              ->OnEnterCall( SLOT(LinkToDevice1()) )
              ->AddTransition( SIGNAL(Device1Linked()) ,State_W_Device1->Fin() )
              ->AddTransition( SIGNAL(ErrorMessage()) ,State_Cooling() )
              ;
        }
        //State_W_Device2 structure and behaviour
        {
          State_W_Device2_Linking 
              = (new sState(State_W_Device2 ,"State_W_Device2_Linking"))
              ->SetAsInitial()
              ->OnEnterCall( SLOT(LinkToDevice2()) )
              ->AddTransition( SIGNAL(Device2Linked()) ,State_W_Device2->Fin() )
              ->AddTransition( SIGNAL(ErrorMessage()) ,State_Cooling() )
              ;
        }
      }
    }
    //evolution of the State_Operating
    {
    }
    //evolution of the State_Cooling
    {
    }
    connect(this ,SIGNAL(Device1Error(QString))
           ,this ,SLOT(ReportDevice1Error(QString)) );
    connect(this ,SIGNAL(Device2Error(QString))
           ,this ,SLOT(ReportDevice2Error(QString)) );
  }
}


void  sNewConsole::AppearanceMakeUp ()
{
  sConsole::AppearanceMakeUp();
  try{ //get config parameters (which are necessary for appearance)
    if( ! get_from<sIntegerNode>(DeviceSettings>>"IntParameter",IntParameter) )  
      throw xDeviceLaunch;
  }catch(...){
    SetInvalid(tr("Something wrong in configuration."));
  }
  //add new widgets
  {
  }
  Translate();
}


void  sNewConsole::Translate ()
{
  RangeField->SetLabel(tr("Range"));
  StepField->SetLabel(tr("Step"));
}


void  sNewConsole::RestoreState ()
{
  sConsole::RestoreState();
  if( WidgetConfig.IsCorrect() ){
    get_from<sIntegerNode>(WidgetConfig>>"IntParameter" ,IntParameter);
    //change the appearance, if it is affected by this parameter
  }
}


void  sNewConsole::StoreState ()
{
  sConsole::StoreState();
  if( WidgetConfig.IsCorrect() ){
    WidgetConfig << attr("IntParameter",IntParameter);
    //no need to store those parameters, which are stored in destructor
  }
}


bool  sNewConsole::CanClose ()
{
//  if(  )
//    return false;
  return true;
}



/*! Called by the core on enter into the State_Operating().  

Do not forget to call the ancestor's slot.  
If you need no extra actions, then remove this method.

Usually, you should connect the signals of the submissive drivers 
with the signals (or slots) of this device:
@code
  Connect(NewDevice1 ,SIGNAL(NewData(sPhysValue)),SIGNAL(ShowData(sPhysValue)));
@endcode
Note, this could not be done, say, in constructor, before the link to
the submissive driver got its value.

Note, there is another way to do something on enter into State_Operating(),
namelly, you can add 
@code
    State_Operating()->OnEnterCall(...);
@endcode
into the constructor.  This will call your slot after all actions,
made by the ancestor's class.  Here, however, you can do something BEFORE
the ancestor's actions.

@sa OnExitFromOperatingState().
*/
void  sNewConsole::OnEnterToOperatingState ()
{
  sConsole::OnEnterToOperatingState();
  Connect(Device1 ,SIGNAL(Angle(sPhysValue)) ,SLOT(ShowAngle(sPhysValue)) );
}



/*! Called by the core on exit from the State_Operating().  

Do not forget to call the ancestor's slot.  
If you need no extra actions, then remove this method.

Usually, you should disconnect the signals of the submissive 
drivers, which were connected on enter into this state:
@code
  Disconnect(NewDevice1 ,SIGNAL(NewData(sPhysValue)) 
            ,SIGNAL(ShowData(sPhysValue)));
@endcode

Note, there is another way to do something on exit from State_Operating(),
namelly, you can add 
@code
    State_Operating()->OnExitCall(...);
@endcode
into the constructor.  This will call your slot after all actions,
made by the ancestor's class.  Here, however, you can do something BEFORE
the ancestor's actions.

@sa OnEnterToCoolingState().
*/
void  sNewConsole::OnExitFromOperatingState ()
{
  Disconnect(Device1 ,SIGNAL(Angle(sPhysValue)) ,SLOT(ShowAngle1(sPhysValue)) );
  sConsole::OnExitFromOperatingState();
}



/*! Called by the core on enter into the State_Cooling().  

Do not forget to call the ancestor's slot.  
If you need no extra actions, then remove this method.

Usually, you should disconnect all submissive drivers, like this:
@code
  Disconnect(NewDevice1 ,0 ,0);
  NewDevice1.ReleaseExclusiveControl();
  NewDevice1.SetToNULL();
@endcode

Note, there is another way to do something on enter into State_Cooling(),
namelly, you can add 
@code
    State_Cooling()->OnEnterCall(...);
@endcode
into the constructor.  This will call your slot after all actions,
made by the ancestor's class.  Here, however, you can do something BEFORE
the ancestor's actions.

@sa OnExitFromOperatingState().
*/
void  sNewConsole::OnEnterToCoolingState ()
{
  Disconnect(Device1 ,0 ,0);
  Device1.ReleaseExclusiveControl();
  Device1.SetToNULL();
  Disconnect(Device2 ,0 ,0);
  Device2.ReleaseExclusiveControl();
  Device2.SetToNULL();
  sConsole::OnEnterToCoolingState();
}


void sNewConsole::ReportDevice1Error (QString  error_text)
{
  emit ErrorMessage( tr("Got error from '%1':\n\n%2")
                       .arg(ToQString(Device1->DeviceIdentifier()))
                       .arg(error_text)     );
}


void sNewConsole::ReportDevice2Error (QString  error_text)
{
  emit ErrorMessage( tr("Got error from '%1':\n\n%2")
                       .arg(ToQString(Device2->DeviceIdentifier()))
                       .arg(error_text)     );
}


void  sNewConsole::LinkToDevice1 ()
{
  try{
    Device1.SetTo(DeviceLinks,"Device1");
    Connect(Device1 ,SIGNAL(Warmed())       ,SIGNAL(Device1Linked()) );
    Connect(Device1 ,SIGNAL(ErrorMessage(QString))
           ,SIGNAL(Device1Error(QString)) );
    //Connect(Device1 ,SIGNAL(...) ,SIGNAL(...) );
  }catch( rxDeviceLaunch E ){
    emit ErrorMessage( tr("Error while linking to a submissive device.\n"
                          "%1\nCheck configuration.")
                         .arg(ToQString(E.MsgText()))   );
  }
}


void  sNewConsole::LinkToDevice2 ()
{
  try{
    Device2.SetTo(DeviceLinks,"Device2");
    Connect(Device2 ,SIGNAL(Warmed())       ,SIGNAL(Device2Linked()) );
    Connect(Device2 ,SIGNAL(ErrorMessage(QString))
           ,SIGNAL(Device2Error(QString)) );
    //Connect(Device2 ,SIGNAL(...) ,SIGNAL(...) );
  }catch( rxDeviceLaunch E ){
    emit ErrorMessage( tr("Error while linking to a submissive device.\n"
                          "%1\nCheck configuration.")
                         .arg(ToQString(E.MsgText()))   );
  }
}

//------------------------------------------------------------------------------
} //namespace RA


