////////////////////////////////////////////////////////////////////////////////
/*! @file NewDriver.cpp A new driver pattern for the developers.
- Part of Examples - Research Assistant Plugin package.
- Uses  RAGUI - Research Assistant Graphical User Interface.
- Uses  RANet - Research Assistant Net Library.
- Uses  Qt v.5  - http://qt-project.org/
- Copyright(C) 2015, Viktor E. Bursian, St.Petersburg, Russia.
                     Viktor_dot_Bursian_at_mail_dot_ioffe_dot_ru
*///////////////////////////////////////////////////////////////////////////////
#include "NewDriver.h"
#include "attr.h"
#include "StringsAndQStrings.h"
#include "Log.h"
namespace RA {
//------------------------------------------------------------------------------

DEFINE_CLASS_TAG(sNewDriver)

csCommand                     sNewDriver::cStandUp;
csCommand                     sNewDriver::cSitDown;

/*!
Приказывает крутиться с заданной скоростью.

Скорость должна быть в единицах, приводимых к "degree/s".

Работает только, если устройство в состоянии State_O_Standing.
*/
const sDirective<sPhysValue>  sNewDriver::cRockNRoll;

//--------------------------------------------------------------- sNewDriver ---
/*! @class sNewDriver


@par The state machine
@dotfile "NewDriver.state_graph"

*/

void  sNewDriver::Preconfigure (sNodePtr  DeviceCfg)
{
  sPtr<sFolderNode>              DeviceSettings;
  DeviceCfg |= "Settings";
  DeviceSettings = DeviceCfg >> "Settings";
  if( DeviceSettings.IsCorrect() ){
    DeviceSettings |= attr("Speed",sPhysValue(3.6,"kdegree/min"));
    DeviceSettings |= attr("CurrentAngle",sPhysValue(0.0,"degree"));
  }
  sDriver::Preconfigure(DeviceCfg);
}


sNewDriver::~sNewDriver ()
{
  if( DeviceSettings.IsCorrect() ){
    // Не все конфигурационные параметры следует сохранять после работы.
    DeviceSettings << attr("CurrentAngle",CurrentAngle);
  }
}



/*! Constructor.

It is declared by DRIVER() macro or ABSTRACT_DRIVER() macro.  
You must provide it's definition
*/
sNewDriver::sNewDriver (sString   DevicetId
                       ,sNodePtr  DeviceCfg)
    :sDriver(DevicetId,DeviceCfg)
    ,Speed("degree/s")
    ,CurrentAngle("degree")
{
}



/*! Called by the core right after constructor 
    (before the start of the state machine).

The purpose of this method is to use virtual methods in the initiation,
which does not work within constructors.

@note
Always call the ancestor's PostConstructor() at the beginning of this function,
if you reimplement it.
*/
void  sNewDriver::PostConstructor ()
{
  sDriver::PostConstructor();
  try{ //get config parameters
    //necessary parameter:
    if( ! get_from<sPhysValueNode>(DeviceSettings>>"Speed",Speed) )
      throw xDeviceLaunch;
    //optional parameter:
    get_from<sPhysValueNode>(DeviceSettings>>"CurrentAngle",CurrentAngle);
  }catch(...){
    SetInvalid(tr("Something wrong in configuration."));
  }
  //evolution of the inherited state machine (behaviour)
  {
    //evolution of the State_Warming
    {
      State_W_Stage1 = new QState(State_Warming());
      State_W_Stage2 = new QState(State_Warming());
      State_W_Stage1
          ->SetAsInitial()
          ;
      //State_W_Stage1 structure and behaviour
      {
      }
      //State_W_Stage2 structure and behaviour
      {
      }
      //...
    }
    //evolution of the State_Operating
    {
      State_O_Sitting = new QState(State_Operating());
      State_O_Standing = new QState(State_Operating());
      State_O_Rolling = new QState(State_Operating());
      State_O_Sitting
          ->SetAsInitial()
          ;
      //State_O_Sitting structure and behaviour
      {
      }
      //State_O_Standing structure and behaviour
      {
      }
      //...
    }
    //evolution of the State_Cooling
    {
    }
  }
  //...
}



/*! Called by the core on enter into the State_Operating().  

Do not forget to call the ancestor's slot.  
If you need no extra actions, then remove this method.

Usually, you should connect the signals of the submissive drivers 
with the signals (or slots) of this device:
@code
  Connect(Slave ,SIGNAL(NewData(sPhysValue)),SIGNAL(ShowData(sPhysValue)));
@endcode
Note, this could not be done, say, in PostConstructor(), before the link to
the submissive driver got its value.

Note, there is another way to do something on enter into State_Operating(),
namelly, you can add 
@code
    State_Operating()->OnEnterCall(...);
@endcode
into the PostConstructor().  This will call your slot after all actions,
made by the ancestor's class.  Here, however, you can do something BEFORE
the ancestor's actions.

@sa OnExitFromOperatingState().
*/
void  sNewDriver::OnEnterToOperatingState ()
{
  sDriver::OnEnterToOperatingState();
  
  //...
  
}



/*! Called by the core on exit from the State_Operating().  

Do not forget to call the ancestor's slot.  
If you need no extra actions, then remove this method.

Usually, you should disconnect the signals of the submissive 
drivers, which were connected on enter into this state:
@code
  Disconnect(Slave ,SIGNAL(NewData(sPhysValue)) ,SIGNAL(ShowData(sPhysValue)));
@endcode

Note, there is another way to do something on exit from State_Operating(),
namelly, you can add 
@code
    State_Operating()->OnExitCall(...);
@endcode
into the PostConstructor().  This will call your slot after all actions,
made by the ancestor's class.  Here, however, you can do something BEFORE
the ancestor's actions.

@sa OnEnterToCoolingState().
*/
void  sNewDriver::OnExitFromOperatingState ()
{
  
  //...
  
  sDriver::OnExitFromOperatingState();
}



/*! Called by the core on enter into the State_Cooling().  

Do not forget to call the ancestor's slot.  
If you need no extra actions, then remove this method.

Usually, you should disconnect all submissive drivers, like this:
@code
  Disconnect(Slave ,0 ,0);
  Slave.ReleaseExclusiveControl();
  Slave.SetToNULL();
@endcode

Note, there is another way to do something on enter into State_Cooling(),
namelly, you can add 
@code
    State_Cooling()->OnEnterCall(...);
@endcode
into the PostConstructor().  This will call your slot after all actions,
made by the ancestor's class.  Here, however, you can do something BEFORE
the ancestor's actions.

@sa OnExitFromOperatingState().
*/
void  sNewDriver::OnEnterToCoolingState ()
{
  
  //...
  
  sDriver::OnEnterToCoolingState();
}

//------------------------------------------------------------------------------
} //namespace RA
