////////////////////////////////////////////////////////////////////////////////
/*! @file SerialPortFinder.cpp Console showing available ports.
- Part of DevelopmentTools - Research Assistant Plugin package.
- Uses  RAGUI - Research Assistant Graphical User Interface Library.
- Uses  Qt v.5  - http://qt-project.org/
- Uses  RANet - Research Assistant Net Library.
- Copyright(C) 2017, Viktor E. Bursian, St.Petersburg, Russia.
                     Viktor.Bursian@mail.ioffe.ru
*///////////////////////////////////////////////////////////////////////////////
#include "SerialPortFinder.h"
#include "StringsAndQStrings.h"
#include <QVBoxLayout>
#include <QHBoxLayout>
#include <QScrollArea>
#include <QtSerialPort/QSerialPortInfo>
namespace RA {
//------------------------------------------------------------------------------

DEFINE_CLASS_TAG(sSerialPortFinder)

//-------------------------------------------------------- sSerialPortFinder ---

void  sSerialPortFinder::Preconfigure (sNodePtr  DeviceCfg)
{
  sConsole::Preconfigure(DeviceCfg);
}


sSerialPortFinder::~sSerialPortFinder ()
{
}


sSerialPortFinder::sSerialPortFinder (sString   DeviceId
                                   ,sNodePtr  DeviceCfg
                                   ,sNodePtr  WidgetCfg
                                   ,QWidget *  parent    )
    :sConsole(DeviceId,DeviceCfg,WidgetCfg,parent)
    ,ShowAnyway(true)
{
  ShowChangesOnly = new QCheckBox("show changes only",this);
  Total = new QLabel(this);
  Screen = new QTextEdit(this);
  Screen->setReadOnly(true);
  RefreshClock = new QTimer(this);
}


void  sSerialPortFinder::AppearanceMakeUp ()
{
  sConsole::AppearanceMakeUp();

  setLayout(new QVBoxLayout);

  QLayout *                   L = new QHBoxLayout;
  layout()->addItem(L);
  L->addWidget(ShowChangesOnly);
  L->addWidget(Total);

  QScrollArea *               ScrollArea = new QScrollArea(this);
  layout()->addWidget(ScrollArea);
  ScrollArea->setWidget(Screen);
  ScrollArea->setWidgetResizable(true);
}


void sSerialPortFinder::OnEnterToOperatingState ()
{
  sConsole::OnEnterToOperatingState();
  PrevInfoStrList = FormInfoStrList();
  connect(ShowChangesOnly ,SIGNAL(stateChanged(int))
         ,this ,SLOT(MustBeShown()) );
  connect(RefreshClock ,SIGNAL(timeout()) ,this ,SLOT(ShowList()) );
  setEnabled(true);
  RefreshClock->start(1000);
}


void sSerialPortFinder::OnExitFromOperatingState ()
{
  RefreshClock->stop();
  setEnabled(false);
  disconnect(RefreshClock ,SIGNAL(timeout()) ,0 ,0 );
  sConsole::OnExitFromOperatingState();
}


void sSerialPortFinder::MustBeShown ()
{
  ShowAnyway = true;
}


QList<QString> sSerialPortFinder::FormInfoStrList ()
{
  QList<QString>              NewInfoStrList;
  foreach (const QSerialPortInfo & Info, QSerialPortInfo::availablePorts()) {
    QString                     InfoStr;
    InfoStr = "Name='<b>" + Info.portName()
        +"</b>', Location='" + Info.systemLocation() + "'"
        +", Busy=" + ( Info.isBusy() ? "Yes" : "No" );
    if( ! Info.description().isEmpty() )
      InfoStr += ", Description='" + Info.description() + "'";
    if( ! Info.manufacturer().isEmpty() )
      InfoStr += ", Manufacturer='" + Info.manufacturer() + "'";
    if( ! Info.serialNumber().isEmpty() )
      InfoStr += ", SN='" + Info.serialNumber() + "'";

    if( Info.hasVendorIdentifier() )
      InfoStr += ", VID=0x" + QByteArray::number(Info.vendorIdentifier(), 16)
               + "=" + QByteArray::number(Info.vendorIdentifier(), 10);
    if( Info.hasProductIdentifier() )
      InfoStr += ", PID=0x" + QByteArray::number(Info.productIdentifier(), 16)
               + "=" + QByteArray::number(Info.productIdentifier(), 10);
    NewInfoStrList.append(InfoStr);
  }
  return NewInfoStrList;
}


void  sSerialPortFinder::ShowList ()
{
  QList<QString>              NewInfoStrList = FormInfoStrList();
  if( ShowAnyway || (PrevInfoStrList != NewInfoStrList) ){
    ShowAnyway = false;
    Screen->clear();
    foreach (const QString & InfoStr, NewInfoStrList) {
      if( ! PrevInfoStrList.contains(InfoStr) ){
        Screen->append( QString("<font size=-1 color=green>") + InfoStr );
      }else if( ! ShowChangesOnly->isChecked() ){
        Screen->append( QString("<font size=-1>") + InfoStr);
      }
    }
    PrevInfoStrList = NewInfoStrList;
  }
  Total->setText(QString("Total number of ports: %1")
                 .arg(NewInfoStrList.count()));
}

//------------------------------------------------------------------------------
} //namespace RA
