////////////////////////////////////////////////////////////////////////////////
/*! @file TelePipe.cpp One end of a telecommunication pipe.
- Part of DevelopmentTools - Research Assistant Plugin package.
- Uses  RAGUI - Research Assistant Graphical User Interface Library.
- Uses  Qt v.5.0.2  - http://qt-project.org/
- Uses  RANet - Research Assistant Net Library.
- Copyright(C) 2013-2014, Viktor E. Bursian, St.Petersburg, Russia.
                          Viktor.Bursian@mail.ioffe.ru
*///////////////////////////////////////////////////////////////////////////////
#include "TelePipe.h"
#include "attr.h"
#include "StringsAndQStrings.h"
//#include "Trace.h"
namespace RA {
//------------------------------------------------------------------------------

DEFINE_CLASS_TAG(sTelePipe)

//---------------------------------------------------------------- sTelePipe ---
/*! @class sTelePipe

Важный момент: объект этого класса, сам по себе, ещё не труба, а как бы
половинка трубы - L-образное колено.  Верхним концом он притворяется объектом
класса sTeleCom и может быть подключён к драйверу или к консоли в качестве
телекоммуникационного канала.  Боковой конец следует соединить с другим
объектом класса sTelePipe. Тогда получается как бы U-образная труба.  Всё, что
заливается в один конец, фонтанирует из другого, и наоборот.

*/

void  sTelePipe::Preconfigure (sNodePtr  DeviceCfg)
{
  sPtr<sFolderNode>              DeviceLinks;
  DeviceCfg |= "Links";
  DeviceLinks = DeviceCfg >> "Links";
  if( DeviceLinks.IsCorrect() ){
    DeviceLinks |= "Mate";
    DeviceLinks >> "Mate" |= attr("Class","sTelePipe");
  }
  sTeleCom::Preconfigure(DeviceCfg);
}


sTelePipe::~sTelePipe ()
{
  Mate.SetToNULL();
}


sTelePipe::sTelePipe (sString   DevicetId
                     ,sNodePtr  DeviceCfg)
    :sTeleCom(DevicetId,DeviceCfg)
    ,Mate(this)
    ,Mated(false)
    ,Submissive(false)
{
}


void  sTelePipe::PostConstructor ()
{
  sTeleCom::PostConstructor();
  //evolution of the inherited state machine (behaviour)
  {
    //evolution of the State_Warming
    {
      State_Warming()->OnEnterCall(SLOT(LinkToMate()));
    }
    connect(this ,SIGNAL(MateError(QString))
           ,this ,SLOT(ReportMateError(QString)) );
  }
}


void sTelePipe::ReportMateError (QString  error_text)
{
  emit ErrorMessage( tr("My mate '%1' says:\n\n%2")
                       .arg(ToQString(Mate->DeviceIdentifier()))
                       .arg(error_text)     );
}


void sTelePipe::LinkToMate ()
{
  sPtr<sTextNode>             MateId(DeviceLinks >> "Mate" >> "DeviceId");
  if( ! MateId.IsCorrect() ){
    emit ErrorMessage( tr("I have no mate!\nCheck configuration.") );
    SetInvalid();
  }else{
    Mate.SetTo(sString(*MateId));
    if( ! Mate.IsCorrect() ){
      emit ErrorMessage( tr("I have been told to mate to '%1', which does not "
                            "look correct. Check configuration.")
                           .arg(ToQString(sString(*MateId))) );
      SetInvalid();
    }else if( Mated ){
      Q_ASSERT( Submissive );
      Q_ASSERT( Mate->Mated );
      Q_ASSERT( ! Mate->Submissive );
      Mate.SetToNULL();
    }else{
      if( ! Mate->Mated ){
        Connect(Mate ,SIGNAL(Warmed())             ,SIGNAL(MateLinked()) );
        Connect(Mate ,SIGNAL(ErrorMessage(QString)),SIGNAL(MateError(QString)));
        Connect(Mate ,SIGNAL(PuttingTelex(sTelex)) ,SIGNAL(GotTelex(sTelex)) );
        connect(this ,SIGNAL(Warmed())             
                ,Mate.operator ->() ,SIGNAL(MateLinked()) 
                ,Qt::QueuedConnection);
        connect(this ,SIGNAL(ErrorMessage(QString))             
                ,Mate.operator ->() ,SIGNAL(MateError(QString)) 
                ,Qt::QueuedConnection);
        connect(this ,SIGNAL(PuttingTelex(sTelex))              
                ,Mate.operator ->() ,SIGNAL(GotTelex(sTelex))  
                ,Qt::QueuedConnection);
        Mate->Mated = true;
        Mate->Submissive = true;
        Submissive = false;
      }else{
        if( ! (this == Mate->Mate.operator ->() ) ){
          emit ErrorMessage( tr("I am trying to mate '%1', but it is already "
                                "mating someone else. Check configuration.")
                             .arg(ToQString(sString(*MateId))) );
          SetInvalid();
        }
        Q_ASSERT( Submissive );
      }
      Mated = true;
    }
  }
}


void sTelePipe::OnEnterToCoolingState ()
{
  if( Mated ){
    if( ! Submissive ){
      Q_ASSERT( Mate->Mated );
      Q_ASSERT( Mate->Submissive );
      Disconnect(Mate,0,0);//all signals from Mate connected to any of my slots
      disconnect(this,0,Mate.operator ->(),0); //and vice versa
      Mate->Mated = false;
      Mate.ReleaseExclusiveControl();
      Mate.SetToNULL();
    }
  }
  Mated = false;
  Submissive = false;
  sTeleCom::OnEnterToCoolingState();
}


//void sTelePipe::OnEnterToOperatingState()
//{
//  sTeleCom::OnEnterToOperatingState();
//  Connect(Mate ,SIGNAL(PuttingTelex(sTelex))
//         ,SIGNAL(GotTelex(sTelex)) );
//}


//void sTelePipe::OnExitFromOperatingState()
//{
//  Disconnect(Mate ,SIGNAL(PuttingTelex(sTelex))
//            ,SIGNAL(GotTelex(sTelex)) );
//  sTeleCom::OnExitFromOperatingState();
//}


void sTelePipe::PutTelex (sTelex  T)
{
  emit PuttingTelex(T);
}

//------------------------------------------------------------------------------
} //namespace RA
