////////////////////////////////////////////////////////////////////////////////
/*! @file TeleSpy.cpp Шпион-диверсант для телекоммуникационного канала.
- Part of DevelopmentTools - Research Assistant Plugin package.
- Uses  RAGUI - Research Assistant Graphical User Interface Library.
- Uses  Qt v.5.0.2  - http://qt-project.org/
- Uses  RANet - Research Assistant Net Library.
- Copyright(C) 2013, Viktor E. Bursian, St.Petersburg, Russia.
                     Viktor.Bursian@mail.ioffe.ru
*///////////////////////////////////////////////////////////////////////////////
#include "TeleSpy.h"
#include "attr.h"
#include "StringsAndQStrings.h"
#include "TeleType.h"
#include <QHBoxLayout>
#include <QVBoxLayout>
#include <QScrollArea>
#include <QLabel>
#include <QMessageBox>
namespace RA {
//------------------------------------------------------------------------------

DEFINE_CLASS_TAG(sTeleSpy)

//----------------------------------------------------------------- sTeleSpy ---

void  sTeleSpy::Preconfigure (sNodePtr  DeviceCfg)
{
  sPtr<sFolderNode>              DeviceLinks;
  DeviceCfg |= "Links";
  DeviceLinks = DeviceCfg >> "Links";
  if( DeviceLinks.IsCorrect() ){
    DeviceLinks |= "LeftTeleCom";
    DeviceLinks >> "LeftTeleCom" |= attr("Class","sTeleCom");
    DeviceLinks |= "RightTeleCom";
    DeviceLinks >> "RightTeleCom" |= attr("Class","sTeleCom");
  }
  sConsole::Preconfigure(DeviceCfg);
}


sTeleSpy::~sTeleSpy()
{
  LeftTeleCom.SetToNULL();
  RightTeleCom.SetToNULL();
}


sTeleSpy::sTeleSpy (sString   DeviceId
                   ,sNodePtr  DeviceCfg
                   ,sNodePtr  WidgetCfg
                   ,QWidget *  parent    )
    :sConsole(DeviceId,DeviceCfg,WidgetCfg,parent)
    ,LeftTeleCom(this)
    ,RightTeleCom(this)
{
  //evolution of the inherited state machine (behaviour)
  {
    //evolution of the State_Warming
    {
      State_W_Parallel 
          = (new sParallelStateGroup(State_Warming()))
          ->SetAsInitial()
          ->AddTransition( State_Warming()->Fin() );
      //State_W_Parallel internal structure and behaviour
      {
        State_W_LeftTeleCom = new sState(State_W_Parallel);
        State_W_RightTeleCom = new sState(State_W_Parallel);
        //State_W_LeftTeleCom internal structure and behaviour
        {
          State_W_LeftTeleCom_Linking 
              =( new sState(State_W_LeftTeleCom) )
              ->SetAsInitial()
              ->OnEnterCall( SLOT(LinkToLeftTeleCom()) )
              ->AddTransition(SIGNAL(LeftTeleComLinked())
                             ,State_W_LeftTeleCom->Fin() )
              ->AddTransition(SIGNAL(ErrorMessage()) ,State_Cooling() );
        }
        //State_W_RightTeleCom internal structure and behaviour
        {
          State_W_RightTeleCom_Linking 
              =( new sState(State_W_RightTeleCom) )
              ->SetAsInitial()
              ->OnEnterCall(SLOT(LinkToRightTeleCom()))
              ->AddTransition(SIGNAL(RightTeleComLinked())
                             ,State_W_RightTeleCom->Fin())
              ->AddTransition(SIGNAL(ErrorMessage()) ,State_Cooling());
        }
      }
    }
    connect(this ,SIGNAL(LeftTeleComError(QString))
           ,this ,SLOT(ReportLeftTeleComError(QString)) );
    connect(this ,SIGNAL(RightTeleComError(QString))
           ,this ,SLOT(ReportRightTeleComError(QString)) );
  }
}


void  sTeleSpy::AppearanceMakeUp ()
{
  sConsole::AppearanceMakeUp();

  setLayout(new QVBoxLayout);

  Screen = new QTextEdit(this);
  Screen->setReadOnly(true);

  QScrollArea *               ScrollArea = new QScrollArea(this);
  layout()->addWidget(ScrollArea);
  ScrollArea->setWidget(Screen);
  ScrollArea->setWidgetResizable(true);

  QLabel *                    Label = new QLabel(this);
  Label->setText("red: from \"left\", blue: from \"right\", bright: typed");
  layout()->addWidget(Label);
  Label->setSizePolicy(QSizePolicy::Preferred,QSizePolicy::Fixed);

  QHBoxLayout *               L = new QHBoxLayout;
  layout()->addItem(L);

  sTelexField *               LeftField = new sTelexField(this);
  L->addWidget(LeftField);
  LeftField->setSizePolicy(QSizePolicy::Preferred,QSizePolicy::Fixed);
  connect(LeftField ,SIGNAL(TelexEntered(sTelex))
         ,this      ,SLOT(OnTypedInLeft(sTelex)) );

  sTelexField *               RightField = new sTelexField(this);
  L->addWidget(RightField);
  RightField->setSizePolicy(QSizePolicy::Preferred,QSizePolicy::Fixed);
  connect(RightField ,SIGNAL(TelexEntered(sTelex))
         ,this       ,SLOT(OnTypedInRight(sTelex)) );

  setMinimumHeight(5*Label->height());

  setEnabled(false);
}


void sTeleSpy::Show (literal color ,sTelex  T)
{
  if( T.isEmpty() ){
    Screen->append(QString::fromUtf8("<font color=")
                  +QString::fromUtf8(color)
                  +QString::fromUtf8(">∅</font>"));
  }else{
    Screen->append(QString::fromUtf8("<font color=")
                  +QString::fromUtf8(color)
                  +QString::fromUtf8(">")
                  +(QString)T
                  +QString::fromUtf8("</font>"));
  }
}


void sTeleSpy::ReportLeftTeleComError (QString  error_text)
{
  emit ErrorMessage( tr("Got error from '%1':\n\n%2")
                       .arg(ToQString(LeftTeleCom->DeviceIdentifier()))
                       .arg(error_text)     );
}


void sTeleSpy::ReportRightTeleComError (QString  error_text)
{
  emit ErrorMessage( tr("Got error from '%1':\n\n%2")
                       .arg(ToQString(RightTeleCom->DeviceIdentifier()))
                       .arg(error_text)     );
}


void sTeleSpy::LinkToLeftTeleCom ()
{
  try{
    LeftTeleCom.SetTo(DeviceLinks,"LeftTeleCom");
    if( ! LeftTeleCom.TakeExclusiveControl() ){
      emit ErrorMessage(tr("Cannot take exclusive control over '%1'. "
                           "Try again later.")
                       .arg(ToQString( LeftTeleCom->DeviceIdentifier() )) );
    }else{
      Connect(LeftTeleCom ,SIGNAL(Warmed()) ,SIGNAL(LeftTeleComLinked()) );
      Connect(LeftTeleCom ,SIGNAL(ErrorMessage(QString))
             ,SIGNAL(LeftTeleComError(QString)) );
    }
  }catch( rxDeviceLaunch E ){
    emit ErrorMessage( tr("Error while linking to a submissive device.\n"
                          "%1\nCheck configuration.")
                         .arg(ToQString(E.MsgText()))   );
  }
}


void sTeleSpy::LinkToRightTeleCom ()
{
  try{
    RightTeleCom.SetTo(DeviceLinks,"RightTeleCom");
    if( ! RightTeleCom.TakeExclusiveControl() ){
      emit ErrorMessage(tr("Cannot take exclusive control over '%1'. "
                           "Try again later.")
                       .arg(ToQString( RightTeleCom->DeviceIdentifier() )) );
    }else{
      Connect(RightTeleCom ,SIGNAL(Warmed()) ,SIGNAL(RightTeleComLinked()) );
      Connect(RightTeleCom ,SIGNAL(ErrorMessage(QString))
             ,SIGNAL(RightTeleComError(QString)) );
    }
  }catch( rxDeviceLaunch E ){
    emit ErrorMessage( tr("Error while linking to a submissive device.\n"
                          "%1\nCheck configuration.")
                         .arg(ToQString(E.MsgText()))   );
  }
}


void sTeleSpy::OnEnterToCoolingState ()
{
  Disconnect(LeftTeleCom,0,0);
  LeftTeleCom.ReleaseExclusiveControl();
  LeftTeleCom.SetToNULL();
  Disconnect(RightTeleCom,0,0);
  RightTeleCom.ReleaseExclusiveControl();
  RightTeleCom.SetToNULL();
  sConsole::OnEnterToCoolingState();
}


void sTeleSpy::OnEnterToOperatingState()
{
  sConsole::OnEnterToOperatingState();
  Connect(LeftTeleCom  ,SIGNAL(GotTelex(sTelex)) ,SLOT(OnGotFromLeft(sTelex)) );
  Connect(RightTeleCom ,SIGNAL(GotTelex(sTelex)) ,SLOT(OnGotFromRight(sTelex)));
  setEnabled(true);
}


void sTeleSpy::OnExitFromOperatingState()
{
  setEnabled(false);
  Disconnect(LeftTeleCom  ,SIGNAL(GotTelex(sTelex)) ,0 );
  Disconnect(RightTeleCom ,SIGNAL(GotTelex(sTelex)) ,0 );
  sConsole::OnExitFromOperatingState();
}


void sTeleSpy::OnGotFromLeft (sTelex  T)
{
  Show("maroon",T);
  SendDirective<sTelex>(RightTeleCom ,sTeleCom::cSendTelex ,T);
}


void sTeleSpy::OnGotFromRight (sTelex  T)
{
  Show("navy",T);
  SendDirective<sTelex>(LeftTeleCom ,sTeleCom::cSendTelex ,T);
}


void sTeleSpy::OnTypedInLeft (sTelex  T)
{
  Show("blue",T);
  SendDirective<sTelex>(LeftTeleCom ,sTeleCom::cSendTelex ,T);
}


void sTeleSpy::OnTypedInRight (sTelex  T)
{
  Show("red",T);
  SendDirective<sTelex>(RightTeleCom ,sTeleCom::cSendTelex ,T);
}

//------------------------------------------------------------------------------
} //namespace RA


