////////////////////////////////////////////////////////////////////////////////
/*! @file TeleType.cpp Teletype (a terminal for a telecommunication channel).
- Part of DevelopmentTools - Research Assistant Plugin package.
- Uses  RAGUI - Research Assistant Graphical User Interface Library.
- Uses  Qt v.5.0.2  - http://qt-project.org/
- Uses  RANet - Research Assistant Net Library.
- Copyright(C) 2013, Viktor E. Bursian, St.Petersburg, Russia.
                     Viktor.Bursian@mail.ioffe.ru
*///////////////////////////////////////////////////////////////////////////////
#include "TeleType.h"
#include "attr.h"
#include "StringsAndQStrings.h"
#include <QVBoxLayout>
#include <QScrollArea>
#include <QLabel>
#include <QMessageBox>
namespace RA {
//------------------------------------------------------------------------------

DEFINE_CLASS_TAG(sTeleType)

//-------------------------------------------------------------- sTelexField ---

sTelexField::sTelexField (QWidget *  parent)
    :QLineEdit(parent)
{
  connect(this ,SIGNAL(returnPressed()) ,this ,SLOT(OnEnterKeyPressed()));
}

void  sTelexField::OnEnterKeyPressed ()
{
  emit TelexEntered(sTelex(text()));
  setText("");
}

//---------------------------------------------------------------- sTeleType ---
/*! @class sTeleType

Позволяет отправлять сообщения в любой телекоммуникационный канал класса
sTeleCom и получать сообщения из него.  Отправляемые и получаемые сообщения
отображаются в сколлируемом окне.  Для отправки сообщения надо набрать его в
поле ввода и нажать Enter.

Его можно подключить к устройству класса sCOMPortTeleCom и тестировать внешнее
железо.

Можно использовать его и для тестирования драйверов, которые общаются с железом
через телекоммуникационный канал класса sTeleCom. Для этого надо создать
"трубу", т.е. инсталлировать два устройства "P1" и "P2" класса sTelePipe и
соединить "P1" с "P2", а "P2" с "P1" (надо соединить в обе стороны!). Затем
подключить тестируемый драйвер к одному концу "трубы" (скажем, к устройству
"P1"), а телетайп ко второму.  Теперь Ваши глаза и руки будуть работать вместо
железа.

*/

void  sTeleType::Preconfigure (sNodePtr  DeviceCfg)
{
  sPtr<sFolderNode>              DeviceLinks;
  DeviceCfg |= "Links";
  DeviceLinks = DeviceCfg >> "Links";
  if( DeviceLinks.IsCorrect() ){
    DeviceLinks |= "TeleCom";
    DeviceLinks >> "TeleCom" |= attr("Class","sTeleCom");
  };
  sPtr<sFolderNode>              DeviceSettings;
  DeviceCfg |= "Settings";
  DeviceSettings = DeviceCfg >> "Settings";
  if( DeviceSettings.IsCorrect() ){
    DeviceSettings |= attr("ControlExclusively",false);
    DeviceSettings |= attr("SendEchoOnPrefix",false);
    DeviceSettings |= attr("EchoPrefix","&rArr;");
    DeviceSettings |= attr("Irritable",false);
  };
  sConsole::Preconfigure(DeviceCfg);
}


sTeleType::~sTeleType()
{
  TeleCom.SetToNULL();
}


sTeleType::sTeleType (sString   DeviceId
                     ,sNodePtr  DeviceCfg
                     ,sNodePtr  WidgetCfg
                     ,QWidget *  parent    )
    :sConsole(DeviceId,DeviceCfg,WidgetCfg,parent)
    ,TeleCom(this)
    ,ControlExclusively(false)
    ,SendEchoOnPrefix(false)
    ,EchoPrefix("&rArr;")
    ,Irritable(false)
{
  try{ //get config parameters
    get_from<sBooleanNode>(DeviceSettings >> "SendEchoOnPrefix"   
                                             ,SendEchoOnPrefix);
    get_from<sTextNode>(DeviceSettings >> "EchoPrefix"   ,EchoPrefix);
    get_from<sBooleanNode>(DeviceSettings >> "Irritable"   ,Irritable);
  }catch(...){
    SetInvalid( tr("Something wrong in configuration.") );
  }
  //evolution of the inherited state machine (behaviour)
  {
    //evolution of the State_Warming
    {
      State_W_TeleCom 
          =( new sState(State_Warming()) )
          ->SetAsInitial()
          ->AddTransition(State_Warming()->Fin());
      //State_W_TeleCom internal structure and behaviour
      {
        State_W_TeleCom_Linking 
            =( new sState(State_W_TeleCom) )
            ->SetAsInitial()
            ->OnEnterCall(SLOT(LinkToTeleCom()))
            ->AddTransition(SIGNAL(TeleComLinked()) ,State_W_TeleCom->Fin())
            ->AddTransition(SIGNAL(ErrorMessage())  ,State_Cooling());
      }
    }

    State_Operating()
        ->AddReaction<sTelex>(SIGNAL(Got(sTelex)) ,sTelex("dura") 
                             ,SIGNAL(OffenceInTurn()) );
    
    connect(this ,SIGNAL(TeleComError(QString))
           ,this ,SLOT(ReportTeleComError(QString)) );
    connect(this ,SIGNAL(Got(sTelex))
           ,this ,SLOT(ShowNavy(sTelex)) );
    connect(this ,SIGNAL(OffenceInTurn())
           ,this ,SLOT(SaySamDurak()) );
  };
}


void  sTeleType::AppearanceMakeUp ()
{
  sConsole::AppearanceMakeUp();

  setLayout(new QVBoxLayout);

  Screen = new QTextEdit(this);
  Screen->setReadOnly(true);

  QScrollArea *               ScrollArea = new QScrollArea(this);
  layout()->addWidget(ScrollArea);
  ScrollArea->setWidget(Screen);
  ScrollArea->setWidgetResizable(true);

  QLabel *                    Label = new QLabel(this);
  Label->setText("red: typed, blue: received");
  layout()->addWidget(Label);
  Label->setSizePolicy(QSizePolicy::Preferred,QSizePolicy::Fixed);

  sTelexField *               Field = new sTelexField(this);
  layout()->addWidget(Field);
  Field->setSizePolicy(QSizePolicy::Preferred,QSizePolicy::Fixed);
  connect(Field ,SIGNAL(TelexEntered(sTelex)), this ,SLOT(OnSend(sTelex)) );

  setMinimumHeight(5*Label->height());

  setEnabled(false);
}


void sTeleType::ShowNavy (sTelex  T)
{
  if( T.isEmpty() ){
    Screen->append(QString::fromUtf8("<font color=navy>∅</font>"));
  }else{
    Screen->append(QString::fromUtf8("<font color=navy>")
                  +(QString)T
                  +QString::fromUtf8("</font>"));
  }
}


void sTeleType::ShowMaroon (sTelex  T)
{
  if( T.isEmpty() ){
    Screen->append(QString::fromUtf8("<font color=maroon>∅</font>"));
  }else{
    Screen->append(QString::fromUtf8("<font color=maroon>")
                  +(QString)T
                  +QString::fromUtf8("</font>"));
  }
}


void sTeleType::ReportTeleComError (QString  error_text)
{
  emit ErrorMessage( tr("Got error from '%1':\n\n%2")
                       .arg(ToQString(TeleCom->DeviceIdentifier()))
                       .arg(error_text)     );
}


void sTeleType::LinkToTeleCom ()
{
  try{
    TeleCom.SetTo(DeviceLinks,"TeleCom");
    if( ControlExclusively && ! TeleCom.TakeExclusiveControl() ){
      emit ErrorMessage(tr("Cannot take exclusive control over '%1'. "
                           "Try again later.")
                       .arg(ToQString( TeleCom->DeviceIdentifier() )) );
    }else{
      Connect(TeleCom ,SIGNAL(Warmed()) ,SIGNAL(TeleComLinked()) );
      Connect(TeleCom ,SIGNAL(ErrorMessage(QString))
             ,SIGNAL(TeleComError(QString)) );
    }
  }catch( rxDeviceLaunch E ){
    emit ErrorMessage( tr("Error while linking to a submissive device.\n"
                          "%1\nCheck configuration.")
                         .arg(ToQString(E.MsgText()))   );
  }
}


void sTeleType::OnEnterToCoolingState ()
{
  Disconnect(TeleCom,0,0);//all its signals connected to any of my slots
  TeleCom.ReleaseExclusiveControl();
  TeleCom.SetToNULL();
  sConsole::OnEnterToCoolingState();
}


void sTeleType::OnEnterToOperatingState()
{
  sConsole::OnEnterToOperatingState();
  Connect(TeleCom ,SIGNAL(GotTelex(sTelex)) ,SIGNAL(Got(sTelex)) );
  Connect(TeleCom ,SIGNAL(GotTelex(sTelex)) ,SLOT(SendEchoIfAsked(sTelex)) );
  setEnabled(true);
}


void sTeleType::OnExitFromOperatingState()
{
  setEnabled(false);
  Disconnect(TeleCom ,SIGNAL(GotTelex(sTelex)) ,SIGNAL(Got(sTelex)) );
  sConsole::OnExitFromOperatingState();
}


void sTeleType::OnSend (sTelex  T)
{
  ShowMaroon(T);
  SendDirective<sTelex>(TeleCom ,sTeleCom::cSendTelex ,T);
}


void  sTeleType::SendEchoIfAsked (sTelex  telex)
{
  if(  SendEchoOnPrefix ){
    if(  telex.startsWith(literal(EchoPrefix)) ){
      SendDirective(TeleCom ,sTeleCom::cSendTelex 
                   ,sTelex(telex.mid(EchoPrefix.Length())));
    }
  }
}


/*!
Если из канала приходит "dura"
и в конфигурации установлен флажок Irritable,
то автоматом ругается в ответ.
*/
void  sTeleType::SaySamDurak ()
{
  if( Irritable ){
    sTelex                      T ("Sam durak!");
    ShowMaroon(T);
    SendDirective<sTelex>(TeleCom ,sTeleCom::cSendTelex ,T);
  }
}

//------------------------------------------------------------------------------
} //namespace RA


