////////////////////////////////////////////////////////////////////////////////
/*! @file FSDminispectroDriver.cpp   Драйвер спектрометра FSD-10.
- Part of FSDminispectro - Research Assistant Plugin package.
- Uses  RAGUI - Research Assistant Graphical User Interface Library.
- Uses  Qt v.5.0.2  - http://qt-project.org/
- Uses  RANet - Research Assistant Net Library.
- Copyright(C) 2018, Dmitrii Nelson, St.Petersburg, Russia.
                          d.nelson@mail.ioffe.ru
*///////////////////////////////////////////////////////////////////////////////
#include "FSDminispectroDriver.h"
#include "attr.h"
#include <QDebug>
#include "Log.h"

namespace RA {
//------------------------------------------------------------------------------

DEFINE_CLASS_TAG(sFSDminispectroDriver)

//----------------------------------------------- sFSDminispectroDriver ----------


void  sFSDminispectroDriver::Preconfigure (sNodePtr  DeviceCfg)
{
  sPtr<sFolderNode>              DeviceSettings;
  DeviceCfg |= "Settings";
  DeviceSettings = DeviceCfg >> "Settings";
  if( DeviceSettings.IsCorrect() ){
    DeviceSettings |= attr("NumberOfPixels",PIXELSNUM);
    DeviceSettings |= attr("Gain",255);
    DeviceSettings |= attr("Shift",127);
    DeviceSettings |= attr("Pause",100);
  };
  sCCDDriver::Preconfigure(DeviceCfg);
}

sFSDminispectroDriver::~sFSDminispectroDriver ()
{
  DataArray.SetSize(0);
  Wavelengths.SetSize(0);
  FSD_CloseDevice();
  FSD_Lib_Close();
  if (ControlWindow != NULL) ControlWindow->close();
}

sFSDminispectroDriver::sFSDminispectroDriver (sString   DevicetId
                                         ,sNodePtr  DeviceCfg)
  :sCCDDriver(DevicetId,DeviceCfg)
{
}

void  sFSDminispectroDriver::PostConstructor ()
{
  sCCDDriver::PostConstructor();
  try{ //get config parameters
    get_from<sIntegerNode>(DeviceSettings >> "Gain"   ,Parameters.Gain);
    get_from<sIntegerNode>(DeviceSettings >> "Shift"   ,Shift);
    get_from<sIntegerNode>(DeviceSettings >> "Pause"   ,Pause);
    Pause = (Pause /100) * 100;
    if (Pause < 100) Pause = 100;
    Calibration[0] = 0;
    Calibration[1] = 1;
    Calibration[2] = 0;
  }catch(...){
    SetInvalid(tr("Something wrong in configuration."));
  }
  ControlWindow = new sWinAPI(this);
  ControlWindow->setGeometry(50,50,100,100);
  ControlWindow->setWindowState(Qt::WindowNoState);
  ControlWindow->setTitle("FSD Control");
  ControlWindow->setVisible(true);
  //evolution of the inherited state machine (behaviour)
  {
  };
}

void  sFSDminispectroDriver::ConnectUSBDevice ()
{
  HWND _hwnd = (HWND)(ControlWindow->winId());
  qDebug() << "HWND = " << _hwnd;
  FSD_Lib_CreateUSBInterface(_hwnd, 10, 22);
  qDebug() << "USBInterface Created";
  FSD_SetDeviceAplifier(240);//установим по умолчанию усиление сигнала
	FSD_SetDeviceShift(70);	//установим по умолчанию смещение сигнала относительно нуля АЦП
}

void sFSDminispectroDriver::GetCCDDevice (WPARAM wParam)
{
	QString name = QString((char*)wParam);
	if(name == NOCONNECT)
	{
    emit ErrorMessage(tr("Device is not connected."));
	}
	else
	{
    emit Connected();
	}
}

void  sFSDminispectroDriver::GetCCD_Data ()
{

}

void  sFSDminispectroDriver::GetFSD_Data (WPARAM wParam, LPARAM lParam)
{ // автоматическое получение данных CCD устройства
  // wParam -указатель на буфер 16 разрядных данных
  // lParam -размер полученных данных (должно быть 3694 -число пикселей CCD TDC1304)
	int *buffer = (int*) wParam;
	int i;
	for(i = 0;i < lParam; i++)
		DataArray.ArrBody[i] = buffer[i];

	// определим темновой сигнал CCD
	int dark_signal = 0;
	for(i=16; i<29; i++)//считываем заслоненные пикселы CCD (с 16 по 28 -см.документ TCD1304DG.pdf)
		dark_signal += DataArray.ArrBody[i];
	dark_signal /= 13;//-средний темновой сигнал
	// скорректируем полезный сигнал на величину темнового
	double corr = 1 + dark_signal / 65535.0;
	for(i = 0;i < lParam; i++) {
		DataArray.ArrBody[i] -= dark_signal;	// уменьшим полезный сигнал на величину сдвига
		DataArray.ArrBody[i] = (int)(DataArray.ArrBody[i] *corr); // скомпенсируем результат на всю шкалу графика
	}

  emit NewData(DataArray);
}

void  sFSDminispectroDriver::CalculateWavelengths ()
{
  for (int i = 0; i < NumOfPixels; i++) {
    Wavelengths.ArrBody[i] = Calibration[0] + Calibration[1]*(i+1)
            + Calibration[2]*(i+1)*(i+1);
  }
  emit NewWavelengths (Wavelengths);
}

void  sFSDminispectroDriver::StartCCD ()
{
  int _IT = Parameters.IntegrationTime / 100;
  if (_IT < 50) _IT = 50;
  FSD_GetDataSignal(1, _IT, Pause);
}

void  sFSDminispectroDriver::StopCCD ()
{
  FSD_StopScan();
}

void  sFSDminispectroDriver::SetCCD_Parameters (sCCDDriver::sParameters Params)
{
  Parameters =  Params;
  int _IT = Parameters.IntegrationTime / 100;
  if (_IT < 50) _IT = 50;
  bool Success = FSD_SetExposure(_IT);
  Success = Success && FSD_SetPause(Pause);
  Success = Success && FSD_SetDeviceAplifier(Parameters.Gain);
  Success = Success && FSD_SetDeviceShift(Shift);
  if( Success ){
    emit ParamsSet();
  } else
    emit ErrorMessage(tr("Error setting parameters."));
}


//------------------------------------------------------------------------------
} //namespace RA
