////////////////////////////////////////////////////////////////////////////////
/*! @file HamamatsuCCDDriver.cpp   Драйвер спектрометра Hamamatsu.
- Part of CCDDevices - Research Assistant Plugin package.
- Uses  RAGUI - Research Assistant Graphical User Interface Library.
- Uses  Qt v.5.0.2  - http://qt-project.org/
- Uses  RANet - Research Assistant Net Library.
- Copyright(C) 2017, Dmitrii Nelson, St.Petersburg, Russia.
                          d.nelson@mail.ioffe.ru
*///////////////////////////////////////////////////////////////////////////////
#include "HamamatsuCCDDriver.h"
#include "attr.h"
#include <QDebug>
#include "Log.h"

namespace RA {
//------------------------------------------------------------------------------

DEFINE_CLASS_TAG(sHamamatsuCCDDriver)

//----------------------------------------------- sHamamatsuCCDDriver ----------


void  sHamamatsuCCDDriver::Preconfigure (sNodePtr  DeviceCfg)
{
  sPtr<sFolderNode>              DeviceSettings;
  DeviceCfg |= "Settings";
  DeviceSettings = DeviceCfg >> "Settings";
  if( DeviceSettings.IsCorrect() ){
    DeviceSettings |= attr("NumberOfPixels",PIXELSNUM);
  };
  sCCDDriver::Preconfigure(DeviceCfg);
}

sHamamatsuCCDDriver::~sHamamatsuCCDDriver ()
{
  if (PipeHandle != INVALID_HANDLE_VALUE) USB_ClosePipe(PipeHandle);
  if( DeviceHandle != INVALID_HANDLE_VALUE ) USB_CloseDevice(DeviceHandle);
  DataArray.SetSize(0);
  Wavelengths.SetSize(0);
}

sHamamatsuCCDDriver::sHamamatsuCCDDriver (sString   DevicetId
                                         ,sNodePtr  DeviceCfg)
  :sCCDDriver(DevicetId,DeviceCfg)
  ,DeviceHandle(INVALID_HANDLE_VALUE)
  ,PipeHandle(INVALID_HANDLE_VALUE)
{
}

void  sHamamatsuCCDDriver::ConnectUSBDevice ()
{
  DeviceHandle = USB_OpenDevice(PRODUCT_ID);
  if( DeviceHandle == INVALID_HANDLE_VALUE ){
    emit ErrorMessage (tr("Error opening USB device."));
    return;
  };
  PipeHandle = USB_OpenPipe(DeviceHandle);
  if (PipeHandle == INVALID_HANDLE_VALUE) {
    emit ErrorMessage (tr("Error opening the pipe for USB device."));
    USB_CloseDevice(DeviceHandle);
    return;
  };
  ushort nResult = USB_ReadCalibrationValue(DeviceHandle, Calibration);
  if (nResult != USBDEV_SUCCESS) {
    emit ErrorMessage (tr("Error reading calibration data."));
    USB_CloseDevice(DeviceHandle);
    return;
  };
  emit Connected();
}

void  sHamamatsuCCDDriver::GetCCD_Data ()
{
  ushort nResult = USB_GetSensorData(DeviceHandle, PipeHandle
                                    ,NumOfPixels, Buffer);
  if( nResult == USBDEV_SUCCESS ){
    for(int i = 0;i < NumOfPixels; i ++) {
      DataArray.ArrBody[i] = Buffer[i];
    }
  }else{
    if( !(nResult == USBDEV_BULK_NOT_UPDATED) ){
      emit ErrorMessage(tr("Error communicating USB device."));
      emit CoolDown();
    }
  }
  emit NewData(DataArray);
}

void  sHamamatsuCCDDriver::CalculateWavelengths ()
{
  for (int i = 0; i < NumOfPixels; i++) {
    Wavelengths.ArrBody[i] = Calibration[0] + Calibration[1]*(i+1)
            + Calibration[2]*(i+1)*(i+1) + Calibration[3]*(i+1)*(i+1)*(i+1)
            + Calibration[4]*(i+1)*(i+1)*(i+1)*(i+1)
            + Calibration[5]*(i+1)*(i+1)*(i+1)*(i+1)*(i+1);
  }
  qDebug() << "sDummyCCDDriver::CalculateWavelengths";
  emit NewWavelengths (Wavelengths);
}

void  sHamamatsuCCDDriver::StartCCD ()
{
  try{
    int _IT = Round((Parameters.IntegrationTime/(sPhysValue(1.0e-3,"s"))).ToReal());
    if (_IT < 1) _IT = 1;
    State_O_Measuring->SetTimeout(_IT);
  }catch(rxRAlgebra){
    SetInvalid(tr("Something wrong with units."));
  }
}

void  sHamamatsuCCDDriver::StopCCD ()
{
}

void  sHamamatsuCCDDriver::SetCCD_Parameters (sCCDDriver::sParameters Params)
{
  UNIT_PARAMETER  Pars;
  Parameters =  Params;
  try{
    Pars.IntegrationTime = Round((Parameters.IntegrationTime/(sPhysValue(1.0e-6,"s"))).ToReal());
    if (Pars.IntegrationTime < 1) Pars.IntegrationTime = 1;
  }catch(rxRAlgebra){
    SetInvalid(tr("Something wrong with units."));
  }
  Pars.TriggerEdge = Parameters.TriggerEdge;
  Pars.TriggerMode = Parameters.TriggerMode;
  Pars.Gain = NO_FUNCTION;
  ushort nResult = USB_SetParameter(DeviceHandle, &Pars);
  if( nResult != USBDEV_SUCCESS ){
    emit ErrorMessage(tr("Error setting parameters."));
  } else
    emit ParamsSet();
}


//------------------------------------------------------------------------------
} //namespace RA
