////////////////////////////////////////////////////////////////////////////////
/*! @file Cubic001Film.cpp Formulae for magnetization of a cubic (001) film.
- Part of MagnetizationCalc - Research Assistant Plugin package.
- Uses  RAGUI - Research Assistant Graphical User Interface.
- Uses  RANet - Research Assistant Net Library.
- Uses  Qt v.5  - http://qt-project.org/
- Copyright(C) 2017, Viktor E. Bursian, St.Petersburg, Russia.
                     Viktor_dot_Bursian_at_mail_dot_ioffe_dot_ru
*///////////////////////////////////////////////////////////////////////////////
#include "Cubic001Film.h"
#include "BasicMath.h"
#include "attr.h"
//#include "Log.h"
//#include "Trace.h"
#include <math.h>
namespace RA {
//------------------------------------------------------------------------------

//------------------------------------------------------------------------------
/*
U=
  -H*Mₛ*(cos(η)*cos(θ)+sin(η)*sin(θ)*cos(ξ)*cos(φ)+sin(η)*sin(θ)*sin(ξ)*sin(φ))
  +2*%pi*Mₛ^2*cos(θ)^2
  +K₁*(sin(θ)^4*cos(φ)^2*sin(φ)^2
      +cos(θ)^2*sin(θ)^2*sin(φ)^2
      +cos(θ)^2*sin(θ)^2*cos(φ)^2)
  +Kᵤ*sin(θ)^2
*/
double  Cubic001Film_U_polar (const gsl_vector *  theta_phi
                             ,void *              params)
{
  double                      theta = gsl_vector_get(theta_phi,0);
  double                      phi   = gsl_vector_get(theta_phi,1);
  psCubic001FilmParams        P = (psCubic001FilmParams)params;
  double                      sin_eta = std::sin(P->eta);
  double                      cos_eta = std::cos(P->eta);
  double                      sin_xi = std::sin(P->xi);
  double                      cos_xi = std::cos(P->xi);
  double                      sin_theta = std::sin(theta);
  double                      cos_theta = std::cos(theta);
  double                      sin_phi = std::sin(phi);
  double                      cos_phi = std::cos(phi);
  double                      sin_theta_2 = sin_theta*sin_theta;
  double                      cos_theta_2 = cos_theta*cos_theta;
  double                      sin_phi_2 = sin_phi*sin_phi;
  double                      cos_phi_2 = cos_phi*cos_phi;
  double                      sin_theta_4 = sin_theta_2*sin_theta_2;
  return
      - P->H0 * P->Ms *  ( sin_eta*sin_theta*cos_xi*cos_phi
                        + sin_eta*sin_theta*sin_xi*sin_phi
                        + cos_eta*cos_theta
                        )
      + 2 * sMath::pi * P->Ms*P->Ms * cos_theta*cos_theta
      + P->kc1*P->Ms *  ( sin_theta_4 * cos_phi_2 * sin_phi_2
                  + cos_theta_2 * sin_theta_2 * sin_phi_2
                  + cos_theta_2 * sin_theta_2 * cos_phi_2
                  )
      + P->ku1*P->Ms * sin_theta_2;
}


/*
dU/dθ=
-H₀*Mₛ*(-cos(η)*sin(θ)+sin(η)*cos(θ)*cos(ξ)*cos(φ)+sin(η)*cos(θ)*sin(ξ)*sin(φ))
-4*%pi*Mₛ^2*cos(θ)*sin(θ)
+K₁*(4*cos(θ)*sin(θ)^3*cos(φ)^2*sin(φ)^2
    -2*cos(θ)*sin(θ)^3*sin(φ)^2
    +2*cos(θ)^3*sin(θ)*sin(φ)^2
    -2*cos(θ)*sin(θ)^3*cos(φ)^2
    +2*cos(θ)^3*sin(θ)*cos(φ)^2)
+2*Kᵤ*cos(θ)*sin(θ)

dU/dφ=
-H₀*Mₛ*(sin(η)*sin(θ)*sin(ξ)*cos(φ)-sin(η)*sin(θ)*cos(ξ)*sin(φ))
+0
+K₁*(2*sin(θ)^4*cos(φ)^3*sin(φ)
    -2*sin(θ)^4*cos(φ)*sin(φ)^3)
+0

*/
void  Cubic001Film_dU_polar (const gsl_vector *  theta_phi
                            ,void *              params
                            ,gsl_vector *        gradient
                            )
{
  double                      theta = gsl_vector_get(theta_phi,0);
  double                      phi   = gsl_vector_get(theta_phi,1);
  psCubic001FilmParams        P = (psCubic001FilmParams)params;
  double                      sin_eta = std::sin(P->eta);
  double                      cos_eta = std::cos(P->eta);
  double                      sin_xi = std::sin(P->xi);
  double                      cos_xi = std::cos(P->xi);
  double                      sin_theta = std::sin(theta);
  double                      cos_theta = std::cos(theta);
  double                      sin_phi = std::sin(phi);
  double                      cos_phi = std::cos(phi);
  double                      sin_theta_2 = sin_theta*sin_theta;
  double                      cos_theta_2 = cos_theta*cos_theta;
  double                      sin_phi_2 = sin_phi*sin_phi;
  double                      cos_phi_2 = cos_phi*cos_phi;
  double                      sin_theta_3 = sin_theta*sin_theta_2;
  double                      cos_theta_3 = cos_theta*cos_theta_2;
  double                      sin_phi_3 = sin_phi*sin_phi_2;
  double                      cos_phi_3 = cos_phi*cos_phi_2;
  double                      sin_theta_4 = sin_theta_2*sin_theta_2;

  gsl_vector_set(gradient,0, // dU/dθ
      - P->H0 * P->Ms * ( sin_eta*cos_theta*cos_xi*cos_phi
                        + sin_eta*cos_theta*sin_xi*sin_phi
                        - cos_eta*sin_theta
                        )
      - 4*sMath::pi*P->Ms*P->Ms* cos_theta*sin_theta
      + P->kc1*P->Ms * (  4*cos_theta*sin_theta_3*cos_phi_2*sin_phi_2
                        - 2*cos_theta*sin_theta_3*sin_phi_2
                        + 2*cos_theta_3*sin_theta*sin_phi_2
                        - 2*cos_theta*sin_theta_3*cos_phi_2
                        + 2*cos_theta_3*sin_theta*cos_phi_2
                       )
      + 2 * P->ku1*P->Ms * cos_theta*sin_theta
      );
  gsl_vector_set(gradient,1, // dU/dφ
      - P->H0 * P->Ms * (-sin_eta*sin_theta*cos_xi*sin_phi
                         +sin_eta*sin_theta*sin_xi*cos_phi
                        )
      + P->kc1*P->Ms * ( 2*sin_theta_4*cos_phi_3*sin_phi
                        -2*sin_theta_4*cos_phi*sin_phi_3
                       )
      );
}


void  Cubic001Film_UdU_polar (const gsl_vector *  theta_phi
                             ,void *              params
                             ,double *            energy
                             ,gsl_vector *        gradient
                             )
{
  *energy = Cubic001Film_U_polar(theta_phi,params);
  Cubic001Film_dU_polar(theta_phi,params,gradient);
}


/*
ddU/dθ/dθ=
-H₀*Mₛ*(-cos(η)*cos(θ)-sin(η)*sin(θ)*cos(ξ)*cos(φ)-sin(η)*sin(θ)*sin(ξ)*sin(φ))
+4*%pi*Mₛ^2*(sin(θ)^2-cos(θ)^2)
+K₁*( -4*sin(θ)^4*cos(φ)^2*sin(φ)^2
      +12*cos(θ)^2*sin(θ)^2*cos(φ)^2*sin(φ)^2
      +2*sin(θ)^4*sin(φ)^2
      -12*cos(θ)^2*sin(θ)^2*sin(φ)^2
      +2*cos(θ)^4*sin(φ)^2
      +2*sin(θ)^4*cos(φ)^2
      -12*cos(θ)^2*sin(θ)^2*cos(φ)^2
      +2*cos(θ)^4*cos(φ)^2)
+2*Kᵤ*(cos(θ)^2-sin(θ)^2)

ddU/dφ/dφ=
-H₀*Mₛ*(-sin(η)*sin(θ)*cos(ξ)*cos(φ)-sin(η)*sin(θ)*sin(ξ)*sin(φ))
+0
+K₁*(2*sin(θ)^4*sin(φ)^4
    -12*sin(θ)^4*cos(φ)^2*sin(φ)^2
    +2*sin(θ)^4*cos(φ)^4)
+0

ddU/dφ/dθ=
-H₀*Mₛ*(sin(η)*cos(θ)*sin(ξ)*cos(φ)-sin(η)*cos(θ)*cos(ξ)*sin(φ))
+0
+K₁*(8*cos(θ)*sin(θ)^3*cos(φ)^3*sin(φ)-8*cos(θ)*sin(θ)^3*cos(φ)*sin(φ)^3)
+0
*/
int  Cubic001Film_ddU_polar (const gsl_vector *  theta_phi
                            ,void *              params
                            ,gsl_matrix *        hessian
                            )
{
  double                      theta = gsl_vector_get(theta_phi,0);
  double                      phi   = gsl_vector_get(theta_phi,1);
  psCubic001FilmParams        P = (psCubic001FilmParams)params;
  double                      sin_eta = std::sin(P->eta);
  double                      cos_eta = std::cos(P->eta);
  double                      sin_xi = std::sin(P->xi);
  double                      cos_xi = std::cos(P->xi);
  double                      sin_theta = std::sin(theta);
  double                      cos_theta = std::cos(theta);
  double                      sin_phi = std::sin(phi);
  double                      cos_phi = std::cos(phi);
  double                      sin_theta_2 = sin_theta*sin_theta;
  double                      cos_theta_2 = cos_theta*cos_theta;
  double                      sin_phi_2 = sin_phi*sin_phi;
  double                      cos_phi_2 = cos_phi*cos_phi;
  double                      sin_theta_3 = sin_theta*sin_theta_2;
//  double                      cos_theta_3 = cos_theta*cos_theta_2;
  double                      sin_phi_3 = sin_phi*sin_phi_2;
  double                      cos_phi_3 = cos_phi*cos_phi_2;
  double                      sin_theta_4 = sin_theta_2*sin_theta_2;
  double                      cos_theta_4 = cos_theta_2*cos_theta_2;
  double                      sin_phi_4 = sin_phi_2*sin_phi_2;
  double                      cos_phi_4 = cos_phi_2*cos_phi_2;

  gsl_matrix_set(hessian,0,0, // ddU/dθ/dθ
      -P->H0*P->Ms*(-cos_eta*cos_theta
              -sin_eta*sin_theta*cos_xi*cos_phi
              -sin_eta*sin_theta*sin_xi*sin_phi)
      +4*sMath::pi*P->Ms*P->Ms*(sin_theta_2-cos_theta_2)
      +P->kc1*P->Ms*( -4*sin_theta_4*cos_phi_2*sin_phi_2
            +12*cos_theta_2*sin_theta_2*cos_phi_2*sin_phi_2
            +2*sin_theta_4*sin_phi_2
            -12*cos_theta_2*sin_theta_2*sin_phi_2
            +2*cos_theta_4*sin_phi_2
            +2*sin_theta_4*cos_phi_2
            -12*cos_theta_2*sin_theta_2*cos_phi_2
            +2*cos_theta_4*cos_phi_2)
      +2*P->ku1*P->Ms*(cos_theta_2-sin_theta_2)
                );
  gsl_matrix_set(hessian,1,1, // ddU/dφ/dφ
      -P->H0*P->Ms*(-sin_eta*sin_theta*cos_xi*cos_phi
                    -sin_eta*sin_theta*sin_xi*sin_phi)
      +P->kc1*P->Ms*(2*sin_theta_4*sin_phi_4
              -12*sin_theta_4*cos_phi_2*sin_phi_2
              +2*sin_theta_4*cos_phi_4)
                );
  gsl_matrix_set(hessian,0,1, // ddU/dφ/dθ
      -P->H0*P->Ms*(sin_eta*cos_theta*sin_xi*cos_phi
                   -sin_eta*cos_theta*cos_xi*sin_phi)
      +P->kc1*P->Ms*(8*cos_theta*sin_theta_3*cos_phi_3*sin_phi
              -8*cos_theta*sin_theta_3*cos_phi*sin_phi_3)
                );
  gsl_matrix_set(hessian,1,0, gsl_matrix_get(hessian,0,1) );
  return GSL_SUCCESS;
}

//------------------------------------------------------------------------------

//!@patch not implemented: calculations in Cartesian system
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wunused-variable"
#pragma GCC diagnostic ignored "-Wuninitialized"
double  Cubic001Film_U_xy (const gsl_vector *  mxmy
                          ,void *              params)
{
  double                      mx = gsl_vector_get(mxmy,0);
  double                      my = gsl_vector_get(mxmy,1);
  psCubic001FilmParams        P = (psCubic001FilmParams)params;
  double                      U;

  return U;
}


void  Cubic001Film_dU_xy (const gsl_vector *  mxmy
                         ,void *              params
                         ,gsl_vector *        gradient
                         )
{
  double                      mx = gsl_vector_get(mxmy,0);
  double                      my = gsl_vector_get(mxmy,1);
  psCubic001FilmParams        P = (psCubic001FilmParams)params;
  double                      dUdmx;
  double                      dUdmy;

  gsl_vector_set(gradient,0, dUdmx );
  gsl_vector_set(gradient,1, dUdmy );
}
#pragma GCC diagnostic pop


void  Cubic001Film_UdU_xy (const gsl_vector *  mxmy
                          ,void *              params
                          ,double *            energy
                          ,gsl_vector *        gradient
                          )
{
  *energy = Cubic001Film_U_xy(mxmy,params);
  Cubic001Film_dU_xy(mxmy,params,gradient);
}


//------------------------------------------------------ sCubic001FilmParams ---

sCubic001FilmParams::sCubic001FilmParams ()
    :sMagnetizationParams()
    ,_kc1(-240,0.5,"G")
    ,_ku1(0,0.5,"G")
{
  U_polar   = & Cubic001Film_U_polar;
  dU_polar  = & Cubic001Film_dU_polar;
  UdU_polar = & Cubic001Film_UdU_polar;
  ddU_polar = & Cubic001Film_ddU_polar;
  kc1 = (_kc1/sPhysValue("G")).ToReal();
  ku1 = (_ku1/sPhysValue("G")).ToReal();
}


psCubic001FilmParams  sCubic001FilmParams::Set_kc1 (rcsPhysValue  value)
{
  _kc1 = value;
  kc1 = (_kc1/sPhysValue("G")).ToReal();
  return this;
}


psCubic001FilmParams  sCubic001FilmParams::Set_ku1 (rcsPhysValue  value)
{
  _ku1 = value;
  ku1 = (_ku1/sPhysValue("G")).ToReal();
  return this;
}


void  sCubic001FilmParams::ProvideAttributesTo (sPtr<sFolderNode>  folder)
{
  folder << attr("4&pi;M<sub>s</sub>",_4piMs())
         << attr("k<sub>c1</sub>",_kc1)
         << attr("k<sub>u1</sub>",_ku1);
//  sMagnetizationParams::ProvideAttributesTo(folder);
}

//------------------------------------------------------------------------------
} //namespace RA


