////////////////////////////////////////////////////////////////////////////////
/*! @file Magneto.cpp Common things for magnetization and its effects.
- Part of MagnetizationCalc - Research Assistant Plugin package.
- Uses  RANet - Research Assistant Net Library.
- Copyright(C) 2017, Viktor E. Bursian, St.Petersburg, Russia.
                     Viktor.Bursian@mail.ioffe.ru
*///////////////////////////////////////////////////////////////////////////////
#include "Magneto.h"
//#include "Cubic001Film.h"
#include "SymbolsGUI.h"
#include "attr.h"
#include "Log.h"
#include <gsl/gsl_multimin.h>
#include <gsl/gsl_roots.h>
namespace RA {
//------------------------------------------------------------------------------

const int                     xMinimizationError::cNotAMinimum = 1000;
std::map<sMagnetoTask::eTaskType
        ,sMagnetoTask>        sMagnetoTask::Tasks;

//------------------------------------------------------------------------------

sStaticSolution  StaticSolution (rcsPoint               initial_point
                                ,psMagnetizationParams  params
                                )
{
  size_t                      MaxIterationNumber = 100;
  double                      step_size = 1e-3;
  double                      tolerance = 1e-7;
  sStaticSolution                         Result;
  const gsl_multimin_fdfminimizer_type *  MinimizerType;
  gsl_multimin_fdfminimizer *             Minimizer;
  gsl_vector *                            MinimizerPos;
  gsl_multimin_function_fdf               MinimizerFunc;
  size_t                                  Iteration = 0;
  int                                     Status;

  MinimizerFunc.n = 2;
  MinimizerFunc.f =   params->U_polar;
  MinimizerFunc.df =  params->dU_polar;
  MinimizerFunc.fdf = params->UdU_polar;
  MinimizerFunc.params = params;

  MinimizerPos = gsl_vector_alloc (2);
  gsl_vector_set(MinimizerPos, 0, initial_point.X);
  gsl_vector_set(MinimizerPos, 1, initial_point.Y);

//  MinimizerType = gsl_multimin_fdfminimizer_steepest_descent;
//  MinimizerType = gsl_multimin_fdfminimizer_conjugate_pr;
  MinimizerType = gsl_multimin_fdfminimizer_conjugate_fr;
//  MinimizerType = gsl_multimin_fdfminimizer_vector_bfgs;
//  MinimizerType = gsl_multimin_fdfminimizer_vector_bfgs2;

  Minimizer = gsl_multimin_fdfminimizer_alloc (MinimizerType, 2);
  gsl_multimin_fdfminimizer_set(Minimizer ,&MinimizerFunc ,MinimizerPos
                               ,step_size ,tolerance );
  do{
      Iteration++;
      Status = gsl_multimin_fdfminimizer_iterate (Minimizer);
      if( Status ){
        throw xMinimizationError(Status);
      }
      Status = gsl_multimin_test_gradient (Minimizer->gradient, 1e-3);
//      if( Status == GSL_SUCCESS )
//        printf ("Minimum found at:\n");
//      printf ("%5d %.5f %.5f %10.5f\n", Iteration,
//              gsl_vector_get (Minimizer->x, 0),
//              gsl_vector_get (Minimizer->x, 1),
//              Minimizer->f);
  }while( Status == GSL_CONTINUE && Iteration < MaxIterationNumber );
  if( Status ){
    throw xMinimizationError(Status);
  }
  if( ! params->IsMinimum_polar(Minimizer->x) ){
    throw xMinimizationError(xMinimizationError::cNotAMinimum);
  }

  Result.theta = gsl_vector_get (Minimizer->x, 0);
  Result.phi =   gsl_vector_get (Minimizer->x, 1);
  Result.energy = Minimizer->f;
  Result.Normalize();

  gsl_multimin_fdfminimizer_free (Minimizer);
  gsl_vector_free (MinimizerPos);
  return Result;
}


std::list<sStaticSolution>  StaticSolutions(psMagnetizationParams  params)
{
  int                         N = 5;
  int                         M = 16;
  real                        AngleRounding = 1e-3;
  std::list<sStaticSolution>  Solutions;
  sStaticSolution             S;
  for( int I = 0 ; I < N ; ++I ){
    real                        theta0 = sMath::pi/N*(0.5+I);
    for( int J = 0 ; J < M ; ++J ){
      real                        phi0 = (2*sMath::pi/M)*J;
      try{
        S = StaticSolution(sPoint(theta0,phi0),params);
        bool                        Found = false;
        for( std::list<sStaticSolution>::iterator K = Solutions.begin()
                                                ; K != Solutions.end()  ; ++K ){
          real                        Delta_theta = std::abs(S.theta-K->theta);
          real                        Delta_phi = std::abs(S.phi-K->phi);
          while( Delta_phi > 2*sMath::pi )  Delta_phi -= 2*sMath::pi;
          if( Delta_theta < AngleRounding ){
            if( Delta_phi < AngleRounding ){
              if( S.energy < K->energy ){
                Solutions.erase(K);
              }else{
                Found = true;
              }
              break;
            }else if( std::abs(S.theta) < AngleRounding ){
              if( (S.energy < K->energy)
                  || (std::abs(S.phi-params->xi) < std::abs(K->phi-params->xi)) ){
                Solutions.erase(K);
              }else{
                Found = true;
              }
              break;
            }
          }
        }
        if( ! Found ){
          Solutions.push_back(S);
        }
      }catch( rxMinimizationError E ){
//        RANet::Log.Put(sLog::Conditionless,"Magneto"
//            ,sString("StaticSolutions: ")+E.Title+": "<<E.Code);
      }
    }
  }
  return Solutions;
}


double FMR_ResEq (double H, void * params)
{
  psMagnetizationParams       P = (psMagnetizationParams)params;
  double                      gamma = 17.6 / 2.0 * P->g;
  std::list<sStaticSolution>  Solutions;
  sStaticSolution             S;
  double                      theta;
  gsl_vector *                theta_phi = gsl_vector_alloc(2);
  gsl_matrix *                Hessian = gsl_matrix_alloc(2,2);
  int                         Status;
  double                      D;
  ASSERT( P->ddU_polar );
  P->H0 = H;
  Solutions = StaticSolutions(P);
  if( Solutions.empty() ){
    throw xMinimizationError(1001);
  }
  Solutions.sort();
  S = Solutions.front();
  theta = rmax( S.theta , 1e-6 );
  gsl_vector_set(theta_phi,0 ,theta);
  gsl_vector_set(theta_phi,1 ,S.phi);
  Status = (*(P->ddU_polar))(theta_phi,P,Hessian);
  if( Status ){
    throw xMinimizationError(Status+2000);
  }
  D =  gsl_matrix_get(Hessian,0,0)*gsl_matrix_get(Hessian,1,1)
      -gsl_matrix_get(Hessian,0,1)*gsl_matrix_get(Hessian,0,1);
  if( D < 0 ){
    throw xMinimizationError(1002);
  }
  return gamma / ( 2* sMath::pi * P->Ms * std::sin(theta) ) * std::sqrt(D)
       - P->f0;
}

real  FMR_Hres (psMagnetizationParams  params)
{
  int                         MaxIterationNumber = 100;
  int                            Status;
  const gsl_root_fsolver_type *  RootFinderType;
  gsl_root_fsolver *             RootFinder;
  double r = 3333.0;
  double x_lo = 100.0, x_hi = 20.0e3;
  gsl_function                   RootFinderFunc;
  int                            Iteration = 0;

  RootFinderFunc.function = &FMR_ResEq;
  RootFinderFunc.params = params;

  RootFinderType = gsl_root_fsolver_bisection;
//  RootFinderType = gsl_root_fsolver_brent;
//  RootFinderType = gsl_root_fsolver_falsepos;

  RootFinder = gsl_root_fsolver_alloc (RootFinderType);
  gsl_root_fsolver_set (RootFinder, &RootFinderFunc, x_lo, x_hi);

//  printf ("using %s method\n", gsl_root_fsolver_name (RootFinder));
//  printf ("%5s [%9s, %9s] %9s  %9s\n",
//          "iter", "lower", "upper", "root", "err(est)");
  do{
    Iteration++;
    Status = gsl_root_fsolver_iterate(RootFinder);
    if( Status ){
      throw xMinimizationError(Status);
    }
    r = gsl_root_fsolver_root(RootFinder);
    x_lo = gsl_root_fsolver_x_lower(RootFinder);
    x_hi = gsl_root_fsolver_x_upper(RootFinder);
    Status = gsl_root_test_interval (x_lo, x_hi, 0, 0.001);
//    if (Status == GSL_SUCCESS)
//      printf ("Converged:\n");
//    printf ("%5d [%.7f, %.7f] %.7f %.7f\n",
//            Iteration, x_lo, x_hi, r, x_hi - x_lo);
  }while( (Status == GSL_CONTINUE) && (Iteration < MaxIterationNumber) );
  gsl_root_fsolver_free (RootFinder);
  if( Status ){
    throw xMinimizationError(Status);
  }
  return r;
}


sAttribute  M_AllSolutions (psMagnetizationParams  params)
{
  sPtr<sRecord>               Record(new sRecord());
//  RANet::Log.Put(sLog::Conditionless,"Magneto"
//      ,"---------------------------------------------------");
  std::list<sStaticSolution>  Solutions;
  integer                     I = 0;
  Solutions = StaticSolutions(params);
//  RANet::Log.Put(sLog::Conditionless,"Magneto","Done");

  Record->SetFinishTime();
  Record->AllocateRecord();
  (Record >= folder("pars",""))
      << attr("H<sub>0</sub>",params->_H0)
      << attr("&eta;",params->_eta)
      << attr("&xi;",params->_xi);
  params->ProvideAttributesTo(Record >= folder("pars",""));
  for( std::list<sStaticSolution>::iterator K = Solutions.begin()
                                          ; K != Solutions.end()  ; ++K ){
    (Record >= folder(++I,""))
        << attr("&theta;",sPhysValue(K->theta_degree(),Symbol::Degree))
        << attr("&phi;",sPhysValue(K->phi_degree(),Symbol::Degree))
        << attr("&epsilon;",sPhysValue(K->energy,"G2"));
  }
  return sAttribute(sAttrName(sString("M (all solutions) at")
                             +" H<sub>0</sub>=" + params->_H0.Text()
                             +" &eta;="         + params->_eta.Text()
                             +" &xi;="          + params->_xi.Text())
                   ,Record);
}


sAttribute  M_vs_H (psMagnetizationParams  params)
{
  real                        _4piMs = 4*sMath::pi*params->Ms;
  real                        Hmax = 1.5 * _4piMs;
  real                        Hmin = -Hmax;
  real                        Hstep = (Hmax-Hmin)/200;
  sPtr<sRecord>               Record(new sRecord());
  sPtr<sDataListFolder>       Mx;
  sPtr<sDataListFolder>       My;
  sPtr<sDataListFolder>       Mz;
  bool                        HasOther = false;
  sPtr<sDataListFolder>       Mx_other;
  sPtr<sDataListFolder>       My_other;
  sPtr<sDataListFolder>       Mz_other;
  std::list<sStaticSolution>  Solutions;
  Mx=sNodePtr(new sDataListFolder(sPhysPair(sPhysValue("G"),sPhysValue("G"))));
  My=sNodePtr(new sDataListFolder(sPhysPair(sPhysValue("G"),sPhysValue("G"))));
  Mz=sNodePtr(new sDataListFolder(sPhysPair(sPhysValue("G"),sPhysValue("G"))));
  for( real H = Hmin ; H < Hmax+Hstep ; H += ( std::abs(H)>20*Hstep
                                             ? Hstep
                                             : ( std::abs(H)>10*Hstep
                                               ? Hstep/4
                                               : Hstep/16 ) ) ){
    params->Set_H0(sPhysValue(H,"G"));
    Solutions = StaticSolutions(params);
    Solutions.sort();
    int I = 0;
    for( std::list<sStaticSolution>::iterator K = Solutions.begin()
                                            ; K != Solutions.end()  ; ++K ){
      if( I == 0 ){
        Mx->AddNewPoint(H , _4piMs * std::sin(K->theta) * std::cos(K->phi) );
        My->AddNewPoint(H , _4piMs * std::sin(K->theta) * std::sin(K->phi) );
        Mz->AddNewPoint(H , _4piMs * std::cos(K->theta) );
      }else{
        if( ! HasOther ){
          HasOther = true;
          Mx_other=sNodePtr(new sDataListFolder(sPhysPair(sPhysValue("G")
                                                         ,sPhysValue("G"))));
          My_other=sNodePtr(new sDataListFolder(sPhysPair(sPhysValue("G")
                                                         ,sPhysValue("G"))));
          Mz_other=sNodePtr(new sDataListFolder(sPhysPair(sPhysValue("G")
                                                         ,sPhysValue("G"))));
        }
        Mx_other->AddNewPoint(H , _4piMs*std::sin(K->theta)*std::cos(K->phi) );
        My_other->AddNewPoint(H , _4piMs*std::sin(K->theta)*std::sin(K->phi) );
        Mz_other->AddNewPoint(H , _4piMs*std::cos(K->theta) );
      }
      ++I;
    }
  }

  Record->SetFinishTime();
  Record->AllocateRecord();
  (Record >= folder("pars",""))
//      << attr("H<sub>0</sub>",H0_Field->Value())
      << attr("&eta;",params->_eta)
      << attr("&xi;",params->_xi);
  params->ProvideAttributesTo(Record >= folder("pars",""));
  Mx->SetAppearance(new sCurveAppearance(sColor::cBlue ,2));
  My->SetAppearance(new sCurveAppearance(sColor::cGreen,2));
  Mz->SetAppearance(new sCurveAppearance(sColor::cRed  ,2));
  Record << attr("4&pi;M<sub>x</sub>",Mx)
         << attr("4&pi;M<sub>y</sub>",My)
         << attr("4&pi;M<sub>z</sub>",Mz);
  if( HasOther ){
    Mx_other->SetAppearance(new sCurveAppearance(3,sColor::cBlue ));
    My_other->SetAppearance(new sCurveAppearance(3,sColor::cGreen));
    Mz_other->SetAppearance(new sCurveAppearance(3,sColor::cRed  ));
    Mx << attr("4&pi;M<sub>x</sub> other",Mx_other);
    My << attr("4&pi;M<sub>y</sub> other",My_other);
    Mz << attr("4&pi;M<sub>z</sub> other",Mz_other);
  }
  return sAttribute(sAttrName(sString("M(H) at &eta;=")+params->_eta.Text()
                             +" &xi;=" +params->_xi.Text())
                   ,Record);
}


sAttribute  FMR_Hres_vs_xi (psMagnetizationParams  params)
{
  sPtr<sRecord>               Record(new sRecord());
  sPtr<sDataListFolder>       Hres;
  Hres=sNodePtr(new sDataListFolder(sPhysPair(sPhysValue(Symbol::Degree)
                                             ,sPhysValue("G")           )));
  params->Set_eta(sPhysValue(90,0.5,Symbol::Degree));
  for( real xi_degree = 0.0 ; xi_degree < 46.0 ; xi_degree += 2.5 ){
    params->Set_xi(sPhysValue(xi_degree,Symbol::Degree));
    try{
      Hres->AddNewPoint(xi_degree,FMR_Hres(params));
    }catch( rxMinimizationError E ){
      RANet::Log.Put(sLog::Conditionless,"Magneto"
          ,(sString("FMR_Hres_vs_xi: at xi=")<<xi_degree)
           +"   "+E.MsgText());
    }catch(...){
      RANet::Log.Put(sLog::Conditionless,"Magneto"
          ,(sString("FMR_Hres_vs_xi: at xi=")<<xi_degree)
           +" unknown exception");
    }
  }

  Record->SetFinishTime();
  Record->AllocateRecord();
  (Record >= folder("pars",""))
      << attr("&eta;",sPhysValue(90,0.5,Symbol::Degree));
  params->ProvideAttributesTo(Record >= folder("pars",""));
  (Record >= folder("pars",""))
          << attr("g",params->_g)
          << attr("&nu;<sub>0</sub>",params->_f0);
  Record << attr("H<sub>res</sub>(&xi;)",Hres);
  return sAttribute(sAttrName("FMR in-plane"),Record);
}


sAttribute  FMR_Hres_vs_eta (psMagnetizationParams  params)
{
  sPtr<sRecord>               Record(new sRecord());
  sPtr<sDataListFolder>       Hres;
  Hres=sNodePtr(new sDataListFolder(sPhysPair(sPhysValue(Symbol::Degree)
                                             ,sPhysValue("G")           )));
  for( real eta_degree = 0.01 ; eta_degree < 90.01
                              ; eta_degree += ( eta_degree < 2.0
                                              ? 0.1
                                              : ( eta_degree < 8.0
                                                ? 0.5
                                                : ( eta_degree < 16.0
                                                  ? 1.0
                                                  : ( eta_degree < 32.0
                                                    ? 5.0
                                                    : 10.0 ) ) ) ) ){
    params->Set_eta(sPhysValue(eta_degree,Symbol::Degree));
    try{
      Hres->AddNewPoint(eta_degree,FMR_Hres(params));
    }catch( rxMinimizationError E ){
      RANet::Log.Put(sLog::Conditionless,"Magneto"
          ,(sString("FMR_Hres_vs_eta: at eta=")<<eta_degree)
           +"   "+E.MsgText());
    }catch(...){
      RANet::Log.Put(sLog::Conditionless,"Magneto"
          ,(sString("FMR_Hres_vs_eta: at eta=")<<eta_degree)
           +" unknown exception");
    }
  }

  Record->SetFinishTime();
  Record->AllocateRecord();
  (Record >= folder("pars",""))
      << attr("&xi;",params->_xi);
  params->ProvideAttributesTo(Record >= folder("pars",""));
  (Record >= folder("pars",""))
          << attr("g",params->_g)
          << attr("&nu;<sub>0</sub>",params->_f0);
  Record << attr("H<sub>res</sub>(&eta;)",Hres);
  return sAttribute(sAttrName(sString("FMR out-of-plane at &xi;=")
                             +params->_xi.Text())
                   ,Record);
}


sAttribute  LMOKE_shoulders (psMagnetizationParams  params)
{
  real                        _4piMs = 4*sMath::pi*params->Ms;
  real                        Hmax = 1.5 * _4piMs;
  real                        Hmin = -Hmax;
  real                        Hstep = (Hmax-Hmin)/200;
  sPtr<sRecord>               Record(new sRecord());
  sPtr<sDataListFolder>       MuMv;
  bool                        HasOther = false;
  sPtr<sDataListFolder>       MuMv_other;
  std::list<sStaticSolution>  Solutions;
  MuMv=sNodePtr(new sDataListFolder(sPhysPair(sPhysValue("G")
                                             ,sPhysValue("G2"))));
  for( real H = Hmin ; H < Hmax+Hstep ; H += ( std::abs(H)>20*Hstep
                                             ? Hstep
                                             : ( std::abs(H)>10*Hstep
                                               ? Hstep/4
                                               : Hstep/16 ) ) ){
    params->Set_H0(sPhysValue(H,"G"));
    Solutions = StaticSolutions(params);
    Solutions.sort();
    int I = 0;
    for( std::list<sStaticSolution>::iterator K = Solutions.begin()
                                            ; K != Solutions.end()  ; ++K ){
      real                        Mx,My,Mu,Mv;
      Mx = _4piMs * std::sin(K->theta) * std::cos(K->phi);
      My = _4piMs * std::sin(K->theta) * std::sin(K->phi);
      Mu =   Mx*cos(params->xi) + My*sin(params->xi);
      Mv = - Mx*sin(params->xi) + My*cos(params->xi);
      if( I == 0 ){
        MuMv->AddNewPoint(H , Mu*Mv );
      }else{
        if( ! HasOther ){
          HasOther = true;
          MuMv_other=sNodePtr(new sDataListFolder(sPhysPair(sPhysValue("G")
                                                           ,sPhysValue("G2"))));
        }
        MuMv_other->AddNewPoint(H , Mu*Mv );
      }
      ++I;
    }
  }

  Record->SetFinishTime();
  Record->AllocateRecord();
  (Record >= folder("pars",""))
//      << attr("H<sub>0</sub>",H0_Field->Value())
      << attr("&eta;",params->_eta)
      << attr("&xi;",params->_xi);
  params->ProvideAttributesTo(Record >= folder("pars",""));
  MuMv->SetAppearance(new sCurveAppearance(sColor::cBlue ,2));
  sString AN("16&pi;<sup>2</sup>M<sub>u</sub>");
  AN += Symbol::MiddleDot;
  AN += "M<sub>v</sub>";
  Record << attr(AN,MuMv);
  if( HasOther ){
    MuMv_other->SetAppearance(new sCurveAppearance(3,sColor::cBlue ));
    MuMv << attr(AN+" other",MuMv_other);
  }
  return sAttribute(sAttrName(sString("LMOKE shoulders at &eta;=")
                             +params->_eta.Text()
                             +" &xi;=" +params->_xi.Text())
                   ,Record);
}


sAttribute  VSM (psMagnetizationParams  params)
{
  real                        Hmax = 1.5 * params->Ms;
  real                        Hmin = -Hmax;
  real                        Hstep = (Hmax-Hmin)/500;
  sPtr<sRecord>               Record(new sRecord());
  sPtr<sDataListFolder>       Mu_vs_H;
  bool                        HasOther = false;
  sPtr<sDataListFolder>       Mu_vs_H_other;
  std::list<sStaticSolution>  Solutions;
  Mu_vs_H=sNodePtr(new sDataListFolder(sPhysPair(sPhysValue("G")
                                                ,sPhysValue("erg/G/cm3"))));
  for( real H = Hmin ; H < Hmax+Hstep ; H += ( std::abs(H)>50*Hstep
                                             ? Hstep
                                             : Hstep/10 )       ){
    params->Set_H0(sPhysValue(H,"G"));
    Solutions = StaticSolutions(params);
    Solutions.sort();
    int I = 0;
    for( std::list<sStaticSolution>::iterator K = Solutions.begin()
                                            ; K != Solutions.end()  ; ++K ){
      real                        Mx,My,Mu;
      Mx = params->Ms * std::sin(K->theta) * std::cos(K->phi);
      My = params->Ms * std::sin(K->theta) * std::sin(K->phi);
      Mu =   Mx*cos(params->xi) + My*sin(params->xi);
      if( I == 0 ){
        Mu_vs_H->AddNewPoint(H , Mu );
      }else{
        if( ! HasOther ){
          HasOther = true;
          Mu_vs_H_other=sNodePtr(new sDataListFolder(sPhysPair(sPhysValue("G")
                                                    ,sPhysValue("erg/G/cm3"))));
        }
        Mu_vs_H_other->AddNewPoint(H , Mu );
      }
      ++I;
    }
  }

  Record->SetFinishTime();
  Record->AllocateRecord();
  (Record >= folder("pars",""))
//      << attr("H<sub>0</sub>",H0_Field->Value())
      << attr("&eta;",params->_eta)
      << attr("&xi;",params->_xi);
  params->ProvideAttributesTo(Record >= folder("pars",""));
  Mu_vs_H->SetAppearance(new sCurveAppearance(sColor::cBlue ,2));
  sString AN("M<sub>u</sub>");
  Record << attr(AN,Mu_vs_H);
  if( HasOther ){
    Mu_vs_H_other->SetAppearance(new sCurveAppearance(3,sColor::cBlue ));
    Mu_vs_H << attr(AN+" other",Mu_vs_H_other);
  }
  return sAttribute(sAttrName(sString("VSM at &eta;=")
                             +params->_eta.Text()
                             +" &xi;=" +params->_xi.Text())
                   ,Record);
}


sMagnetizationParams::sMagnetizationParams ()
    :U_polar(NULL),dU_polar(NULL),UdU_polar(NULL),ddU_polar(NULL)
    ,_Ms(13.5,0.05,"G")
    ,_g(2.1,0.05,_Unitsless_)
    ,_f0(9.34,0.01,"GHz")
    ,_H0(0,0.5,"G")
    ,_eta(0,0.5,Symbol::Degree)
    ,_xi(0,0.5,Symbol::Degree)
{
  Ms = (_Ms/sPhysValue("G")).ToReal();
  g = _g.ToReal();
  f0 = (_f0/sPhysValue("MHz")).ToReal();
  H0 = (_H0/sPhysValue("G")).ToReal();
  eta = (sMath::pi/180) * (_eta/sPhysValue(Symbol::Degree)).ToReal();
  xi = (sMath::pi/180) * (_xi/sPhysValue(Symbol::Degree)).ToReal();
}


psMagnetizationParams  sMagnetizationParams::Set_g (rcsPhysValue  value)
{
  _g = value;
  g = _g.ToReal();
  return this;
}


psMagnetizationParams  sMagnetizationParams::Set_4piMs (rcsPhysValue  value)
{
  _Ms = value/(4*sMath::pi);
  Ms = (_Ms/sPhysValue("G")).ToReal();
  return this;
}

psMagnetizationParams sMagnetizationParams::Set_f0(rcsPhysValue  value)
{
  _f0 = value;
  f0 = (_f0/sPhysValue("MHz")).ToReal();
  return this;
}

psMagnetizationParams sMagnetizationParams::Set_H0(rcsPhysValue  value)
{
  _H0 = value;
  H0 = (_H0/sPhysValue("G")).ToReal();
  return this;
}

psMagnetizationParams sMagnetizationParams::Set_eta(rcsPhysValue  value)
{
  _eta = value;
  eta = (sMath::pi/180) * (_eta/sPhysValue(Symbol::Degree)).ToReal();
  return this;
}

psMagnetizationParams sMagnetizationParams::Set_xi(rcsPhysValue  value)
{
  _xi = value;
  xi = (sMath::pi/180) * (_xi/sPhysValue(Symbol::Degree)).ToReal();
  return this;
}


void  sMagnetizationParams::ProvideAttributesTo (sPtr<sFolderNode>  folder)
{
  folder << attr("4&pi;M<sub>s</sub>",_4piMs());
}


bool  sMagnetizationParams::IsMinimum_polar (const gsl_vector * x)
{
  int                         Status;
  gsl_matrix *                Hessian = gsl_matrix_alloc(2,2);
  ASSERT( ddU_polar );
  Status = (*ddU_polar)(x,this,Hessian);
  if( Status ){

  }
  return ( gsl_matrix_get(Hessian,0,0) > 0 )
      && ( gsl_matrix_get(Hessian,0,0)*gsl_matrix_get(Hessian,1,1)
          -gsl_matrix_get(Hessian,0,1)*gsl_matrix_get(Hessian,0,1) > 0 );
}

void  sStaticSolution::Normalize ()
{
  while( theta <= -sMath::pi )  theta += 2*sMath::pi;
  while( theta >   sMath::pi )  theta -= 2*sMath::pi;
  if( theta < 0 ){
    theta = -theta;
    phi += sMath::pi;
  }
  while( phi <= -sMath::pi )  phi += 2*sMath::pi;
  while( phi >   sMath::pi )  phi -= 2*sMath::pi;
}

//------------------------------------------------------------------------------
} //namespace RA
