////////////////////////////////////////////////////////////////////////////////
/*! @file Magneto.h   Common things for magnetization and its effects.
- Part of MagnetizationCalc - Research Assistant Plugin package.
- Uses  RANet - Research Assistant Net Library.
- Copyright(C) 2017, Viktor E. Bursian, St.Petersburg, Russia.
                     Viktor.Bursian@mail.ioffe.ru
*///////////////////////////////////////////////////////////////////////////////
#ifndef Magneto_H
#define Magneto_H
#include "Record.h"
#include "DataListFolder.h"
#include <gsl/gsl_vector.h>
#include <gsl/gsl_matrix.h>
#include <list>
#include <map>
namespace RA {
//------------------------------------------------------------------------------

ANNOUNCE_CLASS(xMinimizationError)
ANNOUNCE_CLASS(sMagnetizationParams)
ANNOUNCE_CLASS(sMagnetoTask)
ANNOUNCE_CLASS(s3DVector)
ANNOUNCE_CLASS(sStaticSolution)

typedef sAttribute  (* pfTask) (psMagnetizationParams);

//sStaticSolution  StaticSolution (rcsPoint               initial_point
//                                ,psMagnetizationParams  params
//                                );
//std::list<sStaticSolution>  StaticSolutions (psMagnetizationParams);

sAttribute  M_AllSolutions  (psMagnetizationParams);
sAttribute  M_vs_H          (psMagnetizationParams);
sAttribute  FMR_Hres_vs_xi  (psMagnetizationParams);
sAttribute  FMR_Hres_vs_eta (psMagnetizationParams);
sAttribute  LMOKE_shoulders (psMagnetizationParams);
sAttribute  VSM             (psMagnetizationParams);

//------------------------------------------------------------------------------

struct  xMinimizationError : xProgramError
{
    static const int          cNotAMinimum;
                              xMinimizationError (int  error_code)
                                  :xProgramError(sString("minimization error")
                                                <<error_code
                                                ,"",0,"xMinimizationError")
                                {}
};


//------------------------------------------------------------------------------

struct  sMagnetizationParams
{
    virtual                   ~sMagnetizationParams () {}
                              sMagnetizationParams ();

    typedef double (* p_U)  (const gsl_vector * ,void *);
    typedef void (* p_dU) (const gsl_vector * ,void * ,gsl_vector *);
    typedef void (* p_UdU) (const gsl_vector * ,void * ,double * ,gsl_vector *);
    typedef int (* p_ddU) (const gsl_vector * ,void * ,gsl_matrix *);

    psMagnetizationParams     Set_4piMs (rcsPhysValue);
    psMagnetizationParams     Set_g (rcsPhysValue);
    psMagnetizationParams     Set_f0 (rcsPhysValue);
    psMagnetizationParams     Set_H0 (rcsPhysValue);
    psMagnetizationParams     Set_eta (rcsPhysValue);
    psMagnetizationParams     Set_xi (rcsPhysValue);

    sPhysValue                _4piMs () const
                                { return 4*sMath::pi*_Ms; }

    virtual void              ProvideAttributesTo (sPtr<sFolderNode>);

    bool                      IsMinimum_polar (const gsl_vector *) /*const*/;

    p_U                       U_polar;
    p_dU                      dU_polar;
    p_UdU                     UdU_polar;
    p_ddU                     ddU_polar;

    double                    Ms;
    double                    g;
    double                    f0;
    double                    H0;
    double                    eta;
    double                    xi;

    sPhysValue                _Ms;
    sPhysValue                _g;
    sPhysValue                _f0;
    sPhysValue                _H0;
    sPhysValue                _eta;
    sPhysValue                _xi;
};

//------------------------------------------------------------------------------

class  sMagnetoTask
{
  public:
    enum                      eTaskType {cM_AllSolutions
                                        ,cM_vs_H
                                        ,cFMR_Hres_vs_xi
                                        ,cFMR_Hres_vs_eta
                                        ,cLMOKE_shoulders
                                        ,cVSM
                                        };
    static
    std::map<eTaskType
            ,sMagnetoTask>    Tasks;

  public:
                              sMagnetoTask (rcsString  name
                                           ,pfTask     function)
                                  :Name(name),Function(function)
                                {}
    sString                   Name;
    pfTask                    Function;
};

//------------------------------------------------------------------------------

class  s3DVector
{
  public:
    double                      x () const
                                  {
                                    return cartesian
                                          ? C.x
                                          : P.m*std::sin(P.p)*std::cos(P.a);
                                  }
    double                      y () const
                                  {
                                    return cartesian
                                          ? C.y
                                          : P.m*std::sin(P.p)*std::sin(P.a);
                                  }
    double                      z () const
                                  {
                                    return cartesian ? C.z : P.m*std::cos(P.p);
                                  }
    double                      module () const;
    double                      theta () const;
    double                      phi () const;

  private:
    struct sCartesian { double x,y,z; };
    struct sPolar { double m,p,a; };
    bool                        cartesian;
    union {sCartesian C; sPolar P; };
};

//------------------------------------------------------------------------------

class  sStaticSolution
{
  public:
    double                      theta;
    double                      phi;
    double                      energy;
    void                        Normalize ();
    bool                        operator < (rcsStaticSolution  other)
                                  { return energy < other.energy; }
    double                      theta_degree ()
                                  { return theta/sMath::pi*180; }
    double                      phi_degree ()
                                  { return phi/sMath::pi*180; }
};


//------------------------------------------------------------------------------
} //namespace RA
#endif
