////////////////////////////////////////////////////////////////////////////////
/*! @file MagnetoCalculator.cpp Calculator for magnetization and its effects.
- Part of MagnetizationCalc - Research Assistant Plugin package.
- Uses  RAGUI - Research Assistant Graphical User Interface.
- Uses  RANet - Research Assistant Net Library.
- Uses  Qt v.5  - http://qt-project.org/
- Copyright(C) 2017, Viktor E. Bursian, St.Petersburg, Russia.
                     Viktor_dot_Bursian_at_mail_dot_ioffe_dot_ru
*///////////////////////////////////////////////////////////////////////////////
#include "MagnetoCalculator.h"
#include "Cubic001Film.h"
#include "SymbolsGUI.h"
#include "attr.h"
#include "StringsAndQStrings.h"
#include "Log.h"
#include "Trace.h"
#include <QMimeData>
#include <QVBoxLayout>
#include <QHBoxLayout>
#include <QMessageBox>
#include <iostream>

#include <stdio.h>
#include <gsl/gsl_errno.h>
#include <gsl/gsl_math.h>
#include <gsl/gsl_roots.h>

namespace RA {
//------------------------------------------------------------------------------

DEFINE_CLASS_TAG(sMagnetoCalculator)

//------------------------------------------------------- sMagnetoCalculator ---
/*! @class sMagnetoCalculator

В зависимости от значения Linked и Topic имеет 3 состояния. Это не
состояния statemachine, а то, что показывается на графике, и, соответственно,
куда идёт импорт.
- (A) -- Linked && Topic.IsNULL() -- ничего не показывается (окна отсутствуют)
- (B) -- ! Linked && Topic.IsNULL() -- в точности как sPlot создаёт собственный
  sFolderNode, размещённый в WidgetConfig, как значение WidgetConfig>>"Content",
  и его показывает
- (C) -- Linked && Topic.IsCorrect()

*/


void  sMagnetoCalculator::Preconfigure (sNodePtr  DeviceCfg)
{
  sPtr<sFolderNode>              DeviceSettings;
  DeviceCfg |= "Settings";
  DeviceSettings = DeviceCfg >> "Settings";
//  if( DeviceSettings.IsCorrect() ){
//    DeviceSettings |= attr("Symmetry","Bursian");
//    DeviceSettings |= attr("PreserveColors",true);
//  }
  sConsole::Preconfigure(DeviceCfg);
}


sMagnetoCalculator::~sMagnetoCalculator()
{
//  if( DeviceSettings.IsCorrect() ){
//    DeviceSettings << attr("PhysParameter",PhysParameter);
//  }
  Clear();
}


sMagnetoCalculator::sMagnetoCalculator (sString    DeviceId
                           ,sNodePtr   DeviceCfg
                           ,sNodePtr   WidgetCfg
                           ,QWidget *  parent   )
    :sConsole(DeviceId,DeviceCfg,WidgetCfg,parent)
    ,g_Field( new sPhysValueField(this,"g-factor",sPhysValue(2.0,0.005)) )
    ,_4piMs_Field(new sPhysValueField(this,"4&pi;M<sub>s</sub>"
                                     ,sPhysValue("G")) )
    ,kc1_Field(new sPhysValueField(this,"K<sub>c1</sub>/M<sub>s</sub>"
                                  ,sPhysValue(-240,0.5,"G")))
    ,ku1_Field(new sPhysValueField(this,"K<sub>u1</sub>/M<sub>s</sub>"
                                  ,sPhysValue(0,0.5,"G")))
    ,f0_Field (new sPhysValueField(this,"&nu;<sub>0</sub>"
                                  ,sPhysValue(9.34e9,0.5e7,"Hz")))
    ,H0_Field (new sPhysValueField(this,"H<sub>0</sub>"
                                  ,sPhysValue(0,0.5,"G")))
    ,eta_Field(new sPhysValueField(this,"&eta;"
                                  ,sPhysValue(0,0.5,Symbol::Degree)))
    ,xi_Field (new sPhysValueField(this,"&xi;"
                                  ,sPhysValue(0,0.5,Symbol::Degree)))
    ,Evolution(NULL)
    ,Chart(NULL)
    ,Legend(NULL)
{
  try{ //get config parameters
//    sString                     R;
//    if( get_from<sTextNode>(DeviceSettings>>"Symmetry",R) ){
//      if( R == "Bursian" )  Symmetry=sSOSFileImport::Symmetry_Bursian;
//      if( R == "Nelson" )  Symmetry=sSOSFileImport::Symmetry_Nelson;
//    };
//    get_from<sBooleanNode>(DeviceSettings>>"PreserveColors",PreserveColors);
  }catch(...){
    SetInvalid(tr("Something wrong in configuration."));
  }
  //evolution of the inherited state machine
  {
    //evolution of the State_Operating
    {
    }
  }
}


void  sMagnetoCalculator::AppearanceMakeUp ()
{
  sConsole::AppearanceMakeUp();
  try{ //get config parameters (which are necessary for appearance)
//    if( ! get_from<sIntegerNode>(DeviceSettings>>"IntParameter",IntParameter) )
//      throw xDeviceLaunch;
  }catch(...){
    SetInvalid(tr("Something wrong in configuration."));
  }
  //add new widgets
  {
    TheLinkLabel = new QLabel("Store in:",this);
    TheLinkField = new sDropNodeWidget(this);

    SymmetryLabel = new QLabel("Symmetry:",this);
    SymmetryComboBox = new QComboBox(this);
    SymmetryComboBox->addItem("cubic (001) film",0);

    TaskLabel = new QLabel("Task:",this);
    TaskComboBox = new QComboBox(this);
    for( std::map<sMagnetoTask::eTaskType,sMagnetoTask>::const_iterator T
           = sMagnetoTask::Tasks.begin()
           ; T != sMagnetoTask::Tasks.end() ; ++T ){
      TaskComboBox->addItem(ToQString(T->second.Name),T->first);
    }

    CalcButton = new QPushButton("Calculate",this);

//    SampleBindingCheckBox = new QCheckBox("Replace text",this);
//    SampleLabel = new QLabel("=> sample:",this);
//    SampleTextField = new QLineEdit(this);

    Splitter = new QSplitter(this);
    Splitter->setOrientation(Qt::Vertical);

    QActionGroup *              AG = new QActionGroup(this);
    ActnLegendBottom = new QAction(AG);
    ActnLegendBottom->setIcon(QIcon(":/RAGUI/icons/LegendBottom.png"));
    ActnLegendBottom->setCheckable(true);
    ActnLegendBottom->setChecked(true);
    addAction(ActnLegendBottom);
    ActnLegendRight = new QAction(AG);
    ActnLegendRight->setIcon(QIcon(":/RAGUI/icons/LegendRight.png"));
    ActnLegendRight->setCheckable(true);
    addAction(ActnLegendRight);
  }
  //layout widgets
  {
    QVBoxLayout *               ConsoleLayout( new QVBoxLayout );
    QHBoxLayout *               TopLayout( new QHBoxLayout );
    QHBoxLayout *               MidLayout( new QHBoxLayout );
    QVBoxLayout *               CrystalLayout( new QVBoxLayout );
    QVBoxLayout *               TaskLayout( new QVBoxLayout );
    setLayout(ConsoleLayout);
    ConsoleLayout->addItem(TopLayout);
    ConsoleLayout->setStretchFactor(TopLayout,0);
    {
      TopLayout->addWidget(TheLinkLabel);
      TopLayout->addWidget(TheLinkField);
      TopLayout->addWidget(CalcButton);
      TopLayout->setStretchFactor(TheLinkLabel,0);
      TopLayout->setStretchFactor(TheLinkField,1);
      TopLayout->setStretchFactor(CalcButton,0);
    }
    ConsoleLayout->addItem(MidLayout);
    ConsoleLayout->setStretchFactor(MidLayout,0);
    {
      MidLayout->addItem(CrystalLayout);
      {
        {
          QHBoxLayout *               L( new QHBoxLayout );
          CrystalLayout->addItem(L);
//          L->addWidget(SymmetryLabel ,0,Qt::AlignLeft);
//          L->addWidget(SymmetryComboBox ,0,Qt::AlignLeft);
          L->addWidget(SymmetryLabel);
          L->addWidget(SymmetryComboBox);
          L->addStretch(100);
        }
        CrystalLayout->addWidget(g_Field ,0,Qt::AlignLeft);
        CrystalLayout->addWidget(_4piMs_Field ,0,Qt::AlignLeft);
        CrystalLayout->addWidget(kc1_Field ,0,Qt::AlignLeft);
        CrystalLayout->addWidget(ku1_Field ,0,Qt::AlignLeft);
        CrystalLayout->addStretch(100);
      }

      MidLayout->addItem(TaskLayout);
      {
        {
          QHBoxLayout *               L( new QHBoxLayout );
          TaskLayout->addItem(L);
          L->addWidget(TaskLabel);
          L->addWidget(TaskComboBox);
          L->addStretch(100);
        }
        TaskLayout->addWidget(f0_Field ,0,Qt::AlignLeft);
        TaskLayout->addWidget(H0_Field ,0,Qt::AlignLeft);
        TaskLayout->addWidget(eta_Field ,0,Qt::AlignLeft);
        TaskLayout->addWidget(xi_Field ,0,Qt::AlignLeft);
        TaskLayout->addStretch(100);
      }
    }
    ConsoleLayout->addWidget(Splitter);
    ConsoleLayout->setStretchFactor(Splitter,1);
  }
  connect(CalcButton,SIGNAL(clicked())
         ,this      ,SLOT(Calculate()));
  connect(TheLinkField ,SIGNAL(Edited())
         ,this     ,SLOT(OnTheLinkFieldEdit()));
  connect(this  ,SIGNAL(CursorPositionChanged(sPhysPair,sPhysPair
                                             ,sString,sString))
         ,MainWindow() ,SLOT(ShowCursorInfo(sPhysPair,sPhysPair
                                           ,sString,sString))
         ,Qt::QueuedConnection);
  UpdatingTimer = new QTimer(this);
  UpdatingTimer->start(sNavigator::UpdatingPeriod);
  connect(ActnLegendBottom,SIGNAL(toggled(bool))
         ,this,SLOT(PutLegendAtTheBottom(bool)));
  PutLegendAtTheBottom(false);
}


void  sMagnetoCalculator::RestoreState ()
{
  sConsole::RestoreState();
  try{
//    bool                      B;
    int                       I,I1,I2;
    sPhysValue                V;
    if( get_from<sIntegerNode>(WidgetConfig>>"Symmetry",I) ){
      SymmetryComboBox->setCurrentIndex(I);
    }
    if( get_from<sIntegerNode>(WidgetConfig>>"Task",I) ){
      TaskComboBox->setCurrentIndex(I);
    }
    if( get_from<sPhysValueNode>(WidgetConfig>>"g",V) ){
      g_Field->SetValue(V);
    }
    if( get_from<sPhysValueNode>(WidgetConfig>>"4&pi;M<sub>s</sub>",V) ){
      _4piMs_Field->SetValue(V);
    }
    if( get_from<sPhysValueNode>(WidgetConfig>>"k<sub>c1</sub>",V) ){
      kc1_Field->SetValue(V);
    }
    if( get_from<sPhysValueNode>(WidgetConfig>>"k<sub>u1</sub>",V) ){
      ku1_Field->SetValue(V);
    }
    if( get_from<sPhysValueNode>(WidgetConfig>>"&nu;<sub>0</sub>",V) ){
      f0_Field->SetValue(V);
    }
    if( get_from<sPhysValueNode>(WidgetConfig>>"H<sub>0</sub>",V) ){
      H0_Field->SetValue(V);
    }
    if( get_from<sPhysValueNode>(WidgetConfig>>"&eta;",V) ){
      eta_Field->SetValue(V);
    }
    if( get_from<sPhysValueNode>(WidgetConfig>>"&xi;",V) ){
      xi_Field->SetValue(V);
    }
//    if( get_from<sBooleanNode>(WidgetConfig>>"SampleBinding",B) ){
//      SampleBindingCheckBox->setChecked(B);
//    }
//    SampleField->SetValue(WidgetConfig>>"Sample");
    if( ! Evolution ){ /*! @patch This is a workaround to prevent second
                                  execution on second call of ResoreState()
                       */
      TheLinkField->SetValue(WidgetConfig >> "Content");
      OnTheLinkFieldEdit();
    }
    if( get_from<sIntegerNode>(WidgetConfig>>"Splitter.Orientation" ,I) ){
      Splitter->setOrientation(Qt::Orientation(I));
      ActnLegendBottom->setChecked(Splitter->orientation() == Qt::Vertical);
      ActnLegendRight->setChecked(Splitter->orientation() == Qt::Horizontal);
    }
    if(    get_from<sIntegerNode>(WidgetConfig>>"Splitter.Part1" ,I1)
        && get_from<sIntegerNode>(WidgetConfig>>"Splitter.Part2" ,I2) ){
      QList<int>    List;
      List.append(I1);
      List.append(I2);
      Splitter->setSizes(List);
    }
    if( Chart )
      Chart->RestoreScales(WidgetConfig);
//    sString                   S;
//    if( get_from<sTextNode>(WidgetConfig>>"SampleText" ,S) ){
//      SampleTextField->setText(ToQString(S));
//    }
  }catch(...){
  }
}


void  sMagnetoCalculator::StoreState ()
{
  disconnect(TheLinkField,0,0,0);
  sConsole::StoreState();
  try{
    WidgetConfig << attr("Symmetry",integer(SymmetryComboBox->currentIndex()));
    WidgetConfig << attr("Task",integer(TaskComboBox->currentIndex()));
    WidgetConfig << attr("g",g_Field->Value());
    WidgetConfig << attr("4&pi;M<sub>s</sub>",_4piMs_Field->Value());
    WidgetConfig << attr("k<sub>c1</sub>",kc1_Field->Value());
    WidgetConfig << attr("k<sub>u1</sub>",ku1_Field->Value());
    WidgetConfig << attr("&nu;<sub>0</sub>",f0_Field->Value());
    WidgetConfig << attr("H<sub>0</sub>",H0_Field->Value());
    WidgetConfig << attr("&eta;",eta_Field->Value());
    WidgetConfig << attr("&xi;",xi_Field->Value());
//    WidgetConfig << attr("SampleBinding"
//                        ,SampleBindingCheckBox->isChecked());
//    WidgetConfig << attr("SampleText" ,FromQString(SampleTextField->text()));
//    if( SampleField->Value.IsNULL() ){
//      WidgetConfig.DelAttr("Sample");
//    }else{
//      WidgetConfig << attr("Sample" ,SampleField->Value);
//    }
    WidgetConfig << attr("Splitter.Orientation" ,(int)Splitter->orientation());
    if( Chart ){
      WidgetConfig << attr("Splitter.Part1" ,(int)Splitter->sizes().at(0));
      WidgetConfig << attr("Splitter.Part2" ,(int)Splitter->sizes().at(1));
      Chart->StoreScales(WidgetConfig);
    }
  }catch(...){
  }
}


void  sMagnetoCalculator::contextMenuEvent (QContextMenuEvent *  event)
{
  QMenu                       PopupMenu/*(this)*/;
  PopupMenu.addActions(actions());
  PopupMenu.exec(event->globalPos());
}


void  sMagnetoCalculator::PutLegendAtTheBottom (bool  AtTheBottom)
{
  if( AtTheBottom ){
    Splitter->setOrientation(Qt::Vertical);
  }else{
    Splitter->setOrientation(Qt::Horizontal);
  }
}


void  sMagnetoCalculator::OnTheLinkFieldEdit ()
{
  if( Evolution ){
    RemoveChartAndLegend();
  }
  if( ! TheLinkField->Value.IsNULL() ){
    if( ! CreateChartAndLegend(TheLinkField->Value) ){ //case A to case C
      TheLinkField->Clear();
    }
  }
}


void  sMagnetoCalculator::Clear ()
{
  if( Legend ){ delete Legend;  Legend=NULL; };
  if( Chart ){ delete Chart;  Chart=NULL; };
  if( Evolution ){ delete Evolution;  Evolution=NULL; };
  TheLinkField->Clear();
}


void  sMagnetoCalculator::RemoveChartAndLegend ()
{
  setAcceptDrops(false);
  try{
    if( Chart )
      Chart->StoreScales(WidgetConfig);
    Clear();
    WidgetConfig.DelAttr("Content");
  }catch(...){
  }
}


bool  sMagnetoCalculator::CreateChartAndLegend (sNodePtr  folder)
{
  sPtr<sFolderNode>           Content(folder);
  if( Content.IsCorrect() ){
    WidgetConfig << attr("Content",Content);
    Evolution = new sNet2TreeEvolution(sAttribute(sAttrName(Title()),Content));
    Chart = new sChart(Evolution ,this);
    Legend = new sNavigator(Evolution ,this);

    Splitter->addWidget(Chart);
    Splitter->addWidget(Legend);

    connect(Chart ,SIGNAL(CursorPositionChanged(sPhysPair,sPhysPair
                                               ,sString,sString))
           ,this  ,SIGNAL(CursorPositionChanged(sPhysPair,sPhysPair
                                               ,sString,sString)));
    connect(Evolution,SIGNAL(Changed()),Legend,SLOT(OnNetChange()));
    connect(Evolution,SIGNAL(Changed()),Chart,SLOT(OnNetChange()));
    Evolution->Initiate();
    connect(UpdatingTimer, SIGNAL(timeout()), Evolution, SLOT(Update()));
    Chart->RestoreScales(WidgetConfig);
    setAcceptDrops(true);
    return true;
  }
  return false;
}


void  sMagnetoCalculator::Calculate ()
{
  sMagnetoTask::eTaskType     TaskType = (sMagnetoTask::eTaskType)
                                         (TaskComboBox->currentIndex());
  psMagnetizationParams       MParams(NULL);
  try{
    MParams = (new sCubic001FilmParams())
              ->Set_kc1(kc1_Field->Value())
              ->Set_ku1(ku1_Field->Value())
              ->Set_g(g_Field->Value())
              ->Set_4piMs(_4piMs_Field->Value())
              ->Set_f0(f0_Field->Value())
              ->Set_H0(H0_Field->Value())
              ->Set_eta(eta_Field->Value())
              ->Set_xi(xi_Field->Value())
              ;
  }catch(...){
  }
  try{
    WidgetConfig >> "Content"
        << (*(sMagnetoTask::Tasks.at(TaskType).Function))(MParams);
  }catch(...){
  }
  if( MParams ){ delete MParams; MParams=NULL; }
}

//------------------------------------------------------------------------------
} //namespace RA


