////////////////////////////////////////////////////////////////////////////////
/*! @file KineticsNIDAQ.cpp  Усреднитель данных кинетики NIDAQ.
- Part of NIDAQmxDrivers - a Research Assistant Plugin package.
- Uses  RAGUI - the Research Assistant Graphical User Interface.
- Uses  QtGui v.5.0  -  http://qt.nokia.com/
- Uses  RANet - Research Assistant Net Library (based on ANSI C++).
- Copyright(C) 2010-2013, Dmitry K. Nelson, St.Petersburg, Russia.
                          D.Nelson@mail.ioffe.ru
*///////////////////////////////////////////////////////////////////////////////
#include "KineticsNIDAQ.h"
#include "attr.h"
#include "StringsAndQStrings.h"
#include "Log.h"
#include <QMessageBox>
namespace RA {
//------------------------------------------------------------------------------

DEFINE_CLASS_TAG(sKineticsNIDAQ)

//--------------------------------------------------- sKineticsNIDAQ ---

void  sKineticsNIDAQ::Preconfigure (sNodePtr  DeviceCfg)
{
  sPtr<sFolderNode>              DeviceSettings;
  sPtr<sFolderNode>              DeviceLinks;
  DeviceCfg |= "Links";
  DeviceLinks = DeviceCfg >> "Links";
  if( DeviceLinks.IsCorrect() ){
    DeviceLinks |= "Gauge";
    DeviceLinks >> "Gauge" |= attr("Class","sNIDAQCounter");
  };
  DeviceCfg |= "Settings";
  DeviceSettings = DeviceCfg >> "Settings";
  if( DeviceSettings.IsCorrect() ){
      //изменение умолчательных шкал для sScanner
    sChartScales      ChartScales(sChartScale("Time"
                                              ,sPhysRange(0,1.0e-4,"s"))
                                 ,sChartScale(""
                                             ,sPhysRange(0,200,"kphoton/s"))
                                 );
    ChartScales.StoreSettings(DeviceSettings);

    DeviceSettings |= attr("ValidScanRange",sPhysRange(0.0,real_inf,"s"));
    DeviceSettings |= attr("ValidScanSteps",sPhysRange(5.0e-7,10,"s"));
    DeviceSettings |= attr("ScanRange",sPhysRange(0.0,1.0e-4,"s",5.0e-9));
    DeviceSettings |= attr("ScanStep",sPhysValue(5.0e-7,5.0e-9,"s"));
  };
  sScanner::Preconfigure(DeviceCfg);
}


sKineticsNIDAQ::~sKineticsNIDAQ()
{
  if( DeviceSettings.IsCorrect() ){
/*
    DeviceSettings << attr("ScanRange",ScanRangeField->Value())
                   << attr("ScanStep",ScanStepField->Value());
*/
  };
  Gauge.SetToNULL();
}


sKineticsNIDAQ::sKineticsNIDAQ (sString    DeviceId
                               ,sNodePtr   DeviceCfg
                               ,sNodePtr   WidgetCfg
                               ,QWidget *  parent   )
    :sScanner(DeviceId,DeviceCfg,WidgetCfg,parent)
    ,Gauge(this)
    ,ValidScanRange(sPhysRange(0.0,real_inf,"s"))
    ,ValidScanSteps(sPhysRange(5.0e-7,10,"s"))
    ,ScanRange(sPhysRange(0.0,1.0,"s"))
    ,ScanStep(sPhysValue(5.0e-7,"s"))
    ,NumOfPoints(0)
    ,CurrentSamplingNo(0)
{
  Multiplier.X = sPhysValue(1,"s");
  //evolution of the inherited state machine
  {
    //evolution of the State_Warming
    {
      State_W_Gauge = new sState(State_Warming());
      State_W_Gauge
          ->SetAsInitial()
          ->OnEnterCall( SLOT(LinkToGaugeAndMakeItWarming()) )
          ->AddTransition( SIGNAL(GaugeWarmed()) ,State_Warming()->Fin() )
          ->AddTransition( SIGNAL(ErrorMessage()),State_Cooling()      );
    };
    //evolution of the State_O_Scanning
    {
      State_O_Scanning->OnExitCall(SLOT(ReleaseGaugeControl()));
        
      State_O_Sc_SettingGauge = new sState(State_O_Scanning
                                          ,"State_O_Sc_SettingGauge");
      State_O_Sc_ResettingGauge = new sState(State_O_Scanning
                                            ,"State_O_Sc_ResettingGauge");
                                            
      State_O_Sc_SettingGauge
          ->SetAsInitial()
          ->AddTransition( SIGNAL(Break()) ,State_O_Sc_ResettingGauge )
          ->AddTransition(State_O_Sc_Acquisition);
      
      State_O_Sc_Deciding
          ->RemoveOnFinishTransition()
          ->AddTransition(State_O_Sc_ResettingGauge);
      
      State_O_Sc_ResettingGauge
          ->OnEnterCall(SLOT(ResetGaugeRegime()))
          ->AddTransition(State_O_Scanning->Fin());
      
      //State_O_Sc_SettingGauge internal structure and behavior
      {
        State_O_Sc_SG_Waiting = new sState(State_O_Sc_SettingGauge
                                            ,"State_O_Sc_SG_Waiting");
        State_O_Sc_SG_Waiting
            ->SetAsInitial()
            ->OnEnterCall(SLOT(SetGaugeRegime()))
            ->AddTransition(SIGNAL(GaugeBusy()) ,State_O_Idle)
            ->AddTransition<sPhysValue>(SIGNAL(GaugePeriodChanged(sPhysValue))
                                       ,SLOT(CorrectStep(sPhysValue))
                                       ,State_O_Sc_SettingGauge->Fin());
      }
      //State_O_Sc_ResettingGauge internal structure and behavior
      {
        State_O_Sc_RG_Waiting = new sState(State_O_Sc_ResettingGauge
                                            ,"State_O_Sc_RG_Waiting");
                                            
        State_O_Sc_RG_Waiting
            ->SetAsInitial()
            ->AddTransition(SIGNAL(GaugePeriodChanged(sPhysValue))
                           ,SLOT(RestartGauge())
                           ,State_O_Sc_ResettingGauge->Fin());
      }
    };
    connect(this ,SIGNAL(GaugeError(QString))
           ,this ,SLOT(ReportGaugeError(QString)) );
  };
}

void  sKineticsNIDAQ::AppearanceMakeUp ()
{
  sScanner::AppearanceMakeUp();

  try{ //get config parameters
    get_from<sPhysRangeNode>(DeviceSettings >> "ValidScanRange",ValidScanRange);
    get_from<sPhysRangeNode>(DeviceSettings >> "ValidScanSteps",ValidScanSteps);
    get_from<sPhysRangeNode>(DeviceSettings >> "ScanRange" ,ScanRange);
    get_from<sPhysValueNode>(DeviceSettings >> "ScanStep" ,ScanStep);
  }catch(...){
    SetInvalid(tr("Something wrong in configuration."));
  }
  //add new widgets
  {
    ScanRangeField = new sPhysRangeField(ParamArea,sPhysRange(),false,false);
    ParamLayout->addWidget(ScanRangeField,Row++,0,1,3,Qt::AlignLeft);

    ScanStepField = new sPhysValueField(ParamArea,sPhysValue(),false,false,false);
    ParamLayout->addWidget(ScanStepField,Row++,0,1,3,Qt::AlignLeft);
  }
  ScanRangeField->SetValidRange(ValidScanRange);
  ScanRangeField->SetValue(ScanRange);
  ScanStepField->SetValidRange(ValidScanSteps);
  ScanStepField->SetValue(ScanStep);
  Translate();
}

void  sKineticsNIDAQ::Translate ()
{
  ScanRangeField->SetLabel(tr("Range"));
  ScanStepField->SetLabel(tr("Step"));
}

void  sKineticsNIDAQ::RestoreState ()
{
  sScanner::RestoreState();
  if( WidgetConfig.IsCorrect() ){
    try{ //get stored parameters
      get_from<sPhysRangeNode>(WidgetConfig >> "ScanRange" ,ScanRange);
      get_from<sPhysValueNode>(WidgetConfig >> "ScanStep" ,ScanStep);
    }catch(...){
      SetInvalid(tr("Something wrong in desktop stored."));
    }
  }
  ScanRangeField->SetValidRange(ValidScanRange);
  ScanRangeField->SetValue(ScanRange);
  ScanStepField->SetValidRange(ValidScanSteps);
  ScanStepField->SetValue(ScanStep);
}

void  sKineticsNIDAQ::StoreState ()
{
  sScanner::StoreState();
  if( WidgetConfig.IsCorrect() ){
    WidgetConfig << attr("ScanRange",ScanRangeField->Value())
                 << attr("ScanStep",ScanStepField->Value());
  }
}

void sKineticsNIDAQ::ReportGaugeError (QString  error_text)
{
  emit ErrorMessage( tr("Got error from '%1':\n\n%2")
                       .arg(ToQString(Gauge->DeviceIdentifier()))
                       .arg(error_text)     );
}

void sKineticsNIDAQ::LinkToGaugeAndMakeItWarming ()
{
  try{
    Gauge.SetTo(DeviceLinks,"Gauge");
    Connect(Gauge ,SIGNAL(Warmed())       ,SIGNAL(GaugeWarmed()) );
    Connect(Gauge ,SIGNAL(ErrorMessage(QString))
           ,SIGNAL(GaugeError(QString)) );
    Connect(Gauge ,SIGNAL(PeriodChanged(sPhysValue))
           ,SIGNAL(GaugePeriodChanged(sPhysValue)) );
  }catch( rxDeviceLaunch E ){
    emit ErrorMessage( tr("Error while linking to a submissive device.\n"
                          "%1\nCheck configuration.")
                         .arg(ToQString(E.MsgText()))   );
  }
}

void sKineticsNIDAQ::OnEnterToCoolingState ()
{
  Disconnect(Gauge ,0 ,0);
  Gauge.ReleaseExclusiveControl();
  Gauge.SetToNULL();
  sScanner::OnEnterToCoolingState();
}

void  sKineticsNIDAQ::CreateNewDependence ()
{
  if( Multiplier.X.IsNaN() ){
    Multiplier.X = sPhysValue(ScanStep.Units());
    sPhysRange                  RX(0*Multiplier.X,100*Multiplier.X);
    SetChartScales(sChartScales(sChartScale(ChartScales().X.Name,RX)
                               ,ChartScales().Y
                               ));
  };
  NumOfPoints = Round((ScanRange.Size()/ScanStep).ToReal());
  real ArbFrom = (ScanRange.From() / Multiplier.X).ToReal();
  real ArbStep = (ScanStep / Multiplier.X).ToReal();

  DataArrayRecord =
      sNodePtr(new sDataArrayRecord(NumOfPoints,ArbFrom,ArbStep));
  DataArrayRecord->SetMultiplier(sPhysPair(Multiplier.X,Multiplier.Y));
  Dependence = DataArrayRecord;
}

void  sKineticsNIDAQ::StartScanning ()
{
  CurrentSamplingNo=0;

  Connect(Gauge ,SIGNAL(NewDataArray (sNIDAQCounter::sArrayData))
                ,SLOT(AddNewDataArray(sNIDAQCounter::sArrayData)) );
  Gauge->SetAutoRepeat(true);
  SendDirective(Gauge,sNIDAQCounter::cStartKinetics,NumOfPoints);
}

void  sKineticsNIDAQ::StopScanning ()
{
  SendCommand(Gauge,sGauge::cStop);
  Disconnect(Gauge,SIGNAL(NewDataArray (sNIDAQCounter::sArrayData))
                  ,SLOT(AddNewDataArray(sNIDAQCounter::sArrayData)));
}

void  sKineticsNIDAQ::EarlySetAttributes ()
{
  sScanner::EarlySetAttributes ();
  Dependence << attr("Scan range", ScanRange.Text())
             << attr("Scan step", ScanStep.Text());
}

void  sKineticsNIDAQ::LateSetAttributes ()
{
  sScanner::LateSetAttributes ();
  Dependence << attr("Number of samples", CurrentSamplingNo);
}

void  sKineticsNIDAQ::SetGaugeRegime ()
{
  try{
    ScanStep = ScanStepField->Value();
    ScanRange = ScanRangeField->Value();
    if( ! Gauge.TakeExclusiveControl() ){
      QMessageBox::critical(this 
          ,tr("'%1' cannot scan")
          .arg(ToQString(DeviceIdentifier()))
          ,tr("Gauge '%1' is busy now.\n\n  Try again later.")
          .arg(ToQString(Gauge->DeviceIdentifier()))
          );
      emit GaugeBusy();
    }else{
      //вычисляем желательное время на шаг:
      sPhysValue                  TimePerStep("s");
      ScanStep.FixTheUnits();
      TimePerStep = ScanStep;
      GaugeWasStarted = true;
      GaugeWasInAutoRepeatMode = Gauge->AutoRepeat();
      OldPeriod = Gauge->Period();
      SendCommand(Gauge ,sGauge::cStop);
      SendDirective(Gauge ,sGauge::cSetPeriod ,TimePerStep);
    }
  }catch(rxRAlgebra){
    QMessageBox::critical(this
        ,tr("%1 failure.").arg(ToQString(DeviceIdentifier()))
        ,tr("Wrong units in the configuration settings.")
        );
    SetInvalid();
  }
}

void  sKineticsNIDAQ::CorrectStep (sPhysValue  gauge_period)
{
  try{
    //правим ScanStep и поле ScanStepField:
    ScanStep = gauge_period;
    ScanStepField->SetValue( ScanStep );
    //правим поле ScanRangeField (его часть To):
    if( ScanRange.To() < ScanRange.From() + ScanStep ){
      ScanRange = sPhysRange( ScanRange.From() , ScanRange.From() + ScanStep );
      ScanRangeField->SetValue(ScanRange);
    }
  }catch(rxRAlgebra){
    QMessageBox::critical(this
        ,tr("%1 failure.").arg(ToQString(DeviceIdentifier()))
        ,tr("Wrong units in the configuration settings.")
        );
    SetInvalid();
  }
}

void  sKineticsNIDAQ::ResetGaugeRegime ()
{
  if( GaugeWasInAutoRepeatMode != Gauge->AutoRepeat() ){
    Gauge->SetAutoRepeat(GaugeWasInAutoRepeatMode);
  }
  SendDirective(Gauge ,sGauge::cSetPeriod ,OldPeriod);
}


void  sKineticsNIDAQ::RestartGauge ()
{
  if( GaugeWasStarted ){
    SendCommand(Gauge ,sGauge::cStart ,cReleaseControl);
  }
}


void  sKineticsNIDAQ::ReleaseGaugeControl ()
{
  Gauge.ReleaseExclusiveControl();
}


void  sKineticsNIDAQ::AddNewDataArray (sNIDAQCounter::sArrayData Data)
{
  if( ! DataArrayRecord.IsCorrect() ){
    return;
  };
  if( Multiplier.X.IsNaN() ){
    return;
  };
  if( Multiplier.Y.IsNaN() ){
    Multiplier.Y = sPhysValue(_Unitsless_);
//    sPhysRange                  RY(-Abs(Data),3*Abs(Data));
//    SetChartScales(sChartScales(ChartScales().X
//                               ,sChartScale(ChartScales().Y.Name,RY)
//                               ));
  };
  if( DataArrayRecord->Multiplier().Y.IsNaN() ){
    DataArrayRecord->SetMultiplier(sPhysPair(Multiplier.X,Multiplier.Y));
  };
  CurrentSamplingNo++;
  for (int PointNo = 0; PointNo < NumOfPoints; PointNo++ ){
      if (CurrentSamplingNo == 1){
          DataArrayRecord->SetValue(PointNo, Data[PointNo]);
      }else{
          DataArrayRecord
              ->SetValue(PointNo
                        ,( DataArrayRecord->PointY(PointNo)
                                                  *(CurrentSamplingNo-1)
                          + Data[PointNo]
                         )/CurrentSamplingNo);
      }
  };
  //!@todo{Nelson} А как сделать ровно 100 накоплений?
}

//------------------------------------------------------------------------------
} //namespace RA
