////////////////////////////////////////////////////////////////////////////////
/*! @file NIDAQCounter.cpp   Работа со счетчиками платы NIDAQ.
- Part of NIDAQmxDrivers - a Research Assistant Plugin package.
- Uses  RAGUI - the Research Assistant Graphical User Interface.
- Uses  QtGui v.4.6  -  http://qt.nokia.com/
- Uses  RANet - Research Assistant Net Library (based on ANSI C++).
- Copyright(C) 2010-2013, Dmitry K. Nelson, St.Petersburg, Russia.
                          D.Nelson@mail.ioffe.ru
*///////////////////////////////////////////////////////////////////////////////
#include "NIDAQCounter.h"
#include "attr.h"
#include <QDebug>
namespace RA {
//------------------------------------------------------------------------------

DEFINE_CLASS_TAG(sNIDAQCounter)

/*! Директива, запускающая плату NIDAQ для измерения
    кинетики в режиме "медленного времени".
*/
const sDirective<int>         sNIDAQCounter::cStartKinetics("cStartKinetics");


/*! Директива, запускающая плату NIDAQ для измерения
    кинетики в режиме "Старт-стоп" с одним стопом на старт.
*/
const sDirective<sNIDAQCounter::sSSParams>  sNIDAQCounter::cStart_StartStop
                                                           ("cStart_StartStop");

//------------------------------------------------------------ sNIDAQCounter ---

void  sNIDAQCounter::Preconfigure (sNodePtr  DeviceCfg)
{
  sPtr<sFolderNode>              DeviceSettings;
  DeviceCfg |= "Settings";
  DeviceSettings = DeviceCfg >> "Settings";
  if( DeviceSettings.IsCorrect() ){
    DeviceSettings |= attr("StrobeTerminalName","/Dev1/PFI9");
    DeviceSettings |= attr("SignalTerminalName","/Dev1/PFI3");
  }
  sGauge::Preconfigure(DeviceCfg);
}


sNIDAQCounter::~sNIDAQCounter ()
{
  StopCounting();
}


sNIDAQCounter::sNIDAQCounter (sString   DevicetId
                       ,sNodePtr  DeviceCfg)
    :sGauge(DevicetId,DeviceCfg)
    ,Counting(false)
    ,KinMeasuring(false)
    ,ArraySize(0)
    ,PrevCount (0)
    ,CounterHandle(0)
    ,TimerHandle(0)
    ,readArray(NULL)
    ,StartStopCounts(1000)
    ,SSMode(0)
    ,StrobeTermName("")
    ,SignalTermName("")
{
}


void  sNIDAQCounter::PostConstructor ()
{
  sGauge::PostConstructor();
  try{ //get config parameters
    get_from<sTextNode>(DeviceSettings >> "StrobeTerminalName" ,StrobeTermName);
    get_from<sTextNode>(DeviceSettings >> "SignalTerminalName" ,SignalTermName);
  }catch(...){
    SetInvalid(tr("Something wrong in configuration."));
  }
  //evolution of the inherited state machine (behaviour)
  {
    //evolution of the State_Operating
    {
      State_O_Kinetics = new sState(State_O_Measuring);
      State_O_StartStop = new sState(State_O_Measuring);

      State_O_M_Stopped
          ->AddTransition<int>(cStartKinetics
                              ,SLOT(StartKineticsCounting (int))
                              ,State_O_Kinetics)
          ->AddTransition<sSSParams>(cStart_StartStop
                                    ,SLOT(Start_StartStopCounting (sSSParams))
                                    ,State_O_StartStop);
      State_O_M_Acquiring
          ->OnEnterCall(SLOT(StartCounting()))
          ->OnExitCall(SLOT(StopCounting()))
          ->AddTransition<sPhysValue>(SIGNAL(GotSingleCount(sPhysValue))
                                     ,SIGNAL(NewData(sPhysValue))
                                     ,State_O_M_Cycling);
      State_O_Kinetics
          ->AddTransition(cStop ,State_O_M_Stopped)
          ->OnExitCall(SLOT(StopCounting()));
      State_O_StartStop
          ->AddTransition(cStop ,State_O_M_Stopped)
          ->OnExitCall(SLOT(StopCounting()));
    };
  };
}

void  sNIDAQCounter::StartCounting ()
{
  if (AutoRepeat()) StartContCounting ();
  else StartSingleCount ();
}

void  sNIDAQCounter::StartContCounting ()
{
  PrevCount = 0;
  real TimeInSec;
  try {
    TimeInSec = (ThePeriod/(sPhysValue(1.0,"s"))).ToReal();
  }catch(rxRAlgebra){
    emit ErrorMessage(tr("Cannot start counting. Illegal units"));
    emit CoolDown();
    Counting = false;
    return;
  }
  bool Error;
  long  ErrorCode;
  ErrorCode = DAQmxCreateTask ("", &TimerHandle);
  Error = !(ErrorCode == 0);
  ErrorCode = DAQmxCreateCOPulseChanTime (TimerHandle, "Dev1/ctr0"
                                         ,"", DAQmx_Val_Seconds
                                         ,DAQmx_Val_Low
                                         ,TimeInSec
                                         ,0.5*TimeInSec
                                         ,0.5*TimeInSec);
// qDebug() << " Timer Create Chanel, ErrorCode = " << ErrorCode;
  Error = Error || !(ErrorCode == 0);
  ErrorCode = DAQmxCfgImplicitTiming (TimerHandle
                                     ,DAQmx_Val_ContSamps
                                     ,1000);
//  qDebug() << " Timer Configure Chanel, ErrorCode = " << ErrorCode;
  Error = Error || !(ErrorCode == 0);

  ErrorCode = DAQmxCreateTask ("", &CounterHandle);
  Error = Error || !(ErrorCode == 0);
  ErrorCode = DAQmxCreateCICountEdgesChan(CounterHandle,"Dev1/ctr1"
                                         ,"",DAQmx_Val_Rising
                                         ,0,DAQmx_Val_CountUp);
//  qDebug() << " Counter Create Chanel, ErrorCode = " << ErrorCode;
  Error = Error || !(ErrorCode == 0);
  ErrorCode = DAQmxCfgSampClkTiming (CounterHandle,"/Dev1/Ctr0InternalOutput", 1000
                                    ,DAQmx_Val_Rising
                                    ,DAQmx_Val_ContSamps , 1000);
  Error = Error || !(ErrorCode == 0);

  ErrorCode = DAQmxRegisterEveryNSamplesEvent (CounterHandle
                                        ,DAQmx_Val_Acquired_Into_Buffer
                                        ,1, 0, &RA::sNIDAQCounter::OnReadyToReadC, this);
//  qDebug() << " Counter Callback Initiated, ErrorCode = " << ErrorCode;
  if (ErrorCode != 0) {
                        char EString[128];
                        DAQmxGetErrorString (ErrorCode, EString, 128);
                        qDebug() << EString;
                      };
  Error = Error || !(ErrorCode == 0);

  if (!Error)
      ErrorCode = DAQmxStartTask(TimerHandle);
  bool TimerStarted = (ErrorCode == 0);
  Error = Error || !(ErrorCode == 0);
  if (!Error)
      ErrorCode = DAQmxStartTask(CounterHandle);
  bool CounterStarted = (ErrorCode == 0);
  Error = Error || !(ErrorCode == 0);

  if (Error) {
      if (TimerStarted) ErrorCode = DAQmxStopTask(TimerHandle);
      if (CounterStarted) ErrorCode = DAQmxStopTask(CounterHandle);
      if (TimerHandle != 0) ErrorCode = DAQmxClearTask (TimerHandle);
      if (CounterHandle != 0) ErrorCode = DAQmxClearTask (CounterHandle);
      emit ErrorMessage(tr("Cannot start counting. NIDAQ-card is absent or out of service."));
      emit CoolDown();
      qDebug() << "Cooling Down";
      Counting = false;
  } else{
      Counting = true;
      qDebug() << " Counting Successfully Started !";
  }
}

void  sNIDAQCounter::StartSingleCount ()
{
    real TimeInSec;
    try {
      TimeInSec = (ThePeriod/(sPhysValue(1.0,"s"))).ToReal();
    }catch(rxRAlgebra){
      emit ErrorMessage(tr("Cannot start counting. Illegal units"));
      emit CoolDown();
      Counting = false;
      return;
    }
    bool Error;
    long  ErrorCode;
    ErrorCode = DAQmxCreateTask ("", &TimerHandle);
    Error = !(ErrorCode == 0);
    ErrorCode = DAQmxCreateCOPulseChanTime (TimerHandle, "Dev1/ctr0"
                                           ,"", DAQmx_Val_Seconds
                                           ,DAQmx_Val_Low
                                           ,0.02*TimeInSec
                                           ,0.5*TimeInSec
                                           ,0.5*TimeInSec);
  // qDebug() << " Timer Create Chanel, ErrorCode = " << ErrorCode;
    Error = Error || !(ErrorCode == 0);
    ErrorCode = DAQmxCfgImplicitTiming (TimerHandle
                                       ,DAQmx_Val_ContSamps
                                       ,3);
  //  qDebug() << " Timer Configure Chanel, ErrorCode = " << ErrorCode;
    Error = Error || !(ErrorCode == 0);

    ErrorCode = DAQmxCreateTask ("", &CounterHandle);
    Error = Error || !(ErrorCode == 0);
    ErrorCode = DAQmxCreateCICountEdgesChan(CounterHandle,"Dev1/ctr1"
                                           ,"",DAQmx_Val_Rising
                                           ,0,DAQmx_Val_CountUp);
  //  qDebug() << " Counter Create Chanel, ErrorCode = " << ErrorCode;
    Error = Error || !(ErrorCode == 0);
    ErrorCode = DAQmxCfgSampClkTiming (CounterHandle,"/Dev1/Ctr0InternalOutput", 1000
                                      ,DAQmx_Val_Rising
                                      ,DAQmx_Val_FiniteSamps, 2);
    Error = Error || !(ErrorCode == 0);

    ErrorCode = DAQmxRegisterDoneEvent (CounterHandle, 0
                                           , &RA::sNIDAQCounter::OnReadyToReadSC, this);
  //  qDebug() << " Counter Callback Initiated, ErrorCode = " << ErrorCode;
    if (ErrorCode != 0) {
                          char EString[128];
                          DAQmxGetErrorString (ErrorCode, EString, 128);
                          qDebug() << EString;
                        };
    Error = Error || !(ErrorCode == 0);

    if (!Error)
        ErrorCode = DAQmxStartTask(CounterHandle);
    bool CounterStarted = (ErrorCode == 0);
    Error = Error || !(ErrorCode == 0);
    if (!Error)
        ErrorCode = DAQmxStartTask(TimerHandle);
    bool TimerStarted = (ErrorCode == 0);
    Error = Error || !(ErrorCode == 0);

    if (Error) {
        if (TimerStarted) ErrorCode = DAQmxStopTask(TimerHandle);
        if (CounterStarted) ErrorCode = DAQmxStopTask(CounterHandle);
        if (TimerHandle != 0) ErrorCode = DAQmxClearTask (TimerHandle);
        if (CounterHandle != 0) ErrorCode = DAQmxClearTask (CounterHandle);
        emit ErrorMessage(tr("Cannot start counting. NIDAQ-card is absent or out of service."));
        emit CoolDown();
        qDebug() << "Cooling Down";
        Counting = false;
    } else{
        Counting = true;
        qDebug() << " Single Count Successfully Started !";
    }

}

void  sNIDAQCounter::StartKineticsCounting (int size)
{
  PrevCount = 0;
  ArraySize = size;
  readArray = new unsigned long [ArraySize];
  DataArray.SetSize(ArraySize);
  real TimeInSec;
  try {
    TimeInSec = (ThePeriod/(sPhysValue(1.0,"s"))).ToReal();
  }catch(rxRAlgebra){
    emit ErrorMessage(tr("Cannot start kinetics. Illegal units"));
    emit CoolDown();
    KinMeasuring = false;
    return;
  }
  bool Error;
  long  ErrorCode;
  ErrorCode = DAQmxCreateTask ("", &TimerHandle);
  Error = !(ErrorCode == 0);
  ErrorCode = DAQmxCreateCOPulseChanTime (TimerHandle, "Dev1/ctr0"
                                         ,"", DAQmx_Val_Seconds
                                         ,DAQmx_Val_Low
                                         ,TimeInSec
                                         ,0.5*TimeInSec
                                         ,0.5*TimeInSec);
//  qDebug() << " Kinetics Timer Create Chanel, ErrorCode = " << ErrorCode;
  Error = Error || !(ErrorCode == 0);
  ErrorCode = DAQmxCfgImplicitTiming (TimerHandle
                                     ,DAQmx_Val_ContSamps
                                     ,ArraySize);
//  qDebug() << " Kinetics Timer Configure Chanel, ErrorCode = " << ErrorCode;
  Error = Error || !(ErrorCode == 0);

  ErrorCode = DAQmxCreateTask ("", &CounterHandle);
  Error = Error || !(ErrorCode == 0);
  ErrorCode = DAQmxCreateCICountEdgesChan(CounterHandle,"Dev1/ctr1"
                                         ,"",DAQmx_Val_Rising
                                         ,0,DAQmx_Val_CountUp);
//  qDebug() << " Kinetics Counter Create Chanel, ErrorCode = " << ErrorCode;
  Error = Error || !(ErrorCode == 0);
  ErrorCode = DAQmxCfgSampClkTiming (CounterHandle,"/Dev1/Ctr0InternalOutput", 1/TimeInSec
                                    ,DAQmx_Val_Rising
                                    ,DAQmx_Val_FiniteSamps , ArraySize);
//  qDebug() << " Kinetics Counter Configure Chanel, ErrorCode = " << ErrorCode;
  Error = Error || !(ErrorCode == 0);

  ErrorCode = DAQmxRegisterDoneEvent (CounterHandle, 0
                                         , &RA::sNIDAQCounter::OnReadyToReadK, this);
//  qDebug() << " Kinetics Callback Initiated, ErrorCode = " << ErrorCode;
  Error = Error || !(ErrorCode == 0);

  ErrorCode = DAQmxSetArmStartTrigType (TimerHandle, DAQmx_Val_DigEdge);
//  qDebug() << " Kinetics Timer SetArmStartTrigType, ErrorCode = " << ErrorCode;
  Error = Error || !(ErrorCode == 0);
  ErrorCode = DAQmxSetDigEdgeArmStartTrigSrc (TimerHandle, ptr_to_const_char(StrobeTermName));
//  qDebug() << " Kinetics Timer SetArmStartTrigSrc, ErrorCode = " << ErrorCode;
  Error = Error || !(ErrorCode == 0);
  ErrorCode = DAQmxSetDigEdgeArmStartTrigEdge (TimerHandle, DAQmx_Val_Rising );
//  qDebug() << " Kinetics Timer SetArmStartTrigEdge, ErrorCode = " << ErrorCode;
  Error = Error || !(ErrorCode == 0);

  ErrorCode = DAQmxSetArmStartTrigType (CounterHandle, DAQmx_Val_DigEdge);
//  qDebug() << " Kinetics Counter SetArmStartTrigType, ErrorCode = " << ErrorCode;
  Error = Error || !(ErrorCode == 0);
  ErrorCode = DAQmxSetDigEdgeArmStartTrigSrc (CounterHandle, ptr_to_const_char(StrobeTermName));
//  qDebug() << " Kinetics Counter SetArmStartTrigSrc, ErrorCode = " << ErrorCode;
  Error = Error || !(ErrorCode == 0);
  ErrorCode = DAQmxSetDigEdgeArmStartTrigEdge (CounterHandle, DAQmx_Val_Rising );
//  qDebug() << " Kinetics Counter SetArmStartTrigEdge, ErrorCode = " << ErrorCode;
  Error = Error || !(ErrorCode == 0);

  if (!Error)
      ErrorCode = DAQmxStartTask(TimerHandle);
//  qDebug() << " Kinetics Timer Start Task, ErrorCode = " << ErrorCode;
  if (ErrorCode != 0) {
                        char EString[128];
                        DAQmxGetErrorString (ErrorCode, EString, 128);
                        qDebug() << EString;
                      };
  bool TimerStarted = (ErrorCode == 0);
  Error = Error || !(ErrorCode == 0);
  if (!Error)
      ErrorCode = DAQmxStartTask(CounterHandle);
//  qDebug() << " Kinetics Counter Start Task, ErrorCode = " << ErrorCode;
  if (ErrorCode != 0) {
                        char EString[128];
                        DAQmxGetErrorString (ErrorCode, EString, 128);
                        qDebug() << EString;
                      };
  bool CounterStarted = (ErrorCode == 0);
  Error = Error || !(ErrorCode == 0);

  if (Error) {
      if (TimerStarted) ErrorCode = DAQmxStopTask(TimerHandle);
      if (CounterStarted) ErrorCode = DAQmxStopTask(CounterHandle);
      if (TimerHandle != 0) ErrorCode = DAQmxClearTask (TimerHandle);
      if (CounterHandle != 0) ErrorCode = DAQmxClearTask (CounterHandle);
      emit ErrorMessage(tr("Fail to initialize NIDAQ-card for kinetics measurement"));
      emit CoolDown();
      KinMeasuring = false;
  } else{
      KinMeasuring = true;
      qDebug() << " Kinetics Successfully Started !";
  }
}

void  sNIDAQCounter::Start_StartStopCounting (sSSParams params)
{
  SSMode = params.Mode;
  StartStopCounts = params.NumCounts;
  PrevCount = 0;
  ArraySize = params.Size;
  readArray = new unsigned long [StartStopCounts];
  DataArray.SetSize(ArraySize);
  bool Error = false;
  long  ErrorCode = 0;
  ErrorCode = DAQmxCreateTask ("", &CounterHandle);
  Error = Error || !(ErrorCode == 0);
  ErrorCode = DAQmxCreateCITwoEdgeSepChan (CounterHandle,"Dev1/ctr1"
                                     ,"StartStopCounter",0,1000000,DAQmx_Val_Ticks
                                     ,DAQmx_Val_Rising,DAQmx_Val_Rising,NULL);
  qDebug() << " StartStop Counter Create Chanel, ErrorCode = " << ErrorCode;
  Error = Error || !(ErrorCode == 0);
  ErrorCode = DAQmxCfgImplicitTiming  (CounterHandle, DAQmx_Val_FiniteSamps, StartStopCounts);
  qDebug() << " StartStop Counter Configure Chanel, ErrorCode = " << ErrorCode;
  Error = Error || !(ErrorCode == 0);

  char InputName[128];

  if (SSMode == 1){
    ErrorCode = DAQmxSetCICtrTimebaseSrc(CounterHandle, "StartStopCounter"
                                        ,"/Dev1/20MHzTimebase");
    ErrorCode = DAQmxGetCICtrTimebaseSrc(CounterHandle, "StartStopCounter", InputName, 128);
  } else {
    ErrorCode = DAQmxSetCICtrTimebaseSrc(CounterHandle, "StartStopCounter"
                                        ,"/Dev1/80MHzTimebase");
    ErrorCode = DAQmxGetCICtrTimebaseSrc(CounterHandle, "StartStopCounter", InputName, 128);
  }
  qDebug() << " StartStop Counter Configure Source, ErrorCode = " << ErrorCode;
  qDebug() << "Mode - " << SSMode << " Input - " << InputName;
  Error = Error || !(ErrorCode == 0);

  ErrorCode = DAQmxRegisterDoneEvent (CounterHandle, 0
                                         , &RA::sNIDAQCounter::OnReadyToReadSS, this);
  qDebug() << " StartStop Callback Initiated, ErrorCode = " << ErrorCode;
  Error = Error || !(ErrorCode == 0);
  ErrorCode = DAQmxSetCITwoEdgeSepFirstTerm (CounterHandle,"StartStopCounter"
                                            ,ptr_to_const_char(StrobeTermName));
  qDebug() << " StartStop Counter Set Start Source, ErrorCode = " << ErrorCode;
  Error = Error || !(ErrorCode == 0);
  ErrorCode = DAQmxSetCITwoEdgeSepSecondTerm (CounterHandle,"StartStopCounter"
                                             ,ptr_to_const_char(SignalTermName));
  qDebug() << " StartStop Counter Set Stop Source, ErrorCode = " << ErrorCode;
  Error = Error || !(ErrorCode == 0);

  if (!Error)
      ErrorCode = DAQmxStartTask(CounterHandle);
  qDebug() << " StartStop Counter Start Task, ErrorCode = " << ErrorCode;
  if (ErrorCode != 0) {
                        char EString[128];
                        DAQmxGetErrorString (ErrorCode, EString, 128);
                        qDebug() << EString;
                      };
  bool CounterStarted = (ErrorCode == 0);
  Error = Error || !(ErrorCode == 0);
  if (Error) {
      if (CounterStarted) ErrorCode = DAQmxStopTask(CounterHandle);
      if (CounterHandle != 0) ErrorCode = DAQmxClearTask (CounterHandle);
      emit ErrorMessage(tr("Fail to initialize NIDAQ-card for start-stop measurement"));
      emit CoolDown();
      KinMeasuring = false;
  } else{
      KinMeasuring = true;
      qDebug() << " StartStop Successfully Started !";
  }
}

void  sNIDAQCounter::StopCounting ()
{
//  long  ErrorCode;
  if (TimerHandle != 0){
      /*ErrorCode =*/ DAQmxStopTask(TimerHandle);
      /*ErrorCode =*/ DAQmxClearTask (TimerHandle);
  }
  if (CounterHandle != 0){
      /*ErrorCode =*/ DAQmxStopTask(CounterHandle);
      /*ErrorCode =*/ DAQmxClearTask (CounterHandle);
  }
  if (readArray){
    delete [] readArray;
    readArray = NULL;
  }
  if (DataArray.Size() > 0){
    DataArray.SetSize(0);
  }
  Counting = false;
  KinMeasuring = false;
  qDebug() << " Counting Stoped !";
}

void sNIDAQCounter::ReadCounterSampling ()
{
  if (Counting){
//    qDebug() << " Read Counter Sampling";
    unsigned long CurrentCount = 0;
    long ErrorCode = DAQmxReadCounterScalarU32 (CounterHandle, 1.0, &CurrentCount, NULL);
    if (ErrorCode == 0){
        unsigned long C = CurrentCount - PrevCount;
        PrevCount = CurrentCount;
        sPhysValue Intensity("photon/s");
        Intensity = sPhysValue(1.0,"photon")*C/ThePeriod;
        emit NewData (Intensity);
    } else{
        StopCounting ();
        qDebug() << " Counter Read Error, ErrorCode = " << ErrorCode;
        char EString[128];
        DAQmxGetErrorString (ErrorCode, EString, 128);
        qDebug() << EString;
        emit ErrorMessage(" Read Counter: " + QString(EString));
    }
  }
}

void sNIDAQCounter::ReadSingleCount ()
{
  if (Counting){
//    qDebug() << " Read Single Count";
    unsigned long _counts[2];
    long PNumRead = 0;
    long ErrorCode = DAQmxReadCounterU32 (CounterHandle, 2, 1.0, _counts, 2, &PNumRead, NULL);
    if (ErrorCode == 0){
        unsigned long C = _counts[1] - _counts[0];
        sPhysValue Intensity("photon/s");
        Intensity = sPhysValue(1.0,"photon")*C/ThePeriod;
        emit GotSingleCount (Intensity);
    } else{
        StopCounting ();
        qDebug() << " Counter Read Error, ErrorCode = " << ErrorCode;
        char EString[128];
        DAQmxGetErrorString (ErrorCode, EString, 128);
        qDebug() << EString;
        emit ErrorMessage(" Read Counter: " + QString(EString));
    }
  }
}

void sNIDAQCounter::ReadKineticsSampling ()
{
  if (KinMeasuring){
//    qDebug() << " Begin to read Kinetics Sampling";
    bool Error = true;
    long PNumRead = 0;
    long ErrorCode = DAQmxReadCounterU32 (CounterHandle, ArraySize, 1.0, readArray, ArraySize, &PNumRead, NULL);
//    DAQmxStopTask(TimerHandle);
    DAQmxStopTask(CounterHandle);
    real TimeInSec = (ThePeriod/(sPhysValue(1.0,"s"))).ToReal();
    if ((ErrorCode == 0) && (ArraySize == PNumRead)){
        for (int I = 0; I < ArraySize; I++){
          long C = readArray[I] - PrevCount;
          PrevCount = readArray[I];
          real Intensity = C;
          Intensity = Intensity/TimeInSec;
          DataArray.ArrBody[I] = Intensity;
        }
        emit NewDataArray (DataArray);
        PrevCount = 0;
        ErrorCode = DAQmxStartTask(CounterHandle);
        Error = !(ErrorCode == 0);
    }
    if (Error) {
        StopCounting ();
        qDebug() << " Counter Read Error, ErrorCode = " << ErrorCode;
        qDebug() << " Want to read = " << ArraySize << "Actually read" << PNumRead;
        char EString[128];
        DAQmxGetErrorString (ErrorCode, EString, 128);
        qDebug() << EString;
        emit ErrorMessage(" Read Kinetics: " + QString(EString));
        return;
    }
  }
}

void sNIDAQCounter::ReadStartStopSampling ()
{
  if (KinMeasuring){
      double TimeStep = 0.0125*0.000001;
      if (SSMode == 1) TimeStep = 0.05*0.000001;
      long PNumRead = 0;
      long ErrorCode = DAQmxReadCounterU32 (CounterHandle, -1, 0, readArray
                                           ,2*StartStopCounts, &PNumRead, NULL);
      bool ReadError = !(ErrorCode == 0);
      if (ReadError) qDebug() << " Error reading Start-Stop Array !!! ";
      DAQmxStopTask(CounterHandle);
      if (!ReadError) {
          for (int I = 0; I < ArraySize; I++){
              DataArray.ArrBody[I] = 0;
          }
          for (int I = 0; I < PNumRead; I++){
               int N = readArray[I];
               if ((N < ArraySize) && (N >= 0))
                 DataArray.ArrBody[N] = DataArray.ArrBody[N] + 1.0/TimeStep/StartStopCounts;
          }
          emit NewDataArray (DataArray);
      }
      ErrorCode = DAQmxStartTask(CounterHandle);
      bool Error = !(ErrorCode == 0);
      if (Error) {
          char EString[128];
          DAQmxGetErrorString (ErrorCode, EString, 128);
          qDebug() << EString;
          StopCounting ();
          emit ErrorMessage(" Read Start-Stop: " + QString(EString));
          return;
      }
  }
}

int32  CVICALLBACK sNIDAQCounter::OnReadyToReadC (TaskHandle /*theHandle*/
                                                   ,int32      /*eventType*/
                                                   ,uInt32     /*nSamples*/
                                                   ,void *     DataPtr)
{
  sNIDAQCounter* obj = static_cast<sNIDAQCounter*>(DataPtr);
  QMetaObject::invokeMethod(obj, "ReadCounterSampling",Qt::QueuedConnection);
  return 0;
}

int32  CVICALLBACK sNIDAQCounter::OnReadyToReadSC (TaskHandle /*theHandle*/
                                                  ,int32      /*OK*/
                                                  ,void *     DataPtr)
{
  sNIDAQCounter* obj = static_cast<sNIDAQCounter*>(DataPtr);
  QMetaObject::invokeMethod(obj, "ReadSingleCount", Qt::QueuedConnection);

  return 0;
}

int32  CVICALLBACK sNIDAQCounter::OnReadyToReadK (TaskHandle /*theHandle*/
                                                   ,int32      /*OK*/
                                                   ,void *     DataPtr)
{
  sNIDAQCounter* obj = static_cast<sNIDAQCounter*>(DataPtr);
  QMetaObject::invokeMethod(obj, "ReadKineticsSampling", Qt::QueuedConnection);

  return 0;
}

int32  CVICALLBACK sNIDAQCounter::OnReadyToReadSS (TaskHandle /*theHandle*/
                                                    ,int32      /*OK*/
                                                    ,void *     DataPtr)
{
  sNIDAQCounter* obj = static_cast<sNIDAQCounter*>(DataPtr);
  QMetaObject::invokeMethod(obj, "ReadStartStopSampling", Qt::QueuedConnection);

  return 0;
}

//------------------------------------------------------------------------------
} //namespace RA
