////////////////////////////////////////////////////////////////////////////////
/*! @file NIDAQDoubleCounter.cpp  Использование двух счетчиков платы NIDAQ
          для счета сигналов. Время счета задается микроконтроллером.
- Part of NIDAQmxDrivers - a Research Assistant Plugin package.
- Uses  RAGUI - the Research Assistant Graphical User Interface.
- Uses  QtGui v.4.6  -  http://qt.nokia.com/
- Uses  RANet - Research Assistant Net Library (based on ANSI C++).
- Copyright(C) 2010-2016, Dmitry K. Nelson, St.Petersburg, Russia.
                          D.Nelson@mail.ioffe.ru
*///////////////////////////////////////////////////////////////////////////////
#include "NIDAQDoubleCounter.h"
#include "attr.h"
#include <QDebug>
namespace RA {
//------------------------------------------------------------------------------

DEFINE_CLASS_TAG(sNIDAQDoubleCounter)

/*! Директива изменения режима измерения
1 - сигналом NewData передаются данные счетчика A
2 - сигналом NewData передаются данные счетчика B
3 - сигналом NewData передается величина степени поляризации (A-B)/(A+B)
*/
const sDirective<int>  sNIDAQDoubleCounter::cChangeCountingMode("cChangeCountingMode");



//------------------------------------------------------------ sNIDAQDoubleCounter ---

void  sNIDAQDoubleCounter::Preconfigure (sNodePtr  DeviceCfg)
{
  sPtr<sFolderNode>              DeviceSettings;
  sPtr<sFolderNode>              DeviceLinks;
  DeviceCfg |= "Links";
  DeviceLinks = DeviceCfg >> "Links";
  if( DeviceLinks.IsCorrect() ){
    DeviceLinks |= "TimerPort";
    DeviceLinks >> "TimerPort" |= attr("Class","sTeleCom");
  }
  DeviceCfg |= "Settings";
  DeviceSettings = DeviceCfg >> "Settings";
  if( DeviceSettings.IsCorrect() ){
    DeviceSettings |= attr("TimerName","@MSDN204C");
    DeviceSettings |= attr("Timer1TerminalName","/Dev1/PFI11");
    DeviceSettings |= attr("Timer0TerminalName","/Dev1/PFI10");
  }
  sGauge::Preconfigure(DeviceCfg);
}


sNIDAQDoubleCounter::~sNIDAQDoubleCounter ()
{
  StopCounting();
  TimerPort.SetToNULL();
}


sNIDAQDoubleCounter::sNIDAQDoubleCounter (sString   DevicetId
                       ,sNodePtr  DeviceCfg)
    :sGauge(DevicetId,DeviceCfg)
    ,TimerPort(this)
    ,Counting(false)
    ,PrevCount1 (0)
    ,PrevCount0 (0)
    ,Count1(0)
    ,Count0(0)
    ,C1Ready(false)
    ,C0Ready(false)
    ,Counter1Handle(0)
    ,Counter0Handle(0)
    ,MeasuringMode(0)
    ,Timer1TermName("")
    ,Timer0TermName("")
{
}


void  sNIDAQDoubleCounter::PostConstructor ()
{
  sGauge::PostConstructor();
  try{ //get config parameters
    get_from<sTextNode>(DeviceSettings >> "TimerName" ,TimerName);
    get_from<sTextNode>(DeviceSettings >> "Timer1TerminalName" ,Timer1TermName);
    get_from<sTextNode>(DeviceSettings >> "Timer0TerminalName" ,Timer0TermName);
  }catch(...){
    SetInvalid(tr("Something wrong in configuration."));
  }
  WantedPeriod = ThePeriod;
  MeasuringMode = 0;

  //evolution of the inherited state machine (behaviour)
  {
    //evolution of the State_Warming
    {
      State_W_TimerPortLinking = new sState(State_Warming()
                                    ,"State_W_TimerPortLinking");
      State_W_TimerPortChecking = new sState(State_Warming()
                                     ,sPhysValue(1,"s")
                                     ,"State_W_TimerPortChecking");
      State_W_SettingPeriod = new sState(State_Warming()
                                 ,"State_W_SettingPeriod");
      State_W_TimerPortLinking
         ->SetAsInitial()
         ->OnEnterCall(SLOT(LinkToTimerPort()))
         ->AddTransition( SIGNAL(TimerPortLinked()) ,State_W_TimerPortChecking )
         ->AddTransition( SIGNAL(ErrorMessage()) ,State_Cooling() )
         ;
      State_W_TimerPortChecking
         ->OnEnterCall(SLOT(AskTimerName()))
         ->AddTransition<sTelex>(SIGNAL(GotTimerMessage(sTelex))
                                ,State_W_SettingPeriod
                                ,sTelex(TimerName))
         ->AddTransition(SIGNAL(NoAnswerFromTimer())
                        ,SLOT(OnNoAnswerFromTimer())
                        ,State_Cooling() )
         ->AddTransition(SIGNAL(ErrorMessage()) ,State_Cooling() )
         ;
      State_W_SettingPeriod
          ->OnEnterCall(SLOT(SetWantedPeriod()))
          ->AddReaction<sTelex>(SIGNAL(GotTimerMessage(sTelex))
                               ,SLOT(CorrectPeriod(sTelex)) )
          ->AddTransition(SIGNAL(PeriodChanged(sPhysValue))
                         ,State_Warming()->Fin())
          ;
    connect(State_W_TimerPortChecking,SIGNAL(Timeout())
           ,this,SIGNAL(NoAnswerFromTimer()));
    }
    //evolution of the State_Operating
    {
      State_O_M_SetP_Checking = new sState(State_O_M_SettingPeriod
                                         ,"State_O_M_SetP_Checking");
      State_O_M_SetP_Checking
          ->SetAsInitial()
          ->AddTransition<sTelex>(SIGNAL(GotTimerMessage(sTelex))
                                 ,SLOT(CorrectPeriod(sTelex))
                                 ,State_O_M_SettingPeriod->Fin() )
              ;
      State_O_M_Stopped
          ->AddReaction<int>(cChangeCountingMode
                            ,SLOT(SetCountingMode(int)));
      State_O_M_Acquiring
          ->OnEnterCall(SLOT(StartCounting()))
          ->OnExitCall(SLOT(StopCounting()))
          ->AddTransition<sPhysValue>(SIGNAL(GotSingleCount(sPhysValue))
                                     ,SIGNAL(NewData(sPhysValue))
                                     ,State_O_M_Cycling);
    }
    connect(this ,SIGNAL(TimerPortError(QString))
           ,this ,SLOT(ReportTimerPortError(QString)) );
  }
}

void  sNIDAQDoubleCounter::SetInitialMode (int _mode)
{
  MeasuringMode =  _mode;
}

void  sNIDAQDoubleCounter::OnEnterToCoolingState ()
{
  Disconnect(TimerPort ,0 ,0);
  TimerPort.ReleaseExclusiveControl();
  TimerPort.SetToNULL();
  sGauge::OnEnterToCoolingState();
}

void  sNIDAQDoubleCounter::SetWantedPeriod ()
{
  try{
    real TimeInMicroSec; // Период передается в МК в виде целого числа микросекунд.
    TimeInMicroSec = (WantedPeriod/(sPhysValue(1.0e-6,"s"))).ToReal();
    int TimePar = Round(TimeInMicroSec);
    QByteArray _comm;
    _comm.append(QByteArray::number((qlonglong)TimePar, 16).toUpper().rightJustified(8, '0'));
    _comm.prepend("T");
    SendDirective<sTelex>(TimerPort ,sTeleCom::cSendTelex ,sTelex(_comm));
  }catch(rxRAlgebra){
    //!@todo{Instruments} define Failure signal for drivers?
    emit ErrorMessage(tr("Cannot set period, period=%1")
                     .arg(ToQString(WantedPeriod.Text())));
    emit CoolDown();
  }
}

void  sNIDAQDoubleCounter::ReportTimerPortError (QString  error_text)
{
  emit ErrorMessage(tr("Got error from %1:"
                       "\n\n%2")
                      .arg(ToQString(TimerPort->DeviceIdentifier()))
                      .arg(error_text)     );
}

void  sNIDAQDoubleCounter::LinkToTimerPort ()
{
  try{
    TimerPort.SetTo(DeviceLinks,"TimerPort");
    if( ! TimerPort.TakeExclusiveControl() ){
      emit ErrorMessage(tr("Cannot take exclusive control over '%1'. "
                           "Try again later.")
                       .arg(ToQString( TimerPort->DeviceIdentifier() )) );
    }else{
      Connect(TimerPort ,SIGNAL(Warmed())
             ,SIGNAL(TimerPortLinked()) );
      Connect(TimerPort ,SIGNAL(ErrorMessage(QString))
             ,SIGNAL(TimerPortError(QString)) );
      Connect(TimerPort ,SIGNAL(GotTelex(sTelex))
             ,SIGNAL(GotTimerMessage(sTelex)) );
    }
  }catch( rxDeviceLaunch E ){
    emit ErrorMessage( tr("Error while linking to a submissive device.\n"
                          "%1\nCheck configuration.")
                         .arg(ToQString(E.MsgText()))   );
  }
}

void  sNIDAQDoubleCounter::OnNoAnswerFromTimer ()
{
  emit ErrorMessage("No answer from Timer of Double NIDAQ counter");
}

void  sNIDAQDoubleCounter::AskTimerName ()
{
  SendDirective(TimerPort  ,sTeleCom::cSendTelex ,sTelex("?") );
}


void  sNIDAQDoubleCounter::CorrectPeriod (sTelex text)
{
  bool OK;
  int TimeInMicroSec = text.toLong(&OK,16);
  if( OK ){
    ThePeriod = sPhysValue(TimeInMicroSec*1.0e-6,"s");
    emit PeriodChanged(ThePeriod);
  }else{
    emit ErrorMessage(QString("Unknown telex: ")+text);
  }
}

void  sNIDAQDoubleCounter::SetCountingMode (int _theMode)
{
  MeasuringMode = _theMode;
}

void  sNIDAQDoubleCounter::StartCounting ()
{
  if (AutoRepeat()) StartContCounting ();
  else StartSingleCount ();
}

void  sNIDAQDoubleCounter::StartContCounting ()
{
  PrevCount1 = 0;
  PrevCount0 = 0;
  Count1 = 0;
  Count0 = 0;
  C1Ready =false;
  C0Ready =false;
  bool Error;
  long  ErrorCode;
  ErrorCode = DAQmxCreateTask ("", &Counter1Handle);
  Error = !(ErrorCode == 0);
  ErrorCode = DAQmxCreateCICountEdgesChan(Counter1Handle,"Dev1/ctr1"
                                         ,"",DAQmx_Val_Rising
                                         ,0,DAQmx_Val_CountUp);
//  qDebug() << " Counter Create Chanel, ErrorCode = " << ErrorCode;
  Error = Error || !(ErrorCode == 0);
  ErrorCode = DAQmxCfgSampClkTiming (Counter1Handle,ptr_to_const_char(Timer1TermName), 1000
                                    ,DAQmx_Val_Rising
                                    ,DAQmx_Val_ContSamps , 1000);
  Error = Error || !(ErrorCode == 0);

  ErrorCode = DAQmxRegisterEveryNSamplesEvent (Counter1Handle
                                        ,DAQmx_Val_Acquired_Into_Buffer
                                        ,1, 0, &RA::sNIDAQDoubleCounter::OnReadyToReadC1, this);
//  qDebug() << " Counter1 Callback Initiated, ErrorCode = " << ErrorCode;
  if (ErrorCode != 0) {
                        char EString[128];
                        DAQmxGetErrorString (ErrorCode, EString, 128);
                        qDebug() << EString;
                      };
  Error = Error || !(ErrorCode == 0);

  ErrorCode = DAQmxCreateTask ("", &Counter0Handle);
  Error = !(ErrorCode == 0);
  ErrorCode = DAQmxCreateCICountEdgesChan(Counter0Handle,"Dev1/ctr0"
                                         ,"",DAQmx_Val_Rising
                                         ,0,DAQmx_Val_CountUp);
//  qDebug() << " Counter Create Chanel, ErrorCode = " << ErrorCode;
  Error = Error || !(ErrorCode == 0);
  ErrorCode = DAQmxCfgSampClkTiming (Counter0Handle,ptr_to_const_char(Timer0TermName), 1000
                                    ,DAQmx_Val_Rising
                                    ,DAQmx_Val_ContSamps , 1000);
  Error = Error || !(ErrorCode == 0);

  ErrorCode = DAQmxRegisterEveryNSamplesEvent (Counter0Handle
                                        ,DAQmx_Val_Acquired_Into_Buffer
                                        ,1, 0, &RA::sNIDAQDoubleCounter::OnReadyToReadC0, this);
//  qDebug() << " Counter1 Callback Initiated, ErrorCode = " << ErrorCode;
  if (ErrorCode != 0) {
                        char EString[128];
                        DAQmxGetErrorString (ErrorCode, EString, 128);
                        qDebug() << EString;
                      };
  Error = Error || !(ErrorCode == 0);

  if (!Error)
      ErrorCode = DAQmxStartTask(Counter1Handle);
  bool Counter1Started = (ErrorCode == 0);
  Error = Error || !(ErrorCode == 0);
  if (!Error)
      ErrorCode = DAQmxStartTask(Counter0Handle);
  bool Counter0Started = (ErrorCode == 0);
  Error = Error || !(ErrorCode == 0);

  // Запуск таймера в микроконтроллере
  SendDirective<sTelex>(TimerPort ,sTeleCom::cSendTelex ,sTelex("S"));

  if (Error) {
      if (Counter1Started) ErrorCode = DAQmxStopTask(Counter1Handle);
      if (Counter0Started) ErrorCode = DAQmxStopTask(Counter0Handle);
      if (Counter1Handle != 0) ErrorCode = DAQmxClearTask (Counter1Handle);
      if (Counter0Handle != 0) ErrorCode = DAQmxClearTask (Counter0Handle);
      // Остановка таймера в микроконтроллере
      SendDirective<sTelex>(TimerPort ,sTeleCom::cSendTelex ,sTelex("X"));
      emit ErrorMessage(tr("Cannot start counting. NIDAQ-card is absent or out of service."));
      emit CoolDown();
      qDebug() << "Cooling Down";
      Counting = false;
  } else{
      Counting = true;
      qDebug() << " Counting Successfully Started !";
  }
}

void  sNIDAQDoubleCounter::StartSingleCount ()
{
  emit ErrorMessage(tr("Single count is not realised yet"));
}

void  sNIDAQDoubleCounter::StopCounting ()
{
  // Остановка таймера в микроконтроллере
  SendDirective<sTelex>(TimerPort ,sTeleCom::cSendTelex ,sTelex("X"));
  if (Counter1Handle != 0){
      DAQmxStopTask(Counter1Handle);
      DAQmxClearTask (Counter1Handle);
  }
  if (Counter0Handle != 0){
      DAQmxStopTask(Counter0Handle);
      DAQmxClearTask (Counter0Handle);
  }
  Counting = false;
  qDebug() << " Counting Stoped !";
}

void sNIDAQDoubleCounter::ReadCounter1Sampling ()
{
  if (Counting){
//    qDebug() << " Read Counter1 Sampling";
    unsigned long CurrentCount1 = 0;
    long ErrorCode = DAQmxReadCounterScalarU32 (Counter1Handle, 1.0, &CurrentCount1, NULL);
    if (ErrorCode == 0){
        Count1 = CurrentCount1 - PrevCount1;
        PrevCount1 = CurrentCount1;
        C1Ready = true;
    } else{
        StopCounting ();
        qDebug() << " Counter1 Read Error, ErrorCode = " << ErrorCode;
        char EString[128];
        DAQmxGetErrorString (ErrorCode, EString, 128);
        qDebug() << EString;
        emit ErrorMessage(" Read Counter1: " + QString(EString));
    }
    if ((C1Ready) && (C0Ready)) CalcOutput ();
  }
}

void sNIDAQDoubleCounter::ReadCounter0Sampling ()
{
  if (Counting){
//    qDebug() << " Read Counter0 Sampling";
    unsigned long CurrentCount0 = 0;
    long ErrorCode = DAQmxReadCounterScalarU32 (Counter0Handle, 1.0, &CurrentCount0, NULL);
    if (ErrorCode == 0){
        Count0 = CurrentCount0 - PrevCount0;
        PrevCount0 = CurrentCount0;
        C0Ready = true;
    } else{
        StopCounting ();
        qDebug() << " Counter0 Read Error, ErrorCode = " << ErrorCode;
        char EString[128];
        DAQmxGetErrorString (ErrorCode, EString, 128);
        qDebug() << EString;
        emit ErrorMessage(" Read Counter0: " + QString(EString));
    }
    if ((C1Ready) && (C0Ready)) CalcOutput ();
  }
}

void sNIDAQDoubleCounter::CalcOutput ()
{
    C1Ready = false;
    C0Ready = false;
    sPhysValue Intensity(1.0,"photon/s");
    sPhysValue Polarization(1.0,"%");
    switch (MeasuringMode) {
      case(0):
        Intensity = sPhysValue(1.0,"photon")*Count1/ThePeriod;
        emit NewData (Intensity);
        Intensity = sPhysValue(1.0,"photon")*Count0/ThePeriod;
        emit NewData_2 (Intensity);
        break;
      case(1):
        Intensity = sPhysValue(1.0,"photon")*Count0/ThePeriod;
        emit NewData (Intensity);
        Intensity = sPhysValue(1.0,"photon")*Count1/ThePeriod;
        emit NewData_2 (Intensity);
        break;
        break;
      case(2):
        long _I = Count1+Count0;
        if (_I == 0) { Polarization = sPhysValue(0,"%"); }
        else { Polarization = sPhysValue(100,"%")*(Count1-Count0)/_I; }
        emit NewData (Polarization);
        Intensity = sPhysValue(1.0,"photon")*_I/ThePeriod;
        emit NewData_2 (Intensity);
        break;
    }
}

int32  CVICALLBACK sNIDAQDoubleCounter::OnReadyToReadC1 (TaskHandle /*theHandle*/
                                                   ,int32      /*eventType*/
                                                   ,uInt32     /*nSamples*/
                                                   ,void *     DataPtr)
{
  sNIDAQDoubleCounter* obj = static_cast<sNIDAQDoubleCounter*>(DataPtr);
  QMetaObject::invokeMethod(obj, "ReadCounter1Sampling",Qt::QueuedConnection);
  return 0;
}

int32  CVICALLBACK sNIDAQDoubleCounter::OnReadyToReadC0 (TaskHandle /*theHandle*/
                                                   ,int32      /*eventType*/
                                                   ,uInt32     /*nSamples*/
                                                   ,void *     DataPtr)
{
  sNIDAQDoubleCounter* obj = static_cast<sNIDAQDoubleCounter*>(DataPtr);
  QMetaObject::invokeMethod(obj, "ReadCounter0Sampling",Qt::QueuedConnection);
  return 0;
}

//------------------------------------------------------------------------------
} //namespace RA
