////////////////////////////////////////////////////////////////////////////////
/*! @file NIDAQVoltmeter.cpp   Вольтметр на базе платы NIDAQ.
- Part of NIDAQmxDrivers - a Research Assistant Plugin package.
- Uses  RAGUI - the Research Assistant Graphical User Interface.
- Uses  QtGui v.5.0  -  http://qt.nokia.com/
- Uses  RANet - Research Assistant Net Library (based on ANSI C++).
- Copyright(C) 2010-2014, Dmitry K. Nelson, St.Petersburg, Russia.
                          D.Nelson@mail.ioffe.ru
*///////////////////////////////////////////////////////////////////////////////
#include "NIDAQVoltmeter.h"
#include "attr.h"
#include "StringsAndQStrings.h"
#include <QDebug>
namespace RA {
//------------------------------------------------------------------------------

DEFINE_CLASS_TAG(sNIDAQVoltmeter)

//---------------------------------------------------------- sNIDAQVoltmeter ---

void  sNIDAQVoltmeter::Preconfigure (sNodePtr  DeviceCfg)
{
  sPtr<sFolderNode>              DeviceSettings;
  DeviceCfg |= "Settings";
  DeviceSettings = DeviceCfg >> "Settings";
  if( DeviceSettings.IsCorrect() ){
    DeviceSettings |= attr("SamplingPeriod",sPhysValue(0.01,"s"));
    DeviceSettings |= attr("MeasuringRange",sPhysRange(-10.0,10.0,"V"));
  }
  sGauge::Preconfigure(DeviceCfg);
}


sNIDAQVoltmeter::~sNIDAQVoltmeter ()
{
  Stop();
}


sNIDAQVoltmeter::sNIDAQVoltmeter (sString   DevicetId
                       ,sNodePtr  DeviceCfg)
    :sGauge(DevicetId,DeviceCfg)
    ,SamplingPeriod(sPhysValue("s"))
    ,VoltmeterHandle(0)
    ,NumOfSamples(0)
    ,readArray(NULL)
{
}


void  sNIDAQVoltmeter::PostConstructor ()
{
  sGauge::PostConstructor();
  
  if( DeviceSettings.IsCorrect() ){ //get config parameters
      get_from<sPhysValueNode>(DeviceSettings >> "SamplingPeriod" ,SamplingPeriod);
      get_from<sPhysRangeNode>(DeviceSettings >> "MeasuringRange" ,MeasuringRange);
  }
  //! Реализован только режим AutoRepeat:
  AutoRepeatFlag = true;

  //evolution of the inherited state machine (behaviour)
  {
    //evolution of the State_Operating
    {
      State_O_M_Acquiring
          ->OnEnterCall(SLOT(Start()))
          ->OnExitCall(SLOT(Stop()))
          ->AddTransition<sPhysValue>(SIGNAL(GotSingleShot(sPhysValue))
                                     ,SIGNAL(NewData(sPhysValue))
                                     ,State_O_M_Cycling);
    };
  };
}

void  sNIDAQVoltmeter::Start ()
{
  sPhysValue  Min(_Unitsless_);
  sPhysValue  Max(_Unitsless_);
  sPhysValue  Smpls(_Unitsless_);
  try{
    Min = MeasuringRange.From()/sPhysValue(1.0,"V");
    Max = MeasuringRange.To()/sPhysValue(1.0,"V");
    Smpls = ThePeriod/SamplingPeriod;
  }catch(rxRAlgebra){
    emit ErrorMessage(tr("NIDAQVoltmeter.MeasuringRange: Wrong units in the configuration settings."));
    SetInvalid();
    return;
  };
  real MinV = Min.ToReal();
  real MaxV = Max.ToReal();
  NumOfSamples = Round(Smpls.ToReal());
  float64 _SamplingRate = (sPhysValue(1.0,"s")*NumOfSamples/ThePeriod).ToReal();
  readArray = new float64 [NumOfSamples];
  bool Error;
  long  ErrorCode;

  ErrorCode = DAQmxCreateTask ("", &VoltmeterHandle);
  Error = !(ErrorCode == 0);
  ErrorCode = DAQmxCreateAIVoltageChan (VoltmeterHandle,"Dev1/ai0"
                                       ,"",DAQmx_Val_RSE,MinV,MaxV
                                       ,DAQmx_Val_Volts,NULL );
//  qDebug() << " Voltmeter Create Chanel, ErrorCode = " << ErrorCode;
  Error = Error || !(ErrorCode == 0);
  if (AutoRepeat()) {
    ErrorCode = DAQmxCfgSampClkTiming (VoltmeterHandle,NULL, _SamplingRate
                                      ,DAQmx_Val_Rising
                                      ,DAQmx_Val_ContSamps , 10*NumOfSamples);
    Error = Error || !(ErrorCode == 0);

//    qDebug() << "Setting SampleRate" << _SamplingRate;
//    DAQmxGetSampClkRate(VoltmeterHandle, &_SamplingRate);
//    qDebug() << "Getting SampleRate" << _SamplingRate;

    Error = DAQmxSetBufInputBufSize(VoltmeterHandle,10*NumOfSamples);
    Error = Error || !(ErrorCode == 0);

    ErrorCode = DAQmxRegisterEveryNSamplesEvent (VoltmeterHandle
                                          ,DAQmx_Val_Acquired_Into_Buffer
                                          ,NumOfSamples, 0, &RA::sNIDAQVoltmeter::OnReadyToRead, this);
//    qDebug() << " Voltmeter Callback Initiated, ErrorCode = " << ErrorCode;
    Error = Error || !(ErrorCode == 0);
  } else {
    ErrorCode = DAQmxCfgSampClkTiming (VoltmeterHandle,NULL, _SamplingRate
                                      ,DAQmx_Val_Rising
                                      ,DAQmx_Val_FiniteSamps , NumOfSamples);
    Error = Error || !(ErrorCode == 0);

//    qDebug() << "Setting SampleRate" << _SamplingRate;
//    DAQmxGetSampClkRate(VoltmeterHandle, &_SamplingRate);
//    qDebug() << "Getting SampleRate" << _SamplingRate;

    Error = Error || !(ErrorCode == 0);

    ErrorCode = DAQmxRegisterDoneEvent (VoltmeterHandle, 0
                                       , &RA::sNIDAQVoltmeter::OnReadyToReadSS, this);
//    qDebug() << " Voltmeter Callback Initiated, ErrorCode = " << ErrorCode;
    Error = Error || !(ErrorCode == 0);
  }

  if (!Error)
      ErrorCode = DAQmxStartTask(VoltmeterHandle);
  Error = Error || !(ErrorCode == 0);
  qDebug() << " Voltmeter StartTask, ErrorCode = " << ErrorCode;
  if (ErrorCode != 0) {
                        char EString[128];
                        DAQmxGetErrorString (ErrorCode, EString, 128);
                        qDebug() << EString;
                      };
  if (Error) {
      ErrorCode = DAQmxStopTask(VoltmeterHandle);
      ErrorCode = DAQmxClearTask (VoltmeterHandle);
      emit ErrorMessage(tr("Cannot start Voltmeter. NIDAQ-card is absent or out of service."));
      emit CoolDown();
      qDebug() << "Cooling Down";
  } else{
      qDebug() << " Voltmeter Successfully Started !";
  }
}

void  sNIDAQVoltmeter::Stop ()
{
//  long  ErrorCode;
  if (VoltmeterHandle != 0){
      /*ErrorCode =*/ DAQmxStopTask(VoltmeterHandle);
      /*ErrorCode =*/ DAQmxClearTask (VoltmeterHandle);
  }
  if (readArray){
    delete [] readArray;
    readArray = NULL;
  }
  NumOfSamples = 0;
  qDebug() << " Voltmeter Stoped !";
}

void sNIDAQVoltmeter::ReadSampling ()
{
//  qDebug() << " Read Voltmeter Sampling";
  int32 ReallyRead = 0;
  long ErrorCode = DAQmxReadAnalogF64  (VoltmeterHandle, NumOfSamples, 1.0
                                      , DAQmx_Val_GroupByChannel, readArray
                                      , NumOfSamples,&ReallyRead, NULL);
  if (ErrorCode == 0){
      real _Data = 0;
      for (int I = 0; I < ReallyRead; I++){
          _Data = _Data + readArray[I];
      }
      sPhysValue Intensity(1.0,"V");
      Intensity = Intensity * _Data / ReallyRead;
      emit NewData (Intensity);
      if (ReallyRead < NumOfSamples){
          qDebug() << "Really read less than ordered";
        emit ErrorMessage(" Read Voltmeter: Really read less than ordered");
      };
  } else{
      Stop ();
      qDebug() << " Voltmeter Read Error, ErrorCode = " << ErrorCode;
      char EString[128];
      DAQmxGetErrorString (ErrorCode, EString, 128);
      qDebug() << EString;
      emit ErrorMessage(" Read Voltmeter: " + QString(EString));
  }
}

void sNIDAQVoltmeter::ReadSingleShotSampling ()
{
//  qDebug() << " Read Voltmeter Single Shot Sampling";
  int32 ReallyRead = 0;
  long ErrorCode = DAQmxReadAnalogF64  (VoltmeterHandle, NumOfSamples, 1.0
                                      , DAQmx_Val_GroupByChannel, readArray
                                      , NumOfSamples,&ReallyRead, NULL);
  if (ErrorCode == 0){
      real _Data = 0;
      for (int I = 0; I < ReallyRead; I++){
          _Data = _Data + readArray[I];
      }
      sPhysValue Intensity(1.0,"V");
      Intensity = Intensity * _Data / ReallyRead;
      emit GotSingleShot (Intensity);
      if (ReallyRead < NumOfSamples){
          qDebug() << "Really read less than ordered";
        emit ErrorMessage(" Read Voltmeter: Really read less than ordered");
      };
  } else{
      Stop ();
      qDebug() << " Voltmeter Read Error, ErrorCode = " << ErrorCode;
      char EString[128];
      DAQmxGetErrorString (ErrorCode, EString, 128);
      qDebug() << EString;
      emit ErrorMessage(" Read Voltmeter: " + QString(EString));
  }
}

int32  CVICALLBACK sNIDAQVoltmeter::OnReadyToRead (TaskHandle /*theHandle*/
                                                  ,int32      /*eventType*/
                                                  ,uInt32     /*nSamples*/
                                                  ,void *     DataPtr)
{
  sNIDAQVoltmeter* obj = static_cast<sNIDAQVoltmeter*>(DataPtr);
  QMetaObject::invokeMethod(obj, "ReadSampling",Qt::QueuedConnection);
  return 0;
}

int32  CVICALLBACK sNIDAQVoltmeter::OnReadyToReadSS (TaskHandle /*theHandle*/
                                                    ,int32      /*OK*/
                                                    ,void *     DataPtr)
{
  sNIDAQVoltmeter* obj = static_cast<sNIDAQVoltmeter*>(DataPtr);
  QMetaObject::invokeMethod(obj, "ReadSingleShotSampling",Qt::QueuedConnection);
  return 0;
}

//------------------------------------------------------------------------------
} //namespace RA
