////////////////////////////////////////////////////////////////////////////////
/*! @file NIDAQmxDrivers.cpp Пакет драйверов и консолей для работы
                             с платой NIDAQ.
- Part of NIDAQmxDrivers - a Research Assistant Plugin package.
- Uses  RAGUI - the Research Assistant Graphical User Interface.
- Uses  QtGui v.4.6  -  http://qt.nokia.com/
- Uses  RANet - Research Assistant Net Library (based on ANSI C++).
- Copyright(C) 2010, Viktor E. Bursian, St.Petersburg, Russia.
                     Viktor.Bursian@mail.ioffe.ru
- Copyright(C) 2010-2013, Dmitry K. Nelson, St.Petersburg, Russia.
                          D.Nelson@mail.ioffe.ru
*///////////////////////////////////////////////////////////////////////////////
#include "NIDAQmxDrivers.h"
#include "attr.h"
#include "NIDAQCounter.h"
#include "NIDAQDoubleCounter.h"
#include "NIDAQVoltmeter.h"
#include "P138Magnet.h"
#include "FixedUnitsLCD.h"
#include "DoubleCounterDisplay.h"
#include "KineticsNIDAQ.h"
#include "StartStopNIDAQ.h"
#include "PeriodicGaugeRecorder.h"
#include <QMetaType>

namespace RA {
//------------------------------------------------------------------------------

DEFINE_CLASS_TAG(sNIDAQmxDrivers)

//------------------------------------------------------------------------------

sVersion  sNIDAQmxDrivers::Version ()
{
  return sVersion(RAP_VERS_MAJOR
                 ,RAP_VERS_MINOR
                 ,ZAKAVYCH(RAP_VERS_PATCH)
                 ,__DATE__
                 );
}

sVersion  sNIDAQmxDrivers::CoreVersionUsed ()
{
  return sVersion( RA_VERS_MAJOR , RA_VERS_MINOR );
}

sVersion  sNIDAQmxDrivers::CoreVersionMin ()
{
  return sVersion( RA_VERS_MAJOR , RA_VERS_MINOR );
}

QString  sNIDAQmxDrivers::UserFriendlyName ()
{
  return trUtf8("Drivers for Counters of NIDAQ-board");
}

QString  sNIDAQmxDrivers::About ()
{
  return trUtf8(
      "Provides 4 driver and 4 consoles:\n"
      "sNIDAQCounter -- driver for counters on NIDAQ-board;\n"
      "sNIDAQDoubleCounter -- driver for two counters on NIDAQ-board;\n"
      "sNIDAQVoltmeter -- driver for Voltmeter on NIDAQ-board;\n"
      "sP138Magnet -- driver for superconductive magnet based on NIDAQ-board;\n"
      "sKineticsNIDAQ -- user interface (console) for kinetics measurements;\n"
      "sStartStopNIDAQ -- user interface (console) for start-stop technique.\n"
      "Console based on sFixedUnitsLCD for Photon counter display"
      "Console based on sPeriodicGageRecorder for slow time dependence.\n\n"
      "Copyright (C) 2011, Dmitry K. Nelson, St.Petersburg, Russia.\n"
      "                    D.Nelson@mail.ioffe.ru\n"
      "Copyright (C) 2011, Viktor E. Bursian, St.Petersburg, Russia.\n"
      "                    Viktor.Bursian@mail.ioffe.ru"
                );
}

void  sNIDAQmxDrivers::InstallDefaults (sPtr<sFolderNode>  DeviceList)
{
  try{
    sPtr<sFolderNode>            DeviceCfg;

    DeviceList |= "NIDAQ Counter";
    DeviceCfg=DeviceList >> "NIDAQ Counter";
    DeviceCfg << attr("Class","sNIDAQCounter");
    sNIDAQCounter::Preconfigure(DeviceCfg);

    DeviceList |= "NIDAQ Double Counter";
    DeviceCfg=DeviceList >> "NIDAQ Double Counter";
    DeviceCfg << attr("Class","sNIDAQDoubleCounter");
    sNIDAQDoubleCounter::Preconfigure(DeviceCfg);

    DeviceList |= "NIDAQ Voltmeter";
    DeviceCfg=DeviceList >> "NIDAQ Voltmeter";
    DeviceCfg << attr("Class","sNIDAQVoltmeter");
    sNIDAQVoltmeter::Preconfigure(DeviceCfg);

    DeviceList |= "NIDAQ Magnet Driver";
    DeviceCfg=DeviceList >> "NIDAQ Magnet Driver";
    DeviceCfg << attr("Class","sP138Magnet");
    sP138Magnet::Preconfigure(DeviceCfg);

    DeviceList |= "NIDAQ counter display";
    DeviceCfg=DeviceList >> "NIDAQ counter display";
    DeviceCfg << attr("Class","sFixedUnitsLCD");
    sFixedUnitsLCD::Preconfigure(DeviceCfg);
    DeviceCfg >> "Links" >> "Gauge" << attr("DeviceId","NIDAQ Counter");

    DeviceList |= "Double counter display";
    DeviceCfg=DeviceList >> "Double counter display";
    DeviceCfg << attr("Class","sDoubleCounterDisplay");
    sDoubleCounterDisplay::Preconfigure(DeviceCfg);
    DeviceCfg >> "Links" >> "Gauge" << attr("DeviceId","NIDAQ Double Counter");

    DeviceList |= "Kinetics NIDAQ";
    DeviceCfg=DeviceList >> "Kinetics NIDAQ";
    DeviceCfg << attr("Class","sKineticsNIDAQ");
    sKineticsNIDAQ::Preconfigure(DeviceCfg);
    DeviceCfg >> "Links" >> "Counter" << attr("DeviceId","NIDAQ Counter");

    DeviceList |= "Start-Stop NIDAQ";
    DeviceCfg=DeviceList >> "Start-Stop NIDAQ";
    DeviceCfg << attr("Class","sStartStopNIDAQ");
    sStartStopNIDAQ::Preconfigure(DeviceCfg);
    DeviceCfg >> "Links" >> "Counter" << attr("DeviceId","NIDAQ Counter");

    DeviceList |= "Recorder NIDAQ";
    DeviceCfg=DeviceList >> "Recorder NIDAQ";
    DeviceCfg << attr("Class","sPeriodicGaugeRecorder");
    sPeriodicGaugeRecorder::Preconfigure(DeviceCfg);
    DeviceCfg >> "Links" >> "Gauge" << attr("DeviceId","NIDAQ Counter");


    /*________________________________________template for device
    DeviceList |= "__DEVICE_ID__";
    DeviceCfg = DeviceList >> "__DEVICE_ID__";
    DeviceCfg << attr("Class","__DRIVER_CLASS_NAME__");
    __DRIVER_CLASS_NAME__::Preconfigure(DeviceCfg);
    //if there are links:
    DeviceCfg >> "Links" >> "__LINK_NAME__"
        << attr("DeviceId","__DEVICE_ID__FOR_THE_LINK__");
    //if you wanna change some of the default settings:
    DeviceCfg >> "Settings" << attr("",);
    DeviceCfg >> "Settings" << attr("",);
    */
  }catch(...){
    ;
  };
}

sNIDAQmxDrivers::sNIDAQmxDrivers ()
{
//  qRegisterMetaType<RA::sMyDataClass>("sMyDataClass");
    qRegisterMetaType<RA::sNIDAQCounter::sSSParams>("sNIDAQCounter::sSSParams");
    qRegisterMetaType<RA::sNIDAQCounter::sArrayData>("sNIDAQCounter::sArrayData");

  sUnit::Register("photon");
  sUnit::Register("T");
  sUnit::Register("%");

//  Этот вызов не всегда установленной библиотеки в конструкторе плагина
//  ломает программу (по крайней мере в Ubuntu)!!!
//  long  ErrorCode = DAQmxSelfCal("Dev1");
//  qDebug()<< "SelfCal ErrorCode = " << ErrorCode;
}

//------------------------------------------------------------------------------
} //namespace RA
