////////////////////////////////////////////////////////////////////////////////
/*! @file P138Magnet.cpp Driver for Superconducting magnet.
- Part of NIDAQmxDrivers - Research Assistant Plugin package.
- Uses  RAGUI - Research Assistant Graphical User Interface.
- Uses  QtGui v.5.0  -  http://qt.nokia.com/
- Uses  RANet - Research Assistant Net Library (based on ANSI C++).
- Copyright(C) 2010-2013, Viktor E. Bursian, St.Petersburg, Russia.
                          Viktor.Bursian@mail.ioffe.ru
- Copyright(C) 2014, Dmitry K. Nelson, St.Petersburg, Russia.
                     D.Nelson@mail.ioffe.ru
*///////////////////////////////////////////////////////////////////////////////
#include "P138Magnet.h"
#include "attr.h"
#include "StringsAndQStrings.h"
#include <QDebug>
namespace RA {
//------------------------------------------------------------------------------

DEFINE_CLASS_TAG(sP138Magnet)

//--------------------------------------------------------------- StepperMDR ---

void  sP138Magnet::Preconfigure (sNodePtr  DeviceCfg)
{
  sPtr<sFolderNode>              DeviceSettings;
  DeviceCfg |= "Settings";
  DeviceSettings = DeviceCfg >> "Settings";
  if( DeviceSettings.IsCorrect() ){
    DeviceSettings |= attr("NumOfSamples",20);
    DeviceSettings |= attr("MeasuringTime",sPhysValue(0.2,"ms"));
    DeviceSettings |= attr("MaxInput",sPhysValue(0.2,"V"));
    DeviceSettings |= attr("MaxOutput",sPhysValue(10.0,"V"));
    DeviceSettings |= attr("PositiveOutput",false);
    DeviceSettings |= attr("MagnetCoeff",sPhysValue(1,"T/mV"));
    DeviceSettings |= attr("Tolerance",sPhysValue(0.01,"mV"));
    DeviceSettings |= attr("TimeDelay",sPhysValue(1.0,"ms"));
  };
  sEffector::Preconfigure(DeviceCfg);
}


sP138Magnet::sP138Magnet(sString   DevicetId
                        ,sNodePtr  DeviceCfg)
    :sEffector(DevicetId,DeviceCfg)
    ,Tolerance(0.01,"mV")
    ,Destination(0.0,"T")
    ,OutValue(0)
    ,MaxOut(10.0)
    ,OutputSign(-1)
    ,TimeDelay(100)
    ,MeasureHandle(0)
    ,SettingHandle(0)
    ,NumOfSamples(20)
    ,readArray(NULL)
    ,Up(false)
    ,Down(false)
{
}


sP138Magnet::~sP138Magnet()
{
    if (MeasureHandle != 0){
        DAQmxStopTask(MeasureHandle);
        DAQmxClearTask (MeasureHandle);
    }
    if (SettingHandle != 0){
        DAQmxStopTask(SettingHandle);
        DAQmxClearTask (SettingHandle);
    }
    if (readArray){
      delete [] readArray;
      readArray = NULL;
    }
}

void sP138Magnet::SetWantedPeriod()
{
  int                         P;
  ThePeriod = WantedPeriod;
  if (ThePeriod < MeasuringTime)
      ThePeriod = MeasuringTime + sPhysValue(1,"ms");
  P = Round( (ThePeriod/sPhysValue(1,"ms")).ToReal() );
  ThePeriod = P * sPhysValue(1,"ms");
  State_O_M_Acquiring->SetTimeout(P);
}

void  sP138Magnet::PostConstructor ()
{
  sEffector::PostConstructor();

  try{ //get config parameters
    get_from<sIntegerNode>(DeviceSettings >> "NumOfSamples" ,NumOfSamples);
    get_from<sPhysValueNode>(DeviceSettings >> "MeasuringTime" ,MeasuringTime);
    get_from<sPhysValueNode>(DeviceSettings >> "MaxInput" ,MaxInput);
    get_from<sPhysValueNode>(DeviceSettings >> "MaxOutput" ,MaxOutput);
    bool PosOut;
    get_from<sBooleanNode>(DeviceSettings>>"PositiveOutput",PosOut);
    if (PosOut) OutputSign = 1; else OutputSign = -1;
    get_from<sPhysValueNode>(DeviceSettings >> "MagnetCoeff" ,MagnetCoeff);
    get_from<sPhysValueNode>(DeviceSettings >> "Tolerance" ,Tolerance);
    sPhysValue PV;
    get_from<sPhysValueNode>(DeviceSettings >> "TimeDelay",PV);
    TimeDelay = Round( (PV/sPhysValue(1,"ms")).ToReal() );
  }catch(...){
    SetInvalid(tr("Something wrong in configuration."));
  }

  //evolution of the inherited state machine (behaviour)
  {
    //evolution of the State_Warming
    {
      State_W_CreateTask = new sState(State_Warming());
      State_W_CreateTask
          ->SetAsInitial()
          ->OnEnterCall(SLOT(CreateNIDAQTask()))
          ->AddTransition(SIGNAL(TaskOK()),State_Warming()->Fin())
          ->AddTransition(SIGNAL(ErrorMessage(QString)),State_Cooling());
    }
    //evolution of the State_Operating
    {
      State_O_SettingTo = new sState(State_Operating(),"State_O_SettingTo");
      State_O_M_Pausing
          = new sState(State_O_Measuring,ThePeriod,"State_O_M_Pausing");

      State_O_M_Acquiring
          ->OnEnterCall( SLOT(MeasureField()) )
          ->AddTransition(State_O_M_Pausing)
          ->AddTransition<sPhysValue>(cSetTo ,SLOT(SetDestination(sPhysValue))
                         ,State_O_SettingTo)
          ->AddTransition(SIGNAL(ErrorMessage(QString)),State_Cooling())
          ;
      State_O_M_Pausing
          ->AddTransition(State_O_M_Pausing,SIGNAL(Timeout())
                         ,State_O_M_Cycling)
          ->AddTransition<sPhysValue>(cSetTo ,SLOT(SetDestination(sPhysValue))
                         ,State_O_SettingTo)
          ->AddTransition(SIGNAL(ErrorMessage(QString)),State_Cooling());
          ;
      State_O_SettingTo
          ->OnFinishCall(SLOT(SayDone()))
          ->AddTransition(State_O_M_Acquiring)
          ->AddTransition(cBreak ,State_O_M_Acquiring)
          ->AddTransition(SIGNAL(ErrorMessage(QString)),State_Cooling())
          ;
      //State_O_SettingTo internal structure and behaviour
      {
        State_O_SettingTo_CheckValue = new sState(State_O_SettingTo,"State_O_SettingTo_CheckValue");
        State_O_SettingTo_ChangeValue = new sState(State_O_SettingTo,TimeDelay,"State_O_SettingTo_ChangeValue");

        State_O_SettingTo_CheckValue
            ->SetAsInitial()
            ->OnEnterCall( SLOT(MeasureField()) )
            ->AddReaction<sPhysValue>(SIGNAL(NewData(sPhysValue))
                                     ,SLOT(CheckField(sPhysValue)))
            ->AddTransition(SIGNAL(FieldOK()),State_O_SettingTo->Fin())
            ->AddTransition(SIGNAL(FieldNotYet()),State_O_SettingTo_ChangeValue);

        State_O_SettingTo_ChangeValue
            ->OnEnterCall( SLOT(ChangeOutput()) )
            ->AddTransition(State_O_SettingTo_ChangeValue
                           ,SIGNAL(Timeout())
                           ,State_O_SettingTo_CheckValue);
      }
    }
  }
}

void sP138Magnet::CreateNIDAQTask ()
{
  real MaxV;
  real _SamplingRate;
  sPhysValue  Max(_Unitsless_);
  sPhysValue SmplRate(_Unitsless_);
  try{
    Max = MaxInput/sPhysValue(1.0,"V");
    MaxV = Max.ToReal();
    SmplRate = sPhysValue(1.0,"s")*NumOfSamples/MeasuringTime;
    _SamplingRate = SmplRate.ToReal();
  }catch(rxRAlgebra){
    emit ErrorMessage(tr("Measuring task: Wrong units in configuration settings."));
    SetInvalid();
    return;
  };
  readArray = new float64 [NumOfSamples];
  bool Error;
  long  ErrorCode;
  ErrorCode = DAQmxCreateTask ("", &MeasureHandle);
  Error = !(ErrorCode == 0);
  ErrorCode = DAQmxCreateAIVoltageChan (MeasureHandle,"Dev1/ai2"
                                       ,"",DAQmx_Val_Diff ,-MaxV,MaxV
                                       ,DAQmx_Val_Volts,NULL );
  //  qDebug() << "Field measuring Create Chanel, ErrorCode = " << ErrorCode;
  Error = Error || !(ErrorCode == 0);
  ErrorCode = DAQmxCfgSampClkTiming (MeasureHandle,"", _SamplingRate
                                    ,DAQmx_Val_Rising
                                    ,DAQmx_Val_FiniteSamps, NumOfSamples);
  Error = Error || !(ErrorCode == 0);

  ErrorCode = DAQmxRegisterDoneEvent (MeasureHandle, 0, &RA::sP138Magnet::OnReadyToRead, this);
  //  qDebug() << "Field measuring Callback Initiated, ErrorCode = " << ErrorCode;
  if (ErrorCode != 0) {
    char EString[128];
    DAQmxGetErrorString (ErrorCode, EString, 128);
    qDebug() << EString;
  };
  Error = Error || !(ErrorCode == 0);

  if (Error) {
    ErrorCode = DAQmxClearTask (MeasureHandle);
    emit ErrorMessage(tr("Cannot create field measuring task. NIDAQ-card is absent or out of service."));
    emit CoolDown();
    qDebug() << "Cooling Down";
    return;
  } else{
    qDebug() << "Field measuring task is successfully formed !";
  }

  try{
    Max = MaxOutput/sPhysValue(1.0,"V");
  }catch(rxRAlgebra){
    emit ErrorMessage(tr("OutputRange: Wrong units in the configuration settings."));
    DAQmxClearTask (MeasureHandle);
    SetInvalid();
    return;
  };
  MaxOut = Max.ToReal();
  ErrorCode = DAQmxCreateTask ("", &SettingHandle);
  Error = !(ErrorCode == 0);
  ErrorCode = DAQmxCreateAOVoltageChan  (SettingHandle,"Dev1/ao0"
                                       ,"",-MaxOut,MaxOut
                                       ,DAQmx_Val_Volts,NULL );
  //  qDebug() << "Create Output Chanel, ErrorCode = " << ErrorCode;
  Error = Error || !(ErrorCode == 0);
  ErrorCode = DAQmxStartTask(SettingHandle);
  Error = Error || !(ErrorCode == 0);
  OutValue = 0;
  ErrorCode = DAQmxWriteAnalogScalarF64 (SettingHandle, true,0.1,OutputSign*OutValue,NULL);
  Error = Error || !(ErrorCode == 0);
  if (Error) {
    ErrorCode = DAQmxClearTask (MeasureHandle);
    ErrorCode = DAQmxStopTask (SettingHandle);
    ErrorCode = DAQmxClearTask (SettingHandle);
    emit ErrorMessage(tr("Cannot create output task. NIDAQ-card is absent or out of service."));
    emit CoolDown();
    qDebug() << "Cooling Down";
    return;
  } else{
    qDebug() << "Output task is successfully started !";
  }
  emit TaskOK();
}

void sP138Magnet::MeasureField ()
{
  long  ErrorCode =0;
  if (MeasureHandle !=0) ErrorCode = DAQmxStopTask(MeasureHandle);
  if (ErrorCode != 0){
    ErrorCode = DAQmxClearTask (MeasureHandle);
    ErrorCode = DAQmxStopTask(SettingHandle);
    ErrorCode = DAQmxClearTask (SettingHandle);
    emit ErrorMessage(tr("Cannot stop field measuring. NIDAQ-card is out of service."));
    emit CoolDown();
    qDebug() << "Cooling Down";
  }
  if (MeasureHandle !=0) ErrorCode = DAQmxStartTask(MeasureHandle);
  if (ErrorCode != 0){
    ErrorCode = DAQmxStopTask(MeasureHandle);
    ErrorCode = DAQmxClearTask (MeasureHandle);
    ErrorCode = DAQmxStopTask(SettingHandle);
    ErrorCode = DAQmxClearTask (SettingHandle);
    emit ErrorMessage(tr("Cannot start field measuring. NIDAQ-card is out of service."));
    emit CoolDown();
    qDebug() << "Cooling Down";
  } else{
//    qDebug() << "Field measuring is successfully started !";
  }
}

void sP138Magnet::SetDestination(sPhysValue  _newField)
{
  Destination = _newField;
  Up = false;
  Down = false;
}

void sP138Magnet::SayDone ()
{
  emit Done(Destination);
}

void sP138Magnet::ChangeOutput ()
{
  long  ErrorCode;
  ErrorCode = DAQmxWriteAnalogScalarF64 (SettingHandle, true,0.1,OutputSign*OutValue,NULL);
  if (!(ErrorCode == 0)) {
    ErrorCode = DAQmxClearTask (MeasureHandle);
    ErrorCode = DAQmxStopTask (SettingHandle);
    ErrorCode = DAQmxClearTask (SettingHandle);
    emit ErrorMessage(tr("Error writing to output."));
    emit CoolDown();
    qDebug() << "Cooling Down";
    return;
  } else{
//    qDebug() << "New output value is set: " << OutValue;
  }
}

void sP138Magnet::CheckField (sPhysValue _field)
{
  if (!Up && !Down){
      if ( std::abs(((_field - Destination)/(Tolerance*MagnetCoeff)).ToReal()) < 1.0){
        emit FieldOK();
      } else if (Destination < _field) {
        Up = false;
        Down = true;
        qDebug() << "Need to go Down";
        emit FieldNotYet();
      } else {
          Down = false;
          Up = true;
          qDebug() << "Need to go Up";
          emit FieldNotYet();
      }
  } else if (Up) {
      if ( (((_field - Destination)/(Tolerance*MagnetCoeff)).ToReal() >= 0) || (!(_field < Range().To())) ) {
        Up = false;
        Down = false;
        emit FieldOK();
      } else {
        qDebug() << "Going Up, OutValue = " << OutValue;
        OutValue += MaxOut/10000;
        emit FieldNotYet();
      }
  } else {    //go Down
      if ( (((Destination - _field)/(Tolerance*MagnetCoeff)).ToReal() >= 0) || (OutValue == 0) ) {
        Up = false;
        Down = false;
        emit FieldOK();
      } else {
          OutValue -= MaxOut/10000;
          emit FieldNotYet();
      }
  }
}

void sP138Magnet::ReadSampling ()
{
//  qDebug() << " Read Field measuring Sampling";
  int32 ReallyRead = 0;
  long ErrorCode = DAQmxReadAnalogF64  (MeasureHandle, NumOfSamples, 1.0
                                      , DAQmx_Val_GroupByChannel, readArray
                                      , NumOfSamples,&ReallyRead, NULL);
  if (ErrorCode == 0){
      double _Data = 0;
      for (int I = 0; I < ReallyRead; I++){
          _Data = _Data + readArray[I];
//          qDebug() << I << " Data - " << _Data;
      }
//  qDebug() << ReallyRead << " samples, Data = " << _Data;
      sPhysValue _Volts = sPhysValue(1.0,"V");
      _Volts.SetError(Tolerance);
      _Volts = _Volts* _Data / ReallyRead;
      sPhysValue _magneticField = _Volts * MagnetCoeff;
      emit NewData (_magneticField);
      if (ReallyRead < NumOfSamples){
          qDebug() << "Really read less than ordered";
        emit ErrorMessage(" Read Field measuring: Really read less than ordered");
      };
  } else{
      qDebug() << " Field measuring Read Error, ErrorCode = " << ErrorCode;
      char EString[128];
      DAQmxGetErrorString (ErrorCode, EString, 128);
      qDebug() << EString;
      emit ErrorMessage(" Read Field measuring: " + QString(EString));
  }
}

int32  CVICALLBACK sP138Magnet::OnReadyToRead (TaskHandle /*theHandle*/
                                                ,int32      /*eventType*/
                                                ,void *     DataPtr)
{
  sP138Magnet* obj = static_cast<sP138Magnet*>(DataPtr);
  QMetaObject::invokeMethod(obj, "ReadSampling",Qt::QueuedConnection);
  return 0;
}

//------------------------------------------------------------------------------
} //namespace RA




