////////////////////////////////////////////////////////////////////////////////
/*! @file DNSpectroScanner.cpp   Спектрометр непрерывного сканирования.
- Снабжен драйвером шагового двигателя, сообщающим свое положение.
- Part of Nelson - Research Assistant Plugin package.
- Uses  RAGUI - Research Assistant Graphical User Interface.
- Uses  QtGui v.5.0  -  http://qt.nokia.com/
- Uses  RANet - Research Assistant Net Library (based on ANSI C++).
- Copyright(C) 2010-2013, Viktor E. Bursian, St.Petersburg, Russia.
                          Viktor.Bursian@mail.ioffe.ru
- Copyright(C) 2014, Dmitry K. Nelson, St.Petersburg, Russia.
                     D.Nelson@mail.ioffe.ru
*///////////////////////////////////////////////////////////////////////////////
#include "DNSpectroScanner.h"
#include "attr.h"
#include "StringsAndQStrings.h"
#include "Trace.h"
#include "Log.h"
#include <QMessageBox>
namespace RA {
//------------------------------------------------------------------------------

DEFINE_CLASS_TAG(sDNSpectroScanner)

//--------------------------------------------------- sDNSpectroScanner ---

void  sDNSpectroScanner::Preconfigure (sNodePtr  DeviceCfg)
{
  sPtr<sFolderNode>              DeviceSettings;
  sPtr<sFolderNode>              DeviceLinks;
  DeviceCfg |= "Links";
  DeviceLinks = DeviceCfg >> "Links";
  if( DeviceLinks.IsCorrect() ){
    DeviceLinks |= "Stepper";
    DeviceLinks >> "Stepper" |= attr("Class","sStepper");
  }
  DeviceCfg |= "Settings";
  DeviceSettings = DeviceCfg >> "Settings";
  if( DeviceSettings.IsCorrect() ){
    DeviceSettings |= attr("ValidScanRange",sPhysRange(200.0,900.0,"nm"));
    DeviceSettings |= attr("ValidScanSteps",sPhysRange(0.01,10,"nm"));
    DeviceSettings |= attr("ScanRange",sPhysRange(400.0,800.0,"nm",0.005));
    DeviceSettings |= attr("ScanStep",sPhysValue(0.1,0.005,"nm"));
    DeviceSettings |= attr("MeasureTime",sPhysValue(1.0,"s"));
    DeviceSettings |= attr("StepperTimeCorrCoef",sPhysValue(_Unitsless_));
    //изменение умолчательных шкал для sScanner
    sChartScales      ChartScales(sChartScale("Wavelength"
                                              ,sPhysRange(400,850,"nm"))
                                 ,sChartScale(""
                                             ,sPhysRange(0,200,"kphoton/s"))
                                 );
    ChartScales.StoreSettings(DeviceSettings);
  }
  sPeriodicGaugeRecorder::Preconfigure(DeviceCfg);
}


sDNSpectroScanner::~sDNSpectroScanner()
{
  if( DeviceSettings.IsCorrect() ){
    DeviceSettings << attr("MeasureTime",MeasureTimeField->Value());
  }
  Stepper.SetToNULL();
}


sDNSpectroScanner::sDNSpectroScanner (sString    DeviceId
                                     ,sNodePtr   DeviceCfg
                                     ,sNodePtr   WidgetCfg
                                     ,QWidget *  parent   )
    :sPeriodicGaugeRecorder(DeviceId,DeviceCfg,WidgetCfg,parent)
    ,Stepper(this)
    ,MeasureTime(sPhysValue(1.0,"s"))
    ,TimePerMotorStep(sPhysValue(1.0,"s"))
    ,TimeCorrCoef(_Unitsless_)
    ,NumOfPoints(0)
    ,NumOfStepsFrom(0)
    ,NumOfStepsToScan(0)
{
  Multiplier.X = sPhysValue(1,"nm");
  //evolution of the inherited state machine
  {
    //evolution of the State_W_Parallel
    {
      State_W_Stepper = new sState(State_W_Parallel,"State_W_Stepper");
      //State_W_Stepper structure and behaviour
      {
        State_W_Stepper_Warming = new sState(State_W_Stepper,"State_W_Stepper_Warming");
        State_W_Stepper_Warming
            ->SetAsInitial()
            ->OnEnterCall(SLOT(LinkToStepper()))
            ->AddTransition(SIGNAL(StepperWarmed()),State_W_Stepper->Fin())
            ->AddTransition(SIGNAL(ErrorMessage()),State_Cooling());
      }
    }
    // evolution of State_O_Scanning
    {
      State_O_Scanning->OnExitCall(SLOT(ReleaseStepperControl()));
        
      State_O_Sc_GoToTheStart = new sState(State_O_Scanning,"State_O_Sc_GoToTheStart");
      State_O_Sc_SettingTimePerStep = new sState(State_O_Scanning,"State_O_Sc_SettingTimePerStep");
       
      State_O_Sc_GoToTheStart
          ->SetAsInitial()
          ->OnEnterCall(SLOT(SetStepperToStartPos()))
          ->AddTransition(SIGNAL(DevicesBusy()),State_O_Idle)
          ->AddTransition(SIGNAL(Break()),SLOT(StopStepper())
                         ,State_O_Scanning->Fin() )
          ->AddTransition(SIGNAL(StepperIsAtPos(sPhysValue))
                         ,State_O_Sc_SettingTimePerStep);
      
      State_O_Sc_SettingTimePerStep
          ->OnEnterCall(SLOT(SetStepperTimePerStep()))
          ->AddTransition(SIGNAL(Break()),State_O_Scanning->Fin() )
          ->AddTransition<sPhysValue>(SIGNAL(StepperTimePerStepSet(sPhysValue))
                                     ,SLOT(AcceptStepperTimePerStep(sPhysValue))
                                     ,State_O_Sc_SettingGauge);
    }
    // evolution of State_O_Sc_Acquisition
    {
      State_O_Sc_A_Parallel = new sParallelStateGroup(State_O_Sc_Acquisition,"State_O_Sc_A_Parallel");
      State_O_Sc_A_Parallel
          ->SetAsInitial()
          ->AddTransition(SIGNAL(Break()),SLOT(BreakScanning()) ,State_O_Sc_Deciding )
          ->OnFinishCall(SLOT(SayScanIsCompleted()));
      //State_O_Sc_A_Parallel internal structure and behaviour
      {
        State_O_Sc_A_P_Measuring = new sState(State_O_Sc_A_Parallel,"State_O_Sc_A_P_Measuring");
        State_O_Sc_A_P_Stepping = new sState(State_O_Sc_A_Parallel,"State_O_Sc_A_P_Stepping");

        //State_O_Sc_A_P_Measuring internal structure and behaviour
        {
          State_O_Sc_A_P_M_WaitingFin = new sState(State_O_Sc_A_P_Measuring,"State_O_Sc_A_P_M_WaitingFin");
          State_O_Sc_A_P_M_WaitingFin
              ->SetAsInitial()
              ->AddTransition(SIGNAL(MeasuringCompleted()),State_O_Sc_A_P_Measuring->Fin());
        }
        //State_O_Sc_A_P_Stepping internal structure and behaviour
        {
          State_O_Sc_A_P_S_WaitingFin = new sState(State_O_Sc_A_P_Stepping,"State_O_Sc_A_P_S_WaitingFin");
          State_O_Sc_A_P_S_WaitingFin
              ->SetAsInitial()
              ->AddTransition(SIGNAL(SteppingCompleted()),State_O_Sc_A_P_Stepping->Fin());
        }
      }
    }
    connect(this ,SIGNAL(StepperError(QString))
           ,this ,SLOT(ReportStepperError(QString)) );
  }
}

void  sDNSpectroScanner::AppearanceMakeUp ()
{
  sPeriodicGaugeRecorder::AppearanceMakeUp();

  try{ //get config parameters
    get_from<sPhysValueNode>(DeviceSettings >> "MeasureTime" ,MeasureTime);
    get_from<sPhysValueNode>(DeviceSettings >> "StepperTimeCorrCoef" ,TimeCorrCoef);
  }catch(...){
    SetInvalid(tr("Something wrong in configuration."));
  }
  //add new widgets
  {
    MeasureTimeField = new sPhysValueField(ParamArea,sPhysValue(),false,false,false);
    ParamLayout->addWidget(MeasureTimeField,Row++,0,1,3,Qt::AlignLeft);
  }
  MeasureTimeField->SetValue(MeasureTime);
  Translate();
}

void  sDNSpectroScanner::Translate ()
{
  MeasureTimeField->SetLabel(tr("MeasureTime"));
}

void  sDNSpectroScanner::RestoreState ()
{
  sPeriodicGaugeRecorder::RestoreState();
  if( WidgetConfig.IsCorrect() ){
    try{ //get stored parameters
      get_from<sPhysValueNode>(WidgetConfig >> "MeasureTime" ,MeasureTime);
    }catch(...){
      SetInvalid(tr("Something wrong in desktop stored."));
    }
  }
  MeasureTimeField->SetValue(MeasureTime);
}


void  sDNSpectroScanner::StoreState ()
{
  sPeriodicGaugeRecorder::StoreState();
  if( WidgetConfig.IsCorrect() ){
    WidgetConfig << attr("MeasureTime",MeasureTimeField->Value());
  };
}


void sDNSpectroScanner::ReportStepperError (QString  error_text)
{
  QMessageBox::critical(this ,tr("Stepper error")
                       ,tr("Device '%1' says:\n\n%2"
                           "\n\nTry to turn on again.")
                       .arg(ToQString(Stepper->DeviceIdentifier()))
                       .arg(error_text)     );
}


void sDNSpectroScanner::LinkToStepper ()
{
  try{
    Stepper.SetTo(DeviceLinks,"Stepper");
    Connect(Stepper ,SIGNAL(Warmed())       ,SIGNAL(StepperWarmed()) );
    Connect(Stepper ,SIGNAL(ErrorMessage(QString)),SIGNAL(StepperError(QString)) );
    Connect(Stepper ,SIGNAL(Done(sPhysValue)),SIGNAL(StepperIsAtPos(sPhysValue)) );
    Connect(Stepper ,SIGNAL(Done()),SIGNAL(SteppingCompleted()) );
    Connect(Stepper ,SIGNAL(TimePerStepSet(sPhysValue))
           ,SIGNAL(StepperTimePerStepSet(sPhysValue)) );
  }catch( rxDeviceLaunch E ){
    emit ErrorMessage( tr("Error while linking to a submissive device.\n"
                          "%1\nCheck configuration.")
                         .arg(ToQString(E.MsgText()))   );
  }
}

void sDNSpectroScanner::SetStepperToStartPos()
{
  if( ! Gauge.TakeExclusiveControl() ){
    QMessageBox::critical(this
        ,tr("'%1' cannot scan")
        .arg(ToQString(DeviceIdentifier()))
        ,tr("Gauge '%1' is busy now.\n\n  Try again later.")
        .arg(ToQString(Gauge->DeviceIdentifier()))
        );
    emit DevicesBusy();
  }else if( ! Stepper.TakeExclusiveControl() ){
    Gauge.ReleaseExclusiveControl();
    QMessageBox::critical(this,tr("Denied")
                         ,tr("Device '%1' is busy.\nTry again later.")
                         .arg(ToQString(Stepper->DeviceIdentifier())) );
    emit DevicesBusy();
  }else{
    ScanRange = ScanRangeField->Value();
    SendDirective<sPhysValue>(Stepper,sStepper::cSetTo,ScanRange.From());
  };
}

void sDNSpectroScanner::SetStepperTimePerStep()
{
  try{
    ScanStep = ScanStepField->Value();
    ScanStep.FixTheUnits();
    NumOfPoints = Round((ScanRange.Size() / ScanStep).ToReal());
    MeasureTime = MeasureTimeField->Value();
    MeasureTime.FixTheUnits();
    NumOfStepsFrom = Stepper->Steps(ScanRange.From());
    NumOfStepsToScan = Stepper->Steps(ScanRange.To()) - NumOfStepsFrom;
    TimePerMotorStep = (MeasureTime*NumOfPoints) / NumOfStepsToScan;
    SendDirective<sPhysValue>(Stepper,sStepper::cSetTimePerStep,TimePerMotorStep);
  }catch(rxRAlgebra){
    QMessageBox::critical(this
        ,tr("%1 failure.").arg(ToQString(DeviceIdentifier()))
        ,tr("Wrong units in the configuration settings.")
        );
    SetInvalid();
    return;
  };
}

void  sDNSpectroScanner::SetGaugeRegime ()
{
  GaugeWasStarted = true;
  GaugeWasInAutoRepeatMode = Gauge->AutoRepeat();
  OldPeriod = Gauge->Period();
  SendCommand(Gauge ,sGauge::cStop);
//  SendDirective(Gauge,sGauge::cSetPeriod,MeasureTime*TimeCorrCoef);
  SendDirective(Gauge,sGauge::cSetPeriod,MeasureTime);
}

void sDNSpectroScanner::AcceptStepperTimePerStep(sPhysValue TPSmod)
{
  try{
    TimePerMotorStep = TPSmod;
    MeasureTime = TimePerMotorStep * NumOfStepsToScan / NumOfPoints;
  }catch(rxRAlgebra){
    QMessageBox::critical(this
        ,tr("%1 failure.").arg(ToQString(DeviceIdentifier()))
        ,tr("Wrong units in the configuration settings.")
        );
    SetInvalid();
    return;
  }
/*
    qDebug()<< "Stepper - Corrected TimePerMotorStep)= " << TimePerMotorStep.Text()
          << "MeasureTime= " << MeasureTime.Text();
    qDebug()<< "Stepper - NumOfStepsToScan = "<< NumOfStepsToScan;
    qDebug()<< "Stepper - Full Scan time = "
          << (TimePerMotorStep * NumOfStepsToScan).Text();
*/
}

void  sDNSpectroScanner::CorrectStep (sPhysValue GPmod)
{
  try{
    MeasureTime = GPmod;
    // В соответствии с реальным временем измерения точки корректируем
    // параметры NumOfStepsToScan и NumOfPoints
    // и правим поля ScanStepField и ScanRangeField
    NumOfPoints = Round((TimePerMotorStep/MeasureTime).ToReal()*NumOfStepsToScan);
    NumOfStepsToScan = Round((MeasureTime/TimePerMotorStep).ToReal()*NumOfPoints);
    sPhysValue CorrectedTo = Stepper->Position(NumOfStepsFrom + NumOfStepsToScan);
    ScanRange = sPhysRange( ScanRange.From() , CorrectedTo );
    ScanRangeField->SetValue(ScanRange);
    ScanStep = ScanRange.Size() / NumOfPoints;
    ScanStepField->SetValue( ScanStep );
  }catch(rxRAlgebra){
    QMessageBox::critical(this
        ,tr("%1 failure.").arg(ToQString(DeviceIdentifier()))
        ,tr("Wrong units in the configuration settings.")
        );
    SetInvalid();
    return;
  }
/*
  qDebug()<< "Gauge - NumOfStepsToScan = "<< NumOfStepsToScan;
  qDebug()<< "Gauge - Full Scan time(steps) = "
          << (TimePerMotorStep * NumOfStepsToScan).Text();
  qDebug()<< "Gauge - Full Scan time(measure) = "
          << (MeasureTime * NumOfPoints).Text();
*/
}

void sDNSpectroScanner::OnEnterToCoolingState ()
{
  Disconnect(Stepper ,0 ,0);
  Stepper.ReleaseExclusiveControl();
  Stepper.SetToNULL();
  sPeriodicGaugeRecorder::OnEnterToCoolingState();
}


void  sDNSpectroScanner::StartScanning ()
{
  CurrentPointNo=0;
  Connect(Gauge ,SIGNAL(NewData(sPhysValue)) ,SLOT(AddNewData(sPhysValue)) );

/*// ---- For Scan time measuring --------------
  Connect(Stepper ,SIGNAL(Done()),SLOT(ScanningTimeShow()) );
  DataCollectTime.start();
  SteppingTime.start();
// ----------------------------- */
  SendCommand(Gauge,sGauge::cStart);
  SendDirective<int>(Stepper,sStepper::cMoveBy,NumOfStepsToScan);
}


void  sDNSpectroScanner::StopScanning ()
{
  SendCommand(Gauge,sGauge::cStop);
  Disconnect(Gauge,SIGNAL(NewData(sPhysValue)) ,SLOT(AddNewData(sPhysValue)));
}

void  sDNSpectroScanner::BreakScanning ()
{
  SendCommand(Stepper,sStepper::cBreak);
  SendCommand(Gauge,sGauge::cStop);
  Disconnect(Gauge,SIGNAL(NewData(sPhysValue)) ,SLOT(AddNewData(sPhysValue)));
}

void  sDNSpectroScanner::EarlySetAttributes ()
{
  sScanner::EarlySetAttributes ();
  Dependence << attr("Scan range", ScanRange.Text())
             << attr("Scan step", ScanStep.Text())
             << attr("Time per point", MeasureTime.Text());
}


void  sDNSpectroScanner::StopStepper ()
{
  SendCommand(Stepper,sStepper::cBreak);
}


void  sDNSpectroScanner::ReleaseStepperControl ()
{
  Stepper.ReleaseExclusiveControl();
}

void  sDNSpectroScanner::SayScanIsCompleted()
{
  emit ScanningHasCompleted();
}

void  sDNSpectroScanner::AddNewData (sPhysValue  Data)
{
  if( ! DataListRecord.IsCorrect() ){
    return;
  };
  if( Multiplier.X.IsNaN() ){
    return;
  }
  if( Multiplier.Y.IsNaN() ){
    Multiplier.Y = sPhysValue(Data.Units());
    sPhysRange                  RY(-Abs(Data),3*Abs(Data));
    SetChartScales(sChartScales(ChartScales().X
                               ,sChartScale(ChartScales().Y.Name,RY)
                               ));
  }else{
    try{ //the gauge could change units between the scans or on the fly
      (Data / Multiplier.Y).ToReal(); // did he?
    }catch(rxRAlgebra){
      if( CurrentPointNo == 0 ){ // if he did before the scan started
        Multiplier.Y = sPhysValue(Data.Units());
        sPhysRange                  RY(-Abs(Data),3*Abs(Data));
        SetChartScales(sChartScales(ChartScales().X
                                   ,sChartScale(ChartScales().Y.Name,RY)
                                   ));
      }else{
        Data = Multiplier.Y * 0.00 ;
      }
    }
  }
  if( DataListRecord->Multiplier().Y.IsNaN() ){
    DataListRecord->SetMultiplier(sPhysPair(Multiplier.X,Multiplier.Y));
  };
  sPhysValue                  X(_Unitsless_);
  sPhysValue                  Y(_Unitsless_);
  real X_in_steps = NumOfStepsFrom + 0.5*NumOfStepsToScan/NumOfPoints
                  + 1.0*CurrentPointNo*NumOfStepsToScan/NumOfPoints;
  X = Stepper->Position(Round(X_in_steps))/ Multiplier.X;
  Y = Data / Multiplier.Y;
  DataListRecord->AddNewPoint(X.ToReal(),Y.ToReal());

//  CurrentPointNo++;

  if( CurrentPointNo < (NumOfPoints - 1) ){
    CurrentPointNo++;
  }else{
    emit MeasuringCompleted();
/*
    // ---- Scan time measuring --------------
    qDebug() << "Measuring time = " << DataCollectTime.elapsed();
    // -----------------------------
*/
  };

}

/*
// ---- Scan time measuring --------------
void  sDNSpectroScanner::ScanningTimeShow ()
{
  qDebug() << "Stepping time = " << SteppingTime.elapsed();
  Disconnect(Stepper ,SIGNAL(Done()),SLOT(ScanningTimeShow()) );

}
// -----------------------------
*/

//------------------------------------------------------------------------------
} //namespace RA
