////////////////////////////////////////////////////////////////////////////////
/*! @file DigitalLIA.cpp   Driver for the Digital Lock-In-Amplifier.
- Part of Nelson - Research Assistant Plugin package.
- Uses  RAGUI - Research Assistant Graphical User Interface.
- Uses  QtGui v.5.0  -  http://qt.nokia.com/
- Uses  RANet - Research Assistant Net Library (based on ANSI C++).
- Copyright(C) 2010-2013, Viktor E. Bursian, St.Petersburg, Russia.
                          Viktor.Bursian@mail.ioffe.ru
- Copyright(C) 2017, Dmitry K. Nelson, St.Petersburg, Russia.
                     D.Nelson@mail.ioffe.ru
*///////////////////////////////////////////////////////////////////////////////
#include "DigitalLIA.h"
#include <QtMath>
#include "attr.h"
#include "StringsAndQStrings.h"
#include <QDebug>
namespace RA {
//------------------------------------------------------------------------------

DEFINE_CLASS_TAG(sDigitalLIA)

/*! Директива изменения усиления
0 - усиление = 1 (Общее усиление равно 10)
1 - усиление = 10^1 (Общее усиление равно 100)
2 - усиление = 10^2 (Общее усиление равно 1000)
3 - усиление = 10^3 (Общее усиление равно 10000)
*/
const sDirective<int>  sDigitalLIA::cSetGain("cSetGain");

//--------------------------------------------------------------- DigitalLIA ---

void  sDigitalLIA::Preconfigure (sNodePtr  DeviceCfg)
{
  sPtr<sFolderNode>              DeviceSettings;
  sPtr<sFolderNode>              DeviceLinks;
  DeviceCfg |= "Links";
  DeviceLinks = DeviceCfg >> "Links";
  if( DeviceLinks.IsCorrect() ){
    DeviceLinks |= "LIADeviceCom";
    DeviceLinks >> "LIADeviceCom" |= attr("Class","sTelecom");
  };
  DeviceCfg |= "Settings";
  DeviceSettings = DeviceCfg >> "Settings";
  if( DeviceSettings.IsCorrect() ){
    DeviceSettings |= attr("LIADeviceName","@MSDN204A");
    DeviceSettings |= "Multipliers";
    DeviceSettings >> "Multipliers" |= attr(1 ,sPhysValue(0.2441,0.01,"mV"));
    DeviceSettings >> "Multipliers" |= attr(2 ,sPhysValue(0.02441,0.001,"mV"));
    DeviceSettings >> "Multipliers" |= attr(3 ,sPhysValue(0.002441,0.0001,"mV"));
    DeviceSettings >> "Multipliers" |= attr(4 ,sPhysValue(0.0002441,0.00001,"mV"));
  };
  sGauge::Preconfigure(DeviceCfg);
}


sDigitalLIA::sDigitalLIA(sString   DevicetId
                        ,sNodePtr  DeviceCfg)
    :sGauge(DevicetId,DeviceCfg)
    ,LIADeviceCom(this)
    ,GainIndex(0)
{
}


sDigitalLIA::~sDigitalLIA()
{
    LIADeviceCom.SetToNULL();
}


void  sDigitalLIA::SetWantedPeriod ()
{
  try{
    real TimeInMicroSec; // Период передается в МК в виде целого числа микросекунд.
    TimeInMicroSec = (WantedPeriod/(sPhysValue(1.0e-6,"s"))).ToReal();
    int TimePar = Round(TimeInMicroSec);
    QByteArray _comm;
    _comm.append(QByteArray::number((qlonglong)TimePar, 16).toUpper().rightJustified(8, '0'));
    _comm.prepend("T");
    SendDirective<sTelex>(LIADeviceCom ,sTeleCom::cSendTelex ,sTelex(_comm));
  }catch(rxRAlgebra){
    //!@todo{Instruments} define Failure signal for drivers?
    emit ErrorMessage(tr("Cannot set period, period=%1")
                     .arg(ToQString(WantedPeriod.Text())));
    emit CoolDown();
  }
}


void  sDigitalLIA::PostConstructor ()
{
  sGauge::PostConstructor();

  Multiplier[0] = sPhysValue(0.2441,0.01,"mV");
  Multiplier[1] = sPhysValue(0.02441,0.001,"mV");
  Multiplier[2] = sPhysValue(0.002441,0.0001,"mV");
  Multiplier[3] = sPhysValue(0.0002441,0.00001,"mV");

  try{ //get config parameters
    get_from<sTextNode>(DeviceSettings >> "LIADeviceName" ,LIADeviceName);
    for( int i = 0 ; i < 4 ; ++i ){
      get_from<sPhysValueNode>(DeviceSettings >> "Multipliers" >> (i+1) ,Multiplier[i]);
    };
  }catch(...){
    SetInvalid(tr("Something wrong in configuration."));
  }

  WantedPeriod = ThePeriod;

  //evolution of the inherited state machine (behaviour)
  {
    //evolution of the State_Warming
    {
      State_W_Linking = new sState(State_Warming(),"State_W_Linking");
      State_W_Checking = new sState(State_Warming()
                                   ,sPhysValue(1,"s")
                                   ,"State_W_Checking");
      State_W_SettingPeriod = new sState(State_Warming()
                                 ,"State_W_SettingPeriod");
      State_W_Linking
          ->SetAsInitial()
          ->OnEnterCall( SLOT(LinkToLIADevice()) )
          ->AddTransition( SIGNAL(LIADeviceLinked()) ,State_W_Checking )
          ->AddTransition( SIGNAL(ErrorMessage()) ,State_Cooling() )
          ;
      State_W_Checking
          ->OnEnterCall(SLOT(AskLIADeviceName()))
          ->AddTransition(State_W_Checking,SIGNAL(Timeout())
                         ,State_Cooling())
          ->AddTransition<sTelex>(SIGNAL(GotLIADeviceMessage(sTelex))
                         ,State_W_SettingPeriod
                         ,sTelex(LIADeviceName))
          ->AddTransition(SIGNAL(ErrorMessage()) ,State_Cooling() )
          ;
      State_W_SettingPeriod
          ->OnEnterCall(SLOT(SetWantedPeriod()))
          ->AddReaction<sTelex>(SIGNAL(GotLIADeviceMessage(sTelex))
                               ,SLOT(CorrectPeriod(sTelex)) )
          ->AddTransition(SIGNAL(PeriodChanged(sPhysValue))
                                     ,State_Warming()->Fin())
          ;
    }
    //evolution of the State_Operating
    {
      State_O_M_SetP_Checking = new sState(State_O_M_SettingPeriod
                                         ,"State_O_M_SetP_Checking");
      State_O_M_SetP_Checking
          ->SetAsInitial()
          ->AddTransition<sTelex>(SIGNAL(GotLIADeviceMessage(sTelex))
                                 ,SLOT(CorrectPeriod(sTelex))
                                 ,State_O_M_SettingPeriod->Fin() )
              ;
      State_Operating()
          ->AddReaction<int>(cSetGain,SLOT(SetGain(int)));
      State_O_M_Acquiring
          ->OnEnterCall(SLOT(StartMeasuring()))
          ->OnExitCall(SLOT(StopMeasuring()))
          ->AddReaction<sTelex>(SIGNAL(GotLIADeviceMessage(sTelex))
                                 ,SLOT(CalcOutput(sTelex)));
    }
    connect(this ,SIGNAL(LIADeviceError(QString))
           ,this ,SLOT(ReportLIADeviceError(QString)) );
  }
}


void sDigitalLIA::LinkToLIADevice ()
{
  sString                     TeleComId;
  if( ! get_from<sTextNode>(DeviceLinks>>"LIADeviceCom">>"DeviceId",TeleComId) ){
    emit ErrorMessage( tr("The following link is absent: 'LIADeviceCom'.\n"
                          "Check configuration.") );
  }else{
    LIADeviceCom.SetTo(TeleComId);
    if( ! LIADeviceCom.IsCorrect() ){
      emit ErrorMessage( tr("The 'LIADeviceCom' link is set to '%1', which "
                            "does not look correct. Check configuration.")
                           .arg(ToQString(TeleComId)) );
    }else{
      Connect(LIADeviceCom ,SIGNAL(Warmed()) ,SIGNAL(LIADeviceLinked()) );
      Connect(LIADeviceCom ,SIGNAL(ErrorMessage(QString))
             ,SIGNAL(LIADeviceError(QString)) );
      Connect(LIADeviceCom ,SIGNAL(GotTelex(sTelex))
             ,SIGNAL(GotLIADeviceMessage(sTelex)) );
    };
  };
}

void sDigitalLIA::ReportLIADeviceError (QString  error_text)
{
  emit ErrorMessage( tr("Digital LIA '%1' says:\n\n"
                        "Got error from device connected to '%2':\n\n%3")
                       .arg(ToQString(DeviceIdentifier()))
                       .arg(ToQString(LIADeviceCom->DeviceIdentifier()))
                       .arg(error_text)     );
}

void sDigitalLIA::AskLIADeviceName ()
{
    SendDirective<sTelex>(LIADeviceCom ,sTeleCom::cSendTelex ,sTelex("?"));
}

void sDigitalLIA::CorrectPeriod (sTelex text)
{
  bool OK;
  int TimeInMicroSec = text.toLong(&OK,16);
  if( OK ){
    ThePeriod = sPhysValue(TimeInMicroSec*1.0e-6,"s");
    emit PeriodChanged(ThePeriod);
  }else{
    emit ErrorMessage(QString("Unknown telex: ")+text);
  }
}

void sDigitalLIA::SetGain (int _gainIndex)
{
  QByteArray _comm;
  _comm.append(QByteArray::number(_gainIndex, 16)
                          .toUpper().rightJustified(2, '0'));
  _comm.prepend("G");
  SendDirective<sTelex>(LIADeviceCom ,sTeleCom::cSendTelex ,sTelex(_comm));
  GainIndex = _gainIndex;
}

void sDigitalLIA::CalcOutput(sTelex text)
{
  QByteArray text1;
  sPhysValue _ADCV = sPhysValue(2.441,0.1,"mV");
  double _ADCA = 0;
  double _ADCB = 0;
  qint32 Num = 0;
  bool OK = false;
  int _i = 0;
  char _C = text.at(0);
  text.remove(0,1);
  if (!(_C == 'V')) return;
  { // -----------------Reading Amplitude (from V to X)
    _i = text.indexOf('X');
    text1 = text.left(_i);
    text.remove(0,_i+1);
//    qDebug () << "MsgV= " << text1;
    if ( !(_i > 0)) return;
    Num = text1.toLongLong(&OK,16);
    if (!(OK)) return;
    _ADCV = _ADCV*qAbs(Num);
  }
  { // -----------------Reading Signal A (from X to Y)
    _i = text.indexOf('Y');
    text1 = text.left(_i);
    text.remove(0,_i+1);
//    qDebug () << "MsgX= " << text1;
    if ( !(_i > 0)) return;
    Num = text1.toLongLong(&OK,16);
    if (!(OK)) return;
    _ADCA = Num;
  }
  { // -----------------Reading Signal B (from Y to N)
    _i = text.indexOf('N');
    text1 = text.left(_i);
    text.remove(0,_i+1);
//    qDebug () << "MsgB= " << text1;
    if ( !(_i > 0)) return;
    Num = text1.toLongLong(&OK,16);
    if (!(OK)) return;
    _ADCB = Num;
  }
  Num = text.toLongLong(&OK,16);
  if (!(OK)) return;
  _ADCA = _ADCA/Num;
  _ADCB = _ADCB/Num;
  qDebug() << "Num of Samples = " << Num;
//  qDebug() << "ADC A = " << _ADCA;
//  qDebug() << "ADC B = " << _ADCB;
  emit NewAmplitude(_ADCV);
  emit NewData(Multiplier[GainIndex]*qSqrt(_ADCA*_ADCA + _ADCB*_ADCB));

}

void sDigitalLIA::StartMeasuring ()
{
  SendDirective<sTelex>(LIADeviceCom ,sTeleCom::cSendTelex ,sTelex("R0"));
  SendDirective<sTelex>(LIADeviceCom ,sTeleCom::cSendTelex ,sTelex("M"));
}

void sDigitalLIA::StopMeasuring ()
{
  SendDirective<sTelex>(LIADeviceCom ,sTeleCom::cSendTelex ,sTelex("X"));
}


//------------------------------------------------------------------------------
} //namespace RA




