////////////////////////////////////////////////////////////////////////////////
/*! @file MSDN2044.c  Firmware for Step Motor Driver.
- Part of the MSDN firmware.
- Part of the Research Assistant project.
- Uses  LUFA library written by
   - Copyright(C) 2012, Dean Camera, dean [at] fourwalledcubicle [dot] com,
                                     www.lufa-lib.org.

- Copyright(C) 2014, Dmitry K. Nelson, St.Petersburg, Russia.
                     D.Nelson@mail.ioffe.ru
- Copyright(C) 2016, Viktor E. Bursian, St.Petersburg, Russia.
                     Viktor.Bursian@mail.ioffe.ru
*///////////////////////////////////////////////////////////////////////////////
#include "Microsin_USB_IO.h"
#include "CommandParser.h"
#include <avr/wdt.h>
#include <avr/power.h>
#include <util/delay.h>
#include <avr/iousb162.h>
#include <avr/interrupt.h>
//! @todo{MSDN} Неясная роль inttypes здесь
#include <inttypes.h>


//=============================================================================
// Работа с шаговым двигателем. Тактовый сигнал шагов подается на пин 0 порта D
//                              Сигнал направления подается на пин 1 порта D
//                              Сигнал разрешения (L) подается на пин 2 порта D
//                              Сигналы инвертируются оптической развязкой
//=============================================================================

uint16_t  TicksPerStep;               // В данной реализации на полшага, так как период шагания
                                      // задается двумя срабатываниями таймера - переключающими
                                      // выход вверх и вниз.
uint16_t       CurTPS;                // Переменная для текущей паузы между шагами
uint16_t       AccelAddition;         // Коэф увеличения начальной паузы при разгоне
uint32_t       CurrentRamp = 0;       // Текущее значение в процессе разгона и торможения
uint16_t       RampCoef;              // Коэф уменьшения периода шагов за шаг
uint16_t       StoppingZone;          // Параметр области (в шагах) разгона-торможения
uint16_t       RealZone =0;	          // Реальное используемое значение
uint32_t       NumOfSteps;            // Заданное число шагов
bool           Ramp;                  // Признак разгона-торможения
volatile bool  StepsFinished = false; // Признак окончания шагания

// -------- Макросы установки направления вращения --------------------------
// Устанавливают бит направления вращения
// Что соответствует какому направлению, зависит от кинематической схемы прибора
// Данные установки для ДФС-12 с шаговым двигателем
#define SETDIRFORWARD PORTD &= ~(1 << 1)
#define SETDIRBACKWARD PORTD |= (1 << 1)
// ---------------------------------------------------------------------------

int8_t  ExecuteCommand0 (const char  command)
{
  switch( command ){
    case 'X':
      TIMSK1 &= (uint8_t)(~(1 << OCIE1A)); // Запрещаем прерывания по таймеру TIMER1
      PORTD &= ~(1 << 2); // Выводим 0 на пин PORTD2 (Disable)
      PORTD |= (1);		  // Выводим 1 на пин PORTD0
      SendCharToUSB ('N');
      SendCharToUSB ('=');
      SendNumberToUSB (NumOfSteps);
      Send_EOL_ToUSB ();
      NumOfSteps = 0;
      break;
    case 'P':
      SendCharToUSB ('T');
      SendCharToUSB ('=');
      SendNumberToUSB (TicksPerStep);
      SendCharToUSB ('A');
      SendCharToUSB ('=');
      SendNumberToUSB (AccelAddition);
      SendCharToUSB ('K');
      SendCharToUSB ('=');
      SendNumberToUSB (RampCoef);
      SendCharToUSB ('Z');
      SendCharToUSB ('=');
      SendNumberToUSB (StoppingZone);
      Send_EOL_ToUSB ();
      break;
    default:
      return 1;
  }
  return 0;
}

int8_t  ExecuteCommand2 (const char     command
                        ,const uint8_t  argument)
{
  return 1;
}

int8_t  ExecuteCommand4 (const char      command
                        ,const uint16_t  argument)
{
  switch (command){
    case 'T':
      TicksPerStep = argument - 1; // Устанавливается величина, на единицу меньшая,
      break;          // так как счетчик в контроллере срабатывает по значению N+1.
    case 'A':
      AccelAddition = argument;
      break;
    case 'K':
      RampCoef = argument;
      break;
    case 'Z':
      StoppingZone = argument;
      break;
  default:
      return 1;
  }
  return 0;
}

int8_t  ExecuteCommand8 (const char      command
                        ,const uint32_t  argument)
{
  switch( command ){
    case 'G':
      Ramp = true;
      NumOfSteps = argument;
      SETDIRFORWARD;
      PORTD |= (1);			// Выводим 1 на пин PORTD0
      PORTD |= (1 << 2);	// Выводим 1 на пин PORTD2 (Enable)
      if (NumOfSteps >= 2*StoppingZone) RealZone = StoppingZone;
      else RealZone = NumOfSteps / 2;
      CurrentRamp = AccelAddition << 10;
      CurTPS = (uint16_t)(((CurrentRamp*TicksPerStep) >> 10) & 0xFFFF);
      OCR1AH = (uint8_t)((CurTPS >> 8) & 0xFF);
      OCR1AL = (uint8_t)(CurTPS & 0xFF);
      TCNT1H = 0x00;
      TCNT1L = 0x00;
      TIMSK1 |= (1 << OCIE1A);    // Разрешаем прерывания по таймеру TIMER1
      break;
    case 'R':
      Ramp = true;
      NumOfSteps = argument;
      SETDIRBACKWARD;
      PORTD |= (1);			// Выводим 1 на пин PORTD0
      PORTD |= (1 << 2);	// Выводим 1 на пин PORTD2 (Enable)
      if (NumOfSteps >= 2*StoppingZone) RealZone = StoppingZone;
      else RealZone = NumOfSteps / 2;
      CurrentRamp = AccelAddition << 10;
      CurTPS = (uint16_t)(((CurrentRamp*TicksPerStep) >> 10) & 0xFFFF);
      OCR1AH = (uint8_t)((CurTPS >> 8) & 0xFF);
      OCR1AL = (uint8_t)(CurTPS & 0xFF);
      TCNT1H = 0x00;
      TCNT1L = 0x00;
      TIMSK1 |= (1 << OCIE1A);    // Разрешаем прерывания по таймеру TIMER1
      break;
    case 'F':
      Ramp = false;
      NumOfSteps = argument;
      SETDIRFORWARD;
      PORTD |= (1);			// Выводим 1 на пин PORTD0
      PORTD |= (1 << 2);	// Выводим 1 на пин PORTD2 (Enable)
      OCR1AH = (uint8_t)((TicksPerStep >> 8) & 0xFF);
      OCR1AL = (uint8_t)(TicksPerStep & 0xFF);
      TCNT1H = 0x00;
      TCNT1L = 0x00;
      TIMSK1 |= (1 << OCIE1A);    // Разрешаем прерывания по таймеру TIMER1
      break;
    case 'B':
      Ramp = false;
      NumOfSteps = argument;
      SETDIRBACKWARD;
      PORTD |= (1);			// Выводим 1 на пин PORTD0
      PORTD |= (1 << 2);	// Выводим 1 на пин PORTD2 (Enable)
      OCR1AH = (uint8_t)((TicksPerStep >> 8) & 0xFF);
      OCR1AL = (uint8_t)(TicksPerStep & 0xFF);
      TCNT1H = 0x00;
      TCNT1L = 0x00;
      TIMSK1 |= (1 << OCIE1A);    // Разрешаем прерывания по таймеру TIMER1
      break;
    default:
      return 1;
  }
  SendCharToUSB ('<');
  Send_EOL_ToUSB ();
  return 0;
}

void  BeforeLoop (void)
{
  TicksPerStep = 20;
  AccelAddition = 4;
  RampCoef = 1031;   // (1.007 * 1024);
  StoppingZone = 200;
  NumOfSteps = 0;
  Ramp = false;
  StepsFinished = false;
}

void  InLoop (void)
{
  if (StepsFinished){
    SendCharToUSB('D');
    SendCharToUSB('o');
    SendCharToUSB('n');
    SendCharToUSB('e');
    Send_EOL_ToUSB ();
    StepsFinished = false;
  }
}

void SetupHardware(void)
{
  /* Disable watchdog if enabled by bootloader/fuses */
  MCUSR &= ~(1 << WDRF);
  wdt_disable();

  /* Disable clock division */
  clock_prescale_set(clock_div_1);

// Установки для управления шаговым двигателем --------------------------------
  TCCR1A = (0 << COM1A1) |  //  Инициализация таймера микроконтроллера TIMER1
           (0 << COM1A0) |
           (0 << COM1B1) |
           (0 << COM1B0) |
           (0 << FOC1A) |
           (0 << FOC1B) |
           (0 << WGM11) |
           (0 << WGM10);

  TCCR1B = (0 << ICNC1) |
           (0 << ICES1) |
           (0 << WGM13) |
           (1 << WGM12) |     // Режим сброса при совпадении OCR1A
           (1 << CS12) |      // Выбор тактовой частоты
           (0 << CS11) |      // Данными битами
           (0 << CS10);       // установлено деление на 256 (время на полный шаг получается 32 мкс)

  OCR1AH = (uint8_t)((TicksPerStep >> 8) & 0xFF);  // Заносим количество тактов
  OCR1AL = (uint8_t)(TicksPerStep & 0xFF);         // в регистр счетчика

// Установки порта ввода-вывода ------------------------------------------------
  DDRD |= 0b00000111;  // Биты 0,1,2 на вывод (управление шаговыми двигателем)
                       //  Сигналы инвертируются оптической развязкой
  PORTD &= ~(1 << 2);  // Выводим 0 на пин PORTD2 (1 на вход микросхемы - Disable)
  PORTD |= (1);		   // Выводим 1 на пин PORTD0 (0 на вход микросхемы)
  PORTD |= (1<< 1);	   // Выводим 1 на пин PORTD1 (0 на вход микросхемы)
//------------------------------------------------------------------------------
}

//-----------------------------------------------------------------------------
// Обработчик прерывания от таймера 1 (COMPA)
//-----------------------------------------------------------------------------
ISR(TIMER1_COMPA_vect){
  sei();                  // Дадим возможность обработать прерывания от USB...
  if (NumOfSteps) {
    if (Ramp) {
      if (NumOfSteps <= RealZone) {
        if ((NumOfSteps == RealZone) && (CurrentRamp == 0)) CurrentRamp = 1024;
        if (PORTD & 0x01) {
          PORTD &= 0xFE;  // Выводим 0 на пин PORTD0 - делаем шаг
          NumOfSteps--;
          CurrentRamp *= RampCoef;
          CurrentRamp = CurrentRamp >> 10;
          CurTPS = (uint16_t)((CurrentRamp*TicksPerStep >> 10) & 0xFFFF);
          OCR1AH = (uint8_t)((CurTPS >> 8) & 0xFF);
          OCR1AL = (uint8_t)(CurTPS & 0xFF);
        } else {
          PORTD |= 0x01;  // Выводим 1 на пин PORTD0 - изготовка для шага
        };
      } else {
        if (PORTD & 0x01) {
          PORTD &= 0xFE;  // Выводим 0 на пин PORTD0 - делаем шаг
          NumOfSteps--;
          if (CurrentRamp > 0) {
            CurrentRamp = CurrentRamp << 10;
            CurrentRamp /= RampCoef;
            if (CurrentRamp > 1024) {
              CurTPS = (uint16_t)(((CurrentRamp*TicksPerStep) >> 10) & 0xFFFF);
            } else {
              CurTPS = TicksPerStep;
              CurrentRamp = 0;
            }
            OCR1AH = (uint8_t)((CurTPS >> 8) & 0xFF);
            OCR1AL = (uint8_t)(CurTPS & 0xFF);
          }
        } else {
          PORTD |= 0x01; // Выводим 1 на пин PORTD0 - изготовка для шага
        };
      };
    } else {
      if (PORTD & 0x01) {
        PORTD &= 0xFE;   // Выводим 0 на пин PORTD0 - делаем шаг
        NumOfSteps--;
      } else {
        PORTD |= 0x01;  // Выводим 1 на пин PORTD0 - изготовка для шага
      };
    };
  } else {
    TIMSK1 &= (uint8_t)(~(1 << OCIE1A)); // Запрещаем прерывания по таймеру TIMER1
    PORTD &= ~(1 << 2);		// Выводим 0 на пин PORTD2 (Disable)
    PORTD |= (1);			// Выводим 1 на пин PORTD0
    StepsFinished = true;
  };
}

