////////////////////////////////////////////////////////////////////////////////
/*! @file MSDN2046.c  Firmware for MDR Driver with DUK.
- Part of the MSDN firmware.
- Part of the Research Assistant project.
- Uses  LUFA library written by
   - Copyright(C) 2012, Dean Camera, dean [at] fourwalledcubicle [dot] com,
                                     www.lufa-lib.org.

- Copyright(C) 2014, Dmitry K. Nelson, St.Petersburg, Russia.
                     D.Nelson@mail.ioffe.ru
- Copyright(C) 2016, Viktor E. Bursian, St.Petersburg, Russia.
                     Viktor.Bursian@mail.ioffe.ru
*///////////////////////////////////////////////////////////////////////////////

//=============================================================================
// Контроллер управления штатным шаговым двигателем и датчиком угол-код спектрометра МДР.
// Четыре фазы управления шаговым двигателем подаются на пины 0,1,2,3 порта D.
// Управление датчиком угол-код:
// Тактовый сигнал сдвигового регистра подается на пин 0 порта B
// Бит данных для сдвигового регистра подается на пин 1 порта B
// Управление A и B подается, соответственно, на пин 4 и пин 5 порта B
// Считывание датчика угол-код с пина 2 порта C
// Пины 4,5,6,7 порта C используются для ввода номера установленной решетки.
//=============================================================================

#include "Microsin_USB_IO.h"
#include "CommandParser.h"
#include <avr/wdt.h>
#include <avr/power.h>
#include <util/delay.h>
#include <avr/iousb162.h>
#include <avr/interrupt.h>
#include <stdlib.h>
//! @todo{MSDN} Неясная роль inttypes здесь
#include <inttypes.h>
//! @todo{MSDN} Неясная роль math
#include <math.h>

uint8_t  GratingNo;

uint16_t  TicksPerStep;      // Задает время на шаг для сканирования
uint16_t  MaxSpeedTPS;       // Задает максимальную скорость перемотки
uint16_t  CurTPS;            // Переменная для текущей паузы между шагами
uint16_t  AccelAddition; 	// Коэф увеличения начальной паузы при разгоне
uint32_t  CurrentRamp = 0;   // Текущее значение в процессе разгона и торможения
uint16_t  RampCoef;          // Коэф уменьшения периода шагов за шаг
uint16_t  StoppingZone;      // Параметр области (в шагах) разгона-торможения
uint16_t  RealZone =0;	    // Реальное используемое значение
uint32_t  NumOfSteps;        // Заданное число шагов
bool      Ramp=false;         // Признак разгона-торможения
volatile bool    RunFinished=false; // Признак окончания шагания
static bool      Running=false;
static bool      OnTarget=false;

static uint32_t  WLIndex;
static uint32_t  NewWLIndex;


#define PhaseMax 3
const  uint8_t Phases[4] = {0b00000011, 0b00000110, 0b00001100, 0b00001001};
static signed char Inc = 1;
static signed char PhaseI = 0;

#define StepsPerIndex 3
#define MinIndex 18500
#define MaxIndex 102000
#define Understepping 5
#define Overstepping 30

void SendDone(void)
{
    SendCharToUSB('D');
    SendCharToUSB('o');
    SendCharToUSB('n');
    SendCharToUSB('e');
    Send_EOL_ToUSB ();
}

void GetIndex (void)
{
    PORTB &= 0b11001100;   // Устанавливаем в ноль биты 	0,1,4,5
    WLIndex = 0;
    _delay_us(1);
    PORTB |= (1<<1); // Выводим единицу в порт данных
    _delay_us(1);
    PORTB |= (1<<5); // Включаем канал считывания B
    _delay_us(1);
    PORTB |= 1;      // Такт - записываем единицу в регистр
    for (char k = 0; k < 18; k++)
    {
        _delay_us(1);
        PORTB &= 0b11111100; // Выводим 0 на пины PORTB 0,1 - возврат в исх. сост.
        _delay_us(200);
        if (PINC & (1<<2)){
            WLIndex = (WLIndex / 2) + 0x20000;
            PORTB &= 0b11001111; // Выводим 0 на пины PORTB 4,5 - возврат в исх. сост.
            _delay_us(1);
            PORTB |= (1<<5);    // Включаем канал считывания B
        } else {
            WLIndex = (WLIndex / 2);
            PORTB &= 0b11001111; // Выводим 0 на пины PORTB 4,5 - возврат в исх. сост.
            _delay_us(1);
            PORTB |= (1<<4);    // Включаем канал считывания A
        }
        PORTB |= 1;      // Такт - продвигаем единицу в следующий разряд регистра
    }
    PORTB &= 0b11001100;   // Устанавливаем в ноль биты 	0,1,4,5
}

void DoOneStep (void)
{
    PhaseI += Inc;       // Циклически (из интервала 0,1,2,3) изменяем индекс фазы
    if (PhaseI < 0)  PhaseI = PhaseMax;
    if (PhaseI > PhaseMax)  PhaseI = 0;
    PORTD = (PORTD & 0b11110000) | Phases[PhaseI]; // Выводим следующую фазу
}

void StartMultiSteps(void)
{
    Ramp=true;
    TicksPerStep = MaxSpeedTPS;
    RunFinished=false;
    if (NumOfSteps >= 2*StoppingZone) RealZone = StoppingZone;
    else RealZone = NumOfSteps / 2;
    CurrentRamp = AccelAddition << 10;
    CurTPS = (uint16_t)(((CurrentRamp*TicksPerStep) >> 10) & 0xFFFF);
    OCR1AH = (uint8_t)((CurTPS >> 8) & 0xFF);
    OCR1AL = (uint8_t)(CurTPS & 0xFF);
    TCNT1H = 0x00;
    TCNT1L = 0x00;
    TIMSK1 |= (1 << OCIE1A);    // Разрешаем прерывания по таймеру TIMER1
}

void Run (void)
{
  Running = false;
  GetIndex();
  if (NewWLIndex < WLIndex) {
    Running = true;
    Inc = -1;    // Direction - backward
    NumOfSteps = (WLIndex-NewWLIndex)*StepsPerIndex + Overstepping;
    StartMultiSteps();
  } else {
    if ((NewWLIndex-WLIndex)*StepsPerIndex > Understepping){
      Running = true;
      Inc = 1;    // Direction - forward
      NumOfSteps = (NewWLIndex - WLIndex)*StepsPerIndex - Understepping;
      StartMultiSteps();
    }
  }
}

void MoveToTarget(void)
{
    if (Running) {
        Running = !RunFinished;
        if (!Running) {
            GetIndex();
            if (Inc > 0) {
                if ((NewWLIndex-WLIndex)*StepsPerIndex > Understepping) Run();
            } else {
                if ((NewWLIndex-WLIndex)*StepsPerIndex < Overstepping) Run();
            };
        };
    } else {
        GetIndex();
        if (WLIndex >= NewWLIndex) {
            OnTarget = true;
            //if ((WLIndex-NewWLIndex) > 1) Message(4,'Incorrect Wavelength position');
        } else {
            Inc = 1;
            _delay_ms(3);
            DoOneStep();
        };
    };
}

int8_t  ExecuteCommand0 (const char  command)
{
  switch( command ){
    case 'X':
      TIMSK1 &= (uint8_t)(~(1 << OCIE1A)); // Запрещаем прерывания по таймеру TIMER1
      RunFinished=true;
      OnTarget = true;
      NumOfSteps = 0;
      break;
    case 'W':{
      uint8_t Input = ~(PINC & 0b11110000);
      if (Input & (1<<4)) GratingNo = '1';
      else if (Input & (1<<5)) GratingNo = '2';
      else if (Input & (1<<6)) GratingNo = '3';
      else if (Input & (1<<7)) GratingNo = '4';
      else GratingNo = 'E';
      GetIndex();
      SendCharToUSB(GratingNo);
      SendNumberToUSB(WLIndex);
      Send_EOL_ToUSB ();
      break;}
    case 'P':
      SendCharToUSB ('T');
      SendCharToUSB ('=');
      SendNumberToUSB (TicksPerStep);
      SendCharToUSB ('M');
      SendCharToUSB ('=');
      SendNumberToUSB (MaxSpeedTPS);
      SendCharToUSB ('A');
      SendCharToUSB ('=');
      SendNumberToUSB (AccelAddition);
      SendCharToUSB ('K');
      SendCharToUSB ('=');
      SendNumberToUSB (RampCoef);
      SendCharToUSB ('Z');
      SendCharToUSB ('=');
      SendNumberToUSB (StoppingZone);
      Send_EOL_ToUSB ();
      break;
    default:
      return 1;
  }
  return 0;
}

int8_t  ExecuteCommand2 (const char     command
                        ,const uint8_t  argument)
{
  return 1;
}

int8_t  ExecuteCommand4 (const char      command
                        ,const uint16_t  argument)
{
  switch (command){
    case 'T':
      TicksPerStep = argument - 1; // Устанавливается величина, на единицу меньшая,
      break;          // так как счетчик в контроллере срабатывает по значению N+1.
    case 'M':
      MaxSpeedTPS = argument - 1; // Устанавливается величина, на единицу меньшая,
      break;          // так как счетчик в контроллере срабатывает по значению N+1.
    case 'A':
      AccelAddition = argument;
      break;
    case 'K':
      RampCoef = argument;
      break;
    case 'Z':
      StoppingZone = argument;
      break;
  default:
      return 1;
  }
  return 0;
}

int8_t  ExecuteCommand8 (const char      command
                        ,const uint32_t  argument)
{
  switch( command ){
    case 'G':
      NewWLIndex = argument;
      RunFinished=true;
      GetIndex();
      if (abs(NewWLIndex - WLIndex) > 1) {
        OnTarget = false;
        Run();
      } else OnTarget = true;
      break;
    case 'S':
      NumOfSteps = argument;
      Ramp=false;
      OnTarget = true;
      RunFinished=false;
      Inc = 1;    // Direction - forward
      OCR1AH = (uint8_t)((TicksPerStep >> 8) & 0xFF);
      OCR1AL = (uint8_t)(TicksPerStep & 0xFF);
      TCNT1H = 0x00;
      TCNT1L = 0x00;
      TIMSK1 |= (1 << OCIE1A);    // Разрешаем прерывания по таймеру TIMER1
      break;
    default:
      return 1;
  }
  SendCharToUSB ('<');
  Send_EOL_ToUSB ();
  return 0;
}


void  SetupHardware (void)
{
  /* Disable watchdog if enabled by bootloader/fuses */
  MCUSR &= ~(1 << WDRF);
  wdt_disable();

  /* Disable clock division */
  clock_prescale_set(clock_div_1);

  // Установки для управления шаговым двигателем --------------------------------
  TCCR1A = (0 << COM1A1) |  //  Инициализация таймера микроконтроллера TIMER1
           (0 << COM1A0) |
           (0 << COM1B1) |
           (0 << COM1B0) |
           (0 << FOC1A) |
           (0 << FOC1B) |
           (0 << WGM11) |
           (0 << WGM10);

  TCCR1B = (0 << ICNC1) |
           (0 << ICES1) |
           (0 << WGM13) |
          (1 << WGM12) |     // Режим сброса при совпадении OCR1A
          (1 << CS12) |      // Выбор тактовой частоты
          (0 << CS11) |      // Данными битами
          (1 << CS10);       // установлено деление на 1024

 OCR1AH = (uint8_t)((TicksPerStep >> 8) & 0xFF);  // Заносим количество тактов
 OCR1AL = (uint8_t)(TicksPerStep & 0xFF);         // в регистр счетчика

 // Установки ввода-вывода -----------------------------------------------------
     DDRD = 0b00001111;    // Биты 0-3 на вывод (управление шаговыми двигателем)
     PhaseI = 0;
     PORTD = (PORTD & 0b11110000) | Phases[PhaseI]; // Выводим в PORTD начальную фазу
     DDRB |= 0b00110011;    // Биты 0,1,4,5 на вывод (0 - такт сдвига регистра, 1 - бит данных регистра
                            // 4,5 - биты управления A и B)
     PORTB &= 0b11001100;   // Устанавливаем в ноль биты 0,1,4,5
     DDRC &= 0b00001011;    // Бит 2 порта C на вход - считывание угол-код
                            // и биты 4-7 на вход (определение решетки)
     PORTC |= 0b11110100;   // входы с pullup
 //------------------------------------------------------------------------------
}


void  BeforeLoop (void)
{
  TicksPerStep = 100;
  MaxSpeedTPS = 30;
  AccelAddition = 4;
  RampCoef = 1032;   // (1.008 * 1024);
  StoppingZone = 100;
  OnTarget = true;

  //---- Обнуление сдвигового регистра -------------------------------------
      PORTB &= 0b11001100;   // Устанавливаем в ноль биты 	0,1,4,5
      for (char i = 0; i < 20; i++)
      {
          _delay_us(1);
          PORTB |= 1;          // Такт - продвигаем ноль по регистру
          _delay_us(1);
          PORTB &= 0b11001100; // Выводим 0 - возврат в исх. сост.
      }
  //------------------------------------------------------------------------
}


void  InLoop (void)
{
  // Moving to target position if necessary
  if (!OnTarget) MoveToTarget();

  // Forming StepsFinished Echo if necessary
  if (OnTarget && RunFinished){
    SendDone();
    RunFinished=false;
  }
}

//-----------------------------------------------------------------------------
// Обработчик прерывания от таймера 1 (COMPA)
//-----------------------------------------------------------------------------
ISR(TIMER1_COMPA_vect){
  sei();                  // Дадим возможность обработать прерывания от USB...
  if (NumOfSteps) {
    if (Ramp) {
      if (NumOfSteps <= RealZone) {
        if ((NumOfSteps == RealZone) && (CurrentRamp == 0)) CurrentRamp = 1024;
          DoOneStep();
          NumOfSteps--;
          CurrentRamp *= RampCoef;
          CurrentRamp = CurrentRamp >> 10;
          CurTPS = (uint16_t)((CurrentRamp*TicksPerStep >> 10) & 0xFFFF);
          OCR1AH = (uint8_t)((CurTPS >> 8) & 0xFF);
          OCR1AL = (uint8_t)(CurTPS & 0xFF);
        } else {
          DoOneStep();
          NumOfSteps--;
          if (CurrentRamp > 0) {
            CurrentRamp = CurrentRamp << 10;
            CurrentRamp /= RampCoef;
            if (CurrentRamp > 1024) {
              CurTPS = (uint16_t)(((CurrentRamp*TicksPerStep) >> 10) & 0xFFFF);
            } else {
              CurTPS = TicksPerStep;
              CurrentRamp = 0;
            }
            OCR1AH = (uint8_t)((CurTPS >> 8) & 0xFF);
            OCR1AL = (uint8_t)(CurTPS & 0xFF);
          }
        }
    } else {
      DoOneStep();
      NumOfSteps--;
    }
  } else {
    TIMSK1 &= (uint8_t)(~(1 << OCIE1A)); // Запрещаем прерывания по таймеру TIMER1
    RunFinished=true;
  }
}

