////////////////////////////////////////////////////////////////////////////////
/*! @file MSDN204A.c  Firmware for Digital Lock-in amplifier.
- Part of the MSDN firmware.
- Part of the Research Assistant project.
- Uses  LUFA library written by
   - Copyright(C) 2012, Dean Camera, dean [at] fourwalledcubicle [dot] com,
                                     www.lufa-lib.org.

- Copyright(C) 2014, Dmitry K. Nelson, St.Petersburg, Russia.
                     D.Nelson@mail.ioffe.ru
- Copyright(C) 2016, Viktor E. Bursian, St.Petersburg, Russia.
                     Viktor.Bursian@mail.ioffe.ru
*///////////////////////////////////////////////////////////////////////////////
#include "Microsin_USB_IO.h"
#include "CommandParser.h"
#include <avr/wdt.h>
#include <avr/power.h>
#include <util/delay.h>
#include <avr/iousb162.h>
#include <avr/interrupt.h>
#include <stdlib.h>
//#include <avr/io.h>
//#include <stdio.h>
//#include <string.h>
//! @todo{MSDN} Неясная роль inttypes здесь
#include <inttypes.h>
//! @todo{MSDN} Неясная роль math
#include <math.h>

//=============================================================================
// Вход опорного сигнала A    Бит 5 порта D
// Вход опорного сигнала B    Бит 6 порта D
// Ввод данных АЦП            Бит 0 порта D - тактовый сигнал для АЦП
//                            Бит 1 порта D - инициализация передачи АЦП
//                            Бит 2 порта D - данные из АЦП (ADC_DO)
//                            Бит 3 порта D - данные в АЦП (ADC_DI)
//
// Управление усилением       Бит 0 порта B - усиление 10^1 = 10
//                            Бит 1 порта B - усиление 10^2 = 100
//                            Бит 4 порта B - усиление 10^3 = 1000
// Когда все биты установлены в 0, усиление равно 10^0 = 1.
//=============================================================================
#define ADC_CLK PD0
#define ADC_CS PD1
#define ADC_DO PD2
#define ADC_DI PD3
#define ADCRANGE10 0b1010000000000000
#define ADCRANGE5  0b1010100000000000
#define ADCRANGE25 0b1011000000000000
#define ADCCONTROL 0b1000000000010000  // Канал 0 несимметричный, в дополнительном коде

uint16_t NumOfSamples;  	// Число отсчетов АЦП
uint16_t ADCRegisterData;	// Набор битов для записи в регистр АЦП
int16_t InputADC;	       	// Переменная считывания из АЦП
int32_t ADCData;			// Накопитель данных АЦП
int32_t ADCData_A;			// Накопитель данных АЦП, канал A
int32_t ADCData_B;			// Накопитель данных АЦП, канал B

uint8_t DelayVar;       	// Переменная для задержки

bool Reference_A;
bool Reference_B;
bool MeasureEnabled;
bool SampleDone;

uint32_t PeriodInTicks;	// Число тиков (по 1/16 мкс) на период
uint16_t CycleInTicks;  // Число тиков (по 1/16 мкс) на цикл
uint16_t CycleNo;       // Номер цикла счета
uint16_t NumOfCycles;   // Количество циклов
uint32_t RealPeriod;    // Установленный период в мкс

void StartTimer(void)
{
  CycleNo = 0;
  TCNT1H = 0x00;			// Обнуляем
  TCNT1L = 0x00;			// счетчик
  TIFR1 |= (1 << OCF1A);    // Сбрасываем прерывания таймера TIMER1
  TIMSK1 |= (1 << OCIE1A);	// Разрешаем прерывания по таймеру TIMER1
}

void WriteADCRegister(void)
{
    unsigned char i;
    PORTD &= ~(1<<ADC_CS); //ADC_CS=0 (0 на АЦП) активируем передачу данных
    DelayVar++;DelayVar++;	// Для задержки
    for(i=0; i < 16;i++) 	// Передаем данные - старший бит первый
    {
        if (ADCRegisterData & 0x8000)
            PORTD |= (1<<ADC_DI); //(выводим 1 на АЦП)
        else
            PORTD &= ~(1<<ADC_DI); //(выводим 0 на АЦП)
        DelayVar++;DelayVar++; // Для задержки
        PORTD &= ~(1<<ADC_CLK);//ADC_CLK=0 (0 на АЦП)
        DelayVar++;DelayVar++; // Для задержки
        PORTD |= (1<<ADC_CLK); //ADC_CLK=1 (1 на АЦП)
        DelayVar++;DelayVar++; // Для задержки
        ADCRegisterData<<=1;
    }
    DelayVar++;DelayVar++;DelayVar++;// Для задержки
    PORTD &= ~(1<<ADC_DI); // устанавливаем для DI состояние 0 (соотв. 0 на входе АЦП).
}

void ReadADC(void)  // For AD7321
{
    unsigned char i;
    InputADC = 0;
// ---------------Ввод данных из АЦП в переменную InputADC ----------------------------
    //ADC_DI=0 (0 на АЦП) отключаем запись в регистр АЦП
    PORTD &= ~(1<<ADC_DI);
    //ADC_CS=0 (0 на АЦП) активируем передачу данных
    PORTD &= ~(1<<ADC_CS);
    DelayVar++;DelayVar++; // Для задержки
// ------------- 2 такта пропуска нулевых битов и бита номера канала ---------------
    for(i=0; i < 2;i++)
    {
        PORTD &= ~(1<<ADC_CLK);//ADC_CLK=0 (0 на АЦП)
        DelayVar++;DelayVar++; // Для задержки
        PORTD |= (1<<ADC_CLK); //ADC_CLK=1 (1 на АЦП)
        DelayVar++;DelayVar++; // Для задержки
    }
// --------------- чтение знака и 12 бит результата АЦП, старший бит первый ---------------
    for(i=0;i<13;i++)
    {
        PORTD &= ~(1<<ADC_CLK);//ADC_CLK=0 (0 на АЦП)
        DelayVar++;DelayVar++; // Для задержки
        InputADC<<=1;
        if (PIND & (1<<ADC_DO))
            InputADC |= 1;
        PORTD |= (1<<ADC_CLK); //ADC_CLK=1 (1 на АЦП)
        DelayVar++;DelayVar++; // Для задержки
    }
    PORTD |= (1<<ADC_CS); // Отключение передачи данных
    if (InputADC & 0b0001000000000000) InputADC |= 0b1111000000000000;
// ---------------конец ввода АЦП---------------------------------------------------------
}

int8_t  ExecuteCommand0 (const char  command)
{
  switch( command ){
    case 'M':
      NumOfSamples = 0;
      ADCData = 0;
      ADCData_A = 0;
      ADCData_B = 0;
      MeasureEnabled = true;
      SampleDone = false;
      Reference_A = true;
      Reference_B = false;
      StartTimer();
      break;
    case 'X':
      MeasureEnabled = false;
      TIMSK1 &= ~(1 << OCIE1A); 	// Запрещаем прерывания по таймеру TIMER1
      break;
	default:
      return 1;
  }
  return 0;
}

int8_t  ExecuteCommand2 (const char     command
                        ,const uint8_t  argument)
{
  switch( command ){
    case 'G':
      if (argument == 0) {
          PORTB &= ~(1<<PB0);
          PORTB &= ~(1<<PB1);
          PORTB &= ~(1<<PB4);
      } else if (argument == 1) {
          PORTB |= (1<<PB0);
          PORTB &= ~(1<<PB1);
          PORTB &= ~(1<<PB4);
      } else if (argument == 2) {
          PORTB |= (1<<PB1);
          PORTB &= ~(1<<PB0);
          PORTB &= ~(1<<PB4);
      } else if (argument == 3) {
          PORTB |= (1<<PB4);
          PORTB &= ~(1<<PB0);
          PORTB &= ~(1<<PB1);
      } else {
          return 2;
      }
      break;
    case 'R':
      if (argument == 0) {
          ADCRegisterData = ADCRANGE10;
          WriteADCRegister();
      } else if (argument == 1) {
          ADCRegisterData = ADCRANGE5;
          WriteADCRegister();
      } else if (argument == 2) {
          ADCRegisterData = ADCRANGE25;
          WriteADCRegister();
      } else {
          return 2;
      }
      break;
    default:
      return 1;
  }
  return 0;
}

int8_t  ExecuteCommand4 (const char      command
                        ,const uint16_t  argument)
{
  return 1;
}

int8_t  ExecuteCommand8 (const char      command
                        ,const uint32_t  argument)
{
  switch( command ){
    case 'T':
      PeriodInTicks = argument;                  // период в мкс
      PeriodInTicks = PeriodInTicks * 16;        // Число тиков (по 1/16 мкс) на период
      NumOfCycles  = (PeriodInTicks / 0xFFFF) + 1;  // Число циклов длительностью не более FFFF тактов
      CycleInTicks = PeriodInTicks / NumOfCycles -1;
      OCR1AH = (uint8_t)((CycleInTicks >> 8) & 0xFF);	// Заносим количество тактов
      OCR1AL = (uint8_t)(CycleInTicks & 0xFF);			// в регистр счетчика
      uint32_t RealPeriod = CycleInTicks;
      RealPeriod = (RealPeriod * NumOfCycles) >> 4;
      SendNumberToUSB(RealPeriod);
      Send_EOL_ToUSB ();
      break;
    default:
      return 1;
  }
  return 0;
}

void  SetupHardware (void)
{
  /* Disable watchdog if enabled by bootloader/fuses */
  MCUSR &= ~(1 << WDRF);
  wdt_disable();

  /* Disable clock division */
  clock_prescale_set(clock_div_1);
  // Установки таймеров --------------------------------

      TCCR1A = (0 << COM1A1) |	//  Инициализация таймера микроконтроллера TIMER1
               (0 << COM1A0) |
               (0 << COM1B1) |
               (0 << COM1B0) |
               (0 << FOC1A) |
               (0 << FOC1B) |
               (0 << WGM11) |
               (0 << WGM10);

      TCCR1B = (0 << ICNC1) |
               (0 << ICES1) |
               (0 << WGM13) |
               (1 << WGM12) |   // Режим сброса при совпадении OCR1A
               (0 << CS12) |    // Выбор тактовой частоты
               (0 << CS11) |    // Данными битами
               (1 << CS10);     // установлена частота без деления (время на такт 1/16 мкс)
      PeriodInTicks = 0;
      OCR1AH = (uint8_t)((PeriodInTicks >> 8) & 0xFF);	// Заносим количество тактов
      OCR1AL = (uint8_t)(PeriodInTicks & 0xFF);			// в регистр счетчика
      TIMSK1 &= ~(1 << OCIE1A); 				    	// Запрещаем прерывания по таймеру TIMER1
  // -----------------------------------------------------------------------------
  // Ввод данных с АЦП
    //настройка ADC_DO как входа без pullup
    DDRD &= ~(1<<ADC_DO);
    PORTD &= ~(1<<ADC_DO);
    //настройка ADC_CLK, ADC_CS и ADC_DI как выходов.
    DDRD |= ((1<<ADC_CS)|(1<<ADC_CLK)|(1<<ADC_DI));
    PORTD |= (1<<ADC_CLK); // у CLK начальное состояние 1 (соотв. 1 на входе АЦП).
    PORTD |= (1<<ADC_CS);  // у CS начальное состояние 1 (соотв. 1 на входе АЦП).
    PORTD &= ~(1<<ADC_DI); // у DI начальное состояние 0 (соотв. 0 на входе АЦП).

    //настройка PD5 и PD6 как входов c pullup.
    DDRD &= ~(1<<PD5);		 // бит 5 для входа опорного канала A
    PORTD |= (1<<PD5); //   включение pullup.
    DDRD &= ~(1<<PD6);		 // бит 6 для входа опорного канала B
    PORTD |= (1<<PD6); //   включение pullup.

  // Управление усилением       Бит 0 порта B - усиление 10^1 = 10
    //настройка PB0, PB1 и PB4 как выходов.
    DDRB |= ((1<<PB0)|(1<<PB1)|(1<<PB4));
    //все биты устанавливаются в 0, усиление равно 10^0 = 1.
    PORTB &= ~(1<<PB0);
    PORTB &= ~(1<<PB1);
    PORTB &= ~(1<<PB4);
  //------------------------------------------------------------------------------
}


void  BeforeLoop (void)
{
  ADCRegisterData = ADCCONTROL; // Установка режима АЦП
  WriteADCRegister();
  ADCRegisterData = ADCRANGE10; // Установка предела АЦП
  WriteADCRegister();
}


void  InLoop (void)
{
  if (MeasureEnabled) {
    if (PIND & (1<<PD5)) Reference_A = true; else Reference_A = false;
    if (PIND & (1<<PD6)) Reference_B = true; else Reference_B = false;
    ReadADC();
    if (abs(InputADC) > abs(ADCData)) ADCData = InputADC;
//    ADCData += InputADC;
//    ADCData = InputADC;
    if (Reference_A) ADCData_A += InputADC; else ADCData_A -= InputADC;
    if (Reference_B) ADCData_B += InputADC; else ADCData_B -= InputADC;
    NumOfSamples++;
    if (SampleDone) {
      SendCharToUSB ('V');
      SendNumberToUSB (ADCData);// Данные АЦП, амплитуда;
      SendCharToUSB ('X');
      SendNumberToUSB (ADCData_A);// Данные АЦП канал A;
      SendCharToUSB ('Y');
      SendNumberToUSB (ADCData_B);// Данные АЦП канал B;
      SendCharToUSB ('N');
      SendNumberToUSB (NumOfSamples);// Число отсчетов;
      Send_EOL_ToUSB ();
      SampleDone = false;
      NumOfSamples = 0;
      ADCData = 0;
      ADCData_A = 0;
      ADCData_B = 0;
    }
  }
}

//-----------------------------------------------------------------------------
// Обработчик прерывания от таймера 1 (COMPA)
//-----------------------------------------------------------------------------
ISR(TIMER1_COMPA_vect){
  sei();         // Дадим возможность обработать прерывания от USB...
  CycleNo++;
  if (CycleNo == NumOfCycles){
    SampleDone = true;
    CycleNo = 0;
  }
}
