////////////////////////////////////////////////////////////////////////////////
/*! @file MSDN204B.c  Firmware for MDR Driver without DUK.
- Part of the MSDN firmware.
- Part of the Research Assistant project.
- Uses  LUFA library written by
   - Copyright(C) 2012, Dean Camera, dean [at] fourwalledcubicle [dot] com,
                                     www.lufa-lib.org.

- Copyright(C) 2014, Dmitry K. Nelson, St.Petersburg, Russia.
                     D.Nelson@mail.ioffe.ru
- Copyright(C) 2016, Viktor E. Bursian, St.Petersburg, Russia.
                     Viktor.Bursian@mail.ioffe.ru
*///////////////////////////////////////////////////////////////////////////////

//=============================================================================
// Контроллер управления шаговым двигателем спектрометра МДР.
// Для инициализации используется метка, которая подается на пин 2 порта C.
// Завершение инициализации по размыканию контакта метки.
// Работа с шаговым двигателем. Тактовый сигнал шагов подается на пин 0 порта D
//                              Сигнал направления подается на пин 1 порта D
//                              Сигнал разрешения (L) подается на пин 2 порта D
// Пины 4,5,6,7 порта C используются для ввода номера установленной решетки.
//=============================================================================

#include "Microsin_USB_IO.h"
#include "CommandParser.h"
#include <avr/wdt.h>
#include <avr/power.h>
#include <util/delay.h>
#include <avr/iousb162.h>
#include <avr/interrupt.h>
//#include <avr/io.h>
//#include <stdio.h>
//#include <string.h>
//! @todo{MSDN} Неясная роль inttypes здесь
#include <inttypes.h>
//! @todo{MSDN} Неясная роль math
#include <math.h>

uint8_t  GratingNo;

uint16_t  TicksPerStep;      // Задает время на шаг для сканирования
uint16_t  MaxSpeedTPS;       // Задает максимальную скорость перемотки
uint16_t  CurTPS;            // Переменная для текущей паузы между шагами
uint16_t  AccelAddition; 	// Коэф увеличения начальной паузы при разгоне
uint32_t  CurrentRamp = 0;   // Текущее значение в процессе разгона и торможения
uint16_t  RampCoef;          // Коэф уменьшения периода шагов за шаг
uint16_t  StoppingZone;      // Параметр области (в шагах) разгона-торможения
uint16_t  RealZone =0;	    // Реальное используемое значение
uint32_t  NumOfSteps;        // Заданное число шагов
bool      Ramp=false;         // Признак разгона-торможения
volatile bool    RunFinished=false; // Признак окончания шагания
bool      Initiated=false;
bool      RunningBackward;
bool      MoveToZeroPoint = false;
bool      NearTarget = false;
bool      OnTarget = false;

// Индекс задается в шагах, 1 нм = 1000 шагов.
uint32_t  WLIndex;
uint32_t  NewWLIndex;

#define MinIndex 185000
#define MaxIndex 1020000
#define Overstepping 300
#define Understepping 250

void SetDirForward(void)
{
PORTD |= (1 << 1);
RunningBackward = false;
}

void SetDirBackward(void)
{
PORTD &= ~(1 << 1);
RunningBackward = true;
}

void SendDone(void)
{
    SendCharToUSB('D');
    SendCharToUSB('o');
    SendCharToUSB('n');
    SendCharToUSB('e');
    Send_EOL_ToUSB ();
}

void GoWithRamp(void)
{
    Ramp=1;
    PORTD |= (1);			// Выводим 1 на пин PORTD0
    PORTD |= (1 << 2);	// Выводим 1 на пин PORTD2 (Enable)
    if (NumOfSteps >= 2*StoppingZone) RealZone = StoppingZone;
    else RealZone = NumOfSteps / 2;
    CurrentRamp = AccelAddition << 10;
    CurTPS = (uint16_t)(((CurrentRamp*MaxSpeedTPS) >> 10) & 0xFFFF);
    OCR1AH = (uint8_t)((CurTPS >> 8) & 0xFF);
    OCR1AL = (uint8_t)(CurTPS & 0xFF);
    TCNT1H = 0x00;
    TCNT1L = 0x00;
    TIMSK1 |= (1 << OCIE1A);    // Разрешаем прерывания по таймеру TIMER1
}

void MoveToMarker(void)
{
    if (!NearTarget){
        PORTD |= 0x01;  // Выводим 1 на пин PORTD0 - изготовка для шага
        _delay_us(10);
        PORTD &= 0xFE;  // Выводим 0 на пин PORTD0 - делаем шаг
        _delay_ms(5);
        if (PINC & (1<<2)) NearTarget = true;
    } else {
      if (!OnTarget){
        PORTD |= 0x01;  // Выводим 1 на пин PORTD0 - изготовка для шага
        _delay_us(10);
        PORTD &= 0xFE;  // Выводим 0 на пин PORTD0 - делаем шаг
        _delay_ms(5);
        if ( !((PINC & (1<<2))>0) ) OnTarget = true;
      }
    }
}

int8_t  ExecuteCommand0 (const char  command)
{
  switch( command ){
    case 'I':
      Initiated=false;
      MoveToZeroPoint = false;
      NumOfSteps = Overstepping;
      SetDirBackward();
      GoWithRamp();        // Поехали назад на Overstepping шагов.
      break;
    case 'X':
      TIMSK1 &= (uint8_t)(~(1 << OCIE1A)); // Запрещаем прерывания по таймеру TIMER1
      PORTD &= ~(1 << 2); // Выводим 0 на пин PORTD2 (Disable)
      PORTD |= (1);		  // Выводим 1 на пин PORTD0
      SendDone();
      NumOfSteps = 0;
      break;
    case 'W':
      if (Initiated){
          GratingNo = ~(PINC & 0b11110000);
          if (GratingNo & (1<<4)) GratingNo = '1';
          else if (GratingNo & (1<<5)) GratingNo = '2';
          else if (GratingNo & (1<<6)) GratingNo = '3';
          else if (GratingNo & (1<<7)) GratingNo = '4';
          else GratingNo = 'E';
          SendCharToUSB(GratingNo);
          cli();
          uint32_t TheIndex = WLIndex;
          sei();
          SendNumberToUSB(TheIndex);
      } else {
          SendCharToUSB('?');
      }
      Send_EOL_ToUSB ();
      break;
    case 'P':
      SendCharToUSB ('T');
      SendCharToUSB ('=');
      SendNumberToUSB (TicksPerStep);
      SendCharToUSB ('M');
      SendCharToUSB ('=');
      SendNumberToUSB (MaxSpeedTPS);
      SendCharToUSB ('A');
      SendCharToUSB ('=');
      SendNumberToUSB (AccelAddition);
      SendCharToUSB ('K');
      SendCharToUSB ('=');
      SendNumberToUSB (RampCoef);
      SendCharToUSB ('Z');
      SendCharToUSB ('=');
      SendNumberToUSB (StoppingZone);
      Send_EOL_ToUSB ();
      break;
    default:
      return 1;
  }
  return 0;
}

int8_t  ExecuteCommand2 (const char     command
                        ,const uint8_t  argument)
{
  return 1;
}

int8_t  ExecuteCommand4 (const char      command
                        ,const uint16_t  argument)
{
  switch (command){
    case 'T':
      TicksPerStep = argument - 1; // Устанавливается величина, на единицу меньшая,
      break;          // так как счетчик в контроллере срабатывает по значению N+1.
    case 'M':
      MaxSpeedTPS = argument - 1; // Устанавливается величина, на единицу меньшая,
      break;          // так как счетчик в контроллере срабатывает по значению N+1.
    case 'A':
      AccelAddition = argument;
      break;
    case 'K':
      RampCoef = argument;
      break;
    case 'Z':
      StoppingZone = argument;
      break;
  default:
      return 1;
  }
  return 0;
}

int8_t  ExecuteCommand8 (const char      command
                        ,const uint32_t  argument)
{
  switch( command ){
    case 'V':
      WLIndex = argument;
      Initiated=true;
      break;
    case 'G':
      if (Initiated){
          NewWLIndex = argument;
          if (NewWLIndex == WLIndex){
              RunFinished = true;
              SetDirForward();
              break;
          }
          if (NewWLIndex > WLIndex) {
              NumOfSteps = NewWLIndex - WLIndex;
              SetDirForward();
          } else {
              NumOfSteps = (WLIndex - NewWLIndex) + Overstepping;
              SetDirBackward();
          }
          GoWithRamp();
      }
      break;
    case 'S':
      if (Initiated){
          Ramp=0;
          NumOfSteps = argument;
          SetDirForward();
          PORTD |= (1);			// Выводим 1 на пин PORTD0
          PORTD |= (1 << 2);	// Выводим 1 на пин PORTD2 (Enable)
          OCR1AH = (uint8_t)((TicksPerStep >> 8) & 0xFF);
          OCR1AL = (uint8_t)(TicksPerStep & 0xFF);
          TCNT1H = 0x00;
          TCNT1L = 0x00;
          TIMSK1 |= (1 << OCIE1A);    // Разрешаем прерывания по таймеру TIMER1
      }
      break;
    default:
      return 1;
  }
  SendCharToUSB ('<');
  Send_EOL_ToUSB ();
  return 0;
}


void  SetupHardware (void)
{
  /* Disable watchdog if enabled by bootloader/fuses */
  MCUSR &= ~(1 << WDRF);
  wdt_disable();

  /* Disable clock division */
  clock_prescale_set(clock_div_1);

  // Установки для управления шаговым двигателем --------------------------------
  TCCR1A = (0 << COM1A1) |  //  Инициализация таймера микроконтроллера TIMER1
           (0 << COM1A0) |
           (0 << COM1B1) |
           (0 << COM1B0) |
           (0 << FOC1A) |
           (0 << FOC1B) |
           (0 << WGM11) |
           (0 << WGM10);

  TCCR1B = (0 << ICNC1) |
           (0 << ICES1) |
           (0 << WGM13) |
           (1 << WGM12) |   // Режим сброса при совпадении OCR1A
           (1 << CS12) |    // Выбор тактовой частоты
           (0 << CS11) |    // Данными битами
           (0 << CS10);     // установлено деление на 256
                            // (время на полный шаг получается 32 мкс)

  OCR1AH = (uint8_t)((MaxSpeedTPS >> 8) & 0xFF);  // Заносим количество тактов
  OCR1AL = (uint8_t)(MaxSpeedTPS & 0xFF);         // в регистр счетчика

  // Установки ввода-вывода -----------------------------------------------------
  DDRD |= 0b00000111;   // Биты 0,1,2 на вывод (управление шаговыми двигателем)
                        // Эти биты на драйвер ШДР подаются через опторазвязку
  PORTD |= (1);			// Выводим 1 на пин PORTD0 - 0 на драйвер ШДР с учетом инверсии
  PORTD |= (1<< 1);		// Выводим 1 на пин PORTD1 - 0 на драйвер ШДР с учетом инверсии
  PORTD &= ~(1 << 2);	// Выводим 0 на пин PORTD2 - 1 на драйвер ШДР с учетом инверсии (Disable)
  DDRC &= 0b00001011;   // Бит 2 порта C на вход - считывание метки
                        // и биты 4-7 на вход (определение решетки)
  PORTC |= 0b11110100;  // входы с pullup
  //------------------------------------------------------------------------------
}


void  BeforeLoop (void)
{
  TicksPerStep = 50;
  MaxSpeedTPS = 20;
  AccelAddition = 4;
  RampCoef = 1032;   // (1.008 * 1024);
  StoppingZone = 200;
  WLIndex = 500000;
  Initiated=false;
  RunFinished=false;
}


void  InLoop (void)
{
  if (RunFinished) {
    RunFinished=false;
    if (Initiated) {
      if (RunningBackward) {         // Move to target position
        SetDirForward();
        NumOfSteps = Overstepping;
        GoWithRamp();
      } else {
        SendDone();                 // Forming StepsFinished Echo
      }
    } else {
      if (RunningBackward) {         // Move to target position
        SetDirForward();
        NumOfSteps = Understepping;
        GoWithRamp();
      } else {
        MoveToZeroPoint = true;
        NearTarget = false;
        OnTarget = false;
        PORTD |= (1 << 2);	// Выводим 1 на пин PORTD2 (Enable)
      }
    }
  }
  if (MoveToZeroPoint) {
    MoveToMarker();
    if (OnTarget) {
      MoveToZeroPoint = false;
      PORTD &= ~(1 << 2); // Выводим 0 на пин PORTD2 (Disable)
      PORTD |= (1);		  // Выводим 1 на пин PORTD0
      SendCharToUSB ('<');
      Send_EOL_ToUSB ();
    }
  }
}

//-----------------------------------------------------------------------------
// Обработчик прерывания от таймера 1 (COMPA)
//-----------------------------------------------------------------------------
ISR(TIMER1_COMPA_vect){
  sei();                  // Дадим возможность обработать прерывания от USB...
  if (NumOfSteps) {
    if (Ramp) {
      if (NumOfSteps <= RealZone) {
        if ((NumOfSteps == RealZone) && (CurrentRamp == 0)) CurrentRamp = 1024;
        if (PORTD & 0x01) {
           PORTD &= 0xFE;  // Выводим 0 на пин PORTD0 - делаем шаг
           NumOfSteps--;
           if (RunningBackward) WLIndex--; else WLIndex++;
           CurrentRamp *= RampCoef;
           CurrentRamp = CurrentRamp >> 10;
           CurTPS = (uint16_t)((CurrentRamp*MaxSpeedTPS >> 10) & 0xFFFF);
           OCR1AH = (uint8_t)((CurTPS >> 8) & 0xFF);
           OCR1AL = (uint8_t)(CurTPS & 0xFF);
        } else {
          PORTD |= 0x01;  // Выводим 1 на пин PORTD0 - изготовка для шага
        };
      } else {
        if (PORTD & 0x01) {
          PORTD &= 0xFE;  // Выводим 0 на пин PORTD0 - делаем шаг
          NumOfSteps--;
          if (RunningBackward) WLIndex--; else WLIndex++;
          if (CurrentRamp > 0) {
            CurrentRamp = CurrentRamp << 10;
            CurrentRamp /= RampCoef;
            if (CurrentRamp > 1024) {
              CurTPS = (uint16_t)(((CurrentRamp*MaxSpeedTPS) >> 10) & 0xFFFF);
            } else {
              CurTPS = MaxSpeedTPS;
              CurrentRamp = 0;
            }
            OCR1AH = (uint8_t)((CurTPS >> 8) & 0xFF);
            OCR1AL = (uint8_t)(CurTPS & 0xFF);
          }
        } else {
          PORTD |= 0x01;  // Выводим 1 на пин PORTD0 - изготовка для шага
        };
      };
    } else {
      if (PORTD & 0x01) {
        PORTD &= 0xFE;  // Выводим 0 на пин PORTD0 - делаем шаг
        NumOfSteps--;
        if (RunningBackward) WLIndex--; else WLIndex++;
      } else {
        PORTD |= 0x01;  // Выводим 1 на пин PORTD0 - изготовка для шага
      };
    };
  } else {
    TIMSK1 &= (uint8_t)(~(1 << OCIE1A)); // Запрещаем прерывания по таймеру TIMER1
    PORTD &= ~(1 << 2);		// Выводим 0 на пин PORTD2 (Disable)
    PORTD |= (1);			// Выводим 1 на пин PORTD0
    RunFinished=true;
  };
}

