////////////////////////////////////////////////////////////////////////////////
/*! @file MSDN204C.c  Firmware for Programmable Timer.
- Part of the MSDN firmware.
- Part of the Research Assistant project.
- Uses  LUFA library written by
   - Copyright(C) 2012, Dean Camera, dean [at] fourwalledcubicle [dot] com,
                                     www.lufa-lib.org.

- Copyright(C) 2014, Dmitry K. Nelson, St.Petersburg, Russia.
                     D.Nelson@mail.ioffe.ru
- Copyright(C) 2016, Viktor E. Bursian, St.Petersburg, Russia.
                     Viktor.Bursian@mail.ioffe.ru
*///////////////////////////////////////////////////////////////////////////////
#include "Microsin_USB_IO.h"
#include "CommandParser.h"
#include <avr/wdt.h>
#include <avr/power.h>
#include <avr/iousb162.h>
#include <avr/interrupt.h>
//#include <avr/io.h>
//#include <stdio.h>
//#include <string.h>
//! @todo{MSDN} Неясная роль inttypes здесь
#include <inttypes.h>

#define PULSE_DURATION 30  // Длительность импульса, задается в тактах по 1/16 мкс
//=============================================================================
// Выход сигнала таймера      Бит 0 порта B - PB0 (контакт P1 на плате)
//=============================================================================

uint32_t PeriodInTicks;	// Число тиков (по 1/16 мкс) на период
uint16_t CycleInTicks;  // Число тиков (по 1/16 мкс) на цикл
uint16_t CycleNo;       // Номер цикла счета
uint16_t NumOfCycles;   // Количество циклов
uint32_t RealPeriod;    // Установленный период в мкс
bool OneShot = false;   // Признак однократного запуска

void StartTimer(void)
{
  CycleNo = 0;
  TCNT1H = 0x00;			// Обнуляем
  TCNT1L = 0x00;			// счетчик
  PORTB &= ~(1<<PB0);       // Устанавливаем порт в исходное состояние
  TIFR0 |= (1 << OCF0A);    // Сбрасываем прерывания таймера TIMER0
  TIFR1 |= (1 << OCF1A);    // Сбрасываем прерывания таймера TIMER1
  TIMSK0 &= ~(1 << OCIE0A);	// Запрещаем прерывания по таймеру TIMER0
  TIMSK1 |= (1 << OCIE1A);	// Разрешаем прерывания по таймеру TIMER1
}

int8_t  ExecuteCommand0 (const char  command)
{
  switch( command ){
    case 'S':
      StartTimer();
      OneShot = false;
      break;
    case 'O':
      StartTimer();
      OneShot = true;
      break;
    case 'X':
      TIMSK0 &= ~(1 << OCIE0A);	// Запрещаем прерывания по таймеру TIMER0
      TIMSK1 &= ~(1 << OCIE1A);	// Запрещаем прерывания по таймеру TIMER1
      PORTB &= ~(1<<PB0);       // Возвращаем порт в исходное состояние
      break;
    default:
        return 1;
  }
  return 0;
}

int8_t  ExecuteCommand2 (const char     command
                        ,const uint8_t  argument)
{
  return 1;
}

int8_t  ExecuteCommand4 (const char      command
                        ,const uint16_t  argument)
{
  return 1;
}

int8_t  ExecuteCommand8 (const char      command
                        ,const uint32_t  argument)
{
  switch( command ){
    case 'T':{
      PeriodInTicks = argument;                  // период в мкс
      PeriodInTicks = PeriodInTicks * 16;        // Число тиков (по 1/16 мкс) на период
      NumOfCycles  = (PeriodInTicks / 0xFFFF) + 1;  // Число циклов длительностью не более FFFF тактов
      CycleInTicks = PeriodInTicks / NumOfCycles -1;
      OCR1AH = (uint8_t)((CycleInTicks >> 8) & 0xFF);	// Заносим количество тактов
      OCR1AL = (uint8_t)(CycleInTicks & 0xFF);			// в регистр счетчика
      uint32_t RealPeriod = CycleInTicks;
      RealPeriod = (RealPeriod * NumOfCycles) >> 4;
      SendNumberToUSB(RealPeriod);
      Send_EOL_ToUSB ();
      break;
    }
    default:
      return 1;
  }
  return 0;
}



void  SetupHardware (void)
{
  /* Disable watchdog if enabled by bootloader/fuses */
  MCUSR &= ~(1 << WDRF);
  wdt_disable();

  /* Disable clock division */
  clock_prescale_set(clock_div_1);
  // Установки таймеров --------------------------------

  TCCR0A = (0 << COM0A1) |	//  Инициализация таймера микроконтроллера TIMER0
           (0 << COM0A0) |
           (0 << COM0B1) |
           (0 << COM0B0) |
           (1 << WGM01) |	// Режим сброса при совпадении OCR0A
           (0 << WGM00);

  TCCR0B = (0 << FOC0A) |
           (0 << FOC0B) |
           (0 << WGM02) |
           (0 << CS12) |    // Выбор тактовой частоты
           (0 << CS11) |    // Данными битами
           (1 << CS10);     // установлена частота без деления (время на такт 1/16 мкс)

  OCR0A = PULSE_DURATION;	// Заносим количество тактов в регистр счетчика
  TIMSK0 &= ~(1 << OCIE0A);	// Запрещаем прерывания по таймеру TIMER0

  TCCR1A = (0 << COM1A1) |	//  Инициализация таймера микроконтроллера TIMER1
           (0 << COM1A0) |
           (0 << COM1B1) |
           (0 << COM1B0) |
           (0 << FOC1A) |
           (0 << FOC1B) |
           (0 << WGM11) |
           (0 << WGM10);


  TCCR1B = (0 << ICNC1) |
           (0 << ICES1) |
           (0 << WGM13) |
           (1 << WGM12) |   // Режим сброса при совпадении OCR1A
           (0 << CS12) |    // Выбор тактовой частоты
           (0 << CS11) |    // Данными битами
           (1 << CS10);     // установлена частота без деления (время на такт 1/16 мкс)
  PeriodInTicks = 0;
  OCR1AH = (uint8_t)((PeriodInTicks >> 8) & 0xFF);	// Заносим количество тактов
  OCR1AL = (uint8_t)(PeriodInTicks & 0xFF);			// в регистр счетчика
  TIMSK1 &= ~(1 << OCIE1A); 				    	// Запрещаем прерывания по таймеру TIMER1
  // -----------------------------------------------------------------------------
  // Настройка выхода
  DDRB |= (1<<PB0);	  // бит 0 порта B - выход сигнала таймера
  PORTB &= ~(1<<PB0); // начальное состояние 0 - активный перепад Rise.
  //------------------------------------------------------------------------------
}


void  BeforeLoop (void)
{
}


void  InLoop (void)
{
}

//-----------------------------------------------------------------------------
// Обработчик прерывания от таймера 1 (COMPA)
//-----------------------------------------------------------------------------
ISR(TIMER1_COMPA_vect){
  sei();                      // Дадим возможность обработать прерывания от USB...
  CycleNo++;
  if (CycleNo == NumOfCycles){
    PORTB |= (1<<PB0);          // формируем активный перепад
    TCNT0 = 0x00;               // Обнуляем счетчик таймера TIMER0
    TIMSK0 |= (1 << OCIE0A);    // Разрешаем прерывания по таймеру TIMER0
    CycleNo = 0;
    if (OneShot) TIMSK1 &= ~(1 << OCIE1A); // Запрещаем прерывания по таймеру TIMER1
  }
}

//-----------------------------------------------------------------------------
// Обработчик прерывания от таймера 0 (COMPA)
//-----------------------------------------------------------------------------
ISR(TIMER0_COMPA_vect){
    sei();                      // Дадим возможность обработать прерывания от USB...
    PORTB &= ~(1<<PB0);         // Возвращаем порт в исходное состояние
    TIMSK0 &= ~(1 << OCIE0A);	// Запрещаем прерывания по таймеру TIMER0
}
