////////////////////////////////////////////////////////////////////////////////
/*! @file MSDN204D.c  Firmware for Temperature regulator.
- Part of the MSDN firmware.
- Part of the Research Assistant project.
- Uses  LUFA library written by
   - Copyright(C) 2012, Dean Camera, dean [at] fourwalledcubicle [dot] com,
                                     www.lufa-lib.org.

- Copyright(C) 2017, Dmitry K. Nelson, St.Petersburg, Russia.
                     D.Nelson@mail.ioffe.ru
- Copyright(C) 2016, Viktor E. Bursian, St.Petersburg, Russia.
                     Viktor.Bursian@mail.ioffe.ru
*///////////////////////////////////////////////////////////////////////////////
#include "Microsin_USB_IO.h"
#include "CommandParser.h"
#include <avr/wdt.h>
#include <avr/power.h>
#include <avr/io.h>
#include <avr/interrupt.h>
#include <util/delay.h>     /* для _delay_ms() */
//! @todo{MSDN} Неясная роль inttypes здесь
#include <inttypes.h>
#include <math.h>

// Используемые контакты ------------------------------------------------
// PF0 на JP2-8 - вход АЦП, измерение датчика температуры
// PB5 на JP1-11 - выход OC1A таймера 1, выход ШИМ питания печки (16-разр.)
// ----------------------------------------------------------------------


/* [ВХОДЫ И РЕЖИМ РАБОТЫ ADC]
1. Используется несимметричный вход ADC0. MUX5..0 = 000000. Усиление = 1.
2. Результат аналого-цифрового преобразования в регистровой паре ADCH:ADCL выравнен вправо
   (бит ADLAR = 0).
3. Тактовая частота ADC выбирается минимально возможная - выбран максимальный коэффициент
   деления XTAL - на 128 (все биты ADPS2, ADPS1, ADPS0 регистра ADCSRA установлены в 1).
   Таким образом, при частоте кварца 16 МГц тактовая частота ADC составит 16000/128=125 кГц.
   При этом max частота циклов ADC составит около 12.5 кГц (одно преобразование состоит
   из 10 шагов последовательного приближения, так как ADC 10-разрядный).
4. Режим работы - single conversion mode, когда запуск преобразования осуществляется
   записью в бит ADSC регистра ADCSRA логической 1.
6. Готовность результата (в регистровой паре ADCH:ADCL) отслеживаем по обнулению бита ADSC
   в регистре ADCSRA. Как только этот бит == 0, считываем результат и запускаем новое
   преобразование, записав в бит ADSC лог. 1.
*/
#define ADC_CLOCK_PRESCALER128 (1<<ADPS2)|(1<<ADPS1)|(1<<ADPS0)
#define MAX_LONG         INT32_MAX
#define MAX_OUTPUT 100
#define SCALEOUTPUT      INT32_MAX/MAX_OUTPUT

uint16_t ADCdata = 0;
uint32_t Dacc = 0;
uint16_t Dout = 0;
uint16_t CurrentValue = 0;
uint16_t SettingValue = 0;
uint32_t HeatingPowerValue = 0;
uint16_t HeatingOutVoltage = 0;
bool  MeasuringFlag = false;
bool  RegulatingFlag = false;
bool  Regulate = false;

uint32_t PeriodInTicks;	// Число тиков (по 4 мкс) на период
uint8_t  CycleInTicks;  // Число тиков (по 4 мкс) на цикл
uint16_t CycleNo;       // Номер цикла счета
uint16_t NumOfCycles;   // Количество циклов
uint16_t MPeriodNo;     // Номер периода счета для измерения
uint16_t MNumOfPeriods; // Количество периодов для измерения

//! Parameters for regulator
double _K = 1;
double result;
uint16_t RoomTCode;
double _ExtraH;

uint16_t sqrt_newton (uint32_t x)  // Hordware algorithm [GLS]
{
  uint32_t m, y, b;
  uint16_t m16, y16, b16, x16;
  if (x & 0xffff0000) {
    if (x & 0xff000000) {
      m = 0x40000000;
    } else {
      m = 0x00400000;
    }
    y = 0;
    do {           // Do 12-16 times.
      b = y | m;
      y >>= 1;
      if (x >= b) {
        x -= b;
        y |=  m;
      }
      m >>= 2;
    } while (m);
    return y;
  } else {
    if(x & 0x0000ff00) {
      m16 = 0x4000;
    } else {
      m16 = 0x0040;
    }
    y16 = 0;
    x16 = x;
    do {          // Do 4-8 times.
      b16 = y16 | m16;
      y16 >>=  1;
      if (x16 >= b16) {
        x16 -= b16;
        y16 |= m16;
      }
      m16 >>= 2;
    } while (m16);
  return y16;
  }
}

int8_t  ExecuteCommand0 (const char  command)
{
  switch( command ){
  case 'M':
    MeasuringFlag = false;
    TIMSK0 |= (1 << OCIE0A);    // Разрешаем прерывания по таймеру TIMER0
    break;
  case 'R':
    RegulatingFlag = false;
    Regulate = true;
    OCR1A = 0;	// Заносим нулевую величину сигнала в регистр счетчика
    TIMSK0 |= (1 << OCIE0A);    // Разрешаем прерывания по таймеру TIMER0
    break;
  case 'X':
    Regulate = false;
    TIMSK0 &= ~(1 << OCIE0A);	// Запрещаем прерывания по таймеру TIMER0
    OCR1A = 0;	// Заносим нулевую величину сигнала в регистр счетчика
    break;
  default:
    return 1;
  }
  return 0;
}

int8_t  ExecuteCommand2 (const char     command
                        ,const uint8_t  argument)
{
  return 1;
}

int8_t  ExecuteCommand4 (const char      command
                        ,const uint16_t  argument)
{
  switch( command ){
  case 'P':
    _K = argument;
    _K = _K/1000;        // Значение пропорционального коэффициента
    break;
  case 'H':
    _ExtraH = argument;
    _ExtraH = _ExtraH/1000;  // Значение коэффициента добавочного нагрева
    break;
  case 'T':
    PeriodInTicks = argument;                 // период в миллисекундах
    MNumOfPeriods = 2000/PeriodInTicks;       // Передавать измерения раз в 2 с
    PeriodInTicks = PeriodInTicks*250;         // Число тиков (по 4 мкс) на период
    NumOfCycles  = (PeriodInTicks / 0xFF) + 1; // Число циклов длительностью не более FF тактов
    CycleInTicks = PeriodInTicks / NumOfCycles -1;
    CycleNo = 0;
    MPeriodNo = 0;
    TCNT0 = 0x00;			// Обнуляем счетчик
    OCR0A = CycleInTicks;	// Заносим количество тактов в регистр счетчика
    break;
  case 'S':
    SettingValue = argument;   // требуемое значение в квантах АЦП
    break;
  case 'O': // устанавливает вых. мощность равной (argument/100)% от макс.
    result = argument;
    result = (result/100)*SCALEOUTPUT;
    if(result > MAX_LONG){
      result = MAX_LONG;
    }
    else if(result < 0){
      result = 0;
    }
    HeatingPowerValue = lround(result);
    HeatingOutVoltage = sqrt_newton(HeatingPowerValue);
    OCR1A = HeatingOutVoltage;	// Заносим величину сигнала в регистр счетчика
    break;
  default:
    return 1;
  }
  return 0;
}

int8_t  ExecuteCommand8 (const char      command
                        ,const uint32_t  argument)
{
  return 1;
}

void  SetupHardware (void)
{
  /* Disable watchdog if enabled by bootloader/fuses */
  MCUSR &= ~(1 << WDRF);
  wdt_disable();

  /* Disable clock division */
  clock_prescale_set(clock_div_1);

  // Настройка ADC --------------------------------
  ADCSRA  = (1<<ADEN);		//включаем ADC. Важно - это первое, что надо сделать
//  ADMUX = 0; // External AREF
  ADMUX = (0b11<<REFS0);    //Internal 2.56V Voltage Reference with external capacitor at AREF pin
                            //Используется несимметричный вход ADC0. Усиление = 1.
  ADCSRA |= ADC_CLOCK_PRESCALER128;	//задаем тактовую частоту ADC
  _delay_ms(1);
  ADCSRA |= (1<<ADSC);		//запускаем пустое первое преобразование ADC
  while (bit_is_set(ADCSRA, ADSC)) {}	//ждем окончания преобразования
  ADCSRA |= (1<<ADSC);		//запускаем на всякий случай второе преобразование ADC
  while (bit_is_set(ADCSRA, ADSC)) {}	//ждем окончания преобразования
  ADCSRA |= (1<<ADSC);		//запускаем преобразование ADC

  // Установки таймеров --------------------------------

  //  Инициализация таймера микроконтроллера TIMER0 - период регулирования
  TCCR0A = (0 << COM0A1) |	//  Инициализация таймера микроконтроллера TIMER0
           (0 << COM0A0) |
           (0 << COM0B1) |
           (0 << COM0B0) |
           (1 << WGM01) |	// Режим сброса при совпадении OCR0A
           (0 << WGM00);

  TCCR0B = (0 << FOC0A) |
           (0 << FOC0B) |
           (0 << WGM02) |
           (0 << CS12) |    // Выбор тактовой частоты
           (1 << CS11) |    // Данными битами установлено деление частоты на 64
           (1 << CS10);     // (время на такт 4 мкс)

  TIMSK0 &= ~(1 << OCIE0A);	// Запрещаем прерывания по таймеру TIMER0

  //  Инициализация таймера микроконтроллера TIMER1 - формирование ШИМ
  TCCR1A = (1 << COM1A1) |  // Сброс OC1 при совпадении
           (0 << COM1A0) |  // (PWM без инверсии)
           (0 << COM1B1) |
           (0 << COM1B0) |
           (0 << COM1C1) |
           (0 << COM1C0) |
           (1 << WGM11) |   // Fast PWM
           (0 << WGM10);

  TCCR1B = (0 << ICNC1) |
           (0 << ICES1) |
           (1 << WGM13) |   // Верхний предел счета
           (1 << WGM12) |   // задается регистром ICR1
           (0 << CS12) |    // Выбор тактовой частоты
           (0 << CS11) |    // Установлена частота без деления 16 МГц
           (1 << CS10);     // Частота PWM 244 Гц при диапазоне 16 бит

  ICR1 = 0xFFFF;	        // Заносим верхний предел счета
  TIMSK1 &= ~(1 << OCIE1A);	// Запрещаем прерывания по таймеру TIMER1

  // -----------------------------------------------------------------------------
  // Настройка выхода
  DDRB |= (1<<PB5);	  // бит 5 порта B - выход сигнала таймера OC1A
  //------------------------------------------------------------------------------
}


void  BeforeLoop (void)
{
  SettingValue = 512;
  _K = 3.0;
  RoomTCode = 444;
  _ExtraH = 0.1;

  PeriodInTicks = 100;                       // период в миллисекундах
  MNumOfPeriods = 2000/PeriodInTicks;       // Передавать измерения раз в 2 с
  PeriodInTicks = PeriodInTicks*250;         // Число тиков (по 4 мкс) на период
  NumOfCycles  = (PeriodInTicks / 0xFF) + 1; // Число циклов длительностью не более FF тактов
  CycleInTicks = PeriodInTicks / NumOfCycles -1;
  CycleNo = 0;
  MPeriodNo = 0;
  MeasuringFlag = false;
  RegulatingFlag = false;
  Regulate = false;
  TCNT0 = 0x00;			    // Обнуляем счетчик
  OCR0A = CycleInTicks;	    // Заносим количество тактов в регистр счетчика
  TIFR0 |= (1 << OCF0A);    // Сбрасываем прерывания таймера TIMER0
  TIMSK0 &= ~(1 << OCIE0A);	// Запрещаем прерывания по таймеру TIMER0
  HeatingPowerValue = 0;
  HeatingOutVoltage = 0;
  OCR1A = HeatingOutVoltage;	// Заносим величину сигнала в регистр счетчика
}


void  InLoop (void)
{
  if (bit_is_clear(ADCSRA, ADSC)) {
     ADCdata  = ADCL;					//ВАЖНО: сначала считываем младший байт
     ADCdata += ((uint16_t)ADCH)<<8;	//затем считываем старший байт
     ADCSRA |= (1<<ADSC);			    //запускаем новое преобразование ADC
     Dacc = Dacc + ADCdata - Dout;      //усредняем по 512-м отсчетам
     Dout = Dacc >> 9;
     CurrentValue = Dacc >> 5; // для повышения точности данные АЦП
                               // умножены на 16 (4 бита)
  }
  if (MeasuringFlag) {
    MeasuringFlag = false;
    SendNumberToUSB(CurrentValue);
    Send_EOL_ToUSB ();
  }
  if (Regulate && RegulatingFlag) {
    RegulatingFlag = false;
    double _SV, _CV;
    _SV = SettingValue;
    _SV = _SV/16;
    _CV = CurrentValue;
    _CV = _CV/16;
//    result = _K*(_SV - _CV)*SCALEOUTPUT;
    result = (_K*(_SV - _CV) + _ExtraH*(_SV - RoomTCode))*SCALEOUTPUT;
    if(result > MAX_LONG){
      result = MAX_LONG;
    }
    else if(result < 0){
      result = 0;
    }
    HeatingPowerValue = lround(result);
    HeatingOutVoltage = sqrt_newton(HeatingPowerValue);
    OCR1A = HeatingOutVoltage;	// Заносим величину сигнала в регистр счетчика
  }
}

//-----------------------------------------------------------------------------
// Обработчик прерывания от таймера 0 (COMPA)
//-----------------------------------------------------------------------------
ISR(TIMER0_COMPA_vect){
  sei();                      // Дадим возможность обработать прерывания от USB...
  CycleNo++;
  if (CycleNo == NumOfCycles){
    RegulatingFlag = true;
    MPeriodNo++;
    if (MPeriodNo == MNumOfPeriods){
      MeasuringFlag = true;
      MPeriodNo = 0;
    }
    CycleNo = 0;
  }
}

