////////////////////////////////////////////////////////////////////////////////
/*! @file pid.c  Firmware for PID regulator.
* \brief PID implementation for AVR.
- PID controller implementation. Set up by giving P/I/D terms to Init_PID()
- Uses a structure PID_DATA to store internal values.
- Uses the following formula U = K * ( Err + (1/Ti)*Int + Td*dErr )
- Part of the MSDN firmware.
- Part of the Research Assistant project.

- Copyright(C) 2017, Dmitry K. Nelson, St.Petersburg, Russia.
                     D.Nelson@mail.ioffe.ru
*///////////////////////////////////////////////////////////////////////////////

#include "pid.h"
#include "stdint.h"

/*! \brief Initialisation of PID controller parameters.
 *
 *  Initialise the variables used by the PID algorithm.
 *
 *  \param k_factor  Proportional term.
 *  \param i_time  Integration time in regulating periods.
 *  \param d_time  Derivative time in regulating periods.
 *  \param pid  Struct with PID status.
 */
void pid_Init(double k_factor, int16_t i_time, int16_t d_time, struct PID_DATA *pid)
// Set up PID controller parameters
{
  // Start values for PID controller
  pid->sumError = 0;
  pid->lastError = 0;
  // Tuning constants for PID loop
  pid->K_Factor = k_factor;
  if (i_time==0) {
    pid->I_Factor = 0;
  } else {
    double var = i_time;
    pid->I_Factor = 1/var;
  }
  pid->D_Factor = d_time;
}


/*! \brief PID control algorithm.
 *
 *  Calculates output from setpoint, process value and PID status.
 *
 *  \param setPoint  Desired value.
 *  \param processValue  Measured value.
 *  \param pid_st  PID status struct.
 */
double pid_Controller(double setPoint, double processValue, struct PID_DATA *pid_st)
{
  double error, i_term, d_term, total;

  error = setPoint - processValue;

  // Calculate Iterm
  pid_st->sumError = pid_st->sumError + error;
  i_term = pid_st->I_Factor * pid_st->sumError;

  // Calculate Dterm
  d_term = pid_st->D_Factor * (error - pid_st->lastError);

  pid_st->lastError = error;

  total = pid_st->K_Factor*(error + i_term + d_term);
  if(total > MAX_OUTPUT){
    total = MAX_OUTPUT;
  }
  else if(total < 0){
    total = 0;
  }
  return(total);
}

/*! \brief Resets the integrator.
 *
 *  Calling this function will reset the integrator in the PID regulator.
 */
void pid_Reset_Integrator(pidData_t *pid_st)
{
  pid_st->sumError = 0;
}
