////////////////////////////////////////////////////////////////////////////////
/*! @file ManualWLDriver.cpp Driver for the spectrometer with manual wavelength setting.
- Part of Nelson - Research Assistant Plugin package.
- Uses  RAGUI - Research Assistant Graphical User Interface.
- Uses  QtGui v.5.0  -  http://qt.nokia.com/
- Uses  RANet - Research Assistant Net Library (based on ANSI C++).
- Copyright(C) 2010-2013, Viktor E. Bursian, St.Petersburg, Russia.
                          Viktor.Bursian@mail.ioffe.ru
- Copyright(C) 2020, Dmitry K. Nelson, St.Petersburg, Russia.
                     D.Nelson@mail.ioffe.ru
*///////////////////////////////////////////////////////////////////////////////
#include "ManualWLDriver.h"
#include "attr.h"
#include "StringsAndQStrings.h"
#include <QDebug>
namespace RA {
//------------------------------------------------------------------------------

DEFINE_CLASS_TAG(sManualWLDriver)

//--------------------------------------------------------------- ManualWLDriver ---

void  sManualWLDriver::Preconfigure (sNodePtr  DeviceCfg)
{
  sPtr<sFolderNode>              DeviceSettings;
  DeviceCfg |= "Settings";
  DeviceSettings = DeviceCfg >> "Settings";
  if( DeviceSettings.IsCorrect() ){
    DeviceSettings |= attr("Range",sPhysRange(10,1000,"nm"));
    DeviceSettings |= attr("InitValue",sPhysValue(500.0,0.01,"nm"));
  };
  sEffector::Preconfigure(DeviceCfg);
}

sManualWLDriver::sManualWLDriver(sString   DevicetId
                        ,sNodePtr  DeviceCfg)
    :sEffector(DevicetId,DeviceCfg)
    ,InitValuePar(sPhysValue(500.0,0.01,"nm"))
    ,CurrentPosition(0.0,0.01,"nm")
{
}

sManualWLDriver::~sManualWLDriver()
{
  if( DeviceSettings.IsCorrect() ){
    DeviceSettings
        << attr("InitValue",InitValuePar);
  }
}

void  sManualWLDriver::PostConstructor ()
{
  sEffector::PostConstructor();

  try{ //get config parameters
    get_from<sPhysValueNode>(DeviceSettings >> "InitValue" ,InitValuePar);
  }catch(...){
    SetInvalid(tr("Something wrong in configuration."));
  }

  //evolution of the inherited state machine (behaviour)
  {
    //evolution of the State_Warming
    {
      State_W_Initiating = new sState(State_Warming(),"State_W_Initiating");
      State_W_SendingWL = new sState(State_Warming(),"State_W_SendingWL");
      State_W_Initiating
          ->SetAsInitial()
          ->OnEnterCall(SLOT(AskInitialPosition()))
          ->AddTransition(SIGNAL(Initiated ()),State_W_SendingWL)
          ;
      State_W_SendingWL
          ->OnEnterCall(SLOT(SayPosition()))
          ->AddTransition(State_Warming()->Fin())
          ;
    }
    //evolution of the State_Operating
    {
      State_O_M_Pausing
          = new sState(State_O_Measuring,ThePeriod,"State_O_M_Pausing");
      State_O_Initiating
          = new sState(State_Operating(),"State_O_Initiating");
      State_O_SettingTo
          = new sState(State_Operating(),"State_O_SettingTo");

      State_O_Initiating
          ->OnEnterCall(SLOT(AskInitialPosition()))
          ->AddTransition(State_O_M_Acquiring)
          ->OnExitCall(SLOT(SendingReady()))
          ;
      State_O_M_Acquiring
          ->OnEnterCall( SLOT(SayPosition()) )
          ->AddTransition(cTune ,State_O_Initiating)
          ->AddTransition(State_O_M_Pausing)
          ->AddTransition(cSetTo ,SLOT(SetDestination(sPhysValue))
                         ,State_O_SettingTo)
          ->AddTransition(SIGNAL(ErrorMessage(QString))
                         ,State_Cooling())
          ;
      State_O_M_Pausing
          ->AddTransition(cTune ,State_O_Initiating)
          ->AddTransition(State_O_M_Pausing,SIGNAL(Timeout())
                         ,State_O_M_Cycling)
          ->AddTransition(cSetTo ,SLOT(SetDestination(sPhysValue))
                         ,State_O_SettingTo)
          ->AddTransition(SIGNAL(ErrorMessage(QString))
                         ,State_Cooling())
          ;
      State_O_SettingTo
          ->OnFinishCall(SLOT(SayDone()))
          ->AddTransition(State_O_M_Acquiring)
          ;
    }
  }
}

void sManualWLDriver::AskInitialPosition()
{
  sEffectorInitDialog         D(InitValuePar);
  while( ! D.exec() );
  CurrentPosition = D.InitValueField->Value();
  emit Initiated();
}

void sManualWLDriver::SayPosition()
{
  emit NewData(CurrentPosition);
  InitValuePar = CurrentPosition;
}

void sManualWLDriver::SetDestination(sPhysValue  destination)
{
  sEffectorInitDialog         D(destination);
//  D.slblWL->setText(tr("Set manually the desired position and enter the value"));
  while( ! D.exec() );
  CurrentPosition = D.InitValueField->Value();
}

void sManualWLDriver::SayDone()
{
  emit Done(CurrentPosition);
}

void sManualWLDriver::SendingReady()
{
  emit TuneFinished();
}

//------------------------------------------------------------------------------
} //namespace RA




