////////////////////////////////////////////////////////////////////////////////
/*! @file PeriodicXYRecorder.cpp   Самописец.
- Регистрирует периодический сигнал Y в зависимости от поступающего значения X.
- Part of Nelson - Research Assistant Plugin package.
- Uses  RAGUI - Research Assistant Graphical User Interface.
- Uses  QtGui v.5.0  -  http://qt.nokia.com/
- Uses  RANet - Research Assistant Net Library (based on ANSI C++).
- Copyright(C) 2010-2013, Viktor E. Bursian, St.Petersburg, Russia.
                          Viktor.Bursian@mail.ioffe.ru
- Copyright(C) 2014, Dmitry K. Nelson, St.Petersburg, Russia.
                     D.Nelson@mail.ioffe.ru
*///////////////////////////////////////////////////////////////////////////////
#include "PeriodicXYRecorder.h"
#include "attr.h"
#include "StringsAndQStrings.h"
#include "Trace.h"
#include "Log.h"
#include <QMessageBox>
namespace RA {
//------------------------------------------------------------------------------

DEFINE_CLASS_TAG(sPeriodicXYRecorder)

//--------------------------------------------------- sPeriodicXYRecorder ---

void  sPeriodicXYRecorder::Preconfigure (sNodePtr  DeviceCfg)
{
//  sPtr<sFolderNode>              DeviceSettings;
  sPtr<sFolderNode>              DeviceLinks;
  DeviceCfg |= "Links";
  DeviceLinks = DeviceCfg >> "Links";
  if( DeviceLinks.IsCorrect() ){
    DeviceLinks |= "XGauge";
    DeviceLinks >> "XGauge" |= attr("Class","sGauge");
  }
  sPeriodicGaugeRecorder::Preconfigure(DeviceCfg);
}


sPeriodicXYRecorder::~sPeriodicXYRecorder()
{
  XGauge.SetToNULL();
}


sPeriodicXYRecorder::sPeriodicXYRecorder (sString    DeviceId
                                     ,sNodePtr   DeviceCfg
                                     ,sNodePtr   WidgetCfg
                                     ,QWidget *  parent   )
    :sPeriodicGaugeRecorder(DeviceId,DeviceCfg,WidgetCfg,parent)
    ,XGauge(this)
    ,NumOfPoints(0)
    ,FirstPoint(true)
{
  //evolution of the inherited state machine
  {
    //evolution of the State_W_Parallel
    {
      State_W_XGauge = new sState(State_W_Parallel,"State_W_XGauge");
      //State_W_XGauge structure and behaviour
      {
        State_W_XGauge_Warming = new sState(State_W_XGauge,"State_W_XGauge_Warming");
        State_W_XGauge_Warming
            ->SetAsInitial()
            ->OnEnterCall(SLOT(LinkToXGauge()))
            ->AddTransition(SIGNAL(XGaugeWarmed()),State_W_XGauge->Fin())
            ->AddTransition(SIGNAL(ErrorMessage()),State_Cooling());
      }
    }
    connect(this ,SIGNAL(XGaugeError(QString))
           ,this ,SLOT(ReportXGaugeError(QString)) );
  }
}

void sPeriodicXYRecorder::ReportXGaugeError (QString  error_text)
{
  QMessageBox::critical(this ,tr("XGauge error")
                       ,tr("Device '%1' says:\n\n%2"
                           "\n\nTry to turn on again.")
                       .arg(ToQString(XGauge->DeviceIdentifier()))
                       .arg(error_text)     );
}


void sPeriodicXYRecorder::LinkToXGauge ()
{
  try{
    XGauge.SetTo(DeviceLinks,"XGauge");
    Connect(XGauge ,SIGNAL(Warmed())       ,SIGNAL(XGaugeWarmed()) );
    Connect(XGauge ,SIGNAL(ErrorMessage(QString)),SIGNAL(XGaugeError(QString)) );
    Connect(XGauge ,SIGNAL(NewData(sPhysValue)),SLOT(StoreXValue(sPhysValue)) );
  }catch( rxDeviceLaunch E ){
    emit ErrorMessage( tr("Error while linking to a submissive device.\n"
                          "%1\nCheck configuration.")
                         .arg(ToQString(E.MsgText()))   );
  }
}

void sPeriodicXYRecorder::OnEnterToCoolingState ()
{
  Disconnect(XGauge ,0 ,0);
  XGauge.SetToNULL();
  sPeriodicGaugeRecorder::OnEnterToCoolingState();
}

void  sPeriodicXYRecorder::StoreXValue (sPhysValue _Xdata)
{
  XValue =  _Xdata;
}

void  sPeriodicXYRecorder::CreateNewDependence ()
{
  try{
    sPhysValue NP(_Unitsless_);
    NP = ScanRange.Size() / ScanStep;
    NumOfPoints = Round(NP.ToReal());
  }catch(rxRAlgebra){
    QMessageBox::critical(this
        ,tr("%1 failure.").arg(ToQString(DeviceIdentifier()))
        ,tr("Wrong units in the configuration settings.")
        );
    SetInvalid();
    return;
  };
  Dependence = sNodePtr(new sDataListRecord);
  DataListRecord = Dependence;
  FirstPoint = true;
}

void  sPeriodicXYRecorder::AddNewData (sPhysValue  Data)
{
  if( ! DataListRecord.IsCorrect() ){
    return;
  };
  bool NewXUnits = false;
  bool NewYUnits = false;
  try{
    (XValue/ Multiplier.X).ToReal();
  }catch(rxRAlgebra){
    if (FirstPoint) {
      NewXUnits = true;
    } else {
      QMessageBox::critical(this
          ,tr("%1 failure.").arg(ToQString(DeviceIdentifier()))
          ,tr("Wrong units of X-Data.")
          );
      return;
    }
  };
  try{
    (Data / Multiplier.Y).ToReal();
  }catch(rxRAlgebra){
    if (FirstPoint) {
      NewYUnits = true;
    } else {
      QMessageBox::critical(this
          ,tr("%1 failure.").arg(ToQString(DeviceIdentifier()))
          ,tr("Wrong units of Y-Data.")
          );
      return;
    }
  };
  if( Multiplier.X.IsNaN() || NewXUnits ){
    Multiplier.X = sPhysValue(XValue.Units());
    sPhysRange                  RX(-Abs(XValue),3*Abs(XValue));
    SetChartScales(sChartScales(sChartScale(ChartScales().X.Name,RX)
                               ,ChartScales().Y
                               ));
  };
  if( Multiplier.Y.IsNaN() || NewYUnits ){
    Multiplier.Y = sPhysValue(Data.Units());
    sPhysRange                  RY(-Abs(Data),3*Abs(Data));
    SetChartScales(sChartScales(ChartScales().X
                               ,sChartScale(ChartScales().Y.Name,RY)
                               ));
  };
  if( DataListRecord->Multiplier().X.IsNaN() ){
    DataListRecord->SetMultiplier(sPhysPair(Multiplier.X,Multiplier.Y));
  };
  if( DataListRecord->Multiplier().Y.IsNaN() ){
    DataListRecord->SetMultiplier(sPhysPair(Multiplier.X,Multiplier.Y));
  };
  sPhysValue                  X(_Unitsless_);
  sPhysValue                  Y(_Unitsless_);
  X = XValue/ Multiplier.X;
  Y = Data / Multiplier.Y;
  DataListRecord->AddNewPoint(X.ToReal(),Y.ToReal());
  FirstPoint = false;
  if( CurrentPointNo < (NumOfPoints - 1) ){
    CurrentPointNo++;
  }else{
    emit ScanningHasCompleted();
  };
}


//------------------------------------------------------------------------------
} //namespace RA
