////////////////////////////////////////////////////////////////////////////////
/*! @file Stepper428.cpp Driver for Stepping motor microcontroller board.
- Part of Nelson - Research Assistant Plugin package.
- Uses  RAGUI - Research Assistant Graphical User Interface.
- Uses  QtGui v.5.0  -  http://qt.nokia.com/
- Uses  RANet - Research Assistant Net Library (based on ANSI C++).
- Copyright(C) 2010-2013, Viktor E. Bursian, St.Petersburg, Russia.
                          Viktor.Bursian@mail.ioffe.ru
- Copyright(C) 2014, Dmitry K. Nelson, St.Petersburg, Russia.
                     D.Nelson@mail.ioffe.ru
*///////////////////////////////////////////////////////////////////////////////
#include "Stepper428.h"
#include "attr.h"
#include "StringsAndQStrings.h"
#include <QDebug>
namespace RA {
//------------------------------------------------------------------------------

DEFINE_CLASS_TAG(sStepper428)

//--------------------------------------------------------------- Stepper428 ---

void  sStepper428::Preconfigure (sNodePtr  DeviceCfg)
{
  sPtr<sFolderNode>              DeviceSettings;
  sPtr<sFolderNode>              DeviceLinks;
  DeviceCfg |= "Links";
  DeviceLinks = DeviceCfg >> "Links";
  if( DeviceLinks.IsCorrect() ){
    DeviceLinks |= "StepperCom";
    DeviceLinks >> "StepperCom" |= attr("Class","sTelecom");
    DeviceLinks |= "WLIndexCom";
    DeviceLinks >> "WLIndexCom" |= attr("Class","sTelecom");
  };
  DeviceCfg |= "Settings";
  DeviceSettings = DeviceCfg >> "Settings";
  if( DeviceSettings.IsCorrect() ){
    DeviceSettings |= attr("StepperName","@MSDN2044");
    DeviceSettings |= attr("WLIndexName","@MSDN2045");
    DeviceSettings |= attr("Range",sPhysRange(200,1000,"nm"));
    DeviceSettings |= attr("MinimalTicksPerStep",7);
    DeviceSettings |= attr("AccelAdditionCoef",9);
    DeviceSettings |= attr("RampCoef",1031);
    DeviceSettings |= attr("StoppingZone",400);
    DeviceSettings |= attr("Overstep",400);
    DeviceSettings |= attr("WLCoefA",1.26359E-5);
    DeviceSettings |= attr("WLCoefB",-0.02761);
    DeviceSettings |= attr("StepperTimePerStepQuantum",sPhysValue(32e-6,"s"));
  };
  sStepper::Preconfigure(DeviceCfg);
}


sStepper428::sStepper428(sString   DevicetId
                        ,sNodePtr  DeviceCfg)
    :sStepper(DevicetId,DeviceCfg)
    ,StepperCom(this)
    ,WLIndexCom(this)
    ,CoefA(_Unitsless_)
    ,CoefB(_Unitsless_)
    ,TimePerStepQuantum("s")
    ,Destination(0)
{
}


sStepper428::~sStepper428()
{
    StepperCom.SetToNULL();
    WLIndexCom.SetToNULL();
}


sPhysValue  sStepper428::Index2WL (double  index) const
{
  return  sPhysValue((1.0/6.0)*1.0E4*sin(CoefA.ToReal()*index + CoefB.ToReal())
                    ,0.01
                    ,"nm");
/*    return  sPhysValue(index/100
                    ,0.01
                    ,"nm"); */
}


double  sStepper428::WL2Index (sPhysValue  wavelength) const
{
  return (  asin( (wavelength/sPhysValue(1,"nm")).ToReal() * 6.0e-4 )
          - CoefB.ToReal()
         )
         / CoefA.ToReal();
/*  return ( (wavelength/sPhysValue(1,"nm")).ToReal() )*100; */
}

double  sStepper428::Steps2Index (int steps) const
{
    return 1024.0*steps/8000;
}

int  sStepper428::Index2Steps (int index) const
{
    return 8000*index/1024;
}

int sStepper428::Index2Steps(double index) const
{
    return Round(8000*index/1024);
}

sPhysValue  sStepper428::Position (int  steps) const
{
    return  Index2WL(Steps2Index(steps));
}


int  sStepper428::Steps (sPhysValue  position) const
{
    return Index2Steps(WL2Index(position));
}

sPhysValue  sStepper428::Wavelength (int  index) const
{
    return  Index2WL(index);
}


int  sStepper428::Index (sPhysValue  wavelength) const
{
    return Round(WL2Index(wavelength));
}


void sStepper428::SetWantedPeriod ()
{
  int                         P;
  P = Round( (WantedPeriod/sPhysValue(1,"ms")).ToReal() );
  ThePeriod = P * sPhysValue(1,"ms");
  State_O_M_Acquiring->SetTimeout(P);
}


void  sStepper428::PostConstructor ()
{
  sStepper::PostConstructor();

  try{ //get config parameters
    get_from<sTextNode>(DeviceSettings >> "StepperName" ,StepperName);
    get_from<sTextNode>(DeviceSettings >> "WLIndexName" ,WLIndexName);
    get_from<sIntegerNode>(DeviceSettings >> "MinimalTicksPerStep"
                                             ,MinimalTicksPerStep);
    get_from<sIntegerNode>(DeviceSettings >> "AccelAdditionCoef"
                                             ,AccelAddition);
    get_from<sIntegerNode>(DeviceSettings >> "RampCoef" ,RampCoef);
    get_from<sIntegerNode>(DeviceSettings >> "StoppingZone" ,StoppingZone);
    get_from<sIntegerNode>(DeviceSettings >> "Overstep" ,Overstep);
    get_from<sPhysValueNode>(DeviceSettings >> "WLCoefA" ,CoefA);
    get_from<sPhysValueNode>(DeviceSettings >> "WLCoefB" ,CoefB);
    get_from<sPhysValueNode>(DeviceSettings >> "StepperTimePerStepQuantum" ,TimePerStepQuantum);
  }catch(...){
    SetInvalid(tr("Something wrong in configuration."));
  }

  //evolution of the inherited state machine (behaviour)
  {
    //evolution of the State_Warming
    {
      State_W_Parallel = new sParallelStateGroup(State_Warming());
      State_W_Parallel 
          ->SetAsInitial()
          ->AddTransition(State_Warming()->Fin());
      //State_W_Parallel internal structure and behaviour
      {
        State_W_StepperCom = new sState(State_W_Parallel);
        State_W_WLIndexCom = new sState(State_W_Parallel);
        
        State_W_StepperCom
            ->OnFinishCall(SLOT(StepperControllerInitiation()));
        //State_W_StepperCom internal structure and behaviour
        {
          State_W_StepperCom_Linking = new sState(State_W_StepperCom);
          State_W_StepperCom_Checking = new sState(State_W_StepperCom,100);

          State_W_StepperCom_Linking
              ->SetAsInitial()
              ->OnEnterCall(SLOT(LinkToStepperComAndMakeItWarming()))
              ->AddTransition(SIGNAL(StepperComLinked())
                             ,State_W_StepperCom_Checking)
              ->AddTransition(SIGNAL(ErrorMessage()),State_Cooling());
          
          State_W_StepperCom_Checking
              ->OnEnterCall(SLOT(StepperControllerAskName()))
              ->AddTransition(State_W_StepperCom_Checking,SIGNAL(Timeout())
                             ,State_Cooling())
              ->AddTransition(SIGNAL(GotStepperMessage(sTelex))
                             ,State_W_StepperCom->Fin()
                             ,sTelex(StepperName));
          
        };
        //State_W_WLIndexCom structure and behaviour
        {
          State_W_WLIndexCom_Linking = new sState(State_W_WLIndexCom);
          State_W_WLIndexCom_Checking = new sState(State_W_WLIndexCom,100);
        
          State_W_WLIndexCom_Linking
              ->SetAsInitial()
              ->OnEnterCall(SLOT(LinkToWLIndexComAndMakeItWarming()))
              ->AddTransition(SIGNAL(WLIndexComLinked())
                             ,State_W_WLIndexCom_Checking)
              ->AddTransition(SIGNAL(ErrorMessage()),State_Cooling());
                           
          State_W_WLIndexCom_Checking
              ->OnEnterCall(SLOT(WavelengthCounterAskName()))
              ->AddTransition(State_W_WLIndexCom_Checking,SIGNAL(Timeout())
                             ,State_Cooling())
              ->AddTransition(SIGNAL(GotWLIndexMessage(sTelex))
                             ,State_W_WLIndexCom->Fin()
                             ,sTelex(WLIndexName));
        };
      };
    };
    //evolution of the State_Operating
    {
      State_O_Parallel = new sParallelStateGroup(State_Operating());
      State_O_Parallel 
          ->SetAsInitial();
      //State_O_Parallel internal structure and behaviour
      {
        State_O_Measuring->setParent(State_O_Parallel);
        State_O_Control = new sState(State_O_Parallel);
        
        //evolution of the inherited state State_O_Measuring
        {
          State_O_M_Pausing
              = new sState(State_O_Measuring,ThePeriod,"State_O_M_Pausing");

          State_O_M_Acquiring
              ->OnEnterCall( SLOT(AskWL()) )
              ->AddTransition<sTelex>(SIGNAL(GotWLIndexMessage(sTelex))
                                     ,SLOT(RecalculateWL(sTelex))
                                     ,State_O_M_Pausing)
              ;

          State_O_M_Pausing
              ->AddTransition(State_O_M_Pausing,SIGNAL(Timeout())
                             ,State_O_M_Cycling)
              ->AddTransition(SIGNAL(PushWLMeter()),State_O_M_Acquiring)
              ;
        };
        //State_O_Control internal structure and behaviour
        {
          State_OC_Idle = new sState(State_O_Control);
          State_OC_SettingTo = new sState(State_O_Control);
          State_OC_ScanningTo = new sState(State_O_Control);
          State_OC_Breaking = new sState(State_O_Control);
        
          State_OC_Idle
              ->SetAsInitial()
              ->AddTransition(cSetTo ,SLOT(SetDestination(sPhysValue))
                             ,State_OC_SettingTo)
              ->AddReaction(cSetTimePerStep
                             ,SLOT(SetScanTimePerStep(sPhysValue)) )
              ->AddTransition<int>(cMoveBy,SLOT(StartScanning(int))
                                  ,State_OC_ScanningTo);
      
          State_OC_SettingTo
              ->OnFinishCall(SLOT(SayDone()))
              ->AddTransition(State_OC_Idle)
              ->AddTransition(cBreak ,State_OC_Breaking);
          
          State_OC_ScanningTo
              ->OnFinishCall(SLOT(SayDone()))
              ->OnFinishCall(SIGNAL(PushWLMeter()))
              ->AddTransition(State_OC_Idle)
              ->AddTransition(cBreak ,State_OC_Breaking);
          
          State_OC_Breaking
              ->OnFinishCall(SLOT(SayDone()))
              ->AddTransition(State_OC_Idle);
      
          //State_OC_SettingTo internal structure and behaviour
          {
            State_OC_SettingTo_CheckPos = new sState(State_OC_SettingTo);
            State_OC_SettingTo_WaitingRoger = new sState(State_OC_SettingTo);
            State_OC_SettingTo_WaitingDone = new sState(State_OC_SettingTo);
            State_OC_SettingTo_MovingBackwardCheckPos =
                new sState(State_OC_SettingTo);
            State_OC_SettingTo_MovingBackwardWaitingRoger =
                new sState(State_OC_SettingTo);
            State_OC_SettingTo_MovingBackwardWaitingDone =
                new sState(State_OC_SettingTo);
        
            State_OC_SettingTo_CheckPos
                ->SetAsInitial()
                ->OnEnterCall(SIGNAL(PushWLMeter()))
                ->AddTransition(SIGNAL(DoneSetTo())
                               ,State_OC_SettingTo->Fin())
                ->AddTransition(SIGNAL(MoveForward())
                               ,State_OC_SettingTo_WaitingRoger)
                ->AddTransition(SIGNAL(MoveBackward())
                               ,State_OC_SettingTo_MovingBackwardCheckPos)
                ->AddReaction<int>(SIGNAL(TheIndex(int))
                                   ,SLOT(MakeDecision(int))  );
        
            State_OC_SettingTo_WaitingRoger
                ->OnEnterCall(SLOT(StartMoving()))
                ->AddTransition(SIGNAL(GotStepperMessage(sTelex))
                               ,State_OC_SettingTo_WaitingDone
                               ,sTelex("<"));
            
            State_OC_SettingTo_WaitingDone
                ->AddTransition(SIGNAL(GotStepperMessage(sTelex))
                               ,State_OC_SettingTo_CheckPos
                               ,sTelex("Done"));
        
            State_OC_SettingTo_MovingBackwardCheckPos
                ->OnEnterCall(SIGNAL(PushWLMeter()))
                ->AddTransition(SIGNAL(DoneSetTo())
                               ,State_OC_SettingTo_CheckPos)
                ->AddTransition(SIGNAL(MoveBackward())
                               ,State_OC_SettingTo_MovingBackwardWaitingRoger)
                ->AddReaction<int>(SIGNAL(TheIndex(int))
                                   ,SLOT(MakeDecisionBackward(int))  );
        
            State_OC_SettingTo_MovingBackwardWaitingRoger
                ->OnEnterCall(SLOT(StartMovingBackward()))
                ->AddTransition(SIGNAL(GotStepperMessage(sTelex))
                               ,State_OC_SettingTo_MovingBackwardWaitingDone
                               ,sTelex("<"));
        
            State_OC_SettingTo_MovingBackwardWaitingDone
                ->AddTransition(SIGNAL(GotStepperMessage(sTelex))
                               ,State_OC_SettingTo_MovingBackwardCheckPos
                               ,sTelex("Done"));
          };
          //State_OC_ScanningTo internal structure and behaviour
          {
            State_OC_ScanningTo_WaitingRoger = new sState(State_OC_ScanningTo);
            State_OC_ScanningTo_WaitingDone = new sState(State_OC_ScanningTo);

            State_OC_ScanningTo_WaitingRoger
                ->SetAsInitial()
                ->AddTransition(SIGNAL(GotStepperMessage(sTelex))
                               ,State_OC_ScanningTo_WaitingDone
                               ,sTelex("<"));
            State_OC_ScanningTo_WaitingDone
                ->AddTransition(SIGNAL(GotStepperMessage(sTelex))
                               ,State_OC_ScanningTo->Fin()
                               ,sTelex("Done"));
          }
          //content of the State_OC_Breaking
          {
            State_OC_Breaking_WaitingConfirm = new sState(State_OC_Breaking);
            State_OC_Breaking_WaitingConfirm
                ->SetAsInitial()
                ->OnEnterCall(SLOT(SendBreakCommand()))
                ->AddReaction<sTelex>(SIGNAL(GotStepperMessage(sTelex))
                                     ,SLOT(CheckIfBreaked(sTelex))  )
                ->AddTransition(SIGNAL(BreakConfirmed())
                               ,State_OC_Breaking->Fin());
          }
        };
      };
    };
    connect(this ,SIGNAL(StepperComError(QString))
           ,this ,SLOT(ReportStepperComError(QString)) );
    connect(this ,SIGNAL(WLIndexComError(QString))
           ,this ,SLOT(ReportWLIndexComError(QString)) );
  }
}


void sStepper428::LinkToStepperComAndMakeItWarming ()
{
  sString                     TeleComId;
  if( ! get_from<sTextNode>(DeviceLinks>>"StepperCom">>"DeviceId",TeleComId) ){
    emit ErrorMessage( tr("The following link is absent: 'StepperCom'.\n"
                          "Check configuration.") );
  }else{
    StepperCom.SetTo(TeleComId);
    if( ! StepperCom.IsCorrect() ){
      emit ErrorMessage( tr("The 'StepperCom' link is set to '%1', which "
                            "does not look correct. Check configuration.")
                            .arg(ToQString(TeleComId)) );
    }else{
      Connect(StepperCom ,SIGNAL(Warmed()) ,SIGNAL(StepperComLinked()) );
      Connect(StepperCom ,SIGNAL(ErrorMessage(QString))
             ,SIGNAL(StepperComError(QString)) );
      Connect(StepperCom ,SIGNAL(GotTelex(sTelex))
             ,SIGNAL(GotStepperMessage(sTelex)) );
    }
  }
}


void sStepper428::ReportStepperComError (QString  error_text)
{
  emit ErrorMessage( tr("Got error from stepper connected to '%1':\n\n%2")
                        .arg(ToQString(StepperCom->DeviceIdentifier()))
                        .arg(error_text)     );
}


void sStepper428::ReportWLIndexComError (QString  error_text)
{
  emit ErrorMessage( tr("Got error from WL-meter connected to '%1':\n\n%2")
                       .arg(ToQString(WLIndexCom->DeviceIdentifier()))
                       .arg(error_text)     );
}


void sStepper428::LinkToWLIndexComAndMakeItWarming ()
{
  sString                     TeleComId;
  if( ! get_from<sTextNode>(DeviceLinks>>"WLIndexCom">>"DeviceId",TeleComId) ){
    emit ErrorMessage( tr("The following link is absent: 'WLIndexCom'.\n"
                          "Check configuration.") );
  }else{
    WLIndexCom.SetTo(TeleComId);
    if( ! WLIndexCom.IsCorrect() ){
      emit ErrorMessage( tr("The 'WLIndexCom' link is set to '%1', which "
                            "does not look correct. Check configuration.")
                            .arg(ToQString(TeleComId)) );
    }else{
      Connect(WLIndexCom ,SIGNAL(Warmed()) ,SIGNAL(WLIndexComLinked()) );
      Connect(WLIndexCom ,SIGNAL(ErrorMessage(QString))
             ,SIGNAL(WLIndexComError(QString)) );
      Connect(WLIndexCom ,SIGNAL(GotTelex(sTelex))
             ,SIGNAL(GotWLIndexMessage(sTelex)) );
    }
  }
}


void sStepper428::StepperControllerAskName ()
{
    SendDirective<sTelex>(StepperCom ,sTeleCom::cSendTelex ,sTelex("?"));
}


void sStepper428::WavelengthCounterAskName ()
{
    SendDirective<sTelex>(WLIndexCom ,sTeleCom::cSendTelex ,sTelex("?"));
}


void sStepper428::StepperControllerInitiation ()
{
    QByteArray _comm;
    _comm.append(QByteArray::number(AccelAddition, 16)
                    .toUpper().rightJustified(4, '0'));
    _comm.prepend("A");
    SendDirective<sTelex>(StepperCom ,sTeleCom::cSendTelex ,sTelex(_comm));
    _comm.clear();
    _comm.append(QByteArray::number(RampCoef, 16)
               .toUpper().rightJustified(4, '0'));
    _comm.prepend("K");
    SendDirective<sTelex>(StepperCom ,sTeleCom::cSendTelex ,sTelex(_comm));
    _comm.clear();
    _comm.append(QByteArray::number(StoppingZone, 16)
                   .toUpper().rightJustified(4, '0'));
    _comm.prepend("Z");
    SendDirective<sTelex>(StepperCom ,sTeleCom::cSendTelex ,sTelex(_comm));
}


void sStepper428::AskWL()
{
    SendDirective<sTelex>(WLIndexCom ,sTeleCom::cSendTelex ,sTelex("W"));
}


void sStepper428::RecalculateWL(sTelex theIndex)
{
    bool OK;
    int CurrentIndex = theIndex.toLong(&OK,16);
    if (OK){
        emit TheIndex(CurrentIndex);
        emit NewData(Wavelength(CurrentIndex));
    } else
        emit ErrorMessage( tr("Illegal format of WLIndex message"));
}


void sStepper428::SetDestination(sPhysValue  destination)
{
  Destination = Index(destination);
}

void sStepper428::MakeDecision(int current_index)
{
  CurrentIndex = current_index;
  if( CurrentIndex < Destination-1 ){
    emit MoveForward();
  }else if( CurrentIndex > Destination+1 ){
    emit MoveBackward();
  }else{
    emit DoneSetTo();
    qDebug() << "Set to index " << CurrentIndex;
  }
}

void sStepper428::SayDone()
{
  emit Done(Wavelength(CurrentIndex));
}

void sStepper428::StartMoving()
{
  //Go forward on (Destination-CurrentIndex) with Ramp
    QByteArray _comm;
    _comm.append(QByteArray::number(MinimalTicksPerStep, 16)
                          .toUpper().rightJustified(4, '0'));
    _comm.prepend("T");
    SendDirective<sTelex>(StepperCom ,sTeleCom::cSendTelex ,sTelex(_comm));
    _comm.clear();
    _comm.append(QByteArray::number(Index2Steps(Destination-CurrentIndex), 16)
                                            .toUpper().rightJustified(8, '0'));
    _comm.prepend("G");
    SendDirective<sTelex>(StepperCom ,sTeleCom::cSendTelex ,sTelex(_comm));
}


void sStepper428::MakeDecisionBackward (int current_index)
{
  CurrentIndex = current_index;
  if( CurrentIndex > (Destination-Overstep) ){
    emit MoveBackward();
  }else{
    emit DoneSetTo();
  }
}

void sStepper428::StartMovingBackward()
{
  //Go backward on (CurrentIndex-(Destination-Overstep)) with Ramp
    QByteArray _comm;
    _comm.append(QByteArray::number(MinimalTicksPerStep, 16)
                          .toUpper().rightJustified(4, '0'));
    _comm.prepend("T");
    SendDirective<sTelex>(StepperCom ,sTeleCom::cSendTelex ,sTelex(_comm));
    _comm.clear();
    _comm.append(QByteArray::number(Index2Steps(CurrentIndex-(Destination-Overstep)), 16)
                                                       .toUpper().rightJustified(8, '0'));
    _comm.prepend("R");
    SendDirective<sTelex>(StepperCom ,sTeleCom::cSendTelex ,sTelex(_comm));
}

void sStepper428::SetScanTimePerStep(sPhysValue time_per_step)
{
/* Число шагов сканирования и время на шаг вычисляются так:
  int NumStepsToGo = Steps(Scan.To) - Steps(Scan.From);
  sPhysValue TimePerStep = (TimePerPoint * NumOfPoints)/ NumStepsToGo;
*/
  int ticksPerStep = Round((time_per_step/TimePerStepQuantum).ToReal());
  if (ticksPerStep < MinimalTicksPerStep) ticksPerStep = MinimalTicksPerStep;
  if (ticksPerStep > 65500) ticksPerStep = 65500;
  QByteArray _comm;
  _comm.append(QByteArray::number(ticksPerStep, 16)
                   .toUpper().rightJustified(4, '0'));
  _comm.prepend("T");
  SendDirective<sTelex>(StepperCom ,sTeleCom::cSendTelex ,sTelex(_comm));
  sPhysValue time_per_stepSet = TimePerStepQuantum * ticksPerStep;
  emit TimePerStepSet (time_per_stepSet);
}

void sStepper428::StartScanning(int numSteps)
{
  QByteArray _comm;
  _comm.append(QByteArray::number(numSteps, 16)
             .toUpper().rightJustified(8, '0'));
  _comm.prepend("F");
  SendDirective<sTelex>(StepperCom ,sTeleCom::cSendTelex ,sTelex(_comm));
}

void sStepper428::SendBreakCommand()
{
  SendDirective<sTelex>(StepperCom ,sTeleCom::cSendTelex ,sTelex("X"));
}

void sStepper428::CheckIfBreaked(sTelex Msg)
{
  if (Msg.contains("N=")) emit BreakConfirmed();
/*  {
    Msg.remove(0,2);
    bool OK;
    int StepsLeft = Msg.toLong(&OK,16);
    emit StepsNotExecuted(StepsLeft);
    emit BreakConfirmed();
  }
*/
}

//------------------------------------------------------------------------------
} //namespace RA




