////////////////////////////////////////////////////////////////////////////////
/*! @file StepperMDR.cpp Driver for Stepping motor microcontroller board.
- Part of Nelson - Research Assistant Plugin package.
- Uses  RAGUI - Research Assistant Graphical User Interface.
- Uses  QtGui v.5.0  -  http://qt.nokia.com/
- Uses  RANet - Research Assistant Net Library (based on ANSI C++).
- Copyright(C) 2010-2013, Viktor E. Bursian, St.Petersburg, Russia.
                          Viktor.Bursian@mail.ioffe.ru
- Copyright(C) 2014, Dmitry K. Nelson, St.Petersburg, Russia.
                     D.Nelson@mail.ioffe.ru
*///////////////////////////////////////////////////////////////////////////////
#include "StepperMDR.h"
#include "attr.h"
#include "StringsAndQStrings.h"
#include <QDebug>
namespace RA {
//------------------------------------------------------------------------------

DEFINE_CLASS_TAG(sStepperMDR)

//--------------------------------------------------------------- StepperMDR ---

void  sStepperMDR::Preconfigure (sNodePtr  DeviceCfg)
{
  sPtr<sFolderNode>              DeviceSettings;
  sPtr<sFolderNode>              DeviceLinks;
  DeviceCfg |= "Links";
  DeviceLinks = DeviceCfg >> "Links";
  if( DeviceLinks.IsCorrect() ){
    DeviceLinks |= "StepperCom";
    DeviceLinks >> "StepperCom" |= attr("Class","sTelecom");
  };
  DeviceCfg |= "Settings";
  DeviceSettings = DeviceCfg >> "Settings";
  if( DeviceSettings.IsCorrect() ){
    DeviceSettings |= attr("StepperName","DriverMDR");
    DeviceSettings |= attr("Range",sPhysRange(200,1000,"nm"));
    DeviceSettings |= attr("MaxSpeedTicksPerStep",30);
    DeviceSettings |= attr("MinimalTicksPerStep",50);
    DeviceSettings |= attr("AccelAdditionCoef",4);
    DeviceSettings |= attr("RampCoef",1032);
    DeviceSettings |= attr("StoppingZone",100);
    DeviceSettings |= attr("StepperTimePerStepQuantum",sPhysValue(64e-6,"s"));
    DeviceSettings |= attr("GratingNo",1);
    DeviceSettings |= "WLCoefs";
    DeviceSettings >> "WLCoefs" |= attr(1 ,sPhysValue(  0.01  ,"nm"));
    DeviceSettings >> "WLCoefs" |= attr(2 ,sPhysValue(  0.01  ,"nm"));
    DeviceSettings >> "WLCoefs" |= attr(3 ,sPhysValue(  0.02  ,"nm"));
    DeviceSettings >> "WLCoefs" |= attr(4 ,sPhysValue(  0.04  ,"nm"));
    DeviceSettings |= "WLShifts";
    DeviceSettings >> "WLShifts" |= attr(1 ,sPhysValue(  0.0  ,"nm"));
    DeviceSettings >> "WLShifts" |= attr(2 ,sPhysValue(  0.0  ,"nm"));
    DeviceSettings >> "WLShifts" |= attr(3 ,sPhysValue(  0.0  ,"nm"));
    DeviceSettings >> "WLShifts" |= attr(4 ,sPhysValue(  0.0  ,"nm"));
  };
  sStepper::Preconfigure(DeviceCfg);
}


sStepperMDR::sStepperMDR(sString   DevicetId
                        ,sNodePtr  DeviceCfg)
    :sStepper(DevicetId,DeviceCfg)
    ,StepperCom(this)
    ,TimePerStepQuantum("s")
    ,GratingNo(1)
{
}


sStepperMDR::~sStepperMDR()
{
    StepperCom.SetToNULL();
}

sPhysValue  sStepperMDR::Position (int  steps) const
{
    return  WLCoef[GratingNo]*steps/3 - WLShift[GratingNo];
}


int  sStepperMDR::Steps (sPhysValue  position) const
{
    return Round( ((position+WLShift[GratingNo])
                   /WLCoef[GratingNo]).ToReal() ) * 3;
}

sPhysValue  sStepperMDR::Wavelength (int  index) const
{
    return  WLCoef[GratingNo]*index - WLShift[GratingNo];
}


int  sStepperMDR::Index (sPhysValue  wavelength) const
{
    return Round( ((wavelength+WLShift[GratingNo])
                   /WLCoef[GratingNo]).ToReal() );
}


void  sStepperMDR::SetWantedPeriod ()
{
  int                         P;
  P = Round( (WantedPeriod/sPhysValue(1,"ms")).ToReal() );
  ThePeriod = P * sPhysValue(1,"ms");
  State_O_M_Acquiring->SetTimeout(P);
}


void  sStepperMDR::PostConstructor ()
{
  sStepper::PostConstructor();

  WLCoef[0] = sPhysValue(  0.01  ,"nm");
  WLCoef[1] = sPhysValue(  0.01  ,"nm");
  WLCoef[2] = sPhysValue(  0.02  ,"nm");
  WLCoef[3] = sPhysValue(  0.04  ,"nm");
  WLShift[0] = sPhysValue( 0.0   ,"nm");
  WLShift[1] = sPhysValue( 0.0   ,"nm");
  WLShift[2] = sPhysValue( 0.0   ,"nm");
  WLShift[3] = sPhysValue( 0.0   ,"nm");

  try{ //get config parameters
    get_from<sTextNode>(DeviceSettings >> "StepperName" ,StepperName);
    get_from<sIntegerNode>(DeviceSettings >> "MaxSpeedTicksPerStep"
                                             ,MaxSpeedTicksPerStep);
    get_from<sIntegerNode>(DeviceSettings >> "MinimalTicksPerStep"
                                             ,MinimalTicksPerStep);
    get_from<sIntegerNode>(DeviceSettings >> "AccelAdditionCoef"
                                             ,AccelAddition);
    get_from<sIntegerNode>(DeviceSettings >> "RampCoef" ,RampCoef);
    get_from<sIntegerNode>(DeviceSettings >> "StoppingZone" ,StoppingZone);
    get_from<sPhysValueNode>(DeviceSettings >> "StepperTimePerStepQuantum"
                                               ,TimePerStepQuantum);
    get_from<sIntegerNode>(DeviceSettings >> "GratingNo" ,GratingNo);
    for( int i = 0 ; i < 4 ; ++i ){
      get_from<sPhysValueNode>(DeviceSettings >> "WLCoefs" >> (i+1) ,WLCoef[i]);
    };
    for( int i = 0 ; i < 4 ; ++i ){
      get_from<sPhysValueNode>(DeviceSettings >> "WLShifts" >> (i+1) ,WLShift[i]);
    };
    IndexRange = sPhysRange(TheRange.From()/WLCoef[0],TheRange.To()/WLCoef[0]);
  }catch(...){
    SetInvalid(tr("Something wrong in configuration."));
  }

  //evolution of the inherited state machine (behaviour)
  {
    //evolution of the State_Warming
    {
      State_W_Linking = new sState(State_Warming(),"State_W_Linking");
      State_W_Checking = new sState(State_Warming(),"State_W_Checking");
      State_W_Linking
          ->SetAsInitial()
          ->OnEnterCall( SLOT(LinkToStepperComAndMakeItWarming()) )
          ->AddTransition( SIGNAL(StepperComLinked()) ,State_W_Checking )
          ->AddTransition( SIGNAL(ErrorMessage()) ,State_Cooling() )
          ;
      State_W_Checking
          ->OnEnterCall(SLOT(StepperControllerAskName()))
          ->AddTransition(State_Warming()->Fin())
          ->OnFinishCall(SLOT(StepperControllerInitiation()));
          ;
      State_W_Ch_Waiting = new sState(State_W_Checking,"State_W_Ch_Waiting");
      State_W_Ch_Waiting
          ->SetAsInitial()
          ->AddTransition(State_W_Ch_Waiting,SIGNAL(Timeout())
                         ,State_Cooling())
          ->AddTransition(SIGNAL(GotStepperMessage(sTelex))
                         ,State_W_Checking->Fin()
                         ,sTelex(StepperName))
          ;
    }
    //evolution of the State_Operating
    {
      State_O_M_Pausing
          = new sState(State_O_Measuring,ThePeriod,"State_O_M_Pausing");
      State_O_SettingTo
          = new sState(State_Operating(),"State_O_SettingTo");
      State_O_ScanningTo
          = new sState(State_Operating(),"State_O_ScanningTo");
      State_O_Breaking
          = new sState(State_Operating(),"State_O_Breaking");

      State_O_SettingTo_WaitingRoger = new sState(State_O_SettingTo
                                             ,"State_O_SettingTo_WaitingRoger");
      State_O_SettingTo_WaitingDone = new sState(State_O_SettingTo
                                              ,"State_O_SettingTo_WaitingDone");
      State_O_SettingTo_WaitingWLData = new sState(State_O_SettingTo
                                            ,"State_O_SettingTo_WaitingWLData");

      State_O_ScanningTo_WaitingRoger = new sState(State_O_ScanningTo
                                            ,"State_O_ScanningTo_WaitingRoger");
      State_O_ScanningTo_WaitingDone = new sState(State_O_ScanningTo
                                             ,"State_O_ScanningTo_WaitingDone");
      State_O_ScanningTo_WaitingWLData = new sState(State_O_ScanningTo
                                           ,"State_O_ScanningTo_WaitingWLData");

      State_O_M_Acquiring
          ->OnEnterCall( SLOT(AskWL()) )
          ->AddTransition<sTelex>(SIGNAL(GotStepperMessage(sTelex))
                                 ,SLOT(RecalculateWL(sTelex))
                                 ,State_O_M_Pausing)
          ->AddReaction(cSetTimePerStep
                       ,SLOT(SetScanTimePerStep(sPhysValue)) )
          ->AddTransition(cSetTo ,SLOT(SetDestination(sPhysValue))
                         ,State_O_SettingTo_WaitingWLData)
          ->AddTransition(cMoveBy,SLOT(StartScanning(int))
                         ,State_O_ScanningTo_WaitingWLData)
          ->AddTransition(SIGNAL(ErrorMessage(QString))
                         ,State_Cooling())
          ;
      State_O_M_Pausing
          ->AddTransition(State_O_M_Pausing,SIGNAL(Timeout())
                         ,State_O_M_Cycling)
          ->AddTransition(cSetTo ,SLOT(SetDestination(sPhysValue))
                         ,State_O_SettingTo)
          ->AddReaction(cSetTimePerStep
                       ,SLOT(SetScanTimePerStep(sPhysValue)) )
          ->AddTransition<int>(cMoveBy,SLOT(StartScanning(int))
                              ,State_O_ScanningTo)
          ->AddTransition(SIGNAL(ErrorMessage(QString))
                         ,State_Cooling())
          ;
      State_O_SettingTo
          ->OnFinishCall(SLOT(SayDone()))
          ->AddTransition(State_O_M_Acquiring)
          ->AddTransition(cBreak ,State_O_Breaking)
          ;
      State_O_ScanningTo
          ->OnFinishCall(SLOT(SayDone()))
          ->AddTransition(State_O_M_Acquiring)
          ->AddTransition(cBreak ,State_O_Breaking)
          ;
      State_O_Breaking
          ->OnFinishCall(SLOT(SayDone()))
          ->AddTransition(State_O_M_Acquiring)
          ;

      //State_O_SettingTo internal structure and behaviour
      {

        State_O_SettingTo_WaitingRoger
            ->SetAsInitial()
            ->AddTransition(SIGNAL(GotStepperMessage(sTelex))
                           ,State_O_SettingTo_WaitingDone
                           ,sTelex("<"));

        State_O_SettingTo_WaitingDone
            ->AddTransition(SIGNAL(GotStepperMessage(sTelex))
                           ,State_O_SettingTo->Fin()
                           ,sTelex("Done"));

        State_O_SettingTo_WaitingWLData
            ->AddTransition<sTelex>(SIGNAL(GotStepperMessage(sTelex))
                                   ,SLOT(RecalculateWL(sTelex))
                                   ,State_O_SettingTo_WaitingRoger);

      }
      //State_O_ScanningTo internal structure and behaviour
      {

        State_O_ScanningTo_WaitingRoger
            ->SetAsInitial()
            ->AddTransition(SIGNAL(GotStepperMessage(sTelex))
                           ,State_O_ScanningTo_WaitingDone
                           ,sTelex("<"));

        State_O_ScanningTo_WaitingDone
            ->AddTransition(SIGNAL(GotStepperMessage(sTelex))
                           ,State_O_ScanningTo->Fin()
                           ,sTelex("Done"));

        State_O_ScanningTo_WaitingWLData
            ->AddTransition<sTelex>(SIGNAL(GotStepperMessage(sTelex))
                                   ,SLOT(RecalculateWL(sTelex))
                                   ,State_O_ScanningTo_WaitingRoger);
      }
      //content of the State_O_Breaking
      {
        State_O_Breaking_WaitingConfirm = new sState(State_O_Breaking
                                                   ,"State_O_Breaking_WaitingConfirm");
        State_O_Breaking_WaitingConfirm
            ->SetAsInitial()
            ->OnEnterCall(SLOT(SendBreakCommand()))
            ->AddTransition(SIGNAL(GotStepperMessage(sTelex))
                           ,State_O_Breaking->Fin()
                           ,sTelex("Done"));
      }
    }
    connect(this ,SIGNAL(StepperComError(QString))
           ,this ,SLOT(ReportStepperComError(QString)) );
  }
}


void sStepperMDR::LinkToStepperComAndMakeItWarming ()
{
  sString                     TeleComId;
  if( ! get_from<sTextNode>(DeviceLinks>>"StepperCom">>"DeviceId",TeleComId) ){
    emit ErrorMessage( tr("The following link is absent: 'StepperCom'.\n"
                          "Check configuration.") );
  }else{
    StepperCom.SetTo(TeleComId);
    if( ! StepperCom.IsCorrect() ){
      emit ErrorMessage( tr("The 'StepperCom' link is set to '%1', which "
                            "does not look correct. Check configuration.")
                           .arg(ToQString(TeleComId)) );
    }else{
      Connect(StepperCom ,SIGNAL(Warmed()) ,SIGNAL(StepperComLinked()) );
      Connect(StepperCom ,SIGNAL(ErrorMessage(QString))
             ,SIGNAL(StepperComError(QString)) );
      Connect(StepperCom ,SIGNAL(GotTelex(sTelex))
             ,SIGNAL(GotStepperMessage(sTelex)) );
    };
  };
}

void sStepperMDR::ReportStepperComError (QString  error_text)
{
  emit ErrorMessage( tr("StepperMDR '%1' says:\n\n"
                        "Got error from stepper connected to '%2':\n\n%3")
                       .arg(ToQString(DeviceIdentifier()))
                       .arg(ToQString(StepperCom->DeviceIdentifier()))
                       .arg(error_text)     );
}

void sStepperMDR::StepperControllerAskName ()
{
    SendDirective<sTelex>(StepperCom ,sTeleCom::cSendTelex ,sTelex("?"));
}

void sStepperMDR::StepperControllerInitiation ()
{
  QByteArray _comm;
  _comm.append(QByteArray::number(MaxSpeedTicksPerStep, 16)
                         .toUpper().rightJustified(4, '0'));
  _comm.prepend("M");
  SendDirective<sTelex>(StepperCom ,sTeleCom::cSendTelex ,sTelex(_comm));
  _comm.clear();
  _comm.append(QByteArray::number(AccelAddition, 16)
                  .toUpper().rightJustified(4, '0'));
  _comm.prepend("A");
  SendDirective<sTelex>(StepperCom ,sTeleCom::cSendTelex ,sTelex(_comm));
  _comm.clear();
  _comm.append(QByteArray::number(RampCoef, 16)
             .toUpper().rightJustified(4, '0'));
  _comm.prepend("K");
  SendDirective<sTelex>(StepperCom ,sTeleCom::cSendTelex ,sTelex(_comm));
  _comm.clear();
  _comm.append(QByteArray::number(StoppingZone, 16)
                 .toUpper().rightJustified(4, '0'));
  _comm.prepend("Z");
  SendDirective<sTelex>(StepperCom ,sTeleCom::cSendTelex ,sTelex(_comm));
}


void sStepperMDR::AskWL()
{
    SendDirective<sTelex>(StepperCom ,sTeleCom::cSendTelex ,sTelex("W"));
}


void sStepperMDR::RecalculateWL(sTelex theIndex)
{
//    qDebug()<< "Full Index Telex" << theIndex;
    int GrNoStore = GratingNo;
    char GrN = theIndex[0];
    if (GrN == '1') GratingNo = 0;
    else if (GrN == '2') GratingNo = 1;
    else if (GrN == '3') GratingNo = 2;
    else if (GrN == '4') GratingNo = 3;
    else GratingNo = 0;
    emit TheGratingNo(GrN);
    if (!(GrNoStore == GratingNo)) {
      TheRange = sPhysRange( WLCoef[GratingNo]*IndexRange.From()
                           , WLCoef[GratingNo]*IndexRange.To() );
      emit RangeChanged(TheRange);
    }
    bool OK;
    QByteArray Ind = QByteArray(theIndex);
    Ind.remove(0,1);
    int CurrentIndex = Ind.toLong(&OK,16);
    if (OK){
        emit NewData(Wavelength(CurrentIndex));
    } else
        emit ErrorMessage( tr("StepperMDR '%1' says:\n\n"
                          "Illegal format of WLIndex message")
                         .arg(ToQString(DeviceIdentifier()))     );
}


void sStepperMDR::SetDestination(sPhysValue  destination)
{
  Destination = Index(destination);
  QByteArray _comm;
  _comm.append(QByteArray::number((qlonglong)Destination, 16)
                           .toUpper().rightJustified(8, '0'));
  _comm.prepend("G");
  SendDirective<sTelex>(StepperCom ,sTeleCom::cSendTelex ,sTelex(_comm));
}

void sStepperMDR::SayDone()
{
  emit Done(Wavelength(Destination));
}

void sStepperMDR::SetScanTimePerStep(sPhysValue time_per_step)
{
/* Число шагов сканирования и время на шаг вычисляются так:
  int NumStepsToGo = Steps(Scan.To) - Steps(Scan.From);
  sPhysValue TimePerStep = (TimePerPoint * NumOfPoints)/ NumStepsToGo;
*/
  int ticksPerStep = Round((time_per_step/TimePerStepQuantum).ToReal());
  if (ticksPerStep < MinimalTicksPerStep) ticksPerStep = MinimalTicksPerStep;
  if (ticksPerStep > 65500) ticksPerStep = 65500;
  QByteArray _comm;
  _comm.append(QByteArray::number(ticksPerStep, 16)
                 .toUpper().rightJustified(4, '0'));
  _comm.prepend("T");
  SendDirective<sTelex>(StepperCom ,sTeleCom::cSendTelex ,sTelex(_comm));
  sPhysValue time_per_stepSet = TimePerStepQuantum * ticksPerStep;
  emit TimePerStepSet (time_per_stepSet);
}

void sStepperMDR::StartScanning(int numSteps)
{
  QByteArray _comm;
  _comm.append(QByteArray::number((qlonglong)numSteps, 16)
                        .toUpper().rightJustified(8, '0'));
  _comm.prepend("S");
  SendDirective<sTelex>(StepperCom ,sTeleCom::cSendTelex ,sTelex(_comm));
}

void sStepperMDR::SendBreakCommand()
{
  SendDirective<sTelex>(StepperCom ,sTeleCom::cSendTelex ,sTelex("X"));
}

//------------------------------------------------------------------------------
} //namespace RA




