////////////////////////////////////////////////////////////////////////////////
/*! @file StepperMDRNew.cpp Driver for Stepping motor microcontroller board.
                            Version without DUK, but with steps counting.
- Part of Nelson - Research Assistant Plugin package.
- Uses  RAGUI - Research Assistant Graphical User Interface.
- Uses  QtGui v.5.0  -  http://qt.nokia.com/
- Uses  RANet - Research Assistant Net Library (based on ANSI C++).
- Copyright(C) 2010-2013, Viktor E. Bursian, St.Petersburg, Russia.
                          Viktor.Bursian@mail.ioffe.ru
- Copyright(C) 2016, Dmitry K. Nelson, St.Petersburg, Russia.
                     D.Nelson@mail.ioffe.ru
*///////////////////////////////////////////////////////////////////////////////
#include "StepperMDRNew.h"
#include "attr.h"
#include "StringsAndQStrings.h"
#include <QDebug>
#include <QDialogButtonBox>
namespace RA {
//------------------------------------------------------------------------------

DEFINE_CLASS_TAG(sStepperMDRNew)


//------------------------------------------------------------ StepperMDRNew ---

void  sStepperMDRNew::Preconfigure (sNodePtr  DeviceCfg)
{
  sPtr<sFolderNode>              DeviceSettings;
  DeviceCfg |= "Settings";
  DeviceSettings = DeviceCfg >> "Settings";
  if( DeviceSettings.IsCorrect() ){
    DeviceSettings |= attr("StepperName","@MSDN204B");
    DeviceSettings |= attr("MaxSpeedTicksPerStep",30);
    DeviceSettings |= attr("MinimalTicksPerStep",50);
    DeviceSettings |= attr("StoppingZone",200);
    DeviceSettings |= attr("StepperTimePerStepQuantum",sPhysValue(32e-6,"s"));
    DeviceSettings |= "WLCoefs";
    DeviceSettings >> "WLCoefs" |= attr(1 ,sPhysValue(  0.001  ,"nm"));
    DeviceSettings >> "WLCoefs" |= attr(2 ,sPhysValue(  0.001  ,"nm"));
    DeviceSettings >> "WLCoefs" |= attr(3 ,sPhysValue(  0.002  ,"nm"));
    DeviceSettings >> "WLCoefs" |= attr(4 ,sPhysValue(  0.004  ,"nm"));
    DeviceSettings |= "WLShifts";
    DeviceSettings >> "WLShifts" |= attr(1 ,sPhysValue(  0.0  ,"nm"));
    DeviceSettings >> "WLShifts" |= attr(2 ,sPhysValue(  0.0  ,"nm"));
    DeviceSettings >> "WLShifts" |= attr(3 ,sPhysValue(  0.0  ,"nm"));
    DeviceSettings >> "WLShifts" |= attr(4 ,sPhysValue(  0.0  ,"nm"));
    DeviceSettings |= attr("InitWavelength",sPhysValue(500.0,0.01,"nm"));
  };
  sStepperMDR::Preconfigure(DeviceCfg);
}


sStepperMDRNew::sStepperMDRNew(sString   DevicetId
                              ,sNodePtr  DeviceCfg)
    :sStepperMDR(DevicetId,DeviceCfg)
    ,InitWavelengthPar(sPhysValue(500.0,0.01,"nm"))
{
}

sStepperMDRNew::~sStepperMDRNew()
{
  if( DeviceSettings.IsCorrect() ){
    DeviceSettings
        << attr("InitWavelength",InitWavelengthPar);
  }
}

sPhysValue  sStepperMDRNew::Position (int  steps) const
{
    return  WLCoef[GratingNo]*steps - WLShift[GratingNo];
}


int  sStepperMDRNew::Steps (sPhysValue  position) const
{
    return Round( ((position+WLShift[GratingNo])
                   /WLCoef[GratingNo]).ToReal() );
}


void  sStepperMDRNew::PostConstructor ()
{
  sStepperMDR::PostConstructor();

  try{ //get config parameters
    get_from<sPhysValueNode>(DeviceSettings >> "InitWavelength" ,InitWavelengthPar);
  }catch(...){
    SetInvalid(tr("Something wrong in configuration."));
  }

  //evolution of the inherited state machine (behaviour)
  {
    //evolution of the State_Warming
    {
      State_W_Initiating = new sState(State_Warming(),"State_W_Initiating");
      State_W_Checking
          ->AddTransition(State_W_Initiating);
      State_W_Initiating
          ->AddTransition(State_Warming()->Fin())
          ;
      //evolution of the State_W_Initiating
      {
          State_W_Initiating_WaitingRoger1 = new sState(State_W_Initiating
                                         ,"State_W_Initiating_WaitingRoger1");
          State_W_Initiating_WaitingInput = new sState(State_W_Initiating
                                          ,"State_W_Initiating_WaitingInput");
          State_W_Initiating_WaitingRoger2 = new sState(State_W_Initiating
                                         ,"State_W_Initiating_WaitingRoger2");
          State_W_Initiating_WaitingRoger1
              ->SetAsInitial()
              ->AddTransition(SIGNAL(GotStepperMessage(sTelex))
                             ,State_W_Initiating_WaitingInput
                             ,sTelex("<"))
              ;
          State_W_Initiating_WaitingInput
              ->OnEnterCall(SLOT(AskInitialWavelength()))
              ->AddTransition(SIGNAL(InitialWLSent())
                             ,State_W_Initiating_WaitingRoger2)
              ;
          State_W_Initiating_WaitingRoger2
              ->AddTransition(SIGNAL(GotStepperMessage(sTelex))
                             ,State_W_Initiating->Fin()
                             ,sTelex("<"))
              ;
       }
    }
    //evolution of the State_Operating
    {
      State_O_Initiating
          = new sState(State_Operating(),"State_O_Initiating");

      State_O_M_Acquiring
          ->AddTransition(cTune ,State_O_Initiating)
          ;
      State_O_M_Pausing
          ->AddTransition(cTune ,State_O_Initiating)
          ;
      State_O_Initiating
          ->OnEnterCall(SLOT(StepperControllerInitiation()))
          ->AddTransition(State_O_M_Acquiring)
          ->OnExitCall(SLOT(SendingReady()))
          ;

      //State_O_Initiating internal structure and behaviour
      {
          State_O_Initiating_WaitingRoger1 = new sState(State_O_Initiating
                                          ,"State_O_Initiating_WaitingRoger1");
          State_O_Initiating_WaitingInput = new sState(State_O_Initiating
                                           ,"State_O_Initiating_WaitingInput");
          State_O_Initiating_WaitingRoger2 = new sState(State_O_Initiating
                                          ,"State_O_Initiating_WaitingRoger2");
          State_O_Initiating_WaitingRoger1
              ->SetAsInitial()
              ->AddTransition(SIGNAL(GotStepperMessage(sTelex))
                             ,State_W_Initiating_WaitingInput
                             ,sTelex("<"))
              ;
          State_O_Initiating_WaitingInput
              ->OnEnterCall(SLOT(AskInitialWavelength()))
              ->AddTransition(SIGNAL(InitialWLSent())
                             ,State_O_Initiating_WaitingRoger2)
              ;
          State_O_Initiating_WaitingRoger2
              ->AddTransition(SIGNAL(GotStepperMessage(sTelex))
                             ,State_O_Initiating->Fin()
                             ,sTelex("<"))
              ;
       }
    }
  }
}


void sStepperMDRNew::StepperControllerInitiation ()
{
  sStepperMDR::StepperControllerInitiation ();
  SendDirective<sTelex>(StepperCom ,sTeleCom::cSendTelex ,sTelex("I"));
}

void sStepperMDRNew::AskInitialWavelength()
{
  sEffectorInitDialog         D(InitWavelengthPar);
  while( ! D.exec() );
  int initIndex = Round( (D.InitValueField->Value()/WLCoef[0]).ToReal() );
  QByteArray _comm;
  _comm.append(QByteArray::number((qlonglong)initIndex, 16)
                         .toUpper().rightJustified(8, '0'));
  _comm.prepend("V");
  SendDirective<sTelex>(StepperCom ,sTeleCom::cSendTelex ,sTelex(_comm));
  emit InitialWLSent();
}

void sStepperMDRNew::RecalculateWL(sTelex theIndex)
{
//    qDebug()<< "Full Index Telex" << theIndex;
    int GrNoStore = GratingNo;
    char GrN = theIndex[0];
    if (GrN == '1') GratingNo = 0;
    else if (GrN == '2') GratingNo = 1;
    else if (GrN == '3') GratingNo = 2;
    else if (GrN == '4') GratingNo = 3;
    else GratingNo = 0;
    emit TheGratingNo(GrN);
    if (!(GrNoStore == GratingNo)) {
      TheRange = sPhysRange( WLCoef[GratingNo]*IndexRange.From()
                           , WLCoef[GratingNo]*IndexRange.To() );
      emit RangeChanged(TheRange);
    }
    bool OK;
    QByteArray Ind = QByteArray(theIndex);
    Ind.remove(0,1);
    int CurrentIndex = Ind.toLong(&OK,16);
    if (OK){
        emit NewData(Wavelength(CurrentIndex));
        InitWavelengthPar = WLCoef[0]*CurrentIndex;
    } else
        emit ErrorMessage( tr("StepperMDR '%1' says:\n\n"
                          "Illegal format of WLIndex message")
                         .arg(ToQString(DeviceIdentifier()))     );
}

void sStepperMDRNew::SendingReady()
{
  emit TuneFinished();
}

//------------------------------------------------------------------------------
} //namespace RA




