////////////////////////////////////////////////////////////////////////////////
/*! @file ZScanDriver.cpp Driver for Stepping motor microcontroller board.
- Part of Nelson - Research Assistant Plugin package.
- Uses  RAGUI - Research Assistant Graphical User Interface.
- Uses  QtGui v.5.0  -  http://qt.nokia.com/
- Uses  RANet - Research Assistant Net Library (based on ANSI C++).
- Copyright(C) 2010-2013, Viktor E. Bursian, St.Petersburg, Russia.
                          Viktor.Bursian@mail.ioffe.ru
- Copyright(C) 2018, Dmitry K. Nelson, St.Petersburg, Russia.
                     D.Nelson@mail.ioffe.ru
*///////////////////////////////////////////////////////////////////////////////
#include "ZScanDriver.h"
#include "attr.h"
#include "StringsAndQStrings.h"
#include <QDebug>
namespace RA {
//------------------------------------------------------------------------------

DEFINE_CLASS_TAG(sZScanDriver)

//--------------------------------------------------------------- ZScanDriver ---

void  sZScanDriver::Preconfigure (sNodePtr  DeviceCfg)
{
  sPtr<sFolderNode>              DeviceSettings;
  sPtr<sFolderNode>              DeviceLinks;
  DeviceCfg |= "Links";
  DeviceLinks = DeviceCfg >> "Links";
  if( DeviceLinks.IsCorrect() ){
    DeviceLinks |= "StepperCom";
    DeviceLinks >> "StepperCom" |= attr("Class","sTelecom");
  };
  DeviceCfg |= "Settings";
  DeviceSettings = DeviceCfg >> "Settings";
  if( DeviceSettings.IsCorrect() ){
    DeviceSettings |= attr("StepperName","@MSDN204E");
    DeviceSettings |= attr("Range",sPhysRange(0,20,"mm"));
    DeviceSettings |= attr("LengthPerStep",sPhysValue(7.8125e-5,"mm"));
    DeviceSettings |= attr("TicksPerStep",20);
    DeviceSettings |= attr("InitValue",sPhysValue(10.0,0.01,"mm"));
  };
  sEffector::Preconfigure(DeviceCfg);
}

sZScanDriver::sZScanDriver(sString   DevicetId
                        ,sNodePtr  DeviceCfg)
    :sEffector(DevicetId,DeviceCfg)
    ,StepperCom(this)
    ,InitValuePar(sPhysValue(10.0,0.01,"mm"))
    ,TicksPerStep(20)
    ,CurrentPosition(0.0,0.01,"mm")
    ,NewPosition(0.0,0.01,"mm")
{
}

sZScanDriver::~sZScanDriver()
{
  if( DeviceSettings.IsCorrect() ){
    DeviceSettings
        << attr("InitValue",InitValuePar);
  }
    StepperCom.SetToNULL();
}

void  sZScanDriver::SetWantedPeriod ()
{
  int                         P;
  P = Round( (WantedPeriod/sPhysValue(1,"ms")).ToReal() );
  ThePeriod = P * sPhysValue(1,"ms");
  State_O_M_Acquiring->SetTimeout(P);
}


void  sZScanDriver::PostConstructor ()
{
  sEffector::PostConstructor();

  LengthPerStep = sPhysValue(7.8125e-5,"mm");
  TicksPerStep = 20;

  try{ //get config parameters
    get_from<sTextNode>(DeviceSettings >> "StepperName", StepperName);
    get_from<sIntegerNode>(DeviceSettings >> "TicksPerStep", TicksPerStep);
    get_from<sPhysValueNode>(DeviceSettings >> "LengthPerStep", LengthPerStep);
    get_from<sPhysValueNode>(DeviceSettings >> "InitValue" ,InitValuePar);
  }catch(...){
    SetInvalid(tr("Something wrong in configuration."));
  }

  //evolution of the inherited state machine (behaviour)
  {
    //evolution of the State_Warming
    {
      State_W_Linking = new sState(State_Warming(),"State_W_Linking");
      State_W_Checking = new sState(State_Warming(),"State_W_Checking");
      State_W_Initiating = new sState(State_Warming(),"State_W_Initiating");
      State_W_Linking
          ->SetAsInitial()
          ->OnEnterCall( SLOT(LinkToStepperComAndMakeItWarming()) )
          ->AddTransition( SIGNAL(StepperComLinked()) ,State_W_Checking )
          ->AddTransition( SIGNAL(ErrorMessage()) ,State_Cooling() )
          ;
      State_W_Checking
          ->OnEnterCall(SLOT(StepperControllerAskName()))
          ->AddTransition(State_W_Initiating)
          ->OnFinishCall(SLOT(StepperControllerInitiation()))
          ;
      State_W_Ch_Waiting = new sState(State_W_Checking,"State_W_Ch_Waiting");
      State_W_Ch_Waiting
          ->SetAsInitial()
          ->AddTransition(State_W_Ch_Waiting,SIGNAL(Timeout())
                         ,State_Cooling())
          ->AddTransition<sTelex>(SIGNAL(GotStepperMessage(sTelex))
                                 ,State_W_Checking->Fin()
                                 ,sTelex(StepperName))
          ;
      State_W_Initiating
          ->OnEnterCall(SLOT(AskInitialPosition()))
          ->AddTransition(State_Warming()->Fin())
          ;
    }
    //evolution of the State_Operating
    {
      State_O_M_Pausing
          = new sState(State_O_Measuring,ThePeriod,"State_O_M_Pausing");
      State_O_Initiating
          = new sState(State_Operating(),"State_O_Initiating");
      State_O_SettingTo
          = new sState(State_Operating(),"State_O_SettingTo");
      State_O_SettingTo_WaitingRoger = new sState(State_O_SettingTo
                                             ,"State_O_SettingTo_WaitingRoger");
      State_O_SettingTo_WaitingDone = new sState(State_O_SettingTo
                                              ,"State_O_SettingTo_WaitingDone");

      State_O_Initiating
          ->OnEnterCall(SLOT(AskInitialPosition()))
          ->AddTransition(State_O_M_Acquiring)
          ->OnExitCall(SLOT(SendingReady()))
          ;
      State_O_M_Acquiring
          ->OnEnterCall( SLOT(SayPosition()) )
          ->AddTransition(cTune ,State_O_Initiating)
          ->AddTransition(State_O_M_Pausing)
          ->AddTransition(cSetTo ,SLOT(SetDestination(sPhysValue))
                         ,State_O_SettingTo)
          ->AddTransition(SIGNAL(ErrorMessage(QString))
                         ,State_Cooling())
          ;
      State_O_M_Pausing
          ->AddTransition(cTune ,State_O_Initiating)
          ->AddTransition(State_O_M_Pausing,SIGNAL(Timeout())
                         ,State_O_M_Cycling)
          ->AddTransition(cSetTo ,SLOT(SetDestination(sPhysValue))
                         ,State_O_SettingTo)
          ->AddTransition(SIGNAL(ErrorMessage(QString))
                         ,State_Cooling())
          ;
      State_O_SettingTo
          ->OnFinishCall(SLOT(SayDone()))
          ->AddTransition(State_O_M_Acquiring)
          ;
      //State_O_SettingTo internal structure and behaviour
      {

        State_O_SettingTo_WaitingRoger
            ->SetAsInitial()
            ->AddTransition<sTelex>(SIGNAL(GotStepperMessage(sTelex))
                                   ,State_O_SettingTo_WaitingDone
                                   ,sTelex("<"));

        State_O_SettingTo_WaitingDone
            ->AddTransition<sTelex>(SIGNAL(GotStepperMessage(sTelex))
                                   ,State_O_SettingTo->Fin()
                                   ,sTelex("Done"));
      }
    }
    connect(this ,SIGNAL(StepperComError(QString))
           ,this ,SLOT(ReportStepperComError(QString)) );
  }
}


void sZScanDriver::LinkToStepperComAndMakeItWarming ()
{
  sString                     TeleComId;
  if( ! get_from<sTextNode>(DeviceLinks>>"StepperCom">>"DeviceId",TeleComId) ){
    emit ErrorMessage( tr("The following link is absent: 'StepperCom'.\n"
                          "Check configuration.") );
  }else{
    StepperCom.SetTo(TeleComId);
    if( ! StepperCom.IsCorrect() ){
      emit ErrorMessage( tr("The 'StepperCom' link is set to '%1', which "
                            "does not look correct. Check configuration.")
                           .arg(ToQString(TeleComId)) );
    }else{
      Connect(StepperCom ,SIGNAL(Warmed()) ,SIGNAL(StepperComLinked()) );
      Connect(StepperCom ,SIGNAL(ErrorMessage(QString))
             ,SIGNAL(StepperComError(QString)) );
      Connect(StepperCom ,SIGNAL(GotTelex(sTelex))
             ,SIGNAL(GotStepperMessage(sTelex)) );
    };
  };
}

void sZScanDriver::ReportStepperComError (QString  error_text)
{
  emit ErrorMessage( tr("ZScanDriver '%1' says:\n\n"
                        "Got error from stepper connected to '%2':\n\n%3")
                       .arg(ToQString(DeviceIdentifier()))
                       .arg(ToQString(StepperCom->DeviceIdentifier()))
                       .arg(error_text)     );
}

void sZScanDriver::StepperControllerAskName ()
{
    SendDirective<sTelex>(StepperCom ,sTeleCom::cSendTelex ,sTelex("?"));
}

void sZScanDriver::StepperControllerInitiation ()
{
  QByteArray _comm;
  _comm.append(QByteArray::number(TicksPerStep, 16)
                         .toUpper().rightJustified(4, '0'));
  _comm.prepend("T");
  SendDirective<sTelex>(StepperCom ,sTeleCom::cSendTelex ,sTelex(_comm));
}

void sZScanDriver::AskInitialPosition()
{
  sEffectorInitDialog         D(InitValuePar);
  while( ! D.exec() );
  CurrentPosition = D.InitValueField->Value();
}

void sZScanDriver::SayPosition()
{
  emit NewData(CurrentPosition);
  InitValuePar = CurrentPosition;
}

void sZScanDriver::SetDestination(sPhysValue  destination)
{
  NewPosition = destination;
  sPhysValue _theUnit = sPhysValue(1.0,destination.Units());
  try{
    if ((NewPosition/_theUnit).ToReal() > (Range().To()/_theUnit).ToReal())
       NewPosition = Range().To();
    if ((NewPosition/_theUnit).ToReal() < (Range().From()/_theUnit).ToReal())
      NewPosition = Range().From();
  }catch(...){
    emit ErrorMessage(tr("Something wrong with units."));
  }
  int StepsToGo;
  StepsToGo = Round( ((destination - CurrentPosition)/LengthPerStep).ToReal() );
  QByteArray ComPref;
  if (StepsToGo < 0) {
    ComPref = "B";
    StepsToGo = -StepsToGo;
  } else {
    ComPref = "F";
  }
  QByteArray _comm;
  _comm.append(QByteArray::number((qlonglong)StepsToGo, 16)
                           .toUpper().rightJustified(8, '0'));
  _comm.prepend(ComPref);
  SendDirective<sTelex>(StepperCom ,sTeleCom::cSendTelex ,sTelex(_comm));
}

void sZScanDriver::SayDone()
{
  CurrentPosition = NewPosition;
  emit Done(CurrentPosition);
}

void sZScanDriver::SendingReady()
{
  emit TuneFinished();
}

//------------------------------------------------------------------------------
} //namespace RA




