////////////////////////////////////////////////////////////////////////////////
/*! @file BursiPanHandles.cpp Joint console for RadioPAN handles.
- Part of PanBursian - Research Assistant Plugin package.
- Uses  RAGUI - Research Assistant Graphical User Interface Library.
- Uses  RANet - Research Assistant Net Library.
- Uses  Qt v.5  -  http://qt-project.org/
- Copyright(C) 2013-2014, Viktor E. Bursian, St.Petersburg, Russia.
                          Viktor.Bursian@mail.ioffe.ru
*///////////////////////////////////////////////////////////////////////////////
#include "BursiPanHandles.h"
#include "attr.h"
#include "StringsAndQStrings.h"
#include "Log.h"
#include <QVBoxLayout>
#include "flowlayout.h"
#include <QMessageBox>
namespace RA {
//------------------------------------------------------------------------------

DEFINE_CLASS_TAG(sBursiPanHandles)

//---------------------------------------------------- sBursiPanHandleSplash ---

sBursiPanHandleSplash::sBursiPanHandleSplash ()
    :QSplashScreen()
    ,Exposition(90)
    ,ExpositionCounter(0)
    ,Clock(new QTimer(this))
    ,Text(new QLabel(this))
{
  setStyleSheet("font-size: 80px; background: yellow; color: navy");
  QBoxLayout *       Layout = new QVBoxLayout;
  Layout->setAlignment(Qt::AlignCenter | Qt::AlignHCenter);
  setLayout(Layout);
  Layout->addWidget(Text);
  connect(Clock,SIGNAL(timeout()),this,SLOT(OnClock()));
}


void  sBursiPanHandleSplash::Display (QString  text)
{
  ExpositionCounter = Exposition;
  Text->setText(text);
  if( ! isVisible() ){
    show();
    Clock->start(100);
  }
}


void  sBursiPanHandleSplash::OnClock ()
{
  ExpositionCounter--;
  if( ExpositionCounter > 0 ){
    raise();
  }else{
    Clock->stop();
    hide();
    Text->setText("");
  }
}

//--------------------------------------------------------- sBursiPanHandles ---

void  sBursiPanHandles::Preconfigure (sNodePtr  DeviceCfg)
{
  sPtr<sFolderNode>              DeviceLinks;
  DeviceCfg |= "Links";
  DeviceLinks = DeviceCfg >> "Links";
  if( DeviceLinks.IsCorrect() ){
    DeviceLinks |= "BursiPan";
    DeviceLinks >> "BursiPan" |= attr("Class","sBursiPan");
  };
  sConsole::Preconfigure(DeviceCfg);
}


sBursiPanHandles::~sBursiPanHandles ()
{
  BursiPan.SetToNULL();
}


sBursiPanHandles::sBursiPanHandles (sString   DeviceId
                                   ,sNodePtr  DeviceCfg
                                   ,sNodePtr  WidgetCfg
                                   ,QWidget *  parent    )
    :sConsole(DeviceId,DeviceCfg,WidgetCfg,parent)
    ,BursiPan(this)
{
  //evolution of the inherited state machine
  {
    //evolution of the State_Warming
    {
      State_W_BursiPan = new sState(State_Warming());
      State_W_BursiPan
          ->SetAsInitial()
          ->AddTransition(State_Warming()->Fin());
      //State_W_BursiPan internal structure and behaviour
      {
        State_W_BursiPan_Linking
            =( new sState(State_W_BursiPan) )
            ->SetAsInitial()
            ->OnEnterCall( SLOT(LinkToBursiPan()) )
            ->AddTransition( SIGNAL(BursiPanLinked()) ,State_W_BursiPan->Fin() )
            ->AddTransition( SIGNAL(ErrorMessage()) ,State_Cooling() );
      }
    }
    //State_Operating internal structure and behaviour
    {
      State_O_Watching = new sState(State_Operating(),"State_O_Watching");
      State_O_Controling = new sState(State_Operating(),"State_O_Controling");

      State_O_Watching
          ->SetAsInitial()
          ->AddReaction(SIGNAL(SetHo()) ,SLOT(OnSetHo()) )
          ->AddReaction(SIGNAL(SetDeltaH()) ,SLOT(OnSetDeltaH()) )
          ->AddReaction(SIGNAL(SetFrequency()) ,SLOT(OnSetFrequency()) )
          ->AddReaction(SIGNAL(SetAttenuation()) ,SLOT(OnSetAttenuation()) )
//          ->AddReaction(SIGNAL(SetGain()) ,SLOT(OnSetGain()) )
//          ->AddReaction(SIGNAL(SetModulation()) ,SLOT(OnSetModulation()) )
//          ->AddReaction(SIGNAL(SetModulationMode())
//                       ,SLOT(OnSetModulationMode()) )
          ->AddTransition(SIGNAL(BursiPanControlTaken()) ,State_O_Controling)
          ;
      State_O_Controling
//          ->AddReaction(SIGNAL(Break()) ,SLOT(OnBreak())   )
          ->AddTransition(SIGNAL(RodgerFromBursiPan()),State_O_Watching)
          ->OnExitCall(SLOT(DisconnectRodger()))
          ->OnExitCall(SLOT(ReleaseBursiPanControl()))
          ;
    }
    connect(this ,SIGNAL(BursiPanError(QString))
           ,this ,SLOT(ReportBursiPanError(QString)) );
  }
}


void  sBursiPanHandles::AppearanceMakeUp ()
{
  sConsole::AppearanceMakeUp();

  SplashScreen = new sBursiPanHandleSplash;

  HoField = new sPhysValueField(this,sPhysValue());
  connect(HoField ,SIGNAL(Edited()) ,this ,SIGNAL(SetHo()));
  //!@todo{Instruments} Review Connect
  connect(this ,SIGNAL(NewHo(sPhysValue))
         ,HoField ,SLOT(SetValue(sPhysValue)) );

  DeltaHField = new sPhysValueField(this,sPhysValue());
  connect(DeltaHField ,SIGNAL(Edited()) ,this ,SIGNAL(SetDeltaH()));
  connect(this ,SIGNAL(NewDeltaH(sPhysValue))
         ,DeltaHField ,SLOT(SetValue(sPhysValue)) );

  FrequencyField = new sPhysValueField(this,sPhysValue());
  connect(FrequencyField ,SIGNAL(Edited()) ,this ,SIGNAL(SetFrequency()));
  connect(this ,SIGNAL(NewFrequency(sPhysValue))
         ,FrequencyField ,SLOT(SetValue(sPhysValue)) );

  AttenuationField = new sPhysValueField(this,sPhysValue());
  connect(AttenuationField ,SIGNAL(Edited()) ,this ,SIGNAL(SetAttenuation()));
  connect(this ,SIGNAL(NewAttenuation(sPhysValue))
         ,AttenuationField ,SLOT(SetValue(sPhysValue)) );

  GainField = new sPhysValueField(this,sPhysValue());
  GainField->SetReadOnly();
//  connect(GainField ,SIGNAL(Edited()) ,this ,SIGNAL(SetGain()));

  ModulationField = new sPhysValueField(this,sPhysValue());
  ModulationField->SetReadOnly();
//  connect(ModulationField ,SIGNAL(Edited()) ,this ,SIGNAL(SetModulation()));

  ModulationModeField = new sFlexibleField(this);
  ModulationModeField->SetReadOnly();

//  CavityField = new sSelectNodeWidget(this);

  QBoxLayout *       Layout = new QVBoxLayout;
  setLayout(Layout);
  //
  {
    QLayout *       L = new FlowLayout(0,30,3);
    layout()->addItem(L);
    L->addWidget(HoField);
    L->addWidget(DeltaHField);
    L->addWidget(AttenuationField);
    L->addWidget(FrequencyField);
    L->addWidget(GainField);
    L->addWidget(ModulationField);
    L->addWidget(ModulationModeField);
  }

  Translate();

  HoField->SetValidRange(sPhysRange(0,10000.0,"Oe",0.1));
  DeltaHField->SetValidRange(sPhysRange(0,7000,"Oe",1));
  FrequencyField->SetValidRange(sPhysRange(8.8e9,9.7e9,"Hz",0.0001e9));
  AttenuationField->SetValidRange(sPhysRange(0,30,_Unitsless_,0.1));
//  GainField->SetValidRange(sPhysRange(0,4e6,_Unitsless_,1));
//  ModulationField->SetValidRange(sPhysRange(0,10,"Oe",0.0001));
}


void  sBursiPanHandles::Translate ()
{
  HoField->SetLabel("H<sub>0</sub>");
  HoField->setToolTip(tr("central field"));
  DeltaHField->SetLabel("<!DOCTYPE HTML>&Delta;H");
  DeltaHField->setToolTip(tr("field range"));
  FrequencyField->SetLabel("<!DOCTYPE HTML>&nu;");
  FrequencyField->setToolTip(tr("microwave frequency"));
  AttenuationField->SetLabel("A");
  AttenuationField->setToolTip(tr("attenuation"));
  GainField->SetLabel("G");
  GainField->setToolTip(tr("gain"));
  ModulationField->SetLabel("M");
  ModulationField->setToolTip(tr("modulation amplitude"));
  ModulationModeField->setToolTip(tr("modulation mode"));
}


void  sBursiPanHandles::ReportBursiPanError (QString  error_text)
{
  emit ErrorMessage( tr("Got error from '%1':\n\n%2")
                       .arg(ToQString(BursiPan->DeviceIdentifier()))
                       .arg(error_text)     );
}


void  sBursiPanHandles::LinkToBursiPan ()
{
  try{
    BursiPan.SetTo(DeviceLinks,"BursiPan");
    Connect(BursiPan ,SIGNAL(Warmed())       ,SIGNAL(BursiPanLinked()) );
    Connect(BursiPan ,SIGNAL(ErrorMessage(QString))
           ,SIGNAL(BursiPanError(QString)) );
    Connect(BursiPan ,SIGNAL(NewHo(sPhysValue))
           ,SIGNAL(NewHo(sPhysValue)) );
    Connect(BursiPan ,SIGNAL(NewDeltaH(sPhysValue))
           ,SIGNAL(NewDeltaH(sPhysValue)) );
    Connect(BursiPan ,SIGNAL(NewFrequency(sPhysValue))
           ,SIGNAL(NewFrequency(sPhysValue)) );
    Connect(BursiPan ,SIGNAL(NewAttenuation(sPhysValue))
           ,SIGNAL(NewAttenuation(sPhysValue)) );
    Connect(BursiPan ,SIGNAL(NewGain(sPhysValue))
           ,SLOT(OnNewGain(sPhysValue)) );
    Connect(BursiPan ,SIGNAL(NewModulation(sPhysValue))
           ,SLOT(OnNewModulation(sPhysValue)) );
    Connect(BursiPan ,SIGNAL(NewModulationMode(sBursiPan::eModulationMode))
           ,SLOT(OnNewModulationMode(sBursiPan::eModulationMode)) );
  }catch( rxDeviceLaunch E ){
    emit ErrorMessage( tr("Error while linking to a submissive device.\n"
                          "%1\nCheck configuration.")
                         .arg(ToQString(E.MsgText()))   );
  }
}


void  sBursiPanHandles::ReleaseBursiPanControl ()
{
  BursiPan.ReleaseExclusiveControl();
}


void  sBursiPanHandles::OnEnterToCoolingState ()
{
  Disconnect(BursiPan,0,0);      //disconnect all signals from BursiPan,
                                 //which were connected to any of my slots
  BursiPan.ReleaseExclusiveControl();
  BursiPan.SetToNULL();
  sConsole::OnEnterToCoolingState();
}


void  sBursiPanHandles::OnEnterToOperatingState()
{
  sConsole::OnEnterToOperatingState();
  HoField->SetValue(BursiPan->Ho());
  DeltaHField->SetValue(BursiPan->DeltaH());
  FrequencyField->SetValue(BursiPan->Frequency());
  AttenuationField->SetValue(BursiPan->Attenuation());
  GainField->SetValue(BursiPan->Gain());
  ModulationField->SetValue(BursiPan->Modulation());
  ModulationModeField->setText(ToQString(
                        sBursiPan::cModeText[ BursiPan->ModulationMode() ] ));
}


void sBursiPanHandles::DisconnectRodger()
{
  Disconnect(BursiPan,0,SIGNAL(RodgerFromBursiPan()));
}


void  sBursiPanHandles::OnSetHo ()
{
  if( SendDirective(BursiPan ,sBursiPan::cSetHo
                             ,HoField->Value() ,cTakeControl) ){

    Connect(BursiPan ,SIGNAL(NewHo(sPhysValue)) ,SIGNAL(RodgerFromBursiPan()) );
    emit BursiPanControlTaken();
  }else{
    HoField->SetValue(BursiPan->Ho());
  }
}


void  sBursiPanHandles::OnSetDeltaH ()
{
  if( SendDirective(BursiPan ,sBursiPan::cSetDeltaH
                             ,DeltaHField->Value() ,cTakeControl) ){
    Connect(BursiPan ,SIGNAL(NewDeltaH(sPhysValue))
           ,SIGNAL(RodgerFromBursiPan()) );
    emit BursiPanControlTaken();
  }else{
    DeltaHField->SetValue(BursiPan->DeltaH());
  }
}


void  sBursiPanHandles::OnSetFrequency ()
{
  if( SendDirective(BursiPan ,sBursiPan::cSetFrequency
                             ,FrequencyField->Value() ,cTakeControl) ){
    Connect(BursiPan ,SIGNAL(NewFrequency(sPhysValue))
           ,SIGNAL(RodgerFromBursiPan()) );
    emit BursiPanControlTaken();
  }else{
    FrequencyField->SetValue(BursiPan->Frequency());
  }
}


void  sBursiPanHandles::OnSetAttenuation ()
{
  if( SendDirective(BursiPan ,sBursiPan::cSetAttenuation
                             ,AttenuationField->Value() ,cTakeControl) ){
    Connect(BursiPan ,SIGNAL(NewAttenuation(sPhysValue))
           ,SIGNAL(RodgerFromBursiPan()) );
    emit BursiPanControlTaken();
  }else{
    AttenuationField->SetValue(BursiPan->Attenuation());
  }
}


//void  sBursiPanHandles::OnSetGain ()
//{
//  if( SendDirective(BursiPan ,sBursiPan::cSetGain
//                             ,GainField->Value() ,cTakeControl) ){
//    Connect(BursiPan ,SIGNAL(NewGain(sPhysValue))
//           ,SIGNAL(RodgerFromBursiPan()) );
//    emit BursiPanControlTaken();
//  }
//}


//void  sBursiPanHandles::OnSetModulation ()
//{
//  if( SendDirective(BursiPan ,sBursiPan::cSetModulation
//                             ,ModulationField->Value() ,cTakeControl) ){
//    Connect(BursiPan ,SIGNAL(NewModulation(sPhysValue))
//           ,SIGNAL(RodgerFromBursiPan()) );
//    emit BursiPanControlTaken();
//  }
//}


//void  sBursiPanHandles::OnSetModulationMode ()
//{
//  if( SendDirective(BursiPan ,sBursiPan::cSetModulationMode
//                             ,ModulationModeField->Value() ,cTakeControl) ){
//    Connect(BursiPan ,SIGNAL(NewModulationMode(eModulationMode))
//           ,SIGNAL(RodgerFromBursiPan()) );
//    emit BursiPanControlTaken();
//  }
//}


void  sBursiPanHandles::OnNewGain (sPhysValue  g)
{
  GainField->SetValue(g);
  SplashScreen->Display( GainField->Label() + "= "
                         + ToQString(g.Text()) );
}


void  sBursiPanHandles::OnNewModulation (sPhysValue  m)
{
  ModulationField->SetValue(m);
  SplashScreen->Display( ModulationField->Label() + "= "
                       + ToQString(m.Text()) );
}


void  sBursiPanHandles::OnNewModulationMode (sBursiPan::eModulationMode  m)
{
  ModulationModeField->setText(ToQString(sBursiPan::cModeText[m]));
  SplashScreen->Display( ToQString(sBursiPan::cModeText[m]) );
}

//------------------------------------------------------------------------------
} //namespace RA
