////////////////////////////////////////////////////////////////////////////////
/*! @file BursiPanMagnet.cpp BursiPan magnetic field effector.
- Part of PanBursian - Research Assistant Plugin package.
- Uses  RAGUI - Research Assistant Graphical User Interface Library.
- Uses  RANet - Research Assistant Net Library.
- Uses  Qt v.5  -  http://qt-project.org/
- Copyright(C) 2014, Viktor E. Bursian, St.Petersburg, Russia.
                     Viktor.Bursian@mail.ioffe.ru
*///////////////////////////////////////////////////////////////////////////////
#include "BursiPanMagnet.h"
#include "attr.h"
#include "StringsAndQStrings.h"
#include "Log.h"
namespace RA {
//------------------------------------------------------------------------------

DEFINE_CLASS_TAG(sBursiPanMagnet)

//---------------------------------------------------------- sBursiPanMagnet ---

void  sBursiPanMagnet::Preconfigure (sNodePtr  DeviceCfg)
{
  sPtr<sFolderNode>              DeviceLinks;
  DeviceCfg |= "Links";
  DeviceLinks = DeviceCfg >> "Links";
  if( DeviceLinks.IsCorrect() ){
    DeviceLinks |= "BursiPan";
    DeviceLinks >> "BursiPan" |= attr("Class","sBursiPan");
  }
  sPtr<sFolderNode>              DeviceSettings;
  DeviceCfg |= "Settings";
  DeviceSettings = DeviceCfg >> "Settings";
  if( DeviceSettings.IsCorrect() ){
    DeviceSettings |= attr("Range",sPhysRange(0.0,10000.0,"Oe",0.01));
  }
  sEffector::Preconfigure(DeviceCfg);
}


sBursiPanMagnet::~sBursiPanMagnet ()
{
  BursiPan.SetToNULL();
}


sBursiPanMagnet::sBursiPanMagnet (sString   DevicetId
                                 ,sNodePtr  DeviceCfg)
    :sEffector(DevicetId,DeviceCfg)
    ,BursiPan(this)
{
}


void  sBursiPanMagnet::PostConstructor ()
{
  sEffector::PostConstructor();
  //evolution of the inherited state machine (behaviour)
  {
    //evolution of the State_Warming
    {
      State_W_BursiPan = new sState(State_Warming());
      State_W_BursiPan
          ->SetAsInitial()
          ->AddTransition(State_Warming()->Fin())
          ;
      //State_W_BursiPan structure and behaviour
      {
        State_W_BursiPan_Linking = new sState(State_W_BursiPan);
        State_W_BursiPan_Linking
            ->SetAsInitial()
            ->OnEnterCall( SLOT(LinkToBursiPan()) )
            ->AddTransition( SIGNAL(BursiPanLinked()) ,State_W_BursiPan->Fin() )
            ->AddTransition( SIGNAL(ErrorMessage()) ,State_Cooling() )
            ;
      }
    }
    //evolution of the State_Operating
    {
      State_O_ParallelMeasuringAndControlling 
          = new sParallelStateGroup(State_Operating());
      State_O_ParallelMeasuringAndControlling->SetAsInitial();

      //State_O_Measuring evolution
      {
        State_O_Measuring->setParent(State_O_ParallelMeasuringAndControlling);
        State_O_M_Acquiring
            ->AddTransition<sPhysValue>(SIGNAL(GotMagneticField(sPhysValue)) 
                                       ,SIGNAL(NewData(sPhysValue)) 
                                       ,State_O_M_Cycling);
      }
        
      State_O_Controlling = new sState(State_O_ParallelMeasuringAndControlling
                                      ,"State_O_Controlling");
      //State_O_Controlling internal structure and behaviour
      {
        State_O_C_Idle = new sState(State_O_Controlling,"State_O_C_Idle");
        State_O_C_SettingTo = new sState(State_O_Controlling
                                                       ,"State_O_C_SettingTo");
      
        State_O_C_Idle
            ->SetAsInitial()
            ->AddTransition(cSetTo ,SLOT(SetH(sPhysValue))
                           ,State_O_C_SettingTo
                           )
            ;
        State_O_C_SettingTo
            ->AddTransition(cBreak ,State_O_C_Idle)
            ->AddTransition(State_O_C_Idle)
            ->OnExitCall(SLOT(ReleaseBursiPanControl()))
            ;
        //State_O_C_SettingTo internal structure and behaviour
        {
          State_O_C_SettingTo_Check = new sState(State_O_C_SettingTo
                                                ,"State_O_C_SettingTo_Check");
      
          State_O_C_SettingTo_Check
              ->SetAsInitial()
              ->AddTransition<sPhysValue>(SIGNAL(GotMagneticField(sPhysValue))
                                         ,SIGNAL(Done(sPhysValue))
                                         ,State_O_C_SettingTo->Fin())
              ;
        }
      }
    }
    connect(this ,SIGNAL(BursiPanError(QString))
           ,this ,SLOT(ReportBursiPanError(QString)) );
  }
}


void  sBursiPanMagnet::OnEnterToCoolingState ()
{
  Disconnect(BursiPan ,0 ,0);
  BursiPan.ReleaseExclusiveControl();
  BursiPan.SetToNULL();
  sEffector::OnEnterToCoolingState();
}


void  sBursiPanMagnet::ReleaseBursiPanControl ()
{
  BursiPan.ReleaseExclusiveControl();
}


void  sBursiPanMagnet::SetH (sPhysValue  H)
{
  SendDirective(BursiPan ,sBursiPan::cSetMagneticField ,H ,cTakeControl);
}


void  sBursiPanMagnet::ReportBursiPanError (QString  error_text)
{
  emit ErrorMessage( tr("Got error from '%1':\n\n%2")
                       .arg(ToQString(BursiPan->DeviceIdentifier()))
                       .arg(error_text)     );
}


void  sBursiPanMagnet::LinkToBursiPan ()
{
  try{
    BursiPan.SetTo(DeviceLinks,"BursiPan");
    Connect(BursiPan ,SIGNAL(Warmed()) ,SIGNAL(BursiPanLinked()) );
    Connect(BursiPan ,SIGNAL(ErrorMessage(QString))
                     ,SIGNAL(BursiPanError(QString)) );
    Connect(BursiPan ,SIGNAL(MagneticField(sPhysValue))
                     ,SIGNAL(GotMagneticField(sPhysValue)));
    Connect(BursiPan ,SIGNAL(RangeChanged(sPhysRange))
                     ,SIGNAL(RangeChanged(sPhysRange)));
  }catch( rxDeviceLaunch E ){
    emit ErrorMessage( tr("Error while linking to a submissive device.\n"
                          "%1\nCheck configuration.")
                         .arg(ToQString(E.MsgText()))   );
  }
}

//------------------------------------------------------------------------------
} //namespace RA
