////////////////////////////////////////////////////////////////////////////////
/*! @file BursiPanScanner.cpp EPR spectrometer BursiPan Scanner.
- Part of PanBursian - Research Assistant Plugin package.
- Uses  RAGUI - Research Assistant Graphical User Interface Library.
- Uses  RANet - Research Assistant Net Library.
- Uses  Qt v.5  -  http://qt-project.org/
- Copyright(C) 2014, Viktor E. Bursian, St.Petersburg, Russia.
                     Viktor.Bursian@mail.ioffe.ru
*///////////////////////////////////////////////////////////////////////////////
#include "BursiPanScanner.h"
#include "attr.h"
#include "Symbols.h"
#include <QMessageBox>
#include <QApplication>
#include "Trace.h"
#include <QDebug>
namespace RA {
//------------------------------------------------------------------------------

DEFINE_CLASS_TAG(sBursiPanScanner)

//--------------------------------------------------------- sBursiPanScanner ---
/*! @class sBursiPanScanner

@par The state machine

The upper-level state machine is inherited from sScanner.
Below is the evolution of some states.
@dotfile "BursiPanScanner.state_graph"

*/


void  sBursiPanScanner::Preconfigure (sNodePtr  DeviceCfg)
{
  sPtr<sFolderNode>              DeviceLinks;
  DeviceCfg |= "Links";
  DeviceLinks = DeviceCfg >> "Links";
  if( DeviceLinks.IsCorrect() ){
    DeviceLinks |= "BursiPan";
    DeviceLinks >> "BursiPan" |= attr("Class","sBursiPan");
  };
  sPtr<sFolderNode>              DeviceSettings;
  DeviceCfg |= "Settings";
  DeviceSettings = DeviceCfg >> "Settings";
  if( DeviceSettings.IsCorrect() ){
    DeviceSettings |= attr("NumOfPasses",1);
    DeviceSettings |= attr("ScanRange",sPhysRange(2000,4000,"Oe",0.5));
    DeviceSettings |= attr("ScanStep",sPhysValue(0.5,0.05,"Oe"));
    DeviceSettings |= attr("TimePerStep",sPhysValue(30,0.5,"ms"));
  };
  sScanner::Preconfigure(DeviceCfg);
}


sBursiPanScanner::~sBursiPanScanner()
{
  if( DeviceSettings.IsCorrect() ){
//    DeviceSettings << attr("ScanRange",ScanRangeField->Value())
//                   << attr("ScanStep",ScanStepField->Value());
  };
  BursiPan.SetToNULL();
  DataArrayRecord.SetToNULL();
}


sBursiPanScanner::sBursiPanScanner (sString    DevicetId
                                   ,sNodePtr   DeviceCfg
                                   ,sNodePtr   WidgetCfg
                                   ,QWidget *  parent   )
    :sScanner(DevicetId,DeviceCfg,WidgetCfg,parent)
    ,BursiPan(this)
    ,NumOfPasses(1)
    ,ScanningFinished(new QSoundEffect(this))
{
  ScanningFinished->setSource(QUrl::fromLocalFile(qApp->applicationDirPath()
      +"/../extrafiles/PanBursian/BursiPan_scanning_has_been_finished.wav"));

  //evolution of the inherited state machine
  {
    //evolution of the State_Warming
    {
      State_W_Parallel
          = (new sParallelStateGroup(State_Warming()))
          ->SetAsInitial()
          ->AddTransition(State_Warming()->Fin());
      //State_W_Parallel internal structure and behaviour
      {
        State_W_BursiPan = new sState(State_W_Parallel,"State_W_BursiPan");
        //State_W_BursiPan structure and behaviour
        {
          State_W_BursiPan_Linking
              = (new sState(State_W_BursiPan,"State_W_BursiPan_Linking"))
              ->SetAsInitial()
              ->OnEnterCall( SLOT(LinkToBursiPan()) )
              ->AddTransition( SIGNAL(BursiPanLinked()) ,State_W_BursiPan->Fin() )
              ->AddTransition( SIGNAL(ErrorMessage())  ,State_Cooling()  );
        }
      }
    }
    //evolution of the State_Operating()
    {
      State_O_StepSetting = new sState(State_Operating()
                                          ,"State_O_ScanStepSetting");
      State_O_TimePerStepSetting = new sState(State_Operating()
                                             ,"State_O_TimePerStepSetting");
      State_O_AllScanParamResetting = new sState(State_Operating()
                                              ,"State_O_AllScanParamResetting");

      State_O_AllScanParamResetting->SetAsInitial();

      //evolution of the State_O_Idle
      {
      State_O_Idle
          ->OnEnterCall(SLOT(EnableParamSetting()))
          ->OnExitCall(SLOT(DisableParamSetting()))
          ->AddReaction(SIGNAL(StepEdited())
                       ,SLOT(OnStepEdited()) )
          ->AddReaction(SIGNAL(TimePerStepEdited())
                       ,SLOT(OnTimePerStepEdited()) )
          ->AddTransition(SIGNAL(StepSetting())
                         ,State_O_StepSetting)
          ->AddTransition(SIGNAL(TimePerStepSetting())
                         ,State_O_TimePerStepSetting)
          ->AddTransition(SIGNAL(BursiPanRangeChanged(sPhysRange))
                         ,State_O_AllScanParamResetting)
          ;
      }

      State_O_StepSetting
          ->AddTransition(SIGNAL(StepChanged(sPhysValue))
//          ->AddTransition<sPhysValue>(SIGNAL(StepChanged(sPhysValue))
//                                     ,SLOT(OnStepChanged(sPhysValue))
                                     ,State_O_Idle)
          ->OnExitCall(SLOT(ShowParamComments()))
          ->OnExitCall(SLOT(ReleaseBursiPanControl()))
          ;
      State_O_TimePerStepSetting
          ->AddTransition(SIGNAL(TimePerStepChanged(sPhysValue))
//          ->AddTransition<sPhysValue>(SIGNAL(TimePerStepChanged(sPhysValue))
//                                     ,SLOT(OnTimePerStepChanged(sPhysValue))
                                     ,State_O_Idle)
          ->OnExitCall(SLOT(ShowParamComments()))
          ->OnExitCall(SLOT(ReleaseBursiPanControl()))
          ;

      //State_O_AllScanParamResetting internal structure and behaviour
      {
        State_O_ASPR_Step = new sState(State_O_AllScanParamResetting
                                      ,"State_O_ASPR_Step");
        State_O_ASPR_TimePerStep = new sState(State_O_AllScanParamResetting
                                             ,"State_O_ASPR_Step");
        State_O_AllScanParamResetting
            ->OnExitCall(SLOT(OnScanRangeEdited()))
            ->OnExitCall(SLOT(ShowParamComments()))
            ->OnExitCall(SLOT(ReleaseBursiPanControl()))
            ->AddTransition(State_O_Idle)
            ;
        State_O_ASPR_Step
            ->SetAsInitial()
            ->OnEnterCall(SLOT(OnStepEdited()))
            ->AddTransition(SIGNAL(StepChanged(sPhysValue))
//            ->AddTransition<sPhysValue>(SIGNAL(StepChanged(sPhysValue))
//                                       ,SLOT(OnStepChanged(sPhysValue))
                                       ,State_O_ASPR_TimePerStep)
            ;
        State_O_ASPR_TimePerStep
            ->OnEnterCall(SLOT(OnTimePerStepEdited()))
            ->AddTransition(SIGNAL(TimePerStepChanged(sPhysValue))
//            ->AddTransition<sPhysValue>(SIGNAL(TimePerStepChanged(sPhysValue))
//                                       ,SLOT(OnTimePerStepChanged(sPhysValue))
                                       ,State_O_AllScanParamResetting->Fin())
            ;
      }
      //evolution of the State_O_Scanning
      {
        State_O_Scanning
            ->OnExitCall(SLOT(ReleaseBursiPanControl()))
//            ->OnExitCall(ScanningFinished,SLOT(play()))
            ;
        State_O_Sc_D_Waiting
            ->SetTimeout(sPhysValue(15,"s"));
        State_O_Sc_D_Waiting
            ->OnEnterCall(ScanningFinished,SLOT(play()))
            ->AddTransition(State_O_Sc_D_Waiting,SIGNAL(Timeout())
                           ,State_O_Sc_D_Waiting)
            ;

  /*
  поправить и задать начальное поле (надо ждать)

  остановить, запомнить период, задать макс интервал, запустить
  ждать подтверждения
  запустить шагание
*/

//        State_O_Sc_Starting = new sState(State_O_Scanning
//                                            ,"State_O_Sc_Starting");
//        State_O_Sc_Breaking = new sState(State_O_Scanning
//                                            ,"State_O_Sc_Breaking");
//        State_O_Sc_Starting
//            ->SetAsInitial()
//            ->AddTransition( SIGNAL(Break()) , State_O_Sc_Breaking )
//            ->AddTransition( State_O_Sc_Acquisition )
//            ;
//        State_O_Sc_Breaking
//            ->AddTransition( State_O_Scanning->Fin() )
//            ;
//        //State_O_Sc_Starting internal structure and behavior
//        {
//          State_O_Sc_S_Waiting
//              =( new sState(State_O_Sc_Starting,"State_O_Sc_S_Waiting") );
//          State_O_Sc_S_Waiting
//              ->SetAsInitial()
//              ->OnEnterCall( SLOT(StartScanJob()) )
//              ->AddTransition(SIGNAL(BursiPanBusy()) ,State_O_Idle)
//              ->AddTransition<sBursiPan::sScanStartData>
//                         (SIGNAL(ScanStarted(sBursiPan::sScanStartData))
//                         ,SLOT(JustifyScanParams(sBursiPan::sScanStartData))
//                         ,State_O_Sc_Starting->Fin())
//              ;
//        }
        //State_O_Sc_Breaking internal structure and behavior
        {
//          State_O_Sc_B_Waiting
//              =( new sState(State_O_Sc_Breaking,"State_O_Sc_B_Waiting") )
//              ->SetAsInitial()
//              ->OnEnterCall(SLOT(ResetGaugeRegime()))
//              ->AddTransition(SIGNAL(GaugePeriodChanged(sPhysValue))
//                             ,SLOT(RestartGauge())
//                             ,State_O_Sc_Breaking->Fin());
        }
        //State_O_Sc_Acquisition evolution
        {
          State_O_Sc_A_TakingPoints = new sState(State_O_Sc_Acquisition
                                                ,"State_O_Sc_A_TakingPoints");
          State_O_Sc_A_TakingPoints
              ->SetAsInitial()
//              ->AddReaction<sBursiPan::sScanPointData>
//                           (SIGNAL(GotScanPoint(sBursiPan::sScanPointData))
//                           ,SLOT(AddNewData(sBursiPan::sScanPointData)) )
              ->AddReaction<sPhysValue>(SIGNAL(EPR_Signal(sPhysValue))
                                       ,SLOT(AddNewData(sPhysValue)) )
              ->AddReaction(SIGNAL(Break()),SLOT(OnBreak()))
              ->AddTransition(SIGNAL(ScanningHasCompleted())
                             ,State_O_Sc_Deciding)
              ->AddTransition(SIGNAL(ScanningInterrupted())
//                             ,State_O_Sc_Breaking)
                             ,State_O_Sc_Deciding)
              ;
        }
      }
    }
    connect(this ,SIGNAL(BursiPanError(QString))
           ,this ,SLOT(ReportBursiPanError(QString)) );
  }
}


void  sBursiPanScanner::AppearanceMakeUp ()
{
  sScanner::AppearanceMakeUp();
  //add new widgets
  {
    FldNumOfPasses = new QSpinBox(this);
    FldNumOfPasses->setRange(0,999);
    FldNumOfPasses->setSpecialValueText(Symbol::Infinity);
    FldNumOfPasses->setValue(NumOfPasses);
//    NumOfPassesField->setMaximumWidth(fontMetrics().width("99900"));

    NumOfPassesSuffix = new QLabel;

    DirectionField = new QComboBox(this);
    DirectionField->insertItem(0,"forwards");
    DirectionField->insertItem(1,"backwards");
    DirectionField->insertItem(2,"forth&back");
    DirectionField->insertItem(3,"back&forth");
    DirectionField->setCurrentIndex(0);

    ScanRangeField = new sPhysRangeField(ParamArea
                                        ,sPhysRange(2000,4000,"Oe",0.5)
                                        ,false,false);

    ScanStepField = new sPhysValueField(ParamArea
                                       ,sPhysValue(0.5,0.05,"Oe")
                                       ,false,false,false);

    ScanStepSuffix = new QLabel;

    TimePerStepField = new sPhysValueField(ParamArea
                                          ,sPhysValue(30,0.5,"ms")
                                          ,false,false,false);

    TimePerStepSuffix = new QLabel;

    FldTheta = new sPhysValueField(ParamArea
                                  ,sPhysValue(real_nan,Symbol::Degree) );

    FldPhi   = new sPhysValueField(ParamArea
                                  ,sPhysValue(real_nan,Symbol::Degree) );

    LblCavity = new QLabel(ParamArea);
    LblCavity->setObjectName("LblCavity");
    FldCavity = new sDropNodeWidget(ParamArea);
    FldCavity->setObjectName("FldSample");
  }
  //lay the new widgets out
  {
    ParamLayout->addWidget(FldNumOfPasses  ,++Row,0,1,1,Qt::AlignLeft);
    ParamLayout->addWidget(NumOfPassesSuffix ,  Row,1,1,1,Qt::AlignLeft);
    ParamLayout->addWidget(DirectionField    ,  Row,2,1,2,Qt::AlignLeft);
    ParamLayout->addWidget(ScanRangeField    ,++Row,0,1,4,Qt::AlignLeft);
    ParamLayout->addWidget(ScanStepField     ,++Row,0,1,2,Qt::AlignLeft);
    ParamLayout->addWidget(ScanStepSuffix    ,  Row,2,1,2,Qt::AlignLeft);
    ParamLayout->addWidget(TimePerStepField  ,++Row,0,1,2,Qt::AlignLeft);
    ParamLayout->addWidget(TimePerStepSuffix ,  Row,2,1,7,Qt::AlignLeft);
    ParamLayout->addWidget(FldTheta          ,++Row,0,1,2,Qt::AlignLeft);
    ParamLayout->addWidget(FldPhi            ,  Row,2,1,1,Qt::AlignLeft);
    ParamLayout->addWidget(LblCavity         ,   94,0,1,1);
    ParamLayout->addWidget(FldCavity         ,   94,1,1,8);
  }
  Translate();
  try{ //get config parameters
    sPhysRange                  R;
    sPhysValue                  V;
    if( get_from<sIntegerNode>(DeviceSettings >> "NumOfPasses",NumOfPasses) )
      FldNumOfPasses->setValue(NumOfPasses);
    if( get_from<sPhysRangeNode>(DeviceSettings >> "ScanRange" ,R) )
      ScanRangeField->SetValue(R);
    if( get_from<sPhysValueNode>(DeviceSettings >> "ScanStep" ,V) )
      ScanStepField->SetValue(V);
    if( get_from<sPhysValueNode>(DeviceSettings >> "TimePerStep" ,V) )
      TimePerStepField->SetValue(V);
  }catch(...){
    SetInvalid( tr("Something wrong in configuration.") );
  }
  ScanRangeField->SetValidRange(sPhysRange(0,10500,"Oe",0.5));
  ScanStepField->SetValidRange(sPhysRange(0,10,"Oe",0.0005));
  connect(ScanStepField ,SIGNAL(Edited())
         ,this          ,SIGNAL(StepEdited()));
  connect(TimePerStepField ,SIGNAL(Edited())
         ,this             ,SIGNAL(TimePerStepEdited()));
}


void  sBursiPanScanner::Translate ()
{
  ScanRangeField->SetLabel(tr("Range"));
  ScanStepField->SetLabel(tr("Step"));
  TimePerStepField->SetLabel(tr("Time"));
  FldTheta->SetLabel(QString("Angles  ")+Symbol::Theta);
  FldPhi->SetLabel(QString("  ")+Symbol::Phi);
  LblCavity->setText(tr("Cavity"));
  NumOfPassesSuffix->setText(tr("pass(es)",0,NumOfPasses));
}


void  sBursiPanScanner::RestoreState ()
{
  sScanner::RestoreState();
  try{
    integer                     I;
    sPhysRange                  R;
    sPhysValue                  V;
    if( get_from<sIntegerNode>(WidgetConfig >> "NumOfPasses",NumOfPasses) )
      FldNumOfPasses->setValue(NumOfPasses);
    if( get_from<sIntegerNode>(WidgetConfig >> "Direction",I)
        && (0 <= I) && (I <= 3) ){
      DirectionField->setCurrentIndex(I);
    }
    if( get_from<sPhysRangeNode>(WidgetConfig >> "ScanRange" ,R) )
      ScanRangeField->SetValue(R);
    if( get_from<sPhysValueNode>(WidgetConfig >> "ScanStep" ,V) )
      ScanStepField->SetValue(V);
    if( get_from<sPhysValueNode>(WidgetConfig >> "TimePerStep" ,V) )
      TimePerStepField->SetValue(V);
    if( get_from<sPhysValueNode>(WidgetConfig >> "Theta" ,V) )
      FldTheta->SetValue(V);
    if( get_from<sPhysValueNode>(WidgetConfig >> "Phi" ,V) )
      FldPhi->SetValue(V);
    FldCavity->SetValue( WidgetConfig>>"Cavity" );
  }catch(...){
    SetInvalid( tr("Something wrong in configuration.") );
  }
}


void  sBursiPanScanner::StoreState ()
{
  sScanner::StoreState();
  if( WidgetConfig.IsCorrect() ){
    WidgetConfig
        << attr("NumOfPasses",FldNumOfPasses->value())
        << attr("Direction",DirectionField->currentIndex())
        << attr("ScanRange",ScanRangeField->Value())
        << attr("ScanStep",ScanStepField->Value())
        << attr("TimePerStep",TimePerStepField->Value())
        << attr("Theta",FldTheta->Value())
        << attr("Phi",FldPhi->Value())
        ;
    WidgetConfig.DelAttr("Cavity");
    if( ! FldCavity->Value.IsNULL() ){
      WidgetConfig << attr("Cavity",FldCavity->Value);
    }
  }
}


void sBursiPanScanner::OnEnterToOperatingState()
{
//  OnFullRangeChange(BursiPan->FieldRange());
  connect(FldNumOfPasses ,SIGNAL(valueChanged(int))
         ,this             ,SLOT(OnNumOfPassesEdited()));
  connect(ScanRangeField ,SIGNAL(Edited())
         ,this           ,SLOT(OnScanRangeEdited()) );
}


void sBursiPanScanner::OnExitFromOperatingState()
{
  disconnect(FldNumOfPasses ,SIGNAL(valueChanged(int)) ,0,0);
  disconnect(ScanRangeField ,SIGNAL(Edited()) ,0,0);
}


void sBursiPanScanner::OnEnterToCoolingState()
{
  Disconnect(BursiPan,0,0);
  BursiPan.ReleaseExclusiveControl();
  sScanner::OnEnterToCoolingState();
}


void  sBursiPanScanner::CreateNewDependence ()
{
//  sPhysPair                   M (sPhysValue("Oe"),sPhysValue("Spin/Oe2"));
  sPhysPair                   M (sPhysValue("Oe"),sPhysValue("mV"));
  DataArrayRecord = sNodePtr(new sDataArrayRecord
      (BursiPan->NumOfPoints()
      ,(BursiPan->ScanRange().From()/M.X).ToReal()
      ,(BursiPan->ScanStep()/M.X).ToReal()
      ));
  DataArrayRecord->SetMultiplier(M);
  Dependence = DataArrayRecord;
}


void sBursiPanScanner::EarlySetAttributes()
{
//  sScanner::EarlySetAttributes();
  {
    psGraphObject  GrOb = dynamic_cast<psGraphObject>(Dependence.operator->());
    if( GrOb && ! LineColorSet->Empty() )
      GrOb->SetAppearance(new sCurveAppearance(LineColorSet->CurrentColor()
                                              ,LineThickness));
  }
  Sample=FldSample->Value;
  Topic=FldTopic->Value;
  Conditions=FromQString(FldConditions->toPlainText());
  //! @todo{sString} Trim
  //if( ! Conditions.Trim().Empty() )
  if( ! Conditions.Empty() )
    Dependence << attr(Conditions);
  if( ! FldTheta->Value().IsNaN() )
    Dependence << attr("&Theta;",FldTheta->Value());
  if( ! FldPhi->Value().IsNaN() )
    Dependence << attr("&Phi;",FldPhi->Value());
  if( ! Sample.IsNULL() )
    Dependence << attr("sample",Sample);
  if( ! FldCavity->Value.IsNULL() )
    Dependence << attr("cavity",FldCavity->Value);
  Dependence << attr("&nu;",BursiPan->Frequency());
  Dependence << attr("A",BursiPan->Attenuation());
  Dependence << attr("M",BursiPan->Modulation());
  Dependence << attr("G",BursiPan->Gain());
  Dependence << attr("mode",BursiPan->cModeText[BursiPan->ModulationMode()]);
  Dependence << attr("speed",BursiPan->ScanStep()/BursiPan->TimePerStep());
  sPtr<sFolderNode>           Technique( Dependence >=
                                         folder("technique",DeviceIdentifier())
                                       );
  Technique << attr("H<sub>0</sub>",BursiPan->Ho()); //H₀
  Technique << attr("&Delta;H",BursiPan->DeltaH());  //ΔH
  Technique << attr("range",ScanRangeField->Value());
  Technique << attr("step",ScanStepField->Value());
  Technique << attr("time/step",TimePerStepField->Value());
  Technique << attr("v.",sVersion(RAP_VERS_MAJOR
                                 ,RAP_VERS_MINOR
                                 ,ZAKAVYCH(RAP_VERS_PATCH)
                                 ,__DATE__
                                 ).Text());
}


void sBursiPanScanner::LateSetAttributes()
{
  sScanner::LateSetAttributes();
  (Dependence >> "passes");
}


void  sBursiPanScanner::StartScanning ()
{
  PassNo = 1;
  PointNo = 0;
  StopRequested = false;
  SendCommand(BursiPan,sBursiPan::cScanStart);
}


void  sBursiPanScanner::StopScanning ()
{
}


//void  sBursiPanScanner::OnFullRangeChange (sPhysRange)
//{
//  if( ValidateRangeAndStep() ){
//    if( SendDirective(BursiPan,sBursiPan::cSetScanStep
//                       ,ScanStepField->Value(),cTakeControl)
//        &&  SendDirective(BursiPan,sBursiPan::cSetTimePerStep
//                       ,TimePerStepField->Value()) ){
//      Connect(BursiPan ,SIGNAL(TimePerStepChanged(sPhysValue))
//             ,SIGNAL(ParamSettingFinished()) );
//      connect(this,SIGNAL(ParamSettingFinished())
//             ,this,SLOT(OnScanRangeEdited()),Qt::QueuedConnection);
//      emit AllScanParamResetting();
//    }else{
//    }
//  }
//}


bool  sBursiPanScanner::ValidateRangeAndStep ()
{
  bool                        RangeOK = false;
  bool                        StepOK = false;
  try{
    RangeOK = (ScanRangeField->Value() <= BursiPan->FieldRange());
    StepOK = ! (ScanRangeField->Value().Size() < ScanStepField->Value());
  }catch(...){};
  ScanRangeField->SetAttention( ! RangeOK );
  ScanStepField->SetAttention( ! StepOK );
  return RangeOK && StepOK;
}


void  sBursiPanScanner::ShowParamComments ()
{
  if( ValidateRangeAndStep() ){
//    sPhysValue                  TimePerPass(BursiPan->NumOfPoints()
//                                           *BursiPan->TimePerStep());
//    sPhysValue                  ElapsedTime(NumOfPasses*TimePerPass);
//    ScanRangeField->setToolTip(ToQString(BursiPan->ScanRange().Text()));
//    ScanStepField->setToolTip(ToQString(BursiPan->ScanStep().Text()));
//    ScanStepSuffix->setText(tr(" %1 points")
//                           .arg(BursiPan->NumOfPoints()));
//    TimePerStepSuffix->setText(
//        tr(" per step, %1 per pass")
//        .arg(ToQString(
//              ( sPhysValue(100,"s") < TimePerPass
//              ? (sPhysValue("min") = TimePerPass).SetRelativeError(0.01)
//              : (sPhysValue("s")   = TimePerPass).SetRelativeError(0.01)
//              ).Text() ))
//                              );
//    if( NumOfPasses > 1 )
//      TimePerStepSuffix->setText(TimePerStepSuffix->text()
//          +tr(", %2 total")
//          .arg(ToQString(
//                ( sPhysValue(100,"min") < ElapsedTime
//                ? (sPhysValue("hour") = ElapsedTime).SetRelativeError(0.01)
//                : (sPhysValue("min")  = ElapsedTime).SetRelativeError(0.01)
//                ).Text() ))
//                                );
    sPhysValue                  TimePerPass("s/pass");
    sPhysValue                  ElapsedTime("min");
    TimePerPass = BursiPan->TimePerStep()
                // /////////////////////////
                /sPhysValue("step")
                // /////////////////////////
                * sPhysValue(BursiPan->NumOfPoints(),"step/pass");
    ElapsedTime = TimePerPass * sPhysValue(NumOfPasses,"pass");
    ScanRangeField->setToolTip(ToQString(BursiPan->ScanRange().Text()));
    ScanStepField->setToolTip(ToQString(BursiPan->ScanStep().Text()));
    ScanStepSuffix->setText(tr(" %1 points")
                           .arg(BursiPan->NumOfPoints()));
    TimePerStepSuffix->setText(
        tr("/step, %1 ")
        .arg(ToQString(
              ( sPhysValue(100,"s/pass") < TimePerPass
              ? (sPhysValue("min/pass") = TimePerPass).SetRelativeError(0.01)
              : (sPhysValue("s/pass")   = TimePerPass).SetRelativeError(0.01)
              ).Text() ))
                              );
    if( NumOfPasses > 1 )
      TimePerStepSuffix->setText(TimePerStepSuffix->text()
          +tr(", %2 total")
          .arg(ToQString(
                ( sPhysValue(100,"min") < ElapsedTime
                ? (sPhysValue("hour") = ElapsedTime).SetRelativeError(0.01)
                : (sPhysValue("min")  = ElapsedTime).SetRelativeError(0.01)
                ).Text() ))
                                );
    TimePerStepField->setToolTip(ToQString(BursiPan->TimePerStep().Text())
                                +TimePerStepSuffix->text());
    TimePerStepSuffix->setToolTip(TimePerStepField->toolTip());
  }else{
    NumOfPassesSuffix->setText("");
    ScanRangeField->setToolTip("");
    ScanStepField->setToolTip("");
    ScanStepSuffix->setText("");
    TimePerStepField->setToolTip("");
    TimePerStepSuffix->setText("");
    TimePerStepSuffix->setToolTip("");
  }
}


void  sBursiPanScanner::EnableParamSetting ()
{
  DirectionField->setEnabled(true);
  ScanRangeField->UnsetReadOnly();
  ScanStepField->UnsetReadOnly();
  TimePerStepField->UnsetReadOnly();
}


void  sBursiPanScanner::DisableParamSetting ()
{
  DirectionField->setEnabled(false);
  ScanRangeField->SetReadOnly();
  ScanStepField->SetReadOnly();
  TimePerStepField->SetReadOnly();
}


void  sBursiPanScanner::OnNumOfPassesEdited ()
{
  NumOfPasses = FldNumOfPasses->value();
  NumOfPassesSuffix->setText(tr("pass(es)",0,NumOfPasses));
  ShowParamComments();
}


void  sBursiPanScanner::OnScanRangeEdited ()
{
  BursiPan->SetScanRange( ScanRangeField->Value() );
//  ScanRange = BursiPan->ScanRange();
//  ShowParamComments();
}


void  sBursiPanScanner::OnStepEdited ()
{
  if( SendDirective(BursiPan,sBursiPan::cSetScanStep
                   ,ScanStepField->Value(),cTakeControl) ){
    emit StepSetting();
  }else{
    emit ErrorMessage(tr("Cannot change, because BursiPan is busy."
                         "\nTry again later.") );
    emit BursiPanBusy();
  }
}


void  sBursiPanScanner::OnTimePerStepEdited ()
{
  if( SendDirective(BursiPan,sBursiPan::cSetTimePerStep
                     ,TimePerStepField->Value(),cTakeControl) ){
    emit TimePerStepSetting();
  }else{
    emit ErrorMessage(tr("Cannot change, because BursiPan is busy."
                         "\nTry again later.") );
    emit BursiPanBusy();
  }
}


void  sBursiPanScanner::AddNewData (sPhysValue  data)
{
  real  Y = DataArrayRecord->PointY(PointNo);
  Y = (PassNo-1)*Y;
  Y += (data/DataArrayRecord->Multiplier().Y).ToReal();
  Y /= PassNo;
  TRACE(Y);
  DataArrayRecord->SetValue(PointNo,Y);
  PointNo++;
  if( PointNo > BursiPan->NumOfPoints() - 1 ){
    if( StopRequested || (PassNo >= NumOfPasses) ){
      emit ScanningHasCompleted();
    }else{
      PassNo++;
      PointNo = 0;
      SendCommand(BursiPan,sBursiPan::cScanStart);
    }
  }
}


void  sBursiPanScanner::OnBreak ()
{
  QMessageBox                 D;
  QAbstractButton *           BreakBtn;
  QAbstractButton *           StopBtn = NULL;
  D.setWindowTitle(tr("Break requested"));
  D.setIcon(QMessageBox::Question);
  D.setText(tr("What whould you like to do?"));
//  D.setDetailedText(tr("........"));
  BreakBtn = D.addButton(tr("Break right NOW")
                        ,QMessageBox::NoRole);
  if( NumOfPasses > 1 ){
    StopBtn = D.addButton(tr("Stop at the end of the pass")
                         ,QMessageBox::NoRole);
  }
  D.exec();
  if( D.clickedButton() == BreakBtn ){
    SendCommand(BursiPan,sBursiPan::cScanBreak);
    emit ScanningInterrupted();
  }else if( D.clickedButton() == StopBtn ){
    StopRequested = true;
  }
}


void  sBursiPanScanner::ReportBursiPanError (QString  error_text)
{
  emit ErrorMessage( tr("Got error from '%1':\n\n%2")
                       .arg(ToQString(BursiPan->DeviceIdentifier()))
                       .arg(error_text)     );
}


void  sBursiPanScanner::ReleaseBursiPanControl ()
{
  BursiPan.ReleaseExclusiveControl();
}


void  sBursiPanScanner::LinkToBursiPan ()
{
  try{
    BursiPan.SetTo(DeviceLinks,"BursiPan");
    Connect(BursiPan ,SIGNAL(Warmed()) ,SIGNAL(BursiPanLinked()) );
    Connect(BursiPan ,SIGNAL(ErrorMessage(QString))
                     ,SIGNAL(BursiPanError(QString)) );
    Connect(BursiPan ,SIGNAL(RangeChanged(sPhysRange))
                     ,SIGNAL(BursiPanRangeChanged(sPhysRange)) );
    Connect(BursiPan ,SIGNAL(StepChanged(sPhysValue))
                     ,SIGNAL(StepChanged(sPhysValue)) );
    Connect(BursiPan ,SIGNAL(TimePerStepChanged(sPhysValue))
                     ,SIGNAL(TimePerStepChanged(sPhysValue)) );
//    Connect(BursiPan ,SIGNAL(ScanStarts(sBursiPan::sScanStartData))
//                     ,SIGNAL(ScanStarted(sBursiPan::sScanStartData)) );
//    Connect(BursiPan ,SIGNAL(ScanPoint(sBursiPan::sScanPointData))
//                     ,SIGNAL(GotScanPoint(sBursiPan::sScanPointData)) );
    Connect(BursiPan ,SIGNAL(EPR_Signal(sPhysValue))
                     ,SIGNAL(EPR_Signal(sPhysValue)) );
    Connect(BursiPan ,SIGNAL(ScanDone())
                     ,SIGNAL(ScanningHasCompleted()) );
  }catch( rxDeviceLaunch E ){
    emit ErrorMessage( tr("Error while linking to a submissive device.\n"
                          "%1\nCheck configuration.")
                         .arg(ToQString(E.MsgText()))   );
  }
}

//------------------------------------------------------------------------------
} //namespace RA




