////////////////////////////////////////////////////////////////////////////////
/*! @file BursiPan_MSDN.cpp BursiPan implementation
                            with field control through MSDNx48 and
                            with signal and handles from MSDN2047.
- Part of PanBursian - Research Assistant Plugin package.
- Uses  RAGUI - Research Assistant Graphical User Interface Library.
- Uses  RANet - Research Assistant Net Library.
- Uses  Qt v.5  -  http://qt-project.org/
- Copyright(C) 2014, Viktor E. Bursian, St.Petersburg, Russia.
                     Viktor.Bursian@mail.ioffe.ru
*///////////////////////////////////////////////////////////////////////////////
#include "BursiPan_MSDN.h"
#include "attr.h"
#include "StringsAndQStrings.h"
#include "Log.h"
#include <limits>
namespace RA {
//------------------------------------------------------------------------------

DEFINE_CLASS_TAG(sBursiPan_MSDN)

const sPhysValue   sBursiPan_MSDN::PeriodConversion(1,"ns");

//----------------------------------------------------------- sBursiPan_MSDN ---

void  sBursiPan_MSDN::Preconfigure (sNodePtr  DeviceCfg)
{
  sPtr<sFolderNode>              DeviceLinks;
  DeviceCfg |= "Links";
  DeviceLinks = DeviceCfg >> "Links";
  if( DeviceLinks.IsCorrect() ){
    DeviceLinks |= "MSDNx48";
    DeviceLinks >> "MSDNx48" |= attr("Class","sTeleCom");
    DeviceLinks |= "MSDNx47";
    DeviceLinks >> "MSDNx47" |= attr("Class","sTeleCom");
  }
  sPtr<sFolderNode>              DeviceSettings;
  DeviceCfg |= "Settings";
  DeviceSettings = DeviceCfg >> "Settings";
  if( DeviceSettings.IsCorrect() ){
    DeviceSettings |= attr("Emulate",false);
//    DeviceSettings |= attr("Calibration",sPhysValue(1.0,"Spin/Oe/Oe"));
    DeviceSettings |= attr("Calibration",sPhysValue() =
                                         sPhysValue(2.4414,"mV"));
    DeviceSettings |= attr("ZeroCorrection",sPhysValue() =
                                            sPhysValue(0,"mV"));
    DeviceSettings |= attr("Ho",sPhysValue(0.0,0.1,"Oe"));
    DeviceSettings |= attr("DeltaH",sPhysValue(0.0,0.1,"Oe"));
    DeviceSettings |= attr("Frequency",sPhysValue(8.8e9,0.1e9,"Hz"));
    DeviceSettings |= attr("Attenuation",sPhysValue(30,0.1));
  }
  sBursiPan::Preconfigure(DeviceCfg);
}


sBursiPan_MSDN::~sBursiPan_MSDN ()
{
  if( DeviceSettings.IsCorrect() ){
    DeviceSettings
        << attr("Ho",LatestHo)
        << attr("DeltaH",LatestDeltaH)
        << attr("Frequency",LatestFrequency)
        << attr("Attenuation",LatestAttenuation)
        ;
  }
  MSDNx48.SetToNULL();
  MSDNx47.SetToNULL();
}


sBursiPan_MSDN::sBursiPan_MSDN (sString   DevicetId
                               ,sNodePtr  DeviceCfg)
    :sBursiPan(DevicetId,DeviceCfg)
//    ,Calibration("Spin/Oe/Oe")
    ,Calibration()
    ,ZeroCorrection()
    ,MSDNx48(this)
    ,MSDNx47(this)
    ,LatestHo("Oe")
    ,LatestDeltaH("Oe")
    ,LatestFrequency("Hz")
    ,LatestAttenuation(_Unitsless_)
    ,GainMantissaIndex(-1)
    ,GainOrder(-1)
    ,ModulationMantissaIndex(-1)
    ,ModulationOrder(-1)
    ,TheModulationMode(cModUnknown)
    ,CurrentFieldIndex(0)
    ,ScanSettingsOKFlag(false)
    ,TimePerFieldQuantum(0)
    ,StepInQuanta(1)
    ,FromIndex(0)
    ,ToIndex(0)
    ,TheNumOfPoints(1)
    ,Emulate(false)
{
  Calibration = sPhysValue(2.4414,"mV");
  ZeroCorrection = sPhysValue(-46.2,"mV");
}


void  sBursiPan_MSDN::PostConstructor ()
{
  sBursiPan::PostConstructor();
  try{ //get config parameters
    get_from<sBooleanNode>(DeviceSettings>>"Emulate",Emulate);
    get_from<sPhysValueNode>(DeviceSettings>>"Calibration",Calibration);
    get_from<sPhysValueNode>(DeviceSettings>>"ZeroCorrection",ZeroCorrection);
    get_from<sPhysValueNode>(DeviceSettings>>"Ho",LatestHo);
    get_from<sPhysValueNode>(DeviceSettings>>"DeltaH",LatestDeltaH);
    get_from<sPhysValueNode>(DeviceSettings>>"Frequency",LatestFrequency);
    get_from<sPhysValueNode>(DeviceSettings>>"Attenuation",LatestAttenuation);
    (Calibration/ZeroCorrection).ToReal();
  }catch(...){
    SetInvalid( tr("Something wrong in configuration.") );
  }
  //evolution of the inherited state machine
  {
    //evolution of the State_Warming
    {
      State_W_Parallel = new sParallelStateGroup(State_Warming());
      State_W_Parallel
          ->SetAsInitial()
          ->AddTransition(State_Warming()->Fin());
      //State_W_Parallel internal structure and behaviour
      {
        State_W_MSDNx48 = new sState(State_W_Parallel,"State_W_MSDNx48");
        //State_W_MSDNx48 structure and behaviour
        {
          State_W_MSDNx48_Linking = new sState(State_W_MSDNx48
                                            ,"State_W_MSDNx48_Linking");
          State_W_MSDNx48_Checking = new sState(State_W_MSDNx48
                                             ,sPhysValue(15,"min")
                                             ,"State_W_MSDNx48_Checking");

          State_W_MSDNx48_Linking
              ->SetAsInitial()
              ->OnEnterCall( SLOT(LinkToMSDNx48()) )
              ->AddTransition( SIGNAL(MSDNx48Linked()) ,State_W_MSDNx48_Checking )
              ->AddTransition( SIGNAL(ErrorMessage()) ,State_Cooling() )
              ;
          State_W_MSDNx48_Checking
              ->OnEnterCall( SLOT(GetMSDNx48Version()) )
              ->AddReaction<sTelex>( SIGNAL(GotMSDNx48Message(sTelex))
                                   , SLOT(ProcessMSDNx48Signature(sTelex))  )
              ->AddTransition( SIGNAL(NoAnswerFromMSDNx48())
                             , SLOT(OnNoAnswerFromMSDNx48())
                             , State_W_MSDNx48_Checking )
//              ->AddTransition( SIGNAL(MSDNx48Error()) ,State_Cooling() )
              ->AddTransition(SIGNAL(GotMSDNx48Version())
                             ,State_W_MSDNx48->Fin()       )
              ;
          connect(State_W_MSDNx48_Checking,SIGNAL(Timeout())
                 ,this,SIGNAL(NoAnswerFromMSDNx48()));

        State_W_MSDNx47 = new sState(State_W_Parallel,"State_W_MSDNx47");
        //State_W_MSDNx47 structure and behaviour
        {
          State_W_MSDNx47_Linking = new sState(State_W_MSDNx47
                                            ,"State_W_MSDNx47_Linking");
          State_W_MSDNx47_Checking = new sState(State_W_MSDNx47
                                             ,sPhysValue(15,"min")
                                             ,"State_W_MSDNx47_Checking");

          State_W_MSDNx47_Linking
              ->SetAsInitial()
              ->OnEnterCall( SLOT(LinkToMSDNx47()) )
              ->AddTransition( SIGNAL(MSDNx47Linked()) ,State_W_MSDNx47_Checking )
              ->AddTransition( SIGNAL(ErrorMessage()) ,State_Cooling() )
              ;
          State_W_MSDNx47_Checking
              ->OnEnterCall( SLOT(GetMSDNx47Version()) )
              ->AddReaction<sTelex>( SIGNAL(GotMSDNx47Message(sTelex))
                                   , SLOT(ProcessMSDNx47Signature(sTelex))  )
              ->AddTransition( SIGNAL(NoAnswerFromMSDNx47())
                             , SLOT(OnNoAnswerFromMSDNx47())
                             , State_W_MSDNx47_Checking )
//              ->AddTransition( SIGNAL(MSDNx47Error()) ,State_Cooling() )
              ->AddTransition(SIGNAL(GotMSDNx47Version())
                             ,State_W_MSDNx47->Fin()       )
              ;
          }
          connect(State_W_MSDNx47_Checking,SIGNAL(Timeout())
                 ,this,SIGNAL(NoAnswerFromMSDNx47()));
        }
      }
    }
    //evolution of the State_Operating
    {
      State_Operating()
          ->AddReaction<sTelex>( SIGNAL(GotMSDNx48Message(sTelex))
                               , SLOT(ProcessMSDNx48Message(sTelex))  )
          ;
      State_Operating()
          ->AddReaction<sTelex>( SIGNAL(GotMSDNx47Message(sTelex))
                               , SLOT(ProcessMSDNx47Message(sTelex))  )
          ;

      State_O_Idle = new sState(State_Operating(),"State_O_Idle");
      State_O_Scanning = new sState(State_Operating(),"State_O_Scanning");

      State_O_Idle
          ->SetAsInitial()
          ->AddReaction(cSetHo,SLOT(SetHo(sPhysValue)))
          ->AddReaction(cSetDeltaH,SLOT(SetDeltaH(sPhysValue)))
          ->AddReaction(cSetModulation,SLOT(SetModulation(sPhysValue)))
          ->AddReaction(cSetGain,SLOT(SetGain(sPhysValue)))
          ->AddReaction(cSetAttenuation,SLOT(SetAttenuation(sPhysValue)))
          ->AddReaction(cSetFrequency,SLOT(SetFrequency(sPhysValue)))
          ->AddReaction(cSetModulationMode
                       ,SLOT(SetModulationMode(eModulationMode)))

          ->AddReaction(cSetEPRMeasurePeriod
                       ,SLOT(SetWantedEPRMeasurePeriod(sPhysValue)))
          ->AddReaction(cSetMagneticField,SLOT(SetMagneticField(sPhysValue)))

          ->AddReaction(cSetScanStep,SLOT(SetScanStep(sPhysValue)))
          ->AddReaction(cSetTimePerStep,SLOT(SetTimePerStep(sPhysValue)))

          ->AddTransition(cScanStart ,State_O_Scanning)
          ;
      State_O_Scanning
          ->AddReaction(cScanBreak,SLOT(BreakScan()))
          ->AddTransition(SIGNAL(ScanBroken()) ,State_O_Idle)
          ->AddTransition(State_O_Idle)
//          ->AddTransition(SIGNAL(ScanDone()) ,State_O_Idle)
          ;
      //State_O_Scanning internal structure and behaviour
      {
        State_O_Sc_Prepare = new sParallelStateGroup(State_O_Scanning);
        State_O_Sc_Stepping = new sState(State_O_Scanning
                                        ,"State_O_Sc_Stepping");
        State_O_Sc_Prepare
            ->SetAsInitial()
            ->AddTransition(State_O_Sc_Stepping);
        //State_O_Sc_Prepare internal structure and behaviour
        {
          State_O_Sc_SettingInitialField = new sState(State_O_Sc_Prepare
                                             ,"State_O_Sc_SettingInitialField");
          //State_O_Sc_SettingInitialField internal structure and behaviour
          {
            State_O_Sc_SIF_Set = new sState(State_O_Sc_SettingInitialField
                                           ,"State_O_Sc_SIF_Set");
            State_O_Sc_SIF_Pause = new sState(State_O_Sc_SettingInitialField
                                            ,sPhysValue(10,"s")
                                            ,"State_O_Sc_SIF_Pause");
            State_O_Sc_SIF_Set
                ->SetAsInitial()
                ->OnEnterCall(SLOT(SetInitialField()))
                ->AddTransition(SIGNAL(MagneticField(sPhysValue))
                               ,State_O_Sc_SIF_Pause)
                ;
            State_O_Sc_SIF_Pause
                ->AddTransition(State_O_Sc_SIF_Pause,SIGNAL(Timeout())
                               ,State_O_Sc_SettingInitialField->Fin())
                ;
          }
          State_O_Sc_SettingGauge = new sState(State_O_Sc_Prepare
                                              ,"State_O_Sc_SettingGauge");
          //State_O_Sc_SettingGauge internal structure and behaviour
          {
            State_O_Sc_SG_Interval = new sState(State_O_Sc_SettingGauge
                                           ,"State_O_Sc_SG_Interval");
            State_O_Sc_SG_Start = new sState(State_O_Sc_SettingGauge
                                            ,sPhysValue(10,"s")
                                            ,"State_O_Sc_SG_Start");
            State_O_Sc_SG_Interval
                ->SetAsInitial()
                ->OnEnterCall(SLOT(SetAveragingInterval()))
                ->AddTransition(SIGNAL(EPRMeasurePeriodChanged(sPhysValue))
                               ,State_O_Sc_SG_Start)
                ;
            State_O_Sc_SG_Start
                ->OnEnterCall(SLOT(MakeAveragingReady()))
                ->AddTransition(SIGNAL(AveragingIsReady())
                               ,State_O_Sc_SettingGauge->Fin())
                ;
          }
        }
        State_O_Sc_Stepping
            ->OnEnterCall(SLOT(StartScan()))
            ->AddTransition(SIGNAL(SteppingDone())
                           ,State_O_Scanning->Fin())
            ;
      }
    }
    //evolution of the State_Cooling
    {
    }
    connect(this ,SIGNAL(MSDNx48Error(QString))
           ,this ,SLOT(ReportMSDNx48Error(QString)) );
    connect(this ,SIGNAL(MSDNx47Error(QString))
           ,this ,SLOT(ReportMSDNx47Error(QString)) );
//    connect(this ,SIGNAL(NewHo(sPhysValue))
//           ,this ,SLOT(SignalRangeChanging(sPhysValue))
//           ,Qt::QueuedConnection );
//    connect(this ,SIGNAL(NewDeltaH(sPhysValue))
//           ,this ,SLOT(SignalRangeChanging(sPhysValue))
//           ,Qt::QueuedConnection );
    connect(this ,SIGNAL(NewHo(sPhysValue))
           ,this ,SLOT(SignalRangeChanging())
           ,Qt::QueuedConnection );
    connect(this ,SIGNAL(NewDeltaH(sPhysValue))
           ,this ,SLOT(SignalRangeChanging())
           ,Qt::QueuedConnection );
  }
}


//void  sBursiPan_MSDN::AfterStartInitiation ()
//{
//}


void  sBursiPan_MSDN::OnEnterToOperatingState ()
{
  sBursiPan::OnEnterToOperatingState();
  SendDirective( MSDNx48 , sTeleCom::cSendTelex , sTelex("/") );
  SendDirective( MSDNx48 , sTeleCom::cSendTelex , sTelex("F0") );
  SendDirective( MSDNx47 , sTeleCom::cSendTelex , sTelex("/") );
  SendDirective( MSDNx47 , sTeleCom::cSendTelex , sTelex("H") );
  SendDirective( MSDNx47 , sTeleCom::cSendTelex , sTelex("S") );
}


void  sBursiPan_MSDN::OnExitFromOperatingState ()
{
  //...
  SendDirective( MSDNx48 , sTeleCom::cSendTelex , sTelex("/") );
  SendDirective( MSDNx47 , sTeleCom::cSendTelex , sTelex("/") );
  sBursiPan::OnExitFromOperatingState();
}


void  sBursiPan_MSDN::OnEnterToCoolingState ()
{
  Disconnect(MSDNx48 ,0 ,0);
  MSDNx48.ReleaseExclusiveControl();
  MSDNx48.SetToNULL();
  Disconnect(MSDNx47 ,0 ,0);
  MSDNx47.ReleaseExclusiveControl();
  MSDNx47.SetToNULL();
  sBursiPan::OnEnterToCoolingState();
}


void  sBursiPan_MSDN::ReportMSDNx48Error (QString  error_text)
{
  emit ErrorMessage(tr("Got error from Field Control Unit Microcontroller %1:"
                       "\n\n%2")
                      .arg(ToQString(MSDNx48->DeviceIdentifier()))
                      .arg(error_text)     );
}


void  sBursiPan_MSDN::OnNoAnswerFromMSDNx48 ()
{
  emit ErrorMessage
          ("No answer from MSDN2048 (Field Control Unit Microcontroller");
}


void  sBursiPan_MSDN::ReportMSDNx47Error (QString  error_text)
{
  emit ErrorMessage(tr("Got error from Registration Unit Microcontroller %1:"
                       "\n\n%2")
                      .arg(ToQString(MSDNx47->DeviceIdentifier()))
                    .arg(error_text)     );
}


void  sBursiPan_MSDN::OnNoAnswerFromMSDNx47 ()
{
  emit ErrorMessage
          ("No answer from MSDN2047 (Registration Unit Microcontroller");
}


void  sBursiPan_MSDN::LinkToMSDNx48 ()
{
  try{
    MSDNx48.SetTo(DeviceLinks,"MSDNx48");
    if( ! MSDNx48.TakeExclusiveControl() ){
      emit ErrorMessage(tr("Cannot take exclusive control over '%1'. "
                           "Try again later.")
                       .arg(ToQString( MSDNx48->DeviceIdentifier() )) );
    }else{
      Connect(MSDNx48 ,SIGNAL(Warmed()) ,SIGNAL(MSDNx48Linked()) );
      Connect(MSDNx48 ,SIGNAL(ErrorMessage(QString))
             ,SIGNAL(MSDNx48Error(QString)) );
      Connect(MSDNx48 ,SIGNAL(GotTelex(sTelex))
             ,SIGNAL(GotMSDNx48Message(sTelex)) );
    }
  }catch( rxDeviceLaunch E ){
    emit ErrorMessage( tr("Error while linking to a submissive device.\n"
                          "%1\nCheck configuration.")
                         .arg(ToQString(E.MsgText()))   );
  }
}

void  sBursiPan_MSDN::LinkToMSDNx47 ()
{
  try{
    MSDNx47.SetTo(DeviceLinks,"MSDNx47");
    if( ! MSDNx47.TakeExclusiveControl() ){
      emit ErrorMessage(tr("Cannot take exclusive control over '%1'. "
                           "Try again later.")
                       .arg(ToQString( MSDNx47->DeviceIdentifier() )) );
    }else{
      Connect(MSDNx47 ,SIGNAL(Warmed()) ,SIGNAL(MSDNx47Linked()) );
      Connect(MSDNx47 ,SIGNAL(ErrorMessage(QString))
             ,SIGNAL(MSDNx47Error(QString)) );
      Connect(MSDNx47 ,SIGNAL(GotTelex(sTelex))
             ,SIGNAL(GotMSDNx47Message(sTelex)) );
    }
  }catch( rxDeviceLaunch E ){
    emit ErrorMessage( tr("Error while linking to a submissive device.\n"
                          "%1\nCheck configuration.")
                         .arg(ToQString(E.MsgText()))   );
  }
}


void  sBursiPan_MSDN::GetMSDNx48Version ()
{
  SendDirective(MSDNx48  ,sTeleCom::cSendTelex ,sTelex("?") );
}


void  sBursiPan_MSDN::GetMSDNx47Version ()
{
  SendDirective(MSDNx47  ,sTeleCom::cSendTelex ,sTelex("?") );
}


void  sBursiPan_MSDN::SetWantedEPRMeasurePeriod (sPhysValue  period)
{
  real                          Period;
  quint32                       P;
  try{
    Period = (period/PeriodConversion).ToReal();
    if( Period < 1 ){
      P = 1;
    }else if( Period > std::numeric_limits<quint32>::max() ){
      P = std::numeric_limits<quint32>::max();
    }else{
      P = Period;
    }
    sTelex                      T("P");
    T.append(QByteArray::number(P,16).toUpper());
    SendDirective(MSDNx47  ,sTeleCom::cSendTelex ,T );
  }catch(...){

  }
}


void  sBursiPan_MSDN::SetMagneticField (sPhysValue  value)
{
  try{
    sTelex                      T("F");
    T.append(QByteArray::number(FieldIndex(value), 16).toUpper());
    SendDirective(MSDNx48  ,sTeleCom::cSendTelex ,T );
  }catch(...){

  }
}


void  sBursiPan_MSDN::SetHo (sPhysValue  value)
{
  LatestHo = value;
  emit NewHo(LatestHo);
}


void  sBursiPan_MSDN::SetDeltaH (sPhysValue  value)
{
  LatestDeltaH = value;
  emit NewDeltaH(LatestDeltaH);
}


void  sBursiPan_MSDN::SetFrequency (sPhysValue  value)
{
  LatestFrequency = value;
  emit NewFrequency(LatestFrequency);
}


void  sBursiPan_MSDN::SetAttenuation (sPhysValue  value)
{
  LatestAttenuation = value;
  emit NewAttenuation(LatestAttenuation);
}


void  sBursiPan_MSDN::SetGain (sPhysValue)
{
  emit ErrorMessage("Unsupported cSetGain directive");
}


void  sBursiPan_MSDN::SetModulation (sPhysValue)
{
  emit ErrorMessage("Unsupported cSetModulation directive");
}


void  sBursiPan_MSDN::SetModulationMode (eModulationMode)
{
  emit ErrorMessage("Unsupported cSetModulationMode directive");
}


void  sBursiPan_MSDN::ProcessMSDNx48Signature (sTelex  text)
{
  if( text.startsWith("@MSDN2048") ){
    MSDNx48_Version = text.mid(1);
    emit GotMSDNx48Version();
  }else if( text.startsWith("@MSDN") ){
    emit MSDNx48Error(QString("Wrong card at the given port: ")+text.mid(1));
  }
}


void  sBursiPan_MSDN::ProcessMSDNx48Message (sTelex  text)
{
  QString                     ConversionErr(tr("Got '%1'. Cannot convert."));
  bool                        OK = true;
  qint32                      Reading;
  if( text.startsWith('F') ){
    Reading = text.mid(1).toLongLong(&OK,16);
    if( ! OK ){
      emit MSDNx48Error(ConversionErr.arg(QString::fromLatin1(text)));
    }else{
      CurrentFieldIndex = Reading;
      emit MagneticField( TrueField(Reading) );
    }

  }else if( text.startsWith('S') ){
    Reading = text.mid(1).toLongLong(&OK,16);
    if( ! OK ){
      emit MSDNx48Error(ConversionErr.arg(QString::fromLatin1(text)));
    }else{
      StepInQuanta = Reading;
      RecalculateNumOfPoints();
      emit StepChanged( ScanStep() );
    }

  }else if( text.startsWith('T') ){
    quint32                     UReading;
    UReading = text.mid(1).toULongLong(&OK,16);
    if( ! OK ){
      emit MSDNx48Error(ConversionErr.arg(QString::fromLatin1(text)));
    }else{
      TimePerFieldQuantum = UReading;
      emit TimePerStepChanged( TimePerStep() );
    }

  }else if( text == "." ){
    emit SteppingDone();

  }else if( text == "b" ){
    emit ScanBroken();

  }else if( text.startsWith("GoTo") ){

  }else if( text == "g" ){

  }else if( text.startsWith("@MSDN2048") ){
    MSDNx48_Version = text.mid(1);
    emit GotMSDNx48Version();

  }else if( text == "i" ){

  }else{
    emit MSDNx48Error(QString("Unknown telex: ")+text);
  }
}


void  sBursiPan_MSDN::ProcessMSDNx47Signature (sTelex  text)
{
  if( text.startsWith("@MSDN2047") ){
    MSDNx47_Version = text.mid(1);
    emit GotMSDNx47Version();
  }else if( text.startsWith("@MSDN") ){
    emit MSDNx47Error(QString("Wrong card at the given port: ")+text.mid(1));
  }
}


void  sBursiPan_MSDN::ProcessMSDNx47Message (sTelex  text)
{
  QString                     ConversionErr(tr("Got '%1'. Cannot convert."));
  bool                        OK = true;
  qint32                      Reading;
  if(  text.startsWith('V') ){
    real                        V = 1.0;
    int                         P;
    if( (P = text.indexOf('/',2)) > 0 ){
      Reading = text.mid(P+1).toLongLong(&OK,16);
      if( ! OK ){
        emit MSDNx47Error(ConversionErr.arg(QString::fromLatin1(text)));
      }else{
        V /= Reading;
      }
    }else{
      P = -1;
    }
    if( OK ){
      Reading = text.mid(1,P-1).toLongLong(&OK,16);
      if( ! OK ){
        emit MSDNx47Error(ConversionErr.arg(QString::fromLatin1(text)));
      }else{
        V *= Reading;
        emit EPR_Signal( V * Calibration - ZeroCorrection );
      }
    }

  }else if(  text.startsWith('H') ){
    bool                        OK = false;
    int                         Handles = text.mid(1).toInt(&OK,16);
    if( ! OK || (Handles < 0) || (Handles > 0xFFFFFF) ){
      emit MSDNx47Error(QString("Wrong telex: ")+text);
    }else{
      ParseHandles(Handles);
    }

  }else if(  text.startsWith('P') ){
    Reading = text.mid(1).toLongLong(&OK,16);
    if( ! OK ){
      emit MSDNx47Error(ConversionErr.arg(QString::fromLatin1(text)));
    }else{
      emit EPRMeasurePeriodChanged( real(Reading) * PeriodConversion );
    }

  }else if(  text.startsWith('o') ){

  }else if(  text.startsWith('D') ){

  }else if(  text.startsWith('f') ){

  }else if(  text.startsWith('p') ){

  }else if(  text.startsWith('i') ){

  }else if(  text.startsWith('I') ){

  }else if(  text == "Yr" ){
    emit AveragingIsReady();

  }else if(  text == "Ys" ){
//    emit AveragingIsReady();

  }else if(  text == "Yf" ){
//    emit AveragingIsReady();

  }else if(  text == "y" ){

  }else if(  text.startsWith("@MSDN2047") ){
    MSDNx47_Version = text.mid(1);
    emit GotMSDNx47Version();
  }else{
    emit MSDNx47Error(QString("Unknown telex: ")+text);
  }
}


void  sBursiPan_MSDN::SignalRangeChanging ()
{
  emit RangeChanged( FieldRange() );
}


void  sBursiPan_MSDN::ParseHandles (int  handles)
{
  int                         Im;
  int                         Io;
  Im = handles & 0xF;
  Io = (handles & 0x30) >> 4;
  if( (Im != ModulationMantissaIndex) || (Io != ModulationOrder) ){
    ModulationMantissaIndex = Im;
    ModulationOrder = Io;
    emit NewModulation(Modulation());
  }
  Im = (handles & 0x3C0) >> 6;
  Io = (handles & 0xC00) >> 10;
  if( handles & 0x1000 )  Io++;
  if( (Im != GainMantissaIndex) || (Io != GainOrder) ){
    GainMantissaIndex = Im;
    GainOrder = Io;
    emit NewGain(Gain());
  }
  eModulationMode              Mode;
  switch( (handles & 0x3E000) >> 11  ){
    case 0x40:  Mode=cMod12kHz;  break;
    case 0x44:  Mode=cMod25kHz;  break;
    case 0x48:  Mode=cMod50kHz;  break;
    case 0x4C:  Mode=cMod100kHz;  break;
    case 0x20:
    case 0x24:
    case 0x28:
    case 0x2C:  Mode=cMod80Hz;    break;
    case 0x60:  Mode=cMod80Hz12kHz;  break;
    case 0x64:  Mode=cMod80Hz25kHz;  break;
    case 0x68:  Mode=cMod80Hz50kHz;  break;
    case 0x6C:  Mode=cMod80Hz100kHz;  break;
    case 0x50:  Mode=cMod35Hz12kHz;  break;
    case 0x54:  Mode=cMod35Hz25kHz;  break;
    case 0x58:  Mode=cMod35Hz50kHz;  break;
    case 0x5C:  Mode=cMod35Hz100kHz;  break;
    default:    Mode=cModUnknown;
  }
  if( Mode != TheModulationMode ){
    TheModulationMode = Mode;
    emit NewModulationMode(Mode);
  }
}


sPhysValue  sBursiPan_MSDN::Modulation ()
{
  real   M = 10 * power10(real( - ModulationOrder
                                - 0.1*ModulationMantissaIndex ));
  return sPhysValue( M , M*0.08 , "Oe" );
}


sPhysValue  sBursiPan_MSDN::Gain ()
{
  real   G = 50 * power10(real( GainOrder + 0.1*GainMantissaIndex ));
  return sPhysValue( G , G*0.1 , _Unitsless_ );
}


sPhysRange  sBursiPan_MSDN::FieldRange ()
{
  return sPhysRange(TrueHo()-TrueDeltaH()/2 ,TrueHo()+TrueDeltaH()/2);
}


void  sBursiPan_MSDN::SetScanRange (sPhysRange  range)
{
  ScanSettingsOKFlag = false;
  try{
    FromIndex = FieldIndex(range.From());
    ToIndex = FieldIndex(range.To());
    RecalculateNumOfPoints();
    ScanSettingsOKFlag = true;
  }catch(...){}
}


sPhysRange  sBursiPan_MSDN::ScanRange ()
{
  return sPhysRange( TrueField(FromIndex) , TrueField(ToIndex) );
}


sPhysValue  sBursiPan_MSDN::ScanStep ()
{
  return StepInQuanta * FieldQuantum();
}


void  sBursiPan_MSDN::SetScanStep (sPhysValue  value)
{
  try{
    real                S = (value / FieldQuantum()).ToReal();
    if( S > std::numeric_limits<qint8>::max() ){
      StepInQuanta = std::numeric_limits<qint8>::max();
    }else if( S < std::numeric_limits<qint8>::min() ){
      StepInQuanta = std::numeric_limits<qint8>::min();
    }else{
      StepInQuanta = Round(S);
      if( StepInQuanta == 0 )
        StepInQuanta = ( S > -0.01 ? 1 : -1);
    }
    sTelex                      T("S");
    T.append(QByteArray::number(StepInQuanta, 16).toUpper());
    SendDirective(MSDNx48  ,sTeleCom::cSendTelex ,T );
  }catch(...){}
}


void  sBursiPan_MSDN::SetTimePerStep (sPhysValue  value)
{
  try{
    TimePerFieldQuantum = Round( (value / sPhysValue("ns")).ToReal()
                               / std::abs(StepInQuanta) );
    sTelex                      T("T");
    T.append(QByteArray::number(TimePerFieldQuantum, 16).toUpper());
    SendDirective(MSDNx48  ,sTeleCom::cSendTelex ,T );
  }catch(...){}
}


void sBursiPan_MSDN::SetInitialField()
{
  real P = std::abs(( (FromIndex-CurrentFieldIndex)*FieldQuantum()
                                               /sPhysValue("Oe") ).ToReal());
  State_O_Sc_SIF_Pause->SetTimeout(sPhysValue(P+100,"ms"));
  SendDirective(MSDNx48  ,sTeleCom::cSendTelex
               ,sTelex(QByteArray::number(FromIndex, 16).toUpper()
                      .prepend("F") )
               );
}


void  sBursiPan_MSDN::SetAveragingInterval ()
{
  SendDirective(MSDNx47  ,sTeleCom::cSendTelex
//               ,sTelex(QByteArray::number(FromIndex, 16).toUpper()
//                      .prepend("M") )
               ,sTelex("P7FFFFFFF")
               );
}

void  sBursiPan_MSDN::MakeAveragingReady ()
{
  SendDirective(MSDNx47  ,sTeleCom::cSendTelex
               ,sTelex(QByteArray::number(TheNumOfPoints, 16).toUpper()
                      .prepend("Y") )
               );
}


void  sBursiPan_MSDN::RecalculateNumOfPoints ()
{
  TheNumOfPoints = (ToIndex-FromIndex)/StepInQuanta + 1;
  if( TheNumOfPoints < 1 ){
    TheNumOfPoints = 1;
    ToIndex = FromIndex;
  }else{
    ToIndex =  FromIndex + (TheNumOfPoints-1)*StepInQuanta;
    if( ToIndex < 0 ){
      ToIndex += StepInQuanta;
      TheNumOfPoints--;
    }else if( ToIndex > 4095 ){
      ToIndex -= StepInQuanta;
      TheNumOfPoints--;
    }
  }
}


void  sBursiPan_MSDN::StartScan ()
{
  SendDirective(MSDNx48  ,sTeleCom::cSendTelex
               ,sTelex(QByteArray::number(TheNumOfPoints, 16).toUpper()
                      .prepend("G") )
               );
}


void  sBursiPan_MSDN::BreakScan ()
{
  SendDirective(MSDNx48  ,sTeleCom::cSendTelex ,sTelex("B") );
}


sPhysValue  sBursiPan_MSDN::TrueHo ()
{
  real  H = (LatestHo/sPhysValue("Oe")).ToReal();
  return  sPhysValue( H
                    - 2.55953
                    - 0.00174    * H
                    + 5.9237E-7  * H*H
                    - 3.8134E-11 * H*H*H
                    + 3.9487E-15 * H*H*H*H
                    ,0.1,"Oe");
}


sPhysValue  sBursiPan_MSDN::TrueDeltaH ()
{
  return  1.034 * LatestDeltaH;
}


sPhysValue sBursiPan_MSDN::FieldQuantum()
{
  return  TrueDeltaH() / 4096.0;
}


sPhysValue  sBursiPan_MSDN::TrueField (int  index)
{
  return  TrueHo() + TrueDeltaH() * ((index-2048)/4096.0)
        ;
}


/*! Changes from 0 to 4095 (so to say, "uint12_t").
*/
int  sBursiPan_MSDN::FieldIndex (sPhysValue  field)
{
  int                         Index = 0;
  try{
    Index = Round( 4096 * ( (field-TrueHo()) / TrueDeltaH() ).ToReal() )
          + 2048;
    if( Index < 0 ){
      Index = 0;
    }else if( Index > 4095 ){
      Index = 4095;
    }
  }catch(...){}
  return Index;
}

//------------------------------------------------------------------------------
} //namespace RA
