////////////////////////////////////////////////////////////////////////////////
/*! @file BursiPan_MSDN.h   BursiPan implementation
                            with field control through MSDNx48 and
                            with signal and handles from MSDN2047.
- Part of PanBursian - Research Assistant Plugin package.
- Uses  RAGUI - Research Assistant Graphical User Interface Library.
- Uses  RANet - Research Assistant Net Library.
- Uses  Qt v.5  -  http://qt-project.org/
- Copyright(C) 2014, Viktor E. Bursian, St.Petersburg, Russia.
                     Viktor.Bursian@mail.ioffe.ru
*///////////////////////////////////////////////////////////////////////////////
#ifndef BursiPan_MSDN_H
#define BursiPan_MSDN_H
#include "PanBursian.h"
#include "BursiPan.h"
#include "TeleCom.h"
namespace RA {
//------------------------------------------------------------------------------

ANNOUNCE_CLASS(sBursiPan_MSDN)

//----------------------------------------------------------- sBursiPan_MSDN ---
/*! BursiPan implementation with field control through MSDNx48
    and with signal and handles from MSDN2047.
*/
class PanBursian_EXPORT  sBursiPan_MSDN : public sBursiPan
{
  Q_OBJECT
  DRIVER(sBursiPan_MSDN)

  public: //const
    static const sPhysValue   PeriodConversion;

  public:
    virtual                   ~sBursiPan_MSDN ();
    virtual sPhysValue        Ho ()
                                { return LatestHo; }
    virtual sPhysValue        DeltaH ()
                                { return LatestDeltaH; }
    virtual sPhysValue        Frequency ()
                                { return LatestFrequency; }
    virtual sPhysValue        Attenuation ()
                                { return LatestAttenuation; }
    virtual sPhysValue        Gain ();
    virtual sPhysValue        Modulation ();
    virtual eModulationMode   ModulationMode ()
                                { return TheModulationMode; }

    virtual sPhysRange        FieldRange ();

  public:
    virtual void              SetScanRange (sPhysRange);
    virtual bool              ScanSettingsOK ()
                                { return ScanSettingsOKFlag; }
    virtual sPhysRange        ScanRange ();
    virtual sPhysValue        ScanStep ();
    virtual int               NumOfPoints ()
                                { return TheNumOfPoints; }
    virtual sPhysValue        TimePerStep ()
                                { return sPhysValue(TimePerFieldQuantum
                                                   *std::abs(StepInQuanta)
                                         ,"ns");
                                }

  protected: //evolution of the inherited state machine
    psParallelStateGroup      State_W_Parallel;
    psState                   State_W_MSDNx48;
    psState                   State_W_MSDNx48_Linking;
    psState                   State_W_MSDNx48_Checking;
    psState                   State_W_MSDNx47;
    psState                   State_W_MSDNx47_Linking;
    psState                   State_W_MSDNx47_Checking;
    psState                   State_O_Idle;
    psState                   State_O_Scanning;
    psParallelStateGroup      State_O_Sc_Prepare;
    psState                   State_O_Sc_SettingInitialField;
    psState                   State_O_Sc_SIF_Set;
    psState                   State_O_Sc_SIF_Pause;
    psState                   State_O_Sc_SettingGauge;
    psState                   State_O_Sc_SG_Interval;
    psState                   State_O_Sc_SG_Start;
    psState                   State_O_Sc_Stepping;
    psState                   State_O_Sc_ResettingGauge;

  signals: //for internal needs
    void                      MSDNx48Linked ();
    void                      MSDNx48Error (QString = QString());
    void                      GotMSDNx48Message (sTelex);
    void                      NoAnswerFromMSDNx48 ();
    void                      GotMSDNx48Version ();
    void                      SteppingDone ();
    void                      MSDNx47Linked ();
    void                      MSDNx47Error (QString = QString());
    void                      GotMSDNx47Message (sTelex);
    void                      NoAnswerFromMSDNx47 ();
    void                      GotMSDNx47Version ();
    void                      AveragingIsReady ();

  protected: //reimplementing inherited
    virtual void              PostConstructor ();

  protected slots: //reimplementing inherited
//    virtual void              AfterStartInitiation ();
    virtual void              OnEnterToOperatingState ();
    virtual void              OnExitFromOperatingState ();
    virtual void              OnEnterToCoolingState ();

  protected slots:
    virtual void              ReportMSDNx48Error (QString);
    virtual void              ReportMSDNx47Error (QString);
    virtual void              SetWantedEPRMeasurePeriod (sPhysValue);
    virtual void              SetMagneticField (sPhysValue);
    virtual void              SetScanStep (sPhysValue);
    virtual void              SetTimePerStep (sPhysValue);
    virtual void              SetInitialField ();
    virtual void              SetAveragingInterval ();
    virtual void              MakeAveragingReady ();
    virtual void              StartScan ();
    virtual void              BreakScan ();

  protected: //fields
    sPhysValue                Calibration;
    sPhysValue                ZeroCorrection;
    sTeleCom::sLink           MSDNx48;
    sTeleCom::sLink           MSDNx47;
    QByteArray                MSDNx48_Version;
    QByteArray                MSDNx47_Version;
    sPhysValue                LatestHo;
    sPhysValue                LatestDeltaH;
    sPhysValue                LatestFrequency;
    sPhysValue                LatestAttenuation;
    int                       GainMantissaIndex;
                                //!< 0: 0.50; ...; 9: 4.0
    int                       GainOrder;
                                //!< 0: *100; ...; 4: *1000000
    int                       ModulationMantissaIndex;
                                //!< 0: 1.00; ...; 9: 0.125
    int                       ModulationOrder;
                                //!< 0: *10; ...; 3: *0.01
    eModulationMode           TheModulationMode;
    int                       CurrentFieldIndex;
    bool                      ScanSettingsOKFlag;
    quint32                   TimePerFieldQuantum; /*ns*/
    int                       StepInQuanta;
    int                       FromIndex;
    int                       ToIndex;
    int                       TheNumOfPoints;

  private slots:
    void                      LinkToMSDNx48 ();
    void                      LinkToMSDNx47 ();
    void                      GetMSDNx48Version ();
    void                      GetMSDNx47Version ();
    void                      OnNoAnswerFromMSDNx48 ();
    void                      OnNoAnswerFromMSDNx47 ();

    void                      SetHo (sPhysValue);
    void                      SetDeltaH (sPhysValue);
    void                      SetFrequency (sPhysValue);
    void                      SetAttenuation (sPhysValue);
    void                      SetGain (sPhysValue);
    void                      SetModulation (sPhysValue);
    void                      SetModulationMode (eModulationMode);

    void                      ProcessMSDNx48Signature (sTelex);
    void                      ProcessMSDNx48Message (sTelex);
    void                      SignalRangeChanging ();
    void                      ProcessMSDNx47Signature (sTelex);
    void                      ProcessMSDNx47Message (sTelex);
    void                      ParseHandles (int  handles);

  private:
    void                      RecalculateNumOfPoints ();
    sPhysValue                TrueHo ();
    sPhysValue                TrueDeltaH ();
    sPhysValue                FieldQuantum ();
    sPhysValue                TrueField (int  index);
    int                       FieldIndex (sPhysValue  field);

  private: //fields
    bool                      Emulate;
};

//------------------------------------------------------------------------------
} //namespace RA
#endif
