////////////////////////////////////////////////////////////////////////////////
/*! @file BursiPan_RASOS.cpp BursiPan implementation with RASOS on Genie.
- Part of PanBursian - Research Assistant Plugin package.
- Uses  RAGUI - Research Assistant Graphical User Interface Library.
- Uses  RANet - Research Assistant Net Library.
- Uses  Qt v.5  -  http://qt-project.org/
- Copyright(C) 2014, Viktor E. Bursian, St.Petersburg, Russia.
                     Viktor.Bursian@mail.ioffe.ru
*///////////////////////////////////////////////////////////////////////////////
#include "BursiPan_RASOS.h"
#include "attr.h"
#include "StringsAndQStrings.h"
#include "Log.h"
namespace RA {
//------------------------------------------------------------------------------

DEFINE_CLASS_TAG(sBursiPan_RASOS)

//---------------------------------------------------------- sBursiPan_RASOS ---

void  sBursiPan_RASOS::Preconfigure (sNodePtr  DeviceCfg)
{
  sPtr<sFolderNode>              DeviceLinks;
  DeviceCfg |= "Links";
  DeviceLinks = DeviceCfg >> "Links";
  if( DeviceLinks.IsCorrect() ){
    DeviceLinks |= "Genie";
    DeviceLinks >> "Genie" |= attr("Class","sTeleCom");
  }
  sPtr<sFolderNode>              DeviceSettings;
  DeviceCfg |= "Settings";
  DeviceSettings = DeviceCfg >> "Settings";
  if( DeviceSettings.IsCorrect() ){
    DeviceSettings |= attr("Emulate",false);
    DeviceSettings |= attr("Calibration",sPhysValue(1.0,"Spin/Oe/Oe"));
    DeviceSettings |= attr("Ho",sPhysValue(0.0,0.1,"Oe"));
    DeviceSettings |= attr("DeltaH",sPhysValue(0.0,0.1,"Oe"));
    DeviceSettings |= attr("Frequency",sPhysValue(8.8e9,0.1e9,"Hz"));
    DeviceSettings |= attr("Attenuation",sPhysValue(30,0.1));
    DeviceSettings |= attr("Gain",sPhysValue(2e4,1));
    DeviceSettings |= attr("Modulation",sPhysValue(1,0.05,"Oe"));
    DeviceSettings |= attr("ModulationMode",integer(sBursiPan::cMod100kHz));
  }
  sBursiPan::Preconfigure(DeviceCfg);
}


sBursiPan_RASOS::~sBursiPan_RASOS ()
{
  if( DeviceSettings.IsCorrect() ){
    DeviceSettings
        << attr("Ho",LatestHo)
        << attr("DeltaH",LatestDeltaH)
        << attr("Frequency",LatestFrequency)
        << attr("Attenuation",LatestAttenuation)
        << attr("Gain",LatestGain)
        << attr("Modulation",LatestModulation)
        << attr("ModulationMode",integer(LatestModulationMode))
        ;
  }
  Genie.SetToNULL();
}


sBursiPan_RASOS::sBursiPan_RASOS (sString   DevicetId
                                 ,sNodePtr  DeviceCfg)
    :sBursiPan(DevicetId,DeviceCfg)
    ,Calibration("Spin/Oe/Oe")
    ,Genie(this)
    ,LatestHo("Oe")
    ,LatestDeltaH("Oe")
    ,LatestFrequency("Hz")
    ,LatestAttenuation(_Unitsless_)
    ,LatestGain(_Unitsless_)
    ,LatestModulation("Oe")
    ,LatestModulationMode(cMod100kHz)
    ,Emulate(false)
{
}


void  sBursiPan_RASOS::PostConstructor ()
{
  sBursiPan::PostConstructor();
  try{ //get config parameters
    get_from<sBooleanNode>(DeviceSettings>>"Emulate",Emulate);
    get_from<sPhysValueNode>(DeviceSettings>>"Calibration",Calibration);
    get_from<sPhysValueNode>(DeviceSettings>>"Ho",LatestHo);
    get_from<sPhysValueNode>(DeviceSettings>>"DeltaH",LatestDeltaH);
    get_from<sPhysValueNode>(DeviceSettings>>"Frequency",LatestFrequency);
    get_from<sPhysValueNode>(DeviceSettings>>"Attenuation",LatestAttenuation);
    get_from<sPhysValueNode>(DeviceSettings>>"Gain",LatestGain);
    get_from<sPhysValueNode>(DeviceSettings>>"Modulation",LatestModulation);
    integer  I;
    if( get_from<sIntegerNode>(DeviceSettings>>"ModulationMode",I) )
      LatestModulationMode = eModulationMode(I);
  }catch(...){
    SetInvalid( tr("Something wrong in configuration.") );
  }
  //evolution of the inherited state machine
  {
    //evolution of the State_Warming
    {
      State_W_Parallel = new sParallelStateGroup(State_Warming());
      State_W_Parallel
          ->SetAsInitial()
          ->AddTransition(State_Warming()->Fin());
      //State_W_Parallel internal structure and behaviour
      {
        State_W_Genie = new sState(State_W_Parallel,"State_W_Genie");
        //State_W_Genie structure and behaviour
        {
          State_W_Genie_Linking = new sState(State_W_Genie
                                            ,"State_W_Genie_Linking");
          State_W_Genie_Checking = new sState(State_W_Genie
                                             ,sPhysValue(15,"min")
                                             ,"State_W_Genie_Checking");

          State_W_Genie_Linking
              ->SetAsInitial()
              ->OnEnterCall( SLOT(LinkToGenie()) )
              ->AddTransition( SIGNAL(GenieLinked()) ,State_W_Genie_Checking )
              ->AddTransition( SIGNAL(ErrorMessage()) ,State_Cooling() )
              ;
          State_W_Genie_Checking
              ->OnEnterCall( SLOT(GetSOSVersion()) )
              ->AddReaction<sTelex>( SIGNAL(GotGenieMessage(sTelex))
                                   , SLOT(ProcessGenieMessage(sTelex))  )
              ->AddTransition( SIGNAL(NoAnswerFromSOS())
                             , SLOT(OnNoAnswerFromSOS())
                             , State_W_Genie_Checking )
//              ->AddTransition( State_W_Genie_Checking ,SIGNAL(Timeout())
//                             , SLOT(OnNoAnswer())
//                             , State_W_Genie_Checking )
              ->AddTransition( SIGNAL(GenieError()) ,State_Cooling() )
              ->AddTransition(SIGNAL(GotSOSVersion())
                             ,State_W_Genie->Fin()       )
              ;
          connect(State_W_Genie_Checking,SIGNAL(Timeout())
                 ,this,SIGNAL(NoAnswerFromSOS()));
        }
      }
    }
    //evolution of the State_Operating
    {
      State_Operating()
          ->AddReaction<sTelex>( SIGNAL(GotGenieMessage(sTelex))
                               , SLOT(ProcessGenieMessage(sTelex))  )
          ;

      State_O_Idle = new sState(State_Operating(),"State_O_Idle");
      State_O_WaitingRodger = new sState(State_Operating()
                                        ,"State_O_WaitingRodger");
      State_O_Scanning = new sState(State_Operating(),"State_O_Scanning");

      State_O_Idle
          ->SetAsInitial()
          ->AddReaction(cSetMagneticField,SLOT(SendMagneticField(sPhysValue)))
          ->AddReaction(cSetHo,SLOT(SendHo(sPhysValue)))
          ->AddReaction(cSetDeltaH,SLOT(SendDeltaH(sPhysValue)))
          ->AddReaction(cSetModulation,SLOT(SendModulation(sPhysValue)))
          ->AddReaction(cSetGain,SLOT(SendGain(sPhysValue)))
          ->AddReaction(cSetAttenuation,SLOT(SendAttenuation(sPhysValue)))
          ->AddReaction(cSetFrequency,SLOT(SendFrequency(sPhysValue)))
          ->AddReaction(cSetModulationMode
                       ,SLOT(SendModulationMode(eModulationMode)))
//          ->AddReaction(cSetWhatever,SLOT(SendWhatever(sPhysValue)))
          ->AddTransition(cScan,SLOT(StartScanJob(sScanJobData))
                         ,State_O_Scanning)
          ->AddTransition(SIGNAL(WaitRodgerFromSOS()),State_O_WaitingRodger)
          ;
      State_O_WaitingRodger
          ->AddTransition(SIGNAL(RodgerFromSOS())
                         ,SLOT(DisconnectRodger())
                         ,State_O_Idle)
          ;
      State_O_Scanning
          ->AddTransition(SIGNAL(ScanDone()) ,State_O_Idle)
          ;
//      State_O_Scanning
//          ->AddTransition(State_O_Idle)
//          ;
//      //State_O_Scanning internal structure
//      {
//        State_O_Sc_Starting = new sState(State_O_Scanning,"State_O_Sc_Start");
//        State_O_Sc_Rest = new sState(State_O_Scanning,"State_O_Sc_Rest");

//        State_O_Sc_Starting
//            ->SetAsInitial()
//            ->AddTransition(SIGNAL(ScanStarts(sScanStartData))
//                           ,SLOT(JustifyScanParams(sScanStartData))
//                           ,State_O_Sc_Rest)
//            ;
//        State_O_Sc_Rest
//            ->AddTransition(SIGNAL(GenieScanDone())
//                           ,State_O_Scanning->Fin())
//            ;
//      }
    }
    //evolution of the State_Cooling
    {
    }
    connect(this ,SIGNAL(GenieError(QString))
           ,this ,SLOT(ReportGenieError(QString)) );
    connect(this ,SIGNAL(NewHo(sPhysValue))
           ,this ,SLOT(SignalRangeChanging(sPhysValue)) );
    connect(this ,SIGNAL(NewDeltaH(sPhysValue))
           ,this ,SLOT(SignalRangeChanging(sPhysValue)) );
  }
}


//void  sBursiPan_RASOS::AfterStartInitiation ()
//{
//}


void  sBursiPan_RASOS::OnEnterToOperatingState ()
{
  sBursiPan::OnEnterToOperatingState();

  //...
}


void  sBursiPan_RASOS::OnExitFromOperatingState ()
{
  //...

  sBursiPan::OnExitFromOperatingState();
}


void  sBursiPan_RASOS::OnEnterToCoolingState ()
{
//  SendDirective( Genie , sTeleCom::cSendTelex , sTelex("exit") );
  Disconnect(Genie ,0 ,0);
  Genie.ReleaseExclusiveControl();
  Genie.SetToNULL();
  sBursiPan::OnEnterToCoolingState();
}


void  sBursiPan_RASOS::ReportGenieError (QString  error_text)
{
  emit ErrorMessage( tr("Got error from '%1':\n\n%2")
                       .arg(ToQString(Genie->DeviceIdentifier()))
                       .arg(error_text)     );
}

void  sBursiPan_RASOS::OnNoAnswerFromSOS ()
{
  emit ErrorMessage( "No answer from RASOS" );
}


void  sBursiPan_RASOS::LinkToGenie ()
{
  try{
    Genie.SetTo(DeviceLinks,"Genie");
    if( ! Genie.TakeExclusiveControl() ){
      emit ErrorMessage(tr("Cannot take exclusive control over '%1'. "
                           "Try again later.")
                       .arg(ToQString( Genie->DeviceIdentifier() )) );
    }else{
      Connect(Genie ,SIGNAL(Warmed()) ,SIGNAL(GenieLinked()) );
      Connect(Genie ,SIGNAL(ErrorMessage(QString))
             ,SIGNAL(GenieError(QString)) );
      Connect(Genie ,SIGNAL(GotTelex(sTelex))
             ,SIGNAL(GotGenieMessage(sTelex)) );
    }
  }catch( rxDeviceLaunch E ){
    emit ErrorMessage( tr("Error while linking to a submissive device.\n"
                          "%1\nCheck configuration.")
                         .arg(ToQString(E.Text))   );
  }
}


void  sBursiPan_RASOS::GetSOSVersion ()
{
  SendDirective(Genie  ,sTeleCom::cSendTelex
               ,sTelex(( Emulate ? "dura" : "sayversion" )) );
}


void  sBursiPan_RASOS::DisconnectRodger ()
{
  QObject::disconnect(this,0,this,SIGNAL(RodgerFromSOS()));
}


void  sBursiPan_RASOS::SendMagneticField (sPhysValue  value)
{
  Put(value,sPhysValue("Oe"),4,"\\Spectroscope\\Abscissa");
  QObject::connect(this ,SIGNAL(MagneticField(sPhysValue))
                  ,this ,SIGNAL(RodgerFromSOS()) ,Qt::QueuedConnection);
  QObject::connect(this ,SIGNAL(OutOfRange())
                  ,this ,SIGNAL(RodgerFromSOS()) ,Qt::QueuedConnection);
  emit WaitRodgerFromSOS();
  Get("H=","\\Spectroscope\\Abscissa");
}


void  sBursiPan_RASOS::StartScanJob (sBursiPan::sScanJobData  job)
{
  sString                     Command("average ");
  NumOfCountsPerPoint = Round( (job.TimePerStep/sPhysValue("ms")).ToReal() );
  if( NumOfCountsPerPoint <= 0 )  NumOfCountsPerPoint = 1;
  Command += sString::FromInteger(job.NumOfPasses);
  Command += " RASOS ";
  Command += sString::FromReal( (job.Range.From()/sPhysValue("Oe")).ToReal()
                              , 4 );
  Command += " ";
  Command += sString::FromReal( (job.Range.To()/sPhysValue("Oe")).ToReal()
                              , 4 );
  Command += " ";
  Command += sString::FromReal( (job.Step/sPhysValue("Oe")).ToReal()
                              , 4 );
  Command += " ";
  Command += sString::FromInteger(NumOfCountsPerPoint);
  if( ! SendDirective(Genie  ,sTeleCom::cSendTelex ,sTelex(Command) ) ){
    emit GenieError(tr("My submissive device '%1' is busy now.\n\n"
                       "Try again later.")
                      .arg(ToQString(Genie->DeviceIdentifier()))        );
  }
}


//void sBursiPan_RASOS::JustifyScanParams(sBursiPan::sScanStartData  actual)
//{
//}


void  sBursiPan_RASOS::SendHo (sPhysValue  value)
{
  Put(value,sPhysValue("Oe"),0,"\\Spectroscope\\Ho");
  QObject::connect(this ,SIGNAL(NewHo(sPhysValue))
                  ,this ,SIGNAL(RodgerFromSOS()) ,Qt::QueuedConnection);
  emit WaitRodgerFromSOS();
  Get("Ho=","\\Spectroscope\\Ho");
}


void  sBursiPan_RASOS::SendDeltaH (sPhysValue  value)
{
  Put(value,sPhysValue("Oe"),0,"\\Spectroscope\\dH");
  QObject::connect(this ,SIGNAL(NewDeltaH(sPhysValue))
                  ,this ,SIGNAL(RodgerFromSOS()) ,Qt::QueuedConnection);
  emit WaitRodgerFromSOS();
  Get("dH=","\\Spectroscope\\dH");
}


void  sBursiPan_RASOS::SendFrequency (sPhysValue  value)
{
  Put(value,sPhysValue("GHz"),4,"\\Spectroscope\\Freq");
  LatestFrequency = value;
  emit NewFrequency(value);
}


void  sBursiPan_RASOS::SendAttenuation (sPhysValue  value)
{
  Put(value,sPhysValue(_Unitsless_),1,"\\Spectroscope\\A");
  LatestAttenuation = value;
  emit NewAttenuation(value);
}


void  sBursiPan_RASOS::SendGain (sPhysValue  value)
{
  Put(value,sPhysValue(_Unitsless_),0,"\\Spectroscope\\G");
  LatestGain = value;
  emit NewGain(value);
}


void  sBursiPan_RASOS::SendModulation (sPhysValue  value)
{
  Put(value,sPhysValue("Oe"),4,"\\Spectroscope\\MA");
  LatestModulation = value;
  emit NewModulation(value);
}


void  sBursiPan_RASOS::SendModulationMode (eModulationMode  value)
{
  sString                     Command("put \\Spectroscope\\MF ");
  Command += cModeText[value];
  SendDirective( Genie , sTeleCom::cSendTelex , sTelex(Command) );
  emit NewModulationMode(value);
}


/*
\Spectroscope\Title
\Spectroscope\Rem0
\Spectroscope\Rem1
\Spectroscope\Rem2
\Spectroscope\Rem3
\Spectroscope\Abscissa
\Spectroscope\Ordinate
\\Spectroscope\\Freq"); }
\\Spectroscope\\A"); }
\\Spectroscope\\MA"); }
\\Spectroscope\\MF"); }
\\Spectroscope\\G"); }
\\Spectroscope\\Ho"); }
\\Spectroscope\\dH"); }
\\Spectroscope\\Step"); }
\\Spectroscope\\Times"); }
\\Spectroscope\\BurstDempherCounter");

Scan Name From Til Step[ MeasTimes[ DrawPeriod]]
?or?: Scan Name From Til[ Step[ MeasTimes[ DrawPeriod]]]
Average NumOfPasses SpectrumName From Til Step[ MeasTimes[ DrawPeriod]]
ScanBreak
ScanStop
*/

void  sBursiPan_RASOS::Put (sPhysValue  value
                           ,sPhysValue  unit
                           ,short       digits_after_dot
                           ,literal     SOS_field)
{
  sString                     Command("put ");
  Command += SOS_field;
  Command += " ";
  Command += sString::FromReal( (value/unit).ToReal() , digits_after_dot );
  SendDirective( Genie , sTeleCom::cSendTelex , sTelex(Command) );
  if( Emulate )
    Magic = sString::FromReal( (value/unit).ToReal() , digits_after_dot );
}


void  sBursiPan_RASOS::Get (literal  prefix
                           ,literal  SOS_field)
{
  sString                     Command("say ");
  Command += prefix;
  Command += "$";
  Command += SOS_field;
  Command += "$";
  SendDirective( Genie , sTeleCom::cSendTelex , sTelex(Command) );
  if( Emulate ){
    Command = "&rArr;";
    Command += prefix;
    Command += Magic;
    SendDirective( Genie , sTeleCom::cSendTelex , sTelex(Command) );
  }
}


void  sBursiPan_RASOS::ProcessGenieMessage (sTelex  text)
{
  QString               ConversionErr(tr("Got '%1'. Cannot convert."));
  if(  text.startsWith("P:") ){
    bool                        OK = false;
    int                         C1;
    int                         C2;
    sScanPointData              SPD;
    real                        Signal;
    real                        Dispersion;
    C1=2;  C2=text.indexOf(",",C1);
    SPD.PassNo=text.mid(C1,C2-C1).toInt(&OK);
    if( OK ){
      OK=false;
      C1=C2+1;  C2=text.indexOf(",",C1);
      SPD.PointNo=text.mid(C1,C2-C1).toInt(&OK);
    }
    if( OK ){
      OK=false;
      C1=C2+1;  C2=text.indexOf(",",C1);
      Signal=text.mid(C1,C2-C1).toDouble(&OK);
    }
    if( OK ){
      OK=false;
      Dispersion=text.mid(C2+1).toDouble(&OK);
    }
    if( OK ){
      SPD.Signal=sPhysValue(Signal,sqrt(Dispersion))
                *Calibration;
      emit ScanPoint(SPD);
      emit EPR_Signal(SPD.Signal);
    }else{
      emit GenieError(ConversionErr.arg(QString::fromLatin1(text)));
    }
  }else if(  text.startsWith("S=") ){
    bool                        OK = false;
    real                        V = text.mid(2).toFloat(&OK);
    if( OK ){
      emit EPR_Signal(V*Calibration);
    }else{
      emit GenieError(ConversionErr.arg(QString::fromLatin1(text)));
    }

  }else if(  text.startsWith("H=") ){
    bool                        OK = false;
    real                        V = text.mid(2).toFloat(&OK);
    if( OK ){
      emit MagneticField(sPhysValue(V,0.005,"Oe"));
    }else{
      emit GenieError(ConversionErr.arg(QString::fromLatin1(text)));
    }

  }else if(  text.startsWith("Ho=") ){
    bool                        OK = false;
    real                        V = text.mid(3).toFloat(&OK);
    if( OK ){
      LatestHo = sPhysValue(V,0.05,"Oe");
      emit NewHo(LatestHo);
    }else{
      emit GenieError(ConversionErr.arg(QString::fromLatin1(text)));
    }

  }else if(  text.startsWith("dH=") ){
    bool                        OK = false;
    real                        V = text.mid(3).toFloat(&OK);
    if( OK ){
      LatestDeltaH = sPhysValue(V,0.5,"Oe");
      emit NewDeltaH(LatestDeltaH);
    }else{
      emit GenieError(ConversionErr.arg(QString::fromLatin1(text)));
    }

  }else if(  text.startsWith(":ScanStarts:") ){
    bool                        OK = false;
    int                         C1;
    int                         C2;
    sScanStartData              Params;
    real                        From;
    real                        Step;
    C1=text.indexOf(":",1)+1;  C2=text.indexOf(",",C1);
    From=text.mid(C1,C2-C1).toDouble(&OK);
    if( OK ){
      OK=false;
      C1=C2+1;  C2=text.indexOf(",",C1);
      Step=text.mid(C1,C2-C1).toDouble(&OK);
    }
    if( OK ){
      OK=false;
      Params.NumOfPoints=text.mid(C2+1).toInt(&OK);
    }
    if( OK ){
      Params.From=sPhysValue(From,"Oe");
      Params.Step=sPhysValue(Step,"Oe");
      Params.NumOfCountsPerPoint=NumOfCountsPerPoint;
      emit ScanStarts(Params);
    }else{
      emit GenieError(ConversionErr.arg(QString::fromLatin1(text)));
    }

  }else if(  text.startsWith(":ScanDone:") ){
    emit ScanDone();

  }else if(  text.startsWith("E:Out of range") ){
    emit OutOfRange();
  }else if(  text.startsWith("W:") ){
    emit GenieError(text.mid(2));
  }else if(  text.startsWith("E:") ){
    emit GenieError(text.mid(2));
  }else if(  text.startsWith("S:") ){
    emit GenieError(text.mid(2));
  }else if(  text.startsWith("T:") ){
    emit GenieError(text.mid(2));
  }else if(  text.startsWith("SOS:") ){
    RASOS_Version = text;
    emit GotSOSVersion();
  }else if(  text.startsWith("Sam durak") ){
    RASOS_Version = "dura";
    emit GotSOSVersion();
  }else{
    emit GenieError(QString("Unknown telex: ")+text);
  }
}


void  sBursiPan_RASOS::SignalRangeChanging (sPhysValue)
{
  try{
    emit RangeChanged( sPhysRange( LatestHo-LatestDeltaH/2
                                 , LatestHo+LatestDeltaH/2 ) );
  }catch(...){}
}

//------------------------------------------------------------------------------
} //namespace RA
