////////////////////////////////////////////////////////////////////////////////
/*! @file LakeShore475Gaussmeter.cpp LakeShore Model 475 DSP Gaussmeter.
- Part of PanBursian - Research Assistant Plugin package.
- Uses  RAGUI - Research Assistant Graphical User Interface Library.
- Uses  RANet - Research Assistant Net Library.
- Uses  Qt v.5  -  http://qt-project.org/
- Copyright(C) 2015, Viktor E. Bursian, St.Petersburg, Russia.
                     Viktor.Bursian@mail.ioffe.ru
*///////////////////////////////////////////////////////////////////////////////
#include "LakeShore475Gaussmeter.h"
#include "attr.h"
#include "StringsAndQStrings.h"
namespace RA {
//------------------------------------------------------------------------------

DEFINE_CLASS_TAG(sLakeShore475Gaussmeter)

//-------------------------------------------------- sLakeShore475Gaussmeter ---

void  sLakeShore475Gaussmeter::Preconfigure (sNodePtr  DeviceCfg)
{
  sPtr<sFolderNode>              DeviceLinks;
  DeviceCfg |= "Links";
  DeviceLinks = DeviceCfg >> "Links";
  if( DeviceLinks.IsCorrect() ){
    DeviceLinks |= "GaussmeterPort";
    DeviceLinks >> "GaussmeterPort" |= attr("Class","sTeleCom");
  }
  sGauge::Preconfigure(DeviceCfg);
}


sLakeShore475Gaussmeter::~sLakeShore475Gaussmeter ()
{
  GaussmeterPort.SetToNULL();
  if( Clock ){ delete Clock; Clock=NULL; }
}


sLakeShore475Gaussmeter::sLakeShore475Gaussmeter (sString   DevicetId
                                                 ,sNodePtr  DeviceCfg)
    :sGauge(DevicetId,DeviceCfg)
    ,GaussmeterPort(this)
    ,Clock(new QTimer(this))
    ,Interval(100)
{
}


void  sLakeShore475Gaussmeter::PostConstructor ()
{
  sGauge::PostConstructor();
  //evolution of the inherited state machine
  {
    //evolution of the State_Warming
    {
      State_W_GaussmeterPort = new sState(State_Warming()
                                         ,"State_W_GaussmeterPort");
      State_W_GaussmeterPort
          ->SetAsInitial()
          ->AddTransition(State_Warming()->Fin())
          ;
      //State_W_GaussmeterPort structure and behaviour
      {
        State_W_GaussmeterPort_Linking = new sState(State_W_GaussmeterPort
                                          ,"State_W_GaussmeterPort_Linking");
        State_W_GaussmeterPort_Checking = new sState(State_W_GaussmeterPort
                                           ,sPhysValue(1,"s")
                                           ,"State_W_GaussmeterPort_Checking");

        State_W_GaussmeterPort_Linking
            ->SetAsInitial()
            ->OnEnterCall(SLOT(LinkToGaussmeterPort()))
            ->AddTransition(SIGNAL(GaussmeterPortLinked())
                           ,State_W_GaussmeterPort_Checking )
            ->AddTransition(SIGNAL(ErrorMessage()) ,State_Cooling() )
            ;
        State_W_GaussmeterPort_Checking
            ->OnEnterCall(SLOT(RequestGaussmeterName()))
            ->AddReaction<sTelex>(SIGNAL(GotGaussmeterMessage(sTelex))
                                 ,SLOT(ParseGaussmeterName(sTelex))  )
            ->AddTransition(SIGNAL(NoAnswerFromGaussmeter())
                           ,SLOT(OnNoAnswerFromGaussmeter())
                           ,State_W_GaussmeterPort_Checking )
            ->AddTransition(SIGNAL(GaussmeterPortError()) ,State_Cooling() )
            ->AddTransition(SIGNAL(ErrorMessage()) ,State_Cooling() )
            ->AddTransition(SIGNAL(GotGaussmeterName())
                           ,State_W_GaussmeterPort->Fin()       )
            ;
        connect(State_W_GaussmeterPort_Checking,SIGNAL(Timeout())
               ,this,SIGNAL(NoAnswerFromGaussmeter()));
      }
    }
    //evolution of the State_Operating
    {
      State_Operating()->OnExitCall(Clock,SLOT(stop()));
    }
    //evolution of the State_O_Measuring
    {
      State_O_M_Stopped
          ->OnEnterCall(Clock,SLOT(stop()))
          ;
      State_O_M_Acquiring
          ->OnEnterCall(SLOT(RunTheClockIfNotYet()))
          ->AddTransition(State_O_M_Cycling)
          ;
      State_O_M_A_Pausing
          = new sState(State_O_M_Acquiring,"State_O_M_A_Pausing");
      State_O_M_A_Reading
          = new sState(State_O_M_Acquiring,"State_O_M_A_Reading");
      State_O_M_A_Pausing
          ->SetAsInitial()
          ->AddTransition(Clock,SIGNAL(timeout())
                         ,State_O_M_A_Reading)
          ;
      State_O_M_A_Reading
          ->OnEnterCall(SLOT(RequestField()))
          ->AddTransition<sTelex>(SIGNAL(GotGaussmeterMessage(sTelex))
                                 ,SLOT(ParseField(sTelex))
                                 ,State_O_M_Acquiring->Fin())
          ;
    }
    connect(this ,SIGNAL(GaussmeterPortError(QString))
           ,this ,SLOT(ReportGaussmeterPortError(QString)) );
  }
}


void  sLakeShore475Gaussmeter::OnEnterToCoolingState ()
{
//  SendDirective( GaussmeterPort , sTeleCom::cSendTelex , sTelex("exit") );
  Disconnect(GaussmeterPort ,0 ,0);
  GaussmeterPort.ReleaseExclusiveControl();
  GaussmeterPort.SetToNULL();
  sGauge::OnEnterToCoolingState();
}


void  sLakeShore475Gaussmeter::SetWantedPeriod ()
{
  try{
    Interval = Round( (WantedPeriod / sPhysValue(1.0,"ms")).ToReal() );
    if( Interval < 1 )  Interval=1;
    ThePeriod = sPhysValue(Interval,0.5,"ms");
    emit PeriodChanged(ThePeriod);
  }catch(rxRAlgebra){
    //!@todo{Instruments} define Failure signal for drivers?
    emit ErrorMessage(tr("Cannot set period, period=%1")
                     .arg(ToQString(WantedPeriod.Text())));
    emit CoolDown();
  }
}


void sLakeShore475Gaussmeter::RunTheClockIfNotYet ()
{
  if( ! Clock->isActive() ){
    Clock->setSingleShot( AutoRepeat() );
    Clock->start(Interval);
  }
}


void  sLakeShore475Gaussmeter::ReportGaussmeterPortError (QString  error_text)
{
  emit ErrorMessage(tr("Got error from %1:"
                       "\n\n%2")
                      .arg(ToQString(GaussmeterPort->DeviceIdentifier()))
                      .arg(error_text)     );
}


void  sLakeShore475Gaussmeter::LinkToGaussmeterPort ()
{
  try{
    GaussmeterPort.SetTo(DeviceLinks,"GaussmeterPort");
    if( ! GaussmeterPort.TakeExclusiveControl() ){
      emit ErrorMessage(tr("Cannot take exclusive control over '%1'. "
                           "Try again later.")
                       .arg(ToQString( GaussmeterPort->DeviceIdentifier() )) );
    }else{
      Connect(GaussmeterPort ,SIGNAL(Warmed())
             ,SIGNAL(GaussmeterPortLinked()) );
      Connect(GaussmeterPort ,SIGNAL(ErrorMessage(QString))
             ,SIGNAL(GaussmeterPortError(QString)) );
      Connect(GaussmeterPort ,SIGNAL(GotTelex(sTelex))
             ,SIGNAL(GotGaussmeterMessage(sTelex)) );
    }
  }catch( rxDeviceLaunch E ){
    emit ErrorMessage( tr("Error while linking to a submissive device.\n"
                          "%1\nCheck configuration.")
                         .arg(ToQString(E.MsgText()))   );
  }
}


void  sLakeShore475Gaussmeter::OnNoAnswerFromGaussmeter ()
{
  emit ErrorMessage("No answer from LakeShore 475 DSP Gaussmeter");
}


void  sLakeShore475Gaussmeter::RequestGaussmeterName ()
{
  SendDirective(GaussmeterPort  ,sTeleCom::cSendTelex ,sTelex("*IDN?") );
}


void  sLakeShore475Gaussmeter::ParseGaussmeterName (sTelex  text)
{
  if(  text.startsWith("LSCI,MODEL475") ){
    emit GotGaussmeterName();
  }else{
    emit ErrorMessage(QString("Unknown telex: ")+text);
  }
}


void  sLakeShore475Gaussmeter::RequestField ()
{
  SendDirective(GaussmeterPort  ,sTeleCom::cSendTelex ,sTelex("RDGFIELD?") );
}


void  sLakeShore475Gaussmeter::ParseField (sTelex  text)
{
  bool                        OK = false;
  real                        Field = text.toDouble(&OK);
  if( OK ){
    emit NewData(sPhysValue(Field,0.005,"Oe"));
  }else{
    emit ErrorMessage(QString("Unknown telex: ")+text);
  }
}

//------------------------------------------------------------------------------
} //namespace RA
