/*! @file ADC_AD7321_VB.c  ADC AD7321 on Microsin board, designed for VB.
- Part of the MSDN2047 firmware.
- Part of the Research Assistant project.
- Uses  LUFA library written by
   - Copyright(C) 2012, Dean Camera, dean [at] fourwalledcubicle [dot] com,
                                     www.lufa-lib.org.

- Copyright(C) 2014, Dmitry K. Nelson, St.Petersburg, Russia.
                     D.Nelson@mail.ioffe.ru
- Copyright(C) 2016, Viktor E. Bursian, St.Petersburg, Russia.
                     Viktor.Bursian@mail.ioffe.ru
*///////////////////////////////////////////////////////////////////////////////
#include "ADC_AD7321_VB.h"
#include "ClocksSimple.h"
#include "Microsin_USB_IO.h"
#include <avr/iousb162.h>
#include <avr/interrupt.h>
//#include <stdlib.h>    //for rand()
//------------------------------------------------------------------------------

struct sADC         ADC = {.EvaluateNoise = false
                          ,.MinIntervalBetweenReadings = 100000 /*ns*/
                          ,.GotSynchroimpulse = false
                          ,.ToBeSent = false

                          ,.ReadADC  = true
                          ,.EmulateADC  = false
                          ,.NoiseLevel  = 5
                          ,.OutputCounter = -1
                          ,.ClockSignalCounter = 0
                          ,.ClockProcessCounter = 0
                          ,.UnsentCounter = 0
                          };

void  ADC_WriteRegister (void);


/*=============================== Распайка =====================================
Вход внешнего строба       Бит 6 порта D (INT6)

Ввод данных АЦП            Бит 0 порта D - тактовый сигнал для АЦП
                           Бит 1 порта D - инициализация передачи АЦП
                           Бит 2 порта D - данные из АЦП (ADC_DO)
                           Бит 3 порта D - данные в АЦП (ADC_DI)

==============================================================================*/

//#define INT6ENABLE 0b01000000
#define ADC_CLK PD0
#define ADC_CS PD1
#define ADC_DO PD2
#define ADC_DI PD3

#define ADCRANGE10 0b1010000000000000
#define ADCRANGE5  0b1010100000000000
#define ADCRANGE25 0b1011000000000000
#define ADCCONTROL 0b1000000000010000  // Канал 0 несимметричный,
                                       // в дополнительном коде

//#define WATCH_PERIOD 6250		// Период опроса АЦП в тактах таймера 1 (~100ms)
//#define WATCH_SAMPLES_NUM 1000	// Число отсчетов АЦП в режиме Watch
//                                // (не более, чем период/16мкс)

static uint16_t ADCControlData;			// Набор битов для записи в регистр АЦП

static uint8_t DelayVar = 0;       		// Переменная для задержки

//------------------------------------------------------------------------------


void  ADC_Init (void)
{
  Clock1_Init();

  // ---------------------------------------------------------------------------
  // Установки для внешнего прерывания INT6 (бит 6 порта D) --------------------
  EICRB = 0b00100000; 	// interrupt PCINT1 on falling edge
  EIMSK &= ~INT6ENABLE; 	// Запрет прерывания

  // Ввод данных с АЦП
  //настройка ADC_DO как входа без pullup
  DDRD &= ~(1<<ADC_DO);
  PORTD &= ~(1<<ADC_DO);
  //настройка ADC_CLK, ADC_CS и ADC_DI как выходов.
  //Данные сигналы инвертируются оптронной развязкой
  DDRD |= ((1<<ADC_CS)|(1<<ADC_CLK)|(1<<ADC_DI));
  PORTD &= ~(1<<ADC_CLK); // у CLK начальное состояние 0 (соотв. 1 на входе АЦП)
  PORTD &= ~(1<<ADC_CS);  // у CS начальное состояние 0 (соотв. 1 на входе АЦП).
  PORTD |= (1<<ADC_DI); // у DI начальное состояние 1 (соотв. 0 на входе АЦП).

  DDRD |= (1<<PD5);		 // бит 5 для тестового выхода
  PORTD &= ~(1<<PD5); //   начальное состояние 0.

  ADC_BreakAndReinitialize();
}


void  ADC_BreakAndReinitialize (void)
{
  ADC_Stop();
  ADC_SetRegime();
  ADC_SetRange(0);
//  Clock1_SetTickFrequency(62500);  // interval 16 us
  Clock1_SetTickFrequency(15625);  // interval 64 us
  ADC_SetPeriod(1000000000);       // 1 s
//  ADC.EvaluateNoise = true;
  ADC.EvaluateNoise = false;
  ADC.ToBeSent = false;
  ADC.ReadADC = true;
  ADC.EmulateADC = false;
//  ADC.ReadADC = false;
//  ADC.EmulateADC = true;
  ADC.NoiseLevel = 0xFFF;
  ADC.MinIntervalBetweenReadings = 200000;
  ADC.OutputCounter = -1;
  ADC.UnsentCounter = 0;
  ADC.ClockSignalCounter = 0;
  ADC.ClockProcessCounter = 0;
  LeadingZeroes = 1;
}


void  ADC_ReportParams (void)
{
  SendStringToUSB("ADC ");
  if( ADC.ReadADC ){
    SendStringToUSB("read");
    if( ADC.EmulateADC )  SendCharToUSB('+');
  }
  if( ADC.EmulateADC ){
    SendStringToUSB("emu");
  }
  Send_EOL_ToUSB();

  SendCharToUSB('f');
  SendNumberToUSB(Clock1_TickFrequency());
  Send_EOL_ToUSB();
  SendCharToUSB('p');
  SendNumberToUSB(Clock1_Period());
  Send_EOL_ToUSB();
  SendCharToUSB('i');
  SendNumberToUSB(Clock1_TickInterval());
  Send_EOL_ToUSB();
  SendCharToUSB('P');
  SendNumberToUSB(ADC_Period());
  Send_EOL_ToUSB();
  SendCharToUSB('I');
  SendNumberToUSB(ADC.MinIntervalBetweenReadings);
  Send_EOL_ToUSB();

  SendCharToUSB('O');
  SendNumberToUSB(ADC.OutputCounter);
  Send_EOL_ToUSB();
  SendCharToUSB('-');
  SendNumberToUSB(ADC.UnsentCounter);
  Send_EOL_ToUSB();
}


void  ADC_SetRegime (void)
{
  ADCControlData = ADCCONTROL; // Установка режима АЦП
  ADC_WriteRegister();
}


void  ADC_SetRange (int8_t  range)
{
  ADCControlData = ADCRANGE10;
  if(      range == 0 )  ADCControlData = ADCRANGE10;
  else if( range == 1 )  ADCControlData = ADCRANGE5;
  else if( range ==  2)  ADCControlData = ADCRANGE25;
  ADC_WriteRegister();
}


int8_t  ADC_SetPeriod (uint32_t  period /*ns*/)
{
  ADC.MaxNumOfSamplesToAverage = ((uint32_t)0x1) << 19; //supposingly, maximal allowed
  Clock1_SetPeriod( period / ADC.MaxNumOfSamplesToAverage );
  if( Clock1_Period() < ADC.MinIntervalBetweenReadings )
    Clock1_SetPeriod(ADC.MinIntervalBetweenReadings);
  ADC.MaxNumOfSamplesToAverage = period / Clock1_Period();
  if( ADC.MaxNumOfSamplesToAverage )
    return 0;
  return 1;
}


uint32_t  ADC_Period (void) /*ns*/
{
  return Clock1_Period() * ADC.MaxNumOfSamplesToAverage;
}


int8_t  ADC_MeasurePeriodically (void)
{
  EIMSK &= ~INT6ENABLE;		// Запрещаем внешнее прерывание
  ADC.ReportingMode = cPeriodically;
  Clock1_SetToZero();
  ADC_AveragingClear();
  Clock1_ForgetInterrupts();
  Clock1_AllowInterrupts();
  return 0;
}


int8_t  ADC_MeasureSynchronized (uint16_t  num_of_points)
{
  ADC.NumberOfPointsToCollect = num_of_points;
  ADC.ReportingMode = cWaiting;
  EIFR |= INT6ENABLE; // Забыть про необработанные сигналы прерывания,
                      // случившиеся пока оно было запрещено.
  EIMSK |= INT6ENABLE;		// Разрешаем внешнее прерывание
  SendCharToUSB('Y');
  SendCharToUSB('r');
  Send_EOL_ToUSB();
  return 0;
}


void  ADC_Stop (void)
{
  EIMSK &= ~INT6ENABLE;		// Запрещаем внешнее прерывание
  Clock1_ForbidInterrupts();
  ADC.ReportingMode = cNever;
}


void ADC_AveragingClear (void)
{
  ADC.NumOfAveragedSamples = 0;
  ADC.Sum = 0;
  ADC.NumOfNoiseSamples = 0;
  ADC.NoiseSum = 0;
  ADC.ShiftedDispersion = 0;
  ADC.AveragingJustCleared = true;
}


void ADC_AveragingFinalize (void)
{
  if( ADC.ToBeSent ){
    ADC.UnsentCounter++;
  }else{
    ADC.Value = ADC.Sum;
    ADC.ValueDivider = ADC.NumOfAveragedSamples;
    if( ADC.EvaluateNoise ){
      int32_t  Shift = ADC.NoiseSum / (ADC.NumOfNoiseSamples+1);
      Shift -= ADC.FirstValue;
      Shift *= Shift;
      ADC.Dispersion = ADC.ShiftedDispersion
                     - ((uint32_t)Shift) * (ADC.NumOfNoiseSamples + 1);
      ADC.DispersionDivider = ADC.NumOfNoiseSamples;
    }
    ADC.ToBeSent = true;
  }
  ADC_AveragingClear();
}


void ADC_GetNextReadingAndAverage (void)
{
  ADC_Read();
  ADC.Sum += ADC.InputValue;
  ADC.NumOfAveragedSamples++;
  if( ADC.EvaluateNoise && (ADC.NumOfNoiseSamples < 64) ){
    if( ADC.AveragingJustCleared ){
      ADC.FirstValue = ADC.InputValue;
      ADC.NoiseSum = ADC.InputValue;
      ADC.AveragingJustCleared = false;
    }else{
      ADC.NoiseSum += ADC.InputValue;
      int32_t  XX = ADC.InputValue;
      XX -= ADC.FirstValue;
      XX *= XX;
      ADC.ShiftedDispersion += XX;
      ADC.NumOfNoiseSamples++;
    }
  }
  if( ADC.NumOfAveragedSamples >= ADC.MaxNumOfSamplesToAverage ){
    if( ADC.ReportingMode == cSynchronized ){
      //!@todo{MSDN} no synchroimpulse within max period
    }else{
      ADC_AveragingFinalize();
    }
  }
}


void ADC_Report (void)
{
  if( ADC.OutputCounter >= 0 ){
    ADC.OutputCounter++;
  }else{
    SendCharToUSB('V');
    SendNumberToUSB(ADC.Value);
    SendCharToUSB('/');
    SendNumberToUSB(ADC.ValueDivider);
    Send_EOL_ToUSB();
    if( ADC.EvaluateNoise ){
      SendCharToUSB('D');
      SendNumberToUSB(ADC.Dispersion);
      SendCharToUSB('/');
      SendNumberToUSB(ADC.DispersionDivider);
      Send_EOL_ToUSB();
    }
  }
}


void  ADC_OnClock (void)
{
  ADC.ClockProcessCounter++;
  if( ADC.ReportingMode != cNever ){
    ADC_GetNextReadingAndAverage();
  }
}


void  ADC_OnSynchroimpulse (void)
{
  if( ADC.ReportingMode == cWaiting ){
    ADC.ReportingMode = cSynchronized;
    Clock1_SetToZero();
    ADC_AveragingClear();
    Clock1_ForgetInterrupts();
    Clock1_AllowInterrupts();
    SendCharToUSB('Y');
    SendCharToUSB('s');
    Send_EOL_ToUSB();
  }else if( ADC.ReportingMode == cSynchronized ){
    Clock1_SetToZero();
    ADC_AveragingFinalize();
//    ADC_Report();
//    ADC_AveragingClear();
    ADC.NumberOfPointsToCollect--;
    if( ADC.NumberOfPointsToCollect == 0 ){
      ADC_Stop();
      SendCharToUSB('Y');
      SendCharToUSB('f');
      Send_EOL_ToUSB();
    }
  }else{
    SendCharToUSB('y');
    Send_EOL_ToUSB();
  }
}


//------------------------------------------------------------------------------
/*! WriteADCRegister.

@note Опторазвязка между контроллером и АЦП инвертирует сигнал
*/
void  ADC_WriteRegister (void)
{
	unsigned char i;
	PORTD |= (1<<ADC_CS); //ADC_CS=1 (0 на АЦП) активируем передачу данных
	DelayVar++;DelayVar++;	// Для задержки
	for(i=0; i < 16;i++) 	// Передаем данные - старший бит первый
	{
		if (ADCControlData & 0x8000)
			PORTD &= ~(1<<ADC_DI); //(выводим 1 на АЦП)
		else
			PORTD |= (1<<ADC_DI); //(выводим 0 на АЦП)
		DelayVar++;DelayVar++; // Для задержки
		PORTD |= (1<<ADC_CLK);//ADC_CLK=1 (0 на АЦП)
		DelayVar++;DelayVar++; // Для задержки
		PORTD &= ~(1<<ADC_CLK); //ADC_CLK=0 (1 на АЦП)
		DelayVar++;DelayVar++; // Для задержки
		ADCControlData<<=1;
	}
	DelayVar++;DelayVar++;DelayVar++;// Для задержки
}


//------------------------------------------------------------------------------
/*! Ввод данных из АЦП в переменную ADC.Input (for AD7321).

@note Опторазвязка между контроллером и АЦП инвертирует сигнал.

@note Один акт ввода данных АЦП занимает ~16 мкс.
*/
void  ADC_Read (void)
{
  ADC.InputValue = 0;
  if( ADC.ReadADC ){
    //ADC_DI=1 (0 на АЦП) отключаем запись в регистр АЦП
    PORTD |= (1<<ADC_DI);
    //ADC_CS=1 (0 на АЦП) активируем передачу данных
    PORTD |= (1<<ADC_CS);
    DelayVar++;DelayVar++; // Для задержки
    // ----------------- 3 такта пропуска нулевых битов и бита номера канала ---
    ADC.SkippedBits = 0;
    for( int8_t i=0 ; i < 3 ; i++ ){
      PORTD |= (1<<ADC_CLK);//ADC_CLK=1 (0 на АЦП)
      DelayVar++;DelayVar++; // Для задержки
      ADC.SkippedBits<<=1;
      if (PIND & (1<<ADC_DO))
        ADC.SkippedBits |= 1;
      PORTD &= ~(1<<ADC_CLK); //ADC_CLK=0 (1 на АЦП)
      DelayVar++;DelayVar++; // Для задержки
    }
    // ------------ чтение знака и 12 бит результата АЦП, старший бит первый ---
    for( int8_t i=0 ; i<13 ; i++ ){
      PORTD |= (1<<ADC_CLK);//ADC_CLK=1 (0 на АЦП)
      DelayVar++;DelayVar++; // Для задержки
      ADC.InputValue<<=1;
      if (PIND & (1<<ADC_DO))
        ADC.InputValue |= 1;
      PORTD &= ~(1<<ADC_CLK); //ADC_CLK=0 (1 на АЦП)
      DelayVar++;DelayVar++; // Для задержки
    }
    PORTD &= ~(1<<ADC_CS); // Отключение передачи данных
    if( ADC.InputValue & 0b0001000000000000 )
      ADC.InputValue |=  0b1111000000000000;
  }
  if( ADC.EmulateADC ){
    ADC.InputValue = -666;
    if( ADC.NoiseLevel ){
      int32_t  Noise = TCNT0;// rand();
      Noise *= 2;
      Noise -= OCR0A;// RAND_MAX;
      Noise *= ADC.NoiseLevel;
      Noise /= OCR0A;// RAND_MAX;
      ADC.InputValue += Noise;
    }
  }
}


//------------------------------------------------------------------------------
/*! Обработчик прерывания от таймера 1 (COMPA).
*/
ISR(TIMER1_COMPA_vect){
//  sei();                  // Дадим возможность обработать прерывания от USB...
//  Clock1_Signalled = 1;
  ADC.ClockSignalCounter++;
  ADC_OnClock();
}


//------------------------------------------------------------------------------
/*! Обработчик прерывания INT6.
*/
ISR(INT6_vect){
  sei();                  // Дадим возможность обработать прерывания от USB...
  ADC.GotSynchroimpulse = true;
}
