////////////////////////////////////////////////////////////////////////////////
/*! @file MSDN2047.c  Firmware for Radiopan Field Control
                      microcontroller board.
- Part of the MSDN firmware.
- Part of the Research Assistant project.
- Uses  LUFA library written by
   - Copyright(C) 2012, Dean Camera, dean [at] fourwalledcubicle [dot] com,
                                     www.lufa-lib.org.

- Copyright(C) 2014, Dmitry K. Nelson, St.Petersburg, Russia.
                     D.Nelson@mail.ioffe.ru
- Copyright(C) 2016, Viktor E. Bursian, St.Petersburg, Russia.
                     Viktor.Bursian@mail.ioffe.ru
*///////////////////////////////////////////////////////////////////////////////
#include "Microsin_USB_IO.h"
#include "CommandParser.h"
#include "ClocksSimple.h"
#include "ADC_AD7321_VB.h"
#include <avr/wdt.h>
#include <avr/power.h>
#include <avr/iousb162.h>
#include <avr/interrupt.h>
#include <math.h>
//#include <util/delay.h>

/*================================ Команды =====================================
==============================================================================*/
/*=============================== Распайка =====================================
Ввод данных переключателей. Данные подаются на бит 2 порта C
                            Бит 4 порта C - тактовый сигнал сдвига.
                            Бит 5 порта C - сигнал записи в регистр сдвига
                                            низким уровнем.
Вывод для отладки		  Бит 5 порта D - контакт P17

Про АЦП см. в ADC_AD7321_VB.c
==============================================================================*/

//! Последнее сообщённое значение переключателей
static uint32_t               ReportedSwitches;

//! Предыдущее считанное значение переключателей (для устранения дребезга).
static uint32_t               PreviousSwitches;

//! Переменная считывания переключателей
static uint32_t               InputSwitches;

static uint8_t DelayVar = 0;       		// Переменная для задержки

void  BreakAndReinitialize (void);
void  GetSwitches (void);


uint32_t  LoopCounter;
uint32_t  Clock0SignalCounter;
uint32_t  Clock0ProcessCounter;

//------------------------------------------------------------------------------

void  ReportSettings (void)
{
  SendStringToUSB("U ");
  if( SendByByte ){
    SendCharToUSB('1');
  }else{
    SendCharToUSB('2');
  }
  if( USBByClock ){
    SendCharToUSB('c');
    Send_EOL_ToUSB();
    SendCharToUSB('p');
    SendNumberToUSB(Clock0_Period());
  }else{
    SendCharToUSB('l');
  }
  Send_EOL_ToUSB();
}

//------------------------------------------------------------------------------

int8_t  ExecuteCommand0 (const char  command)
{
  switch( command ){
	case '/':
        BreakAndReinitialize();
        return 0;
    case 'H':
        ReportedSwitches = 0xFF000000;
        PreviousSwitches = 0xFF000000;
        Clock0_SetToZero();
        Clock0_AllowInterrupts();
        return 0;
    case 'h':
        Clock0_ForbidInterrupts();
        return 0;
    case 'P':
        SendCharToUSB('P');
        SendNumberToUSB(ADC_Period());
        Send_EOL_ToUSB();
        return 0;
    case 'S':
        return ADC_MeasurePeriodically();
    case 's':
        ADC_Stop();
        return 0;
	case 'D':
        ADC.EvaluateNoise = true;
        return 0;
	case 'd':
        ADC.EvaluateNoise = false;
        return 0;
    case 'y':
        ADC_OnSynchroimpulse();
        return 0;
	case '$':
        ReportSettings();
        return 0;
	case '%':
        ADC_ReportParams();
        return 0;
    case 'f':
        SendCharToUSB('f');
        SendNumberToUSB(Clock1_TickFrequency());
        Send_EOL_ToUSB();
        return 0;
    case 'p':
        SendCharToUSB('p');
        SendNumberToUSB(Clock1_Period());
        Send_EOL_ToUSB();
        return 0;
    case 'i':
        SendCharToUSB('i');
        SendNumberToUSB(Clock1_TickInterval());
        Send_EOL_ToUSB();
        return 0;
    case 'A':
        ADC.ReadADC = true;
        return 0;
    case 'a':
        ADC.ReadADC = false;
        return 0;
    case 'E':
        ADC.EmulateADC = true;
        return 0;
    case 'e':
        ADC.EmulateADC = false;
        return 0;
    case 'b':
        SendCharToUSB('b');
        SendNumberToUSB(ADC.SkippedBits);
        Send_EOL_ToUSB();
        return 0;
    case 'o':
        ADC.OutputCounter = 0;
        return 0;
    case 'O':
        SendCharToUSB('O');
        SendNumberToUSB(ADC.OutputCounter);
        Send_EOL_ToUSB();
        ADC.OutputCounter = -1;
        return 0;
    case 'w':
        SendCharToUSB('w');
        SendNumberToUSB(Clock0SignalCounter);
        Send_EOL_ToUSB();
        Clock0SignalCounter = 0;
        return 0;
    case 'c':
        SendCharToUSB('c');
        SendNumberToUSB(ADC.ClockSignalCounter);
        Send_EOL_ToUSB();
        SendCharToUSB(',');
        SendNumberToUSB(ADC.ClockProcessCounter);
        Send_EOL_ToUSB();
        SendCharToUSB('w');
        SendNumberToUSB(Clock0SignalCounter);
        Send_EOL_ToUSB();
        SendCharToUSB(',');
        SendNumberToUSB(Clock0ProcessCounter);
        Send_EOL_ToUSB();
        SendCharToUSB('l');
        SendNumberToUSB(LoopCounter);
        Send_EOL_ToUSB();
        return 0;
    case '1':
        SendByByte = 1;
        ReportSettings();
        return 0;
    case '2':
        SendByByte = 0;
        ReportSettings();
        return 0;
	case 'U':
        USBByClock = 1;
        ReportSettings();
        return 0;
	case 'u':
        USBByClock = 0;
        ReportSettings();
        return 0;
	case 'T':
        EIMSK |= INT6ENABLE;		// Разрешаем внешнее прерывание
        return 0;
	case 'N':
        EIFR |= INT6ENABLE; // Забыть про необработанные сигналы прерывания,
                            // случившиеся пока оно было запрещено.
        EIMSK |= INT6ENABLE;		// Разрешаем внешнее прерывание
        return 0;
	case 'n':
        EIMSK &= ~INT6ENABLE;		// Запрещаем внешнее прерывание
        return 0;
    default:
        return 1;
  }
}


int8_t  ExecuteCommand2 (const char     command
                        ,const uint8_t  argument)
{
  switch( command ){
    case 'R':
        ADC_SetRange(argument);
        return 0;
    case 'Y':
        return ADC_MeasureSynchronized(argument);
    case 'E':
        ADC.EmulateADC = true;
        ADC.NoiseLevel = argument;
        return 0;
    case 'c':
        ADC.ClockSignalCounter = 0;
        ADC.ClockProcessCounter = 0;
        Clock0SignalCounter = 0;
        Clock0ProcessCounter = 0;
        LoopCounter = 0;
        return ExecuteCommand0('c');
    default:
        return 1;
  }
}


int8_t  ExecuteCommand4 (const char      command
                        ,const uint16_t  argument)
{
  switch( command ){
    case 'Y':
        return ADC_MeasureSynchronized(argument);
    case 'E':
        ADC.EmulateADC = true;
        ADC.NoiseLevel = argument;
        return 0;
    default:
        return 1;
  }
}


int8_t  ExecuteCommand8 (const char      command
                        ,const uint32_t  argument)
{
  int8_t                      Err;
  switch( command ){
    case 'P':
        Err = ADC_SetPeriod(argument);
        SendCharToUSB('P');
        SendNumberToUSB(ADC_Period());
        Send_EOL_ToUSB();
        return Err;
    case 'M'://max interval for synchronized mode
        return ADC_SetPeriod(argument);
    case 'F':
        Clock1_SetTickFrequency(argument);
        SendCharToUSB('f');
        SendNumberToUSB(Clock1_TickFrequency());
        Send_EOL_ToUSB();
        return 0;
    case 'I':
        ADC.MinIntervalBetweenReadings = argument;
        Err = ADC_SetPeriod(ADC_Period());
        SendCharToUSB('P');
        SendNumberToUSB(ADC_Period());
        Send_EOL_ToUSB();
        return Err;
    default:
        return 1;
  }
}


void  SetupHardware (void)
{
  /* Disable watchdog if enabled by bootloader/fuses */
  MCUSR &= ~(1 << WDRF);
  wdt_disable();

  /* Disable clock division */
  clock_prescale_set(clock_div_1);

  Clock0_Init();
  Clock0_SetPeriod(10000000);      //10ms
  ADC_Init();

  // Установки для управления чтением переключателей
  DDRC |= 0b00110000; 	// Биты 4,5 на вывод (4 - такты сдвига ввода,
                        //                    5- загрузка входных данных)
  DDRC &= 0b11111011;   // Бит 2 на ввод (ввод данных переключателей)
  PORTC &= 0b11101111;	// Устанавливаем в ноль бит 4
  PORTC |= 0b00100100;	// Устанавливаем в единицу бит 5, подтягиваем в 1 бит 2
}


void  BeforeLoop (void)
{
  //! @todo{MSDN} В оригинале это делается после sei() перед самым началом
  //!             цикла обработки входного потока.  Здесь тоже.
  //!             Но правильно ли это?
  BreakAndReinitialize();
}


void  BreakAndReinitialize (void)
{
  Clock0_ForbidInterrupts();
  ADC_BreakAndReinitialize();
  LoopCounter = 0;
  USBByClock = 0;
  SendByByte = 0;
//  Clock0_AllowInterrupts();
}


void  InLoop (void)
{
//  ADC_Read();// 29-12=17 us
//  ADC_OnClock();//34.6-12=22.6 us (без вывода, но с накоплением,
//                //                 т.е. после команд /|S|o )
//                //а максимальный цикл ~1050 us >1ms!

//  if( Clock1_Signalled ){
//    Clock1_Signalled = 0;
//    ADC_OnClock();
//  }

  if( ADC.ToBeSent ){
    ADC_Report();
    ADC.ToBeSent = 0;
  }
  if( ADC.GotSynchroimpulse ){
    ADC.GotSynchroimpulse = false;
    ADC_OnSynchroimpulse();
  }
  if( Clock0_Signalled ){
    Clock0_Signalled = 0;
    Clock0ProcessCounter++;
    USBClockSignal = 1;
    GetSwitches();
    if( InputSwitches == PreviousSwitches ){
      if( InputSwitches != ReportedSwitches ){
        SendCharToUSB('H');
        SendNumberToUSB(InputSwitches);
        Send_EOL_ToUSB();
        ReportedSwitches = InputSwitches;
      }
    }else{
      PreviousSwitches = InputSwitches;
    }
  }
//  if( ++LoopCounter == 1000000 ){
//    SendCharToUSB('o');
// //    SendStringToUSB("ooooooooooooooooooooooooooooooooooooooooooooooo");
//    Send_EOL_ToUSB();
//    LoopCounter = 0;
//  }
}


//------------------------------------------------------------------------------
/*! Обработчик прерывания от таймера 0 (COMPA).
*/
ISR(TIMER0_COMPA_vect){
  sei();                  // Дадим возможность обработать прерывания от USB...
  Clock0_Signalled = 1;
  Clock0SignalCounter++;
}


//------------------------------------------------------------------------------
/*! Ввод данных переключателей в переменную InputSwitches.

@note По Митиным данным (слегка устаревшим):
Один опрос переключателей (вместе с анализом изменения),
стоящий в основном цикле занимает ~26 мкс при тактовой частоте 16 МГц.
*/
void  GetSwitches (void)
{
  //PORTD |= (1<<PD5); //   тестовый выход 1.

  InputSwitches = 0;
  PORTC &= ~(1 << 5);       // Выводим 0 на пин PORTC5 - загрузка регистра
  DelayVar++;DelayVar++;    // Для задержки
  PORTC |= (1 << 5);        // Выводим 1 на пин PORTC5 - возврат в исх. сост.
  for( int8_t i = 0 ; i < 24 ; i++ ){
    InputSwitches <<= 1;
    if( PINC & (1<<PINC2) )
      InputSwitches |= 1;   // Считываем бит 2 и добавляем его в накопитель
    PORTC |= (1 << 4);      // Выводим 1 на пин PORTC4 - сдвиг регистра
    DelayVar++;	DelayVar++; // Для задержки
    PORTC &= ~(1 << 4);     // Выводим 0 на пин PORTC4 - возврат в исх. сост.
  }
  //PORTD &= ~(1<<PD5); //  тестовый выход 0.
}
