////////////////////////////////////////////////////////////////////////////////
/*! @file MSDN2048.c  Firmware for Radiopan CUP202 Field Scanning
                      microcontroller board.
- Part of the MSDN firmware.
- Part of the Research Assistant project.
- Uses  LUFA library written by
   - Copyright(C) 2012, Dean Camera, dean [at] fourwalledcubicle [dot] com,
                                     www.lufa-lib.org.

- Copyright(C) 2014, Dmitry K. Nelson, St.Petersburg, Russia.
                     D.Nelson@mail.ioffe.ru
- Copyright(C) 2016, Viktor E. Bursian, St.Petersburg, Russia.
                     Viktor.Bursian@mail.ioffe.ru
*///////////////////////////////////////////////////////////////////////////////

//=============================================================================
//  Бит 0 порта D - данные для регистра сдвига
//  Бит 1 порта D - тактовый сигнал регистра сдвига
//  Бит 2 порта D - сигнал записи в регистр защелки
//  Бит 3 порта D - строб запуска измерения
//=============================================================================

#include "Microsin_USB_IO.h"
#include "CommandParser.h"
#include "ClocksSimple.h"
#include <avr/wdt.h>
#include <avr/power.h>
//#include <util/delay.h>

#define ERR_WRONG_FIELD 2
#define ERR_WRONG_SPEED 3
#define ERR_WRONG_RANGE 4


static uint8_t   DelayVar = 0;   		// Переменная для задержки

static uint16_t  CurrentFieldIndex = 0; // 12 bit
static int8_t    ScanStep;

static uint16_t  StepsToGo;
static int8_t    WaitingFirstClockTick;
static int8_t    SubstepsMade;
static int8_t    SkipInterrupt = 0;
static int8_t    GotInterrupt  = 0;

void  BreakAndReinitialize (void);
int8_t  SetFieldIndex (uint16_t  field_index);
int8_t  SetScanStep (int8_t  step);
int8_t  SetTimePerSubstep (uint32_t  nanoseconds_per_substep);
void    SendSynchronization (void);
int8_t  Go (uint16_t  number_of_steps);


int8_t  ExecuteCommand0 (const char  command)
{
  switch( command ){
	case '/':
        BreakAndReinitialize();
        return 0;
	case 'F':
        SendCharToUSB('F');
        SendNumberToUSB(CurrentFieldIndex);
        Send_EOL_ToUSB();
        return 0;
	case 'B':
        StepsToGo = 0;
        Clock1_ForbidInterrupts();
        SendCharToUSB('b');
        Send_EOL_ToUSB();
        return 0;
	case 'I':
        SendSynchronization();
        return 0;
	default:
        return 1;
  }
}


int8_t  ExecuteCommand2 (const char     command
                        ,const uint8_t  argument)
{
  switch( command ){
	case 'F':
        return SetFieldIndex(argument);
	case 'S':
        return SetScanStep(argument);
	case 'T': //time per elementary movement by 1 unit in field index
        return SetTimePerSubstep( argument /*ns*/);
	case 'G':
        return Go( argument /*number of steps*/);
	default:
        return 1;
  }
}


int8_t  ExecuteCommand4 (const char      command
                        ,const uint16_t  argument)
{
  switch( command ){
	case 'F':
        return SetFieldIndex(argument);
	case 'T':
        return SetTimePerSubstep( argument /*ns*/);
	case 'G':
        return Go( argument /*number of steps*/);
	default:
        return 1;
  }
}


int8_t  ExecuteCommand8 (const char      command
                        ,const uint32_t  argument)
{
  switch( command ){
	case 'T':
        return SetTimePerSubstep( argument /*ns*/);
	default:
        return 1;
  }
}


int8_t  SetFieldIndex (uint16_t  field_index)
{
  if( field_index >> 12 )
    return ERR_WRONG_FIELD;
  CurrentFieldIndex = field_index;
  for( char k = 0 ; k < 16 ; k++ ){
    if( field_index & 1 ){
      PORTD |= 1;    // Устанавливаем бит для вывода в 1
    }else{
      PORTD &= ~(1); // или обнуляем его (бит PORTD0)
    }
    PORTD |= (1 << PD1);  	// Выводим 1 на пин PORTD1 - сдвиг регистра
    DelayVar++;	DelayVar++;	// Для задержки
    PORTD &= ~(1 << PD1);  	// Выводим 0 на пин PORTD1 - возврат в исх. сост.
    DelayVar++;	DelayVar++;	// Для задержки
    field_index = field_index >> 1;
  };
  PORTD |= (1 << PD2);  	// Выводим 1 на пин PORTD2 - запись в защелку
  DelayVar++;	DelayVar++;	// Для задержки
  PORTD &= ~(1 << PD2);  	// Выводим 0 на пин PORTD2 - возврат в исх. сост.

  SendCharToUSB('F');
  SendNumberToUSB(CurrentFieldIndex);
  Send_EOL_ToUSB();
  return 0;
}


int8_t  SetScanStep (int8_t  step)
{
  ScanStep = step;
  SendCharToUSB('S');
  SendNumberToUSB(ScanStep);
  Send_EOL_ToUSB();
  return 0;
}


int8_t  SetTimePerSubstep (uint32_t  nanoseconds_per_substep)
{
  Clock1_SetPeriod(nanoseconds_per_substep);
  SendCharToUSB('T');
  SendNumberToUSB( Clock1_Period() );
  Send_EOL_ToUSB();
  return 0;
}


void  SendSynchronization (void)
{
//  PORTD |= 0b00001000;	// Подтягиваем в 1 бит 3
//  DelayVar++; DelayVar++; DelayVar++; DelayVar++; DelayVar++;
  PORTD &= 0b11110111;	// Устанавливаем в ноль бит 3
  DelayVar++; DelayVar++; DelayVar++; DelayVar++; DelayVar++;
  PORTD |= 0b00001000;	// Возвращаем в единицу бит 3

  SendCharToUSB('i');
  Send_EOL_ToUSB();

}


int8_t  Go (uint16_t  number_of_steps)
{
  int32_t  TargetIndex = number_of_steps;
  TargetIndex *= ScanStep;
  TargetIndex += CurrentFieldIndex;
  if( ! ( (0 <= TargetIndex) && (TargetIndex <= 0xFFF ) ) )
    return ERR_WRONG_RANGE;

  SendStringToUSB("GoTo");
  SendNumberToUSB(TargetIndex);
  Send_EOL_ToUSB();

  StepsToGo = number_of_steps;
  SubstepsMade = 0;
  WaitingFirstClockTick = 1;
  GotInterrupt = 0;
  SkipInterrupt = 1;
  Clock1_SetToZero();
  Clock1_ForgetInterrupts();
  Clock1_AllowInterrupts();
  return 0;
}


void  SetupHardware (void)
{
  /* Disable watchdog if enabled by bootloader/fuses */
  MCUSR &= ~(1 << WDRF);
  wdt_disable();

  /* Disable clock division */
  clock_prescale_set(clock_div_1);

  Clock1_Init();
  Clock1_SetTickFrequency(15625);

  //настройка PD0 - PD3 как выходов.
  DDRD |= ((1<<PD0)|(1<<PD1)|(1<<PD2)|(1<<PD3));
  PORTD &= ~(1<<PD0); // у PD0 начальное состояние 0
  PORTD &= ~(1<<PD1); // у PD1 начальное состояние 0
  PORTD &= ~(1<<PD2); // у PD2 начальное состояние 0
  PORTD |=  (1<<PD3); // у PD3 начальное состояние 1
}


void  BeforeLoop (void)
{
  SetFieldIndex(0);
  ScanStep = 1;
  StepsToGo = 0;
  Clock1_ForbidInterrupts();
}


void  BreakAndReinitialize (void)
{
  BeforeLoop();
}


void  InLoop (void)
{
  if( GotInterrupt && StepsToGo ){
    GotInterrupt = 0;
    if( WaitingFirstClockTick ){
      WaitingFirstClockTick = 0;
      SendSynchronization();
      SendCharToUSB('g');
      Send_EOL_ToUSB();
    }else{
      if( ScanStep < 0 ){
        SetFieldIndex( --CurrentFieldIndex );
        SubstepsMade--;
      }else{
        SetFieldIndex( ++CurrentFieldIndex );
        SubstepsMade++;
      }
      if( SubstepsMade == ScanStep ){
        SubstepsMade = 0;
        StepsToGo--;
        SendSynchronization();
        if( StepsToGo == 0 ){
          Clock1_ForbidInterrupts();
          SendCharToUSB('.');
          Send_EOL_ToUSB();
        }
      }
    }
  }
}


//------------------------------------------------------------------------------
/*! Обработчик прерывания от таймера 1 (COMPA).
*/
ISR(TIMER1_COMPA_vect){
  sei(); // Дадим возможность обработать прерывания от USB...
         // Лишнее? Они и так старше.
  if( SkipInterrupt ){
    SkipInterrupt = 0;
  }else{
    GotInterrupt = 1;
  }
}
