////////////////////////////////////////////////////////////////////////////////
/*! @file SOSFileImport.cpp Class sSOSFileImport (imports SOP and SOS files).
- Part of SOS2RA - Research Assistant Plugin package.
- Uses  RAGUI - Research Assistant Graphical User Interface.
- Uses  RANet - Research Assistant Net Library.
- Uses  Qt v.5  - http://qt-project.org/
- Copyright(C) 2000-2015, Viktor E. Bursian, St.Petersburg, Russia.
                          Viktor_dot_Bursian_at_mail_dot_ioffe_dot_ru
*///////////////////////////////////////////////////////////////////////////////
#include "SOSFileImport.h"
#include "General.h"
#include "BasicNodes.h"
#include "attr.h"
#include "Symbols.h"
#include "StringsAndQStrings.h"
#include "Log.h"
#include "Trace.h"
namespace RA {
//------------------------------------------------------------------------------

//                       Turbo Pascal types:
typedef bool                  tP_boolean;
typedef quint16               tP_word;
typedef qint16                tP_integer;
typedef qint32                tP_longint;
typedef float                 tP_single;
typedef double                tP_double;

//----------------------------------------------------------- sSOSFileImport ---

sSOSFileImport::sSOSFileImport (eSOSRelease  release
                               ,bool         preserve_colors)
    :sFile()
    ,SOSRelease(release)
    ,PreserveColors(preserve_colors)
{
}


void  sSOSFileImport::SetSample (rcsString          sample_text
                                ,sPtr<sFolderNode>  sample)
{
  SampleText = sample_text;
  Sample = sample;
}


sPtr<sDataArrayRecord>
              sSOSFileImport::Import_SOP (rcsPath  file_path)
{
  RANet::Log.OpenBlock(sLog::Info,"Import"
                      ,sString("Importing ")+file_path.PathStr());
  FileType = FileTypeSOP;
  Assign(file_path);
  Open(Reading);
  sAttribute                  A( Get_SOS_Object() );
  Close();
  sPtr<sDataArrayRecord>  Imported(A.Value());
  sPtr<sDataArrayRecord>  Result;
  if( Imported.IsCorrect() ){
    try{
      Imported.DelAttr("ColorInd");
    }catch(...){}
    sTime                     FileTime( file_path.ModificationTime() );
    psDataArrayRecord     Spectrum(new sDataArrayRecord
                                                  (0
                                                  ,Imported->Multiplier()
                                                  ,Imported->NumberOfPoints()
                                                  ,Imported->X0()
                                                  ,Imported->Step()          ));
    for( int i=0 ; i < Imported->NumberOfPoints() ; i++ ){
      Spectrum->SetValue(i,Imported->PointY(i));
    };
    Result = sNodePtr(Spectrum);
    if( PreserveColors && Imported->Appearance() ){
      Result->SetAppearance(
          new sCurveAppearance( * dynamic_cast<pcsCurveAppearance>
                                              (Imported->Appearance()) ) );
    }
    if( SOSRelease == SOSRelease_Bursian ){
      Modify_SOP_Bursian(Result,Imported,A.Name().Text(),file_path,FileTime);
    }else if( SOSRelease == SOSRelease_Nelson ){
      Modify_SOP_Nelson(Result,Imported,A.Name().Text(),file_path,FileTime);
    }
  }
  RANet::Log.CloseBlock(sLog::Info,"Import");
  return Result;
}


void  sSOSFileImport::Modify_SOP_Bursian (sPtr<sDataArrayRecord> &
                                                          result
                                         ,sPtr<sDataArrayRecord> &
                                                          imported
                                         ,rcsString       picture_name
                                         ,rcsPath         file_path
                                         ,rcsTime         file_time)
{
  //setting date, time, record number
  if( DetermineDateAndTime(imported,picture_name,file_path,file_time,2010) ){
    result->SetStartTime(imported->StartTime());
    result->SetFinishTime(imported->FinishTime());
    result->SetRecordNo(imported->RecordNo());
  }
  if( SEX_Pars_Transform(result,imported) ){
  }
  for_each_attribute( A , imported ){
    TRACE( (*A).Name().Text() );
    TRACE( (*A).Value()->Text(Plain,Verbose) );
    TRACE( (*A).Value()->TrueType()->Name() );
    result |= *A;
  }end_attr_loop()
}


bool  sSOSFileImport::DetermineDateAndTime(sPtr<sDataArrayRecord>
                                                           record
                                          ,rcsString       picture_name
                                          ,rcsPath         file_path
                                          ,rcsTime         file_time
                                          ,int             decade_start)
{
  sString                     Identifier;
  sString                     TimeStamp; //"17:30:41-17:35:13"
  real                        StartTime = 0.0;
  real                        FinishTime = 0.0;
  if( get_from<sTextNode>(record>>"Identifier",Identifier) )
    record.DelAttr("Identifier");
  if( get_from<sTextNode>(record>>"TimeStamp",TimeStamp) ){
    record.DelAttr("TimeStamp");
    StartTime = sTime::ToSeconds( TimeStamp.Substr(0,8) );
    FinishTime = sTime::ToSeconds( TimeStamp.Substr(9,8) );
    if( FinishTime < StartTime )  FinishTime += 24*3600.0;
  }
  if( ! ( (picture_name == Identifier) && (Identifier==file_path.Name()) ) ){
    RANet::Log.Put(sLog::Warning,"Import",
        sString("picture_name='")+picture_name
        +"' != Identifier='"+Identifier+"'");
    return false;
  }
  int                       Y=-1,M=-1,D=-1,N=-1;
  Identifier.Substr(0,1) >> Y;
  Identifier.Substr(1,2) >> M;
  Identifier.Substr(3,2) >> D;
  Identifier.Substr(6,2) >> N;
  if( ! (   (0 <= Y) && (Y <= 9)
         && (1 <= M) && (M <= 12)
         && (1 <= D) && (D <= 31)   && (1 <= N) ) ){
    RANet::Log.Put(sLog::Warning,"Import",
        sString("Identifier='")+Identifier+"' has wrong format");
    return false;
  }else{
    sTime                       IdTime(decade_start+Y,M,D);
//    TRACE(IdTime.Text());
    if( (decade_start==2010)&&(Y==4)&&(M==1)&&(D==28) ){
      IdTime = sTime(2015,1,28);
    }else if( (decade_start==2010)&&(Y==4)&&(M==1)&&(D==21)&&(N==1) ){
      IdTime = sTime(2015,1,21);
    }else if( (decade_start==2010)&&(Y==4)&&(M==1)&&(D==20)&&(N<=18) ){
      IdTime = sTime(2015,1,20);
    }
//    TRACE(IdTime.Text());
    if( file_time < IdTime ){
      RANet::Log.Put(sLog::Warning,"Import",sString("file_time < IdTime ( ")
         +file_time.Text()+" < "+IdTime.Text()+" )");
      return false;
    }else{
      record->SetStartTime(IdTime+StartTime);
      record->SetFinishTime(IdTime+FinishTime);
//      TRACE(record->StartTime().Text());
//      TRACE(record->FinishTime().Text());
      record->SetRecordNo(N);
      return true;
    }
  }
}


bool  sSOSFileImport::SEX_Pars_Transform (sPtr<sDataArrayRecord>  result
                                         ,sPtr<sDataArrayRecord>  imported)
{
  try{
    sPhysValue                  Parameter;
    real                        Precision;
    sString                     Rem0,Rem1,Rem2,Rem3;
    sString                     Command,Comment,Cavity/*,Sample*/;
    bool                        AddSample = false;
    sPhysValue                  TauRC(-1,"ms");
    bool                        RT = false;
    bool                        LNT = false;

//    TRACE("translation begin**************************************************")
    if( get_from<sTextNode>(imported>>"Rem0",Rem0) )
      imported.DelAttr("Rem0");
    if( get_from<sTextNode>(imported>>"Rem1",Rem1) )
      imported.DelAttr("Rem1");
    if( get_from<sTextNode>(imported>>"Rem2",Rem2) )
      imported.DelAttr("Rem2");
    if( get_from<sTextNode>(imported>>"Rem3",Rem3) )
      imported.DelAttr("Rem3");

    Command = Rem0;

//    TRACE(Rem1);
//    TRACE(Rem2);
//    TRACE(Rem3);

    //cavity recognition
    {
      size_t                      P;
      P = Rem3.Pos("RX102#");
      if( P < Rem3.Length() ){
        Cavity = Rem3.Substr(P,9);
        Rem3 = Rem3.Substr(0,P)+Rem3.Tail(P+9);
      }
      if( Rem3.WipeOut("RX102") ){
        Cavity = "RX102";
      //}else if( Rem3.WipeOut("") ){
      //  Cavity = "";
      }else if( Rem3.WipeOut("Tolparov's") ){
        Cavity = "Tolparov";
      }else if( Rem3.WipeOut("Tolparov") ){
        Cavity = "Tolparov";
      }
    }
    //TauRC recognition
    {
      if( Rem3.WipeOut("3ms") ){
        TauRC = sPhysValue(3,0.5,"ms");
      }else if( Rem3.WipeOut("10ms") ){
        TauRC = sPhysValue(10,1,"ms");
      }else if( Rem3.WipeOut("30ms") ){
        TauRC = sPhysValue(30,1,"ms");
      }else if( Rem3.WipeOut("100ms") ){
        TauRC = sPhysValue(100,1,"ms");
      }else if( Rem3.WipeOut("300ms") ){
        TauRC = sPhysValue(300,1,"ms");
      }
    }
    Rem1 = FromQString(ToQString(Rem1).simplified());
    Rem2 = FromQString(ToQString(Rem2).simplified());
    Rem3 = FromQString(ToQString(Rem3).simplified());
    Comment = Rem1;
    if( ! Rem2.Empty() )
      (Comment += "; ") += Rem2;
    if( ! Rem3.Empty() )
      (Comment += "; ") += Rem3;

    //T recognition
    {
      LNT =  Comment.WipeOut(" LNT",true)
          || Comment.WipeOut(";LNT",true);
      RT =   Comment.WipeOut(" RT",true)
          || Comment.WipeOut(";RT",true);
    }
    AddSample = ExtractSample(Comment) && Sample.IsCorrect();
    CleanComment(Comment);
    if( ! Comment.Empty() )
      result << attr( Comment );
    if( RT ){
      result << attr("T","RT");
    }else if( LNT ){
      result << attr("T","LNT");
    }
    if( get_from<sPhysValueNode>(imported>>"Freq",Parameter) ){
      Precision = ( 100*Parameter.ToReal()-Round(100*Parameter.ToReal()) < 0.001
                  ? 1e-2
                  : 1e-5 );
      result << attr("&nu;" ,sPhysValue(Parameter.ToReal(),Precision,"GHz") );
      imported.DelAttr("Freq");
      imported.DelAttr("FreqTitle");
    }
    if( get_from<sPhysValueNode>(imported>>"A",Parameter) ){
      Precision = ( Parameter.ToReal() < 3.0
                  ? 0.1
                  : 1.0 );
      result << attr("A" ,sPhysValue(Parameter.ToReal(),Precision,"dB") );
      imported.DelAttr("A");
      imported.DelAttr("ATitle");
    }
    if( get_from<sPhysValueNode>(imported>>"G",Parameter) ){
      result << attr("G" ,sPhysValue(Parameter.ToReal(),0.05*Parameter.ToReal()
                                    ,_Unitsless_) );
      imported.DelAttr("G");
      imported.DelAttr("GTitle");
    }
    if( get_from<sPhysValueNode>(imported>>"MA",Parameter) ){
      result << attr("M" ,sPhysValue(Parameter.ToReal(),0.05*Parameter.ToReal()
                                    ,"Oe") );
      imported.DelAttr("MA");
      imported.DelAttr("MATitle");
    }
    sPtr<sFolderNode>       Technique = (result >= folder("technique",Command));
    if( get_from<sPhysValueNode>(imported>>"Step",Parameter) ){
      Technique << attr("step" ,sPhysValue(Parameter.ToReal()
                                          ,0.005*Parameter.ToReal(),"Oe") );
      imported.DelAttr("Step");
      imported.DelAttr("StepTitle");
    }
    if( get_from<sPhysValueNode>(imported>>"Ho",Parameter) ){
      Technique << attr("H<sub>0</sub>" ,sPhysValue(Parameter.ToReal()
                                                   ,0.05,"Э") );
      imported.DelAttr("Ho");
      imported.DelAttr("HoTitle");
    }
    if( get_from<sPhysValueNode>(imported>>"dH",Parameter) ){
      Technique << attr("&Delta;H" ,sPhysValue(Parameter.ToReal(),0.5,"Э") );
      imported.DelAttr("dH");
      imported.DelAttr("dHTitle");
    }
    sString                     MMode;
    if( get_from<sTextNode>(imported>>"MF",MMode) ){
      Technique << attr("mode" ,MMode );
      imported.DelAttr("MF");
      imported.DelAttr("MFTitle");
    }
    if( sPhysValue(0,"ms") < TauRC ){
      Technique << attr("&tau;<sub>RC</sub>" ,TauRC );
    }
    integer                     Repetitions = -1;
    if( get_from<sIntegerNode>(imported>>"Times",Repetitions) ){
      Technique << attr("repetitions" ,Repetitions );
      imported.DelAttr("Times");
      imported.DelAttr("TimesTitle");
    }
    sString                     BurstDempheringStr;
    integer                     BurstDemphering;
    if( get_from<sTextNode>(imported>>"BurstDempherCounter"
                                         ,BurstDempheringStr) ){
      BurstDempheringStr >> BurstDemphering;
      if( BurstDemphering != 0 )
        Technique << attr("burst demphering" ,BurstDemphering );
      TRACE(BurstDemphering);
      imported.DelAttr("BurstDempherCounter");
    }
    if( ! Cavity.Empty() ){
      Technique << attr("cavity" ,Cavity );
    }
    if( AddSample ){
      result << attr("sample",Sample);
    }
    result->SetMultiplier(sPhysPair(imported->Multiplier().X * sPhysValue("Э")
                                   ,imported->Multiplier().Y * sPhysValue("mV")
                                   ));
//    TRACE("translation end ***************************************************")
  }catch( rxRAlgebra E ){
    TRACE("SOS spectrum pars parsing xRAlgebra")
    E.Report2cerr();
    return false;
  }catch(...){
    TRACE("SOS spectrum pars parsing error")
    return false;
  }
  return true;
}


bool  sSOSFileImport::ExtractSample (rsString  text)
{
  return SampleText.Empty() || text.WipeOut(SampleText,true);
}


void  sSOSFileImport::CleanComment (rsString  text)
{
  text = FromQString(ToQString(text).simplified());
  if( text[0] == ';' )  text = text.Tail(1);
  if( text[text.Length()-1] == ';' )  text = text.Substr(0,text.Length()-1);
  while( text.Replace("; ;",";") ){}
  text = FromQString(ToQString(text).simplified());
}


void  sSOSFileImport::Modify_SOP_Nelson (sPtr<sDataArrayRecord> &
                                                         result
                                        ,sPtr<sDataArrayRecord> &
                                                         imported
                                        ,rcsString       /*picture_name*/
                                        ,rcsPath         /*file_path*/
                                        ,rcsTime         /*file_time*/)
{
  result->SetMultiplier(sPhysPair(sPhysValue(1,"nm"),sPhysValue(1,"parot")));
  try{
    for_each_attribute( A , imported ){
      result |= *A;
    }end_attr_loop()
  }catch(...){}
}


sAttribute  sSOSFileImport::Get_SOS_Object ()
{
  quint16                     IOC;
  sAttribute                  A((sAttrName()),(sNodePtr()));
  READ_INTO(IOC);
  if( IOC == 0 ){
  }else if( IOC <= 25 ){
    switch( IOC ){
      //case  0:   A= new sAttribute("",NULL);  break;
      //case  1:   A=Load_SOS_AnalogField();  break;
      case  2:   A=Load_SOS_Spectrum();  break;
      case  3:   A=Load_SOS_Figure();  break;
      case  4:   A=Load_SOS_Window();  break;
      //case  5:   A=Load_SOS_Panel();  break;
      //case  6:   A=Load_SOS_Cross();  break;
      case  7:   A=Load_SOS_Curve();  break;
      //case  8:   A=Load_SOS_FileString();  break;
      //case  9:   A=Load_SOS_FileViewer();  break;
      //case 10:   A=Load_SOS_Relief();  break;         !!! happens in SHFS
      case 11:   A=Load_SOS_Remark();  break;
      case 12:   A=Load_SOS_Field();  break;
      case 13:   A=Load_SOS_IntegerField();  break;
      case 14:   A=Load_SOS_LongintField();  break;
      case 15:   A=Load_SOS_StringField();  break;
      case 16:   A=Load_SOS_FixedField();  break;
      case 17:   A=Load_SOS_FloatField();  break;
      //case 18:   A=Load_SOS_CommandProcessor();  break;
      //case 19:   A=Load_SOS_Macro();  break;
      //case 20:   A=Load_SOS_Variable();  break;
      //case 21:   A=Load_SOS_Command();  break;
      //case 22:   A=Load_SOS_MemoryField();  break;
      //case 23:   A=Load_SOS_Cartridge();  break;
      //case 24:   A=Load_SOS_Cartridge();  break;
      //case 25:   A=Load_SOS_Cartridge();  break;
      default:   TRACE(IOC);  break;
    }
  }else if( (IOC <= 28) && (SOSRelease == SOSRelease_Bursian ) ){
    switch( IOC ){
      //case 26:   A=Load_SOS_MagneticFieldDriver();  break;
      //case 27:   A=Load_SOS_ESRMeter();  break;
      //case 28:   A=Load_SOS_SEX();  break;
      default:   TRACE(IOC);  break;
    }
  }else if( (IOC <= 28) && (SOSRelease == SOSRelease_Nelson ) ){
    switch( IOC ){
      //case 26:   A=Load_SOS_WaveLengthDriver();  break;
      //case 27:   A=Load_SOS_DigitalVoltmeter();  break;
      //case 28:   A=Load_SOS_MDR();  break;
      default:   TRACE(IOC);  break;
    }
  }else if( IOC <= 31 ){
    switch( IOC ){
      case 29:   A=Load_SOS_Picture();  break;
      //case 30:   A=Load_SOS_PicturesList();  break;
      case 31:   A=Load_SOS_ColorIndicator();  break;
      default:   TRACE(IOC);  break;
    }
  }else if( SOSRelease == SOSRelease_Bursian ){
    switch( IOC ){
      //case 33:   A=________________();  break;           !!!! happens
      default:   TRACE(IOC);  break;
    }
  }else if( SOSRelease == SOSRelease_Nelson ){
    switch( IOC ){
      //case 32:   A=Load_SOS_GratingNo();  break;
      //case 33:   A=Load_SOS_FreqMeter();  break;
      //case 34:   A=Load_SOS_Fitter();  break;
      //case 35:   A=Load_SOS_Fitter();  break;
      //case 36:   A=Load_SOS_Fitter();  break;
      //case 37:   A=Load_SOS_DFSWLDriver();  break;
      //case 38:   A=Load_SOS_DFS();  break;
      //case 39:   A=Load_SOS_DifMeter();  break;
      default:   TRACE(IOC);  break;
    }
  }
  return A;
}


sString  sSOSFileImport::Load_SOS_Item ()
{
  const int                   CartNameSize = 24;
  byte                        L;
  char                        N[CartNameSize+1];
  for( int i=0 ; i < CartNameSize+1 ; i++ )  N[i]='\0';
  READ_INTO(L);
  if(  L < 0x40  ){
    read((char*)N,CartNameSize);
  }else{
    L-=0x40;
    read((char*)N,L);
  };
  Skip(8+1);
  return sString(N);
}


sAttribute  sSOSFileImport::Load_SOS_Remark ()
{
  sString                     AName( Load_SOS_Item() );
  short int                   Size;
  byte                        L;
  char                        PictureBody[256];

  for( int i=0 ; i < 256 ; i++ )  PictureBody[i]='\0';
  READ_INTO(Size);
  READ_INTO(L);
  read((char*)PictureBody,L);
  Skip(8+16+4);
  return sAttribute(sAttrName(AName)
                   ,sNodePtr(new sTextNode(DOStoUTF8(PictureBody))));
}


sAttribute  sSOSFileImport::Load_SOS_Field ()
{
  sAttribute                  A = Load_SOS_Remark();
  Skip(/*CursorPos*/ sizeof(tP_integer)
      +/*Virgin*/ sizeof(tP_boolean) );
  return A;
}


sAttribute  sSOSFileImport::Load_SOS_IntegerField ()
{
  sAttribute                  A = Load_SOS_Field();
  tP_integer                  Value;
  READ_INTO(Value);
  return sAttribute( A.Name() ,sNodePtr(new sIntegerNode(Value)) );
}


sAttribute  sSOSFileImport::Load_SOS_LongintField ()
{
  sAttribute                  A = Load_SOS_Field();
  tP_longint                  Value;
  READ_INTO(Value);
  return sAttribute( A.Name() ,sNodePtr(new sIntegerNode(Value)) );
}


sAttribute  sSOSFileImport::Load_SOS_StringField ()
{
  sAttribute                  A = Load_SOS_Field();
  byte                        L;
  char                        Value[256];

  for( int i=0 ; i < 256 ; i++ )  Value[i]='\0';
  READ_INTO(L);
  if( L == 0xFF ){
    READ_INTO(L);
    read((char*)Value,L);
  }else{
    read((char*)Value,255);
  };
  return sAttribute( A.Name() ,sNodePtr(new sTextNode(DOStoUTF8(Value))) );
}


sAttribute  sSOSFileImport::Load_SOS_FixedField ()
{
  sAttribute                  A = Load_SOS_Field();
  tP_double                   Value;
  tP_integer                  Decimals;
  READ_INTO(Value);
  READ_INTO(Decimals);
  return sAttribute( A.Name() ,sNodePtr(new sPhysValueNode(sPhysValue(Value))));
}


sAttribute  sSOSFileImport::Load_SOS_FloatField ()
{
  sAttribute                  A = Load_SOS_Field();
  tP_double                   Value;
  READ_INTO(Value);
  return sAttribute( A.Name() ,sNodePtr(new sPhysValueNode(sPhysValue(Value))));
}


sAttribute  sSOSFileImport::Load_SOS_Figure ()
{
  sNodePtr                    FolderNode(new sFolderNode);
  sString                     CName = Load_SOS_Item();
  sAttribute                  A = Get_SOS_Object();
  while( ! A.Value().IsNULL() ){
    FolderNode << A;
    A = Get_SOS_Object();
  };
  return sAttribute(sAttrName(CName),FolderNode);
}


sAttribute  sSOSFileImport::Load_SOS_Window ()
{
  sAttribute                  A = Load_SOS_Figure();
  Skip(4*4+2*2+2);
  return A;
}


psCurveAppearance  sSOSFileImport::Load_SOS_Line (rsString  name)
{
  name = Load_SOS_Item();
  tP_word                     Color,Style,Width;
  sColor                      LineColor;
  psCurveAppearance           CA(NULL);
  bool                        Default = false;
  READ_INTO(Color);
  READ_INTO(Style);
  READ_INTO(Width);
  if( PreserveColors ){
    switch( Color ){
        case  1:  LineColor=sColor(0,0,127);  break;
        case  2:  LineColor=sColor(0,127,0);  break;
        case  3:  LineColor=sColor(0,127,127);  break;
        case  4:  LineColor=sColor(127,0,0);  break;
        case  5:  LineColor=sColor(127,0,127);  break;
        case  6:  LineColor=sColor(127,127,0);  break;
        case  7:  LineColor=sColor(150,150,150);  break;
        case  8:  LineColor=sColor(100,100,100);  break;
        case  9:  LineColor=sColor(0,0,255);  break;
        case 10:  LineColor=sColor(0,255,0);  break;
        case 11:  LineColor=sColor(0,255,255);  break;
        case 12:  LineColor=sColor(255,0,0);  break;
        case 13:  LineColor=sColor(255,0,255);  break;
        case 14:  LineColor=sColor(255,255,0);  break;
        default:  Default=true;  break;
    };
    if( ! Default )
      CA = new sCurveAppearance(LineColor,Width);
  };
  return CA;
}


sAttribute  sSOSFileImport::Load_SOS_Curve ()
{
  sString                     LineName;
  psCurveAppearance           CA = Load_SOS_Line(LineName);
  tP_integer                  Size;
  bool                        Press;
  tP_single                   MultiplierX;
  tP_single                   MultiplierY;
  tP_single                   OriginX;
  tP_single                   Quantum,Shift;
  tP_single                   Val;
  tP_integer                  V;
  psDataArrayRecord       Curve;

  READ_INTO(Size);
  Press=(Size < 0);
  if( Press )
    Size=-Size;
  READ_INTO(MultiplierX);
  READ_INTO(MultiplierY);
  Curve= new sDataArrayRecord(Size,0,MultiplierX);
  if( ! Press ){
    for( int x=0 ; x < Size ; x++ ){
      READ_INTO(Val);
      Curve->SetValue( x ,double(MultiplierY)*Val );
    };
  }else{
    READ_INTO(Shift);
    READ_INTO(Quantum);
    for( int x=0 ; x < Size ; x++ ){
      READ_INTO(V);
      Curve->SetValue( x ,double(MultiplierY)*(Shift + V * Quantum) );
    };
  };
  READ_INTO(OriginX);
  Curve->SetX0(OriginX);
  Curve->SetMultiplier(sPhysPair(sPhysValue(1),sPhysValue(1)));
  Curve->SetAppearance(CA);
  return sAttribute((sAttrName()),sNodePtr(Curve));
}


sAttribute  sSOSFileImport::Load_SOS_Picture ()
{
  sAttribute                  Window = Load_SOS_Window();
  sAttribute                  Body = Get_SOS_Object();
  sPtr<sFolderNode>           Curve( Body.Value() );
  if( Curve.IsCorrect() ){
    for_each_attribute( A , Window.Value() ){
      Curve |= *A;
    }end_attr_loop()
  }
  return sAttribute(Window.Name(),Curve);
}


sAttribute  sSOSFileImport::Load_SOS_ColorIndicator ()
{
  return Load_SOS_Window();
}


sAttribute  sSOSFileImport::Load_SOS_Spectrum ()
{
  sAttribute                  A((sAttrName()),(sNodePtr()));
  sString                     FormatName;
  sAttribute                  Body((sAttrName()),(sNodePtr()));
  sNodePtr                    Spectrum;
  byte                        L;
  tP_integer                  I;
  tP_double                   X;
  char                        SV [256];
  csString                    DummyString( "$%Z" );
  //tP_double                   DummyValue = 123456789e123;

  if( FileType == FileTypeSOS ){
    A=Load_SOS_Picture();
    read((char*)&L,1);  read((char*)SV,L); SV[L]='\0';
    if( ! A.Value().IsNULL() )
      A.Value() << sAttribute(sAttrName("FormatName")
                             ,sNodePtr(new sTextNode(SV)) );
  }else{
    read((char*)&L,1);  read((char*)SV,L); SV[L]='\0';
    FormatName=SV;
    Body=Get_SOS_Object();
    Spectrum=Body.Value();
    A=sAttribute(sAttrName(/*FileLocation.Name*/),Spectrum);
//                                ^^^!
//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    if( FormatName == "MDR-format" ){
      read((char*)&L,1);  read((char*)SV,L); SV[L]='\0';
                     Spectrum<<attr("Title",DOStoUTF8(SV));
      READ_INTO(I);  Spectrum<<attr("GratingNo",I);

      //!@todo{Nelson} Slit, StWL, et cetera should have units

      READ_INTO(X);  Spectrum<<attr("Slit",X);
      READ_INTO(X);  Spectrum<<attr("StWL",X);
      READ_INTO(X);  Spectrum<<attr("FinWL",X);
      READ_INTO(X);  Spectrum<<attr("Step",X);
      READ_INTO(X);  Spectrum<<attr("Termostat",X);
      read((char*)&L,1);  read((char*)SV,L); SV[L]='\0';
                     Spectrum<<attr("Type",DOStoUTF8(SV));
      READ_INTO(I);  Spectrum<<attr("Counts",I);
      read((char*)&L,1);  read((char*)SV,L); SV[L]='\0';
                     Spectrum<<attr("Matter",DOStoUTF8(SV));
      read((char*)&L,1);  read((char*)SV,L); SV[L]='\0';
                     Spectrum<<attr("Sample",DOStoUTF8(SV));
      read((char*)&L,1);  read((char*)SV,L); SV[L]='\0';
                     Spectrum<<attr("Comment",DOStoUTF8(SV));
    }else if( FormatName == "FitFormat" ){
      read((char*)&L,1);  read((char*)SV,L); SV[L]='\0';
                     Spectrum<<attr("Title",DOStoUTF8(SV));
      READ_INTO(X);  Spectrum<<attr("From",X);
      READ_INTO(X);  Spectrum<<attr("Til",X);
      READ_INTO(I);  Spectrum<<attr("Points",I);
      for( int i=1 ; i<=8 ; i++ ){
        read((char*)&L,1);  read((char*)SV,L); SV[L]='\0';
        READ_INTO(X);
        if( X < 1e131 ){
          if( DummyString == SV ){
            switch( i ){
              case 1:  Spectrum<<attr("Par1",X); break;
              case 2:  Spectrum<<attr("Par2",X); break;
              case 3:  Spectrum<<attr("Par3",X); break;
              case 4:  Spectrum<<attr("Par4",X); break;
              case 5:  Spectrum<<attr("Par5",X); break;
              case 6:  Spectrum<<attr("Par6",X); break;
              case 7:  Spectrum<<attr("Par7",X); break;
              case 8:  Spectrum<<attr("Par8",X); break;
            };
          }else{
            Spectrum<<attr(sAttrName(DOStoUTF8(SV)) ,sPhysValue(X) );
          };
        };
      };
    }else{
      //throw       A = new sAttribute("",NULL);
    };
  };
  return A;
}


void  sSOSFileImport::Skip (int  number)
{
  byte B;
  for( int i=0 ; i < number ; i++ )
    READ_INTO(B);
}


sString  sSOSFileImport::DOStoUTF8 (literal  src) const
{
  sString                     R;
  literal                     NaC = "\357\277\275"; // �  U+FFFD
  for( int i = 0 ; src[i] != '\0' ; i++ ){
    if(       (char(0) <= src[i]) && (src[i] <= char(31)) ){
      R+=NaC;
    }else if( (char(32) <= src[i]) && (src[i] <= char(59)) ){
      R += src[i];
    }else if( src[i] == '<' ){
      R += "∠";
//      R += "&lt;";
    }else if( char(61) <= src[i] ){
      R += src[i];
    }else{
      switch( src[i] ){
        case char(128):  R+="А";  break;
        case char(129):  R+="Б";  break;
        case char(130):  R+="В";  break;
        case char(131):  R+="Г";  break;
        case char(132):  R+="Д";  break;
        case char(133):  R+="Е";  break;
        case char(134):  R+="Ж";  break;
        case char(135):  R+="З";  break;
        case char(136):  R+="И";  break;
        case char(137):  R+="Й";  break;
        case char(138):  R+="К";  break;
        case char(139):  R+="Л";  break;
        case char(140):  R+="М";  break;
        case char(141):  R+="Н";  break;
        case char(142):  R+="О";  break;
        case char(143):  R+="П";  break;
        case char(144):  R+="Р";  break;
        case char(145):  R+="С";  break;
        case char(146):  R+="Т";  break;
        case char(147):  R+="У";  break;
        case char(148):  R+="Ф";  break;
        case char(149):  R+="Х";  break;
        case char(150):  R+="Ц";  break;
        case char(151):  R+="Ч";  break;
        case char(152):  R+="Ш";  break;
        case char(153):  R+="Щ";  break;
        case char(154):  R+="Ъ";  break;
        case char(155):  R+="Ы";  break;
        case char(156):  R+="Ь";  break;
        case char(157):  R+="Э";  break;
        case char(158):  R+="Ю";  break;
        case char(159):  R+="Я";  break;
        case char(160):  R+="а";  break;
        case char(161):  R+="б";  break;
        case char(162):  R+="в";  break;
        case char(163):  R+="г";  break;
        case char(164):  R+="д";  break;
        case char(165):  R+="е";  break;
        case char(166):  R+="ж";  break;
        case char(167):  R+="з";  break;
        case char(168):  R+="и";  break;
        case char(169):  R+="й";  break;
        case char(170):  R+="к";  break;
        case char(171):  R+="л";  break;
        case char(172):  R+="м";  break;
        case char(173):  R+="н";  break;
        case char(174):  R+="о";  break;
        case char(175):  R+="п";  break;
        case char(176):  R+=NaC;  break;
        case char(177):  R+=NaC;  break;
        case char(178):  R+=NaC;  break;
        case char(179):  R+="|";  break;
        case char(180):  R+=NaC;  break;
        case char(181):  R+=NaC;  break;
        case char(182):  R+=NaC;  break;
        case char(183):  R+=NaC;  break;
        case char(184):  R+=NaC;  break;
        case char(185):  R+=NaC;  break;
        case char(186):  R+="∥";  break;
        case char(187):  R+=NaC;  break;
        case char(188):  R+=NaC;  break;
        case char(189):  R+=NaC;  break;
        case char(190):  R+=NaC;  break;
        case char(191):  R+=NaC;  break;
        case char(192):  R+=NaC;  break;
        case char(193):  R+="⟂";  break;
        case char(194):  R+=NaC;  break;
        case char(195):  R+=NaC;  break;
        case char(196):  R+=NaC;  break;
        case char(197):  R+=NaC;  break;
        case char(198):  R+=NaC;  break;
        case char(199):  R+=NaC;  break;
        case char(200):  R+=NaC;  break;
        case char(201):  R+=NaC;  break;
        case char(202):  R+=NaC;  break;
        case char(203):  R+=NaC;  break;
        case char(204):  R+=NaC;  break;
        case char(205):  R+=NaC;  break;
        case char(206):  R+=NaC;  break;
        case char(207):  R+=NaC;  break;
        case char(208):  R+=NaC;  break;
        case char(209):  R+=NaC;  break;
        case char(210):  R+=NaC;  break;
        case char(211):  R+=NaC;  break;
        case char(212):  R+=NaC;  break;
        case char(213):  R+=NaC;  break;
        case char(214):  R+=NaC;  break;
        case char(215):  R+=NaC;  break;
        case char(216):  R+=NaC;  break;
        case char(217):  R+=NaC;  break;
        case char(218):  R+=NaC;  break;
        case char(219):  R+=NaC;  break;
        case char(220):  R+=NaC;  break;
        case char(221):  R+=NaC;  break;
        case char(222):  R+=NaC;  break;
        case char(223):  R+=NaC;  break;
        case char(224):  R+="р";  break;
        case char(225):  R+="с";  break;
        case char(226):  R+="т";  break;
        case char(227):  R+="у";  break;
        case char(228):  R+="ф";  break;
        case char(229):  R+="х";  break;
        case char(230):  R+="ц";  break;
        case char(231):  R+="ч";  break;
        case char(232):  R+="ш";  break;
        case char(233):  R+="щ";  break;
        case char(234):  R+="ъ";  break;
        case char(235):  R+="ы";  break;
        case char(236):  R+="ь";  break;
        case char(237):  R+="э";  break;
        case char(238):  R+="ю";  break;
        case char(239):  R+="я";  break;
        case char(240):  R+="≡";  break;
        case char(241):  R+="ё";  break;
        case char(242):  R+="≥";  break;
        case char(243):  R+="≤";  break;
        case char(244):  R+=NaC;  break;
        case char(245):  R+=NaC;  break;
        case char(246):  R+="÷";  break;
        case char(247):  R+="≈";  break;
        case char(248):  R+=Symbol::Degree;  break;
        case char(249):  R+=NaC;  break;
        case char(250):  R+=NaC;  break;
        case char(251):  R+="√";  break;
        case char(252):  R+=NaC;  break;
        case char(253):  R+=NaC;  break;
        case char(254):  R+=NaC;  break;
        case char(255):  R+=NaC;  break;
        default:              break;
      }
    }
  }
  return R;
}

//------------------------------------------------------------------------------
} //namespace RA
