////////////////////////////////////////////////////////////////////////////////
/*! @file SR400Counter.cpp   Управление счетчиком SR400.
- Part of SR400Drivers - a Research Assistant Plugin package.
- Uses  RAGUI - the Research Assistant Graphical User Interface.
- Uses  QtGui v.5.0  -  http://qt.nokia.com/
- Uses  RANet - Research Assistant Net Library (based on ANSI C++).
- Copyright(C) 2010-2016, Dmitry K. Nelson, St.Petersburg, Russia.
                          D.Nelson@mail.ioffe.ru
*///////////////////////////////////////////////////////////////////////////////
#include "SR400Counter.h"
#include "attr.h"
#include "StringsAndQStrings.h"
#include <QDebug>
namespace RA {
//------------------------------------------------------------------------------

DEFINE_CLASS_TAG(sSR400Counter)

//---------------------------------------------------------- sSR400Counter ---

void  sSR400Counter::Preconfigure (sNodePtr  DeviceCfg)
{
  sPtr<sFolderNode>              DeviceSettings;
  sPtr<sFolderNode>              DeviceLinks;
  DeviceCfg |= "Links";
  DeviceLinks = DeviceCfg >> "Links";
  if( DeviceLinks.IsCorrect() ){
    DeviceLinks |= "SR400Port";
    DeviceLinks >> "SR400Port" |= attr("Class","sTeleCom");
  }
  DeviceCfg |= "Settings";
  DeviceSettings = DeviceCfg >> "Settings";
  if( DeviceSettings.IsCorrect() ){
    DeviceSettings |= attr("MinimumPeriod",sPhysValue(0.04,"s"));
  }
  sGauge::Preconfigure(DeviceCfg);
}


sSR400Counter::~sSR400Counter ()
{
  Stop();
  SR400Port.SetToNULL();
  if( Clock ){ delete Clock; Clock=NULL; }
}

sSR400Counter::sSR400Counter (sString   DevicetId
                       ,sNodePtr  DeviceCfg)
    :sGauge(DevicetId,DeviceCfg)
    ,SR400Port(this)
    ,Clock(new QTimer(this))
    ,PollingInterval(100)
    ,CountingTime("s")
    ,MinPeriod(0.04,"s")
    ,ChNum(1)
{
}

void  sSR400Counter::PostConstructor ()
{
  sGauge::PostConstructor();

  try{ //get config parameters
    get_from<sPhysValueNode>(DeviceSettings >> "MinimumPeriod" ,MinPeriod);
  }catch(...){
    SetInvalid(tr("Something wrong in configuration."));
  }
  WantedPeriod = ThePeriod;

  //! Реализован только режим AutoRepeat:
  AutoRepeatFlag = true;

  //evolution of the inherited state machine (behaviour)
  {
    //evolution of the State_Warming
    {
      State_W_SR400Port = new sState(State_Warming()
                                   ,"State_W_SR400Port");
      State_W_SettingPeriod = new sState(State_Warming()
                                   ,"State_W_SettingPeriod");
      State_W_SR400Port
          ->SetAsInitial()
          ->AddTransition(State_W_SettingPeriod)
          ;
      State_W_SettingPeriod
          ->OnEnterCall(SLOT(SetWantedPeriod()))
          ->AddReaction<sTelex>(SIGNAL(GotSR400Message(sTelex))
                               ,SLOT(CorrectPeriod(sTelex)) )
          ->AddTransition(SIGNAL(PeriodChanged(sPhysValue))
                         ,State_Warming()->Fin())
          ;
      //State_W_SR400Port structure and behaviour
      {
        State_W_SR400Port_Linking = new sState(State_W_SR400Port
                                       ,"State_W_SR400Port_Linking");
        State_W_SR400Port_Flashing = new sState(State_W_SR400Port
                                        ,sPhysValue(1,"s")
                                       ,"State_W_SR400Port_Flashing");
        State_W_SR400Port_Checking = new sState(State_W_SR400Port
                                        ,sPhysValue(1,"s")
                                       ,"State_W_SR400Port_Checking");

        State_W_SR400Port_Linking
            ->SetAsInitial()
            ->OnEnterCall(SLOT(LinkToSR400Port()))
            ->AddTransition(SIGNAL(SR400PortLinked())
                           ,State_W_SR400Port_Flashing )
            ->AddTransition(SIGNAL(ErrorMessage()) ,State_Cooling() )
            ;
        State_W_SR400Port_Flashing
            ->OnEnterCall(SLOT(SendDummyCommand()))
            ->AddTransition(SIGNAL(NoAnswerFromSR400())
                           ,State_W_SR400Port_Checking )
            ->AddTransition(SIGNAL(GotSR400Message(sTelex))
                                  ,State_W_SR400Port_Checking)
            ;
        State_W_SR400Port_Checking
            ->OnEnterCall(SLOT(RequestSR400ID()))
            ->AddReaction<sTelex>(SIGNAL(GotSR400Message(sTelex))
                                 ,SLOT(ParseSR400ID(sTelex))  )
            ->AddTransition(SIGNAL(NoAnswerFromSR400())
                           ,SLOT(OnNoAnswerFromSR400())
                           ,State_W_SR400Port_Checking )
            ->AddTransition(SIGNAL(SR400PortError()) ,State_Cooling() )
            ->AddTransition(SIGNAL(ErrorMessage()) ,State_Cooling() )
            ->AddTransition(SIGNAL(GotSR400ID())
                           ,State_W_SR400Port->Fin()       )
            ;
        connect(State_W_SR400Port_Flashing,SIGNAL(Timeout())
               ,this,SIGNAL(NoAnswerFromSR400()));
        connect(State_W_SR400Port_Checking,SIGNAL(Timeout())
               ,this,SIGNAL(NoAnswerFromSR400()));
      }
    }
    //evolution of the State_Operating
    {
      State_O_M_SetP_Checking = new sState(State_O_M_SettingPeriod
                                         ,"State_O_M_SetP_Checking");
      State_O_M_SetP_Checking
          ->SetAsInitial()
          ->AddTransition<sTelex>(SIGNAL(GotSR400Message(sTelex))
                                 ,SLOT(CorrectPeriod(sTelex))
                                 ,State_O_M_SettingPeriod->Fin() )
              ;
      State_O_M_Acquiring
          ->OnEnterCall(SLOT(Start()))
          ->OnExitCall(SLOT(Stop()))
          ;
      State_O_M_A_Waiting
          = new sState(State_O_M_Acquiring,"State_O_M_A_Waiting");
      State_O_M_A_Reading
          = new sState(State_O_M_Acquiring,"State_O_M_A_Reading");

      State_O_M_A_Waiting
          ->SetAsInitial()
          ->AddTransition(Clock,SIGNAL(timeout())
                         ,State_O_M_A_Reading)
          ;
      State_O_M_A_Reading
          ->OnEnterCall(SLOT(RequestCount()))
          ->AddTransition<sTelex>(SIGNAL(GotSR400Message(sTelex))
                                 ,SLOT(ParseCount(sTelex))
                                 ,State_O_M_A_Waiting)
          ;
    }
    connect(this ,SIGNAL(SR400PortError(QString))
           ,this ,SLOT(ReportSR400PortError(QString)) );
  }
}

bool  sSR400Counter::AutoRepeat() const
{
  /*! Реализован только режим AutoRepeat.
   *  Режим однократного запуска при необходимости
   *  должен реализовываться отдельно.
   */
  return true;
}


void  sSR400Counter::SetAutoRepeat (bool  auto_repeat)
{
  //! Реализован только режим AutoRepeat:
  if( ! auto_repeat )
    emit ErrorMessage(tr("sNIDAQVoltmeter works in AutoRepeat mode only"));
  AutoRepeatFlag = true;
}


void  sSR400Counter::SetWantedAutoRepeatMode ()
{
  //! Реализован только режим AutoRepeat:
  if( ! WantedAutoRepeatMode )
    emit ErrorMessage(tr("sNIDAQVoltmeter works in AutoRepeat mode only"));
  AutoRepeatFlag = true;
}

void  sSR400Counter::OnEnterToCoolingState ()
{
  Disconnect(SR400Port ,0 ,0);
  SR400Port.ReleaseExclusiveControl();
  SR400Port.SetToNULL();
  sGauge::OnEnterToCoolingState();
}


void  sSR400Counter::SetWantedPeriod ()
{
  try{
    if (WantedPeriod < MinPeriod) WantedPeriod = MinPeriod;
    sPhysValue CTime("s");
    CTime = WantedPeriod - sPhysValue(2.0,"ms");
    PollingInterval = Round( (CTime/10 / sPhysValue(1.0,"ms")).ToReal() );
qDebug() << "PollingInterval = " << PollingInterval;
    if( PollingInterval < 1 )  PollingInterval=1;
    QByteArray _comm;
    _comm.append(QByteArray::number((CTime / sPhysValue(1.0e-7,"s")).ToReal(),'E',0));
    _comm.prepend("CP2,");
    SendDirective<sTelex>(SR400Port ,sTeleCom::cSendTelex ,sTelex(_comm));
    _comm.clear();
    _comm.append("CP2");
    SendDirective<sTelex>(SR400Port ,sTeleCom::cSendTelex ,sTelex(_comm));
  }catch(rxRAlgebra){
    //!@todo{Instruments} define Failure signal for drivers?
    emit ErrorMessage(tr("Cannot set period, period=%1")
                     .arg(ToQString(WantedPeriod.Text())));
    emit CoolDown();
  }
}

void  sSR400Counter::ReportSR400PortError (QString  error_text)
{
  emit ErrorMessage(tr("Got error from %1:"
                       "\n\n%2")
                      .arg(ToQString(SR400Port->DeviceIdentifier()))
                      .arg(error_text)     );
}


void  sSR400Counter::LinkToSR400Port ()
{
  try{
    SR400Port.SetTo(DeviceLinks,"SR400Port");
    if( ! SR400Port.TakeExclusiveControl() ){
      emit ErrorMessage(tr("Cannot take exclusive control over '%1'. "
                           "Try again later.")
                       .arg(ToQString( SR400Port->DeviceIdentifier() )) );
    }else{
      Connect(SR400Port ,SIGNAL(Warmed())
             ,SIGNAL(SR400PortLinked()) );
      Connect(SR400Port ,SIGNAL(ErrorMessage(QString))
             ,SIGNAL(SR400PortError(QString)) );
      Connect(SR400Port ,SIGNAL(GotTelex(sTelex))
             ,SIGNAL(GotSR400Message(sTelex)) );
    }
  }catch( rxDeviceLaunch E ){
    emit ErrorMessage( tr("Error while linking to a submissive device.\n"
                          "%1\nCheck configuration.")
                         .arg(ToQString(E.MsgText()))   );
  }
}

void  sSR400Counter::OnNoAnswerFromSR400 ()
{
  emit ErrorMessage("No answer from SR400 Gated photon counter");
}

void  sSR400Counter::SendDummyCommand ()
{
  SendDirective(SR400Port  ,sTeleCom::cSendTelex ,sTelex("CP1") );

}

void  sSR400Counter::RequestSR400ID ()
{
  SendDirective(SR400Port  ,sTeleCom::cSendTelex ,sTelex("CP1,7E7;CP1") );
}


void  sSR400Counter::ParseSR400ID (sTelex  text)
{
  if(  text.startsWith("7E7") ){
    emit GotSR400ID();
  }else{
    emit ErrorMessage(QString("Unknown telex: ")+text);
  }
}


void  sSR400Counter::CorrectPeriod (sTelex text)
{
  bool OK;
  real _preset = text.toDouble(&OK);
  if( OK ){
    CountingTime = sPhysValue(_preset*1e-7,"s");
    ThePeriod = CountingTime + sPhysValue(2.0,"ms");
    emit PeriodChanged(ThePeriod);
  }else{
    emit ErrorMessage(QString("Unknown telex: ")+text);
  }
}


void  sSR400Counter::RequestCount ()
{
  QByteArray _comm;
  _comm.append(QByteArray::number(ChNum,10));
  _comm.prepend("QA");
  SendDirective<sTelex>(SR400Port ,sTeleCom::cSendTelex ,sTelex(_comm));
}


void  sSR400Counter::ParseCount (sTelex  text)
{
  bool                        OK = false;
  int                         Count = text.toInt(&OK,10);
  if( OK ){
    if (!(Count < 0)) {
//      qDebug()<<"n="<<ChNum<<"  Count="<<Count;
      ChNum++;
      if (ChNum > 2000) {
        ChNum = 1;
        SendDirective(SR400Port  ,sTeleCom::cSendTelex ,sTelex("CR;CS") );
      }
// For Debug ---------------------------------------
      _C++;
      if (_C == 100) {
          qDebug("Time elapsed: %d ms", _time.elapsed());
          _time.start();
          _C = 0;
      }
// -------------------------------------------------
      emit NewData(sPhysValue(Count,0.5,"photon")/CountingTime);
    }
  }else{
    emit ErrorMessage(QString("Unknown telex: ")+text);
  }
}


void  sSR400Counter::Start ()
{
  SendDirective(SR400Port  ,sTeleCom::cSendTelex ,sTelex("CR") );
  SendDirective(SR400Port  ,sTeleCom::cSendTelex ,sTelex("CM0") );
  SendDirective(SR400Port  ,sTeleCom::cSendTelex ,sTelex("CI0,1") );
  SendDirective(SR400Port  ,sTeleCom::cSendTelex ,sTelex("NP2000") );
  SendDirective(SR400Port  ,sTeleCom::cSendTelex ,sTelex("NE0") );
  SendDirective(SR400Port  ,sTeleCom::cSendTelex ,sTelex("DT2E-3") );
  Clock->setSingleShot(false);
  Clock->start(PollingInterval);
  ChNum = 1;
  // For Debug ------------------------
  _C = 0;
  _time.start();
  // ----------------------------------
  SendDirective(SR400Port  ,sTeleCom::cSendTelex ,sTelex("CS") );
}

void  sSR400Counter::Stop ()
{
  SendDirective(SR400Port  ,sTeleCom::cSendTelex ,sTelex("CH") );
  SendDirective(SR400Port  ,sTeleCom::cSendTelex ,sTelex("CR") );
  Clock->stop();
}


//------------------------------------------------------------------------------
} //namespace RA
