////////////////////////////////////////////////////////////////////////////////
/*! @file UT60EMeter.cpp   Получение данных от мультиметра UT60E.
- Part of UTMeters - a Research Assistant Plugin package.
- Uses  RAGUI - the Research Assistant Graphical User Interface.
- Uses  QtGui v.5.0  -  http://qt.nokia.com/
- Uses  RANet - Research Assistant Net Library (based on ANSI C++).
- Copyright(C) 2016-2017, Dmitry K. Nelson, St.Petersburg, Russia.
                          D.Nelson@mail.ioffe.ru
*///////////////////////////////////////////////////////////////////////////////
#include "UT60EMeter.h"
#include "attr.h"
#include "StringsAndQStrings.h"
#include "Symbols.h"
#include <QDebug>
namespace RA {
//------------------------------------------------------------------------------

DEFINE_CLASS_TAG(sUT60EMeter)

//---------------------------------------------------------- sUT60EMeter ---

void  sUT60EMeter::Preconfigure (sNodePtr  DeviceCfg)
{
  sPtr<sFolderNode>              DeviceSettings;
  DeviceCfg |= "Settings";
  DeviceSettings = DeviceCfg >> "Settings";
  if( DeviceSettings.IsCorrect() ){
    DeviceSettings |= attr("PortName","COM1");
    DeviceSettings |= attr("BaudRate",2400);
    DeviceSettings |= attr("Parity","N");
    DeviceSettings |= attr("DataBits",8);
    DeviceSettings |= attr("StopBits",1);
  };
  sGauge::Preconfigure(DeviceCfg);
}

sUT60EMeter::~sUT60EMeter ()
{
  if( Port ){
    disconnect(Port, 0, this, 0);
    Port->close();
    delete  Port;  Port = NULL;
  };
}

sUT60EMeter::sUT60EMeter (sString   DevicetId
                         ,sNodePtr  DeviceCfg)
    :sGauge(DevicetId,DeviceCfg)
    ,Port(NULL)
    ,PortName("COM1")
{
}

void  sUT60EMeter::PostConstructor ()
{
  sGauge::PostConstructor();

  BaudRate = 2400;
  Parity = QSerialPort::NoParity;
  DataBits = QSerialPort::Data8;
  StopBits = QSerialPort::OneStop;
  try{  //get config parameters
    integer                     I;
    sString                     S;
    if( get_from<sIntegerNode>(DeviceSettings>>"BaudRate",BaudRate) ){
    }
    if( get_from<sTextNode>(DeviceSettings>>"Parity",S) ){
      if( S.UpperCase() == "N" ){
        Parity = QSerialPort::NoParity;
      }else if( S.UpperCase() == "E" ){
        Parity = QSerialPort::EvenParity;
      }else if( S.UpperCase() == "O" ){
        Parity = QSerialPort::OddParity;
      }
    }
    if( get_from<sIntegerNode>(DeviceSettings>>"DataBits",I) ){
      switch( I ){
        case 5:  DataBits = QSerialPort::Data5;  break;
        case 6:  DataBits = QSerialPort::Data6;  break;
        case 7:  DataBits = QSerialPort::Data7;  break;
        case 8:  DataBits = QSerialPort::Data8;  break;
        default: ;
      }
    }
    if( get_from<sIntegerNode>(DeviceSettings>>"StopBits",I) ){
      switch( I ){
        case 1:  StopBits = QSerialPort::OneStop;  break;
        case 3:  StopBits = QSerialPort::OneAndHalfStop;  break;
        case 2:  StopBits = QSerialPort::TwoStop;  break;
        default: ;
      }
    }
    get_from<sTextNode>(DeviceSettings >> "PortName" ,PortName);
  }catch(...){
    SetInvalid(tr("Something wrong in configuration."));
  }
  Port= new QSerialPort(ToQString(PortName),this);

  //evolution of the inherited state machine (behaviour)
  {
    //evolution of the State_Warming
    {
      State_W_PortOpening = new sState(State_Warming(),"State_W_PortOpening");
      State_W_PortOpening
        ->SetAsInitial()
        ->OnEnterCall(SLOT(PortOpen()))
        ->AddTransition(SIGNAL(PortOK()) ,State_Warming()->Fin())
        ->AddTransition(SIGNAL(ErrorMessage()) ,State_Cooling() )
      ;
    }
    //evolution of the State_Operating
    {
      State_O_M_Acquiring
        ->OnEnterCall(SLOT(Start()))
        ->OnExitCall(SLOT(Stop()))
      ;
    }
  }
}

void  sUT60EMeter::PortOpen ()
{
  qDebug()<< "Opening COM port " << Port->portName();

  if( ! Port->open(QIODevice::ReadWrite /*| QIODevice::Unbuffered*/) ){
    emit ErrorMessage( tr("Cannot open '%1'.").arg(Port->portName()) );
    return;
  };
  if( ! Port->setBaudRate(BaudRate) ){
    emit ErrorMessage( tr("Port '%1': set baud rate %2 error.")
                         .arg(Port->portName()).arg(BaudRate) );
    return;
  };
  if( ! Port->setParity(Parity) ){
    emit ErrorMessage( tr("Port '%1': set parity %2 error.")
                         .arg(Port->portName()).arg(Parity) );
    return;
  };
  if( ! Port->setDataBits(DataBits) ){
    emit ErrorMessage( tr("Port '%1': set data bits %2 error.")
                         .arg(Port->portName()).arg(DataBits) );
    return;
  };
  if( ! Port->setStopBits(StopBits) ){
    emit ErrorMessage( tr("Port '%1': set stop bits %2 error.")
                         .arg(Port->portName()).arg(StopBits) );
    return;
  };

  qDebug() << Port->portName() << "is open:";
  qDebug() << "Mode     " << Port->openMode();
  qDebug() << "Baud rate" << Port->baudRate();
  qDebug() << "Data bits" << Port->dataBits();
  qDebug() << "Parity   " << Port->parity();
  qDebug() << "Stop bits" << Port->stopBits();
  qDebug() << "Flow ctrl" << Port->flowControl();

  Port->flush();
  Port->reset();

  emit PortOK();
}

void  sUT60EMeter::OnEnterToCoolingState ()
{
  if( Port ){
    disconnect(Port, 0, this, 0);
    delete  Port;  Port = NULL;
  };
  sGauge::OnEnterToCoolingState();
}

void sUT60EMeter::ReadAvailableBytes ()
{
  while( qint64  AvailableBytes = Port->bytesAvailable() ){
    QByteArray                  BA = Port->read(AvailableBytes);
    InputBuffer.append(BA);
    while( InputBuffer.length() > 0 ){
      int ByteNo=(InputBuffer[0]&0xF0)>>4;
      Sample[ByteNo-1]=InputBuffer[0]&0xF;
      if (ByteNo==1) Has_first=true;
      if (ByteNo==14 && Has_first)  ParseSample();
      InputBuffer.remove(0, 1);
    };
  };
}


void  sUT60EMeter::Start ()
{
  Has_first=false;
  connect(Port, SIGNAL(readyRead()), this, SLOT(ReadAvailableBytes()));
}

void  sUT60EMeter::Stop ()
{
  disconnect(Port, SIGNAL(readyRead()), this, SLOT(ReadAvailableBytes()));
}

void  sUT60EMeter::ParseSample ()
{
  sString _unit = "";
  QByteArray Result = "";
  real Value = 0;
  unsigned char c;

  for (int n=1;n<9;n+=2) {
    c = (Sample[n] << 4) | (Sample[n+1] & 0xf);
    if (c & 0x80) {
      if (n==1) { Result.append('-'); }
      else { Result.append('.'); }
      c &= ~0x80;
    }
    switch (c) {
    case 0x00: c = '0'; break;
    case 0x7d: c = '0'; break;
    case 0x05: c = '1'; break;
    case 0x5b: c = '2'; break;
    case 0x1f: c = '3'; break;
    case 0x27: c = '4'; break;
    case 0x3e: c = '5'; break;
    case 0x7e: c = '6'; break;
    case 0x15: c = '7'; break;
    case 0x7f: c = '8'; break;
    case 0x3f: c = '9'; break;
    case 0x68: c = 'L'; break;
    default: c = '?';
    }
    Result.append(c);
  }
  if (Sample[9] & 0x4)
		_unit += "n";
	else if (Sample[9] & 0x8)
		_unit += Symbol::Micro;
	else if (Sample[10] & 0x8)
		_unit += "m";
	else if (Sample[9] & 0x2)
		_unit += "k";
	else if (Sample[10] & 0x2)
		_unit += "M";

	if (Sample[11] & 0x8)
		_unit += "F";
	else if (Sample[11] & 0x4)
		_unit += Symbol::Ohm;
	else if (Sample[12] & 0x8)
		_unit += "A";
	else if (Sample[12] & 0x4)
		_unit += "V";
	else if (Sample[13] & 0x1)
		_unit = _unit + Symbol::CelsiusDegree;

  bool OK;
  Value = Result.toDouble(&OK);
  if (OK)
    emit NewData(sPhysValue(Value,_unit));
}


//------------------------------------------------------------------------------
} //namespace RA
