////////////////////////////////////////////////////////////////////////////////
/*! @file AdminJob.cpp Background job execution for RAdmin.
- Part of RAdmin project.
- Uses  RANet - Research Assistant Net Library.
- Uses  Qt v.5  - http://qt.io/
- Copyright(C) 2020, Viktor E. Bursian, St.Petersburg, Russia.
                     VBursian AT gmail DOT com
*///////////////////////////////////////////////////////////////////////////////
#include "AdminJob.h"
#include "StringsAndQStrings.h"
#include "DB_SQL_Net.h"
#include "Nodes.h"
#include "Attribute.h"
#include "attr.h"
#include <sstream>
#include <string>
//#include "Log.h"
//#include <QDebug>
namespace RA {
//------------------------------------------------------------------------------
//---------------------------------------------------------------- sAdminJob ---

void  sAdminJob::FindAndExecute (const QString &  job_name
                                ,QStringList &    params)
{
  if(      job_name == "DUMMY"    )  DummyJob(params);
  else if( job_name == "SQLDL"    )  AvailableSQLDrivers(params);
  else if( job_name == "CN"       )  CreateNewPartition(params);
  else if( job_name == "RP"       )  ReplicatePartition(params);
  else if( job_name == "AARP"     )  AbsorbAllRelatedPartitions(params);
  else if( job_name == "CRPT"     )  CureRelatedPartitionsTable(params);
  else                            sScriptJob::FindAndExecute(job_name,params);
}


void  sAdminJob::DummyJob (QStringList &  params)
{
  ReportNewJobStart("Dummy Job");
  emit TotalStages(params.size());
  for( int St = 1 ; St <= params.size() ; ++St ){
    ReportNewStageStart(QString("Stage %1").arg(St));
    bool  OK = false;
    int   N = params.at(St-1).toInt(&OK);
    if( ! OK ){
      Failure(QString("Wrong parameter: "+params.at(St-1)));
      return;
    }
    SetProgressTarget(N);
    for( int P = 0 ; P < N ; ++P ){
      SetProgress(P);
      msleep(10);
      if( P % 1000 == 0 )
        emit OutPar("little red fox jumps: ");
      if( P % 100 == 99 )
        emit Out(" . ");
      if( P % 1000 == 999 )
        emit Out(QString().setNum(P+1));
      if( AbortRequested() )
        break;
    }
    if( AbortRequested() ){
      break;
    }
  }
  if( ! AbortRequested() ){
    ReportJobConclusion("DummyJob completed.");
  }
}


#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wunused-parameter"
void  sAdminJob::AvailableSQLDrivers (QStringList &  params)
{
  ReportNewJobStart("Available SQL drivers list");
  foreach( QString DN , QSqlDatabase::drivers() ){
    QSqlDatabase              D =QSqlDatabase::addDatabase(DN,"test");
    bool                      A = D.isDriverAvailable(DN);
    bool                      V = D.isValid();
    emit OutPar(QString("<big><b>%1</b></big>").arg(DN));
    if( ! A ){
      emit OutPar("&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;<font color=red>"
                  "Qt looks awful shaky with this driver</font>");
    }else if( ! V ){
      emit OutPar("&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;"
                  "Driver is theoretically available, but "
                  "<font color=red>cannot be loaded</font>");
    }else{
      emit OutPar(QString("&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;%1: %2")
                 .arg("Transactions")
                 .arg(D.driver()->hasFeature(QSqlDriver::Transactions)
                      ? "yes" : "<font color=red>no</font>" )
                 );
      emit OutPar(QString("&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;%1: %2")
                 .arg("PreparedQueries")
                 .arg(D.driver()->hasFeature(QSqlDriver::PreparedQueries)
                      ? "yes" : "<font color=red>no</font>" )
                 );
      emit OutPar(QString("&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;%1: %2")
                 .arg("PositionalPlaceholders")
                 .arg(D.driver()->hasFeature(QSqlDriver::PositionalPlaceholders)
                      ? "yes" : "<font color=red>no</font>" )
                 );
      emit OutPar(QString("&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;%1: %2")
                 .arg("NamedPlaceholders")
                 .arg(D.driver()->hasFeature(QSqlDriver::NamedPlaceholders)
                      ? "yes" : "no" )
                 );
      emit OutPar(QString("&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;%1: %2")
                 .arg("FinishQuery")
                 .arg(D.driver()->hasFeature(QSqlDriver::FinishQuery)
                      ? "yes" : "<font color=red>no</font>" )
                 );
      emit OutPar(QString("&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;%1: %2")
                 .arg("BLOB")
                 .arg(D.driver()->hasFeature(QSqlDriver::BLOB)
                      ? "yes" : "<font color=red>no</font>" )
                 );
      emit OutPar(QString("&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;%1: %2")
                 .arg("Unicode")
                 .arg(D.driver()->hasFeature(QSqlDriver::Unicode)
                      ? "yes" : "<font color=red>no</font>" )
                 );
      emit OutPar(QString("&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;%1: %2")
                 .arg("LastInsertId")
                 .arg(D.driver()->hasFeature(QSqlDriver::LastInsertId)
                      ? "yes" : "no" )
                 );
      emit OutPar(QString("&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;%1: %2")
                 .arg("QuerySize")
                 .arg(D.driver()->hasFeature(QSqlDriver::QuerySize)
                      ? "yes" : "no" )
                 );
      emit OutPar(QString("&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;%1: %2")
                 .arg("SimpleLocking")
                 .arg(D.driver()->hasFeature(QSqlDriver::SimpleLocking)
                      ? "yes" : "no" )
                 );
    }
    D.close();
    QSqlDatabase::removeDatabase(D.connectionName());
  }
}
#pragma GCC diagnostic pop


void  sAdminJob::CreateNewPartition (QStringList &  params)
{
  ReportNewJobStart("Create new partition");
  sDB_Info                    PartitionInfo;
  try{
    if( params.size() != 1 ){
      Failure("Error: parameter required.");
      return;
    }
    PartitionInfo.SetByURL( FromQString(params.at(0)) );
    sDB_SQL_Net::CreateNewPartition(PartitionInfo,true);
  }catch( std::exception & E ){
    Failure(QString("Error: ")+QString(E.what()));
    return;
  }catch( ... ){
    Failure("Error: ???");
    return;
  }
  if( ! AbortRequested() ){
    ReportJobConclusion(QString("'%1' created and initiated.")
                        .arg(ToQString(PartitionInfo.URL()))
                       );
  }
}


void  sAdminJob::ReplicatePartition (QStringList &  params)
{
  ReportNewJobStart("Replicate partition");
  sDB_Info                    SourceInfo;
  sDB_Info                    DestinationInfo;
  try{
    if( params.size() != 2 ){
      Failure("Error: 2 parameters required.");
      return;
    }
    SourceInfo.SetByURL(FromQString(params.at(0)));
    DestinationInfo.SetByURL(FromQString(params.at(1)));
    emit TotalStages(6);
    //______ Create new partition ______
    {
      ReportNewStageStart("Create new partition");
      if( ! sDB_SQL_Net::CreateNewPartition(DestinationInfo,false) ){
        Failure("Unable to create the destination partition.");
        return;
      }
    }
    if( AbortRequested() ){
      return;
    }
    sDatabase                   Source(SourceInfo);
    sDatabase                   Destination(DestinationInfo);
    bool                        HasParamsTable = false;
    sSQLQuery                   SQL(Source.Connection("replication"));
    sSQLQuery                   InQ(Destination.Connection("replication"));
    //______  Replicate params ______
    {
      ReportNewStageStart("Replicate params");
      try{
        SQL.exec(sString("SELECT parkey, parval FROM ")
                +Source.TableFullName("params"));
        HasParamsTable = true;
      }catch( rxSQL ){
        SQL.exec(sString("SELECT id FROM ")
                +Source.TableFullName("databases")
                +" WHERE nickname = '"+sDatabase::MyselfNickname+"'");
      }
      if( HasParamsTable ){
        while( SQL.next() ){
          if(       SQL.StringValue(0) == "part_id" ){
            Source.Id = sDB_UUID( SQL.StringValue(1) );
          }else if( SQL.StringValue(0) == "version" ){
            Source.Version = sVersion( SQL.StringValue(1) );
          //!@todo{NetIO} get oldest_node_version, newest_node_version
          //}else if( SQL.StringValue(0) == "oldest_node_version" ){
          //}else if( SQL.StringValue(0) == "newest_node_version" ){
          }
          InQ.prepare(sString("INSERT INTO ")
                     +Destination.TableFullName("params")
                     +" VALUES(?,?)");
          InQ.addBindString( SQL.StringValue(0) );
          InQ.addBindString( SQL.StringValue(1) );
          InQ.exec();
        }
      }else{
        if( ! SQL.first() ){
          Failure(QString("ReplicatePartition: cannot find '")
                 +ToQString(sDatabase::MyselfNickname)
                 +"' in databases of the source partition.");
          return;
        }
        Source.Id = sDB_UUID( SQL.StringValue(0) );
        if( SQL.next() ){
          Failure(QString("ReplicatePartition: ambiguos '")
                 +ToQString(sDatabase::MyselfNickname)
                 +"' in 'databases' table of the source partition.");
          return;
        }
        InQ.SetParamValue( Destination.TableFullName("params")
                         , "part_id" , Source.Id.ToString() );
        InQ.SetParamValue( Destination.TableFullName("params")
                         , "version" , RANet::Version.Text() );
      }
    }
    if( AbortRequested() ){
      return;
    }
    //______  Replicate parts table ______
    {
      ReportNewStageStart("Replicate parts table");
      if( HasParamsTable ){
        SQL.exec(sString("SELECT part_id, part_url, pcode FROM ")
                +Source.TableFullName("parts"));
        while( SQL.next() ){
          InQ.prepare(sString("INSERT INTO ")
                     +Destination.TableFullName("parts")
                     +" VALUES(?,?,?)");
          InQ.addBindString( SQL.StringValue(0) );
          InQ.addBindString( SQL.StringValue(1) );
          InQ.addBindString( SQL.StringValue(2) );
          InQ.exec();
        }
      }else{
        sDB_Link                    Link;
        SQL.exec(sString("SELECT id ,type ,name "
                         ",host ,port ,login ,password FROM ")
                +Source.TableFullName("databases")
                +" WHERE nickname != '"+sDatabase::MyselfNickname+"'");
        while( SQL.next() ){
          Link.Id        = sDB_UUID( SQL.StringValue(0) );
          Link.Type      = sDB_Info::eType( SQL.value(1).toUInt() );
          Link.DBName    = SQL.StringValue(2);
          if( Link.DBName.Pos("#") < Link.DBName.Length() ){
            Link.SchemaName = Link.DBName.Tail(Link.DBName.Pos("#")+1);
            Link.DBName = Link.DBName.Substr(0,Link.DBName.Pos("#"));
          }else if( Link.DBName.Pos("?") < Link.DBName.Length() ){
            Link.SchemaName = Link.DBName.Tail(Link.DBName.Pos("?")+1);
            Link.DBName = Link.DBName.Substr(0,Link.DBName.Pos("?"));
          }
          Link.Host      = SQL.StringValue(3);
          Link.Port      = SQL.value(4).toInt();
          if( Link.Port == 0 )  Link.Port = -1;
          Link.Login     = SQL.StringValue(5);
          Link.Password  = SQL.StringValue(6);
          InQ.prepare(sString("INSERT INTO ")+Destination.TableFullName("parts")
                     +" (part_id, part_url, pcode) VALUES(?,?,?)");
          InQ.addBindString( Link.Id.ToString() );
          InQ.addBindString( Link.URL() );
          InQ.addBindValue( QByteArray((literal)(Link.Password)).toBase64()
                          , QSql::In | QSql::Binary );
          InQ.exec();
        }
      }
    }
    if( AbortRequested() ){
      return;
    }
    //______  Replicate nodes and attributes ______
    {
      ReportNewStageStart("Replicate nodes");
      if( ! ReplicatePartNodes(Source,Destination) )
        return;
      if( AbortRequested() )
        return;
      ReportNewStageStart("Replicate attributes");
      if( ! ReplicatePartAttrs(Source,Destination) )
        return;
      if( AbortRequested() )
        return;
    }
    //______  Finishing ______
    {
      ReportNewStageStart("Editing params");
      InQ.SetParamValue( Destination.TableFullName("params")
                       , "replicated at" , sTime().ToISO_8601() );
      InQ.SetParamValue( Destination.TableFullName("params")
                       , "replicated from" , Source.URL() );
    }
  }catch( rxSQL E ){
    Failure(QString("SQL error in sAdminJob::ReplicatePartition: ")
                +ToQString(E.Error));
    return;
  }catch( rxException E ){
    Failure(QString("Error in sAdminJob::ReplicatePartition: ")
           +ToQString(E.MsgText()));
    return;
  }catch( std::exception & E ){
    Failure(QString("Error in sAdminJob::ReplicatePartition: ")
           +QString(E.what()));
    return;
  }catch( ... ){
    Failure("Error in sAdminJob::ReplicatePartition: ???");
    return;
  }
  if( ! AbortRequested() ){
    ReportJobConclusion("Replication completed.");
  }
}


void  sAdminJob::AbsorbAllRelatedPartitions (QStringList &  params)
{
  ReportNewJobStart("Absorbing all related partitions");
  sDB_Link                    RootPartition;
  std::list<sDB_Link>         Partitions;
  try{
    if( params.size() != 1 ){
      Failure("Error: 1 parameter required.");
      return;
    }
//    SourceInfo.SetByURL(FromQString(params.at(0)));
    sDB_SQL_Net                 Net;
    //______ Checking versions ______
    {
      ReportNewStageStart("Checking versions");
      bool                        OK = false;
      QString                     OutLine;
      if( ! Net.Open(FromQString(params.at(0))) ){
        Failure("Oops! It's very strange: cannot reopen RANet.");
        return;
      }
      for( std::map<sDB_UUID,psDatabase>
                            ::const_iterator D = Net.Databases.begin()
                                           ; D != Net.Databases.end()
                                           ; ++D ){
        psDatabase                P = dynamic_cast<psDatabase>(D->second);
        OK = ! ( P->Version.Principal() < sVersion(1,5).Principal() );
        OutLine = ToQString( P->Id.ToString() );
        if( P->Id == Net.RootDB_UUID ){
          RootPartition = *P;
          OutLine += "(root)";
        }else{
          Partitions.push_back(*P);
        }
        OutLine += "  v.";
        OutLine += ToQString( P->Version.Principal().Text() ) + " - ";
        emit OutPar( OutLine + ( OK ? "OK" : "WRONG!!!" ) );
      }
      if( ! RootPartition.Version.IsValid() ){
        Failure("Oops! It's weird: root partition version is invalid.");
        return;
      }
      if( ! OK ){
        Failure("All partition versions must be 1.5 or above. Some don't.");
        return;
      }
      Net.Close();
    }
    if( AbortRequested() ){
      return;
    }
    emit TotalStages( 1 + 2 * Partitions.size() + 3 );

    //______ Copying ______
    {
      for( std::list<sDB_Link>::const_iterator P = Partitions.begin()
                                             ; P != Partitions.end()
                                             ; ++P ){
        ReportNewStageStart( QString("Copying nodes from ")
                           + ToQString(P->Id.ToString()) );
        if( ! ReplicatePartNodes( *P , RootPartition ) )
          return;
        if( AbortRequested() )
          return;
        ReportNewStageStart( QString("Copying attributes from ")
                           + ToQString(P->Id.ToString()) );
        if( ! ReplicatePartAttrs( *P , RootPartition ) )
          return;
        if( AbortRequested() )
          return;
      }
    }
    if( AbortRequested() ){
      return;
    }
    //______ Changing  attribute links ______
    {
      ReportNewStageStart("Changing  attribute links");
      SetProgressTarget(Partitions.size());
      int  N = 0;
      SetProgress(N);
      for( std::list<sDB_Link>::const_iterator P = Partitions.begin()
                                             ; P != Partitions.end()
                                             ; ++P ){
        OutPar( QString("... to ") + ToQString(P->Id.ToString())
                  + " ..." );
        sDB_SQL_Net::RunSQL( RootPartition
            ,sString("UPDATE ") + RootPartition.TableFullName("attrs")
            +" SET value_db = NULL WHERE value_db = '"
            +P->Id.ToString()
            +"'"
            );
        Out("    DONE;");
        ++N;
        SetProgress(N);
      }
      SetProgress(0);
    }
    if( AbortRequested() ){
      return;
    }
    //______ Cleaning parts table ______
    {
      ReportNewStageStart("Cleaning parts table");
      sDB_SQL_Net::RunSQL(RootPartition
          ,sString("DELETE FROM ")+RootPartition.TableFullName("parts")
          );
    }
    if( AbortRequested() ){
      return;
    }
    //______ Filling the 'Absorbed partitions' folder ______
    {
      ReportNewStageStart("Filling the 'Absorbed partitions' folder");
      ASSERT( Net.Open(RootPartition.URL(true)) );
      sPtr<sFolderNode>           RelatedParts;
      sPtr<sFolderNode>           AbsorbedParts;
      RelatedParts = (Net.Root() >> "Related partitions");
      AbsorbedParts = (Net.Root() >= "Absorbed partitions");
      for( std::list<sDB_Link>::const_iterator P = Partitions.begin()
                                             ; P != Partitions.end()
                                             ; ++P ){
        AbsorbedParts << attr(P->Id.ToString()
                             ,RelatedParts >> P->Id.ToString());
        RelatedParts.DelAttr(P->Id.ToString());
      }
      Net.Close();
    }
  }catch( rxSQL E ){
    Failure(QString("SQL error in sAdminJob::AbsorbAllRelatedPartitions: ")
                +ToQString(E.Error));
    return;
  }catch( rxException E ){
    Failure(QString("Error in sAdminJob::AbsorbAllRelatedPartitions: ")
           +ToQString(E.MsgText()));
    return;
  }catch( std::exception & E ){
    Failure(QString("Error in sAdminJob::AbsorbAllRelatedPartitions: ")
           +QString(E.what()));
    return;
  }catch( ... ){
    Failure("Error in sAdminJob::AbsorbAllRelatedPartitions: ???");
    return;
  }
  if( ! AbortRequested() ){
    ReportJobConclusion("All related partitions have been absorbed.");
  }
}


void sAdminJob::CureRelatedPartitionsTable (QStringList &  params)
{
  ReportNewJobStart("Cure related partitions table");
  sDB_Info                    PartitionInfo;
  try{
    if( params.size() != 1 ){
      Failure("Error: parameter required.");
      return;
    }
    PartitionInfo.SetByURL( FromQString(params.at(0)) );



    Failure("Not yet implemented");
    return;





  }catch( rxSQL E ){
    Failure(QString("SQL error in sAdminJob::CureRelatedPartitionsTable: %1")
            .arg(ToQString(E.Error)));
    return;
  }catch( rxException E ){
    Failure(QString("Error in sAdminJob::CureRelatedPartitionsTable: ")
           +ToQString(E.MsgText()));
    return;
  }catch( std::exception & E ){
    Failure(QString("Error in sAdminJob::CureRelatedPartitionsTable: ")
           +QString(E.what()));
    return;
  }catch( ... ){
    Failure("Error in sAdminJob::CureRelatedPartitionsTable: ???");
    return;
  }
  if( ! AbortRequested() ){
    ReportJobConclusion("The related partitions table has been brought "
                        "in accordance with de-facto existing links.");
  }
}

//...................................................... utilized in jobs ......

bool  sAdminJob::ReplicatePartNodes (rcsDB_Link  source
                                    ,rcsDB_Link  destination)
{
  try{
    sDatabase                   Source(source);
    sDatabase                   Destination(destination);
    sSQLQuery                   SQ(Source.Connection("replication_nodes"));
    sSQLQuery                   DQ(Destination
                                    .Connection("replication_nodes"));
    bool                        SourceVersionIsLessThan_1_5 = false;
    int                         RecordsInDestinatonBefore = 0;
    int                         RecordsInDestinatonAfter;
    int                         RecordsInSource;
    int                         RecordsCopied = 0;

    SQ.exec(sString("SELECT COUNT(*) FROM ")+Source.TableFullName("nodes"));
    if( ! SQ.first() ){
      Failure("Source node counting failed.");
      return false;
    }
    RecordsInSource = SQ.value(0).toInt();
    emit OutPar(QString("Replicating %1 nodes...").arg(RecordsInSource));
    SetProgressTarget(RecordsInSource);
    SetProgress(0);

    DQ.exec(sString("SELECT COUNT(*) FROM ")
             +Destination.TableFullName("nodes"));
    if( DQ.first() )
      RecordsInDestinatonBefore = DQ.value(0).toInt();

    sString                     NodeId;
    sString                     NodeClass;
    QByteArray                  OldNodeBody;
    sString                     TimeText;
    if( Source.Version < sVersion(1,5) ){
      SourceVersionIsLessThan_1_5 = true;
      SQ.exec(sString("SELECT id,class,body,time")
              +" FROM "+Source.TableFullName("nodes"));
    }else{
      SQ.exec(sString("SELECT node_id,node_class,node_body,node_save_time")
              +" FROM "+Source.TableFullName("nodes"));
    }
    while( SQ.next() ){
      NodeId = SQ.StringValue(0);
      if( SourceVersionIsLessThan_1_5 ){
        if( NodeId == sString(QUuid().toByteArray().data()) ){
          NodeId = Source.Id.ToString();
        }
      }
      NodeClass = SQ.StringValue(1);
      OldNodeBody = SQ.value(2).toByteArray();
      psNode                      Node;
      std::stringstream           Stream(std::string(OldNodeBody.data()
                                                    ,OldNodeBody.size()+1   )
                                        ,std::ios_base::in);
      Node = dynamic_cast<psNode>
                         ( sStorable::PolymorphicLoad(NodeClass,Stream) );
      Node->Store(Stream);
      QByteArray                  NewNodeBody (Stream.str().data()
                                              ,Stream.str().size());
      delete Node;
      TimeText = SQ.StringValue(3);
      DQ.prepare(sString("INSERT INTO ")+Destination.TableFullName("nodes")
               +" VALUES(?,?,?,?)" );
      DQ.addBindString( NodeId );
      DQ.addBindString( NodeClass );
      DQ.addBindValue( NewNodeBody , QSql::In | QSql::Binary );
      DQ.addBindString( TimeText );
      DQ.exec();
      RecordsCopied++;
      SetProgress(RecordsCopied);
      if( AbortRequested() )
        break;
    }
    if( AbortRequested() ){
      return false;
    }

    DQ.exec(sString("SELECT COUNT(*) FROM ")
             +Destination.TableFullName("nodes"));
    if( ! DQ.first() ){
      Failure("Destination node counting failed.");
      return false;
    }
    RecordsInDestinatonAfter = DQ.value(0).toInt();
    if( RecordsInDestinatonAfter-RecordsInDestinatonBefore != RecordsInSource ){
      Failure(QString("Number of de-facto replicated (%1) doesn't match.")
              .arg(RecordsInDestinatonAfter-RecordsInDestinatonBefore));
      return false;
    }else{
      emit OutPar(QString("All %1 nodes have been replicated.")
                      .arg(RecordsCopied));
    }
    DQ.SetParamValue( Destination.TableFullName("params")
                     , "oldest_node_version"
                     , RANet::Version.Principal().Text() );
    DQ.SetParamValue( Destination.TableFullName("params")
                     , "newest_node_version"
                     , RANet::Version.Principal().Text() );
  }catch( rxSQL E ){
    Failure(QString("SQL error in sAdminJob::ReplicatePartNodes: ")
                +ToQString(E.Error));
    return false;
  }catch( rxException E ){
    Failure(QString("Error in sAdminJob::ReplicatePartNodes: ")
           +ToQString(E.MsgText()));
    return false;
  }catch( std::exception & E ){
    Failure(QString("Error in sAdminJob::ReplicatePartNodes: ")
           +QString(E.what()));
    return false;
  }catch( ... ){
    Failure("Error in sAdminJob::ReplicatePartNodes: ???");
    return false;
  }
  return true;
}


bool  sAdminJob::ReplicatePartAttrs (rcsDB_Link  source
                                    ,rcsDB_Link  destination)
{
  try{
    sDatabase                   Source(source);
    sDatabase                   Destination(destination);
    sSQLQuery                   SQ(Source.Connection("replication_attrs"));
    sSQLQuery                   DQ(Destination.Connection("replication_attrs"));
    bool                        SourceVersionIsLessThan_1_5 = false;
    int                         RecordsInDestinatonBefore = 0;
    int                         RecordsInDestinatonAfter;
    int                         RecordsInSource;
    int                         RecordsCopied = 0;
    sString                     HostNodeId;
    sString                     PrevHostNodeId;
    int32_t                     OrderingIndex =0;//to avoid compiler warning
    quint8                      Type;
    sString                     ValueNodeId;
    bool                        ValueDBIdIsNull;
    sString                     ValueDBId;
    bool                        NumericIsNull;
    int                         Numeric =0;//to avoid compiler warning
    bool                        SymbolicIsNull;
    sString                     Symbolic;
    sString                     TimeText;
    sString                     Fields;
    sString                     Placeholders;

    if( Source.Version < sVersion(1,5) ){
      SQ.exec(sString("SELECT COUNT(*) FROM ")
              +Source.TableFullName("attributes"));
    }else{
      SQ.exec(sString("SELECT COUNT(*) FROM ")
              +Source.TableFullName("attrs"));
    }
    if( ! SQ.first() ){
      Failure("Source attribute counting failed.");
      return false;
    }
    RecordsInSource = SQ.value(0).toInt();
    emit OutPar(QString("Replicating %1 attributes...")
                    .arg(RecordsInSource));
    SetProgressTarget(RecordsInSource);
    SetProgress(0);

    DQ.exec(sString("SELECT COUNT(*) FROM ")
             +Destination.TableFullName("attrs"));
    if( DQ.first() )
      RecordsInDestinatonBefore = DQ.value(0).toInt();

    if( Source.Version < sVersion(1,5) ){
      SourceVersionIsLessThan_1_5 = true;
      SQ.exec(sString("SELECT host,type,value,value_db,")
              +Source.ColumnNotation("numeric")
              +",symbolic,time"
              +" FROM "+Source.TableFullName("attributes"));
    }else{
      SQ.exec(sString("SELECT host_node,attr_type,value_node,value_db,")
              +"numeric_name,symbolic_name,birth_time"
              +" FROM "+Source.TableFullName("attrs")
              +" ORDER BY host_node, ordering_index");
    }
    while( SQ.next() ){
      Fields = "birth_time,host_node,ordering_index,attr_type,value_node";
      Placeholders = "?,?,?,?,?";
      HostNodeId = SQ.StringValue(0);
      if( HostNodeId != PrevHostNodeId ){
        OrderingIndex = sAttribute::OrderingIndexMin;
        PrevHostNodeId = HostNodeId;
      }else{
        OrderingIndex += sAttribute::OrderingIndexInc;
      }
      Type = SQ.value(1).toUInt();
      ValueNodeId = SQ.StringValue(2);
      ValueDBIdIsNull = SQ.value(3).isNull();
      if( ! ValueDBIdIsNull ){
        ValueDBId = SQ.StringValue(3);
        Fields += ",value_db";
        Placeholders += ",?";
      }
      NumericIsNull = SQ.value(4).isNull();
      if( ! NumericIsNull ){
        Numeric = SQ.value(4).toInt();
        Fields += ",numeric_name";
        Placeholders += ",?";
      }
      SymbolicIsNull = SQ.value(5).isNull();
      if( ! SymbolicIsNull ){
        Symbolic = SQ.StringValue(5);
        Fields += ",symbolic_name";
        Placeholders += ",?";
      }
      TimeText = SQ.StringValue(6);
      if( TimeText.Empty() ){
        TimeText = sTime().ToISO_8601();
      }
      if( SourceVersionIsLessThan_1_5 ){
        if( HostNodeId == sString(QUuid().toByteArray().data()) ){
          HostNodeId = Source.Id.ToString();
        }
        if( ValueNodeId == sString(QUuid().toByteArray().data()) ){
          if( ValueDBIdIsNull ){
            ValueNodeId = Source.Id.ToString();
          }else{
            ValueNodeId = ValueDBId;
          }
        }
      }
      DQ.prepare(sString("INSERT INTO ")+Destination.TableFullName("attrs")
               +" (" + Fields + ") VALUES (" + Placeholders +")" );
      DQ.addBindString( TimeText );
      DQ.addBindString( HostNodeId );
      DQ.addBindValue ( OrderingIndex );
      DQ.addBindValue ( Type );
      DQ.addBindString( ValueNodeId );
      if( ! ValueDBIdIsNull ){
        DQ.addBindString( ValueDBId );
      }
      if( ! NumericIsNull ){
        DQ.addBindValue( Numeric );
      }
      if( ! SymbolicIsNull ){
        DQ.addBindString( Symbolic );
      }

      try{
        DQ.exec();
      //!@patch Замена символов Юникода, на которых у MySQL несварение.
      }catch( rxSQL E ){
        if( (E.Error.Pos("Incorrect string value") < E.Error.Length())
            && ! SymbolicIsNull ){
          emit OutPar(QString("Bad char in Symbolic='%1'")
                          .arg(ToQString(Symbolic)));
          Symbolic = FromQString( ToQString(Symbolic).replace(ToQString("𝝋")
                                                             ,ToQString("φ"))
                                                     .replace(ToQString("𝝑")
                                                             ,ToQString("ϑ"))
                                );
          DQ.prepare(sString("INSERT INTO ")+Destination.TableFullName("attrs")
                   +" (" + Fields + ") VALUES (" + Placeholders +")" );
          DQ.addBindString( TimeText );
          DQ.addBindString( HostNodeId );
          DQ.addBindValue ( OrderingIndex );
          DQ.addBindValue ( Type );
          DQ.addBindString( ValueNodeId );
          if( ! ValueDBIdIsNull ){
            DQ.addBindString( ValueDBId );
          }
          if( ! NumericIsNull ){
            DQ.addBindValue( Numeric );
          }
          DQ.addBindString( Symbolic );
          DQ.exec();
        }
      }

      RecordsCopied++;
      SetProgress(RecordsCopied);
      if( AbortRequested() )
        break;
    }
    if( AbortRequested() ){
      return false;
    }

    DQ.exec(sString("SELECT COUNT(*) FROM ")
             +Destination.TableFullName("attrs"));
    if( ! DQ.first() ){
      Failure("Destination attrs counting failed.");
      return false;
    }
    RecordsInDestinatonAfter = DQ.value(0).toInt();
    if( RecordsInDestinatonAfter-RecordsInDestinatonBefore != RecordsInSource ){
      Failure(QString("Number of de-facto copied (%1) doesn't match.")
              .arg(RecordsInDestinatonAfter-RecordsInDestinatonBefore));
      return false;
    }else{
      emit OutPar(QString("All %1 attributes have been copied.")
                      .arg(RecordsCopied));
    }
  }catch( rxSQL E ){
    Failure(QString("SQL error in sAdminJob::ReplicatePartAttrs: ")
                +ToQString(E.Error));
    return false;
  }catch( rxException E ){
    Failure(QString("Error in sAdminJob::ReplicatePartAttrs: ")
           +ToQString(E.MsgText()));
    return false;
  }catch( std::exception & E ){
    Failure(QString("Error in sAdminJob::ReplicatePartAttrs: ")
           +QString(E.what()));
    return false;
  }catch( ... ){
    Failure("Error in sAdminJob::ReplicatePartAttrs: ???");
    return false;
  }
  return true;
}

/*
substitute PartId with PartId in URL
cure related partition table and 'Related partitions' folder
*/

//------------------------------------------------------------------------------
} //namespace RA
