////////////////////////////////////////////////////////////////////////////////
/*! @file AdminWindow.cpp RANet administrator window.
- Part of RAdmin project.
- Uses  RAGUI - Research Assistant Graphical User Interface.
- Uses  RANet - Research Assistant Net Library.
- Uses  Qt v.5  - http://qt.io/
- Copyright(C) 2013-2020, Viktor E. Bursian, St.Petersburg, Russia.
                          VBursian AT gmail DOT com
*///////////////////////////////////////////////////////////////////////////////
#include "AdminWindow.h"
#include "StringsAndQStrings.h"
#include "Environment.h"
#include "ActionSignallingText.h"
#include "DBLinkingDialogs.h"
#include "attr.h"
#include "NetExceptions.h"
#include "Restyling.h"
#include "Helper.h"
#include "Log.h"
#include "RAMainWindow.h"
#include <QHBoxLayout>
#include <QApplication>
#include <QLabel>
#include <QMessageBox>
#include <QStatusBar>
#include <QSettings>
#include <QFile>
#include <QDebug>
namespace RA {
//------------------------------------------------------------------------------

csString                 sAdminWindow::Title ("RAdmin");
csString                 sAdminWindow::WorkstationAttrName ("Workstation");

//------------------------------------------------------------- sAdminWindow ---

sAdminWindow::~sAdminWindow()
{
  DeleteNavigator();
  RANet::Log.Put(sLog::Debug,"Quitting","sAdminWindow destructor");
}


sAdminWindow::sAdminWindow (psNet  net)
    :QMainWindow(NULL)
    ,OpenMode(cNothingIsOpen)
    ,Net(net)
    ,Evolution(NULL)
    ,Navigator(NULL)
    ,UpdatingTimer(new QTimer(this))
{
  resize(450,600);
  move(0,0);
  setWindowIcon(QIcon(":/RAGUI/icons/ra.png"));
  setWindowTitle(ToQString(Title));

  JobScriptEditor = new sEditScriptDialog(this);

  NameInfo = new QLabel(this);
  ErrorStatus = new QLabel(this);
  statusBar()->addPermanentWidget(ErrorStatus);
  statusBar()->addWidget(NameInfo);

  ShowRA();

  ActnNew = new QAction(tr("New"), this);
  ActnNew->setStatusTip(tr("Create new database"));
  connect(ActnNew ,SIGNAL(triggered()) ,this ,SLOT(OpenNew()));

  ActnOpen = new QAction(tr("Open RANet"), this);
  ActnOpen->setStatusTip(tr("Open a partion (and all related) as a RANet"));
  connect(ActnOpen ,SIGNAL(triggered()) ,this ,SLOT(OpenRANet()));

  ActnOpenPartition = new QAction(tr("Open partition"), this);
  ActnOpenPartition
      ->setStatusTip(tr("Open just one partition for maintainance"));
  connect(ActnOpenPartition ,SIGNAL(triggered()) ,this ,SLOT(OpenPartition()));

  ActnProperties = new QAction(tr("Properties"), this);
  connect(ActnProperties ,SIGNAL(triggered())
         ,this ,SLOT(ShowPartitionProperties()));

  ActnClose = new QAction(tr("Close"), this);
//  ActnClose->setStatusTip(tr("Close database"));
  connect(ActnClose ,SIGNAL(triggered()) ,this ,SLOT(Close()));

  ActnExit = new QAction(tr("E&xit"), this);
  ActnExit->setShortcuts(QKeySequence::Quit);
  ActnExit->setStatusTip(tr("Exit the application"));
  connect(ActnExit, SIGNAL(triggered()), qApp, SLOT(closeAllWindows()));

  ActnShowHelpContents = new QAction(tr("Help"), this);
  ActnShowHelpContents->setShortcuts(QKeySequence::HelpContents);
  connect(ActnShowHelpContents ,SIGNAL(triggered())
         ,this                 ,SLOT(ShowHelpContents()));

  ActnAddDatabaseLink = new QAction(tr("Add ..."), this);
  connect(ActnAddDatabaseLink ,SIGNAL(triggered()) ,this ,SLOT(AddDatabase()));

  ActnDoDummyJobs = new QAction("Do dummy jobs", this);
  connect(ActnDoDummyJobs ,SIGNAL(triggered())
         ,this ,SLOT(DoDummyJobs()));

  ActnShowSQLDrivers = new QAction("Show available SQL drivers", this);
  connect(ActnShowSQLDrivers ,SIGNAL(triggered())
         ,this ,SLOT(ShowSQLDrivers()));

  ActnRunJobScript = new QAction(tr("Run an arbitrary job script"), this);
  connect(ActnRunJobScript ,SIGNAL(triggered())
         ,this ,SLOT(RunJobScript()));

  ActnCureRelatedPartitionsTable = new QAction("Cure related partitions table"
                                              ,this);
  connect(ActnCureRelatedPartitionsTable ,SIGNAL(triggered())
         ,this ,SLOT(CureRelatedPartitionsTable()));

  ActnReplicatePartition = new QAction(tr("Replicate this partition"), this);
  connect(ActnReplicatePartition ,SIGNAL(triggered())
         ,this ,SLOT(ReplicatePartition()));

  ActnAbsorbAllRelatedPartitions
      = new QAction(tr("Absorb all related partitions"), this);
  connect(ActnAbsorbAllRelatedPartitions ,SIGNAL(triggered())
         ,this ,SLOT(AbsorbAllRelatedPartitions()));

  ActnInstallSimple = new QAction(tr("Install everything here"), this);
  connect(ActnInstallSimple ,SIGNAL(triggered()) ,this ,SLOT(InstallSimple()));

  ActnAutocomplete = new QAction(tr("Autocomplete installation"), this);
  connect(ActnAutocomplete ,SIGNAL(triggered()) ,this ,SLOT(Autocomplete()));

  ActnCreateStyles = new QAction(tr("Add editable styles"), this);
  connect(ActnCreateStyles ,SIGNAL(triggered()) ,this ,SLOT(CreateStyles()));

//  ActnUpgrade_1_1_to_1_2 = new QAction(tr("Upgrade 1.1 -> 1.2"), this);
//  connect(ActnUpgrade_1_1_to_1_2 ,SIGNAL(triggered())
//         ,this ,SLOT(Upgrade_1_1_to_1_2()));

  MenuFile = menuBar()->addMenu(tr("&File"));
  MenuFile->addAction(ActnNew);
  MenuFile->addAction(ActnOpen);
  MenuFile->addAction(ActnOpenPartition);
  MenuFile->addSeparator();
  MenuFile->addAction(ActnProperties);
  MenuFile->addSeparator();
  MenuFile->addAction(ActnClose);
  MenuFile->addSeparator();
  MenuFile->addAction(ActnExit);
  connect(MenuFile ,SIGNAL(aboutToShow()) ,this ,SLOT(UpdateMenuFile()));

  if( dynamic_cast<psDB_SQL_Net>(Net) ){
    MenuDatabases = menuBar()->addMenu(tr("Related partitions"));
    connect(MenuDatabases ,SIGNAL(aboutToShow())
           ,this ,SLOT(FormMenuDatabases()));
  }

  QMenu *                     MenuTools = menuBar()->addMenu(tr("Tools"));
  MenuTools->addAction(ActnDoDummyJobs);
  MenuTools->addAction(ActnShowSQLDrivers);
  MenuTools->addSeparator();
  MenuTools->addAction(ActnRunJobScript);
  MenuTools->addSeparator();
//  MenuTools->addAction(ActnCureRelatedPartitionsTable);
  MenuTools->addAction(ActnReplicatePartition);
  MenuTools->addAction(ActnAbsorbAllRelatedPartitions);

  QMenu *                     MenuInstall = menuBar()->addMenu(tr("Install"));
  MenuInstall->addAction(ActnInstallSimple);
  MenuInstall->addSeparator();
  MenuInstall->addAction(ActnAutocomplete);
  MenuInstall->addAction(ActnCreateStyles);
  MenuInstall->addSeparator();
//  MenuInstall->addAction(ActnUpgrade_1_1_to_1_2);

  connect(MenuInstall ,SIGNAL(aboutToShow()) ,this ,SLOT(UpdateMenuInstall()));

  menuBar()->addAction(ActnShowHelpContents);

  UpdateMenu();

  connect(this ,SIGNAL(GotJob(QString))
         ,this ,SLOT(RunJobScript(QString)) ,Qt::QueuedConnection);
  connect(this ,SIGNAL(GotJobFile(QString))
         ,this ,SLOT(RunJobScriptFromFile(QString))
                                              ,Qt::QueuedConnection);
  QString                     JobScript;
  QString                     JobScriptFile;
  RootURL = sEnvironment::StringOption("--root"
                                      ,sEnvironment::StringOption("-r") );
  PartURL = sEnvironment::StringOption("--partition"
                                      ,sEnvironment::StringOption("-p") );
  JobScript = sEnvironment::StringOption("--job"
                                        ,sEnvironment::StringOption("-j") );
  JobScriptFile = sEnvironment::StringOption("--job-file"
                                        ,sEnvironment::StringOption("-jf") );
  if( ! JobScript.isEmpty() ){
    ReopenMode = cJobScript;
    emit GotJob(JobScript);
  }else if( ! JobScriptFile.isEmpty() ){
    ReopenMode = cJobScript;
    emit GotJobFile(JobScriptFile);
  }else if( ! RootURL.Empty() ){
    OpenRANet(RootURL);
  }else if( ! PartURL.Empty() ){
    OpenPartition(PartURL);
  }
}


void  sAdminWindow::OpenRANet ()
{
  try{
    if( Net->PreOpenSetUp() ){
      if( Net->Open() ){
        CreateNavigator();
        OpenMode = cPartitionAndRelated;
        NameInfo->setText(ToQString(Net->DesktopInfoText()));
      }
    }
  }catch( rxException E ){
    QMessageBox::critical(this,"RAdmin: Error while partition opening"
        ,ToQString(E.SorryText()));
  }catch( std::exception & E ){
    QMessageBox::critical(this,"RAdmin: Error while partition opening"
        ,QString("Error: ")+QString(E.what()));
  }catch( ... ){
    QMessageBox::critical(this,"RAdmin: Error while partition opening"
        ,"Unknown error");
  }
  UpdateMenu();
}


void  sAdminWindow::OpenRANet (sString  textual_URL)
{
  try{
    if( Net->Open(textual_URL) ){
      CreateNavigator();
      OpenMode = cPartitionAndRelated;
      NameInfo->setText(ToQString(Net->DesktopInfoText()));
    }
  }catch( rxException E ){
    QMessageBox::critical(this,"RAdmin: Error while partition opening"
        ,ToQString(E.SorryText()));
  }catch( std::exception & E ){
    QMessageBox::critical(this,"RAdmin: Error while partition opening"
        ,QString("Error: ")+QString(E.what()));
  }catch( ... ){
    QMessageBox::critical(this,"RAdmin: Error while partition opening"
        ,"Unknown error");
  }
  UpdateMenu();
}


void  sAdminWindow::OpenPartition ()
{
  try{
    if( Net->PreOpenSetUp() ){
      psDB_SQL_Net                DB_SQL_Net = dynamic_cast<psDB_SQL_Net>(Net);
      ASSERT(DB_SQL_Net);
      if( DB_SQL_Net->LoadDatabases(false) ){
        OpenMode = cPartition;
        NameInfo->setText(ToQString(Net->DesktopInfoText()));
      }
    }
  }catch( rxException E ){
    QMessageBox::critical(this,"RAdmin: Error while partition opening"
        ,ToQString(E.SorryText()));
  }catch( std::exception & E ){
    QMessageBox::critical(this,"RAdmin: Error while partition opening"
        ,QString("Error: ")+QString(E.what()));
  }catch( ... ){
    QMessageBox::critical(this,"RAdmin: Error while partition opening"
        ,"Unknown error");
  }
  UpdateMenu();
}


void  sAdminWindow::OpenPartition (sString  textual_URL)
{
  try{
    psDB_SQL_Net                DB_SQL_Net = dynamic_cast<psDB_SQL_Net>(Net);
    ASSERT(DB_SQL_Net);
    if( DB_SQL_Net->Preset(textual_URL) ){
      if( DB_SQL_Net->LoadDatabases(false) ){
        OpenMode = cPartition;
        NameInfo->setText(ToQString(Net->DesktopInfoText()));
      }
    }
  }catch( rxException E ){
    QMessageBox::critical(this,"RAdmin: Error while partition opening"
        ,ToQString(E.SorryText()));
  }catch( std::exception & E ){
    QMessageBox::critical(this,"RAdmin: Error while partition opening"
        ,QString("Error: ")+QString(E.what()));
  }catch( ... ){
    QMessageBox::critical(this,"RAdmin: Error while partition opening"
        ,"Unknown error");
  }
  UpdateMenu();
}


void  sAdminWindow::Close ()
{
  try{
    ReopenMode = OpenMode;
    if( (OpenMode==cPartition) || (OpenMode==cPartitionAndRelated) ){
      try{
        psDB_SQL_Net                DB_SQL_Net = dynamic_cast<psDB_SQL_Net>(Net);
        ASSERT(DB_SQL_Net);
        std::map<sDB_UUID,psDatabase>::iterator
                                    D = DB_SQL_Net->Databases
                                                 .find(DB_SQL_Net->RootDB_UUID);
        if( D != DB_SQL_Net->Databases.end() ){
          ReopenURL = D->second->URL(true);
          ReopenURLOnSuccess = ReopenURL;
        }
      }catch(...){
        ReopenMode = cNothingIsOpen;
      }
    }
    ReopenModeOnSuccess = ReopenMode;
    DeleteNavigator();
    Net->Close();
    setWindowTitle(ToQString(Title));
    OpenMode = cNothingIsOpen;
  }catch( rxNetIO E ){
    QMessageBox::critical(this,"RANet closing error"
        ,ToQString(E.MsgText()));
  }
  NameInfo->clear();
  ErrorStatus->clear();
  UpdateMenu();
}


void  sAdminWindow::Reopen (bool   on_success)
{
  DeleteConsole();
  if( ReopenMode == cJobScript ){
    qApp->exit(on_success ? 0 : 1);
  }else if( on_success ){
    if( ReopenModeOnSuccess == cPartitionAndRelated ){
      OpenRANet(ReopenURLOnSuccess);
    }else if( ReopenModeOnSuccess == cPartition ){
      OpenPartition(ReopenURLOnSuccess);
    }
  }else{
    if( ReopenMode == cPartitionAndRelated ){
      OpenRANet(ReopenURL);
    }else if( ReopenMode == cPartition ){
      OpenPartition(ReopenURL);
    }
  }
}


void  sAdminWindow::ShowRA ()
{
  QLabel *                  LabelRA = new QLabel(this);
  LabelRA->setObjectName(QString::fromUtf8("LabelRA"));
  LabelRA->setPixmap(QPixmap(":/RAGUI/images/aboutra.png"));
  LabelRA->setText(QString());
  LabelRA->setAlignment(Qt::AlignHCenter|Qt::AlignCenter);
  setCentralWidget(LabelRA);
}


void  sAdminWindow::CreateNavigator ()
{
  Evolution = new sNet2TreeEvolution(sAttribute(sAttrName(),Net->Root()));
  Navigator = new sNavigator(Evolution ,this);
  setCentralWidget(Navigator);

  connect(Evolution,SIGNAL(Changed()),Navigator,SLOT(OnNetChange()));
  Evolution->Initiate();

  connect(UpdatingTimer, SIGNAL(timeout()), Evolution, SLOT(Update()));
  UpdatingTimer->start(sNavigator::UpdatingPeriod);
}


void  sAdminWindow::DeleteNavigator ()
{
  disconnect(UpdatingTimer,0,0,0);
  if( Evolution )
    disconnect(Evolution,0,0,0);
  ShowRA();//deletes Navigator, if it was set as the central widget!
  if( Evolution ){ delete Evolution;  Evolution=NULL; };
}


void  sAdminWindow::ExecuteInConsole (const QString &  script
                                     ,bool     autoclose_on_success
                                     ,bool     autoclose_on_abend
                                     ,int      autoclose_delay_on_success
                                     ,int      autoclose_delay_on_abend
                                     )
{
  menuBar()->setEnabled(false);
  Console = new sExecutionConsole(this);
  connect(Console,SIGNAL(Done(bool)),this,SLOT(Reopen(bool)));
  setCentralWidget(Console);
  Console->Execute(script,autoclose_on_success,autoclose_on_abend
                         ,autoclose_delay_on_success,autoclose_delay_on_abend);
}


void  sAdminWindow::DeleteConsole ()
{
  ShowRA();//deletes Console, if it was set as the central widget!
  menuBar()->setEnabled(true);
}


void  sAdminWindow::ShowHelpContents ()
{
  sHelper::ShowKeyId("RAdmin");
}


void  sAdminWindow::UpdateMenu ()
{
  DEBUG_LOG(NetOpen,sString("sAdminWindow::UpdateMenu: OpenMode")<<OpenMode);
  UpdateMenuFile();
  FormMenuDatabases();
  ActnCureRelatedPartitionsTable->setEnabled( OpenMode == cPartition );
  ActnReplicatePartition->setEnabled( OpenMode == cPartition );
  ActnAbsorbAllRelatedPartitions->setEnabled( OpenMode == cPartitionAndRelated);
  UpdateMenuInstall();
}


void  sAdminWindow::UpdateMenuFile ()
{
  MenuFile->setTitle( OpenMode == cNothingIsOpen
                    ? tr("File")
                    : OpenMode == cPartition
                    ? tr("Partition")
                    : tr("RANet")
                    );
  ActnNew->setEnabled( OpenMode == cNothingIsOpen );
  ActnOpen->setEnabled( OpenMode == cNothingIsOpen );
  ActnOpenPartition->setEnabled( OpenMode == cNothingIsOpen );
  ActnProperties->setEnabled( OpenMode != cNothingIsOpen );
  ActnClose->setEnabled( OpenMode != cNothingIsOpen );
}


void  sAdminWindow::UpdateMenuInstall ()
{
  try{
    bool                        Virgin = false;
//    bool                        UpgradeTo_1_2 = false;
    if( Net->IsOpen() ){
      sPtr<sFolderNode>           Root( Net->Root() );
      if( Root.IsCorrect() )
        Virgin = ! Root->HasAttributes();
    }
    ActnInstallSimple->setEnabled( Virgin );
    ActnAutocomplete->setEnabled( Net->IsOpen() && ! Virgin );
    ActnCreateStyles->setEnabled( Workstation().IsCorrect() );
//    ActnUpgrade_1_0_to_1_1->setEnabled( Net->IsOpen()
//        && ! Virgin
//        && sPtr<sFolderNode>(Net->Root()>>"Program").IsCorrect() );
//    if( Net->IsOpen() && ! ActnUpgrade_1_0_to_1_1->isEnabled() ){
//      psDB_SQL_Net                DB_SQL_Net = dynamic_cast<psDB_SQL_Net>(Net);
//      ASSERT(DB_SQL_Net);
//      for( std::map<sDB_UUID,psDatabase>
//                            ::const_iterator D = DB_SQL_Net->Databases.begin()
//                                           ; D != DB_SQL_Net->Databases.end()
//                                           ; ++D ){
//        psDatabase           DB(D->second);
//        if( ! DB->Version.IsValid() || (DB->Version < sVersion(1,2)) ){
//          UpgradeTo_1_2 = true;
//          break;
//        }
//      }
//    }
//    ActnUpgrade_1_1_to_1_2->setEnabled( UpgradeTo_1_2 );
  }catch(...){
  }
}


void  sAdminWindow::FormMenuDatabases ()
{
  ActnAddDatabaseLink->setEnabled( OpenMode != cNothingIsOpen );
  MenuDatabases->clear();
  MenuDatabases->addAction(ActnAddDatabaseLink);
  MenuDatabases->addSeparator();
  try{
    psDB_SQL_Net                DB_SQL_Net = dynamic_cast<psDB_SQL_Net>(Net);
    ASSERT(DB_SQL_Net);
    for( std::map<sDB_UUID,psDatabase>
                          ::const_iterator D = DB_SQL_Net->Databases.begin()
                                         ; D != DB_SQL_Net->Databases.end()
                                         ; ++D ){
      psDatabase                Partition = dynamic_cast<psDatabase>(D->second);
      if( ! (Partition->Id == DB_SQL_Net->RootDB_UUID) ){
        if( OpenMode == cPartitionAndRelated ){
          psActionSignallingText     Actn = new sActionSignallingText(this);
          Actn->setText(ToQString(Partition->URL()));
          Actn->SetDataString(ToQString(Partition->Id.ToString()));
          MenuDatabases->addAction(Actn);
          connect(Actn,SIGNAL(DataStringSignal(QString))
                 ,this,SLOT(ShowPartitionProperties(QString)));
        }else if( OpenMode == cPartition ){
          QMenu *         MenuPart
                          = MenuDatabases->addMenu(ToQString(Partition->URL()));
          psActionSignallingText     ActnEdit = new sActionSignallingText(this);
          ActnEdit->setText(tr("Edit link"));
          ActnEdit->SetDataString(ToQString(Partition->Id.ToString()));
          MenuPart->addAction(ActnEdit);
          connect(ActnEdit,SIGNAL(DataStringSignal(QString))
                 ,this    ,SLOT(EditDatabase(QString)));
          psActionSignallingText     ActnDel = new sActionSignallingText(this);
          ActnDel->setText(tr("Remove link"));
          ActnDel->SetDataString(ToQString(Partition->Id.ToString()));
          MenuPart->addAction(ActnDel);
          connect(ActnDel,SIGNAL(DataStringSignal(QString))
                 ,this   ,SLOT(DeleteDatabase(QString)));
        }
      }
    }
  }catch(...){
    QMessageBox::critical(this,"RAdmin","ERROR");
  }
//  qDebug() << "QSqlDatabase::connectionNames() = "
//           << QSqlDatabase::connectionNames();
}


void  sAdminWindow::ShowPartitionProperties ()
{
  try{
    psDB_SQL_Net                DB_SQL_Net = dynamic_cast<psDB_SQL_Net>(Net);
    ASSERT(DB_SQL_Net);
    ShowPartitionProperties(ToQString( DB_SQL_Net->RootDB_UUID.ToString() ));
  }catch(...){
    QMessageBox::critical(this,"RAdmin","ERROR");
  }
}


void  sAdminWindow::ShowPartitionProperties (QString  part_id)
{
  try{
    psDB_SQL_Net                DB_SQL_Net = dynamic_cast<psDB_SQL_Net>(Net);
    ASSERT(DB_SQL_Net);
    sDB_UUID                    PartId(FromQString(part_id));
    psDatabase                  Partition(DB_SQL_Net->Databases[PartId]);
    QMessageBox::information(this,tr("RAdmin: partition information")
        ,QString("Partition id: ") + part_id
        +"\nVersion: " + ToQString( Partition->Version.Text() )
        +"\nURL: '"+ ToQString( Partition->URL(true) ) +"'\n"
        +"\ntype=" + ToQString( Partition->TypeName[Partition->Type] )
        +"\nlogin=" + ToQString( Partition->Login )
        +"\npassword=" + ToQString( Partition->Password )
        +"\nhost=" + ToQString( Partition->Host )
        +"\nport=" + QString().setNum( Partition->Port )
        +"\ndatabase=" + ToQString( Partition->DBName )
        +"\nschema=" + ToQString( Partition->SchemaName )
        +"\nprefix=" + ToQString( Partition->TableNamePrefix )
        +"\noptions=" + ToQString( Partition->ConnectOptions )
        );
  }catch(...){
    QMessageBox::critical(this,"RAdmin","ERROR");
  }
}


void  sAdminWindow::DeleteDatabase (QString  part_id)
{
  try{
    psDB_SQL_Net                DB_SQL_Net = dynamic_cast<psDB_SQL_Net>(Net);
    ASSERT(DB_SQL_Net);
    sDB_UUID                    PartId(FromQString(part_id));
    psDatabase                  Partition(DB_SQL_Net->Databases[PartId]);
    if( QMessageBox::question(this,tr("RAdmin: remove partition link")
               ,tr("You are going to remove reference to:\n\n   %1\n   %2\n\n"
                   "Are you damn sure?!!")
               .arg( part_id )
               .arg( ToQString( Partition->URL() ))
               ) == QMessageBox::Yes ){
      try{
        DB_SQL_Net->RemovePartition(Partition);
      }catch( rxNetIO E ){
        QMessageBox::critical(this,"RAdmin: ERROR while removing link"
            ,ToQString(E.MsgText())
            );
      }
    }
  }catch(...){
    QMessageBox::critical(this,"RAdmin","ERROR");
  }
}


void  sAdminWindow::EditDatabase (QString  part_id)
{
  try{
    psDB_SQL_Net                DB_SQL_Net = dynamic_cast<psDB_SQL_Net>(Net);
    ASSERT(DB_SQL_Net);
    sDB_UUID                    PartId(FromQString(part_id));
    psDatabase                  Partition(DB_SQL_Net->Databases[PartId]);
    sDB_Link                    NewLink(*Partition);
    sEditDBLinkDialog *         Dlg = new sEditDBLinkDialog(&NewLink
                                            ,tr("Edit link to partition"));
    Dlg->setAttribute(Qt::WA_DeleteOnClose);
    if( ( Dlg->exec() == QDialog::Accepted )
        && ( QMessageBox::question(this,tr("Edit partition link?")
                             ,tr("This makes sence when the partition"
                                 "\n\n%1"
                                 "\n\nhas been moved to (or replicated into)"
                                 "\n\n%2"
                                 "\n\nAre you sure?")
                             .arg( ToQString( Partition->URL() ))
                             .arg( ToQString( NewLink.URL() ) )
                             ) == QMessageBox::Yes ) ){
      try{
        DB_SQL_Net->EditLink(NewLink);
        *Partition = NewLink;
      }catch( rxNetIO E ){
        QMessageBox::critical(this,"RAdmin: ERROR while editing link"
            ,ToQString(E.MsgText())
            );
      }
    }
  }catch(...){
    QMessageBox::critical(this,"RAdmin","ERROR");
  }
}


void  sAdminWindow::AddDatabase ()
{
  if( (OpenMode == cPartitionAndRelated)
      || ( QMessageBox::question(this,tr("Add related partition?")
                         ,tr("You should better close the partition, "
                             "then reopen it as a root of RANet, unless you "
                             "need to restore an occasionly deleted link.\n\n"
                             "If you proceed now, the link will be added, but "
                             "the corresponding attribute won't appear in the "
                             "'Related partitions' folder.\n\n"
                             "Understood?"
                             )
                         ) == QMessageBox::Yes ) ){
    psDB_SQL_Net                DB_SQL_Net = dynamic_cast<psDB_SQL_Net>(Net);
    ASSERT(DB_SQL_Net);
    sDB_Link                    Link;
    sEditDBLinkDialog *         Dlg = new sEditDBLinkDialog(&Link
                                            ,tr("Add related partition"));
    Dlg->setAttribute(Qt::WA_DeleteOnClose);
    if( Dlg->exec() == QDialog::Accepted ){
      try{
        DB_SQL_Net->AddDBLink(Link);
        if( OpenMode == cPartitionAndRelated ){
          sPtr<sFolderNode>            RelatedParts;
          RelatedParts = (Net->Root() >> "Related databases");
          if( ! RelatedParts.IsCorrect() )
            RelatedParts = (Net->Root() >= "Related partitions");
          RelatedParts << sAttribute(sAttrName(Link.Id.ToString())
                               ,sDB_SQL_Net::sNodeUUID::ForTheRoot(Net
                                                                  ,Link.Id
                                                                  ,Link.Version
                                                                  ).Replica()
                               ,sAttribute::cNewborn);
        }
      }catch( rxNetIO E ){
        QMessageBox::critical(this,"RAdmin: ERROR while addind a related DB"
            ,ToQString(E.SorryText())
            );
      }
    }
  }
}

void sAdminWindow::DoDummyJobs()
{
  Close();
  ExecuteInConsole("DUMMY | 200 | 400 | 300 || DUMMY | 4500 | 2000\n"
                   "DUMMY || DUMMY || DUMMY || DUMMY || DUMMY || "
                   " DUMMY|1000 ||DUMMY | 400 | 300 |200|100|1000\n"
                  );
}


void sAdminWindow::ShowSQLDrivers ()
{
  Close();
  ExecuteInConsole("SQLDL");
}


void  sAdminWindow::OpenNew ()
{
  try{
    sDB_Info                    PartInfo;
    sEditDBInfoDialog *         Dlg = new sEditDBInfoDialog(&PartInfo
                                            ,tr("Create new partition"));
    Dlg->setAttribute(Qt::WA_DeleteOnClose);
    if( Dlg->exec() == QDialog::Accepted ){
      Close();
      ReopenMode = cNothingIsOpen;
      ReopenURLOnSuccess = PartInfo.URL(true);
      ReopenModeOnSuccess = cPartitionAndRelated;
      ExecuteInConsole( QString("CN|") + ToQString(PartInfo.URL(true))
                      , true , false , 0 );
    }
  }catch( rxException E ){
    QMessageBox::critical(this,"New partition creation error"
        ,ToQString(E.SorryText()));
  }
}


void  sAdminWindow::RunJobScript ()
{
  if( JobScriptEditor->exec() == QDialog::Accepted ){
    Close();
    ExecuteInConsole( JobScriptEditor->Text() );
  }
}


void  sAdminWindow::RunJobScript (QString  script)
{
  ExecuteInConsole(script,true,true);
}


void  sAdminWindow::RunJobScriptFromFile (QString  file_path)
{
  QString                     Text;
  QFile                       File(file_path);
  if( ! File.open(QFile::ReadOnly) ){
    QMessageBox::critical(this,"RAdmin"
        ,tr("Cannot open the job script file\n'%1'"
            "\nfor reading. RAdmin will close.")
        .arg(file_path));
    qApp->exit(1);
  }else{
    ExecuteInConsole( QString::fromUtf8( File.readAll() ) , true , true );
  }
}


void sAdminWindow::CureRelatedPartitionsTable()
{
  try{
    psDB_SQL_Net                DB_SQL_Net = dynamic_cast<psDB_SQL_Net>(Net);
    ASSERT(DB_SQL_Net);
    sDB_Info                    Source(*(DB_SQL_Net
                                         ->Databases[DB_SQL_Net->RootDB_UUID]));
    Close();
    ExecuteInConsole( QString("CRPT|") + ToQString(Source.URL(true)) );
  }catch( rxException E ){
    QMessageBox::critical(this,"RAdmin: ERROR",ToQString(E.SorryText()));
  }
}


void  sAdminWindow::ReplicatePartition ()
{
  try{
    psDB_SQL_Net                DB_SQL_Net = dynamic_cast<psDB_SQL_Net>(Net);
    ASSERT(DB_SQL_Net);
    sDB_Info                    Source(*(DB_SQL_Net
                                         ->Databases[DB_SQL_Net->RootDB_UUID]));
    sDB_Info                    Destination(Source);
    Destination.Password = "";
    sEditDBInfoDialog *         Dlg = new sEditDBInfoDialog(&Destination
                                            ,tr("Destination partition"));
    Dlg->setAttribute(Qt::WA_DeleteOnClose);
    if( Dlg->exec() == QDialog::Accepted ){
      Close();
      ExecuteInConsole( QString("RP|")
                      + ToQString(Source.URL(true)) + "|"
                      + ToQString(Destination.URL(true))
                      );
    }
  }catch( rxException E ){
    QMessageBox::critical(this,"RAdmin: ERROR",ToQString(E.SorryText()));
  }
}


void  sAdminWindow::AbsorbAllRelatedPartitions ()
{
  if( QMessageBox::question(this,tr("RAdmin: Absorb all related partitions")
             ,tr("You are going to merge all related partitions into this one."
                 "\n\nAre you sure?")
                           ) != QMessageBox::Yes ){
    return;
  }
  if( QMessageBox::question(this,tr("RAdmin: Absorb all related partitions")
      ,tr("OK. You've got another chance to think about it."
          "\n\nThis partition (the one that has been opened as a root of RANet)"
          " will suck into itself not only all those partitions that have been"
          " added to it as \"related\", but also related to related, and so on."
          "\n\nAll internal part-to-part links inside this enlarged partition"
          " will be corrected accordingly. "
          "\n\nIf the whole RANet lies inside this enlarged partion, then there"
          " will be no problem. If not, you might have to cure some links."
          "\n\nAll the formely related patitions will remain untouched, but"
          " should not be used anymore - neither directly, nor indirectly."
          " You'd better remove them."
          "\n\nDo you surely want to proceed?!!")
                           ) != QMessageBox::Yes ){
    return;
  }
  try{
    psDB_SQL_Net                DB_SQL_Net = dynamic_cast<psDB_SQL_Net>(Net);
    ASSERT(DB_SQL_Net);
    sDB_Info                    Source(*(DB_SQL_Net
                                         ->Databases[DB_SQL_Net->RootDB_UUID]));
    Close();
    ExecuteInConsole( QString("AARP|") + ToQString(Source.URL(true)) );
  }catch( rxException E ){
    QMessageBox::critical(this,"RAdmin: ERROR while absorbing"
        ,ToQString(E.SorryText())
        );
  }
}


void  sAdminWindow::InstallSimple ()
{
  Autocomplete();
}


void  sAdminWindow::Autocomplete ()
{
  try{
    Net->Root() |= GlobalAttr::Data;
    Net->Root() |= WorkstationAttrName;
    AutocompleteData();
    AutocompleteWorkstation();

//  должна [до]создавать в корне;

    if( (Net->Root()>>GlobalAttr::Desktop).IsNULL() ){
      Net->Root() |= GlobalAttr::Desktop;
    }
    if( (Net->Root()>>GlobalAttr::FontSize).IsNULL() ){
      Net->Root() << attr( GlobalAttr::FontSize , font().pointSize() );
    }
    if( (Net->Root()>>GlobalAttr::Language).IsNULL() ){
      Net->Root() << attr( GlobalAttr::Language , "" );
    }
    if( (Net->Root()>>GlobalAttr::CustomColors).IsNULL() ){
      Net->Root() << attr(GlobalAttr::CustomColors
                         ,sNodePtr(new sColorSetNode));
    }
    if( (Net->Root()>>GlobalAttr::CSVExportSettings).IsNULL() ){
      Net->Root() |= GlobalAttr::CSVExportSettings;
    }
    if( (Net->Root()>>GlobalAttr::CSVImportSettings).IsNULL() ){
      Net->Root() |= GlobalAttr::CSVImportSettings;
    }
  }catch(...){}
}


void  sAdminWindow::AutocompleteData ()
{
//  должна создавать только в Data;

  sPtr<sFolderNode>  UnitSystem      (Net->Root()>>GlobalAttr::UnitSystem);
  sPtr<sFolderNode>  RecordNumbering (Net->Root()>>GlobalAttr::RecordNumbering);
  sPtr<sFolderNode>  RecordHistory   (Net->Root()>>GlobalAttr::RecordHistory);
  sPtr<sFolderNode>  RecordsToBrowse (Net->Root()>>GlobalAttr::RecordsToBrowse);
  sPtr<sFolderNode>  Samples         (Net->Root()>>GlobalAttr::Samples);
  sPtr<sFolderNode>  Topics          (Net->Root()>>GlobalAttr::Topics);
  sPtr<sFolderNode>  SOSImportSettings  (Net->Root()>>GlobalAttr
                                                           ::SOSImportSettings);
  try{
    if( ! Data().IsNULL() ){
      if( UnitSystem.IsNULL() ){
        UnitSystem = (Data() >= GlobalAttr::UnitSystem);
        Net->Root() << attr( GlobalAttr::UnitSystem , UnitSystem );
      }
      if( RecordNumbering.IsNULL() ){
        RecordNumbering = (Data() >= GlobalAttr::RecordNumbering);
        Net->Root() << attr( GlobalAttr::RecordNumbering , RecordNumbering );
      }
      if( RecordHistory.IsNULL() ){
        if( RecordsToBrowse.IsNULL() ){
          RecordHistory = (Data() >= GlobalAttr::RecordHistory);
          Net->Root() << attr( GlobalAttr::RecordHistory , RecordHistory );
        }else{
          RecordHistory = RecordsToBrowse;
        }
      }
      if( RecordsToBrowse.IsNULL() ){
        RecordsToBrowse = RecordHistory;
        Net->Root() << attr( GlobalAttr::RecordsToBrowse , RecordsToBrowse );
      }
      if( Samples.IsNULL() ){
        Samples = (Data() >= GlobalAttr::Samples);
        Net->Root() << attr( GlobalAttr::Samples , Samples );
      }
      if( Topics.IsNULL() ){
        Topics = (Data() >= GlobalAttr::Topics);
        Net->Root() << attr( GlobalAttr::Topics , Topics );
      }
      if( SOSImportSettings.IsNULL() ){
        SOSImportSettings = (Data() >= GlobalAttr::SOSImportSettings);
        Net->Root() << attr( GlobalAttr::SOSImportSettings , SOSImportSettings);
      }
    }
  }catch(...){}
}


void  sAdminWindow::AutocompleteWorkstation ()
{
//  должна создавать только в Workstation;

  sPtr<sFolderNode>  Plugins         (Net->Root()>>GlobalAttr::Plugins);
  sPtr<sFolderNode>  Devices         (Net->Root()>>GlobalAttr::Devices);
  sPtr<sTextNode>    FontFamily      (Net->Root()>>GlobalAttr::FontFamily);
  sPtr<sColorSetNode>LineColorSet    (Net->Root()>>GlobalAttr::LineColorSet);
  try{
    if( ! Workstation().IsNULL() ){
      if( Plugins.IsNULL() ){
        Plugins = (Workstation() >= GlobalAttr::Plugins);
        Net->Root() << attr( GlobalAttr::Plugins , Plugins );
      }
      if( Devices.IsNULL() ){
        Devices = (Workstation() >= GlobalAttr::Devices);
        Net->Root() << attr( GlobalAttr::Devices , Devices );
      }
      if( FontFamily.IsNULL() ){
        FontFamily = (Workstation() >= attr(GlobalAttr::FontFamily
                                    ,FromQString(font().family()) ) );
        Net->Root() << attr( GlobalAttr::FontFamily , FontFamily );
      }
      if( LineColorSet.IsNULL() ){
        LineColorSet = (Workstation()
                        >= attr(GlobalAttr::LineColorSet
                               ,sNodePtr(new sColorSetNode(sColorSet::Basic()))
                       ));
        Net->Root() << attr( GlobalAttr::LineColorSet , LineColorSet );
      }
    }
  }catch(...){}
}


void  sAdminWindow::CreateStyles ()
{
  try{
    (Workstation() >= GlobalAttr::Styles)
        << attr("Shapes",CSS::Shapes)
        << attr("NormalStyle",CSS::NormalStyle)
        << attr("NightStyle",CSS::NightStyle);
    Net->Root() << attr(GlobalAttr::Styles,Workstation() >> GlobalAttr::Styles);
  }catch(...){}
}


//void  sAdminWindow::Upgrade_1_0_to_1_1 ()
//{
//  sPtr<sFolderNode>  RecordHistory   (Net->Root()>>GlobalAttr::RecordHistory);
//  sPtr<sFolderNode>  Desktop         (Net->Root()>>GlobalAttr::Desktop);
//  sPtr<sIntegerNode> FontSize        (Net->Root()>>GlobalAttr::FontSize);
//  sPtr<sTextNode>    Language        (Net->Root()>>GlobalAttr::Language);
//  try{
//    if( Workstation().IsNULL() ){
//      sPtr<sFolderNode>  Program         (Net->Root()>>"Program");
//      if( ! Program.IsNULL() ){
//        Net->Root() << attr( WorkstationAttrName , Program );
//        Net->Root().DelAttr(sAttrName("Program"));
//      }
//    }
//    if( Desktop.IsCorrect() ){
//      if( FontSize.IsCorrect() ){
//        Net->Root() << attr( GlobalAttr::FontSize , FontSize->Value() );
//        Desktop.DelAttr(GlobalAttr::FontSize);
//      }
//      if( Language.IsCorrect() ){
//        Net->Root() << attr( GlobalAttr::Language , Language->Value() );
//        Desktop.DelAttr(GlobalAttr::Language);
//      }
//    }

//    //clean trash from older versions
//    Net->Root().DelAttr(sAttrName("To Do List"));
//    Net->Root().DelAttr(sAttrName("Configuration"));
//    Net->Root().DelAttr(sAttrName("Font"));
//    Net->Root().DelAttr(sAttrName("Font Name"));
//    if( ! Workstation().IsNULL() ){
//      Workstation().DelAttr(sAttrName("Display"));
//    }
//    if( ! RecordHistory.IsNULL() ){
//      RecordHistory.DelAttr(sAttrName("LastRecordDate"));
//      RecordHistory.DelAttr(sAttrName("LastRecordNo"));
//    }
//    Autocomplete();
//  }catch(...){}
//}


//void  sAdminWindow::Upgrade_1_1_to_1_2 ()
//{
//  try{
//    // Add version column
//    {
//      bool                        OK = true;
//      psDB_SQL_Net                DB_SQL_Net = dynamic_cast<psDB_SQL_Net>(Net);
//      ASSERT(DB_SQL_Net);
//      for( std::map<sDB_UUID,psDatabase>
//                            ::const_iterator D = DB_SQL_Net->Databases.begin()
//                                           ; D != DB_SQL_Net->Databases.end()
//                                           ; ++D ){
//        psDatabase                    DB( D->second );
//        if( ! DB->Version.IsValid() ){
//          sDBConnection                 DBConnection( DB->Connection() );
//          try{
//            sSQLQuery                   SQL(DBConnection);
//            DBConnection.transaction();

//            SQL.exec(QString("ALTER TABLE databases ADD COLUMN version TEXT"));
//            SQL.exec(QString("UPDATE databases SET version=\"1.2\"")
//                    +"WHERE nickname=\""+ToQString(sDB_Link::MyselfNickname)+"\"");
//            DBConnection.commit();
//            DB->Version = sVersion(1,2);
//          }catch( rxSQL E ){
//            OK = false;
//            try{ DBConnection.rollback(); }catch(...){};
//            QMessageBox::critical(this,"RAdmin: ERROR while upgrading to v.1.2"
//                ,QString("SQL error while addding 'version' to database\n\n[")
//                +ToQString(DB->Nickname)+":]"+ToQString(DB->DBName)+"\n\n"
//                +ToQString(E.Error));
//          }
//        }
//      }
//      if( OK ){
//        QMessageBox::information(this,"RAdmin: upgrading to v.1.2"
//            ,"All databases have Version column now.");
//      }
//    }
//    //sRAMainWindow geometry and state move into RANet knowledge base
//    {
//      sPtr<sFolderNode>     Desktop(Net->Root()>>GlobalAttr::Desktop);
//      QSettings             Settings(QSettings::UserScope,"BursianSoft", "RA");
//      if( Desktop.IsCorrect()
//          && sPtr<sByteArrayNode>(Desktop>>sRAMainWindow::DesktopStateAttrName)
//                                                                    .IsNULL()
//          && Settings.contains("sRAMainWindow/geometry")
//          && Settings.contains("sRAMainWindow/windowState") ){
//        QByteArray                  BA;
//        BA = Settings.value("sRAMainWindow/geometry").toByteArray();
//        Desktop << attr(sRAMainWindow::DesktopGeometryAttrName
//                       ,sNodePtr(new sByteArrayNode(BA.constData(),BA.size())));
//        BA = Settings.value("sRAMainWindow/windowState").toByteArray();
//        Desktop << attr(sRAMainWindow::DesktopStateAttrName
//                       ,sNodePtr(new sByteArrayNode(BA.constData(),BA.size())));
//      }
//    }
//  }catch( rxNet E ){
//      QMessageBox::critical(this,"RAdmin: ERROR while upgrading to v.1.2"
//          ,ToQString(E.SorryText())
//          );
//  }
//}


//void  sAdminWindow::Upgrade_1_2_to_1_3 ()
//{
//  bool                        OK = true;
//  try{
//    psDB_SQL_Net                DB_SQL_Net = dynamic_cast<psDB_SQL_Net>(Net);
//    ASSERT(DB_SQL_Net);
//    for( std::map<sDB_UUID,psDatabase>
//                          ::const_iterator D = DB_SQL_Net->Databases.begin()
//                                         ; D != DB_SQL_Net->Databases.end()
//                                         ; ++D ){
//      psDatabase                    DB( D->second );
//      sDBConnection                 DBConnection( DB->Connection() );
//      try{
//        sSQLQuery                   SQL(DBConnection);
//        DBConnection.transaction();

//        SQL.exec(QString("ALTER TABLE nodes ADD COLUMN owner TEXT"));
//        SQL.exec(QString("ALTER TABLE nodes ADD COLUMN permissions "));
//        SQL.exec(QString("ALTER TABLE attributes ADD COLUMN ")
//                +"version TEXT"
//                );
//                +"(id        "+DBIdFieldDecl+" UNIQUE NOT NULL"
//                +",nickname  TEXT NOT NULL"
//                +",comment   TEXT"
//                +",type      UNSIGNED TINYINT NOT NULL"
//                +",name      TEXT"
//                +",host      TEXT"
//                +",port      UNSIGNED SMALLINT"
//                +",login     TEXT"
//                +",password  TEXT"
//                +")"
//                );
//        if( OK )
//          OK = DBConnection.commit();
//      }catch( rxSQL E ){
//        try{ DBConnection.rollback(); }catch(...){};
//        QMessageBox::critical(this,"RAdmin: ERROR while upgrading to v.1.3"
//            ,QString("SQL error while processing database\n\n[")
//            +ToQString(DB->Nickname)+":]"+ToQString(DB->Name)+"\n\n"
//            +ToQString(E.Error));
//      }
//    }
//    if( OK ){

//    }
//  }catch( rxNet E ){
//      QMessageBox::critical(this,"RAdmin: ERROR while upgrading to v.1.3"
//          ,ToQString(E.Title)+"\n"+ToQString(E.Text)
//          );
//  }
//}

//------------------------------------------------------------------------------
} //namespace RA


