////////////////////////////////////////////////////////////////////////////////
/*! @file ExecutionConsole.cpp Execution watch frame for sAdminWindow.
- Part of RAdmin project.
- Uses  RAGUI - Research Assistant Graphical User Interface.
- Uses  RANet - Research Assistant Net Library.
- Uses  Qt v.5  - http://qt.io/
- Copyright(C) 2013-2020, Viktor E. Bursian, St.Petersburg, Russia.
                          VBursian AT gmail DOT com
*///////////////////////////////////////////////////////////////////////////////
#include "ExecutionConsole.h"
#include "StringsAndQStrings.h"
#include "Environment.h"
#include "attr.h"
#include "NetExceptions.h"
#include <QHBoxLayout>
#include <QVBoxLayout>
#include <QMessageBox>
#include "Log.h"
#include <QDebug>
namespace RA {
//------------------------------------------------------------------------------

const QString      sExecutionConsole::JobDoneSymbols(QString::fromUtf8(
                                   "➊➋➌➍➎➏➐➑➒➓⓫⓬⓭⓮⓯⓰⓱⓲⓳⓴"));
const QString      sExecutionConsole::JobToDoSymbols(QString::fromUtf8(
                                   "①②③④⑤⑥⑦⑧⑨⑩⑪⑫⑬⑭⑮⑯⑰⑱⑲⑳"));
const QString      sExecutionConsole::StageSymbols(QString::fromUtf8("■▥□"));

//-------------------------------------------------------- sExecutionConsole ---

sExecutionConsole::~sExecutionConsole()
{
  RANet::Log.Put(sLog::Debug,"NetTools","sExecutionConsole destructor");
//  QThread: Destroyed while thread is still running
}


sExecutionConsole::sExecutionConsole (QWidget * parent)
    :QFrame(parent)
    ,CurrentJobNum(0)
    ,TotalJobs(0)
    ,CurrentStage(0)
    ,TotalStages(0)
    ,Blinker(0)
    ,CloseDelayOnSuccess(5)
    ,CloseDelayOnAbEnd(20)
    ,CloseDelayCounter(0)
{
  ReportViewer = new QTextEdit(this);
  ReportViewer->setReadOnly(true);
  JobIndicator = new QLabel(this);
  JobIndicator->setToolTip(tr("job #"));
  QFont F(JobIndicator->font().family(),JobIndicator->font().pointSize()*2);
  JobIndicator->setFont(F);
  StageIndicator = new QLabel(this);
  StageIndicator->setToolTip(tr("stage"));
  StageIndicator->setFont(F);
  ProgressBar = new QProgressBar(this);
  ProgressBar->setRange(0,100);
  ProgressBar->setVisible(false);
  CancelButton = new QPushButton(tr("Cancel"),this);
  CloseButton = new QPushButton(tr("Close"),this);
  CloseButton->setVisible(false);
  AutoClosingLabel = new QLabel(this);
  AutoClosingText = tr("<small>closing in %1\"</small>");
  AutoClosingLabel->setToolTip(tr("click text window to have more time"
                                  "\n or button to close immediately"));
  AutoClosingLabel->setVisible(false);
  RefreshTimer = new QTimer(this);

  QHBoxLayout *               FooterLayout = new QHBoxLayout;
  QVBoxLayout *               Layout = new QVBoxLayout;
  setLayout(Layout);
  Layout->addWidget(ReportViewer);
  Layout->addLayout(FooterLayout);
  FooterLayout->addWidget(CancelButton,0,Qt::AlignLeft);
  FooterLayout->addWidget(JobIndicator,0,Qt::AlignLeft);
  FooterLayout->addWidget(StageIndicator,0,Qt::AlignLeft);
  FooterLayout->addWidget(ProgressBar,100);
  FooterLayout->addWidget(CloseButton,0,Qt::AlignRight);
  FooterLayout->addWidget(AutoClosingLabel,0,Qt::AlignRight);

  connect(CloseButton ,SIGNAL(clicked())
         ,this        ,SLOT(Complete()) );
  connect(RefreshTimer ,SIGNAL(timeout())
         ,this         ,SLOT(Refresh()) );
}


void  sExecutionConsole::Execute (const QString &  script
                                 ,bool             autoclose_on_success
                                 ,bool             autoclose_on_abend
                                 ,int              autoclose_delay_on_success
                                 ,int              autoclose_delay_on_abend
                                 )
{
  AutocloseOnSuccess = autoclose_on_success;
  AutocloseOnAbEnd = autoclose_on_abend;
  CloseDelayOnSuccess = autoclose_delay_on_success;
  CloseDelayOnAbEnd = autoclose_delay_on_abend;
  TheJobExecutor = new sAdminJob(script,this);
  connect(CancelButton   ,SIGNAL(clicked())
         ,TheJobExecutor ,SLOT(Abort()) );
  connect(TheJobExecutor ,SIGNAL(TotalJobs(int))
         ,this           ,SLOT(SetTotalJobs(int)) ,Qt::QueuedConnection );
  connect(TheJobExecutor ,SIGNAL(NewJob())
         ,this           ,SLOT(OnNewJob()) ,Qt::QueuedConnection );
  connect(TheJobExecutor ,SIGNAL(TotalStages(int))
         ,this           ,SLOT(SetTotalStages(int)) ,Qt::QueuedConnection );
  connect(TheJobExecutor ,SIGNAL(NewStage())
         ,this           ,SLOT(OnNewStage()) ,Qt::QueuedConnection );
  connect(TheJobExecutor ,SIGNAL(OutPar(QString))
         ,this           ,SLOT(OutPar(QString)) ,Qt::QueuedConnection );
  connect(TheJobExecutor ,SIGNAL(Out(QString))
         ,this           ,SLOT(Out(QString)) ,Qt::QueuedConnection );
  connect(TheJobExecutor ,SIGNAL(AllDone())
         ,this           ,SLOT(CompleteOnAllDone()) ,Qt::QueuedConnection );
  connect(TheJobExecutor ,SIGNAL(Aborted())
         ,this           ,SLOT(CompleteOnAbEnd()) ,Qt::QueuedConnection );
  connect(TheJobExecutor ,SIGNAL(Failed())
         ,this           ,SLOT(CompleteOnAbEnd()) ,Qt::QueuedConnection );
  TheJobExecutor->start(QThread::LowPriority);
  RefreshTimer->start(100);
}


void  sExecutionConsole::Refresh ()
{
  if( TheJobExecutor ){
    int                         P = TheJobExecutor->Progress();
    if( P >= 0 ){
      ProgressBar->setValue(P);
      ProgressBar->setVisible(true);
    }else{
      ProgressBar->setVisible(false);
    }
    Blinker = (Blinker+1) % 10;
    if( (CurrentJobNum > 0) && (CurrentJobNum <= JobDoneSymbols.length()) ){
      if( Blinker == 0 ){
        QString                     S( JobIndicator->text() );
        S[CurrentJobNum-1] = JobToDoSymbols[CurrentJobNum-1];
        JobIndicator->setText(S);
      }
      if( Blinker == 8 ){
        QString                     S( JobIndicator->text() );
        S[CurrentJobNum-1] = JobDoneSymbols[CurrentJobNum-1];
        JobIndicator->setText(S);
      }
    }
  }
}


void  sExecutionConsole::SetTotalJobs (int  num_of_jobs)
{
  TotalJobs = num_of_jobs;
  if( TotalJobs > 1 ){
    JobIndicator->setVisible(true);
  }else{
    JobIndicator->setVisible(false);
  }
  JobIndicator->setText(JobToDoSymbols.left(TotalJobs));
  int W =
      std::max(JobIndicator->fontMetrics().width(JobToDoSymbols.left(TotalJobs))
              ,JobIndicator->fontMetrics().width(JobDoneSymbols.left(TotalJobs))
              );
  JobIndicator->setMinimumWidth(W+10);
  JobIndicator->setMaximumWidth( JobIndicator->minimumWidth() );
  CurrentJobNum = -1; //increments to 0 by the next statement
  OnNewJob();
}


void  sExecutionConsole::OnNewJob ()
{
  QString                     S( JobIndicator->text() );
  ++CurrentJobNum;
  SetTotalStages(0);
  if( (CurrentJobNum > 1) && (CurrentJobNum <= JobDoneSymbols.length()+1) )
    S[CurrentJobNum-2] = JobDoneSymbols[CurrentJobNum-2];
  JobIndicator->setText(S);
  Refresh();
}


void  sExecutionConsole::SetTotalStages (int  num_of_stages)
{
  TotalStages = num_of_stages;
  if( TotalStages > 1 ){
    StageIndicator->setVisible(true);
  }else{
    StageIndicator->setVisible(false);
  }
  CurrentStage = -1; //increments to 0 by the next statement
  OnNewStage();
}


void  sExecutionConsole::OnNewStage ()
{
  if( TotalStages > 1 ){
    ++CurrentStage;
    if( CurrentStage > 0 ){
      StageIndicator->setText(
            QString(StageSymbols[0]).repeated(CurrentStage-1)
          + StageSymbols[1]
          + QString(StageSymbols[2]).repeated(TotalStages-CurrentStage)
          );
    }
  }
  Refresh();
}


void  sExecutionConsole::Out (QString  text)
{
  if( ReportViewer ){
    ReportViewer->insertPlainText(text);
    ReportViewer->moveCursor(QTextCursor::End);
    ReportViewer->repaint();
  }
}


void  sExecutionConsole::OutPar (QString  text)
{
  if( ReportViewer ){
    ReportViewer->append(text.prepend("<p>").append("</p>"));
    ReportViewer->moveCursor(QTextCursor::End);
    ReportViewer->repaint();
  }
}


void  sExecutionConsole::CompleteOnAllDone ()
{
  ExecutedSuccessfully = true;
  disconnect(RefreshTimer ,0,0,0 );
  RefreshTimer->stop();
  CancelButton->setVisible(false);
  JobIndicator->setVisible(false);
  StageIndicator->setVisible(false);
  ProgressBar->setVisible(false);
  CloseButton->setVisible(true);
  CloseButton->setFocus();
  if( AutocloseOnSuccess ){
    CloseDelayCounter = CloseDelayOnSuccess;
    DelayedComplete();
  }
}


void  sExecutionConsole::CompleteOnAbEnd ()
{
  ExecutedSuccessfully = false;
  CancelButton->setVisible(false);
  CloseButton->setVisible(true);
  CloseButton->setFocus();
  if( AutocloseOnAbEnd ){
    CloseDelayCounter = CloseDelayOnAbEnd;
    DelayedComplete();
  }
}


void sExecutionConsole::DelayedComplete ()
{
  if( ! CloseButton->hasFocus() ){
    AutoClosingLabel->setVisible(false);
  }else{
    if( CloseDelayCounter <= 0 ){
      Complete();
    }else{
      AutoClosingLabel->setText( AutoClosingText.arg(CloseDelayCounter) );
      AutoClosingLabel->setVisible(true);
      --CloseDelayCounter;
      QTimer::singleShot(1000,this,SLOT(DelayedComplete()));
    }
  }
}


void sExecutionConsole::Complete ()
{
  emit Done(ExecutedSuccessfully);
}

//------------------------------------------------------------------------------
} //namespace RA


