////////////////////////////////////////////////////////////////////////////////
/*! @file ScriptJob.cpp Background script execution.
- Part of RAdmin project.
- Uses  RANet - Research Assistant Net Library.
- Uses  Qt v.5  - http://qt.io/
- Copyright(C) 2020, Viktor E. Bursian, St.Petersburg, Russia.
                     VBursian AT gmail DOT com
*///////////////////////////////////////////////////////////////////////////////
#include "ScriptJob.h"
#include <QRegExp>
#include "Log.h"
namespace RA {
//------------------------------------------------------------------------------
//--------------------------------------------------------------- sScriptJob ---

sScriptJob::~sScriptJob ()
{
  RANet::Log.Put(sLog::Debug,"NetTools","sScriptJob destructor");
  Mutex.lock();
  if( isRunning() ){
    terminate();
    wait();
  }
  Mutex.unlock();
}


sScriptJob::sScriptJob (const QString &  script
                       ,QObject *        parent)
    :QThread(parent)
    ,AbortRequestedFlag(false)
    ,FailureHappendFlag(false)
    ,TheProgress(-1)
    ,TheProgressTarget(100)
{
  Jobs = QString(script)
      .replace(QRegExp("\\n"),"||")
      .replace(QRegExp("\\r"),"||")
      .replace(QRegExp("\\s+\\|\\|"),"||")
      .replace(QRegExp("\\|\\|\\s+"),"||")
      .split("||",QString::SkipEmptyParts)
      .replaceInStrings(QRegExp("^\\s+"),"")
      .replaceInStrings(QRegExp("\\s+$"),"")
      .replaceInStrings(QRegExp("\\s+\\|"),"|")
      .replaceInStrings(QRegExp("\\|\\s+"),"|")
      ;
}


int  sScriptJob::Progress ()
{
  QMutexLocker locker(&Mutex);
  return (100 * TheProgress) / TheProgressTarget;
}


void  sScriptJob::Abort ()
{
  QMutexLocker locker(&Mutex);
  AbortRequestedFlag = true;
}


void  sScriptJob::Failure (const QString &  message)
{
  emit OutPar(QString("<font color=red><b>%1</b></font>")
              .arg(message));
  {
    QMutexLocker locker(&Mutex);
    FailureHappendFlag = true;
  }
}


void  sScriptJob::ReportNewJobStart (QString  job_title)
{
  emit OutPar(QString("<font color=navy><h1>%1. %2</h1></font>")
              .arg(TheCurrentJobNum)
              .arg(job_title)
              );
}


void  sScriptJob::ReportNewStageStart (QString  stage_title)
{
  emit OutPar(QString("<h3>______ %1 ______</h3>")
              .arg(stage_title));
  emit NewStage();
}


void  sScriptJob::ReportJobConclusion (QString  conclusion)
{
  emit OutPar(QString("<font color=navy><b>____________ ")
                 +conclusion+"</b></font>");
}


bool  sScriptJob::AbortRequested ()
{
  QMutexLocker locker(&Mutex);
  return AbortRequestedFlag;
}


bool  sScriptJob::FailureHappend()
{
  QMutexLocker locker(&Mutex);
  return FailureHappendFlag;
}


void  sScriptJob::SetProgress (int  progress)
{
  QMutexLocker locker(&Mutex);
  TheProgress = progress;
}


void  sScriptJob::SetProgressTarget (int  progress_target)
{
  QMutexLocker locker(&Mutex);
  TheProgressTarget = progress_target;
}


void sScriptJob::FindAndExecute (const QString & job_name
                                ,QStringList &   params)
{
  Failure( QString("Unknown command in the script: '%1|%2'%3.")
           .arg(job_name)
           .arg(params.join("|"))
           .arg( job_name.isEmpty() ? " (empty command name)" : "" )
         );
}


void  sScriptJob::run ()
{
  emit OutPar("<font color=brown><h1>Script</h1></font>");
  for( int i = 0 ; i < Jobs.size() ; ++i ){
    emit OutPar(QString("<font color=brown><code>%1</code></font>")
                .arg(Jobs.at(i)));
  }
  emit TotalJobs(Jobs.size());
  for( int i = 0 ; i < Jobs.size() ; ++i ){
    TheCurrentJobNum = i+1;
    //emit OutPar("<hr>"); -- потом вставляет черту после каждого абзаца
    //emit OutPar("<br/>"); //it works
    emit NewJob();
    QStringList  Pars( Jobs.at(i).split("|") );
    if( Pars.size() > 0 ){
      QString      Job( Pars.front() );
      Pars.pop_front();
      FindAndExecute(Job,Pars);
    }
    if( FailureHappend() || AbortRequested() ){
      break;
    }
  }
  if( AbortRequested() ){
    emit OutPar(QString("<font color=red><big>"
                            "______________________ Execution ABORTED"
                            "</big></font>"));
    emit Aborted();
  }else if( FailureHappend() ){
    emit OutPar(QString("<font color=red><big>"
                            "______________________ Execution FAILED"
                            "</big></font>"));

    emit Failed();
  }else{
    emit OutPar(QString("<font color=green><big>"
                            "______________________ ALL DONE"
                            "</big></font>"));
    emit AllDone();
  }
}

//------------------------------------------------------------------------------
} //namespace RA
