////////////////////////////////////////////////////////////////////////////////
/*! @file BasicNet.cpp   Реализация сети RANet, опирающаяся на файловую систему
                         размещающая сеть в системе многих катологов и файлов
                         и подгружающая части по мере надобности.
- Part of RANet - Research Assistant Net Library (based on ANSI C++).
- Copyright(C) 2006-2010, Viktor E. Bursian, St.Petersburg, Russia.
                     Viktor.Bursian@mail.ioffe.ru
*///////////////////////////////////////////////////////////////////////////////
#include "BasicNet.h"
#include "Version.h"
#include "FolderNode.h"
#include "BasicNodes.h"
#include "Trace.h"
#include "attr.h"
namespace RA {
//------------------------------------------------------------------------------

const sString                 sBasicNet::cLastUIDFileExt(".RAsys");
const sString                 sBasicNet::cStreamFileExt(".RAbin");
const sString                 sBasicNet::cTextFileExt(".RAtxt");

//---------------------------------------------------------------- sBasicNet ---

sBasicNet::~sBasicNet ()
{
  Undo.Clear();
  Root.Strip();
  if( ! LoadedNodes.empty() ){
    for( tNodeIndex::const_iterator P=LoadedNodes.begin(); P!=LoadedNodes.end(); ++P){
      TRACE( P->first );
    };
  };
//  ASSERT( LoadedNodes.empty() );
};


sBasicNet::sBasicNet (sNodeLocation root_location)
    :RootLocation(root_location)
{
  psNode                      root;
  root=GetNode(&RootLocation);
  Root=sNodePtr(root);
  sPtr<sTextNode>             S(root->GetAttr(sAttributeName("LastUIDFName")));
  if( ! S.IsCorrect() )
    throw xNet();
  LastUIDFile.Assign(rsString(*S));
  try{
    LastUIDFile.Open(sObjectFile::Reading);
    LastUIDFile.READ_INTO(LastUsedNodeUID);
    LastUIDFile.Close();
  }catch(...){
    throw xNetInitiation();
  };
};


sBasicNet::sBasicNet (sString        owner
                     ,int            copy
                     ,sNodeLocation  root_location
                     ,sPath          lastUID_path
                     )
    :sNet(owner,copy)
    ,RootLocation(root_location)
{
  /*! ^todo{elegancy} appending extension to sPath; sPath->sString */
  lastUID_path = lastUID_path.PathStr() + cLastUIDFileExt;
  LastUIDFile.Assign(lastUID_path);

  LastUsedNodeUID=LastReservedNodeUID;
  LastUIDFile.Open(sObjectFile::Writing);
  LastUIDFile.WRITE_FROM(LastUsedNodeUID);
  LastUIDFile.Close();
  psNode                    N = new sFolderNode;
  N->Location=RootLocation;
  N->NumberOfHosts=1;
  LoadedNodes[N->Location.NodeUID]=N;
  Root=sNodePtr(N);
  Root<<attr("LastUIDFName",literal(lastUID_path.PathStr()));
};


tNodeUID  sBasicNet::GetNewNodeUID ()
{
  LastUIDFile.Open(sObjectFile::Updating);
  LastUIDFile.READ_INTO(LastUsedNodeUID);
  ++LastUsedNodeUID;
  LastUIDFile.SetWritingPosition(0);
  LastUIDFile.WRITE_FROM(LastUsedNodeUID);
  LastUIDFile.Close();
  return LastUsedNodeUID;
};


psNode  sBasicNet::GetNode (psNodeLocation L)
{
  psNode                      N;
  tNodeIndex::iterator        P = LoadedNodes.find(L->NodeUID);
  if( P != LoadedNodes.end() ){
    N=P->second;
  }else if( L->Type == sNodeLocation::InStreamFile ){
    sObjectFile                 File;
    psStorable                  O = NULL;
    int                         ver_major;
    int                         ver_minor;
    sString                     TypeIndicator;
    sString                     Ownr;
    try{
      File.Assign(L->Path+cStreamFileExt);
      File.Open(sObjectFile::Reading);
      File.READ_INTO(ver_major);
      File.READ_INTO(ver_minor);
      File.ReadString(TypeIndicator);
      File.ReadString(Ownr);
      File.READ_INTO(Copy);
      if( Owner.Empty() ){
        Owner=Ownr;
      }else{
        ASSERT( Ownr == Owner );
      };
      //! ^todo{NetIO} CurrentIOLocation PATCH
      sNode::CurrentIOLocation=L;
      O=File.GetObject();
      sNode::CurrentIOLocation=NULL;
      File.Close();
    }catch(...){
    };
    if( ! sNode::Type()->CanCast(O) )
      throw xNet();  //! ^todo{Net} something in place of xNet?  "not a node"
    N=dynamic_cast<psNode>(O);
    if( ! N ) //перебдёш
      throw xNet();  //! ^todo{Net}  something in place of xNet?  "not a node"
    LoadedNodes[N->Location.NodeUID]=N;
  }else{
    throw xProgramError("sBasicNet::GetNode wrong location type");
  };
  return N;
};


void  sBasicNet::ReleaseNode (tNodeUID UID)
{
//  psNode                      N;
  tNodeIndex::iterator        P = LoadedNodes.find(UID);
  if( P != LoadedNodes.end() ){
//    N=P->second;
    LoadedNodes.erase(P);
  }else{
    throw xProgramError("sBasicNet::ReleaseNode");
  };
};


void  sBasicNet::Allocate (rcsAttribute A ,psNode Host)
{
  sNodePtr                    AValue;
  AValue=A.Value();
  if( AValue.Stripped() ){
    throw xNotImplemented("sBasicNet::Allocate(rcsAttribute,psNode): this should not happen?");
  }else if( AValue.IsNULL() ){
    //nothing to do
  }else if( AValue->Location.NodeUID != NoID ){
    //already allocated
  }else if( AValue->NumberOfHosts == 0 ){
    //the attribute had not been set to Host and does not need allocation
  }else{
    /*/////////DEBUG
    tNodeUID    DEBUG_ID=AValue->Location.NodeUID;
    sLongString DEBUG_LP=AValue->Location.Path+AValue->Location.DefaultSuffix;
    bool        DEBUG_EDITED=AValue->Edited;
    int         DEBUG_NUMOFHOSTS=AValue->NumberOfHosts;
    int         DEBUG_NUMOFPTRS=AValue->NumberOfPointers;
    tNodeUID    DEBUG_Host_ID=Host->Location.NodeUID;
    sLongString DEBUG_Host_LP=Host->Location.Path+Host->Location.DefaultSuffix;
    bool        DEBUG_Host_EDITED=Host->Edited;
    int         DEBUG_Host_NUMOFHOSTS=Host->NumberOfHosts;
    int         DEBUG_Host_NUMOFPTRS=Host->NumberOfPointers;
    */
    AValue->Preallocate(Host,A.Name());
  };
};


void  sBasicNet::AllocateMandatory (psNode                Node
                                   ,sNodeLocation::eType  LType
                                   ,rcsString             LPath)
{
  if( Node->Location.NodeUID == NoID ){
    if( LType == sNodeLocation::InStreamFile ){
      /*/////////DEBUG
      tNodeUID    DEBUG_ID=Node->Location.NodeUID;
      sLongString DEBUG_LP=Node->Location.Path+Node->Location.DefaultSuffix;
      bool        DEBUG_EDITED=Node->Edited;
      int         DEBUG_NUMOFHOSTS=Node->NumberOfHosts;
      int         DEBUG_NUMOFPTRS=Node->NumberOfPointers;
      */
      Node->Location.NodeUID=sNet::Manager->GetNewNodeUID();
      LoadedNodes[Node->Location.NodeUID]=Node;
      Node->Location.Type=LType;
      Node->Location.Path=LPath;
      Node->Location.DefaultSuffix="";
      Node->Location.Offset=0;
      Node->SetEditedFlag();  /*! ^todo{NetIO} SetToBeSavedFlag() */
      Node->PreallocateAttributes();
      Node->Save();
      Node->LocationEdited=true;
    }else if( LType == sNodeLocation::InTextFile ){
      throw xNotImplemented("Allocation InTextFile");
    }else if( LType == sNodeLocation::InRegistry ){
      throw xNotImplemented("Allocation InRegistry");
    }else{
      throw xNotImplemented("Allocation");
    };
  }else{
    throw xNotImplemented("Node reallocation");
  };
};


void  sBasicNet::Allocate (psNode Node) //! ^todo{NetIO} obsolete allocation?
//можно попробовать вернуть этой процедуре основную роль
//знать Host не так уж и обязательно -- его можно поискать среди
//LoadedNodes с таким же Location.Path, затем спуститься и нарастить
//суффикс
{
  if( Node->Location.Type == sNodeLocation::NotAllocated ){
    sString NotAllocated_node;
    NotAllocated_node+=Node->Location.Path;
    (NotAllocated_node+="|")+=Node->Location.DefaultSuffix;
    ((NotAllocated_node+="[")<<Node->Location.NodeUID)+="]";
    NotAllocated_node+=Node->TrueType()->Name();
    TRACE( NotAllocated_node );
  };
//  ASSERT( Node->Location.Type != sNode::sLocation::NotAllocated );
  ASSERT( (Node->NumberOfHosts <= 1) || (Node->Location.NodeUID != NoID) );
  if( Node->Location.NodeUID != NoID ){
    Node->PreallocateAttributes();
  };
};


void  sBasicNet::PreallocateNode (psNode          Node
                                 ,sNodePtr        Host
                                 ,sAttributeName  AN   )
{
  /*/////////DEBUG
  tNodeUID    DEBUG_ID=Node->Location.NodeUID;
  sLongString DEBUG_LP=Node->Location.Path+Node->Location.DefaultSuffix;
  bool        DEBUG_EDITED=Node->Edited;
  int         DEBUG_NUMOFHOSTS=Node->NumberOfHosts;
  int         DEBUG_NUMOFPTRS=Node->NumberOfPointers;
  tNodeUID    DEBUG_Host_ID=Host->Location.NodeUID;
  sLongString DEBUG_Host_LP=Host->Location.Path+Host->Location.DefaultSuffix;
  bool        DEBUG_Host_EDITED=Host->Edited;
  int         DEBUG_Host_NUMOFHOSTS=Host->NumberOfHosts;
  int         DEBUG_Host_NUMOFPTRS=Host->NumberOfPointers;
  */
  ASSERT( Node->NumberOfHosts > 0 );
  if( Node->NumberOfHosts == 1 ){
    //form info for future allocation
    Node->Location.Type=Host->Location.Type;
    Node->Location.Path=Host->Location.Path;
    Node->Location.DefaultSuffix=Host->Location.DefaultSuffix;
    Node->Location.DefaultSuffix+="/";
    sString                   ANT(sPath::ValidateName(AN.Text()));
    if( ANT.Empty() )
      ANT = sPath::ValidateName(Node->Text(Plain,Laconic));
    if( ANT.Empty() )  ANT = "^";
    Node->Location.DefaultSuffix+=ANT;
    Node->Location.Offset=0;
    Node->SetEditedFlag();  /*! ^todo{NetIO} SetToBeSavedFlag() и/или LocationEdited=true*/
  }else{ //------ NumberOfHosts > 1
    if(    (Node->Location.Type != sNodeLocation::NotAllocated)
        || (Host->Location.Type != sNodeLocation::NotAllocated)   ){
      Node->Location.NodeUID=GetNewNodeUID();
      LoadedNodes[Node->Location.NodeUID]=Node;
      if( Node->Location.Type != sNodeLocation::NotAllocated ){
        Node->Location.Path+=Node->Location.DefaultSuffix;
      }else{
        Node->Location.Type=Host->Location.Type;
        Node->Location.Path=Host->Location.Path;
        Node->Location.Path+=Host->Location.DefaultSuffix;
        Node->Location.Path+="/";
        sString                   ANT(sPath::ValidateName(AN.Text()));
        if( ANT.Empty() )
          ANT = sPath::ValidateName(Node->Text(Plain,Laconic));
        if( ANT.Empty() )  ANT = "^";
        Node->Location.Path+=ANT;
      };
      //! ^todo{NetIO} make sure the Path is correct and such file does not exist
      Node->Location.Path=sPath::Validate(Node->Location.Path);
      sPath(Node->Location.Path).CreateDirForIt();
      if(   ( (Node->Location.Type == sNodeLocation::InStreamFile)
            && sPath(Node->Location.Path + cStreamFileExt).Exists()  )
          ||( (Node->Location.Type == sNodeLocation::InTextFile)
            && sPath(Node->Location.Path + cTextFileExt).Exists()    ) ){
        Node->Location.Path+="~";
        Node->Location.Path<<Node->Location.NodeUID;
        Node->Location.Path=sPath::Validate(Node->Location.Path);
      };
      Node->Location.DefaultSuffix="";
      Node->Location.Offset=0;
      Node->SetEditedFlag();  /*! ^todo{NetIO} SetToBeSavedFlag() */
      Node->PreallocateAttributes();
      Node->Save();//! ^todo{NetIO} возможно, лишнее спасение
      Node->LocationEdited=true;
    };
  };
};


bool  sBasicNet::SaveNode (psNode N)
{
  ASSERT( N->Location.Type != sNodeLocation::NotAllocated );
  ASSERT( N->Location.NodeUID != NoID );
  ASSERT( ! N->Location.Path.Empty() );
  if( N->Location.Type == sNodeLocation::InStreamFile ){
    sObjectFile                   File;
    File.Assign(N->Location.Path+cStreamFileExt);
    //! ^todo{NetIO} rename to <...>.~RAnet (recursively)
    //в сетевом варианте надо взять ресурс на файл и не отпускать
    //if( N->NumberOfHosts > 0 ){
    File.Open(sObjectFile::Writing);
    File.WRITE_FROM(RANet::Version.Major);
    File.WRITE_FROM(RANet::Version.Minor);
    File.WriteString(cStreamFileExt);
    File.WriteString(Owner);
    File.WRITE_FROM(Copy);
    File.PutObject(N);
    File.Close();
    //};
  }else if( N->Location.Type == sNodeLocation::InTextFile ){
    throw xNotImplemented("sBasicNet::SaveNode for TextFile");
  }else if( N->Location.Type == sNodeLocation::InRegistry ){
    throw xNotImplemented("sBasicNet::SaveNode for Registry");
  };
  return true;
};


bool  sBasicNet::GetReadingAccess (sNodePtr  NP)
{
  Undo.Push( new sUndoGetReadingAccess(NP) );
  return true;
};


void  sBasicNet::ReleaseReadingAccess (sNodePtr)
{
};


bool  sBasicNet::GetFullAccess (sNodePtr  NP)
{
  Undo.Push( new sUndoGetFullAccess(NP) );
  return true;
};


void  sBasicNet::ReleaseFullAccess (sNodePtr)
{
};


bool  sBasicNet::TooLongWaiting ()
{
  return false;
};

//------------------------------------------------------------------------------
}; //namespace RA
